!------------------------------------------------------------------------------
!
! MODULE: cart_topology
!
!
! DESCRIPTION: 
!>  This module provide a cartesien topology on the parrallel layout.
!! This virtual topology is created by the MPI procedures (and thus use
!! low-level optimisation based on the underlyinfg hardware). It
!! provides the different tools to create, to manipulate and to interface
!! it with the other topology and communicators.
!! The solver use some dimensionnal splitting and this module contains all the
!! method used to solve advection along the Y-axis. This is a parallel
!! implementation using MPI and the cartesien topology it provides.
!!
!!  Nowaday, the domain is only splitted along Y and Z axis. Therefore,
!! we only use a 2D cartesian topology.
!!  A "global" communicator is devoted to the (2D) cartesian structure.
!! Another communicator is added for each direction in order to deal
!! with all 1D communication (along Y or Z).
!! Be careful : the (Y,Z)-axis in the 3D mesh match to the (X,Y) axis on the 2D
!! mpi-topology.
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

module cart_topology

    use precision

    implicit none

    public

    ! ===== Public variables =====

    ! ----- Communicators -----
    !> Communicator associated with the cartesian topology
    integer     :: cart_comm        
    !> Communicators devoted to 1-dimensionnal subgrids (along Y and Z)
    integer     :: X_comm, Y_comm, Z_comm   
    !> Table of the preivious communicators (ie comm devoted to 1D subgrids)
    integer, dimension(3)   :: D_comm

    ! ----- Information about mesh subdivision and on the global grid -----
    !> number of processes in each direction
    integer, dimension(3)   :: nb_proc_dim
    !> information about min and max local indice on the current directory
    integer, dimension(3)   :: begin_proc, end_proc     
    !> space lengh of th domain
    real(WP), dimension(3)  :: length
    !> space step for scalar discretisation
    real(WP), dimension(3)  :: d_sc                     
    !> number of (sub)grid in each direction
    integer, dimension(3)   :: N!, N_proc         
    integer, dimension(3)   :: N_proc
    !> rank of current processus (in the cartesian communicator)
    integer                 :: myrank
    !> coordinate of the current processus
    integer, dimension(3)   :: coord
    !> YZ coordinate of the current processus
    integer, dimension(2)   :: coordYZ
    !> Periodic boundary conditions : logical array, equals true if periodic
    logical, dimension(3)   :: periods                      
    !> Computation are done by group of line. Here we define their size
    integer, dimension(3,2) :: group_size
    !> To concatenate position in order to create unique mpi message tag
    integer, dimension(3,2) :: tag_size
    !> To concatenate rank in order to create unique mpi message tag
    integer                 :: tag_rank


    ! ==== Public procedures ==== 
    !> Creation of the cartesian topology
    public      :: cart_create
    !> Initialise mesh information
    public      :: mesh_default
    public      :: mesh_init
    !> Compute tag for mpi message
    public      :: compute_tag_gap
    public      :: compute_tag_NP

    interface compute_tag
        module procedure compute_tag_gap, compute_tag_NP
    end interface compute_tag

contains

!> Creation of the cartesian mesh and of all communicators
subroutine cart_create(dims, ierr)

    use mpi
    use parallel

    !> @param[in]    dims    = array specifying the number of processes in each dimension
    !> @param[out]   ierr    = error code
    integer, dimension(2), intent(in)   :: dims
    integer, intent(out)                :: ierr

    logical                 :: reorganisation                   ! to choose to reordered or not the processus rank.
    logical, dimension(3)   :: remains_dim                      ! use to create 1D-subdivision : remains_dims(i) equal
                                                                ! true if the i-th dimension is kept in the subgrid.
    ! Creation of the cartesian topology
    reorganisation = .true.
    periods = .true.
    nb_proc_dim = (/ 1, dims(1), dims(2) /)

    call mpi_cart_create(MPI_COMM_WORLD, 3, nb_proc_dim, periods, reorganisation, &
            & cart_comm, ierr)


    ! Subdivision in 1D-subgrids and creation of communicator devoted to
    ! 1D-communication
    ! Communication along X-axis 
    remains_dim = (/.true., .false., .false. /)
    call mpi_cart_sub(cart_comm, remains_dim, X_comm, ierr)
    D_comm(1) = X_comm
    ! Communication along Y-axis (in the 3D mesh, ie the x-axis on the mpi-topology)
    remains_dim = (/.false., .true., .false. /)
    call mpi_cart_sub(cart_comm, remains_dim, Y_comm, ierr)
    D_comm(2) = Y_comm
    ! Communication along Z-axis 
    remains_dim = (/ .false., .false., .true. /)
    call mpi_cart_sub(cart_comm, remains_dim, Z_comm, ierr)
    D_comm(3) = Z_comm

    ! Initialise information about the current processus
    call mpi_comm_rank(cart_comm, myrank, ierr)
    call mpi_cart_coords(cart_comm, myrank, 3, coord, ierr)
    coordYZ = (/ coord(2), coord(3) /)

    ! XXX ajouter un test du code d'erreur mpi
!    if(ierr/=MPI_SUCCESS) then 
!        call parallel_error(ierr, 'cart_create')
!    end if

end subroutine cart_create

!> Defaut mesh setup
subroutine mesh_default()

    ! A cubic geometry : unitary lengh and 100 mesh points in each direction.
    N = 100
    length = 1.
    length = 1
    N_proc = N / nb_proc_dim
    begin_proc = 1
    end_proc = N_proc

    call mesh_init()
    
end subroutine mesh_default

!> To initialize last mesh parameters
subroutine mesh_init()

    integer                 :: direction    ! direction (along X = 1, along Y = 2, along Z = 3)
    integer                 :: group_dir    ! direction "bis"
    integer, dimension(3,2) :: N_group      ! number of group on one processus along one direction

    d_sc = length/(N)

    ! XXX Pour le moment, on travaille par groupe d'une unique ligne
    group_size = 1
    
    ! Compute number of group
    ! Group of line along X
    N_group(1,1) = N_proc(2)/group_size(1,1)
    N_group(1,2) = N_proc(3)/group_size(1,2)
    ! Group of line along X
    N_group(2,1) = N_proc(1)/group_size(2,1)
    N_group(2,2) = N_proc(3)/group_size(2,2)
    ! Group of line along X
    N_group(3,1) = N_proc(1)/group_size(3,1)
    N_group(3,2) = N_proc(2)/group_size(3,2)
    

    ! tag_size = smallest power of ten to ensure tag_size > max ind_group
    do direction = 1,3
        tag_size(direction,:) = 1
        do group_dir = 1,2
            do while (N_group(direction, group_dir)/(10**tag_size(direction, group_dir))>1)
                tag_size(direction, group_dir) = tag_size(direction, group_dir)+1 
            end do
        end do
    end do

    tag_rank = 1
    do while((nb_proc_dim(1)+nb_proc_dim(2)+nb_proc_dim(3))/(10**tag_rank)>=1)
        tag_rank = tag_rank+1
    end do
    if (tag_rank == 1) tag_rank = 2


if(myrank==0) then
    print*, ' '
    print*, ' -- initialisation: tag generation --'
    do direction = 1,3
        print*, 'tag_size(',direction,',:) = ', tag_size(direction,:)
    end do
    print*, 'tag_rank = ', tag_rank
    print*, ' ------------------------------------'
    print*, ' '
end if

end subroutine mesh_init

!> Compute unique tag for mpi message by concatenation of position (ie line coordinate), proc_gap and unique Id
!!    @param[in]    ind_group   = indice of current group of line
!!    @param[in]    tag_param   = couple of int unique for each message (used to create the tag)
!!    @param[in]    direction   = current direction
!!    @param[in]    proc_gap    = number of processus between the sender and the receiver
!!    @return       tag         = unique tag: at each message send during an iteration have a different tag
!!@details 
!!     Use this procedure to compute tag in order to communicate with a distant processus or/and when
!!    you will send more then two message. It produce longer tag compute_tag_NP because rather tyo use 0/1 it 
!!    put the gap between the sender and the receiver (ie the number of processus between them) in the tag.
!!    Using these two procedure allow to obtain more unique tag for communication.
function compute_tag_gap(ind_group, tag_param, direction,proc_gap) result(tag)

    integer                             :: tag
    integer, dimension(2), intent(in)   :: ind_group
    integer, dimension(2), intent(in)   :: tag_param
    integer, intent(in)                 :: direction
    integer, intent(in)                 :: proc_gap

    integer                              :: abs_proc_gap ! absolute value of proc_gap

    abs_proc_gap = max(abs(proc_gap),1)
    tag = (tag_param(1)*10+direction)*(10**(tag_rank+1))
    tag = tag + abs_proc_gap*10 + dim(proc_gap/abs_proc_gap,0) 
    tag = tag*(10**tag_size(direction,1))+(ind_group(1)-1)
    tag = ((tag*(10**tag_size(direction,2)))+(ind_group(2)-1))
    tag = (tag*10)+tag_param(2)

if ((tag<0) .and. (myrank==0))  then
print*, 'ind_group = ', ind_group, ' ; tag_param = ', tag_param
print*, 'direction = ', direction, ' gap = ', proc_gap ,' and tag = ', tag
end if

end function compute_tag_gap


!> Compute unique tag for mpi message by concatenation of position(ie line coordinate), +1 or -1 and unique Id
!!    @param[in]    ind_group   = indice of current group of line
!!    @param[in]    tag_param   = couple of int unique for each message (used to create the tag)
!!    @param[in]    direction   = current direction
!!    @return       tag_table   = unique couple tag: use tag_table(1) for mesage to previous proc. (or first 
!!                                message ) and tag_table(2) for the other message.
!!@details 
!!     Use this procedure to compute tag for communication with your neighbor or when only two message are send:
!!    it produce smaller tag then compute_tag_gap because the gap between sender and receiver are replaced by 1,
!!    for communicate with previous processus (or first of the two message), or 0, for communication with next 
!!    processus (or the second message). It allow to reuse some unique Id.
function compute_tag_NP(ind_group, tag_param, direction) result (tag_table)

    integer, dimension(2)               :: tag_table
    integer, dimension(2), intent(in)   :: ind_group
    integer, dimension(2), intent(in)   :: tag_param
    integer, intent(in)                 :: direction

    tag_table(2) = (tag_param(1)*10+direction)*10
    tag_table(1) = tag_table(2)
    
    tag_table(2) = tag_table(2) +1
   
    tag_table(2) = (tag_table(2)*(10**tag_size(direction,1)))+(ind_group(1)-1)
    tag_table(1) = (tag_table(1)*(10**tag_size(direction,1)))+(ind_group(1)-1)

    tag_table(2) = ((tag_table(2)*(10**tag_size(direction,2)))+(ind_group(2)-1))
    tag_table(1) = ((tag_table(1)*(10**tag_size(direction,2)))+(ind_group(2)-1))

    tag_table(2) = (tag_table(2)*10)+tag_param(2)
    tag_table(1) = (tag_table(1)*10)+tag_param(2)

if ((minval(tag_table)<0) .and. (myrank==0))  then
print*, 'ind_group = ', ind_group, ' ; tag_param = ', tag_param
print*, 'direction = ', direction, ' and tag = ', tag_table
end if

end function compute_tag_NP


end module cart_topology


