module advec

    use string
    implicit none

    ! ===== Variables =====
    !> numerical method use to advect the scalar
    character(len=str_short), private   :: type_part_solv   
    !> dimensionnal splitting (eg classical, Strang or particle)
    character(len=str_short), private   :: dim_splitting    


    ! ===== Public procedures =====
    
    ! Scheme used to advec the scalar (order 2 or 4 ?)
    public                              :: type_part_solver

    ! Advection methods
    public                              :: advec_init   ! initialize the scalar solver
    public                              :: advec_step   ! advec the scalar field during a time step.

contains

! ===== Public methods =====

!> Return the name of the particle method used for the advection
function type_part_solver()
    character(len=str_short)    :: type_part_solver
    
    type_part_solver = type_part_solv
end function

!> Initialise the particle advection methods
!!    @param order      = to choose the remeshing method (and thus the order)
subroutine advec_init(order)
    
    use cart_topology
    use advec_common

    character(len=*), optional, intent(in)  ::  order

    ! Remplissage de la valeur par defaut en l'absence de choix explicite de la méthode numérique
    if(present(order)) then
        type_solv = order
    else 
        type_solv = 'p_O2'
    end if

    ! Initialisation part adapted to each method
    select case(type_solv)
        case('p_O2')
            bl_size = 2
            bl_bound_size = 1
        case('p_O4')
            bl_size = 4
            bl_bound_size = 2
        case default
            bl_size = 2
            bl_bound_size = 1
    end select

    ! Check if the subdomain contain a number of mesh wich could be divided by bl_size
    if ((modulo(N_proc(1),bl_size)/=0).OR.(modulo(N_proc(2),bl_size)/=0).OR.(modulo(N_proc(3),bl_size)/=0)) then
        if (myrank ==0) print*, 'Number of mesh by processus must be a muliple of ', bl_size
        stop
    end if

    ! Compute local number of block along each direction
    bl_number = N_proc/bl_size

    ! Choosing the dimensionnal splitting to use
    ! XXX parser le fichier input
    dim_splitting = 'strang'

end subroutine advec_init


!> advec_scheme
subroutine advec_step(dt, Vx, Vy, Vz, scal, dim_split)

    use cart_topology
    use advec_common
    use advecX
    use advecY
    !use advecZ

    ! @param[in]        dt          = time step
    ! @param[in]        Vx          = velocity along x (could be discretised on a bigger mesh then the scalar)
    ! @param[in]        Vy          = velocity along y 
    ! @param[in]        Vz          = velocity along z
    ! @param[in,out]    scal        = scalar field to advect
    ! @param[in]        dim_split   = dimensionnal splitting (eg classical,
                                        !! Strang splitting or particle splitting)
    real(WP), intent(in)                                    :: dt
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(in)   :: Vx, Vy, Vz
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(inout):: scal
    character(len=str_short), optional, intent(in)                              :: dim_split
   
    character(len=str_short)        :: splitting    ! to choose the splitting

    ! Default dimensionnal splitting if the user do not choose it
    if(present(dim_split)) then
        splitting = dim_split
    else 
        splitting = dim_splitting
    end if

    ! Scheme used for advection : particle method (order 2 or 4) or spectral one.
    if (type_solv=='spectral') then
        print*, 'Solveur non implémenté'
    end if
        
    select case(splitting)
        case('classical')
            !call advecX_calc(dt, Vx, scal, type_part_solv)
            call advecY_calc(dt, Vy, scal, type_part_solv)
            !call advecZ_calc(dt, Vz, scal, type_part_solv)
        case('Strang')
            !call advecX_calc(dt/2.0, Vx, scal, type_part_solv)
            call advecY_calc(dt/2.0, Vy, scal, type_part_solv)
            !call advecZ_calc(dt/2.0, Vz, scal, type_part_solv)
            !call advecZ_calc(dt/2.0, Vz, scal, type_part_solv)
            call advecY_calc(dt/2.0, Vy, scal, type_part_solv)
            !call advecX_calc(dt/2.0, Vz, scal, type_part_solv)
        case default
            !call advecX_calc(dt/2.0, Vx, scal, type_part_solv)
            call advecY_calc(dt/2.0, Vy, scal, type_part_solv)
            !call advecZ_calc(dt/2.0, Vz, scal, type_part_solv)
            !call advecZ_calc(dt/2.0, Vz, scal, type_part_solv)
            call advecY_calc(dt/2.0, Vy, scal, type_part_solv)
            !call advecX_calc(dt/2.0, Vx, scal, type_part_solv)
    end select

end subroutine advec_step
    

!> ===== Private procedure =====


end module advec

