!------------------------------------------------------------------------------
!
! MODULE: advecY
!
!
! DESCRIPTION: 
!> This module is a part of the advection solver based on particles method.
!! The solver use some dimensionnal splitting and this module contains all the
!! method used to solve advection along the Y-axis. This is a parallel
!! implementation using MPI and the cartesien topology it provides.
!!
!! This module can use the method and variables defined in the module
!! "advec_common" which gather information and tools shared for advection along
!! x, y and z-axis.
!!
!! The module "test_advec" can be used in order to validate the procedures
!! embedded in this module.
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

!> @addtogroup part
!! @{
module advecY

    use precision

    implicit none

    ! ===== Public procedures =====
    !> Generique procedure to advect the scalar with a particles solver
    public                  :: advecY_calc
    !----- (corrected) Remeshing method (these methods are set to public in validation purposes) -----
    public                  :: Yremesh_O2       ! order 2
    public                  :: Yremesh_O4       ! order 4

    ! ===== Private porcedures =====
    !> particles solver with remeshing method at order 2
    private                 :: advecY_calc_O2
    private                 :: advecY_init      ! initialize particle position, velocity and weight

    contains

! ===== Public procedure =====

!> Scalar advection (this procedure call the right solver, depending on the simulation setup) 
!!    @param[in]        dt          = time step
!!    @param[in]        Vy          = velocity along y (could be discretised on a bigger mesh then the scalar)
!!    @param[in,out]    SC          = scalar field to advect
!!    @param[in]        type_solver = scheme use for the advection (particle with order 2 or 4)
subroutine advecY_calc(dt, Vy, SC, type_solver)
  
    use string
    use cart_topology

    real(WP), intent(in)                                                :: dt
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(in)    :: Vy
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(inout) :: SC
    character(len=*), intent(in)                                        :: type_solver



    ! Call the right solver depending on the space order we want to.
    select case(type_solver)
        case('p_O2')
            call advecY_calc_O2(dt, Vy, SC)
        case default
            call advecY_calc_O2(dt, Vy, SC)
    end select

end subroutine advecY_calc

! ----- Remeshing procedures -----


! ===== Private procedure =====

! ---- Order 2 solver, with correction for large time step ----

!> remeshing with an order 2 method, corrected to allow large CFL number - untagged particles
!!    @param[in]        ind_group   = coordinate of the current group of lines
!!    @param[in]        p_pos_adim  = adimensionned  particles position
!!    @param[in]        bl_type     = equal 0 (resp 1) if the block is left (resp centered)
!!    @param[in]        bl_tag      = contains information about bloc (is it tagged ?)
!!    @param[in]        i,k         = indice of of the current line (x-coordinate and z-coordinate)
!!    @param[in,out]    scal        = scalar field to advect
subroutine Yremesh_O2(ind_group, p_pos_adim, bl_type, bl_tag,i,k,scal)

    use advec_common
    use cart_topology

    integer, dimension(2), intent(in)                                   :: ind_group
    integer, intent(in)                                                 :: i, k
    integer, dimension(:), intent(in)                                   :: bl_type
    integer, dimension(:), intent(in)                                   :: bl_tag
    real(WP), dimension(:), intent(in)                                  :: p_pos_adim
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(inout) :: scal
  
integer     :: direction =2                            ! current direction

    real(WP), dimension(N_proc(2))  :: p_sc             ! scalar advected by the particles
    ! Variable used to remesh particles in a buffer
    integer     :: j                                    ! indice of the current particle
    integer     :: bl_ind                               ! indice of the current "block end".
    integer     :: bl_info                              ! to know if particles are tagged and left/centered
                                                        ! (they depend on the block type)
    real(WP),dimension(:),allocatable   :: send_buffer  ! buffer use to remesh the scalar before to send it to the right subdomain
    

    ! -- Initialise particle weigth
    do j = 1, N_proc(direction)
        p_sc(j) = scal(i,j,k)
        scal(i,j,k) = 0
    end do

    !  -- Allocate and initialize the buffer --
    if (bl_type(1)==0) then
        ! First particle is a left one
        send_j_min = floor(p_pos_adim(1))-1
    else
        ! First particle is a centered one
        send_j_min = nint(p_pos_adim(1))-1
    end if
    if (bl_type(N_proc(direction)/bl_size +1)==0) then
        ! Last particle is a left one
        send_j_max = floor(p_pos_adim(N_proc(2)))+1
    else
        ! Last particle is a centered one
        send_j_max = nint(p_pos_adim(N_proc(2)))+1
    end if
        
    ! XXX j_min and j_max could be optimized !! (use that sub-domains cut
            ! particles block in two egals parts)
    allocate(send_buffer(send_j_min:send_j_max))
    send_buffer = 0.0;

    ! -- Remesh the particles in the buffer --
    ! The remeshing formula depends on the particle type :
    !     1 - Is the particle tagged ?
    !     2 - Does it belong to a centered or a left block ?
    !   Observe that tagged particles go by group of two : if the particles of a
    !   block end are tagged, the one first one of the following block are
    !   tagged too.
    !   The following algorithm is write for block of minimal size.
    do j = 1, N_proc(direction), bl_size
        bl_ind = j/bl_size + 1
        bl_info = 100*bl_type(bl_ind)+(10*bl_type(bl_ind+1))+bl_tag(bl_ind)
        select case (bl_info)
            case (0)    
            ! no tag and only left particle 
                call AC_remesh_left(p_pos_adim(j),p_sc(j), send_buffer)
                call AC_remesh_left(p_pos_adim(j+1),p_sc(j+1), send_buffer) 
            case (10)    
            ! no tag, the first particle belong to a left block and the last to centered block.
                call AC_remesh_left(p_pos_adim(j),p_sc(j), send_buffer)
                call AC_remesh_center(p_pos_adim(j+1),p_sc(j+1), send_buffer)
            case (110)    
            ! no tag and only centered particle
                call AC_remesh_center(p_pos_adim(j),p_sc(j), send_buffer)
                call AC_remesh_center(p_pos_adim(j+1),p_sc(j+1), send_buffer)
            case (100)    
            ! no tag, the first particle belong to a centered block and the last to left block.
                call AC_remesh_center(p_pos_adim(j),p_sc(j), send_buffer)
                call AC_remesh_left(p_pos_adim(j+1),p_sc(j+1), send_buffer)
            ! XXX add the tagged cases
            case (101)
            ! tagged, the first particle belong to a centered block and the last to left block.
                call AC_remesh_tag_CL(p_pos_adim(j), p_sc(j), p_pos_adim(j+1), p_sc(j+1), send_buffer)
            case (11)
            ! tagged, the first particle belong to a left block and the last to centered block.
                call AC_remesh_tag_LC(p_pos_adim(j), p_sc(j), p_pos_adim(j+1), p_sc(j+1), send_buffer)
            case default
                print*, 'error on remeshing particles : bl_info equals to ', &
                    & bl_info, ' and must be 0, 10, 110, 100, 101 or 11. Mesh point = (',i, ', ', j,', ',k,')'
                print*,  'paramètres du blocs : ind = ', bl_ind, ' , type(ind) = ', bl_type(bl_ind), &
                    & ' , type(ind+1) = ', bl_type(bl_ind+1), ' tag = ', bl_tag(bl_ind)
                stop
        end select
    end do
    
    ! -- Send the buffer to the matching processus and update the scalar field --
    call AC_bufferToScalar(direction, Y_comm, ind_group , send_j_min, send_j_max, send_buffer, scal(i,:,k))

    ! Deallocate all field
    deallocate(send_buffer)

end subroutine Yremesh_O2


!> remeshing with an order 4 method, corrected to allow large CFL number - untagged particles
!!    @param[in]        ind_group   = coordinate of the current group of lines
!!    @param[in]        p_pos_adim  = adimensionned particles position
!!    @param[in]        bl_tag      = contains information about block (is it tagged ?)
!!    @param[in]        i,k         = indice of of the current line (x-coordinate and z-coordinate)
!!    @param[in]        bl_type     = equal 0 (resp 1) if the block is left (resp centered)
!!    @param[in,out]    scal        = scalar field to advect
subroutine Yremesh_O4(ind_group, p_pos_adim, bl_type, bl_tag,i,k,scal)

    use advec_common
    use cart_topology

    integer, dimension(2), intent(in)                       :: ind_group
    integer, intent(in)                                     :: i, k
    integer, dimension(:), intent(in)                       :: bl_type
    integer, dimension(:), intent(in)                       :: bl_tag
    real(WP), dimension(:), intent(in)                      :: p_pos_adim
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(inout) :: scal
  
    ! Variable used to remesh particles ...
    ! ... and to communicate between subdomains. A variable prefixed by "send_"(resp "rece")
    ! design something I send (resp. I receive).
    real(WP), dimension(N_proc(2))  :: p_sc             ! scalar advected by the particles
    integer     :: j                                    ! indice of the current particle
    integer     :: bl_ind                               ! indice of the current "block end".
    integer     :: bl_info                              ! to know if particles are tagged and left/centered
                                                        ! (they depend on the block type)
    real(WP),dimension(:),allocatable   :: send_buffer  ! buffer use to remesh the scalar before to send it to the right subdomain
    integer     :: proc_gap                             ! gap between my Y-coordinate and the one of the processus 
                                                        ! with wich I communicate.
    integer     :: proc_min, proc_max                   ! minimal and maximal gap between my Y-coordinate and the one of
                                                        ! sub-domains wherein the particle will be remeshed.
    integer     :: ierr                                 ! mpi error code
    integer     :: comm_size                            ! number of element to send/receive
    integer     :: tag                                  ! mpi message tag
    integer     :: direction = 2                        ! current direction = along Z

    ! -- Initialise particle weigth
    do j = 1, N_proc(direction)
        p_sc(j) = scal(i,j,k)
        scal(i,j,k) = 0
    end do

    !  -- Allocate and initialize the buffer --
    if (bl_type(1)==0) then
        ! First particle is a left one
        send_j_min = floor(p_pos_adim(1))-2
    else
        ! First particle is a centered one
        send_j_min = nint(p_pos_adim(1))-2
    end if
    if (bl_type(N_proc(direction)/bl_size +1)==0) then
        ! Last particle is a left one
        send_j_max = floor(p_pos_adim(N_proc(2)))+2
    else
        ! Last particle is a centered one
        send_j_max = nint(p_pos_adim(N_proc(2)))+2
    end if
        
    ! -- Remesh the particles in the buffer --
    ! The remeshing formula depends on the particle type :
    !     1 - Is the particle tagged ?
    !     2 - Does it belong to a centered or a left block ?
    !   Observe that tagged particles go by group of two : if the particles of a
    !   block end are tagged, the one first one of the following block are
    !   tagged too.
    !   The following algorithm is write for block of minimal size.
    do j = 1, N_proc(direction), bl_size
        bl_ind = j/2
        bl_info = 100*bl_type(bl_ind)+(10*bl_type(bl_ind+1))+bl_tag(bl_ind)
        select case (bl_info)
            case (0)    
            ! no tag and only left particle
                call AC_remesh_left(p_pos_adim(j),p_sc(j), send_buffer)
                call AC_remesh_left(p_pos_adim(j+1),p_sc(j+1), send_buffer)
                call AC_remesh_left(p_pos_adim(j+2),p_sc(j+2), send_buffer)
                call AC_remesh_left(p_pos_adim(j+3),p_sc(j+3), send_buffer)
            case (10)    
            ! no tag, the 2 first particles belong to a left block and the two last to centered block.
                call AC_remesh_left(p_pos_adim(j),p_sc(j), send_buffer)
                call AC_remesh_left(p_pos_adim(j+1),p_sc(j+1), send_buffer)
                call AC_remesh_center(p_pos_adim(j+2),p_sc(j+2), send_buffer)
                call AC_remesh_center(p_pos_adim(j+3),p_sc(j+3), send_buffer)
            case (110)    
            ! no tag and only left particle
                call AC_remesh_center(p_pos_adim(j),p_sc(j), send_buffer)
                call AC_remesh_center(p_pos_adim(j+1),p_sc(j+1), send_buffer)
                call AC_remesh_center(p_pos_adim(j+2),p_sc(j+2), send_buffer)
                call AC_remesh_center(p_pos_adim(j+3),p_sc(j+3), send_buffer)
            case (100)    
            ! no tag, the 2 first particles belong to a centered block and the two last to left block.
                call AC_remesh_center(p_pos_adim(j),p_sc(j), send_buffer)
                call AC_remesh_center(p_pos_adim(j+1),p_sc(j+1), send_buffer)
                call AC_remesh_left(p_pos_adim(j+2),p_sc(j+2), send_buffer)
                call AC_remesh_left(p_pos_adim(j+3),p_sc(j+3), send_buffer)
            ! XXX add the tagged cases
            case default
                print*, 'error on remeshing particles'
        end select
    end do
    
    ! -- Send the buffer to the matching processus and update the scalar field --
    call AC_bufferToScalar(direction, Y_comm, ind_group, send_j_min, send_j_max, send_buffer, scal(i,:,k))

    ! Deallocate all field
    deallocate(send_buffer)

end subroutine Yremesh_O4


!> Advection during a time step dt - order 2
!!    @param[in]        dt      = time step
!!    @param[in]        Vy      = velocity along y (could be discretised on a bigger mesh then the scalar)
!!    @param[in,out]    scal3D   = scalar field to advect
subroutine advecY_calc_O2(dt,Vy,scal3D)

    use cart_topology
    use advec_common
use mpi


    real(WP), intent(in)                                                :: dt
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(in)    :: Vy
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(inout) :: scal3D

    integer                             :: i,k          ! indice of the currend mesh point
    integer, dimension(2)               :: ind_group    ! indice of the currend group of line (=(i,k) by default)
    integer                             :: direction=2  ! current direction = along Y
    real(WP), dimension(N_proc(2))      :: p_pos_adim   ! adimensionned particles position 
    real(WP), dimension(N_proc(2))      :: p_V          ! particles velocity 
    integer, dimension(bl_number(2)+1)  :: bl_type      ! is the particle block a center block or a left one ?
    integer, dimension(bl_number(2))    :: bl_tag       ! indice of tagged particles
integer :: ierr

    ind_group = 0
    do k = 1, N_proc(3)
        ind_group(2) = ind_group(2) + 1
        ind_group(1) = 0
        do i = 1, N_proc(1)
        ind_group(1) = ind_group(1) + 1

            ! ===== Init particles =====
            call advecY_init(Vy, i, k, p_pos_adim, p_V)

            ! ===== Advection =====
            ! -- Compute velocity (with a RK2 scheme) --
            call AC_particle_velocity(dt, direction, Y_comm, ind_group, p_pos_adim, p_V)
            ! -- Determine blocks type and tag particles --
            call AC_type_and_block(dt, direction, Y_comm, ind_group, p_V, &
                    & bl_type, bl_tag)
            ! -- Advec particles --
            p_pos_adim = p_pos_adim + dt*p_V/d_sc(direction)

            ! ===== Remeshing =====
           call Yremesh_O2(ind_group, p_pos_adim, bl_type, bl_tag,i,k,scal3D)

call mpi_barrier(MPI_COMM_WORLD, ierr)
        end do
    end do

end subroutine advecY_calc_O2

!> Particle creation and initialisation
!!    @param[in]    Vy          = 3D velocity field
!!    @param[in]    i           = X-indice of the current line
!!    @param[in]    k           = Z-indice of the current line
!!    @param[out]   p_pos_adim  = adimensioned particles postion
!!    @param[out]   p_V         = particle velocity 
subroutine advecY_init(Vy, i, k, p_pos_adim, p_V)

    use cart_topology

    integer, intent(in)                         :: i,k
    real(WP), dimension(N_proc(2)), intent(out) :: p_pos_adim, p_V
    real(WP), dimension(:,:,:), intent(in)      :: Vy

    integer                                     :: direction=2  ! current direction
    integer                                     :: ind          ! indice

    do ind = 1, N_proc(direction)
        p_pos_adim(ind) = ind
        p_V(ind)        = Vy(i,ind,k)
    end do

end subroutine advecY_init

end module advecY
!! @}
