
!> @addtogroup part
!! @{
module advec_common

    use precision
    use string

    implicit none

! XXX Si passage au fortran 2003 : basculer toutes ces variables dans le module
! advec (fichier advec.F90) et mettre toutes les variables en protected.
! Seul la procédure "advec_init" doit pouvoir les modifier, mais de nombreuses
! procédures doivent pouvoir y accéder.


    ! Information about the particles and their bloc
    private
    ! XXX Public variables have to be switched to "PROTECTED" when we will use fortran2003 norm

    ! ===== Private variables =====
    ! ----- Minimal and maximal indice of the buffer used in the different communication -----
    !> minimal indice of the send buffer
    integer, public                 :: send_j_min
    !> maximal indice of the send buffer
    integer, public                 :: send_j_max
    ! ------ Block infromation -----
    !> number of particles in a block
    integer, public                 :: bl_size           
    !> number of particles in a block boundary wich used a specific remeshing formula
    !! which is related to the stencil width.
    integer, public                 :: bl_bound_size     
    !> Number of block on each processus along each direction
    integer, dimension(3), public   :: bl_number

    ! ------ To ensure unique mpi message tag -----
    ! Tag generate with a proc_gap
    !> To create tag used in AC_particle_velocity to send range
    integer, dimension(2)           :: tag_velo_range = (/ 0,1 /)
    !> To create tag used in AC_particle_velocity to send velocity field
    integer, dimension(2)           :: tag_velo_V = (/ 0,2 /)
    !> To create tag used in bufferToScalar to send range of buffer which will be send
    integer, dimension(2)           :: tag_bufToScal_range = (/ 0,3 /)
    !> To create tag used in bufferToScalar to send the buffer used to remesh particles
    integer, dimension(2)           :: tag_bufToScal_buffer = (/ 0,4 /)

    ! Tag generate with "compute_gap_NP"
    !> To create tag used in AC_obtain_recevers to send ghost
    integer, dimension(2)           :: tag_obtrec_ghost_NP = (/ 0, 1/)
    !> To create tag used in AC_type_and_bloc to exchange ghost with neighbors
    integer, dimension(2)           :: tag_part_tag_NP = (/ 0, 2/)
    !> To create tag used in AC_obtain_recevers to send message about recevers of minimal and maximal rank
    integer, dimension(2)           :: tag_obtrec_NP = (/ 0, 3/)
    !> To create tag used in AC_obtain_receivers to send message about senders of minimal and maximal rank
    integer, dimension(2)           :: tag_obtsend_NP = (/ 0, 4/)
    
    




    ! ===== Public procedures =====
    character(len=str_short), public    :: type_solv
    !public                              :: tag_particles_O2
    !----- Determine block type and tag particles -----
    public                              :: AC_type_and_block
    !----- To interpolate velocity -----
    public                              :: AC_obtain_receivers
    public                              :: AC_particle_velocity
    !----- To remesh particles -----
    public                              :: AC_obtain_senders
    public                              :: AC_bufferToScalar
    !----- Order 2 remeshing formula -----
    public                              :: AC_remesh_left      ! left remeshing formula
    public                              :: AC_remesh_center    ! centered remeshing formula
    public                              :: AC_remesh_tag_CL    ! corrected formula for tagged particles : transition from C to L block.
    public                              :: AC_remesh_tag_LC    ! corrected formula for tagged particles : transition from L to C block


    contains

! ===== Public procedure =====

! XXX Subroutine à mettre à jour
!    eps_bl=d_sc/10.0
!    cfl = dt/d_sc
!
!    bl_nb_part = 0
!    bl_final_ind = 0
!    bl_lambdaMin = maxval(p_V)*cfl    ! or any big value. We can perhaps used
!                                      ! the cfl condition from the NS solver
!   
!
!    deallocate(bl_nb_part)
!    deallocate(bl_final_ind)
!    deallocate(bl_lambdaMin)
!    deallocate(bl_ind)
!    deallocate(bl_type)
!
!end subroutine tag_particles_O2

! ==================================================================================
! ====================     Compute particle velocity (RK2)      ====================
! ==================================================================================

!> Interpolate the velocity field used in a RK2 scheme for particle advection.
!!    @param[in]        dt          = time step
!!    @param[in]        direction   = current direction (1 = along X, 2 = along Y and 3 = along Z)
!!    @param[in]        comm        = mpi communicator associated to the current direction (Y_comm, Z_comm)
!!    @param[in]        ind_group   = coordinate of the current group of lines
!!    @param[in]        p_pos_adim  = adimensionned particle postion
!!    @param[in,out]    p_V         = particle velocity (along the current direction)
!! @details
!!    A RK2 scheme is used to advect the particles : the midlle point scheme. An
!!    intermediary position "p_pos_bis(i) = p_pos(i) + V(i)*dt/2" is computed and then 
!!    the numerical velocity of each particles is computed as the interpolation of V  in 
!!    this point. This field is used to advect the particles at the seconde order in time : 
!!    p_pos(t+dt, i) = p_pos(i) + p_V(i).
!!    The group line indice is used to ensure using unicity of each mpi message tag.
subroutine AC_particle_velocity(dt, direction, comm, ind_group, p_pos_adim, p_V)

    ! This code involve a recopy of p_V. It is possible to directly use the 3D velocity field but in a such code
    ! a memory copy is still needed to send velocity field to other processus : mpi send contiguous memory values

    use mpi
    use cart_topology

    ! --- In order to create an arry of pointer ---
    type real_pter
        real(WP), pointer                           :: pter
    end type real_pter
    ! ---------------------------------------------

    real(WP), intent(in)                            :: dt       ! time step
    integer, intent(in)                             :: direction
    integer, intent(in)                             :: comm
    integer, dimension(2), intent(in)               :: ind_group
    real(WP), dimension(:), intent(in)              :: p_pos_adim
    real(WP), dimension(:), intent(inout)           :: p_V

    real(WP), dimension(N_proc(direction))          :: p_pos_bis    ! adimensionned position of the middle point
    real(WP), dimension(N_proc(direction)), target  :: p_V_bis      ! velocity of the middle point
    real(WP), dimension(N_proc(direction))          :: weight       ! interpolation weight
    type(real_pter), dimension(N_proc(direction))   :: Vp, Vm       ! Velocity on previous and next mesh point
    real(WP), dimension(:), allocatable, target     :: V_buffer     ! Velocity buffer for postion outside of the local subdomain
    integer                                         :: size_buffer  ! buffer size
    integer                                         :: temp         ! use in some computations
    integer                                         :: rece_ind_min ! the minimal indice used in velocity interpolation
    integer                                         :: rece_ind_max ! the maximal indice used in velocity interpolation
    integer                                         :: ind, ind_com ! indices
    integer                                         :: pos, pos_old ! indices of the mesh point wich preceed the particle position
    integer                                         :: proc_gap, gap! distance between my (mpi) coordonate and coordinate of the 
                                                                    ! processus associated to a given position
    integer, dimension(:), allocatable              :: rece_rank    ! rank of processus wich send me information
    integer                                         :: send_rank    ! rank of processus to wich I send information
    integer                                         :: rankP        ! rank of processus ("source rank" returned by mpi_cart_shift)
    integer, dimension(2)                           :: rece_range   ! range of the velocity fields I want to receive
    integer, dimension(2)                           :: send_range   ! range of the velocity fields I send
    integer, dimension(2)                           :: rece_gap     ! distance between me and processus wich send me information
    integer, dimension(2)                           :: send_gap     ! distance between me and processus to wich I send information
    integer                                         :: msg_size     ! size of message send/receive
    integer                                         :: tag          ! mpi message tag
    integer                                         :: ierr         ! mpi error code
    integer                                         :: send_request ! mpi communication request (handle) of nonblocking send
! XXX debug - to delete XXX
    integer, dimension(:), allocatable              :: s_request    ! mpi communication request (handle) of nonblocking send
    integer, dimension(:), allocatable              :: s_request_bis! mpi communication request (handle) of nonblocking send
! XXX end debug XXX
    integer, dimension(:), allocatable              :: rece_request ! mpi communication request (handle) of nonblocking receive
    integer, dimension(MPI_STATUS_SIZE)             :: rece_status  ! mpi status (for mpi_wait)

    ! -- Initialisation --
    ind_com = 0
    do ind = 1, N_proc(direction)
        nullify(Vp(ind)%pter)
        nullify(Vm(ind)%pter)
    end do
    ! Compute the midlle point
    p_pos_bis = p_pos_adim + (dt/2.0)*p_V/d_sc(direction)
    p_V_bis = p_V
    ! Compute range of the set of point where I need the velocity value
    rece_ind_min = floor(p_pos_bis(1))
    rece_ind_max = floor(p_pos_bis(N_proc(direction))) + 1
    ! Allocate the buffer
    ! If rece_ind_min and rece_ind_max are not in [N_proc(direction);1] then it will change the number of communication
    ! size_buffer = max(temp - N_proc(direction), 0) - min(0, temp)
    !size_buffer = - max(temp - N_proc(direction), 0) - min(0, temp)
    ! It must work, but for first test we prefer compute size_buffer more simply
    size_buffer = 0

    ! -- Exchange non blocking message to do the computations during the
    ! communication process
    call AC_obtain_receivers(direction, comm, ind_group, rece_ind_min, rece_ind_max, send_gap, rece_gap)
    allocate(rece_rank(rece_gap(1):rece_gap(2)))
    ! Send messages about what I want
! XXX debug  - to delete XXX
    allocate(s_request_bis(rece_gap(1):rece_gap(2)))
! XXX end debug XXX
    do proc_gap = rece_gap(1), rece_gap(2)
        call mpi_cart_shift(cart_comm, direction-1, proc_gap, rankP, rece_rank(proc_gap), ierr)
        if (rece_rank(proc_gap) /= myrank) then
            ! Range I want
            gap = proc_gap*N_proc(direction)
            rece_range(1) = max(rece_ind_min, gap+1) ! fortran => indice start from 0
            rece_range(2) = min(rece_ind_max, gap+N_proc(direction)) 
            ! Tag = concatenation of (rank+1), ind_group(1), ind_group(2), direction et unique Id.                          
            tag = compute_tag(ind_group, tag_velo_range, direction, proc_gap)
            ! Send message
            size_buffer = size_buffer + (rece_range(2)-rece_range(1)) + 1
! XXX debug  - to delete XXX
            call mpi_Isend(rece_range(1), 2, MPI_INTEGER, rece_rank(proc_gap), tag, comm, s_request_bis(proc_gap), ierr)
! XXX debug - uncomment the following line XXX
!            call mpi_Isend(rece_range(1), 2, MPI_INTEGER, rece_rank(proc_gap), tag, comm, send_request, ierr)
! XXX end debug XXX
        end if
    end do
    allocate(V_buffer(max(size_buffer,1)))
    V_buffer = 0
    ! Send the velocity field to processus which need it
! XXX debug  - to delete XXX
    allocate(s_request(send_gap(1):send_gap(2)))
! XXX end debug XXX
    do proc_gap = send_gap(1), send_gap(2)
        call mpi_cart_shift(cart_comm, direction-1, proc_gap, rankP, send_rank, ierr)
        if (send_rank /= myrank) then
            ! I - Receive messages about what I have to send
            ! Ia - Compute reception tag = concatenation of (rank+1), ind_group(1), ind_group(2), direction et unique Id.             
            tag = compute_tag(ind_group, tag_velo_range, direction, -proc_gap)
            ! Ib - Receive the message
            call mpi_recv(send_range(1), 2, MPI_INTEGER, send_rank, tag, comm, rece_status, ierr)
            send_range = send_range + proc_gap*N_proc(direction)
            ! II - Send it
            ! IIa - Compute send tag
            tag = compute_tag(ind_group, tag_velo_V, direction, proc_gap)
            ! IIb - Send message
! XXX debug - to delete XXX
            call mpi_Isend(p_V(send_range(1)), send_range(2)-send_range(1)+1, MPI_DOUBLE_PRECISION, &
                    & send_rank, tag, comm, s_request(proc_gap), ierr)
! XXX debug - uncomment the two following line XXX
!            call mpi_Isend(p_V(send_range(1)), send_range(2)-send_range(1)+1, MPI_DOUBLE_PRECISION, &
!                    & send_rank, tag, comm, send_request, ierr)
! XXX end debug XXX
        end if
    end do

    ! Non blocking reception of the velocity field
    ind = 1
    allocate(rece_request(rece_gap(1):rece_gap(2)))
    do proc_gap = rece_gap(1), rece_gap(2)
        if (rece_rank(proc_gap) /= myrank) then
            ! IIa - Compute reception tag
            tag = compute_tag(ind_group, tag_velo_V, direction, -proc_gap)
            ! IIb - Receive message
            rece_range(1) = max(rece_ind_min, gap+1) ! fortran => indice start from 0
            rece_range(2) = min(rece_ind_max, gap+N_proc(direction)) 
            msg_size = rece_range(2)-rece_range(1)+1
            call mpi_Irecv(V_buffer(ind), msg_size, MPI_DOUBLE_PRECISION, rece_rank(proc_gap), tag, comm, &
                        & rece_request(proc_gap), ierr) 
            ind = ind + msg_size
        end if
    end do

    ! -- Compute the interpolated velocity 
    ! Compute the interpolation weight and update the pointers Vp and Vm
    ! Initialisation of reccurence process
    ind = 1
    pos = floor(p_pos_bis(ind))
    weight(ind) = p_pos_bis(ind)-pos
    ! Vm = V(pos)
    proc_gap = floor(real((pos-1)/N_proc(direction))) 
    call mpi_cart_shift(cart_comm, direction-1, proc_gap, rankP, send_rank, ierr)
    if (send_rank == myrank) then
        Vm(ind)%pter => p_V_bis(pos-proc_gap*N_proc(direction))
    else 
        ind_com = ind_com + 1
        Vm(ind)%pter => V_buffer(ind_com)
    end if
    ! Vp = V(pos+1)
    proc_gap = floor(real((pos+1-1)/N_proc(direction))) 
    call mpi_cart_shift(cart_comm, direction-1, proc_gap, rankP, send_rank, ierr)
    if (send_rank == myrank) then
        Vp(ind)%pter => p_V_bis(pos+1-proc_gap*N_proc(direction))
    else 
        ind_com = ind_com + 1
        Vp(ind)%pter => V_buffer(ind_com)
    end if
    pos_old = pos

    ! Following indice : we use previous work (already done)
    do ind = 2, N_proc(direction)
        pos = floor(p_pos_bis(ind))
        weight(ind) = p_pos_bis(ind)-pos
        select case(pos-pos_old)
            case(0)
                ! The particle belongs to the same segment than the previous one
                Vm(ind)%pter => Vm(ind-1)%pter
                Vp(ind)%pter => Vp(ind-1)%pter
            case(1)
                ! The particle follows the previous one
                Vm(ind)%pter => Vp(ind-1)%pter
                ! Vp = V(pos+1)
                proc_gap = floor(real((pos+1-1)/N_proc(direction))) ! fortran -> indice starts from 1
                call mpi_cart_shift(cart_comm, direction-1, proc_gap, rankP, send_rank, ierr)
                if (send_rank == myrank) then
                    Vp(ind)%pter => p_V_bis(pos+1-proc_gap*N_proc(direction))
                else 
                    ind_com = ind_com + 1
                    Vp(ind)%pter => V_buffer(ind_com)
                end if
            case(2)
                ! pos = pos_old +2, wich correspond to "extention"
                ! Vm = V(pos)
                proc_gap = floor(real((pos-1)/N_proc(direction))) 
                call mpi_cart_shift(cart_comm, direction-1, proc_gap, rankP, send_rank, ierr)
                if (send_rank == myrank) then
                    Vm(ind)%pter => p_V_bis(pos-proc_gap*N_proc(direction))
                else 
                    ind_com = ind_com + 1
                    Vm(ind)%pter => V_buffer(ind_com)
                end if
                ! Vp = V(pos+1)
                proc_gap = floor(real((pos+1-1)/N_proc(direction))) 
                call mpi_cart_shift(cart_comm, direction-1, proc_gap, rankP, send_rank, ierr)
                if (send_rank == myrank) then
                    Vp(ind)%pter => p_V_bis(pos+1-proc_gap*N_proc(direction))
                else 
                    ind_com = ind_com + 1
                    Vp(ind)%pter => V_buffer(ind_com)
                end if
            case default
                print*, "unexpected case : pos = ", pos, " , pos_old = ", pos_old, " ind = ", ind
        end select
        pos_old = pos
    end do

    ! -- Compute the interpolate velocity --
    ! Check if communication are done
    do proc_gap = rece_gap(1), rece_gap(2)
        if (rece_rank(proc_gap)/=myrank) then
            call mpi_wait(rece_request(proc_gap), rece_status, ierr)
        end if
    end do


    ! Then compute the field
    do ind = 1, N_proc(direction)
        p_V(ind) = weight(ind)*Vp(ind)%pter + (1-weight(ind))*Vm(ind)%pter
    end do

! XXX debug - to delete XXX
    do ind = 1, N_proc(direction)
        nullify(Vp(ind)%pter)
        nullify(Vm(ind)%pter)
    end do

    do proc_gap = send_gap(1), send_gap(2)
        call mpi_cart_shift(cart_comm, direction-1, proc_gap, rankP, send_rank, ierr)
        if (send_rank /= myrank) then
            call MPI_WAIT(s_request(proc_gap),rece_status,ierr)
        end if
    end do
    deallocate(s_request)
    do proc_gap = rece_gap(1), rece_gap(2)
        if (rece_rank(proc_gap) /= myrank) then
            call MPI_WAIT(s_request_bis(proc_gap),rece_status,ierr)
        end if
    end do
    deallocate(s_request_bis)
! XXX end debug XXX

    ! Deallocation    
    deallocate(rece_rank)
    deallocate(rece_request)
    deallocate(V_buffer)

end subroutine AC_particle_velocity


!> Determine the set of processes wich will send me information during the velocity interpolation.
!!    @param[in]    direction       = current direction (1 = along X, 2 = along Y, 3 = along Z)
!!    @param[in]    comm            = mpi communicator associated with the current direction
!!    @param[in]    ind_group       = coordinate of the current group of lines
!!    @param[in]    rece_ind_min    = minimal indice of mesh involved in remeshing particles (of the my local subdomains)
!!    @param[in]    rece_ind_max    = maximal indice of mesh involved in remeshing particles (of the my local subdomains)
!!    @param[out]   send_gap        = gap between my coordinate and the processes of minimal coordinate which will send information to me
!!    @param[out]   rece_gap        = gap between my coordinate and the processes of maximal coordinate which will receive information from me
!! @details
!!    Obtain the list of processus wich need a part of my local velocity field
!!    to interpolate the velocity used in the RK2 scheme to advect its particles.
subroutine AC_obtain_receivers(direction, comm, ind_group, rece_ind_min, rece_ind_max, send_gap, rece_gap)
! XXX Work only for periodic condition.

    use cart_topology
    use mpi

    integer, intent(in)                 :: rece_ind_min, rece_ind_max
    integer, intent(in)                 :: direction, comm
    integer, dimension(2), intent(in)               :: ind_group
    integer, dimension(2), intent(out)  :: rece_gap, send_gap
    integer, dimension(MPI_STATUS_SIZE) :: statut
    
    integer                             :: proc_gap         ! gap between a processus coordinate (along the current 
                                                            ! direction) into the mpi-topology and my coordinate
    integer                             :: rece_gapP        ! gap between the coordinate of the previous processus (in the current direction)
                                                            ! and the processes of maximal coordinate which will receive information from it
    integer                             :: rece_gapN        ! same as above but for the next processus
    integer                             :: rankP, rankN     ! processus rank for shift (P= previous, N = next)
    integer                             :: tag_min, tag_max ! mpi message tag (for communicate rece_proc(1) and rece_proc(2))
    integer                             :: send_request     ! mpi status of nonblocking send
    integer                             :: send_request_bis ! mpi status of nonblocking send
    integer                             :: ierr             ! mpi error code
    integer                             :: tag              ! mpi message tag
    integer, dimension(2)               :: tag_table        ! some mpi message tag
! XXX debug - to delete XXX
    integer                             :: pr
    logical, dimension(:,:), allocatable:: test_request 
    integer, dimension(:,:), allocatable:: s_request 
! XXX end debug XXX

    tag_min = 5
    tag_max = 6

    send_gap = 3*N(direction)

    rece_gap(1) = floor(real(rece_ind_min-1)/N_proc(direction))
    rece_gap(2) = floor(real(rece_ind_max-1)/N_proc(direction))
    
    ! Communicate with my neigbors
    ! Compute their rank
    call mpi_cart_shift(cart_comm, (direction-1), 1, rankP, rankN, ierr)
    ! Inform that about processus from which I need information
    tag_table = compute_tag(ind_group, tag_obtrec_ghost_NP, direction)
    call mpi_Isend(rece_gap(1), 1, MPI_INTEGER, rankP, tag_table(1), comm, send_request, ierr)
    call mpi_Isend(rece_gap(2), 1, MPI_INTEGER, rankN, tag_table(2), comm, send_request_bis, ierr)
    ! Receive the same message form my neighbors
    call mpi_recv(rece_gapN, 1, MPI_INTEGER, rankN, tag_table(1), comm, statut, ierr)
    call mpi_recv(rece_gapP, 1, MPI_INTEGER, rankP, tag_table(2), comm, statut, ierr)


    ! Send
! XXX debug - to delete XXX
    allocate(s_request(rece_gap(1):rece_gap(2),2))
    allocate(test_request(rece_gap(1):rece_gap(2),2))
    test_request = .false.
! XXX end debug XXX
    do proc_gap = rece_gap(1), rece_gap(2)
        ! Compute the rank of the target processus
        call mpi_cart_shift(cart_comm, (direction-1), proc_gap, rankP, rankN, ierr)
        ! Determine if I am the the first or the last processes (considering the current directory) 
            ! to require information from this processus
        if (proc_gap>rece_gapP-1) then
            if(rankN /= myrank) then
                tag_table = compute_tag(ind_group, tag_obtrec_NP, direction)
! XXX debug - to delete XXX
                call mpi_Isend(-proc_gap, 1, MPI_INTEGER, rankN, tag_table(1), comm, s_request(proc_gap,1), ierr)
                test_request(proc_gap,1) = .true.
! XXX uncomment following line XXX
!                call mpi_Isend(-proc_gap, 1, MPI_INTEGER, rankN, tag_table(1), comm, send_request, ierr)
! XXX end debug XXX
            else
                send_gap(1) = -proc_gap
            end if
        end if
        if (proc_gap<rece_gapN+1) then
            if(rankN /= myrank) then
! XXX debug - to delete XXX
                test_request(proc_gap,2) = .true.
                call mpi_Isend(-proc_gap, 1, MPI_INTEGER, rankN, tag_table(2), comm, s_request(proc_gap,2), ierr)
! XXX uncomment following line XXX
!                call mpi_Isend(-proc_gap, 1, MPI_INTEGER, rankN, tag_table(2), comm, send_request_bis, ierr)
! XXX end debug XXX
            else
                send_gap(2) = -proc_gap
            end if
        end if
    end do


    ! Receive
    if (send_gap(1) == 3*N(direction)) then
        call mpi_recv(send_gap(1), 1, MPI_INTEGER, MPI_ANY_SOURCE, tag_table(1), comm, statut, ierr)
    end if
    if (send_gap(2) == 3*N(direction)) then
        call mpi_recv(send_gap(2), 1, MPI_INTEGER, MPI_ANY_SOURCE, tag_table(2), comm, statut, ierr)
    end if


! XXX debug - to delete XXX
    call MPI_WAIT(send_request,statut,ierr)
    call MPI_WAIT(send_request_bis,statut,ierr)
    do proc_gap = rece_gap(1), rece_gap(2)
        if (test_request(proc_gap,1).eqv. .true.) call MPI_WAIT(s_request(proc_gap,1),statut,ierr)
        if (test_request(proc_gap,2)) call MPI_WAIT(s_request(proc_gap,2),statut,ierr)
    end do
    deallocate(s_request)
    deallocate(test_request)
! XXX end debug XXX

end subroutine AC_obtain_receivers


! ===================================================================================================
! ====================     Others than velocity interpolation and remeshing      ====================
! ===================================================================================================

!> Determine type (center or left) of each block and tag particle to know where
!! corrected remeshing formula are recquired.
!!    @param[in]        dt          = time step
!!    @param[in]        direction   = current direction (1 = along X, 2 = along Y and 3 = along Z)
!!    @param[in]        comm        = mpi communicator associated to the current direction (Y_comm, Z_comm)
!!    @param[in]        ind_group   = coordinate of the current group of lines
!!    @param[in]        p_V         = particle velocity (along the current direction)
!!    @param[out]       bl_type     = table of blocks type (center of left)
!!    @param[out]       bl_tag      = inform about tagged particles (bl_tag(ind_bl)=1 if the end of the bl_ind-th block 
!!                                    and the begining of the following one is tagged)
subroutine AC_type_and_block(dt, direction, comm, ind_group, p_V, &
                    & bl_type, bl_tag)

    use mpi
    use cart_topology
    use precision

    real(WP), intent(in)                                    :: dt           ! time step
    integer, intent(in)                                     :: direction
    integer, intent(in)                                     :: comm
    integer, dimension(2), intent(in)                       :: ind_group
    real(WP), dimension(:), intent(in)                      :: p_V
    integer, dimension(bl_number(direction)+1), intent(out) :: bl_type      ! is the particle block a center block or a left one ?
    integer, dimension(bl_number(direction)), intent(out)   :: bl_tag       ! indice of tagged particles

    real(WP), dimension(bl_number(direction)+1)             :: bl_lambdaMin ! for a particle, lamda = V*dt/dx ;  bl_lambdaMin = min of 
                                                                            ! lambda on a block (take also into account first following particle)
    real(WP)                                                :: lambP, lambN ! buffer to exchange some lambda min with other processus
    integer, dimension(bl_number(direction)+1)              :: bl_ind       ! block index : integer as lambda in (bl_ind,bl_ind+1) for a left block
                                                                            ! and lambda in (bl_ind-1/2, bl_ind+1/2) for a right block 
    integer                                                 :: ind, ind2,i_p! some indices
    real(WP)                                                :: cfl          ! = d_sc
    integer                                                 :: rankP, rankN ! processus rank for shift (P= previous, N = next)
    integer, dimension(2)                                   :: send_request ! mpi status of nonblocking send
    integer, dimension(2)                                   :: rece_request ! mpi status of nonblocking receive
    integer, dimension(MPI_STATUS_SIZE)                     :: rece_status  ! mpi status (for mpi_wait)
    integer                                                 :: tag          ! tag for mpi message
    integer, dimension(2)                                   :: tag_table    ! other tags for mpi message
    integer                                                 :: ierr         ! mpi error code

    ! ===== Initialisation =====
    cfl = dt/d_sc(direction)

    ! ===== Compute bl_lambdaMin =====
    ! -- Compute rank of my neighbor --
    call mpi_cart_shift(cart_comm, (direction-1), 1, rankP, rankN, ierr)

    ! -- For the first block (1/2) --
    ! The domain contains only its second half => exchange ghost with the previous processus
    bl_lambdaMin(1) = p_V(1)*cfl
    do i_p = 2, (bl_size/2)+1
        bl_lambdaMin(1) = min(bl_lambdaMin(1), p_V(i_p)*cfl)
    end do
    tag_table = compute_tag(ind_group, tag_part_tag_NP, direction)
    ! Send message
    call mpi_Isend(bl_lambdaMin(1), 1, MPI_DOUBLE_PRECISION, rankP, tag_table(1), comm, send_request(1), ierr)
    ! Receive it
    call mpi_Irecv(lambN, 1, MPI_DOUBLE_PRECISION, rankN, tag_table(1), comm, rece_request(1), ierr) 

    ! -- For the last block (1/2) --
    ! The processus contains only its first half => exchange ghost with the next processus
    ind = bl_number(direction) + 1
    bl_lambdaMin(ind) = p_V(N_proc(direction))*cfl
    do i_p = N_proc(direction) - (bl_size/2)+2, N_proc(direction)-1
            bl_lambdaMin(ind) = min(bl_lambdaMin(ind), p_V(i_p)*cfl)
    end do
    ! Send message
    call mpi_Isend(bl_lambdaMin(ind), 1, MPI_DOUBLE_PRECISION, rankN, tag_table(2), comm, send_request(2), ierr)
    ! Receive it
    call mpi_Irecv(lambP, 1, MPI_DOUBLE_PRECISION, rankP, tag_table(2), comm, rece_request(2), ierr) 

    ! -- For the "middle" block --
    do ind = 2, bl_number(direction)
        i_p = ((ind-1)*bl_size) + 1 - bl_size/2
        bl_lambdaMin(ind) = p_V(i_p)*cfl
        do ind2 = 2, bl_size+1
            i_p = ((ind-1)*bl_size) + ind2 - bl_size/2
            bl_lambdaMin(ind) = min(bl_lambdaMin(ind), p_V(i_p)*cfl)
        end do
    end do

    ! -- For the first block (1/2) --
    ! The domain contains only its second half => use exchanged ghost
    ! Check reception
    call mpi_wait(rece_request(2), rece_status, ierr)
    bl_lambdaMin(1) = min(bl_lambdaMin(1), lambP)

    ! -- For the last block (1/2) --
    ! The processus contains only its first half => use exchanged ghost
    ! Check reception
    call mpi_wait(rece_request(1), rece_status, ierr)
    ind = bl_number(direction) + 1
    bl_lambdaMin(ind) = min(bl_lambdaMin(ind), lambN)

    ! ===== Compute block type and index =====
    do ind = 1, bl_number(direction) + 1
        bl_ind(ind) = nint(bl_lambdaMin(ind))
        if (bl_lambdaMin(ind)<bl_ind(ind)) then
            ! => center type
            bl_type(ind) = 1
        else
            ! => left type
            bl_type(ind) = 0
        end if
    end do

    ! ===== Tag particles =====

    do ind = 1, bl_number(direction)
        if ((bl_ind(ind)/=bl_ind(ind+1)) .and. (bl_type(ind)/=bl_type(ind+1))) then
            ! the end of ind-th block and the beginning of the following one are tagged
            bl_tag(ind) = 1
        else
            bl_tag(ind) = 0
        end if
    end do

end subroutine AC_type_and_block


! ===================================================================
! ====================     Remesh particles      ====================
! ===================================================================

!> Determine the set of processes wich will send me information during the
!!  scalar remeshing.
!!    @param[in]    send_j_min  = minimal indice of mesh involved in remeshing particles (of the particles in my local subdomains)
!!    @param[in]    send_j_max  = maximal indice of mesh involved in remeshing particles (of the particles in my local subdomains)
!!    @param[in]    direction   = current direction (1 = along X, 2 = along Y, 3 = along Z)
!!    @param[in]    comm        = mpi communicator associated with the current direction
!!    @param[in]    ind_group   = coordinate of the current group of lines
!!    @param[out]   proc_min    = gap between my coordinate and the processes of minimal coordinate which will receive information from me
!!    @param[out]   proc_max    = gap between my coordinate and the processes of maximal coordinate which will receive information from me
!!    @param[out]   rece_proc   = coordinate range of processes which will send me information during the remeshing.
!! @details
!!    Obtain the list of processus which contains some particles which belong to
!!    my subdomains after their advection (and thus which will be remeshing into
!!    my subdomain). This result is return as an interval [send_min; send_max].
!!    All the processus whose coordinate (into the current direction) belong to 
!!    this segment are involved into scalar remeshing into the current
!!    subdomains.
subroutine AC_obtain_senders(send_j_min, send_j_max, direction, comm, ind_group, proc_min, proc_max, rece_proc)
! XXX Work only for periodic condition. For dirichlet conditions : it is
! possible to not receive either rece_proc(1), either rece_proc(2) or none of
! these two => detect it (track the first and the last particles) and deal with it.

    use cart_topology
    use mpi

    integer, intent(in)                 :: send_j_min, send_j_max
    integer, intent(in)                 :: direction, comm
    integer, dimension(2), intent(in)   :: ind_group
    integer(kind=4), intent(out)        :: proc_min, proc_max
    integer, dimension(2), intent(out)  :: rece_proc
    integer, dimension(MPI_STATUS_SIZE) :: statut
    
    integer(kind=4)                     :: proc_gap         ! gap between a processus coordinate (along the current 
                                                            ! direction) into the mpi-topology and my coordinate
    integer                             :: rankP, rankN     ! processus rank for shift (P= previous, N = next)
    integer, dimension(2)               :: tag_table        ! mpi message tag (for communicate rece_proc(1) and rece_proc(2))
    integer                             :: send_request     ! mpi status of nonblocking send
    integer                             :: ierr             ! mpi error code

    tag_table = compute_tag(ind_group, tag_obtsend_NP, direction)

    rece_proc = 3*N(direction)

    proc_min = floor(real(send_j_min-1)/N_proc(direction))
    proc_max = floor(real(send_j_max-1)/N_proc(direction))
    
    ! Send
    do proc_gap = proc_min, proc_max
        ! Compute the rank of the target processus
        call mpi_cart_shift(cart_comm, (direction-1), proc_gap, rankP, rankN, ierr)
        ! Determine if I am the the first or the last processes (considering my
                ! coordinate along the current directory) to send information to
                ! one of these processes.
                ! Note that local indice go from 1 to N_proc (fortran).
        ! I am the first ?
        if ((send_j_min< +1-2*bl_bound_size + proc_gap*N_proc(direction)+1).AND. &
                    & (send_j_max>= proc_gap*N_proc(direction))) then
            if(rankN /= myrank) then
                call mpi_Isend(-proc_gap, 1, MPI_INTEGER, rankN, tag_table(1), comm, send_request, ierr)
            else
                rece_proc(1) = -proc_gap
            end if
        end if
        ! I am the last ?
        if ((send_j_max > -1+2*bl_bound_size + (proc_gap+1)*N_proc(direction)) &
                    & .AND.(send_j_min<= (proc_gap+1)*N_proc(direction))) then
            if(rankN /= myrank) then
                call mpi_Isend(-proc_gap, 1, MPI_INTEGER, rankN, tag_table(2), comm, send_request, ierr)
            else
                rece_proc(2) = -proc_gap
            end if
        end if
    end do


    ! Receive
    if (rece_proc(1) == 3*N(direction)) then
        call mpi_recv(rece_proc(1), 1, MPI_INTEGER, MPI_ANY_SOURCE, tag_table(1), comm, statut, ierr)
    end if
    if (rece_proc(2) == 3*N(direction)) then
        call mpi_recv(rece_proc(2), 1, MPI_INTEGER, MPI_ANY_SOURCE, tag_table(2), comm, statut, ierr)
    end if

end subroutine AC_obtain_senders

!> Common procedure for remeshing wich perform all the communcation and provide
!! the update scalar field.
!!    @param[in]        direction   = current direction (1 = along X, 2 = along Y and 3 = along Z)
!!    @param[in]        comm        = mpi communicator associated to the current direction (Y_comm, Z_comm)
!!    @param[in]        ind_group   = coordinate of the current group of lines
!!    @param[in]        send_i_min  = minimal indice of the send buffer
!!    @param[in]        send_i_max  = maximal indice of the send buffer
!!    @param[in]        send_buffer = buffer use to remesh the scalar before to send it to the right subdomain
!!    @param[in,out]    scal1D      = mono-dimensionnal scalar field to advect
!!                                  (XXX todo : decide if extraction from the 3D or pointeur)
!! @details
!!    Remeshing are done in a local buffer. This subroutine distribute this buffer 
!!    to the right processes, receive the buffer send and update the scalar field.
subroutine AC_bufferToScalar(direction, comm, ind_group, send_i_min, send_i_max, send_buffer, scal1D)

    use cart_topology
    use mpi

    integer, intent(in)                                                 :: direction
    integer, intent(in)                                                 :: comm
    integer, dimension(2), intent(in)                                   :: ind_group
    integer, intent(in)                                                 :: send_i_min
    integer, intent(in)                                                 :: send_i_max
    real(WP), dimension(send_i_min:send_i_max), intent(in)              :: send_buffer
    !XXX - ca ne devrait pas bugger sans les range
    real(WP), dimension(N_proc(direction)), intent(inout)               :: scal1D

! XXX todo vrai scalaire = 3D -> choisir si on envoie un sous-tableau 1D du
! scalaire ou un tableau de pointeur 1D qui pointe vers les bonnes cases du
! champ scalaire

    ! Variables used to communicate between subdomains. A variable prefixed by "send_"(resp "rece")
    ! design something I send (resp. I receive).
    integer                             :: i            ! table indice
    integer                             :: proc_gap     ! gap between my Y-coordinate and the one of the processus 
    integer, dimension(2)               :: rece_proc    ! minimal and maximal gap between my Y-coordinate and the one from which 
                                                        ! I will receive data
    integer                             :: proc_min     ! smaller gap between me and the processes to where I send data
    integer                             :: proc_max     ! smaller gap between me and the processes to where I send data
    real(WP), dimension(:), allocatable :: rece_buffer  ! buffer use to stock received scalar field
    integer                             :: send_gap     ! number of mesh between my and another processus
    integer,dimension(:,:), allocatable :: rece_range   ! range of (local) indice where the received scalar field has to be save
    integer, dimension(2)               :: send_range   ! range of (local) indice where the send scalar field has to be save
    integer, dimension(:), allocatable  :: rece_request ! mpi communication request (handle) of nonblocking receive
    integer, dimension(:), allocatable  :: rece_rank    ! rank of processus from wich I receive data
    integer                             :: send_rank    ! rank of processus to which I send data
    integer                             :: rankP        ! rank used in mpi_cart_shift
    integer, dimension(MPI_STATUS_SIZE) :: rece_status  ! mpi status (for mpi_wait)
    integer                             :: send_request ! mpi status of nonblocking send
    integer                             :: rece_i_min   ! the minimal indice from where belong the scalar field I receive
    integer                             :: rece_i_max   ! the maximal indice from where belong the scalar field I receive
    integer                             :: ierr         ! mpi error code
    integer                             :: comm_size    ! number of element to send/receive
    integer                             :: tag          ! mpi message tag
                                                        ! with wich I communicate.


    ! Determine the communication needed : who will communicate whit who ? (ie compute sender and recer)
    call AC_obtain_senders(send_i_min, send_i_max, direction, comm, ind_group, proc_min, proc_max, rece_proc)
    ! Send the information
    do proc_gap = proc_min, proc_max
            ! Compute the rank of the target processus
            call mpi_cart_shift(cart_comm, direction-1, proc_gap, rankP, send_rank, ierr)
            send_gap = proc_gap*N_proc(direction)
            send_range(1) = max(send_i_min, send_gap+1) ! fortran => indice start from 0
            send_range(2) = min(send_i_max, send_gap+N_proc(direction)) 
        if (send_rank/=myrank) then
            ! Determine quantity of information to send
            comm_size = send_range(2)-send_range(1)+1
            ! Send the range of the scalar field send
            tag = compute_tag(ind_group, tag_bufToScal_range, direction, proc_gap)
            call mpi_Isend(send_range(1), 2, MPI_INTEGER, send_rank, tag, comm, send_request, ierr)
            ! And send the buffer
            tag = compute_tag(ind_group, tag_bufToScal_buffer, direction, proc_gap)
            call mpi_Isend(send_buffer(send_range(1)),comm_size, MPI_DOUBLE_PRECISION, send_rank, &
                        & tag, comm, send_request, ierr)
        else
            ! I have to distribute the buffer in myself
            do i = send_range(1), send_range(2)
                scal1D(i-send_gap) = scal1D(i-send_gap) + send_buffer(i)
            end do
        end if
    end do

    ! Receive information
    ! Allocate field
    allocate(rece_rank(rece_proc(1):rece_proc(2)))
    allocate(rece_range(2,rece_proc(1):rece_proc(2)))  ! be careful that mpi use contiguous memory element
    allocate(rece_request(rece_proc(1):rece_proc(2)))
    ! Receive range
    do proc_gap = rece_proc(1), rece_proc(2)
        call mpi_cart_shift(cart_comm, direction-1, proc_gap, rankP, rece_rank(proc_gap), ierr)
        if (rece_rank(proc_gap)/=myrank) then
            tag = compute_tag(ind_group, tag_bufToScal_range, direction, -proc_gap)
            call mpi_Irecv(rece_range(1,proc_gap), 2, MPI_INTEGER, rece_rank(proc_gap), tag, comm, &
                        & rece_request(proc_gap), ierr) ! we use tag = source rank
        end if
    end do
    ! Check reception
    do proc_gap = rece_proc(1), rece_proc(2)
        if (rece_rank(proc_gap)/=myrank) then
            call mpi_wait(rece_request(proc_gap), rece_status, ierr)
        end if
    end do
    deallocate(rece_request)
    ! Receive buffer and remesh it
        ! XXX Possible optimisation : an optimal code will
        !   1 - have non-blocking reception of scalar buffers
        !   2 - check when a reception is done and then update the scalar
        !   3 - iterate step 2 until all message was rece and that the scalar
        !       field was update with all the scalar buffers
    do proc_gap = rece_proc(1), rece_proc(2)
        if (rece_rank(proc_gap)/=myrank) then
            rece_i_min = rece_range(1,proc_gap)
            rece_i_max = rece_range(2,proc_gap)
            ! Receive information
            comm_size=(rece_i_max-rece_i_min+1)
            allocate(rece_buffer(rece_i_min:rece_i_max)) ! XXX possible optimisation
                ! by allocating one time to the max size, note that the range use in
                ! this allocation instruction is include in (1, N_proc(2))
            tag = compute_tag(ind_group, tag_bufToScal_buffer, direction, -proc_gap)
            call mpi_recv(rece_buffer(rece_i_min), comm_size, MPI_DOUBLE_PRECISION, &
                    & rece_rank(proc_gap), tag, Y_comm, rece_status, ierr)
            ! Update the scalar field
            send_gap = proc_gap*N_proc(direction)
            scal1D(rece_i_min+send_gap:rece_i_max+send_gap) = scal1D(rece_i_min+send_gap:rece_i_max+send_gap) &
                & + rece_buffer(rece_i_min:rece_i_max)
            deallocate(rece_buffer)
        end if
    end do

    deallocate(rece_range)
    deallocate(rece_rank)

end subroutine AC_bufferToScalar


!> Left remeshing formula of order 2
!!      @param[in]       pos_adim= adimensionned particle position
!!      @param[in]       sca     = scalar advected by the particle
!!      @param[in,out]   buffer  = temporaly remeshed scalar field
subroutine AC_remesh_left(pos_adim, sca, buffer)

    use cart_topology
    
    real(WP), intent(in)                                        :: pos_adim, sca
    real(WP), dimension(send_j_min:send_j_max), intent(inout)   :: buffer

    integer     :: j0                       ! indice of the the nearest mesh points
    real(WP)    :: bM, b0, bP               ! interpolation weight for the particles
    real(WP)    :: yM, y0, yP               ! adimensionned distance to mesh points
    ! Mesh point used in remeshing formula
    j0 = floor(pos_adim)
    !j0 = floor(pos/d_sc(2))
            
    ! Distance to mesh points
    y0 = (pos_adim - float(j0))
    !y0 = (pos - float(j0)*d_sc(2))/d_sc(2)
    yM = y0-1
    yP = y0+1

    ! Interpolation weights
    bM=0.5*y0*yM
    b0=1.-y0**2
    bP=0.5*y0*yP

    ! remeshing
    buffer(j0-1) = buffer(j0-1)   + bM*sca
    buffer(j0)   = buffer(j0)     + b0*sca
    buffer(j0+1) = buffer(j0+1)   + bP*sca

end subroutine AC_remesh_left

!> Centered remeshing formula of order 2
!!      @param[in]       pos_adim= adimensionned particle position
!!      @param[in]       sca     = scalar advected by the particle
!!      @param[in,out]   buffer  = temporaly remeshed scalar field
subroutine AC_remesh_center(pos_adim, sca, buffer)

    use cart_topology

    real(WP), intent(in)                                        :: pos_adim, sca
    real(WP), dimension(send_j_min:send_j_max), intent(inout)   :: buffer

    integer     :: j0                       ! indice of the the nearest mesh points
    real(WP)    :: bM, b0, bP               ! interpolation weight for the particles
    real(WP)    :: yM, y0, yP               ! adimensionned distance to mesh points

    j0 = nint(pos_adim)
    !j0 = nint(pos/d_sc(2))

    ! Distance to mesh points
    y0 = (pos_adim - float(j0))
    !y0 = (pos - float(j0)*d_sc(2))/d_sc(2)
    yM = y0-1
    yP = y0+1

    ! Interpolation weights
    bM=0.5*y0*yM
    b0=1.-y0**2
    bP=0.5*y0*yP

    ! remeshing
    buffer(j0-1) = buffer(j0-1)   + bM*sca
    buffer(j0)   = buffer(j0)     + b0*sca
    buffer(j0+1) = buffer(j0+1)   + bP*sca
            
end subroutine AC_remesh_center


!> Corrected remeshing formula for transition from Centered block to a Left block with a different indice (tagged particles)
!!    @param[in]       pos_adim= adimensionned particle position
!!    @param[in]       sca     = scalar advected by this particle
!!    @param[in]       posP_ad = adimensionned position of the second particle
!!    @param[in]       scaP    = scalar advected by this particle
!!    @param[in,out]   buffer  = temporaly remeshed scalar field
!! @details 
!!    Remeshing formula devoted to tagged particles.
!!    The particle group send into argument is composed of a block end and of the 
!!    begining of the next block. The first particles belong to a centered block
!!    and the last to a left one. The block have difference indice (tagged
!!    particles) and we have to use corrected formula.
subroutine AC_remesh_tag_CL(pos_adim, sca, posP_ad, scaP, buffer)

    use cart_topology

    real(WP), intent(in)                                        :: pos_adim, sca, posP_ad, scaP
    real(WP), dimension(send_j_min:send_j_max), intent(inout)   :: buffer

    integer     :: jM, j0, jP               ! indice of the the nearest mesh points
                                            ! (they depend on the block type)
    integer     :: j0_bis                   ! indice of the the nearest mesh point for the indP=ind+1 particle
    real(WP)    :: aM, a0, bP, b0           ! interpolation weight for the particles
    real(WP)    :: yM, y0, y0_bis, yP_bis   ! adimensionned distance to mesh points

    j0 = nint(pos_adim)
    !j0 = nint(pos/d_sc(2))
    j0_bis = floor(posP_ad)
    !j0_bis = floor(posP/d_sc(2))
    jM=j0-1
    jP=j0+1

    y0 = (pos_adim - float(j0))
    !y0 = (pos - float(j0)*d_sc(2))/d_sc(2)
    y0_bis = (posP_ad - float(j0_bis))
    !y0_bis = (posP - float(j0_bis)*d_sc(2))/d_sc(2)
    yM=y0-1.
    yP_bis=y0_bis+1

    aM=0.5*y0*yM
    a0=1.-aM
    bP=0.5*y0_bis*yP_bis
    b0=1.-bP

    ! Remeshing
    buffer(jM)=buffer(jM)+aM*sca
    buffer(j0)=buffer(j0)+a0*sca+b0*scaP
    buffer(jP)=buffer(jP)+bP*scaP

end subroutine AC_remesh_tag_CL


!> Corrected remeshing formula for transition from Left block to a Centered  block with a different indice (tagged particles)
!!    @param[in]       pos_adim= adimensionned particle position
!!    @param[in]       sca     = scalar advected by this particle
!!    @param[in]       posP_ad = adimensionned position of the second particle
!!    @param[in]       scaP    = scalar advected by this particle
!!    @param[in,out]   buffer  = temporaly remeshed scalar field
!! @details 
!!    Remeshing formula devoted to tagged particles.
!!    The particle group send into argument is composed of a block end and of the 
!!    begining of the next block. The first particles belong to a left block
!!    and the last to a centered one. The block have difference indice (tagged
!!    particles) and we have to use corrected formula.
subroutine AC_remesh_tag_LC(pos_adim, sca, posP_ad, scaP, buffer)

    use cart_topology

    real(WP), intent(in)                                        :: pos_adim, sca, posP_ad, scaP
    real(WP), dimension(send_j_min:send_j_max), intent(inout)   :: buffer

    integer     :: jM, j0, jP, jP2, jP3             ! indice of the the nearest mesh points
                                                    ! (they depend on the block type)
    integer     :: j0_bis                           ! indice of the the nearest mesh point for the indP=ind+1 particle
    real(WP)    :: aM, a0, aP,aP2, b0, bP, bP2, bP3 ! interpolation weight for the particles
    real(WP)    :: yM,y0,yP, yM_bis,y0_bis,yP_bis   ! adimensionned distance to mesh points


    ! Indice of mesh point used in order to remesh
    j0 = floor(pos_adim)
    !j0 = floor(pos/d_sc(2))
    j0_bis = nint(posP_ad)
    !j0_bis = nint(posP/d_sc(2))
    jM=j0-1
    jP=j0+1
    jP2=j0+2
    jP3=j0+3
    
    ! Distance to mesh point
    y0 = (pos_adim - float(j0))
    !y0 = (pos - float(j0)*d_sc(2))/d_sc(2)
    y0_bis = (posP_ad - float(j0_bis))
    !y0_bis = (posP - float(j0_bis)*d_sc(2))/d_sc(2)
    yP=y0+1
    yM=y0-1
    yP_bis=y0_bis+1
    yM_bis=y0_bis-1
    
    ! Interpolation weight
    aM=y0*yM/2.
    a0=1-y0**2
    aP=y0
    aP2=y0*yM/2.
    b0=y0_bis*yP_bis/2.
    bP=-y0_bis
    bP2=1-y0_bis**2
    bP3=y0_bis*yP_bis/2.

    ! Remeshing
    buffer(jM)= buffer(jM)+aM*sca
    buffer(j0)= buffer(j0)+a0*sca+b0*scaP
    buffer(jP)= buffer(jP)+aP*sca+bP*scaP
    buffer(jP2)=buffer(jP2)+aP2*sca+bP2*scaP
    buffer(jP3)=buffer(jP3)+bP3*scaP

end subroutine AC_remesh_tag_LC

end module advec_common
!! @}
