!------------------------------------------------------------------------------
!
! MODULE: test_advection
!
! DESCRIPTION: 
!> Validation test for advection method.
!!
!! @details
!! This module provide different test to validate the transport solver. 
!! All these test are unit test : they return a logical value to check if 
!! the code version pass it or not. 
!!
!! That is all these test are logical function, they return true if the result
!! is the right one and false otherwise. 
!! All the "test_part_*" function are devoted to validate the particular solver
!! The following test are included :
!!      A - Validate the particular method, step by step
!!      1 -> Test the procedure "AC_obtain_senders" from advec_common
!!      2 -> Validate the redistribution the buffer during the remeshing (XXX todo)
!!      3 -> Validate the remeshing of untagged particles
!!      4 -> Validate the remeshing of tagged particles (XXX todo)
!!      B - Validate an advection solver (XXX todo)
!!      1 -> advec a ball with a constant velocity
!!      2 -> advec a ball with a spheric velocity field (the ball turns)
!!
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

module advec_aux

    use string
    use precision
    implicit none

    real(WP), private   :: epsilon_error = 1e-4     ! Error tolerance


    ! Public procedures

    ! ===== Test for the particles solver =====
    ! Public function
    public              :: test_part_remesh_no_tag
    public              :: test_part_init
    public              :: test_part_advecO2
    public              :: test_advecY


    ! ===== Generic test for an advection solver =====

    ! Private procedure

contains

!> Particle method: validation of particle creation
!!     @param[in] direction = to select direction to test
!!     @param[in] init_scal = optional parameter to select initialisation
!!     @return      error   = test error (= false if the code pass the test) (= not success)
function test_part_init (direction, init_scal) result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use advecY
    use advec_common
    use cart_topology
    ! Test procedures
    use advec_aux_init
    use test_common

    logical                                     :: success
    character(str_short), intent(in), optional  :: init_scal
    integer, intent(in)                         :: direction        ! current direction

    character(str_short)                        :: initialisation   ! to choose how to initialise the scalar field
    real(WP), dimension(:, :, :), allocatable   :: scalar           ! the scalar field
    real(WP), dimension(:), allocatable         :: p_pos, p_V       ! the particles position and the scalar they advect
    real(WP)                                    :: velocity         ! constant velocity of the flux

    if(present(init_scal)) then
        initialisation = init_scal
    else 
        initialisation = 'constant'
    end if
    success = .true.
    call test_substatus('test of particle initialisation !', success, myrank)

end function test_part_init 





!> Particles method: validation of the remeshing of untagged particles
!!     @param[in] init_scal =  optional parameter to initialise the scalar fied to
!!                                  a constant one or to a sphere shape
!!     @return      error   = test error (= false if the code pass the test) (= not success)
function test_part_remesh_no_tag (init_scal) result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use advecY
    use advec_common
    use cart_topology
    ! Test procedures
    use advec_aux_init
    use test_common

    logical                                     :: success
    character(len=*), intent(in), optional      :: init_scal

    character(len=17)                           :: initialisation   ! to choose how to initialise the scalar field
    character(len=str_short)                    :: order            ! space order of the solveur
    real(WP), dimension(:, :, :), allocatable   :: scalar           ! the scalar field
    real(WP), dimension(:), allocatable         :: p_pos_adim, p_SC ! the adimensionned particles position and the scalar they advect
    integer, dimension(:), allocatable          :: bl_type, bl_tag  ! type and tag of each particle bloc
    real(WP)                                    :: velocity         ! constant velocity of the flux
    integer                                     :: nb_proc          ! number of processes
    integer                                     :: ierr             ! mpi error code
    integer                                     :: i,j,k            ! mesh indice   
    integer, dimension(2)                       :: ind_group        ! indice of the current group of line
    integer                                     :: T_step           ! time
    integer                                     :: T_end            ! final time
    real(WP), dimension(:,:,:), allocatable     :: good_scal        ! analytic solution
    real(WP), dimension(3)                      :: translat         ! to compute analytic solution

    ! Initialize the particular solver
    order = 'p_O2'
    call advec_init(order)

    allocate(scalar(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_scal(N_proc(1), N_proc(2), N_proc(3)))
    allocate(p_pos_adim(N_proc(2)))
    allocate(p_SC(N_proc(2)))
    allocate(bl_type(1+N_proc(2)/bl_size))
    allocate(bl_tag((N_proc(2)/bl_size)))

    if(present(init_scal)) then
        initialisation = init_scal
    else 
        initialisation = 'constant'
    end if
    success = .true.

    !call cart_create((/ 1, nb_proc, 1 /), ierr)
    !call mesh_default()

    ! Choose a velocity
    velocity = N(2)/11*d_sc(2)
    T_step = 1
    T_end = 1
    T_end = T_step
    translat = 0
    translat(2) = - velocity*T_step
    translat = translat/d_sc

    ! Initialise the scalar field
    call scal_init(initialisation, scalar, good_scal, translat)
    call test_substatus('initialisation', success, myrank)

    ! Choose to test remeshing for centered/left and tagged/untagged particles
    bl_type = 1
    bl_tag  = 0

    ! Advec it with a particular method
    do k = begin_proc(3), end_proc(3)
    ind_group(2) = 0
    !k = begin_proc(3)
        ind_group(1) = 0
        ind_group(2) = ind_group(2) + 1
        do i = begin_proc(1), end_proc(1)
        !i = begin_proc(1)
            ind_group(1) = ind_group(1) + 1
            ! Initialise the particles
            p_SC = scalar(i,:,k)
            !scalar(i,:,k)=0
            do j = begin_proc(2), end_proc(2)
                p_pos_adim(j) = j
            end do
!            do T_step = 1, 11
                ! Advection
                p_pos_adim = p_pos_adim + T_step*velocity/d_sc(2)
                ! Remeshing
                call Yremesh_O2(ind_group, p_pos_adim, bl_type, bl_tag, i, k, scalar)
!            end do
        end do
    end do

    ! Check the final scalar field
    call test_check_success(scalar, good_scal, success)

    deallocate(scalar)
    deallocate(p_pos_adim)
    deallocate(p_SC)
    deallocate(bl_type)
    deallocate(bl_tag)

    success = .not.success

end function test_part_remesh_no_tag


!> Particles method: validation of the advection along one direction with the
!! order 2 particle method.
!!     @param[in]   init_scal   =  optional parameter to initialise the scalar fied to
!!                                  a constant one or to a sphere shape
!!     @return      error       = test error (= false if the code pass the test) (= not success)
!! @detail
!!    These tests are devoted to validate the advection solver based on particle
!!    method. They can be used for other advection solvers too. Their specificity
!!    is to test each configuration that could be encoutered in the order 2 solver
!!    based on particle method. Therefore they provide complete test for this
!!    order 2 solver but non necessary for a solver based on another method.
function test_part_advecO2(init_scal, shift) result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use advecY
    use cart_topology
    ! Test procedures
    use advec_aux_init
    use test_common

    logical                                     :: success
    character(len=*), intent(in), optional      :: init_scal
    integer, intent(in), optional               :: shift

    character(str_short)                        :: initialisation   ! to choose how to initialise the fields
    integer                                     :: shift_bis        ! shift effectly used in the test
    character(str_short)                        :: order            ! space order of the solveur
    real(WP), dimension(:, :, :), allocatable   :: scal3D           ! the scalar field
    real(WP), dimension(:, :, :), allocatable   :: velo             ! the flow
    integer                                     :: i,j,k            ! mesh indice   
    integer                                     :: T_step           ! time
    integer                                     :: T_end            ! final time
    real(WP)                                    :: dt               ! time step
    real(WP), dimension(:, :, :), allocatable   :: good_scal        ! analytic solution
    integer                                     :: direction = 2    ! current direction

    ! -- Allocation --
    allocate(scal3D(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_scal(N_proc(1), N_proc(2), N_proc(3)))
    allocate(velo(N_proc(1), N_proc(2), N_proc(3)))

    ! -- Initialisation --
    if(present(init_scal)) then
        initialisation = init_scal
    else 
        initialisation = 'center'
    end if
    if (present(shift)) then
        shift_bis = shift
    else
        shift_bis = 0
    end if
    success = .true.
    dt = 0.1

    call test_substatus('shift', shift_bis, myrank)

    ! Initialize the particular solver
    order = 'p_O2'
    call advec_init(order)

    ! Initialise the velocity, the scalar field and compute the theoritical solution
    call scal_velo_init_part(init_scal, dt, shift_bis, scal3D, velo, direction, good_scal)
    call test_substatus('initialisation', success, myrank)

    ! Advec it with a particular method
    call advecY_calc(dt, velo, scal3D, 'p_O2')
    call test_check_success(scal3D, good_scal, success)
    call test_substatus('advection', success, myrank)


    deallocate(scal3D)
    deallocate(good_scal)
    deallocate(velo)

    success = .not.success

end function test_part_advecO2


!> Test devoted to validate advection solver along one direction
!!     @param[in]   init_scal   =  optional parameter to initialise the scalar fied to
!!                                  a constant one or to a sphere shape
!!     @return      error       = test error (= false if the code pass the test) (= not success)
function test_advecY(init_scal) result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use advecY
    use cart_topology
    ! Test procedures
    use advec_aux_init
    use test_common

    logical                                     :: success
    character(len=*), intent(in), optional      :: init_scal

    character(len=17)                           :: initialisation   ! to choose how to initialise the scalar field
    character(len=str_short)                    :: order            ! space order of the solveur
    real(WP), dimension(:, :, :), allocatable   :: scal3D           ! the scalar field
    real(WP), dimension(:, :, :), allocatable   :: Vy               ! the flow
    real(WP)                                    :: velocity         ! constant velocity of the flux
    integer                                     :: ierr             ! mpi error code
    integer                                     :: i,j,k            ! mesh indice   
    integer                                     :: T_step           ! time
    integer                                     :: T_end            ! final time
    real(WP)                                    :: dt               ! time step
    real(WP), dimension(:, :, :), allocatable   :: good_scal        ! analytic solution
    real(WP), dimension(3)                      :: translat         ! to compute analytic solution

    allocate(scal3D(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_scal(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vy(N_proc(1), N_proc(2), N_proc(3)))

    if(present(init_scal)) then
        initialisation = init_scal
    else 
        initialisation = 'constant'
    end if
    success = .true.


    ! Initialize the particular solver
    order = 'p_O2'
    call advec_init(order)


    ! Initialise the velocity
    velocity = N(2)/11*d_sc(2)
    Vy = velocity
    T_end = 1
    translat = 0
    translat(2) = - velocity*T_end
    translat = translat/d_sc

    ! Initialise the scalar field
    call scal_init(initialisation, scal3D, good_scal, translat)
    call test_substatus('initialisation', success, myrank)

    ! Advec it with a particular method
    dt = 1
    call advecY_calc(dt, Vy, scal3D, 'p_O2')
    call test_check_success(scal3D, good_scal, success)
    call test_substatus('advec along Y', success, myrank)


    deallocate(scal3D)
    deallocate(good_scal)
    deallocate(Vy)

    success = .not.success

end function test_advecY


end module advec_aux
