!------------------------------------------------------------------------------
!
! MODULE: test_part_common
!
! DESCRIPTION: 
!> This module provides tests to validate the particular solver. It is
!! more precisly focused on testing the "common part" (advec_common) used for each
!! directions.
!!
!! @details 
!!  This module is devoted to validate all the procedure from "advec_common". All the
!! tests are unit tests: they return a logical value to check if the code version pass
!! it or not. 
!!
!! That is all these test are logical function, they return true if the result
!! is the right one and false otherwise. 
!! All the "test_part_*" function are devoted to validate the particular
!! solver
!! The following test are included :
!!      A - Validate the particular method, step by step
!!      1   -> Test the procedure "AC_obtain_senders" from advec_common
!!      2a  -> Validate the redistribution the buffer during the remeshing
!!      2b  -> Validate the redistribution the buffer during the remeshing -
!!              debug version : only one processus contains non-zero field.
!!      3   -> Validate the remeshing of untagged particles
!!      4   -> Validate the remeshing of tagged particles (XXX todo)
!!      B - Validate an advection solver (in advec_aux)
!!      1 -> advec a ball with a constant velocity
!!      2 -> advec a ball with a spheric velocity field (the ball turns)
!!
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

module advec_aux_common

    use string
    use precision
    implicit none



    ! ===== Test for the particles solver =====
    ! Public function
    public              :: test_part_AC_obtain_senders
    public              :: test_part_AC_bufferToScalar
    public              :: test_part_AC_bufferToScalar_Deb
    public              :: test_part_AC_interpol_velocity




contains


!> Particles method: validation of the procedure "AC_obtain_senders" wich determine
!! who will comunicate with who during the remeshing
!!     @return      success = test success (= false if the code pass the test)
!!     @param[in]   shift   = global translation of indices (optional)
function test_part_AC_obtain_senders(shift) result(success)

    ! Library
    use mpi
    ! Scale code
    use advec
    use advec_common
    use cart_topology
    ! Test procdure
    use test_common

    integer, intent(in), optional               :: shift
    logical                                     :: success          
    ! Be aware : during this function, success = true if everything is right, but we return not success = error

    integer                                     :: shift_bis        ! global translation of indices
    character(str_short)                        :: order            ! order of the particles solver
    integer                                     :: nb_proc          ! number of processes
    integer                                     :: ierr             ! mpi success code
    integer                                     :: j_min, j_max     ! input argument of the tested procedure
    integer                                     :: proc_min,proc_max! input argument of the tested procedure
    integer                                     :: send_begin       ! theoritical value of proc_min
    integer                                     :: send_end         ! theoritical value of proc_max
    integer, dimension(2)                       :: rece_proc        ! output argument of the tested procedure
    integer                                     :: rece_begin       ! theoritical value of rece_proc(1)
    integer                                     :: rece_end         ! theoritical value of rece_proc(2)
    integer, dimension(2)                       :: ind_group        ! indice of current group of lines
    integer                                     :: direction        ! current direction (alonG X, Y or Z)

    ! Some initialisation
    success = .true.
    ind_group = 1
    if (present(shift)) then
        shift_bis = shift
    else
        shift_bis = 0
    end if
    call test_substatus('shift', shift_bis, myrank)

    ! Initialize the particular solver
    order = 'p_O2'
    call advec_init(order)
    call test_substatus('initialisation solveur', success,  myrank)
    call mpi_barrier(MPI_COMM_WORLD, ierr)

    do direction = 1, 3
        call test_substatus('direction', direction, myrank)
        ! Test the procedure "AC_obtain_senders"
        ! If I communicate just with my self
        bl_bound_size = 0
        j_min = 1 + shift_bis*N_proc(direction)
        j_max = N_proc(direction)*(shift_bis+1)
        proc_min = -1
        proc_max = -1
        call AC_obtain_senders(j_min, j_max, direction, D_comm(direction), ind_group, proc_min, proc_max, rece_proc)
        call mpi_barrier(MPI_COMM_WORLD, ierr)
        if (rece_proc(1)/=-shift_bis) then 
            call test_substatus(' XXX error - just me', myrank)
            call test_substatus('rece_proc(1)', rece_proc(1), myrank)
            call test_substatus('and it must be', -shift_bis, myrank)
            success = .false.
        end if
        if (rece_proc(2)/=-shift_bis) then  
            call test_substatus(' XXX error - just me', myrank)
            call test_substatus('rece_proc(2)', rece_proc(2), myrank)
            call test_substatus('and it must be', -shift_bis, myrank)
            success = .false.
        end if
        if (proc_min/=shift_bis) then 
            call test_substatus(' XXX error - just me', myrank)
            call test_substatus('send_proc_min', proc_min, myrank)
            call test_substatus('and it must be', shift_bis, myrank)
            success = .false.
        end if
        if (proc_max/=shift_bis) then 
            call test_substatus(' XXX error - just me', myrank)
            call test_substatus('send_proc_max', proc_min, myrank)
            call test_substatus('and it must be', shift_bis, myrank)
            success = .false.
        end if
        call mpi_barrier(MPI_COMM_WORLD, ierr)
        call test_substatus('just me', success, myrank)

        ! I communicate with my two neighbors
        call advec_init(order)
        j_min = shift_bis*N_proc(direction)
        j_max = (1+shift_bis)*N_proc(direction) -1+2*bl_bound_size
        proc_min = 0
        proc_max = 0
        call AC_obtain_senders(j_min, j_max, direction, D_comm(direction), ind_group, proc_min, proc_max, rece_proc)
        call mpi_barrier(MPI_COMM_WORLD, ierr)
        if (rece_proc(1)/=-1-shift_bis) then  
            call test_substatus(' XXX error - neighbors', myrank)
            call test_substatus('rece_proc(1)', rece_proc(1), myrank)
            call test_substatus('and it must be', -1-shift_bis, myrank)
            success = .false.
        end if
        if (rece_proc(2)/=1-shift_bis) then 
            call test_substatus(' XXX error - neighbors', myrank)
            call test_substatus('rece_proc(2)', rece_proc(2), myrank)
            call test_substatus('and it must be', 1-shift_bis, myrank)
            success = .false.
        end if
        if (proc_min/=-1+shift_bis) then 
            call test_substatus(' XXX error - neighbors', myrank)
            call test_substatus('send_proc_min', proc_min, myrank)
            call test_substatus('and it must be', -1+shift_bis, myrank)
            success = .false.
        end if
        if (proc_max/=1+shift_bis) then 
            call test_substatus(' XXX error - neighbors', myrank)
            call test_substatus('send_proc_max', proc_min, myrank)
            call test_substatus('and it must be', 1+shift_bis, myrank)
            success = .false.
        end if
        call mpi_barrier(MPI_COMM_WORLD, ierr)
        call test_substatus('me and my neighbors', success, myrank)

        ! Contraction / extention
        if (modulo(nb_proc_dim(direction),2)==0) then
            if (modulo(myrank,2)== 0) then
                ! Contraction
                j_min = 1 + shift_bis*N_proc(2)
                j_max = (1+shift_bis)*N_proc(2)
                send_begin = 0
                send_end = 0
            else
                ! Dilatation
                j_min = +1 - 2*bl_bound_size + shift_bis*N_proc(direction)
                j_max = (1+shift_bis)*N_proc(direction) + 2*bl_bound_size 
                send_begin = -1
                send_end = 1
            end if
            if (modulo(myrank+shift_bis,2)== 0) then
                rece_begin = -1
                rece_end = +1
            else
                rece_begin = 0
                rece_end = 0
            end if
            proc_min = send_begin + 1
            proc_max = send_end + 1
            call AC_obtain_senders(j_min, j_max, direction, D_comm(direction), ind_group, proc_min, proc_max, rece_proc)
            call mpi_barrier(MPI_COMM_WORLD, ierr)
            if (rece_proc(1)/=rece_begin-shift_bis) then  
                call test_substatus(' XXX error contract/dilat', myrank)
                call test_substatus('rece_proc(1)', rece_proc(1), myrank)
                call test_substatus('and it must be', rece_begin-shift_bis, myrank)
                success = .false.
            end if
            if (rece_proc(2)/=rece_end-shift_bis) then 
                call test_substatus(' XXX error contract/dilat', myrank)
                call test_substatus('rece_proc(2)', rece_proc(2), myrank)
                call test_substatus('and it must be', rece_end-shift_bis, myrank)
                success = .false.
            end if
            if (proc_min/=send_begin+shift_bis) then 
                call test_substatus(' XXX error contract/dilat', myrank)
                call test_substatus('send_proc_min', proc_min, myrank)
                call test_substatus('and it must be', send_begin+shift_bis, myrank)
                success = .false.
            end if
            if (proc_max/=send_end+shift_bis) then 
                call test_substatus(' XXX error contract/dilat', myrank)
                call test_substatus('send_proc_max', proc_min, myrank)
                call test_substatus('and it must be', send_end+shift_bis, myrank)
                success = .false.
            end if
            call mpi_barrier(MPI_COMM_WORLD, ierr)
            call test_substatus('contraction/extension', success, myrank)
        else
            call test_substatus('nb of processus is even', myrank)
            call test_substatus('no contraction/dilatation test', myrank)
        end if
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    end do

    success = .not.success

end function test_part_AC_obtain_senders



!> Particles method: validation of the procedure "AC_bufferToScalar".
!!     @return      success = test success (= false if the code pass the test)
!!     @param[in]   shift   = global translation of indices (optional)
!! @details
!! Test the procedure "AC_obtain_senders" wich send local buffer use for remeshing
!! to the right processes. This procedure belong to "advec_common". 
function test_part_AC_bufferToScalar(shift) result(success)

    ! Library
    use mpi
    ! Scale code
    use advec
    use advec_common
    use cart_topology
    ! Test procdure
    use test_common

    integer, intent(in), optional               :: shift
    logical                                     :: success

    integer                                     :: shift_bis        ! global translation of indices
    character(str_short)                        :: order            ! order of the particles solver
    integer                                     :: nb_proc          ! number of processes
    integer                                     :: ierr             ! mpi success code
    integer                                     :: j_min, j_max     ! input argument of the tested procedure
    real(WP), dimension(:), allocatable         :: send_buffer      ! the buffer to redistribute
    real(WP), dimension(:), allocatable         :: scal1D           ! the scalar field
    integer                                     :: direction        ! direction (2 if along Y, 3=along Z)
    integer                                     :: success_inf        ! norm L_inf of the success
    real(WP)                                    :: good_scal        ! theoritical value of scal1D
    integer, dimension(2)                       :: mycoord          ! my coordonate in the mpi topology
    integer, dimension(2)                       :: ind_group        ! indice of current group of lines

    ! Some initialisation
    success = .true.
    ind_group = 1
    if (present(shift)) then
        shift_bis = shift
    else
        shift_bis = 0
    end if
    call test_substatus('shift', shift_bis, myrank)
    
    ! Initialize the particular solver
    order = 'p_O2'
    call advec_init(order)

    do direction = 1, 3
        call test_substatus('direction', direction, myrank)
        good_scal = modulo(coord(direction)-shift_bis, nb_proc_dim(direction))
        
        ! Test 1 - unrealistic case with no communication
        ! Initialize the buffer to remesh
        bl_bound_size = 0
        j_min = 1 + shift_bis*N_proc(direction)
        j_max = N_proc(direction)*(shift_bis+1)
        allocate (send_buffer(j_min:j_max))
        allocate (scal1D(1:N_proc(direction)))
        send_buffer = coord(direction)
        scal1D = 0.0
        ! Let's go !
        call mpi_barrier(MPI_COMM_WORLD, ierr)
        call AC_bufferToScalar(direction, D_comm(direction), ind_group, j_min, j_max, send_buffer, scal1D)
        ! Check the success
        call test_check_success_S(scal1D, good_scal, success)
        deallocate (send_buffer)
        deallocate (scal1D)
        call mpi_barrier(MPI_COMM_WORLD, ierr)
        call test_substatus('just me', success, myrank)

        ! I communicate with my two neighbors
        call advec_init(order)
        j_min = shift_bis*N_proc(direction)
        j_max = (1+shift_bis)*N_proc(direction) -1+2*bl_bound_size
        allocate (send_buffer(j_min:j_max))
        allocate (scal1D(1:N_proc(direction)))
        send_buffer = coord(direction)
        send_buffer(j_min) = modulo(coord(direction)-1, nb_proc_dim(direction))/2.0
        send_buffer(j_min+1) = send_buffer(j_min+1)/2.
        send_buffer(j_max) = modulo(coord(direction)+1, nb_proc_dim(direction))/2.0
        send_buffer(j_max-1) = send_buffer(j_max-1)/2.
        scal1D = 0.0
        call AC_bufferToScalar(direction, D_comm(direction), ind_group, j_min, j_max, send_buffer, scal1D)
        ! Check the success
        call test_check_success_S(scal1D, good_scal, success)
        deallocate (send_buffer)
        deallocate (scal1D)
        call mpi_barrier(MPI_COMM_WORLD, ierr)
        call test_substatus('me and my neighbors', success, myrank)
    end do

    success = .not.success

end function test_part_AC_bufferToScalar



!> Particles method: validation of the procedure "AC_bufferToScalar"
!!     @return      success = test success (= false if the code pass the test)
!!     @param[in]   shift   = global translation of indices (optional)
!!     @param[in]   rank    = rank of the processus which will contains non-zero field.
!! @details
!!    Debugging version : only one processus contains non zero field to remesh.
!!    Each processus communicate with its two neibor. A shift could be add.
function test_part_AC_bufferToScalar_Deb(rank, shift) result(success)

    ! Library
    use mpi
    ! Scale code
    use advec
    use advec_common
    use cart_topology
    ! Test procdure
    use test_common

    integer, intent(in)                         :: rank
    integer, intent(in), optional               :: shift
    logical                                     :: success

    integer                                     :: shift_bis        ! global translation of indices
    character(str_short)                        :: order            ! order of the particles solver
    integer                                     :: nb_proc          ! number of processes
    integer                                     :: ierr             ! mpi success code
    integer                                     :: j_min, j_max     ! input argument of the tested procedure
    real(WP), dimension(:), allocatable         :: send_buffer      ! the buffer to redistribute
    real(WP), dimension(:), allocatable         :: scal1D           ! the scalar field
    real(WP), dimension(:), allocatable         :: good_scal        ! theoritical value of scal1D
    integer                                     :: direction        ! direction (2 if along Y, 3=along Z)
    integer, dimension(3)                       :: mycoord          ! my coordonates in the mpi topology
    integer                                     :: rank_shift       ! rank of processus of rank = rank+shift
    integer, dimension(2)                       :: ind_group        ! indice of current group of lines

    ! Some initialisation
    success = .true.
    ind_group = 1
    if (present(shift)) then
        shift_bis = shift
    else
        shift_bis = 0
    end if
    call test_substatus('shift', shift_bis, myrank)
    call test_substatus('rank tested', rank, myrank)

    ! Initialize the particular solver
    order = 'p_O2'
    call advec_init(order)
    
    do direction = 1, 3
        call test_substatus('direction', direction, myrank)
        ! Initialise the solver environnement and the field
        call advec_init(order)
        j_min = shift_bis*N_proc(direction)
        j_max = (1+shift_bis)*N_proc(direction) -1+2*bl_bound_size
        allocate (send_buffer(j_min:j_max))
        allocate (scal1D(1:N_proc(direction)))
        send_buffer = 0
        if (myrank == rank) then
            send_buffer = 2
            send_buffer(j_min) = 1
            send_buffer(j_max) = 3
        end if
        scal1D = 0.0
        call mpi_barrier(MPI_COMM_WORLD, ierr)

        ! Compute the analytic solution
        allocate (good_scal(1:N_proc(direction)))
        good_scal = 0.0
        ! For the processus wich correcpond to me after the shift
        call mpi_cart_coords(cart_comm, rank, 3, mycoord, ierr)
        mycoord(direction) = mycoord(direction) + shift_bis
        call mpi_cart_rank(cart_comm, mycoord, rank_shift, ierr)
        if (myrank==rank_shift)   good_scal = 2
        ! For the next processus in the current direction
        mycoord(direction) = mycoord(direction) + 1
        call mpi_cart_rank(cart_comm, mycoord, rank_shift, ierr)
        if (myrank==rank_shift)   good_scal(1) = 3
        ! For the previous processus in the current direction
        mycoord(direction) = mycoord(direction) -2
        call mpi_cart_rank(cart_comm, mycoord, rank_shift, ierr)
        if (myrank==rank_shift)   good_scal(N_proc(direction)) = 1

        ! Compute the numerical solution
        call AC_bufferToScalar(direction, D_comm(direction), ind_group, j_min, j_max, send_buffer, scal1D)
        ! Check the success
        call test_check_success_F(scal1D, good_scal, success)
        deallocate (send_buffer)
        deallocate (scal1D)
        call mpi_barrier(MPI_COMM_WORLD, ierr)
        call test_substatus('me and my neighbors', success, myrank)
    end do

    success = .not.success


end function test_part_AC_bufferToScalar_Deb


!> Particles method: validation of the procedure "AC_obtain_senders" wich determine
!! who will comunicate with who during the remeshing
!!     @return      success = test success (= false if the code pass the test)
!!     @param[in]   shift   = global translation of indices (optional)
function test_part_AC_obtain_recevers(shift) result(success)

    ! Library
    use mpi
    ! Scale code
    use advec_common
    use cart_topology
    ! Test procedures
    use advec_aux_init
    use test_common

    integer, intent(in), optional               :: shift
    logical                                     :: success          
    ! Be aware : during this function, success = true if everything is right, but we return not success = error

    integer                                     :: shift_bis        ! global translation of indices
    integer                                     :: direction        ! current direction
    integer                                     :: comm             ! associated mpi communicator
    integer                                     :: rece_ind_min     ! the minimal indice used in velocity interpolation
    integer                                     :: rece_ind_max     ! the maximal indice used in velocity interpolation
    integer, dimension(2)                       :: rece_gap         ! distance between me and processus wich send me information
    integer, dimension(2)                       :: send_gap         ! distance between me and processus to wich I send information
    integer                                     :: nn, clock        ! to generate random number
    integer, dimension(:), allocatable          :: seed             ! to generate random number
    real(WP)                                    :: p_pos_1,p_pos_end! location where the velocity is interpolated
    integer                                     :: ind, indbis      ! indices
    real(WP), dimension(2)                      :: r                ! random numbers to initialise p_pos
    integer                                     :: ierr             ! mpi error code
    integer                                     :: i, k, ite        ! indices

    call test_substatus('test AC_obtain_recevers', myrank)

    ! Some initialisation
    success = .true.
    if (present(shift)) then
        shift_bis = shift
    else
        shift_bis = 0
    end if
    call test_substatus('shift', shift_bis, myrank)
    direction = 2
    comm = Y_comm
    ! Init the particles data
    ! To generate random number
    call RANDOM_SEED(size = nn)
    allocate(seed(nn))
    call SYSTEM_CLOCK(COUNT=clock)   
    seed = clock + 37 * (/ (ind - 1, ind = 1, nn) /)
    call RANDOM_SEED(PUT = seed)   
    deallocate(seed)
    ! Position where the velocity will be interpolated
    CALL RANDOM_NUMBER(r)
    p_pos_1 = (1+shift_bis+r(1))*d_sc(direction)
    p_pos_end = (N_proc(direction)+shift_bis+r(2))*d_sc(direction)

    ! Compute range of the set of point where I need the velocity value
    rece_ind_min = floor(p_pos_1/d_sc(direction))
    rece_ind_max = floor(p_pos_end/d_sc(direction)) + 1
    ! Test the function
    indbis=0
    do ite = 1, 10
        do i = 1, 100
            do k = 1, 100
                ind = ind + 1
                call AC_obtain_receivers(direction, comm, (/i, k/), rece_ind_min, rece_ind_max, send_gap, rece_gap)
                if (ind>indbis*100000) then
                    call test_substatus('iteration/100000', ind/100000, myrank)
                    indbis = indbis+1
                end if
            end do
        end do
        call mpi_barrier(MPI_COMM_WORLD, ierr)
    end do

    success = .not. success

end function test_part_AC_obtain_recevers


!> Particles method: validation of the velocity interpolation
!!     @return      success     = test success (= false if the code pass the test)
!!     @param[in]   direction   = 1 for along X, 2 for along Y, 3 for along Z
!!     @param[in]   comm        = mpi communcator associated to direction
!!     @param[in]   shift       = global translation of indice
function test_part_AC_interpol_velocity(direction, comm, shift) result(success)

    ! External Library
    use mpi
    ! Scales code
    use cart_topology
    ! Test procedures
    use advec_aux_init
    use test_common

    logical                                     :: success          
    integer, intent(in)                         :: direction, comm, shift

    integer                                                 :: ind          ! indice
    integer                                                 :: nn, clock    ! to generate random number
    integer, dimension(:), allocatable                      :: seed         ! to generate random number
    real(WP), dimension(N_proc(direction))                  :: p_pos        ! location where the velocity is interpolated
    real(WP), dimension(N_proc(direction))                  :: r            ! random numbers to initialise p_pos
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3))    :: velo         ! velocity field
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3))    :: good_velo    ! analytic interpolation of velocity field in particles position
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3))    :: num_velo     ! numerical interpolation of velocity field in particles position
    real(WP)                                                :: dt           ! time step
    integer                                                 :: T            ! for solver iteration
    integer                                                 :: ierr         ! mpi error code


    call test_substatus('test velocity interpolation', myrank)
    call test_substatus('shift', shift, myrank)

    ! Initialisation
    success = .true.
    ! To generate random number
    call RANDOM_SEED(size = nn)
    allocate(seed(nn))
    call SYSTEM_CLOCK(COUNT=clock)   
    seed = clock + 37 * (/ (ind - 1, ind = 1, nn) /)
    call RANDOM_SEED(PUT = seed)   
    deallocate(seed)
    ! Position where the velocity will be interpolated
    CALL RANDOM_NUMBER(r)
    do ind = 1, N_proc(direction)
        p_pos(ind) = (ind+shift)
    end do

    ! ===== Particles are on mesh point =====
    call test_substatus('particles are placed on a mesh point', myrank)
    ! - Basic test : constant velocity -
    call velo_init('constant', velo, direction, p_pos, good_velo)
    dt = 0.0
do T = 1, 10
    call aux_com_interpol(dt, direction, comm, p_pos, velo, num_velo)
    ! Check result
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_check_success(num_velo, good_velo, success)
end do
    call test_substatus('constant velocity', success, myrank)
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    ! - Normal test -
    call velo_init('translation_field', velo, direction, p_pos, good_velo)
    call aux_com_interpol(dt, direction, comm, p_pos, velo, num_velo)
    ! Check result
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_check_success(num_velo, good_velo, success)
    call test_substatus('translation_field', success, myrank)
    

    ! ===== Test with p_pos not on a mesh point =====
    call test_substatus('particles are not placed on a mesh point', myrank)
    p_pos = p_pos + r*d_sc(direction)
    ! - Basic test : constant velocity -
    call velo_init('constant', velo, direction, p_pos, good_velo)
    do T = 1, 10
        call aux_com_interpol(dt, direction, comm, p_pos, velo, num_velo)
        call mpi_barrier(MPI_COMM_WORLD, ierr)
    end do
    ! Check result
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_check_success(num_velo, good_velo, success)
    call test_substatus('constant velocity', success, myrank)
    ! - Normal test -
    call velo_init('translation_field', velo, direction, p_pos, good_velo)
    call aux_com_interpol(dt, direction, comm, p_pos, velo, num_velo)
    ! Check result
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_check_success(num_velo, good_velo, success)
    call test_substatus('translation_field', success, myrank)


    ! Return not success
    success = .not. success

end function test_part_AC_interpol_velocity


!> Particle method: validation of particle tag and block type determination
!!     @return      error   = test error (= false if the code pass the test) (= not success)
!!     @param[in]   shift   = global translation of indices (optional)
function test_part_AC_type_and_block_O2(shift) result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use advec_common
    use cart_topology
    ! Test procedures
    use test_common

    integer, intent(in), optional               :: shift
    logical                                     :: success

    integer                                     :: shift_bis    ! global translation of indices
    character(str_short)                        :: order        ! order of the particles solver
    
    integer, dimension(:), allocatable          :: bl_type      ! correct (analytic) type of block
    integer, dimension(:), allocatable          :: bl_tag       ! correct (analytic) tag of block
    integer, dimension(:), allocatable          :: good_type    ! correct (analytic) type of block
    integer, dimension(:), allocatable          :: good_tag     ! correct (analytic) tag of block
    integer                                     :: ierr         ! mpi success code
    integer                                     :: direction    ! direction (2 if along Y, 3=along Z)
    integer, dimension(2)                       :: ind_group    ! indice of the current group of line
    real(WP)                                    :: dt, cfl      ! time step and CFL number
    real(WP), dimension(:), allocatable         :: p_V          ! particle velocity (used to tag particles)
    real(WP)                                    :: p_V_next     ! velocity of the first particle of the next block 
                                                                ! (used to determine block type and to tag particles)
    integer, dimension(3)                       :: comm_table   ! table of mpi communicator dedicaced to each direction
    integer                                     :: ind, ind2    ! indice of the current particle
    integer                                     :: ind_bl       ! indice of the current block
    real(WP)                                    :: lambda_min   ! minimum courant number on a block
    integer                                     :: ind_tag_bl   ! indice of the first tagged block in the test
    integer                                     :: ind_tag_p    ! indice of particle corresponding to the velocity 
                                                                ! variation wich induce a tag of the block ind_tag_bl 
    integer                                     :: ind_tag_bl2  ! indice of the second tagged block in the test
    integer                                     :: ind_tag_p2   ! indice of particle wich induce a tag of the block ind_tag_bl2


    ! Some initialisation
    success = .true.
    ind_group = 1
    if (present(shift)) then
        shift_bis = shift
    else
        shift_bis = 0
    end if
    call test_substatus('shift', shift_bis, myrank)
    call advec_init('p_O2')
    dt = 0.01
    comm_table(1) = X_comm
    comm_table(2) = Y_comm
    comm_table(3) = Z_comm

direction = 2
! do direction = 1, 3
    call test_substatus('direction', direction, myrank)
    allocate(good_type(bl_number(direction)+1))
    allocate(good_tag(bl_number(direction)))
    allocate(bl_type(bl_number(direction)+1))
    allocate(bl_tag(bl_number(direction)))
    allocate(p_V(N_proc(direction)))
    cfl = dt/d_sc(direction)
    if (cfl == 0) cfl = 1

    !  -- Test case 1 : only left block, no tag --
    p_V = 0.4/cfl
    good_tag = 0
    good_type = 0
    call AC_type_and_block(dt, direction, comm_table(direction), (/1,1/), p_V, &
            & bl_type, bl_tag)
    call test_check_success(bl_tag, good_tag, success)
    call test_substatus('test 1.a - no tag ', success, myrank)
    call test_check_success(bl_type, good_type, success)
    call test_substatus('test 1.a - left particle only', success, myrank)
    call MPI_barrier(MPI_COMM_WORLD, ierr)
    ! And for another velocity
    p_V = -2.8/cfl
    call AC_type_and_block(dt, direction, comm_table(direction), (/1,1/), p_V, &
            & bl_type, bl_tag)
    call test_check_success(bl_tag, good_tag, success)
    call test_substatus('test 1.b - no tag ', success, myrank)
    call test_check_success(bl_type, good_type, success)
    call test_substatus('test 1.b - left block only', success, myrank)

    ! Test case 2 : only center block, no tag
    p_V = -0.4/cfl
    good_tag = 0
    good_type = 1
    call AC_type_and_block(dt, direction, comm_table(direction), (/1,1/), p_V, &
            & bl_type, bl_tag)
    call test_check_success(bl_tag, good_tag, success)
    call test_substatus('test 2.a - no tag ', success, myrank)
    call test_check_success(bl_type, good_type, success)
    call test_substatus('test 2.a - center particle only', success, myrank)
    call MPI_barrier(MPI_COMM_WORLD, ierr)
    ! And for another velocity
    p_V = 6.8/cfl
    call AC_type_and_block(dt, direction, comm_table(direction), (/1,1/), p_V, &
            & bl_type, bl_tag)
    call test_check_success(bl_tag, good_tag, success)
    call test_substatus('test 2.b - no tag ', success, myrank)
    call test_check_success(bl_type, good_type, success)
    call test_substatus('test 2.b - center particle only', success, myrank)
    
    ! Test case 3 = 2 tag
    ! one for an extension: on the left part of the domain all block are left
    ! one, on the right side they are center and there is a gap of 1 in the block index
    ! one for a contraction: on the left part of the domain all block are center
    ! one, on the right side they are left and there is a gap of -1 in the block index
    ! => particles are tagged on the location of the switch
    ! Update dt in order to create "chock" without broke the stability condition
    dt = 0.8*d_sc(direction)/sqrt(2*1.2*bl_size)
    cfl = dt/d_sc(direction)
    good_type = 0 
    good_tag = 0
    p_V = shift_bis/cfl
    ind_tag_bl = (nb_proc_dim(direction)*bl_number(direction)/3) - (bl_number(direction)*coord(direction))
    ind_tag_p = (ind_tag_bl-1)*bl_size + bl_size/2 + 1
    ind_tag_p2 = floor((0.3/0.8)*(N(direction)-N_proc(direction)*coord(direction)-ind_tag_p) + ind_tag_p)
    ind_tag_p2 = ind_tag_p2 + 1
    ind_tag_bl2 = (ind_tag_p2 - 1 - bl_size/2)/bl_size + 1
    if (modulo(ind_tag_p2-1-bl_size/2, bl_size)==0) ind_tag_bl2 = ind_tag_bl2 -1
    if (ind_tag_bl <=bl_number(direction)) then
        if (ind_tag_bl>=1) then
            ! Tag the right block transition
            good_tag(ind_tag_bl) = 1
        else 
            ! Change the velocity for the first half block
            lambda_min = max(0,shift_bis) + 10
            do ind = 1, bl_size/2
                p_V(ind) = shift_bis/cfl
                p_V(ind) = p_V(ind) + 0.8*(1.0-float(ind-ind_tag_p)/(N(direction)-N_proc(direction)*coord(direction)-ind_tag_p))/cfl
            end do
            if (ind+1<ind_tag_p2) good_type(1) = 1
        end if
        ! Update velocity and block type
        do ind_bl = max(2, ind_tag_bl+1), bl_number(direction)
            lambda_min = max(0,shift_bis) + 10
            do ind = 1, bl_size
                ind2 = ind+((ind_bl-2)*bl_size)+bl_size/2 ! the first block is only a half block
                p_V(ind2) = shift_bis/cfl
                p_V(ind2) = p_V(ind2)+0.8*(1.-float(ind2-ind_tag_p)/(N(direction)-N_proc(direction)*coord(direction)-ind_tag_p))/cfl
            end do
            if (ind2+1<ind_tag_p2) good_type(ind_bl) = 1
        end do
        ! For the last half block
        lambda_min = max(0,shift_bis) + 10
        ind_bl = bl_number(direction) + 1
        do ind = 1, bl_size/2
            ind2 = ind+((ind_bl-2)*bl_size)+bl_size/2 ! the first block is only a half block
            p_V(ind2) = shift_bis/cfl
            p_V(ind2) = p_V(ind2)+0.8*(1.-float(ind2-ind_tag_p)/(N(direction)-N_proc(direction)*coord(direction)-ind_tag_p))/cfl
        end do
        if (ind2+bl_size/2+1<ind_tag_p2) good_type(ind_bl) = 1
    end if
    if ((ind_tag_bl2 <= bl_number(direction)).and.(ind_tag_bl2>0)) good_tag(ind_tag_bl2) = 1
    call AC_type_and_block(dt, direction, comm_table(direction), (/1,1/), p_V, &
            & bl_type, bl_tag)

! == Pour debugger sur 1 processus ==
!print*, 'ind_tag_p2 = ', ind_tag_p2
!print*, 'V(ind_tag_p2) = ', p_V(ind_tag_p2)*cfl
!print*, 'ind_tag_bl2 = ', ind_tag_bl2
!do ind = 0, 9
!print*, ' V de ', ind*10+1, ' a ', (ind+1)*10
!print*, p_V(ind*10+1:(ind+1)*10)*cfl
!print*, 'tag associé :'
!print*, bl_tag(ind*5+1:(ind+1)*5)
!print*, 'tag prevu :'
!print*, good_tag(ind*5+1:(ind+1)*5)
!print*, 'type associé :'
!print*, bl_type(ind*5+1:(ind+1)*5+1)
!print*, 'type prevu :'
!print*, good_type(ind*5+1:(ind+1)*5+1)
!end do
! == Pour debugger en parallele ==
if (maxval(abs(bl_type-good_type))>0) then
print*, 'myrank = ', myrank
print*, 'p_V = ', p_V*cfl
print*, 'type associé :'
print*, bl_type
print*, 'type prevu :'
print*, good_type
end if

    call test_check_success(bl_tag, good_tag, success)
    call test_substatus('test 3 - one tag ', success, myrank)
    call test_check_success(bl_type, good_type, success)
    call test_substatus('test 3 - center at first, then left', success, myrank)
    call MPI_barrier(MPI_COMM_WORLD, ierr)
    

    deallocate(good_type)
    deallocate(good_tag)
    deallocate(bl_type)
    deallocate(bl_tag)
    deallocate(p_V)
! end do

    success = .not. success

end function test_part_AC_type_and_block_O2


! ===== Private procedure =====


!> Particles method: interpolate the velocity on all the 3D domain
!!     @param[in]   dt          = time step
!!     @param[in]   direction   = 1 for along X, 2 for along Y, 3 for along Z
!!     @param[in]   comm        = mpi communcator associated to direction
!!     @param[in]   p_pos       = particle position
!!     @param[in]   velo        = velocity field
!!     @param[out]  num_velo    = numerical interpolation of the velocity field
!! @details
!!     The function AC_particle_velocity interpolate the velocity on a 1D line
!!     (along X, Y or Z-axis). In order to test it on all direction and on 3D
!!     cases, this subroutine call this interpolation on each line of the domain.
subroutine aux_com_interpol(dt, direction, comm, p_pos, velo, num_velo)

    ! Scales code
    use advec_common
    use cart_topology
    ! Test procedures
    use advec_aux_init
    use test_common

    real(WP), intent(in)                                                :: dt           ! time step
    integer, intent(in)                                                 :: direction, comm
    real(WP), dimension(N_proc(direction)), intent(in)                  :: p_pos        ! location where the velocity is interpolated
    real(WP), dimension(:,:,:), intent(in)                              :: velo         ! velocity field
    real(WP), dimension(N_proc(1), N_proc(2), N_proc(3)), intent(out)   :: num_velo     ! numerical interpolation of velocity field

    integer                                                             :: i,j,k,ind    ! indice
    real(WP), dimension(N_proc(direction))                              :: p_V          ! particles velocity (interpolation of velo in location p_pos)

    select case(direction)
        case(1)
            do k = 1, N_proc(3)
                do j = 1, N_proc(2)
                    do ind = 1, N_proc(1)
                        p_V(ind) = velo(ind, j, k)
                    end do
                    call AC_particle_velocity(dt, direction, comm, (/j, k/), p_pos, p_V)
                    do ind = 1, N_proc(1)
                        num_velo(ind,j,k)= p_V(ind) 
                    end do
                end do
            end do
        case(2)
            do k = 1, N_proc(3)
                do i = 1, N_proc(1)
                    do ind = 1, N_proc(2)
                        p_V(ind) = velo(i, ind, k)
                    end do
                    call AC_particle_velocity(dt, direction, comm, (/i, k/),  p_pos, p_V)
                    do ind = 1, N_proc(2)
                        num_velo(i, ind,k)= p_V(ind) 
                    end do
                end do
            end do
        case(3)
            do j = 1, N_proc(2)
                do i = 1, N_proc(1)
                    do ind = 1, N_proc(2)
                        p_V(ind) = velo(i, j, ind)
                    end do
                    call AC_particle_velocity(dt, direction, comm, (/i, k/), p_pos, p_V)
                    do ind = 1, N_proc(2)
                        num_velo(i, j, ind)= p_V(ind) 
                    end do
                end do
            end do
    end select

end subroutine aux_com_interpol

end module advec_aux_common
