!------------------------------------------------------------------------------
!
! PROGRAM : advec_main
!
! DESCRIPTION: 
!> This program use the function implemented in the module advec_aux to
!! test the advection solver.
!!
!! @details
!! All these test are unit test : they return a logical value to check if 
!! the code version pass it or not. 
!!
!! That is all these test are logical function, they return true if the result
!! is the right one and false otherwise. 
!! All the "test_part_*" function are devoted to validate the particular
!! solver
!! The following test are included :
!!      1 -> advec a ball with a constant velocity
!!      2 -> advec a ball with a spheric velocity field (the ball turns)
!!
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

program advec_main

    ! External Library
    use mpi
    ! Scales code
    use cart_topology
    ! Test procedures
    use test_common
    use advec_aux
    use advec_aux_common

    implicit none

    logical     :: error = .false.  ! logical error
    integer     :: ierr             ! mpi error code
    integer     :: rank_world       ! processus rank on "MPI_COMM_WORLD"
    integer     :: nb_proc          ! number of processus
    integer     :: i                ! some boucle indice
   
    ! ===== Initialisation =====

    ! Set the verbosity
    verbose_test = .true.
    verbose_more = .false.
    ! Initialise mpi
    call mpi_init(ierr)    
    call mpi_comm_rank(MPI_COMM_WORLD, rank_world, ierr)
    call mpi_comm_size(MPI_COMM_WORLD, nb_proc, ierr)

    ! Cut the domain along Y and initialize the toppology
    if (mod(100, nb_proc)/=0) then
        stop 'wrong number of processes : it have to divide 100'
    end if
    call cart_create((/ nb_proc, 1 /), ierr)
    call mesh_default()
    call mpi_barrier(MPI_COMM_WORLD, ierr)


    ! ===== Test about procedures involved in remeshing process =====
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_title('particle method - remeshing', rank_world)
    
    ! Does it well compute who communicate with who during remeshing ?
    error = test_part_AC_obtain_senders()
    call test_status(error, 'obtain_senders sans shift', rank_world)
    error = test_part_AC_obtain_senders(1)
    call test_status(error, 'obtain_senders avec shift', rank_world)
    error = test_part_AC_obtain_senders(3*nb_proc_dim(2)+2)
    call test_status(error, 'obtain_senders avec large shift', rank_world)
    
verbose_more = .true.
    if (nb_proc_dim(2)>1) then 
        ! The remeshing are done in a buffer. Is it well re-distribuate on processes ?
        do i = 0, min(4, nb_proc-1)
            error = test_part_AC_bufferToScalar_Deb(i)
            call test_status(error, 'bufferToScalar deb, no shift, rank =', i, rank_world)
            error = test_part_AC_bufferToScalar_Deb(i,1)
            call test_status(error, 'bufferToScalar deb,shift; rank =', i, rank_world)
        end do
        error = test_part_AC_bufferToScalar()
        call test_status(error, 'bufferToScalar sans shift', rank_world)
        error = test_part_AC_bufferToScalar(1)
        call test_status(error, 'bufferToScalar avec shift', rank_world)
        error = test_part_AC_bufferToScalar(3*nb_proc_dim(2)+2)
        call test_status(error, 'bufferToScalar avec large shift', rank_world)

        ! Test remeshing of untagged particles
        error = test_part_remesh_no_tag()
        call test_status(error, 'remeshing of cst field advected a cst v', rank_world)

        error = test_part_remesh_no_tag('translation_field')
        call test_status(error, 'remeshing of uncst field advected a cst v', rank_world)
    else 
        call test_status(error, 'only one proc along Y, no test on remesh', rank_world)
    end if
verbose_more = .false.

    ! ===== Test about procedures involved in computation of particles advection =====
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_title('particle method - particle advection', rank_world)

    ! Test auxiliary procedures
    error = test_part_AC_obtain_recevers()
    call test_status(error, 'AC_obtain_recevers', rank_world)


    ! Test velocity interpolation
    error = test_part_AC_interpol_velocity(2, Y_comm, 0)
    call test_status(error, 'velocity interpolation along Y, no shift', rank_world)
    error = test_part_AC_interpol_velocity(2, Y_comm, 1)
    call test_status(error, 'velocity interpolation along Y, shift', rank_world)

    ! ===== Test others =====
    call test_title('particle method - tag and bloc type', rank_world)
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    error = test_part_AC_type_and_block_O2()
    call test_status(error, 'determine block type and tag particles', rank_world)

   ! ===== Test solvers =====

    ! Test devoted to solver based on particles method
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_title('particle method - advection test', rank_world)
    error = test_part_advecO2('left')
    call test_status(error, 'advection - left block, no tag', rank_world)
    error = test_part_advecO2()
    call test_status(error, 'advection - center block, no tag', rank_world)

    ! Generic test
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_title('generic advection test - V=constant', rank_world)
    error = test_advecY()
    call test_status(error, 'advection of a constant field', rank_world)
    error = test_advecY('translation_field')
    call test_status(error, 'translation of a unconstant field', rank_world)

    call mpi_finalize(ierr)

end program advec_main

