module transform

  use precision
  use parallel
  implicit none
  include 'fftw3.f'

  real(WP), dimension(:,:,:), pointer :: A    !_real (ns1,ns2,ns3)
  complex(WP), dimension(:,:,:), pointer :: B !_complex (fns1,fns2,fns3)
  complex(WP), dimension(:,:,:), pointer :: C !_dummy (fns1,fns2,fns3)
  complex(WP), dimension(:,:,:), pointer :: D !_dummy (fns1,fns3,fns2)
end module transform


subroutine forward_transform(n1,n2,n3,fn1,fn2,fn3)
    
  use transform
    implicit none

  integer :: nsize(2),idim,istride,idist,ostride,odist
  integer :: idim2,istride2,idist2,ostride2,odist2
  integer :: inem(3),onem(3),howmany
  integer :: n1,n2,n3,fn1,fn2,fn3
  integer(KIND=8) :: plan_mf1,plan_mf2


  integer, parameter :: NULL = 0

  integer :: nsize1

  idim = 2
  nsize(1) = n1
  nsize(2) = n2
  inem(1) = n1
  inem(2) = n2
  inem(3) = n3
  istride = 1
  idist = n1*n2

  onem(1) = fn1
  onem(2) = fn3
  onem(3) = fn2
  ostride = 1
  odist = fn1*fn3

    call dfftw_plan_many_dft_r2c(plan_mf1,idim,nsize,n3,A,inem,istride,idist,D,onem,ostride,odist,FFTW_ESTIMATE)
    call dfftw_execute_dft_r2c(plan_mf1,A,D)
    call dfftw_destroy_plan(plan_mf1)
    
    call transpose_forward(n1,n2,n3,fn1,fn2,fn3)   !(D,B)

  idim2 = 1
  nsize1 = fn3
  howmany = fn1*fn2

  inem(1) = fn1
  inem(2) = fn2
  inem(3) = fn3
  istride2 = fn1*fn2
  idist2 = 1

  onem(1) = fn1
  onem(2) = fn2
  onem(3) = fn3
  ostride2 = fn1*fn2
  odist2 = 1

    call dfftw_plan_many_dft(plan_mf2,idim2,nsize1,howmany,B,inem,istride2,idist2,B,onem,ostride2,&
       &odist2,FFTW_FORWARD,FFTW_ESTIMATE) 
    call dfftw_execute_dft(plan_mf2,B,B)
    call dfftw_destroy_plan(plan_mf2)
    
end subroutine forward_transform


subroutine backward_transform(n1,n2,n3,fn1,fn2,fn3)
  
  use transform
  implicit none
  integer :: k,j,i
  integer :: nsize(2),idim,istride,idist,ostride,odist
  integer :: idim2,istride2,idist2,ostride2,odist2
  integer :: inem(3),onem(3),howmany,nsize1
  integer :: n1,n2,n3,fn1,fn2,fn3
  integer(KIND=8) :: plan_mb1,plan_mb2


  idim2 = 1
  nsize1 = fn3
  howmany = fn1*fn2

  inem(1) = fn1
  inem(2) = fn2
  inem(3) = fn3
  istride2 = fn1*fn2
  idist2 = 1

  onem(1) = fn1
  onem(2) = fn2
  onem(3) = fn3
  ostride2 = fn1*fn2
  odist2 = 1

  call dfftw_plan_many_dft(plan_mb1,idim2,nsize1,howmany,B,inem,istride2,idist2,B,onem,ostride2,&
       &odist2,FFTW_BACKWARD,FFTW_ESTIMATE)
  call dfftw_execute_dft(plan_mb1,B,B)
  call dfftw_destroy_plan(plan_mb1)

  call transpose_backward(n1,n2,n3,fn1,fn2,fn3)           !(B,D)

  idim = 2
  nsize(1) = n1
  nsize(2) = n2
  inem(1) = n1
  inem(2) = n2
  inem(3) = n3
  istride = 1
  idist = n1*n2

  onem(1) = fn1
  onem(2) = fn3
  onem(3) = fn2
  ostride = 1
  odist = fn1*fn3

  call dfftw_plan_many_dft_c2r(plan_mb2,idim,nsize,fn2,D,onem,ostride,odist,A,inem,istride,idist,FFTW_ESTIMATE)
  call dfftw_execute_dft_c2r(plan_mb2,D,A)
  call dfftw_destroy_plan(plan_mb2)
  
end subroutine backward_transform


  
subroutine transpose_forward(n1,n2,n3,fn1,fn2,fn3)



  use transform
  implicit none

  integer :: n1,n2,n3,fn1,fn2,fn3
  integer :: scount,ierr,i,k,ki,jstart,jend,j

  if (nproc.gt.1) then

     scount = fn1*fn2*fn2
     
     
     do i = 1,nproc
        do k = 1,fn2
           ki = (i-1)*fn2 + k
           jstart = (i-1)*fn2 + 1
           jend = i*fn2
           
           C(1:fn1,1:fn2,ki) = D(1:fn1,jstart:jend,k)
        enddo
     enddo
     
     
     call MPI_ALLTOALL(C,scount,MPI_COMPLEX_WP,B,scount,MPI_COMPLEX_WP,MPI_COMM_WORLD,ierr)
  else
     do k = 1,fn3
        do j = 1,fn2
           do i = 1,fn1
              B(i,j,k) = D(i,j,k)
           enddo
        enddo
     enddo
  endif
end subroutine transpose_forward

subroutine transpose_backward(n1,n2,n3,fn1,fn2,fn3)



  use transform
  implicit none

  integer :: n1,n2,n3,fn1,fn2,fn3
  integer :: scount,ierr,i,k,ki,jstart,jend,j

  if (nproc.gt.1) then

     scount = fn1*fn2*fn2
     
     
     call MPI_ALLTOALL(B,scount,MPI_COMPLEX_WP,C,scount,MPI_COMPLEX_WP,MPI_COMM_WORLD,ierr)
     
     
     do i = 1,nproc
        do k = 1,fn2
           ki = (i-1)*fn2 + k
           jstart = (i-1)*fn2 + 1
           jend = i*fn2
           D(1:fn1,jstart:jend,k) =  C(1:fn1,1:fn2,ki) 
        enddo
     enddo
  else
     do k = 1,fn3
        do j = 1,fn2
           do i = 1,fn1
              D(i,j,k) = B(i,j,k)
           enddo
        enddo
     enddo
  endif

end subroutine transpose_backward



  
  
  




subroutine transform_init

  use transform



  implicit none

  integer :: nsize(2),idim,istride,idist,ostride,odist
  integer :: idim2,istride2,idist2,ostride2,odist2
  integer :: inem(3),onem(3),howmany
  integer, parameter :: NULL = 0

  integer :: nsize1

  
!  print *,' PLANS', plan_mf1,plan_mf2,plan_mb1,plan_mb2


end subroutine transform_init


subroutine ftran_wrap(in,out,n1,n2,n3,fn1,fn2,fn3)

  use transform
  use precision
  implicit none

  integer :: n1,n2,n3,fn1,fn2,fn3
  real(WP) :: in(n1,n2,n3)
  complex(WP) :: out(fn1,fn2,fn3)
  integer :: i,j,k

  allocate(A(n1,n2,n3))
  allocate(B(fn1,fn2,fn3))
  allocate(C(fn1,fn2,fn3))
  allocate(D(fn1,fn3,fn2))


  do k = 1,n3
     do j = 1,n2
        do i = 1,n1
           A(i,j,k) = in(i,j,k)
        enddo
     enddo
  enddo
  
  call forward_transform(n1,n2,n3,fn1,fn2,fn3)
  
  do k = 1,fn3
     do j = 1,fn2
        do i = 1,fn1
           out(i,j,k) = B(i,j,k)
        enddo
     enddo
  enddo
  deallocate(A)
  deallocate(B)
  deallocate(C)
  deallocate(D)
 
 
end subroutine ftran_wrap
  
subroutine btran_wrap(in,out,n1,n2,n3,fn1,fn2,fn3)

  use transform
  use precision

  implicit none

  integer :: n1,n2,n3,fn1,fn2,fn3
  real(WP) :: out(n1,n2,n3)
  complex(WP) :: in(fn1,fn2,fn3)


  integer :: i,j,k


  allocate(A(n1,n2,n3))
  allocate(B(fn1,fn2,fn3))
  allocate(C(fn1,fn2,fn3))
  allocate(D(fn1,fn3,fn2))


  do k = 1,fn3
     do j = 1,fn2
        do i = 1,fn1
           B(i,j,k) = in(i,j,k)
        enddo
     enddo
  enddo


  call backward_transform(n1,n2,n3,fn1,fn2,fn3)

  do k = 1,n3
     do j = 1,n2
        do i = 1,n1
           out(i,j,k) = A(i,j,k)
        enddo
     enddo
  enddo

  deallocate(A)
  deallocate(B)
  deallocate(C)
  deallocate(D)


end subroutine btran_wrap
