#!/usr/bin/python
import parmepy as pp
import parmepy.f2py
import numpy as np
import mpi4py.MPI as MPI
import math as m

PARMES_REAL = pp.constants.PARMES_REAL
ORDER = pp.constants.ORDER

#from numpy import linalg as LA

pi = m.pi

# Import scales and fftw solvers
ppfft = parmepy.f2py.fftw2py
scales = parmepy.f2py.scales2py

rank = MPI.COMM_WORLD.Get_rank()
print "Mpi process number ", rank

# ----------- A 3d problem -----------
print " ========= Start test for Navier-Stokes 3D ========="

# Physical Domain description
Lx = Ly = Lz = 2 * pi
myDomain3d = pp.Box(dimension=3, length=[Lx, Ly, Lz], origin=[0., 0., 0.])
resolution3d = np.asarray((65, 65, 65))
ncells = resolution3d - 1
hx = Lx / ncells[0]
hy = Ly / ncells[1]
hz = Lz / ncells[2]

## Obstacle
lambd = np.array([0, 1, 10 ** 8], dtype=PARMES_REAL, order=ORDER)
sphere = pp.Obstacle(myDomain3d, name='sphere', zlayer=0.1,
                     radius=0.1, center=[0.5, 0.5, 0.5],
                     orientation='West', porousLayer=0.05)

## Post
outputFilePrefix = './res/NS_'
outputModulo = 1

# Simulation parameters
timeStep = 1e-8
finalTime = timeStep#1.0

# Fields declaration

# 1 - Poisson/diffusion solvers initialisation.
# See poisson3d.py for a working example.
# poisson = pp.Poisson(vorticity,velocity)
# 
localres, localoffset = ppfft.init_fftw_solver(resolution3d,
                                               myDomain3d.length)

print "FFT solver local resolution/offset: ", localres, localoffset

##topofft = poisson.getTopology()


def computeVel(x, y, z):
    vx = 2. / np.sqrt(3) * np.sin(2. * pi / 3.) * np.sin(x) \
        * np.cos(y) * np.cos(z)
    vy = 2. / np.sqrt(3) * np.sin(-2. * pi / 3.) * np.cos(x) \
        * np.sin(y) * np.cos(z)
    vz = 0.
    return vx, vy, vz


def computeVort(x, y, z):
    wx = - np.cos(x) * np.sin(y) * np.sin(z)
    wy = - np.sin(x) * np.cos(y) * np.sin(z)
    wz = 2. * np.sin(x) * np.sin(y) * np.cos(z)
    return wx, wy, wz


velocity = pp.AnalyticalField(domain=myDomain3d, formula=computeVel,
                              name='Velocity', vector=True)
vorticity = pp.AnalyticalField(domain=myDomain3d, formula=computeVort,
                               name='Vorticity', vector=True)

############ REF ##############
x = np.arange(localoffset[0], localres[0] + localoffset[0],
              dtype='float64') * hx
y = np.arange(localoffset[1], localres[1] + localoffset[1],
              dtype='float64') * hy
z = np.arange(localoffset[2], localres[2] + localoffset[2],
              dtype='float64') * hz

cden = 4 * pi ** 2 * (Ly ** 2 * Lz ** 2 + Lx ** 2 * Lz ** 2 +
                      Lx ** 2 * Ly ** 2) / (Lx ** 2 * Ly ** 2 * Lz ** 2)
cx = 2 * pi / Lx
cy = 2 * pi / Ly
cz = 2 * pi / Lz

# Initialize vorticity (This init should be improved or done if C or Fortran)
## for k in range(localres[2]):
##     for j in range(localres[1]):
##         for i in range(localres[0]):
##             omega_x[i, j, k] = cden * (m.sin(cx * x[i]) *
##                                        m.sin(cy * y[j]) * m.cos(cz * z[k]))
##             omega_y[i, j, k] = cden * (m.cos(cx * x[i]) *
##                                        m.sin(cy * y[j]) * m.sin(cz * z[k]))
##             omega_z[i, j, k] = cden * (m.cos(cx * x[i]) *
##                                        m.cos(cy * y[j]) * m.sin(cz * z[k]))
##             ref_x[i, j, k] = -cy * (m.cos(cx * x[i]) * m.sin(cy * y[j])
##                                     * m.sin(cz * z[k])) - \
##                 cz * (m.cos(cx * x[i]) * m.sin(cy * y[j]) * m.cos(cz * z[k]))

##             ref_y[i, j, k] = -cz * (m.sin(cx * x[i]) * m.sin(cy * y[j])
##                                     * m.sin(cz * z[k])) + \
##                 cx * (m.sin(cx * x[i]) * m.cos(cy * y[j]) * m.sin(cz * z[k]))

##             ref_z[i, j, k] = -cx * (m.sin(cx * x[i]) * m.sin(cy * y[j])
##                                     * m.sin(cz * z[k])) - \
##                 cy * (m.sin(cx * x[i]) * m.cos(cy * y[j]) * m.cos(cz * z[k]))


# 2 - Advection solver initialisation. See testScales for a working example.
# Based on scales JB solver
# Warning : fields input for scales should be of size (ncells), not localres.
nbProcs = MPI.COMM_WORLD.Get_size()
topodims = [1, 1, nbProcs]  # fits with fftw topology

scalesres, scalesoffset, stab_coeff = \
    scales.init_advection_solver(ncells, myDomain3d.length,
                                 topodims, order='p_O2')

##topofft = poisson.getTopology()
##,ghost=2
topofft = pp.CartesianTopology(domain=myDomain3d,
                               resolution=resolution3d, dim=3)

# 3 - Stretching
stretch = pp.Stretching(vorticity, velocity)

# 4 - Penalization
sphere.discretize(topofft)
sphereD = sphere.discreteObstacle[0]
penal = pp.Penalization(velocity, vorticity, sphereD, lambd)

# Define the problem to solve
navierStokes = pp.Problem(topofft, [penal, stretch])

printer = pp.Printer(fields=[vorticity, velocity], frequency=outputModulo,
                     outputPrefix=outputFilePrefix)

navierStokes.setSolver(finalTime, timeStep, solver_type='basic', io=printer)

## Problem => ParticularSover = basic.initialize()
navierStokes.initSolver()

omega_x = vorticity.discreteField[0][0]
omega_y = vorticity.discreteField[0][1]
omega_z = vorticity.discreteField[0][2]
vx = velocity.discreteField[0][0]
vy = velocity.discreteField[0][1]
vz = velocity.discreteField[0][2]


## end of init ##
# Mind that scales works on arrays of size resolution - 1
# --> pointers to subarrays of velocity/vorticity
#somega_x = np.zeros((localcells), dtype='float64', order='Fortran')
# solve advection
omega_x = scales.solve_advection(timeStep, vx, vy, vz, omega_x)
omega_y = scales.solve_advection(timeStep, vx, vy, vz, omega_y)
omega_z = scales.solve_advection(timeStep, vx, vy, vz, omega_z)

# solve stretching
navierStokes.solve()

print 'vorticity', vorticity.discreteField[0][0] , vorticity.discreteField[0][1] ,  vorticity.discreteField[0][2]

# solve diffusion

nudt = 0.0001
omega_x, omega_y, omega_z = \
    ppfft.solve_diffusion_3d(nudt, vx, vy, vz, omega_x, omega_y, omega_z)

# solve poisson
vx, vy, vz = ppfft.solve_poisson_3d(omega_x, omega_y, omega_z, vx, vy, vz)

## end of time loop ##

# Clean memory buffers
ppfft.clean_fftw_solver(myDomain3d.dimension)
