#!/usr/bin/python
#import sys
#sys.path.insert(0,'/scratch/mimeau/install-parmes3/')
import time
import parmepy as pp
from parmepy.f2py import fftw2py
import mpi4py.MPI as MPI
import math as m
from parmepy.constants import np, PARMES_REAL, ORDER
from parmepy.operator.advection import Advection
from parmepy.operator.density import DensityVisco
from parmepy.operator.stretching import Stretching
from parmepy.operator.diffusion import Diffusion
from parmepy.operator.poisson import Poisson
from parmepy.operator.multiphase import Baroclinic
from parmepy.operator.penalization import Penalization
from parmepy.operator.redistribute import Redistribute
from parmepy.problem.navier_stokes import NSProblem
from parmepy.problem.simulation import Simulation
from parmepy.operator.monitors.energy_enstrophy import Energy_enstrophy
from parmepy.operator.monitors.reprojection_criterion import Reprojection_criterion
from parmepy.operator.monitors.printer import Printer
from parmepy.dataloader import DataLoader

#from numpy import linalg as LA

pi = m.pi

rank = MPI.COMM_WORLD.Get_rank()
print "Mpi process number ", rank

## ----------- A 3d problem -----------
print " ========= Start test for Navier-Stokes 3D (Taylor Green benchmark)========="

## Opening/reading input data file
data = DataLoader('inputData.dat')

# Parameters
dim = 3
nbElem = [data.resX, data.resY, data.resZ]
nbElemFilter = [data.resFilterX, data.resFilterY, data.resFilterZ]
densityVal = [data.rho, data.rho2]
viscoVal = [data.visco, data.visco2]

t0 = time.time()

## Domain
box = pp.Box(dim, length=[data.Lx, data.Ly, data.Lz], 
             origin=[data.Ox, data.Oy, data.Oz])

## Fields declaration
def computeVel(x, y, z):
#--------Richtmyer Meshkov------------
    vx = data.uinf 
    vy = 0.
    vz = 0.
    return vx, vy, vz

def computeVort(x, y, z):
#--------Richtmyer Meshkov------------
    wx = 0.
    wy = 0.
    wz = (data.uinf * 2.0 * pi / data.Ly) * \
          m.sin(2.0 * pi * y / data.Ly)
    return wx, wy, wz

def computeDensity(x, y, z):
#--------Richtmyer Meshkov------------
    perturb = 0.2 * np.sin(2.0 * pi * y / 5.9333)
    d = (3.0 + perturb + 0.5 - x)/(2.0 * 0.5)
    c = 1.0
    if d > 0. and d < 1.0 :
        c = 0.5
#        c = m.exp(-(-m.log10(0.)) * m.fabs(d)^8.0)
    elif d >= 1.0 :
        c = 0. 
    rho = densityVal[0] + (densityVal[1] - densityVal[0]) * c
    return rho

def computeVisco(x, y, z):
#--------Richtmyer Meshkov------------
    perturb = 0.2 * np.sin(2.0 * pi * y / 5.9333)
    d = (3.0 + perturb + 0.5 - x)/(2.0 * 0.5)
    c = 1.0
    if d > 0. and d < 1.0 :
        c = 0.5
#        c = m.exp(-(-m.log10(0.)) * m.fabs(d)^8.0)
    elif d >= 1.0 :
        c = 0. 
    nu = viscoVal[0] + (viscoVal[1] - viscoVal[0]) * c
    return nu

## Fields
velo = pp.Field(domain=box, formula=computeVel, 
                name='Velocity', isVector=True)
vorti = pp.Field(domain=box, formula=computeVort, 
                 name='Vorticity', isVector=True)
rho = pp.Field(domain=box, formula=computeDensity,
                 name='Density', isVector=False)
visco = pp.Field(domain=box, formula=computeVisco,
                 name='Viscosity', isVector=False)

## Operators
advecVort = Advection(velo, vorti,
                      resolutions={velo: nbElem,
                                   vorti: nbElem},
                      method=data.methodAdv
                     )

stretch = Stretching(velo, vorti,
                     resolutions={velo: nbElem,
                                  vorti: nbElem},
                     method=data.methodStretch,
                     propertyOp=data.propStretch
                    )

diffusion = Diffusion(vorti,
                      resolution=nbElem,
                      method='fftw',
                      viscosity=data.visco
                     )

baroclinic = Baroclinic(velo, vorti, rho, visco,
                        resolutions={velo: nbElem,
                                     vorti: nbElem,
                                     rho: nbElem,
                                     visco: nbElem},
                        method=data.methodBaro
                       )

advecScalar = Advection(velo, rho,
                        resolutions={velo: nbElem,
                                     rho: nbElem},
                        method=data.methodAdvSc
                       )

density = DensityVisco(rho, visco,
                       resolutions={rho: nbElem,
                                    visco: nbElem},
                       method='',
                       densityVal = densityVal,
                       viscoVal = viscoVal
                      )

poisson = Poisson(velo, vorti,
                  resolutions={velo: nbElem,
                               vorti: nbElem},
                  method='fftw',
                  projection=data.vortProj,
                  multires=data.multires,
                  filterSize=box.length / \
                             (np.asarray(nbElemFilter, 
                                         dtype=PARMES_REAL) - 1)
                 )

## Diagnostics related to the problem
energy = Energy_enstrophy(velo, vorti,
                          resolutions={velo: nbElem,
                                       vorti: nbElem},
                          viscosity=data.visco, 
                          frequency=data.outModuloEnergies,
                          prefix='./res/Energies_test.dat')

#reproj = Reprojection_criterion(velo, vorti,
#                                resolutions={velo: nbElem,
#                                             vorti: nbElem},
#                                method='FD_order4',
#                                frequency=1,
#                                prefix='../res/Reproj.dat')

printer = Printer(fields=[rho],
                  frequency=data.outModuloPrinter,
                  prefix=data.outputPrinter,
                  ext='.vtk'
                 )

distrAdvStr = Redistribute([vorti, velo], advecVort, stretch)
distrStrDiff = Redistribute([vorti], stretch, diffusion)
distrDiffBaro = Redistribute([vorti], diffusion, baroclinic)
distrBaroPoiss = Redistribute([vorti], baroclinic, poisson)
distrDensBaro = Redistribute([rho, visco], density, baroclinic)

## Definition of simulation parameters
simu = Simulation(tinit=0.0, tend=data.finalTime, timeStep=data.timeStep, iterMax=1000000)

## Define the problem to solve
#pb = NSProblem([advecVort, stretch, diffusion, baroclinic, advecScalar, density, poisson],
pb = NSProblem(operators=[advecVort, distrAdvStr, stretch, distrStrDiff, diffusion,
                          distrDiffBaro, baroclinic, distrBaroPoiss, advecScalar, 
                          density, poisson, distrDensBaro],
#pb = NSProblem(operators=[advecVort, distrAdvStr, stretch, distrStrDiff, diffusion,
#                          advecScalar, density, poisson],
               simulation=simu, 
               monitors=[energy], 
               dumpFreq=int(data.dumpfreq), name=data.dumpfile)
## Setting solver to Problem
pb.setUp()

## Restarting Problem from dumped field values 
if data.restart :
    pb.restart(data.dumpfile)

t1 = time.time()

## Solve problem
#poisson.apply(simu)
timings = pb.solve()

tf = time.time()

print "\n"
print "Total time : ", tf - t0, "sec (CPU)"
print "Init time : ", t1 - t0, "sec (CPU)"
print "Solving time : ", tf - t1, "sec (CPU)"

## end of time loop ##

# Clean memory buffers
fftw2py.clean_fftw_solver(box.dimension)
