#!/usr/bin/python

"""
Navier Stokes 3D : flow past bluff bodies (using penalization).

All parameters are set and defined in python module dataNS_bb.

"""

import parmepy as pp
from parmepy.f2py import fftw2py
import numpy as np
from parmepy.fields.continuous import Field
from parmepy.mpi.topology import Cartesian
from parmepy.fields.variable_parameter import VariableParameter
from parmepy.domain.obstacle.sphere import Sphere
from parmepy.operator.advection import Advection
from parmepy.operator.stretching import Stretching
from parmepy.operator.poisson import Poisson
from parmepy.operator.velocity_correction import VelocityCorrection
#from parmepy.operator.vorticity_correction import VorticityCorrection
from parmepy.operator.diffusion import Diffusion
from parmepy.operator.penalization import Penalization
from parmepy.operator.differential import Curl
from parmepy.operator.redistribute import Redistribute
from parmepy.operator.monitors import Printer, Energy_enstrophy, DragAndLift,\
    Reprojection_criterion
from parmepy.problem.simulation import Simulation
from parmepy.constants import VTK, HDF5
from parmepy.domain.obstacle.planes import PlaneBoundaries, SubPlane
from parmepy.domain.obstacle.controlBox import ControlBox
from dataNS_bb import dim, nb, NBGHOSTS, ADVECTION_METHOD, VISCOSITY, \
    OUTPUT_FREQ, FILENAME, PROJ, LCFL, CFL, CURL_METHOD, \
    TIMESTEP_METHOD, OBST_Ox, OBST_Oy, OBST_Oz, RADIUS
import parmepy.tools.numpywrappers as npw
import parmepy.tools.io_utils as io
## ----------- A 3d problem -----------
print " ========= Start Navier-Stokes 3D (Flow past bluff bodies) ========="

## pi constant
pi = np.pi
cos = np.cos
sin = np.sin

## Domain
#boxlength = npw.realarray([8.0 * pi, 2.0 * pi, 2.0 * pi])
#boxorigin = npw.realarray([-2.0, -pi, -pi])
boxlength = npw.realarray([10.24, 5.12, 5.12])
boxorigin = npw.realarray([-2.0, -2.56, -2.56])
box = pp.Box(dim, length=boxlength, origin=boxorigin)

nbElem = [129, 65, 65]


## Global resolutionsimu
#nbElem = [1025, 513, 513]


# Initial upstream flow velocity
uinf = 1.0

def computeFlowrate(simu):
    ##Time-dependant flow rate
    t = simu.tk
    Tstart = 3.0
    flowrate = np.zeros(3)
    flowrate[0] = uinf * box.length[1] * box.length[2]
    if t >= Tstart and t <= Tstart + 1.0:
        flowrate[1] = sin(pi * (t - Tstart)) * \
                      box.length[1] * box.length[2]
    ##constant flow rate
#    flowrate = np.zeros(3)
#    flowrate[0] = uinf * box.length[1] * box.length[2]
    return flowrate

# Function to compute initial velocity
def computeVel(res, x, y, z, t):
    res[0][...] = uinf
    res[1][...] = 0.
    res[2][...] = 0.
    return res


# Function to compute initial vorticity
def computeVort(res, x, y, z, t):
    res[0][...] = 0.
    res[1][...] = 0.
    res[2][...] = 0.
    return res

## Vector Fields
velo = Field(domain=box, formula=computeVel,
             name='Velocity', isVector=True)
vorti = Field(domain=box, formula=computeVort,
              name='Vorticity', isVector=True)

## Usual Cartesian topology definition
# At the moment we use two (or three?) topologies :
# - "topo" for Stretching and all operators based on finite differences.
#    --> ghost layer = 2
# - topo from Advection operator for all the other operators.
#    --> no ghost layer
# - topo from fftw for Poisson and Diffusion.
# Todo : check compat between scales and fft operators topologies.
ghosts = np.ones((box.dimension)) * NBGHOSTS
topostr = Cartesian(box, box.dimension, nbElem,
                    ghosts=ghosts)

## === Obstacles (sphere + up and down plates) ===
#sphere_pos = (boxlength - boxorigin) * 0.5
sphere_pos = [0., 0., 0.]
sphere = Sphere(box, position=sphere_pos, radius=RADIUS)

## === Computational Operators ===
op = {}
op['advection'] = Advection(velo, vorti,
                            resolutions={velo: nbElem, vorti: nbElem},
                            method=ADVECTION_METHOD)

op['stretching'] = Stretching(velo, vorti,
                              resolutions={velo: nbElem, vorti: nbElem},
                              topo=topostr)

op['diffusion'] = Diffusion(vorti, resolution=nbElem, viscosity=VISCOSITY)

op['poisson'] = Poisson(velo, vorti, resolutions={velo: nbElem, vorti: nbElem})

op['curl'] = Curl(velo, vorti, resolutions={velo: nbElem, vorti: nbElem},
                  method=CURL_METHOD)

## Discretisation of computational operators
for ope in op.values():
    ope.discretize()

# Get fft topology and use it for penalization and correction operators
opfft = op['poisson']
topofft = opfft.discreteFields[opfft.vorticity].topology
topocurl = op['curl'].discreteFields[op['curl'].invar].topology
topoadvec = op['advection'].discreteFields[op['advection'].velocity].topology
# space step
ref_step = topofft.mesh.space_step
step_dir = ref_step[0]

## Penalization velocity
op['penalization'] = Penalization(velo, [sphere], coeff=[1e8],
                                  topo=topofft, resolutions={velo: nbElem})

## Time-dependant required-flowrate (Variable Parameter)
req_flowrate = VariableParameter(formula=computeFlowrate)
op['correction'] = VelocityCorrection(velo, vorti,
                                      resolutions={velo: nbElem,
                                                   vorti: nbElem},
                                      req_flowrate=req_flowrate, topo=topofft)

op['penalization'].discretize()
op['correction'].discretize()

## === Bridges between the different topologies ===
distr = {}

distr['fft2curl'] = Redistribute([velo], op['penalization'], op['curl'])
distr['curl2fft'] = Redistribute([vorti], op['curl'], op['poisson'])
distr['adv2str'] = Redistribute([velo, vorti],
                                op['advection'], op['stretching'])
distr['str2diff'] = Redistribute([vorti], op['stretching'], op['diffusion'])
distr['curl2adv'] = Redistribute([velo, vorti], op['curl'], op['advection'])
distr['curl2str'] = Redistribute([velo, vorti], op['curl'], op['stretching'])

#distr['adv2corr'] = Redistribute([velo, vorti],
#                                op['advection'], op['correc_vorti'])

for ope in distr.values():
    ope.discretize()

# === Simulation setup ===
simu = Simulation(tinit=0.0, tend=75., timeStep=0.0125, iterMax=1000000)

## === Monitoring operators ===
monitors = {}
# Save velo and vorti values on topofft
monitors['printerFFT'] = Printer(variables=[velo, vorti], topo=topofft,
                                 formattype=HDF5)

monitors['printerCurl'] = Printer(variables=[velo, vorti], topo=topocurl,
                                  formattype=HDF5, prefix='curl_io_vorti')

monitors['printerAdvec'] = Printer(variables=[velo, vorti], topo=topofft,
                                   formattype=HDF5, prefix='advec_io',
                                   xmfalways=True)

thickSliceXY = ControlBox(box, origin=[-2.0, -2.56, -2.0 * step_dir], 
                          lengths=[10.24, 5.12, 4.0 * step_dir])
#thickSliceXY = ControlBox(box, origin=[-2.0, -pi, -2.0 * step_dir], 
#                        lengths=[8.0*pi, 2.0*pi, 4.0 * step_dir])
monitors['printerSliceXY'] = Printer(variables=[velo, vorti], topo=topofft,
                                   frequency=100, formattype=HDF5, prefix='sliceXY', 
                                   subset=thickSliceXY, xmfalways=True)

thickSliceXZ = ControlBox(box, origin=[-2.0, -2.0 * step_dir, -2.56], 
                        lengths=[10.24, 4.0 * step_dir, 5.12])
monitors['printerSliceXZ'] = Printer(variables=[velo, vorti], topo=topofft,
                                   frequency=100, formattype=HDF5, prefix='sliceXZ', 
                                   subset=thickSliceXZ, xmfalways=True)

subBox = ControlBox(box, origin=[-0.7, -1.25, -1.25], lengths=[7.0, 2.5, 2.5])
monitors['printerSubBox'] = Printer(variables=[velo, vorti], topo=topofft,
                                   frequency=100, formattype=HDF5, prefix='subBox', 
                                   subset=subBox, xmfalways=True)
# Compute and save enstrophy/Energy, on topofft
io_default = {}
monitors['energy'] = Energy_enstrophy(velo, vorti, topo=topofft,
                                      io_params={},
                                      viscosity=VISCOSITY, isNormalized=False)

# Compute velocity and vorticity magnitude on topofft and print it in data file
#io_default = {}
#monitors['profiles'] = ProfilesOutput(velo, vorti, obstacles=[sphere],
#                                 topo=topofft, io_params={"frequency":1000})

# Setup for reprojection
reprojection_constant = 0.04
reprojection_rate = 1
monitors["reprojection"] = Reprojection_criterion(vorti, reprojection_constant,
                                                  reprojection_rate, topofft,
                                                  io_params={})

# Add reprojection for poisson operator
op["poisson"].activateProjection(monitors["reprojection"])

ind = sphere.discretize(topofft)
vdfft = velo.discreteFields[topofft].data

# Compute and save drag and lift on topofft
# 3D Control box
cbpos = boxorigin + 10 * ref_step
cblength = boxlength - 25 * ref_step
cb = ControlBox(box, cbpos, cblength)
coeffForce = 1. / (0.5 * uinf ** 2 * pi * RADIUS ** 2)

#print coeffForce
# Monitor for forces
monitors['forces'] = DragAndLift(velo, vorti, VISCOSITY, coeffForce,
                                  topostr, cb, obstacles=[sphere], io_params={})
#monitors['forces'] = DragAndLift(velo, vorti, VISCOSITY, coeffForce,
#                                 topo, cb, io_params={})

# === Setup for all declared operators ===
for ope in op.values():
    ope.setUp()
for ope in distr.values():
    ope.setUp()
# Set up for monitors
for monit in monitors.values():
    monit.setUp()

allops = dict(op.items() + distr.items() + monitors.items())


## === Fields initialization ===
# Initialisation, mode 1:
# - compute velocity on topofft
# - penalize
# - compute vorticity with curl
def initFields_mode1():
    # The velocity is initialized on the fftw topology
    # penalized and distributed on curl topo
    velo.initialize(topo=topofft)
    op['penalization'].apply(simu)
    distr['fft2curl'].apply(simu)
    # Vorticity is initialized after penalization of velocity
    op['curl'].apply(simu)
    distr['curl2adv'].apply(simu)
    op['advection'].apply(simu)
    # Distribute vorticity on topostr
    distr['adv2str'].apply(simu)
    # From this point both velocity and vorticity are initialized
    # on topocurl and topoadvec. velocity is also init. on topofft.

## Call initialization
initFields_mode1()

##### Check initialize process results #####
norm_vel_fft = velo.norm(topofft)
norm_vel_curl = velo.norm(topocurl)
norm_vort_fft = vorti.norm(topofft)
norm_vort_curl = vorti.norm(topocurl)

#assert np.allclose(norm_vort_curl, norm_vort_fft)
assert np.allclose(norm_vel_curl, norm_vel_fft)

# Print initial state
for mon in monitors.values():
    mon.apply(simu)

## Note Franck : tests init ok, same values on both topologies, for v and w.
# === After init ===
#
# v init on topocurl (which may be equal to topofft)
#
# === Definition of the 'one-step' sequence of operators ===
#
#  1 - w = curl(v), topocurl
#  2 - w = advection(v,w), topo-advec
#  3 - w = stretching(v,w), topostr
#  4 - w = diffusion(w), topofft
#  5 - v = poisson(w), topofft
#  6 - v = correction(v, w), topofft
#  7 - v = penalization(v), topofft
#
#  Required bridges :
# 1 --> 2 : (v,w) on topo-advec
# 2 --> 3 : (v,w) on topostr
# 3 --> 4 : w on topofft
# 7 --> 1 : v on topocurl

#fullseq = ['advection',  # in: v,w out: w, on topoadvec
#            'adv2str', 'stretching',  # in: v,w out: w on topostr
#            # in: w, out: v, w on topofft
#            'str2diff', 'diffusion', 'reprojection', 'poisson', 'correction',
#            'penalization',
#            'fft2curl', 'curl', 'forces',
#            'curl2adv', 'energy', 'profiles', 'printerAdvec',
#            ]

fullseq = ['stretching',
           'str2diff', 'diffusion', 'poisson', 'correction',
           'penalization',
           'fft2curl', 'curl',
           'curl2fft', 'poisson', 'correction',
           'advection',
           # 'printerSliceXY', 'printerSliceXZ', 'printerSubBox',
           'energy', 'adv2str', 'forces']


def run(sequence):
    for name in sequence:
        assert allops.keys().count(name) > 0 or distr.keys().count(name) > 0\
            or monitors.keys().count(name) > 0, 'unknow key:' + name
        allops[name].apply(simu)

seq = fullseq

simu.initialize()

while not simu.isOver:
    if topocurl.rank == 0:
        simu.printState()
    run(seq)
    simu.advance()


## print 'total time (rank):', MPI.Wtime() - time, '(', topo.rank, ')'

# === Finalize for all declared operators ===

# Note FP : bug in fft finalize. To be checked ...
# --> clean_fftw must be called only once
#for ope in op.values():
#    ope.finalize()
## Clean memory buffers
fftw2py.clean_fftw_solver(box.dimension)
for ope in distr.values():
    ope.finalize()
for monit in monitors.values():
    monit.finalize()
