#!/usr/bin/python

"""
Navier Stokes 3D : flow past bluff bodies (using penalization).

All parameters are set and defined in python module dataNS_bb.

"""

import parmepy as pp
from parmepy.f2py import fftw2py
import numpy as np
from parmepy.fields.continuous import Field
from parmepy.variables.variables import Variables
from parmepy.mpi.topology import Cartesian
from parmepy.domain.obstacle.sphere import Sphere
from parmepy.operator.advection import Advection
from parmepy.operator.stretching import Stretching
from parmepy.operator.poisson import Poisson
from parmepy.operator.velocity_correction import VelocityCorrection
from parmepy.operator.diffusion import Diffusion
from parmepy.operator.penalization import Penalization
from parmepy.operator.differential import Curl
from parmepy.operator.redistribute import Redistribute
from parmepy.operator.adapt_timestep import AdaptTimeStep
from parmepy.problem.navier_stokes import NSProblem
from parmepy.operator.monitors.printer import Printer
from parmepy.operator.monitors.energy_enstrophy import Energy_enstrophy
from parmepy.operator.monitors.compute_forces import Forces
from parmepy.problem.simulation import Simulation
from parmepy.constants import VTK
from dataNS_bb import dim, NBGHOSTS, ADVECTION_METHOD, VISCOSITY, \
    OUTPUT_FREQ, FILENAME, PROJ, LCFL, CFL, CURL_METHOD, \
    TIMESTEP_METHOD, OBST_Ox, OBST_Oy, OBST_Oz, RADIUS
from parmepy.domain.obstacle.planes import PlaneBoundaries

## ----------- A 3d problem -----------
print " ========= Start Navier-Stokes 3D (Flow past bluff bodies) ========="

## pi constant
pi = np.pi
cos = np.cos
sin = np.sin

## Domain
#box = pp.Box(dim, length=[1., 1., 1.], origin=[0., 0., 0.])
box = pp.Box(dim, length=[2.0 * pi, pi,  pi], origin=[-pi, -pi, -pi])
#box = pp.Box(dim, length=[12., 10., 10.], origin=[-4., -5., -5.])
#box = pp.Box(dim, length=[2.0 * pi, 2.0 * pi, 2.0 * pi])

## Global resolution
#nbElem = [nb] * dim
nbElem = [65, 33, 33]

# Upstream flow velocity
uinf = 1.0

# Function to compute potential flow past a sphere
def computeVel(res, x, y, z, t):
    module = np.sqrt((x - OBST_Ox) * (x - OBST_Ox) +
                     (y - OBST_Oy) * (y - OBST_Oy) +
                     (z - OBST_Oz) * (z - OBST_Oz))
    
#    print 'module', module, np.shape(module)
    for i in range (np.shape(module)[0]):
        for j in range (np.shape(module)[1]):
            for k in range (np.shape(module)[2]):
                if module[i,j,k]==0.0:
                    module[i,j,k] = RADIUS
    res[0][...] = - ((uinf * x) / module) * (1 - (RADIUS ** 3 / module ** 3))
    res[1][...] = ((uinf * y) / module) * (1 + (RADIUS ** 3 / (2. * module ** 2)))
    res[2][...] = 0.
#    res[0][...] = uinf
#    res[1][...] = 0.
#    res[2][...] = 0.
    return res

# Function to compute initial vorticity
def computeVort(res, x, y, z, t):
    res[0][...] = 0.
    res[1][...] = 0.
    res[2][...] = 0.
    return res

## Function to compute TG velocity
#def computeVel(res, x, y, z, t):
#    res[0][...] = sin(x) * cos(y) * cos(z)
#    res[1][...] = - cos(x) * sin(y) * cos(z)
#    res[2][...] = 0.
#    return res

## Function to compute reference vorticity
#def computeVort(res, x, y, z, t):
#    res[0][...] = - cos(x) * sin(y) * sin(z)
#    res[1][...] = - sin(x) * cos(y) * sin(z)
#    res[2][...] = 2. * sin(x) * sin(y) * cos(z)
#    return res

## Fields
velo = Field(domain=box, formula=computeVel,
             name='Velocity', isVector=True)
vorti = Field(domain=box, formula=computeVort,
              name='Vorticity', isVector=True)

## Variables
dt_adapt = Variables(domain=box, name='adaptative time step', 
                     data=[0.01])

## Usual Cartesian topology definition
# At the moment we use two (or three?) topologies :
# - "topo" for Stretching and all operators based on finite differences.
#    --> ghost layer = 2
# - topo from Advection operator for all the other operators.
#    --> no ghost layer
# - topo from fftw for Poisson and Diffusion.
# Todo : check compat between scales and fft operators topologies.
ghosts = np.ones((box.dimension)) * NBGHOSTS
topo = Cartesian(box, box.dimension, nbElem,
                 ghosts=ghosts)

## Obstacles (sphere + up and down plates)
sphere = Sphere(box, position=[OBST_Ox, OBST_Oy, OBST_Oz],
                radius=RADIUS)

#plates = Plates(box, normal_dir=2, epsilon=0.005)

## Tools Operators
dtAdapt = AdaptTimeStep(velo, vorti,
                        resolutions={velo: nbElem,
                                     vorti: nbElem},
                        dt_adapt=dt_adapt,
                        method=TIMESTEP_METHOD,
                        topo=topo,
                        lcfl=LCFL,
                        cfl=CFL)

## Operators
advec = Advection(velo, vorti,
                  resolutions={velo: nbElem,
                               vorti: nbElem},
                  method=ADVECTION_METHOD
                  )

stretch = Stretching(velo, vorti,
                     resolutions={velo: nbElem,
                                  vorti: nbElem},
                     topo=topo
                     )

diffusion = Diffusion(vorti,
                      resolution=nbElem,
                      viscosity=VISCOSITY
                     )

poisson = Poisson(velo, vorti,
                  resolutions={velo: nbElem,
                               vorti: nbElem},
                  projection=PROJ)

curl = Curl(velo, vorti,
            resolutions={velo: nbElem,
                         vorti: nbElem},
            method=CURL_METHOD)
#            topo=topo)

## Bridges between the different topologies in order to
## redistribute data.
# 1 - Curl_fft to stretching (velocity only)
distrCurlStr_velo = Redistribute([velo], curl, stretch)
distrAdvStr_velo = Redistribute([velo], advec, stretch)
# 2 - Advection to stretching (vorticity only)
distrAdvStr_vorti = Redistribute([vorti], advec, stretch)
# 3 - Stretching to Diffusion
distrStrDiff = Redistribute([vorti], stretch, diffusion)
# + - Brigdes required if Curl op. is solved using a FD method
distrPoissCurl = Redistribute([vorti, velo], poisson, curl)
distrCurlAdv = Redistribute([vorti, velo], curl, advec)

## Simulation with fixed time step
simu = Simulation(tinit=0.0,
                  tend=10., timeStep=0.005,
                  iterMax=1000000)

## Define the problem to solve

#pb = NSProblem(operators=[distrPoissCurl,
#                          curl,
#                          distrCurlAdv,
#                          advec,             # <--------------------:
#                          distrAdvStr_vorti, 
#                          stretch, 
#                          distrStrDiff,
#                          diffusion, 
#                          poisson],
#               simulation=simu, dumpFreq=-1)

pb = NSProblem(operators=[curl,
                          distrCurlAdv,
                          distrCurlStr_velo, # Redistribute while advec
                          advec,             # <--------------------:
                          distrAdvStr_vorti, 
                          stretch, 
                          distrStrDiff,
                          diffusion, 
                          poisson],
               simulation=simu, dumpFreq=-1)

## Setting solver to Problem (only operators for computational tasks)
pb.pre_setUp()

## Topology without ghost points
topofft = poisson.discreteFields[poisson.vorticity].topology

## Add operators related to the problem and depending on FFT topology
correction = VelocityCorrection(velo, vorti,
                                resolutions={velo: nbElem,
                                             vorti: nbElem},
                                uinf=uinf,
                                topo=topofft)

bc = PlaneBoundaries(box, 2, thickness=0.1)

penal = Penalization(velo, [sphere, bc],
                     coeff=[1e8],
                     topo=topofft,
                     resolutions={velo: nbElem})

## Diagnostics/monitors related to the problem

forces = Forces(velo, vorti, topo, 
                obstacle=sphere, boxMin= [0.2, 0.2, 0.2], 
                boxMax= [0.8, 0.8, 0.8], Reynolds=300., uinf=uinf,
                frequency=1, prefix='./res/Noca_sphere.dat')

printer = Printer(variables=[velo, vorti],
                  topo=topofft,
                  frequency=100,
                  prefix='./res/NS_sphere',
                  formattype=VTK)

energy = Energy_enstrophy(velo, vorti,
                          topo=topofft,
                          viscosity=VISCOSITY,
                          isNormalized=True,
                          frequency=OUTPUT_FREQ,
                          prefix=FILENAME)

## Add monitors and setting solver to Problem
vorti.setTopoInit(topofft)
velo.setTopoInit(topofft)
pb.addMonitors([correction, penal, energy, printer])
penal.discretize()
correction.discretize()
correction.setUp()
penal.setUp()
pb.setUp()

penal.apply(simu)

def run():
## Solve problem
    pb.solve()

## end of time loop ##

from parmepy.mpi import MPI
print "Start computation ..."
time = MPI.Wtime()
run()
print 'total time (rank):', MPI.Wtime() - time, '(', topo.rank, ')'

## Clean memory buffers
fftw2py.clean_fftw_solver(box.dimension)
