#!/usr/bin/python

"""
Taylor Green 3D : see paper van Rees 2011.

All parameters are set and defined in python module dataTG.

"""

import parmepy as pp
from parmepy.constants import ORDER, np, PARMES_REAL, HDF5
from parmepy.f2py import fftw2py
from parmepy.fields.continuous import Field
from parmepy.mpi.topology import Cartesian
from parmepy.operator.advection import Advection
from parmepy.operator.stretching import Stretching
from parmepy.operator.poisson import Poisson
from parmepy.operator.diffusion import Diffusion
from parmepy.operator.redistribute import Redistribute
from parmepy.problem.navier_stokes import NSProblem
from parmepy.operator.monitors.printer import Printer
from parmepy.operator.monitors.energy_enstrophy import Energy_enstrophy
from parmepy.problem.simulation import Simulation
from parmepy.operator.differential import Curl
from parmepy.mpi.main_var import main_size
from parmepy.methods_keys import Scales, TimeIntegrator, Interpolation,\
    Remesh, Support, Splitting, dtAdvecCrit, SpaceDiscretisation, GhostUpdate, MultiScale
from parmepy.numerics.integrators.runge_kutta2 import RK2 as RK2
from parmepy.numerics.integrators.runge_kutta3 import RK3 as RK3
from parmepy.numerics.integrators.runge_kutta4 import RK4 as RK4
from parmepy.numerics.finite_differences import FD_C_4, FD_C_2
from parmepy.numerics.interpolation import Linear
from parmepy.numerics.remeshing import L6_4
from parmepy.f2py import fftw2py

# problem dimension
dim = 3
# resolution
nb = 129
nb_s = 257
# number of ghosts in usual cartesian topo
NBGHOSTS = 2
ADVECTION_METHOD = {Scales: 'p_64', MultiScale: 'L4_4'}
# ADVECTION_METHOD = {TimeIntegrator: RK2,
#                     Interpolation: Linear,
#                     Remesh: L6_4,
#                     Support: '',
#                     Splitting: 'o2_FullHalf'}
VISCOSITY = 1e-4


## ----------- A 3d problem -----------
## pi constant
pi = np.pi
cos = np.cos
sin = np.sin
exp = np.exp
abs = np.abs
tanh = np.tanh
## Domain
box = pp.Box(dim, length=[1., ]*3)

## Global resolution
nbElem = [nb] * dim
nbElem_s = [nb_s] * dim
randX = np.asarray(np.random.random([nb - 1, nb - 1,(nb - 1)/main_size]),
                   dtype=PARMES_REAL, order=ORDER) - 0.5
randY = np.asarray(np.random.random([nb - 1, nb - 1,(nb - 1)/main_size]),
                   dtype=PARMES_REAL, order=ORDER) - 0.5
randZ = np.asarray(np.random.random([nb - 1, nb - 1,(nb - 1)/main_size]),
                   dtype=PARMES_REAL, order=ORDER) - 0.5

##initjet.f
width = 0.01
ampl3 = 0.3
ampl = 0.05
ampl2=0.05
def computeVel(res, x, y, z, t):
    yy = abs(y - 0.5)
    aux = (0.1 - 2. * yy) / (4. * width)
    strg1 = exp(-abs(aux ** 2)) * randX
    strg2 = exp(-abs(aux ** 2)) * randY
    strg3 = exp(-abs(aux ** 2)) * randZ
    res[0][...] = 0.5 * (1. + tanh(aux))
    res[0][...] *= (1. + ampl3 * sin(8. * pi * x))
    res[0][...] *= (1. + ampl * strg1)
    res[1][...] = ampl * strg2
    res[2][...] = ampl * strg3
    return res


def initScal(res, x, y, z, t):
    yy = abs(y - 0.5)
    aux = (0.1 - 2. * yy) / (4. * width)
    res[0][...] = 0.5 * (1. + tanh(aux))
    #strg3 = exp(-abs(aux ** 2)) * randZ
    #res[0][...] *= (1. + ampl2 * strg3)
    return res

## Fields
velo = Field(domain=box, formula=computeVel,
             name='Velocity', isVector=True)
vorti = Field(domain=box,
              name='Vorticity', isVector=True)
scal = Field(domain=box, formula=initScal,
              name='PassiveScalar', isVector=False)

## Variables
## Usual Cartesian topology definition
# At the moment we use two (or three?) topologies :
# - "topo" for Stretching and all operators based on finite differences.
#    --> ghost layer = 2
# - topo from Advection operator for all the other operators.
#    --> no ghost layer
# - topo from fftw for Poisson and Diffusion.
# Todo : check compat between scales and fft operators topologies.
ghosts = [NBGHOSTS] * box.dimension
topo = Cartesian(box, box.dimension, nbElem,
                 ghosts=ghosts)

## Navier Stokes Operators
advec = Advection(velo, vorti,
                  resolutions={velo: nbElem,
                               vorti: nbElem},
                  method=ADVECTION_METHOD
                  )
advecScal = Advection(velo, scal,
                      resolutions={velo: nbElem,
                                   scal: nbElem_s},
                      method=ADVECTION_METHOD
                      )

stretch = Stretching(velo, vorti,
                     resolutions={velo: nbElem,
                                  vorti: nbElem},
                     topo=topo
                     )

diffusion = Diffusion(vorti,
                      resolution=nbElem,
                      viscosity=VISCOSITY
                      )

poisson = Poisson(velo, vorti,
                  resolutions={velo: nbElem,
                               vorti: nbElem},
                  projection=[True, 1, False],
                  )

c = Curl(velo, vorti,
         resolutions={velo: nbElem,
                      vorti: nbElem},
         method={SpaceDiscretisation: fftw2py, GhostUpdate: True},
         )

## Simulation
simu = Simulation(tinit=0.0,
                  tend=4.5, timeStep=0.001,
                  iterMax=1000000)


# Bridges between the different topologies in order to
# redistribute data.
# 1 -Advection to stretching
toGhost_vorti = Redistribute([vorti], advec, stretch)
toGhost_velo = Redistribute([velo], poisson, stretch)
# 2 - Stretching to Poisson/Diffusion
fromGhost_vorti = Redistribute([vorti], stretch, diffusion)

#  Define the problem to solve
pb = NSProblem(operators=[
        toGhost_velo,
        advecScal,
        advec,
        toGhost_vorti,
        stretch,
        fromGhost_vorti,
        diffusion,
        poisson,
        ], simulation=simu, dumpFreq=-1)

## Setting solver to Problem (o nly operators for computational tasks)
pb.pre_setUp()
## Diagnostics related to the problem
topofft = poisson.discreteFields[poisson.vorticity].topology
energy = Energy_enstrophy(velo, vorti, isNormalized=True,
                          topo=topofft,
                          viscosity=VISCOSITY,
                          frequency=1,
                          prefix='./res_NS_planeJet/energy_ref_128-256.dat')

# initialisation
vorti.setTopoInit(topofft)
velo.setTopoInit(topofft)
#scal.setTopoInit(topofft)
pb.addMonitors([energy])
pb.setUp()


c.discretize()
c.setUp()

c.apply(simu)

p = Printer(variables=[scal,],
            topo=scal.discreteFields.keys()[0],
            frequency=100,
            prefix='./res_NS_planeJet/scal_ref_128-256',
            formattype=HDF5)
pb.addMonitors([p])
p._printStep()


def run():
    pb.solve()

## Solve problem
from parmepy.mpi import MPI
print "Start computation ..."
time = MPI.Wtime()
run()
print 'total time (rank):', MPI.Wtime() - time, '(', topo.rank, ')'


pb.finalize()
## Clean memory buffers
