import parmepy as pp
from parmepy.constants import ORDER, np, PARMES_REAL, HDF5
from parmepy.f2py import fftw2py
from parmepy.fields.continuous import Field
from parmepy.mpi.topology import Cartesian
from parmepy.operator.advection import Advection
from parmepy.operator.stretching import Stretching
from parmepy.operator.poisson import Poisson
from parmepy.operator.diffusion import Diffusion
from parmepy.operator.redistribute import Redistribute
from parmepy.operator.redistribute_intercomm import RedistributeIntercomm
from parmepy.operator.monitors import Printer, Energy_enstrophy, DragAndLift,\
    Reprojection_criterion
from parmepy.problem.simulation import Simulation
from parmepy.operator.differential import Curl
from parmepy.methods_keys import Scales, TimeIntegrator, Interpolation,\
    Remesh, Support, Splitting, dtCrit, SpaceDiscretisation, GhostUpdate, MultiScale
from parmepy.numerics.integrators.runge_kutta2 import RK2 as RK2
from parmepy.numerics.integrators.runge_kutta3 import RK3 as RK3
from parmepy.numerics.integrators.runge_kutta4 import RK4 as RK4
from parmepy.numerics.finite_differences import FD_C_4, FD_C_2
from parmepy.numerics.interpolation import Linear
from parmepy.numerics.remeshing import L6_4, L4_4
from parmepy.f2py import fftw2py
from parmepy.mpi.main_var import main_comm, main_rank, main_size
from parmepy.problem.problem_tasks import ProblemTasks
from parmepy.fields.variable_parameter import VariableParameter
from parmepy.operator.adapt_timestep import AdaptTimeStep

NS_TASK = 0
PS_TASK = 1

assert main_size > 1
proc_tasks = [NS_TASK,] * main_size
proc_tasks[0] = PS_TASK
if main_size > 5:
    proc_tasks[5] = PS_TASK
comm_s = main_comm.Split(color=proc_tasks[main_rank], key=main_rank)
comm_s_size = comm_s.Get_size()

pi = np.pi
cos = np.cos
sin = np.sin
exp = np.exp
abs = np.abs
tanh = np.tanh

## Global resolution
dim = 3
nb = 65
nb_s = 129
nbElem = [nb] * dim
nbElem_s = [nb_s] * dim
if proc_tasks[main_rank] == NS_TASK:
    try:
        rand_file = "{0}_{1}_{2}".format(*tuple([nb-1] * 3))
        rand_file += "_{0}p_{1}".format(comm_s_size, comm_s.Get_rank())
        randX = np.asarray(np.reshape(np.fromfile('rand/rand_X_' + rand_file),
                                      [nb - 1, nb - 1,(nb - 1)/comm_s_size]),
                           dtype=PARMES_REAL, order=ORDER)
        randY = np.asarray(np.reshape(np.fromfile('rand/rand_Y_' + rand_file),
                                      [nb - 1, nb - 1,(nb - 1)/comm_s_size]),
                           dtype=PARMES_REAL, order=ORDER)
        randZ = np.asarray(np.reshape(np.fromfile('rand/rand_Z_' + rand_file),
                                      [nb - 1, nb - 1,(nb - 1)/comm_s_size]),
                           dtype=PARMES_REAL, order=ORDER)
    except IOError:
        from create_random_arrays import create_arrays
        create_arrays([nb-1] * 3)
else:
    randX, randY, randZ = None, None, None

##initjet.f
width = 0.01
ampl3 = 0.3
ampl = 0.05
ampl2=0.05
def computeVel(res, x, y, z, t):
    yy = abs(y - 0.5)
    aux = (0.1 - 2. * yy) / (4. * width)
    strg1 = exp(-abs(aux ** 2)) * randX
    strg2 = exp(-abs(aux ** 2)) * randY
    strg3 = exp(-abs(aux ** 2)) * randZ
    res[0][...] = 0.5 * (1. + tanh(aux))
    res[0][...] *= (1. + ampl3 * sin(8. * pi * x))
    res[0][...] *= (1. + ampl * strg1)
    res[1][...] = ampl * strg2
    res[2][...] = ampl * strg3
    return res


def initScal(res, x, y, z, t):
    yy = abs(y - 0.5)
    aux = (0.1 - 2. * yy) / (4. * width)
    res[0][...] = 0.5 * (1. + tanh(aux))
    return res

# number of ghosts in usual cartesian topo
ADVECTION_METHOD = {Scales: 'p_64', MultiScale: 'L4_4'}
ADVECTION_METHOD_S = {TimeIntegrator: RK2,
                      Interpolation: Linear,
                      Remesh: L6_4,
                      Support: 'gpu_1k',
                      Splitting: 'o2',
                      MultiScale: L4_4}
VISCOSITY = 1e-4

## Domain
box = pp.Box(dim, length=[1., ]*3)

## Fields
velo = Field(domain=box, formula=computeVel,
             name='Velocity', isVector=True)
vorti = Field(domain=box,
              name='Vorticity', isVector=True)
scal = Field(domain=box, formula=initScal,
              name='PassiveScalar', isVector=False)


topodims = np.ones((dim, ))
topodims[-1] = comm_s.Get_size()
topo_shared = box.getOrCreateTopology( 1, nbElem, topodims, comm=comm_s)

## Navier Stokes Operators
advec = Advection(velo, vorti,
                  resolutions={velo: nbElem,
                               vorti: nbElem},
                  method=ADVECTION_METHOD,
                  task_id=NS_TASK,
                  comm=comm_s,
                  )
advecScal = Advection(velo, scal,
                      resolutions={velo: nbElem,
                                   scal: nbElem_s},
                      method=ADVECTION_METHOD_S,
                      comm=comm_s,
                      task_id=PS_TASK,
                      device_id=main_rank//5,
                      device_type='gpu',
                      diffusion=True, diffusion_coeff=1e-5,
                      )

stretch = Stretching(velo, vorti,
                     resolutions={velo: nbElem,
                                  vorti: nbElem},
                     comm=comm_s,
                     task_id=NS_TASK,
                     )

diffusion = Diffusion(vorti,
                      resolution=nbElem,
                      viscosity=VISCOSITY,
                      comm=comm_s,
                      task_id=NS_TASK,
                      )

poisson = Poisson(velo, vorti,
                  resolutions={velo: nbElem,
                               vorti: nbElem},
                  comm=comm_s,
                  task_id=NS_TASK,
                  )
poisson.activateProjection(
    Reprojection_criterion(vorti, reproj_cst=0, reprojRate=1,
                           topo=None))

c = Curl(velo, vorti,
         resolutions={velo: nbElem,
                      vorti: nbElem},
         method={SpaceDiscretisation: 'fftw', GhostUpdate: True},
         task_id=NS_TASK,
         comm=comm_s,
         )

data = {'dt': 0.001 }
dt = VariableParameter(data)
dt_adapt = AdaptTimeStep(velo, vorti,
                         dt_adapt=dt,
                         resolutions={velo: nbElem,
                                      vorti: nbElem},
                         io_params={"filename": "dt_hybrid_64-128.dat"},
                         task_id=NS_TASK,
                         comm=comm_s)
to_ghostsdt = Redistribute([velo], poisson, dt_adapt, name_suffix='toG_V_dt',
                           task_id=NS_TASK)

toGhost_vorti = Redistribute([vorti], advec, stretch, name_suffix='toG_W',
                             task_id=NS_TASK)
toGhost_velo = Redistribute([velo], poisson, stretch, name_suffix='toG_V',
                            task_id=NS_TASK)
fromGhost_vorti = Redistribute([vorti], stretch, diffusion, name_suffix='fromG_W',
                               task_id=NS_TASK)

distrForAdvecX = Redistribute([velo], None, advecScal.advecDir[0],
                              component=0, name_suffix='toADVEC_X',
                              task_id=PS_TASK)
distrForAdvecY = Redistribute([velo], None, advecScal.advecDir[1],
                              component=1, name_suffix='toADVEC_Y',
                              task_id=PS_TASK)
distrForAdvecZ = Redistribute([velo], None, advecScal.advecDir[2],
                              component=2, name_suffix='toADVEC_Z',
                              task_id=PS_TASK)

redX = RedistributeIntercomm(velo, poisson, distrForAdvecX,
                             proc_tasks, main_comm,
                             topo=topo_shared, name_suffix='INTER_X',
                             component=0)
redY = RedistributeIntercomm(velo, poisson, distrForAdvecY,
                             proc_tasks, main_comm,
                             topo=topo_shared, name_suffix='INTER_Y',
                             component=1)
redZ = RedistributeIntercomm(velo, poisson, distrForAdvecZ,
                             proc_tasks, main_comm,
                             topo=topo_shared, name_suffix='INTER_Z',
                             component=2)
distrForAdvecX.opFrom = redX
distrForAdvecY.opFrom = redY
distrForAdvecZ.opFrom = redZ

## Simulation
simu = Simulation(tinit=0.0, tend=4.5, timeStep=dt.data['dt'], iterMax=10000,
                  bcast_comm=main_comm, bcast_rk=1
                  )

#  Define the problem to solve
pb = ProblemTasks(operators=[
        redX, redY, redZ,
        distrForAdvecX,
        distrForAdvecY,
        distrForAdvecZ,
        toGhost_velo,
        advecScal,
        advec,
        toGhost_vorti,
        stretch,
        fromGhost_vorti,
        diffusion,
        poisson,
        to_ghostsdt, dt_adapt,
        ], simulation=simu, tasks_list=proc_tasks,
                  dumpFreq=-1, main_comm=main_comm)

pb.pre_setUp()

if proc_tasks[main_rank] == NS_TASK:
    topofft = poisson.discreteFields[poisson.vorticity].topology
    topo_scal = None
else:
    topofft = None
    topo_scal = advecScal.advecDir[0].discreteFields[scal].topology

energy = Energy_enstrophy(velo, vorti, isNormalized=True,
                          topo=topofft,
                          viscosity=VISCOSITY,
                          io_params={"filename": "energy_hybrid_64-128.dat"},
                          task_id=NS_TASK)
p = Printer(variables=[scal,],
            topo=topo_scal,
            frequency=200,
            prefix='scal_hybrid_64-128',
            formattype=HDF5,
            task_id=PS_TASK)
pb.addMonitors([energy, p])


# initialisation
vorti.setTopoInit(topofft)
velo.setTopoInit(topofft)
scal.setTopoInit(topo_scal)

pb.setUp()

if proc_tasks[main_rank] == NS_TASK:
    c.discretize()
    c.setUp()
    c.apply(simu)

if proc_tasks[main_rank] == PS_TASK:
    p.apply(simu)

main_comm.Barrier()
from parmepy.mpi import MPI
print "Start computation ..."
time = MPI.Wtime()
pb.solve()

if proc_tasks[main_rank] == PS_TASK:
    p.apply(simu)

print 'total time (rank):', MPI.Wtime() - time, '(', main_rank, ')'

pb.finalize()
print pb.timer
