#!/usr/bin/python

"""
Taylor Green 3D : see paper van Rees 2011.

All parameters are set and defined in python module dataTG.

"""

import parmepy as pp
from parmepy.f2py import fftw2py
import numpy as np
from parmepy.fields.continuous import Field
from parmepy.fields.variable_parameter import VariableParameter
from parmepy.mpi.topology import Cartesian
from parmepy.operator.advection import Advection
from parmepy.operator.stretching import Stretching
from parmepy.operator.poisson import Poisson
from parmepy.operator.diffusion import Diffusion
from parmepy.operator.adapt_timestep import AdaptTimeStep
from parmepy.operator.redistribute import Redistribute
from parmepy.problem.navier_stokes import NSProblem
from parmepy.operator.monitors.printer import Printer
from parmepy.operator.monitors.energy_enstrophy import Energy_enstrophy
from parmepy.operator.monitors.reprojection_criterion import Reprojection_criterion
from parmepy.problem.simulation import Simulation
from parmepy.constants import VTK, HDF5
from parmepy.methods_keys import Scales, TimeIntegrator, Interpolation,\
    Remesh, Support, Splitting, dtCrit, SpaceDiscretisation
from parmepy.numerics.integrators.runge_kutta2 import RK2 as RK2
from parmepy.numerics.integrators.runge_kutta3 import RK3 as RK3
from parmepy.numerics.integrators.runge_kutta4 import RK4 as RK4
from parmepy.numerics.finite_differences import FD_C_4, FD_C_2
from parmepy.numerics.interpolation import Linear
from parmepy.numerics.remeshing import L6_4 as rmsh

## ----------- A 3d problem -----------
print " ========= Start Navier-Stokes 3D (Taylor Green benchmark) ========="
## pi constant
pi = np.pi
cos = np.cos
sin = np.sin

## constants
VISCOSITY = 1. / 1600.

## Domain
dim = 3
box = pp.Box(dim, length=[2.0 * pi, 2.0 * pi, 2.0 * pi])

## Global resolution
nb = 33
nbElem = [nb] * dim

## Function to compute TG velocity
def computeVel(res, x, y, z, t):
    res[0][...] = sin(x) * cos(y) * cos(z)
    res[1][...] = - cos(x) * sin(y) * cos(z)
    res[2][...] = 0.
    return res


## Function to compute reference vorticity
def computeVort(res, x, y, z, t):
    res[0][...] = - cos(x) * sin(y) * sin(z)
    res[1][...] = - sin(x) * cos(y) * sin(z)
    res[2][...] = 2. * sin(x) * sin(y) * cos(z)
    return res

## Fields
velo = Field(domain=box, formula=computeVel,
             name='Velocity', isVector=True)
vorti = Field(domain=box, formula=computeVort,
              name='Vorticity', isVector=True)

## Parameter Variable (adaptative timestep)
data = {'dt': 0.01}
dt_adapt = VariableParameter(data)

## Usual Cartesian topology definition
# At the moment we use two (or three?) topologies :
# - "topo" for Stretching and all operators based on finite differences.
#    --> ghost layer = 2
# - topo from Advection operator for all the other operators.
#    --> no ghost layer
# - topo from fftw for Poisson and Diffusion.
# Todo : check compat between scales and fft operators topologies.
NBGHOSTS = 2
ghosts = [NBGHOSTS] * box.dimension
topo = Cartesian(box, box.dimension, nbElem, ghosts=ghosts)

## Tools Operators
# Adaptative timestep method : dt = min(values(dtCrit))
# where dtCrit is a list of criterions on which the computation 
# of the adaptative time step is based
# ex : dtCrit = ['gradU', 'cfl', 'stretch'], means :
# dt = min (dtAdv, dtCfl, dtStretch), where dtAdv is equal to LCFL / |gradU|
# For dtAdv, the possible choices are the following:
# 'vort' (infinite norm of vorticity) : dtAdv = LCFL / |vort|
# 'gradU' (infinite norm of velocity gradient), dtAdv = LCFL / |gradU|
# 'deform' (infinite norm of deformation tensor), dtAdv = LCFL / (0.5(gradU + gradU^T))
dtAdapt = AdaptTimeStep(velo, vorti,
                        resolutions={velo: nbElem,
                                     vorti: nbElem},
                        dt_adapt=dt_adapt,
                        method={TimeIntegrator: RK3, 
                                SpaceDiscretisation: FD_C_4, 
                                dtCrit: ['deform', 'cfl', 'stretch']},
                        topo=topo,
                        io_params={},
                        lcfl=0.125,
                        cfl=0.5)

## Navier Stokes Operators
advec = Advection(velo, vorti,
                  resolutions={velo: nbElem,
                               vorti: nbElem},
                  method={Scales: 'p_M6', Splitting: 'classic'} # Scales advection
#                  method={TimeIntegrator: RK2,
#                          Interpolation: Linear,
#                          Remesh: rmsh,
#                          Support: '',
#                          Splitting: 'o2_FullHalf'} # pure Python advection
                  )

stretch = Stretching(velo, vorti,
                     resolutions={velo: nbElem,
                                  vorti: nbElem},
                     topo=topo
                     )

diffusion = Diffusion(vorti,
                      resolution=nbElem,
                      viscosity=VISCOSITY
                     )

poisson = Poisson(velo, vorti,
                  resolutions={velo: nbElem,
                               vorti: nbElem}
                 )

#-----------------------------
# Topology without ghost points
poisson.discretize()
topofft = poisson.discreteFields[poisson.vorticity].topology
#-----------------------------

# Bridges between the different topologies in order to
# redistribute data.
# 1 -Advection to stretching
distrAdvStr_vorti = Redistribute([vorti], advec, stretch)
distrPoiStr_velo = Redistribute([velo], poisson, stretch)
# 2 - Stretching to Poisson/Diffusion
distrStrDiff = Redistribute([vorti], stretch, diffusion)
# 3 - Poisson to TimeStep
distrPoissTimeStep = Redistribute([velo, vorti], poisson, dtAdapt)

## Simulation
simu = Simulation(tinit=0.0,
                  tend=10.0, 
                  timeStep=dt_adapt['dt'],
                  iterMax=1000000)

#  Define the problem to solve
## With adaptative time step
pb = NSProblem(operators=[distrPoiStr_velo, 
                          advec, 
                          distrAdvStr_vorti,
                          stretch,
                          distrStrDiff,
                          diffusion,
                          poisson,
                          distrPoissTimeStep,
                          dtAdapt],
               simulation=simu, dumpFreq=-1)

## With fixed time step
#pb = NSProblem(operators=[distrPoiStr_velo, 
#                          advec,
#                          distrAdvStr_vorti,
#                          stretch,
#                          distrStrDiff,
#                          diffusion,
#                          poisson],
#               simulation=simu, dumpFreq=-1)

## Diagnostics related to the problem
printer = Printer(variables=[velo, vorti],
                  topo=topofft,
                  frequency=100,
                  formattype=HDF5,
                  prefix='TG_io',
                  xmfalways=True)

energy = Energy_enstrophy(velo, vorti,
                          viscosity=VISCOSITY,
                          isNormalized=True ,
                          topo=topofft,
                          io_params={})

reproj = Reprojection_criterion(vorti, 
                                reproj_cst=0.04, 
                                reprojRate=1,
                                topo=topofft, 
                                checkCriterion=False,
                                io_params={})

## Add the reprojection of vorticity field (i.e div(vort)=0) for poisson operator
poisson.activateProjection(reproj)

## Setting solver to Problem (only operators for computational tasks)
pb.pre_setUp()

## Initializations on required topos + add monitors to problem + setUp
vorti.setTopoInit(topofft)
velo.setTopoInit(topofft)
pb.addMonitors([energy, printer])
pb.setUp()

printer.apply(simu)
## Time loop
def run():
# =====Automatic problem launch=====
    pb.solve()
# =====Manually problem launch=====
#    print "\n\n Start solving ..."
#    simu.initialize()
#    while not simu.isOver:
#        distrPoiStr_velo.apply(simu)
#        subSimu = Simulation(tinit=simu.tk, tend=10., timeStep=simu.timeStep/2.0,
#                             iterMax=2)
#        subSimu.initialize()
#        print '=============='
#        simu.printState()
#        while not subSimu.isOver:
#            subSimu.printState()
#            advec.apply(subSimu)
#            poisson.apply(subSimu)
#            subSimu.advance()
#        distrAdvStr_vorti.apply(simu)
#        stretch.apply(simu)
#        distrStrDiff.apply(simu)
#        diffusion.apply(simu)
#        poisson.apply(simu)
#        distrPoissTimeStep.apply(simu)
#        dtAdapt.apply(simu)
#        energy.apply(simu)
##        reproj.apply(simu)
#        simu.advance()

## Solve problem
from parmepy.mpi import MPI
print "Start computation ..."
time = MPI.Wtime()
run()
pb.finalize()
print 'total time (rank):', MPI.Wtime() - time, '(', topo.rank, ')'

## Clean memory buffers
#fftw2py.clean_fftw_solver(box.dimension)
