#!/usr/bin/python

"""
Taylor Green 3D : see paper van Rees 2011.

All parameters are set and defined in python module dataTG.

"""

import parmepy as pp
from parmepy.f2py import fftw2py
import numpy as np
from parmepy.fields.continuous import Field
from parmepy.variables.variables import Variables
from parmepy.mpi.topology import Cartesian
from parmepy.operator.advection import Advection
from parmepy.operator.stretching import Stretching
from parmepy.operator.adapt_timestep import AdaptTimeStep
from parmepy.operator.poisson import Poisson
from parmepy.operator.diffusion import Diffusion
from parmepy.operator.redistribute import Redistribute
from parmepy.problem.navier_stokes import NSProblem
from parmepy.operator.monitors.energy_enstrophy import Energy_enstrophy
from parmepy.operator.monitors.reprojection_criterion import Reprojection_criterion
from parmepy.methods_keys import TimeIntegrator, Interpolation, Remesh,\
    Support, Splitting, dtCrit, SpaceDiscretisation
from parmepy.numerics.integrators.runge_kutta2 import RK2
from parmepy.numerics.integrators.runge_kutta3 import RK3
from parmepy.numerics.finite_differences import FD_C_4, FD_C_2
from parmepy.numerics.interpolation import Linear
from parmepy.numerics.remeshing import L4_2 as rmsh
from parmepy.methods_keys import Scales
from parmepy.problem.simulation import Simulation

# problem dimension
dim = 3
# resolution
nb = 65
# number of ghosts in usual cartesian topo
NBGHOSTS = 2
TIMESTEP_METHOD = {TimeIntegrator: RK3, SpaceDiscretisation: FD_C_4,
                   dtCrit: 'deform'}
# Lagrangian CFL for adaptative timestep
LCFL = 0.125
# CFL (if CFL is None, no CFL condition is taken into account
# in the computation of adaptative timesteps)
CFL = 0.5
# Advection method
#ADVECTION_METHOD = {Scales: 'p_M6'
ADVECTION_METHOD = {TimeIntegrator: RK2,
                    Interpolation: Linear,
                    Remesh: rmsh,
                    Support: 'gpu_1k',
                    Splitting: 'o2_FullHalf'}
#
VISCOSITY = 1. / 1600.
# reprojection criterion
REPROJ_CST = 0.04
# Vorticity projection (list of 3 elements):
# - the 1st element determines if the reprojection is needed
# - the 2nd element gives the reprojection frequency (nber of iterations)
# if the 1st element is True
# - the 3rd element indicates if the reprojection frequency given before
# can be reduced to satisfy the reprojection criterion
# (governed by the reprojection constant given above)
# ex1 : PROJ = [True, 10, False] means that the reprojection
# of vorticty will strictly be applied every 10 iterations,
# regardless of reprojection criterion
# ex2 : PROJ = [True, 25, True] means that the reprojection
# of vorticty will be applied by default every 25 iterations
# AND also when the reprojection criterion won't be satisfied
PROJ = [False, 1, False]

# reprojection method
REPROJ_METHOD = {SpaceDiscretisation: FD_C_4}

# post treatment ...
OUTPUT_FREQ = 1
FILENAME = './res/energy_GPU.dat'

## ----------- A 3d problem -----------
print " ========= Start Navier-Stokes 3D (Taylor Green benchmark) ========="
## pi constant
pi = np.pi
cos = np.cos
sin = np.sin
## Domain
box = pp.Box(dim, length=[2.0 * pi, 2.0 * pi, 2.0 * pi])

## Global resolution
nbElem = [nb] * dim


## Function to compute TG velocity
def computeVel(res, x, y, z, t):
    res[0][...] = sin(x) * cos(y) * cos(z)
    res[1][...] = - cos(x) * sin(y) * cos(z)
    res[2][...] = 0.
    return res


## Function to compute reference vorticity
def computeVort(res, x, y, z, t):
    res[0][...] = - cos(x) * sin(y) * sin(z)
    res[1][...] = - sin(x) * cos(y) * sin(z)
    res[2][...] = 2. * sin(x) * sin(y) * cos(z)
    return res

## Fields
velo = Field(domain=box, formula=computeVel,
             name='Velocity', isVector=True)
vorti = Field(domain=box, formula=computeVort,
              name='Vorticity', isVector=True)

## Variables
dt_adapt = Variables(domain=box, name='adaptative time step',
                     data=[0.01])

need_reproj = Variables(domain=box, name='vorticity projection frequency',
                        data=PROJ)

## Usual Cartesian topology definition
# At the moment we use two (or three?) topologies :
# - "topo" for Stretching and all operators based on finite differences.
#    --> ghost layer = 2
# - topo from Advection operator for all the other operators.
#    --> no ghost layer
# - topo from fftw for Poisson and Diffusion.
# Todo : check compat between scales and fft operators topologies.
ghosts = [NBGHOSTS] * box.dimension
topo = Cartesian(box, box.dimension, nbElem,
                 ghosts=ghosts)

## Tools Operators
dtAdapt = AdaptTimeStep(velo, vorti,
                        resolutions={velo: nbElem,
                                     vorti: nbElem},
                        dt_adapt=dt_adapt,
                        method=TIMESTEP_METHOD,
                        topo=topo,
                        lcfl=LCFL,
                        cfl=CFL)

## Operators
advec = Advection(velo, vorti,
                  resolutions={velo: nbElem,
                               vorti: nbElem},
                  method=ADVECTION_METHOD,
                  )
advec.discretize()

stretch = Stretching(velo, vorti,
                     resolutions={velo: nbElem,
                                  vorti: nbElem},
                     topo=topo
                     )

diffusion = Diffusion(vorti,
                      resolution=nbElem,
                      viscosity=VISCOSITY
                      )

poisson = Poisson(velo, vorti,
                  resolutions={velo: nbElem,
                               vorti: nbElem},
                  projection=need_reproj)

# Bridges between the different topologies in order to
# redistribute data.
# 1 -Advection to stretching
# 1.1 -Advection output is vorticity on gpu
vorti_advec_stretch = Redistribute([vorti], advec, stretch)
# 1.2 -Streching input is already uptodate on cpu from Poisson output
velo_poisson_stretch = Redistribute([velo], poisson, stretch)
# 2 -Stretching to Poisson/Diffusion
vorti_stretch_diffusion = Redistribute([vorti], stretch, diffusion)
# 3 -Advection input to GPU from Poisson/Diffusion
vorti_diff_advec = Redistribute([vorti], diffusion, advec)
velo_diff_advec_X = Redistribute([velo], poisson, advec.advecDir[0],
                                 component=0)
velo_diff_advec_Y = Redistribute([velo], poisson, advec.advecDir[1],
                                 component=1)
velo_diff_advec_Z = Redistribute([velo], poisson, advec.advecDir[2],
                                 component=2)
# 4 - Poisson to TimeStep
distrPoissTimeStep = Redistribute([velo, vorti], poisson, dtAdapt)

## Simulation
simu = Simulation(tinit=0.0,
                  tend=10., timeStep=dt_adapt,
                  iterMax=1000000)
##  Define the problem to solve
pb = NSProblem(operators=[velo_poisson_stretch,
                          advec,
                          vorti_advec_stretch,
                          stretch,
                          vorti_stretch_diffusion,
                          diffusion,
                          vorti_diff_advec,
                          poisson,
                          distrPoissTimeStep,
                          velo_diff_advec_X,
                          velo_diff_advec_Y,
                          velo_diff_advec_Z,
                          dtAdapt
                          ],
               simulation=simu, dumpFreq=-1)

## Setting solver to Problem (only operators for computational tasks)
pb.pre_setUp()
## Diagnostics related to the problem
topofft = poisson.discreteFields[poisson.vorticity].topology
energy = Energy_enstrophy(velo, vorti,
                          topo=topofft,
                          viscosity=VISCOSITY,
                          frequency=OUTPUT_FREQ,
                          prefix=FILENAME)

reproj = Reprojection_criterion(vorti,
                                need_reproj=need_reproj,
                                reproj_cst=REPROJ_CST,
                                topo=topo,
                                frequency=OUTPUT_FREQ,
                                method=REPROJ_METHOD,
                                prefix=None)

## printer = Printer(variables=[velo, vorti],
##                   topo=topo,
##                   frequency=500,
##                   prefix='./res/TG_',
##                   ext='.vtk')
vorti.setTopoInit(topofft)
velo.setTopoInit(topofft)
pb.addMonitors([energy])
pb.setUp()
advec.advecDir[0].discreteOperator.velocity.wait()
velo_diff_advec_Z.apply()

# from parmepy.tools.problem2dot import toDot
# toDot(pb,filename='graph.dot')

def run():
    pb.solve()

## Solve problem
from parmepy.mpi import MPI
print "Start computation ..."
time = MPI.Wtime()
run()
print 'total time (rank):', MPI.Wtime() - time, '(', topo.rank, ')'

pb.finalize()

# advec.finalize()
# print pb.timer
# ## Clean memory buffers
fftw2py.clean_fftw_solver(box.dimension)
