"""Testing mpi multiple topologies transfers"""

import parmepy as pp
from parmepy.constants import np, PARMES_MPI_REAL, ORDERMPI, PARMES_REAL
import parmepy.tools.numpywrappers as npw
from parmepy.fields.continuous import Field
from parmepy.mpi.main_var import main_rank, main_size, main_comm, MPI
from parmepy.mpi.topology import Cartesian
from parmepy.operator.analytic import Analytic
from parmepy.operator.redistribute import Redistribute
from parmepy.operator.redistribute_intercomm import RedistributeIntercomm

if PARMES_REAL is np.float32:
    atol = 1.e-4
else:
    atol = 1e-10

def func(res, x, y, z, t=0.):
    res[0][...] = x
    res[1][...] = y ** 2
    res[2][...] = x + y
    return res


def test_topologies(domain, topo_from, topo_to, init_formula):
    """
    Testing the mpi transfers between the two topologies.
    It initialises a field on topo_from and one Analytic operator on each
    topology. A Redistribute operator is created to transfer the data.
    """
    field = Field(domain=domain, name="Test_Vec",
              isVector=True, formula=init_formula)

    if not topo_from.isNew:
        topo_from = topo_to
    if not topo_to.isNew:
        topo_to = topo_from

    op_from = Analytic([field], {field: nb_elem}, topo=topo_from)
    op_to = Analytic([field], {field: nb_elem}, topo=topo_to)
    op_from.discretize()
    op_to.discretize()
    op_from.setUp()
    op_to.setUp()

    ope = Redistribute([field], op_from, op_to)
    ope.discretize()
    ope.setUp()

    field.initialize(topo=topo_from)
    ope.apply()
    ope.wait()

    assert pp.constants.np.allclose(field.norm(topo_from),
                                    field.norm(topo_to),
                                    atol=atol), \
        "[{0}] Data differs: norm, topo_from - topo_to = {1}".format(
            main_rank, field.norm(topo_from) - field.norm(topo_to))


nb_elem = [17, ] * 3
b = pp.Box()

## CASE 1:
## Testing for two topologies with the same MPI size but different dimensions
if main_rank == 0:
    print "Testing CASE 1 ...",
    b.reset()
topo_q = Cartesian(b, dim=2, globalMeshResolution=[17, 17, 17],
                   cutdir=[False, True, True]
                   )
topo_p = Cartesian(b, dim=1, globalMeshResolution=[17, 17, 17],
                   cutdir=[True, False, False]
                   )
if not topo_q.isNew:
    if main_rank == 0:
        print "Reuse topology"
    topo_q = topo_p
if not topo_p.isNew:
    if main_rank == 0:
        print "Reuse topology"
    topo_p = topo_q
test_topologies(b, topo_p, topo_q, func)
test_topologies(b, topo_q, topo_p, func)
if main_rank == 0:
    print "Ok"
main_comm.Barrier()


## CASE 2:
## Testing for two topologies with based on a MPI sub-communicator of same
## size. Only half of total processes are working.
if main_size % 2 == 0:
    if main_rank == 0:
        print "Testing CASE 2 ...",
    b.reset()
    proc_tasks = [0,] * (main_size/2) + [1,] * (main_size/2)
    comm_s = main_comm.Split(color=proc_tasks[main_rank], key=main_rank)
    topo_q = Cartesian(b, dim=2, globalMeshResolution=[17, 17, 17],
                       cutdir=[False, True, True],
                       comm=comm_s
                       )
    topo_p = Cartesian(b, dim=1, globalMeshResolution=[17, 17, 17],
                       cutdir=[True, False, False],
                       comm=comm_s
                       )
    if not topo_q.isNew:
        if main_rank == 0:
            print "Reuse topology"
        topo_q = topo_p
    if not topo_p.isNew:
        if main_rank == 0:
            print "Reuse topology"
        topo_p = topo_q
    if proc_tasks[main_rank] == 0:
        test_topologies(b, topo_p, topo_q, func)
        test_topologies(b, topo_q, topo_p, func)
    if main_rank == 0:
        print "Ok"
else:
    if main_rank == 0:
        print "CASE 2 is not tested (require an even process number)"
main_comm.Barrier()

## CASE 3 (small case):
## Testing for two topologies with based on a MPI sub-communicator of different
## size. Only half of total processes are working. Initialisation is performed
## by each processus in using his topology.
if main_size > 2:
    if main_rank == 0:
        print "Testing CASE 3 ...",
    b.reset()
    proc_tasks = [0,] * main_size
    proc_tasks[0] = 1
    comm_s = main_comm.Split(color=proc_tasks[main_rank], key=main_rank)
    topo = Cartesian(b, dim=2, globalMeshResolution=[17, 17, 17],
                       cutdir=[False, True, True],
                       comm=comm_s
                       )
    field = Field(domain=b, name="Test_Vec",
                  isVector=True, formula=func)
    field.discretize(topo)
    field.initialize(topo=topo)
    norms = npw.zeros((main_size, 3))

    norms[main_rank,:] = field.norm(topo)
    for i in xrange(main_size):
        norms[i,:] = main_comm.bcast(norms[i,:], root=i)
    for i in xrange(main_size):
        assert pp.constants.np.allclose(norms[i,:], norms[main_rank,:],
                                        atol=atol)
    if main_rank == 0:
        print " Ok"
else:
    if main_rank == 0:
        print "CASE 3 is not tested (require process number > 2)"


## CASE 4:
## Extending CASE 3: the field is initialized by process of only one
## task id.
## 'From' topology is 1D and 'To' is 2D
if main_size > 2:
    if main_rank == 0:
        print "Testing CASE 4 ...",
    b.reset()
    proc_tasks = [0,] * main_size
    proc_tasks[0:2] = [1, 1]
    comm_s = main_comm.Split(color=proc_tasks[main_rank], key=main_rank)
    if proc_tasks[main_rank] == 1:
        topo = Cartesian(b, dim=1, globalMeshResolution=[17, 17, 17],
                         cutdir=[True, False, False],
                         comm=comm_s
                         )
    elif proc_tasks[main_rank] == 0:
        topo = Cartesian(b, dim=2, globalMeshResolution=[17, 17, 17],
                         cutdir=[False, True, True],
                         comm=comm_s
                         )
    field = Field(domain=b, name="Test_Vec",
                  isVector=True, formula=func)
    field_r = Field(domain=b, name="Test_Vec_R",
                    isVector=True, formula=func)

    field.discretize(topo)
    field_r.discretize(topo)
    if proc_tasks[main_rank] == 1:
        field.initialize(topo=topo)
    if proc_tasks[main_rank] == 0:
        field_r.initialize(topo=topo)

    op_from = Analytic([field], {field: [17, 17, 17]}, topo=topo, task_id=1)
    op_to = Analytic([field], {field: [17, 17, 17]}, topo=topo, task_id=0)

    red = RedistributeIntercomm(field, topo, op_from, op_to,
                                proc_tasks, main_comm,
                                component=None)
    red_r = RedistributeIntercomm(field_r, topo, op_to, op_from,
                                  proc_tasks, main_comm,
                                  component=None)
    red.discretize()
    red_r.discretize()
    red.setUp()
    red_r.setUp()

    red.apply()
    red_r.apply()

    norms = npw.zeros((main_size, 3))
    norms[main_rank, :] = field.norm(topo)
    for i in xrange(main_size):
        norms[i, :] = main_comm.bcast(norms[i, :], root=i)
    for i in xrange(main_size):
        assert pp.constants.np.allclose(norms[i, :], norms[main_rank, :],
                                        atol=atol)
    norms[main_rank, :] = field_r.norm(topo)
    for i in xrange(main_size):
        norms[i, :] = main_comm.bcast(norms[i, :], root=i)
    for i in xrange(main_size):
        assert pp.constants.np.allclose(norms[i, :], norms[main_rank, :],
                                        atol=atol)

    if main_rank == 0:
        print " Ok"
else:
    if main_rank == 0:
        print "CASE 4 is not tested (require process number > 2)"

## CASE 5:
## Extending CASE 4: Using topology with ghosts
if main_size > 2:
    if main_rank == 0:
        print "Testing CASE 5 ...",
    b.reset()
    proc_tasks = [0,] * main_size
    proc_tasks[0:2] = [1, 1]
    comm_s = main_comm.Split(color=proc_tasks[main_rank], key=main_rank)
    if proc_tasks[main_rank] == 1:
        topo = Cartesian(b, dim=1, globalMeshResolution=[17, 17, 17],
                         cutdir=[True, False, False], ghosts=[2, 0, 1],
                         comm=comm_s
                         )
    elif proc_tasks[main_rank] == 0:
        topo = Cartesian(b, dim=2, globalMeshResolution=[17, 17, 17],
                         cutdir=[False, True, True], ghosts=[2, 0, 1],
                         comm=comm_s
                         )
    field = Field(domain=b, name="Test_Vec",
                  isVector=True, formula=func)
    field_r = Field(domain=b, name="Test_Vec_R",
                    isVector=True, formula=func)

    field.discretize(topo)
    field_r.discretize(topo)
    if proc_tasks[main_rank] == 1:
        field.initialize(topo=topo)
    if proc_tasks[main_rank] == 0:
        field_r.initialize(topo=topo)

    op_from = Analytic([field], {field: [17, 17, 17]}, topo=topo, task_id=1)
    op_to = Analytic([field], {field: [17, 17, 17]}, topo=topo, task_id=0)

    red = RedistributeIntercomm(field, topo, op_from, op_to,
                                proc_tasks, main_comm,
                                component=None)
    red_r = RedistributeIntercomm(field_r, topo, op_to, op_from,
                                  proc_tasks, main_comm,
                                  component=None)
    red.discretize()
    red_r.discretize()
    red.setUp()
    red_r.setUp()

    red.apply()
    red_r.apply()

    norms = npw.zeros((main_size, 3))
    norms[main_rank, :] = field.norm(topo)
    for i in xrange(main_size):
        norms[i, :] = main_comm.bcast(norms[i, :], root=i)
    for i in xrange(main_size):
        assert pp.constants.np.allclose(norms[i, :], norms[main_rank, :],
                                        atol=atol)
    norms[main_rank, :] = field_r.norm(topo)
    for i in xrange(main_size):
        norms[i, :] = main_comm.bcast(norms[i, :], root=i)
    for i in xrange(main_size):
        assert pp.constants.np.allclose(norms[i, :], norms[main_rank, :],
                                        atol=atol)

    # Assert that ghosts have not been exchanged
    if proc_tasks[main_rank] == 0:
        assert np.allclose(field.discreteFields[topo].data[0][0:2, :, :], 0.,
                           atol=atol)
        assert np.allclose(field.discreteFields[topo].data[0][-2:, :, :], 0.,
                           atol=atol)
        assert np.allclose(field.discreteFields[topo].data[0][:, :, 0:1], 0.,
                           atol=atol)
        assert np.allclose(field.discreteFields[topo].data[0][:, :, -1:], 0.,
                           atol=atol)
    if proc_tasks[main_rank] == 1:
        assert np.allclose(field_r.discreteFields[topo].data[0][0:2, :, :], 0.,
                           atol=atol)
        assert np.allclose(field_r.discreteFields[topo].data[0][-2:, :, :], 0.,
                           atol=atol)
        assert np.allclose(field_r.discreteFields[topo].data[0][:, :, 0:1], 0.,
                           atol=atol)
        assert np.allclose(field_r.discreteFields[topo].data[0][:, :, -1:], 0.,
                           atol=atol)

    if main_rank == 0:
        print " Ok"

else:
    if main_rank == 0:
        print "CASE 5 is not tested (require process number > 2)"
