#!/bin/env python
import time
import parmepy as pp
from parmepy.methods_keys import Scales, TimeIntegrator, Interpolation,\
    Remesh, Support, Splitting, dtCrit, SpaceDiscretisation, GhostUpdate
from parmepy.numerics.interpolation import Linear
from parmepy.numerics.integrators.runge_kutta2 import RK2 as RK2
from parmepy.numerics.remeshing import L4_2 as rmsh
from parmepy.fields.continuous import Field
from parmepy.mpi.topology import Cartesian
from parmepy.operator.advection import Advection
from parmepy.operator.analytic import Analytic
from parmepy.problem.simulation import Simulation
from parmepy.problem.transport import TransportProblem
from parmepy.operator.monitors.printer import Printer
from parmepy.constants import VTK
import numpy as np
import math
import sys

## constants
pi = np.pi
cos = np.cos
sin = np.sin
exp = np.exp
abs = np.abs
tanh = np.tanh

def vitesse(res, x, y, z, t):
    res[0] = 2.0 * sin(x * pi) ** 2 * sin(y * pi * 2) * sin(z * pi * 2) * cos(pi * t / 2.)
    res[1] = - sin(x * pi * 2) * sin(y * pi) ** 2 * sin(z * pi * 2) * cos(pi * t / 2.)
    res[2] = - sin(x * pi * 2) * sin(y * pi * 2) * sin(z * pi) ** 2 * cos(pi * t / 2.)
    return res

def scalaire(res, x, y, z, t):
    rr = np.sqrt((x - 0.35) ** 2 + (y - 0.35) ** 2 + (z - 0.35) ** 2)
    ind = np.where(rr < 0.15)
    res[0][...] = 0.
    res[0][ind] = 1.0
    return res

def run(nb=65):
    dim = 3
    boxLength = [1., 1., 1.]
    boxMin = [0., 0., 0.]
    nbElem = [nb, nb, nb]

    timeStep = 0.04
    #period = 10.
    finalTime = 2.0 #55 * timeStep
    outputFilePrefix = './res/levelSet_3D_'
    outputModulo = 10

    t0 = time.time()

    ## Domain
    box = pp.Box(dim, length=boxLength, origin=boxMin)

    ## Fields
    velo = Field(domain=box, name='Velocity', isVector=True)
    scal = Field(domain=box, formula=scalaire, name='Scalar')

    ## Usual Cartesian topology definition
    ghosts = np.ones((box.dimension)) * 0
    topo = Cartesian(box, box.dimension, nbElem,
                     ghosts=ghosts)

    ## Operators
    advec = Advection(velo, scal,
                      resolutions={velo: nbElem,
                                   scal: nbElem},
#                      method={TimeIntegrator: RK2,
#                              Interpolation: Linear,
#                              Remesh: rmsh,
#                              Support: '',
#                              Splitting: 'o1'}
                      method={Scales: 'p_M4', Splitting: 'strang'},
                      topo=topo)

    velocity = Analytic(velo, resolutions={velo: nbElem}, 
                        formula=vitesse, topo=topo)

    ## Definition of simulation parameters
    simu = Simulation(tinit=0.0, tend=2.0, timeStep=0.01, iterMax=1000000)

    ##Problem
    pb = TransportProblem(operators=[velocity, advec],
                          simulation=simu, dumpFreq=-1)

    ## Setting solver to Problem (only operators for computational tasks)
    pb.pre_setUp()

    printer = Printer(variables=[velo, scal],
                      topo=topo,
                      frequency=outputModulo,
                      prefix=outputFilePrefix,
                      formattype=VTK)

    ## Setting solver to Problem
    velo.setTopoInit(topo)
    scal.setTopoInit(topo)
    pb.addMonitors([printer])
    pb.setUp()

    t1 = time.time()

    ## Solve problem
    timings = pb.solve()

    tf = time.time()

    print "\n"
    print "Total time : ", tf - t0, "sec (CPU)"
    print "Init time : ", t1 - t0, "sec (CPU)"
    print "Solving time : ", tf - t1, "sec (CPU)"
    print ""
    return pb.timings_info[0] + "\"Solving time\" \"Init time\" \"Total time\"", pb.timings_info[1] + str(tf - t1) + " " + str(t1 - t0) + " " + str(tf - t0)

if __name__ == "__main__":
    run(65)
