"""
Task parallelism example.

One process is dedicated to file outputs and other are computing the data.
"""
import parmepy as pp
from parmepy.constants import np, VTK, HDF5
from parmepy.mpi.main_var import main_size, main_rank, main_comm
from parmepy.mpi.topology import Cartesian
from parmepy.fields.continuous import Field
from parmepy.operator.advection import Advection
from parmepy.operator.analytic import Analytic
from parmepy.operator.redistribute_intercomm import RedistributeIntercomm
from parmepy.operator.redistribute import Redistribute
from parmepy.operator.monitors.printer import Printer
from parmepy.problem.problem_tasks import ProblemTasks
from parmepy.problem.simulation import Simulation
from parmepy.tools.problem2dot import toDot
from parmepy.methods_keys import TimeIntegrator, Interpolation, Remesh,\
    Support, Splitting, MultiScale, Scales, MultiScale
from parmepy.numerics.integrators.runge_kutta2 import RK2 as RK
#from parmepy.numerics.integrators.runge_kutta4 import RK4 as RK
from parmepy.numerics.interpolation import Linear
from parmepy.numerics.remeshing import L2_1 as rmsh, L4_2, L2_1, L4_4, L8_4
pi = np.pi
cos = np.cos
sin = np.sin
sqrt = np.sqrt

def scalaire(res, x, y, z, t=0.):
    r = sqrt((x - 0.35) ** 2 + (y - 0.35) ** 2 + (z - 0.35) ** 2)
    res[0][...] = 0.
    res[0][r < 0.15] = 1.
    return res


def vitesse(res, x, y, z, t=0.):
    res[0][...] = 2. * sin(pi*x)**2*sin(2.*pi*y)*sin(2.*pi*z)*cos(t*pi/3.)
    res[1][...] = -sin(2.*pi*x)*sin(pi*y)**2*sin(2.*pi*z)*cos(t*pi/3.)
    res[2][...] = -sin(2.*pi*x)*sin(2.*pi*y)*sin(pi*z)**2*cos(t*pi/3.)
    return res

dim = 3
boxLength = [1., 1., 1.]
boxMin = [0., 0., 0.]
nbElem = [129] * 3

assert main_size > 1
proc_tasks = [0,] * main_size
proc_tasks[0] = 1
comm_s = main_comm.Split(color=proc_tasks[main_rank], key=main_rank)


box = pp.Box(dim, length=boxLength, origin=boxMin)

if proc_tasks[main_rank] == 1:
    topo = box.getOrCreateTopology(1, gridResolution=nbElem,
                     comm=comm_s
                     )
elif proc_tasks[main_rank] == 0:
    topo = box.getOrCreateTopology(1, gridResolution=nbElem,
                     comm=comm_s
                     )
## Fields
scal = Field(domain=box, name='Scalar', formula=scalaire
             )
velo = Field(domain=box, name='Velocity', isVector=True, formula=vitesse
             )

op_velo = Analytic([velo], {velo: nbElem}, topo=topo, task_id=1)
advec = Advection(velo, scal,
                  resolutions={velo: nbElem,
                               scal: nbElem},
                  method={TimeIntegrator: RK,
                          Interpolation: Linear,
                          Remesh: L2_1,
                          Support: 'gpu_1k',
                          Splitting: 'o2',
                          MultiScale: L2_1},
                  topo=topo,
                  task_id=0,
                  )
p = Printer(variables=[scal],
            topo=topo,
            frequency=5,
            prefix='./levelset3D_hybrid/',
            formattype=HDF5,
            task_id=0)

distrForAdvecZ = Redistribute([velo], None, advec.advecDir[2], component=2,
                              task_id=0)

redX = RedistributeIntercomm(velo, topo, op_velo, advec.advecDir[0],
                             proc_tasks, main_comm,
                            component=0)
redY = RedistributeIntercomm(velo, topo, op_velo, advec.advecDir[1],
                             proc_tasks, main_comm,
                            component=1)
redZ = RedistributeIntercomm(velo, topo, op_velo, distrForAdvecZ,
                             proc_tasks, main_comm,
                            component=2)
distrForAdvecZ.opFrom = redZ

simu = Simulation(tinit=0.0, tend=3., timeStep=0.025, iterMax=200)


pb = ProblemTasks([op_velo, redX, redY, redZ, distrForAdvecZ, advec, p],
                  simu, proc_tasks, dumpFreq=1000)

from parmepy.tools.problem2dot import toDot
toDot(pb)

pb.setUp()
pb.solve()
pb.finalize()

