#!/usr/bin/python

"""
Test memory and performances for finite differences operations.

"""

import parmepy as pp
import parmepy.tools.numpywrappers as npw
import math as m
from parmepy.fields.continuous import Field
from parmepy.mpi.topology import Cartesian
from parmepy.operator.stretching import Stretching


## ----------- A 3d problem -----------
print " ========= Start Finite Differences tests ========="
dim = 3
pi = m.pi
sin = m.sin
cos = m.cos
## Domain
box = pp.Box(dim, length=[2.0 * pi, 2.0 * pi, 2.0 * pi])

## Global resolution
nb = 65
nbElem = [nb] * dim


## Function to compute velocity
def computeVel(x, y, z):
    vx = sin(x) * cos(y) * cos(z)
    vy = - cos(x) * sin(y) * cos(z)
    vz = 0.
    return vx, vy, vz


## Function to compute vorticity
def computeVort(x, y, z):
    wx = - cos(x) * sin(y) * sin(z)
    wy = - sin(x) * cos(y) * sin(z)
    wz = 2. * sin(x) * sin(y) * cos(z)
    return wx, wy, wz

## Fields
velo = Field(domain=box, formula=computeVel,
             name='Velocity', isVector=True)
vorti = Field(domain=box, formula=computeVort,
              name='VorticityC', isVector=True)

## A topology with ghosts
NBGHOSTS = 2
ghosts = npw.ones((box.dimension)) * NBGHOSTS
topo = Cartesian(box, box.dimension, nbElem,
                 ghosts=ghosts)

## local coordinates
coords = topo.mesh.coords
## Fields discretization
velo.discretize(topo)
vorti.discretize(topo)
velo.initialize()
vorti.initialize()

## alias to discrete fields components
w = vorti.discreteFields.values()[0].data
v = velo.discreteFields.values()[0].data

from parmepy.mpi import MPI
from parmepy.numerics.finite_differences import FD_C_4

## FD scheme creation and init
fd_scheme = FD_C_4((topo.mesh.space_step))
fd_scheme.computeIndices(topo.mesh.iCompute)

import numpy as np
#### Timings for finite differences schemes

iter = 10
# Allocation of work vector to store results
result = npw.zeros_like(w[0])

## Add raw_input to check memory with htop
print 'press any key to start tests'
raw_input()

########### Test 1 : finite diff. direct call ###########

#### First method #####
print 'start ...'
time = MPI.Wtime()
for i in xrange(iter):
    fd_scheme.compute(w[0], 0, result)
print 'FD compute method time ', MPI.Wtime() - time
raw_input()
print 'start ... '
#### Second method #####
time = MPI.Wtime()
for i in xrange(iter):
    fd_scheme.compute_and_add(w[0], 0, result)
print 'FD compute_and_add method time ', MPI.Wtime() - time

del result
iter = 1

#
# Res : first method is faster and cost less in memory
#

########### Test 2 : 'FD, with accumulation of the result ###########

result = [npw.zeros_like(w[0]) for i in xrange(2)]

## fd results are accumulated into result[0]

raw_input()
print 'start'
#### First method ####
## Needs at least two workspaces and one is used for the final result.
time = MPI.Wtime()
for i in xrange(iter):
    #for cdir in xrange(3):
    fd_scheme.compute(w[0], 0, result[0])
    fd_scheme.compute(w[1], 0, result[1])
    np.add(result[0], result[1], result[0])
    fd_scheme.compute(w[2], 0, result[1])
    np.add(result[0], result[1], result[0])
print 'Accumulate/FD Compute meth time ', MPI.Wtime() - time

raw_input()
print 'start ...'
result2 = npw.zeros_like(w[0])
#### Second method ####
## Needs only one workspace, used for the final result.
time = MPI.Wtime()
for i in xrange(iter):
    #for cdir in xrange(3):
    fd_scheme.compute(w[0], 0, result2)
    fd_scheme.compute_and_add(w[1], 0, result2)
    fd_scheme.compute_and_add(w[2], 0, result2)
print 'Accumulate/FD Compute_and_add meth time ', MPI.Wtime() - time

print 'Computation ok? ', np.allclose(result[0], result2)
del result
del result2

#
# Res : first method is faster and cost less in memory
#

########### Test 3 : 'real' case
#            corresponding more or less to divV computation ###########
# Needs an intermediate workspace to compute w.v
raw_input()
print 'start ...'
work = [npw.zeros_like(w[0]) for i in xrange(3)]
raw_input()

## fd results are accumulated into result[0]

#### First method ####
## Needs at least three workspaces and one is used for the final result.
time = MPI.Wtime()
for i in xrange(iter):
    #for cdir in xrange(3):
    work[2][...] = v[0] * w[0]
    fd_scheme.compute(work[2], 0, work[0])
    work[2][...] = v[0] * w[1]
    fd_scheme.compute(work[2], 0, work[1])
    np.add(work[0], work[1], work[0])
    work[2][...] = v[0] * w[2]
    fd_scheme.compute(work[2], 0, work[1])
    np.add(work[0], work[1], work[0])
print 'Div/FD Compute meth time ', MPI.Wtime() - time
time = MPI.Wtime()
for i in xrange(iter):
    #for cdir in xrange(3):
    work[2][...] = v[0] * w[0]
    fd_scheme.compute(work[2], 0, work[0])
    work[2][...] = v[0] * w[1]
    fd_scheme.compute(work[2], 0, work[1])
    work[0][...] += work[1]
    work[2][...] = v[0] * w[2]
    fd_scheme.compute(work[2], 0, work[1])
    work[0][...] += work[1]
print 'Div/FD Compute meth time (bis) ', MPI.Wtime() - time

#### Second method ####
## Needs only two workspace, used for the final result.
raw_input()
print 'start ...'
work2 = [npw.zeros_like(w[0]) for i in xrange(2)]
raw_input()

print 'pre id ...', id(work2[1])
time = MPI.Wtime()
for i in xrange(iter):
    #for cdir in xrange(3):
    work2[1][...] = v[0] * w[0]
    print 'post id ...', id(work2[1])
    fd_scheme.compute(work2[1], 0, work2[0])
    work2[1][...] = v[0] * w[1]
    fd_scheme.compute_and_add(work2[1], 0, work2[0])
    work2[1][...] = v[0] * w[2]
    fd_scheme.compute_and_add(work2[1], 0, work2[0])
print 'Div/FD Compute_and_add meth time ', MPI.Wtime() - time

print 'Computation ok? ', np.allclose(work[0], work2[0])

del work
del work2
#
# Res : first method is faster and cost less in memory
# (but it's almost the same)
#

########### Test 4 : 'real' case
#            corresponding more or less to divT computation ###########
# Needs an intermediate workspace to compute w.v

raw_input()
print 'start ...'
work = [npw.zeros_like(w[0]) for i in xrange(5)]
raw_input()

## fd results are accumulated into work[0:2]

#### First method ####
## Needs at least five workspaces and three are used for the final result.
# work[i] = divV(w vi)
time = MPI.Wtime()
for i in xrange(iter):
    # Compute first component of divT ...
    work[2] = v[0] * w[0]
    fd_scheme.compute(work[2], 0, work[0])
    work[2] = v[0] * w[1]
    fd_scheme.compute(work[2], 0, work[1])
    np.add(work[0], work[1], work[0])
    work[2] = v[0] * w[2]
    fd_scheme.compute(work[2], 0, work[1])
    np.add(work[0], work[1], work[0])
    # Second component ... work[0] can not be used anymore
    work[3] = v[0] * w[0]
    fd_scheme.compute(work[3], 0, work[1])
    work[3] = v[0] * w[1]
    fd_scheme.compute(work[3], 0, work[2])
    np.add(work[1], work[2], work[1])
    work[3] = v[0] * w[2]
    fd_scheme.compute(work[3], 0, work[2])
    np.add(work[1], work[2], work[1])
    # Last component ... work[0] and work[1] can not be used anymore
    work[4] = v[0] * w[0]
    fd_scheme.compute(work[4], 0, work[2])
    work[4] = v[0] * w[1]
    fd_scheme.compute(work[4], 0, work[3])
    np.add(work[2], work[3], work[2])
    work[4] = v[0] * w[2]
    fd_scheme.compute(work[4], 0, work[3])
    np.add(work[2], work[3], work[2])

print 'DivT/FD Compute meth time ', MPI.Wtime() - time
#### Second method ####
## Needs only 4 workspaces, used for the final result.
raw_input()
print 'start ...'
work2 = [npw.zeros_like(w[0]) for i in xrange(4)]
time = MPI.Wtime()
for i in xrange(iter):
    # Compute first component of divT ...
    work2[3] = v[0] * w[0]
    fd_scheme.compute(work2[3], 0, work2[0])
    work2[3] = v[0] * w[1]
    fd_scheme.compute_and_add(work2[3], 0, work2[0])
    work2[3] = v[0] * w[2]
    fd_scheme.compute_and_add(work2[3], 0, work2[0])
    # Second component ... work2[0] can not be used anymore
    work2[3] = v[0] * w[0]
    fd_scheme.compute(work2[3], 0, work2[1])
    work2[3] = v[0] * w[1]
    fd_scheme.compute_and_add(work2[3], 0, work2[1])
    work2[3] = v[0] * w[2]
    fd_scheme.compute_and_add(work2[3], 0, work2[1])
    # Last component ... work2[0] and work2[1] can not be used anymore
    work2[3] = v[0] * w[0]
    fd_scheme.compute(work2[3], 0, work2[2])
    work2[3] = v[0] * w[1]
    fd_scheme.compute_and_add(work2[3], 0, work2[2])
    work2[3] = v[0] * w[2]
    fd_scheme.compute_and_add(work2[3], 0, work2[2])

print 'DivT/FD Compute_and_add meth time ', MPI.Wtime() - time

ind = topo.mesh.iCompute
for i in xrange(3):
    print 'Computation ok? ', np.allclose(work[i][ind], work2[i][ind])
#
# Res : first method is faster and cost less in memory
# (but it's almost the same)
#


## Conclusion (if res > 200 **)-> it seems that first method, although it needs
# more initial workspaces, is always faster and cheaper
# concerning memory (because of hidden tmp alloc)
# BUT for smaller resolutions, second method may be more efficient ...
