"""
@file domain.py

Abstract interface for physical domains description.

"""
from abc import ABCMeta, abstractmethod
from parmepy.constants import debug
from parmepy.mpi.topology import Cartesian


class Domain(object):
    """Abstract base class for description of physical domain. """

    __metaclass__ = ABCMeta

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    @abstractmethod
    def __init__(self, dimension):
        """ Constructor
        @param dimension integer : domain dimension.
        """
        ## Domain dimension.
        self.dimension = dimension
        ## A list of all the topologies defined on this domain.
        ## Each topology is unique in the sense defined by
        ## the comparison operator in the class parmepy.mpi.topology.Cartesian.
        self.topologies = {}

    def getOrCreateTopology(self, topoDim, gridResolution,
                            topoResolution=None, ghosts=None,
                            precomputed=False, localres=None,
                            offset=None, fixedResolution=False,
                            comm=None):
        """
        This routines checks if a topology is present in the list
        of topologies of the domain.
        Either it gets the topology corresponding to the input arguments
        (topodim, gridResolution, ...) if it exists (in the sense of the
        comparison operator defined in parmepy.mpi.topology.Cartesian)
        or it creates a new topology and register it in the topology list.
        @param topoDim : dimension of the underlying MPI topology (i.e.
        process layout)
        @param gridResolution : resolution of the global mesh
        @param topoResolution : resolution of the mpi process distribution
        @return the required topology.
        """
        if topoResolution is None:
            newTopo = Cartesian(self, topoDim, gridResolution, ghosts=ghosts,
                                comm=comm)
        elif precomputed:
            newTopo = Cartesian.withPrecomputedResolution(self, topoResolution,
                                                          gridResolution,
                                                          localres=localres,
                                                          offset=offset,
                                                          ghosts=ghosts,
                                                          comm=comm)

        else:
            if fixedResolution:
                topoCreation = Cartesian.withResolutionFixed
            else:
                topoCreation = Cartesian.withResolution
            newTopo = topoCreation(self, shape=topoResolution,
                                   globalMeshResolution=gridResolution,
                                   ghosts=ghosts, comm=comm)
        newid = newTopo.getId()
        return self.topologies[newid]

    def checkTopo(self, newTopo):
        """
        Return the id of the input topology
        if it exists in the present domain list, else return -1.
        @param newTopo : the topology  to check
        @return id of the topology if present else -1.
        """
        otherid = -1
        for top in self.topologies.values():
            if newTopo == top:
                otherid = top.getId()
                break
        return otherid

    def register(self, newTopo):
        """
        Add a new topology in the list of this domain.
        Do nothing if an equivalent topology is already
        in the list.
        @param newTopo : the topology to be registered.
        @return the id of the new registered topology
        or of the corresponding 'old one' if it already exists.
        """
        otherid = self.checkTopo(newTopo)
        if otherid < 0:
            self.topologies[newTopo.getId()] = newTopo
            newTopo.setUp()
            newTopo.isNew = True
        else:
            # No registration
            newTopo = self.topologies[otherid]
            newTopo.isNew = False

        newid = newTopo.getId()
        return newid

    def remove(self, topo):
        """
        Remove a topology from the list of this domain.
        Do nothing if the topology does not exist in the list.
        Warning : the object topo is not destroyed, only
        removed from self.topologies.
        @param topo : the topology to be removed.
        @return either the od of the removed topology
        or -1 if nothing is done.
        """
        otherid = self.checkTopo(topo)
        if otherid >= 0:
            self.topologies.pop(otherid)
        return otherid

    def reset(self):
        for topo in self.topologies.values():
            del topo
        self.topologies = {}

    def printTopologies(self):
        """
        Print all topologies of the domain.
        """
        from parmepy.mpi import main_rank
        if main_rank == 0:
            for topo in self.topologies.values():
                print (topo)
