## @package parmepy.domain.obstacle
# Obstacles description (geometry).
#
#
# An 'obstacle' is the description of a sub-domain
# of the main physical domain with one or more user-defined
# functions of the space coordinates (at least).
#
# What for? \n
# Mainly to provide some 'index sets' to penalization operator.\n
# For a given obstacle, and a given discrete field, we must be able to call:
# \code
# cond = obstacle.discretize(topo)
# field[cond] = 1.9
# \endcode
# This example means that the field (discretized on topology topo)
# will be set to 1.9 everywhere inside the obstacle.
#
# 
# Obviouslvy the index sets will depend on the discretization
# of the domain (the underlying topology indeed).
# So each obstacle handles a dictionnary of boolean arrays. The keys
# of the dictionnaries are the topologies and the values some boolean arrays
# which values are true on and inside the object and false outside.
# \code
# obstacle.ind[topo] = someBooleanArray
# \endcode
# Each component of the dictionnary is created using the method 'discretize':
# \code
# # Create the boolean array that represents the obstacle for topology topo:
# someBooleanArray = obstacle.discretize(topo)
# # So for a field already discretized on topo, we may call
# field[someBooleanArray]
# \endcode
#
# A more complete example : initialize a scalar field with one inside a sphere
# and zero everywhere else.
# 
# \code
# Lx = Ly = Lz = 2
# dom = pp.Box(dimension=3, length=[Lx, Ly, Lz], origin=[-1., -1., -1.])
# # Definition of a sphere in the middle of the domain
# sphere = Sphere(dom, position=[0., 0., 0.], radius=0.5)
# # A topology
# topo = Cartesian(dom, 3, [33, 33, 33])
# # A scalar field on the domain :
# scal = pp.Field(domain=dom, name='Scalar')
# # Discretization on topo:
# scal_discr = scal.discretize(topo)
# # scal set to 1. inside the obstacle:
# condition = sphere.discretize(topo)
# scal.discreteFields[topo][condition] = 1.
# # equivalent to
# scal_discr[condition] = 1.
# # to set a value everywhere except in the sphere
# scal_discr[not condition] = 8.
# \endcode
