"""
@file controlBox.py
Define a sub-domain with a box-liked shape.
"""
from parmepy.domain.obstacle.obstacle import Obstacle
from parmepy.domain.obstacle.planes import SubSpace, SubPlane
from parmepy.mpi.mesh import SubMesh
import numpy as np
import parmepy.tools.numpywrappers as npw


class ControlBox(Obstacle):
    """
    Build a sub-domain, box-shaped
    ==> define set of indices inside this domain (ind member)
    and set of indices belonging to surfaces of this domain (slices members).
    Useful to define control volume to perform integration.
    See for example parmepy.operator.monitor.forces
    """

    def __init__(self, domain, origin, lengths, vd=0):
        """
        Build the volume of control
        @param domain : parmepy.domain of defintion
        @param origin : coordinates of the lowest point in the sub-domain
        @param lengths : lengths of box sides.
        @param velocity of the center of mass of the domain (rigid body),
        default = 0.
        """
        Obstacle.__init__(self, domain, vd=vd)

        self._dim = self.domain.dimension
        ## Lowest point of the box
        self.origin = npw.realarray(origin)
        ## Box's sides dimension
        self.lengths = npw.realarray(lengths)
        ## Dictionnary of local meshes, such that
        ## mesh[topo] is the restriction of topo.mesh
        ## to the current control box.
        self.mesh = {}
        self.upper = None
        self.lower = None
        self.upperS = None
        self.lowerS = None
        self.slices = {}
        self.indReduced = {}
        self._boxCreated = False
        ## Check if the defined box contains points
        ## for a given topology. self.isEmpty[topo] = False
        ## if some grid points are inside the box on
        ## the current processor for topo discretization.
        self.isEmpty = {}
        ## Dict of local coordinates for a given topology
        self.coords = {}
                ## Global resolution of the obstacle (plane, sub space ...)
        ## At the time, it's computed only for subspaces, by calling
        ## globalResolution method.
        self.gRes = None
        ## Global index in the original topology of the 'lowest' point
        ## of the obstacle. Only for subspaces.
        self.gstart = None

    def createVolumeAndSides(self, spaceStep):
        """
        @param[in] array of size self._dim, space step size in each direction
        This value will be used to compute a tolerance and detect
        points inside the box.
        """
        # Build Half-spaces indices list in all directions
        normalUp = np.identity(self._dim)
        normalDown = np.identity(self._dim) * -1.
        pointsUp = npw.zeros((self._dim, self._dim))
        # Control box will be used for integration, so we remove the
        # last point in the grid.
        boxlengths = self.lengths - spaceStep
        tol = spaceStep * 0.5

        for i in xrange(self._dim):
            pointsUp[:, i] = self.origin
        pointsUp.flat[::self._dim + 1] += self.lengths
        # -- Control volume : union of two halfspaces --
        if self.upper is None:
            self.upper = [SubSpace(self.domain, normal=normalUp[:, i],
                                   point=pointsUp[:, i],
                                   lengths=boxlengths,
                                   epsilon=tol[i])
                          for i in xrange(self._dim)]
        if self.lower is None:
            self.lower = [SubSpace(self.domain, normal=normalDown[:, i],
                                   point=self.origin, lengths=boxlengths,
                                   epsilon=tol[i])
                          for i in xrange(self._dim)]

        # Create objects to describe the sides of the box
        if self.upperS is None:
            self.upperS = [SubPlane(self.domain, normal=normalUp[:, i],
                                    point=pointsUp[:, i],
                                    lengths=boxlengths,
                                    epsilon=tol[i])
                           for i in xrange(self._dim)]

        if self.lowerS is None:
            self.lowerS = [SubPlane(self.domain, normal=normalDown[:, i],
                                    point=self.origin, lengths=boxlengths,
                                    epsilon=tol[i])
                           for i in xrange(self._dim)]
        self._boxCreated = True

    def discretize(self, topo):
        """
        Discretize the box volume and its surfaces.
        @param topo : the topology that described the discretization.
        """
        # Check if already done. If so, this function has no effect.
        if topo not in self.ind.keys():
            spaceStep = topo.mesh.space_step
            # -- Control volume : union of two halfspaces --
            if not self._boxCreated:
                self.createVolumeAndSides(spaceStep)

            # Discretize all volume and surfaces of
            # the box for topo
            for i in xrange(self._dim):
                self.lower[i].discretize(topo)
                self.upper[i].discretize(topo)
                self.lowerS[i].discretize(topo)
                self.upperS[i].discretize(topo)

            # 1 -- Compute list of indices inside the box,
            #  for topo --> ind[topo]
            self.ind[topo] = []

            self.ind[topo].append(np.logical_and(self.upper[0].ind[topo][0],
                                                 self.lower[0].ind[topo][0]))
            for i in xrange(1, self._dim):
                cond = np.logical_and(self.upper[i].ind[topo][0],
                                      self.lower[i].ind[topo][0])
                self.ind[topo][0] = np.logical_and(self.ind[topo][0], cond)

            ind = np.where(self.ind[topo][0])

            # 2 -- Convert ind[topo] (array of bool) to slices
            # which may be more convenient for computations
            # --> slices[topo]
            # + mesh[topo], a parmepy.mpi.SubMesh, useful
            # to get local coordinates and so on
            if ind[0].size == 0:
                self.slices[topo] = [slice(0, 0) for i in xrange(self._dim)]
                self.mesh[topo] = None
                self.isEmpty[topo] = True
            else:
                self.isEmpty[topo] = False
                ic = topo.mesh.iCompute
                lstart = [ind[i].min() if ind[i].size > 0 else None
                          for i in xrange(self._dim)]
                lstart = npw.integerarray([max(lstart[i], ic[i].start)
                                          for i in xrange(self._dim)])
                end = [ind[i].max() for i in xrange(self._dim)]
                end = npw.integerarray([min(end[i], ic[i].stop - 1)
                                       for i in xrange(self._dim)])
                # slice(start,end) --> end not included, so +1
                end += 1
                resol = end - lstart + 2 * topo.ghosts
                gstart = lstart + topo.mesh.global_start - topo.ghosts
                self.mesh[topo] = SubMesh(topo, gstart, resol)
                self.slices[topo] = [slice(lstart[i], end[i])
                                     for i in xrange(self._dim)]
                coords = []
                for i in xrange(self._dim):
                    cc = topo.mesh.coords[i].flat[self.slices[topo][i]]
                    coords.append(cc)
                coords = tuple(coords)
                self.coords[topo] = np.ix_(*coords)

                # --> self.ind[topo][0] components are True
                # for points inside the volume
                # --> self.slices[topo] represent the same thing
                # but using slices of numpy arrays.
                # Usage (vd being a numpy array discretized
                # on the whole domain, cb a control box):
                # # Set values to all points inside the control box
                # vd[cb.ind[topo][0]] = ...
                # # Get a sub-array of vd representing the control box
                # # and use it
                # result[...] = vd[cb.slices] + ...
                # The important difference between slices and ind is:
                # 1 - vd[ind] returns a 1D array whatever vd shape is.
                # 2 - vd[slices] return an array of the same dim as vd,
                # with shape given by slices.

            return self.ind[topo]

    def sub(self, obstacle, topo):
        """
        Remove all points corresponding to the input obstacle from
        the current control box
        """
        if topo not in self.ind.keys():
            obstacle.discretize(topo)
            self.discretize(topo)
            self.indReduced[topo] = []
            # Warning : obstacle may have several layers
            cond = obstacle.ind[topo][0]
            for i in xrange(1, len(obstacle.ind[topo])):
                cond = npw.asarray(np.logical_or(cond, obstacle.ind[topo][i]))
            cond = np.logical_not(cond)
            self.indReduced[topo].append(np.logical_and(self.ind[topo][0],
                                                        cond))
        return self.indReduced[topo][-1]

    def integrate_on_proc(self, field, topo, useSlice=True, component=0):
        """
        integrate field on the box
        """
        if useSlice:
            cond = self.slices[topo]
        else:
            iC = topo.mesh.iCompute
            cond = self.ind[topo][0][iC]
        dvol = npw.prod(topo.mesh.space_step)
        result = npw.sum(field.discretize(topo)[component][cond])
        result *= dvol
        return result

    def integrate(self, field, topo, useSlice=True,
                  component=0, root=0, mpiall=True):
        res = self.integrate_on_proc(field, topo, useSlice, component)
        if mpiall:
            return topo.comm.allreduce(res)
        else:
            return topo.comm.reduce(res, root=root)

    def integrateOnSurface(self, field, topo, normalDir=0, up=True,
                           useSlice=True, component=0, root=0, mpiall=True):
        """
        integrate field on top (if up is True) or down surface
        normal to a direction
        """
        res = self.integrateOnSurf_proc(field, topo, normalDir, up, useSlice,
                                        component)
        if mpiall:
            return topo.comm.allreduce(res)
        else:
            return topo.comm.reduce(res, root=root)

    def integrateOnSurf_proc(self, field, topo, normalDir=0,
                             up=True, useSlice=True, component=0):
        """
        integrate field on top and down surfaces normal to a direction
        """
        if up:
            surf = self.upperS[normalDir]
        else:
            surf = self.lowerS[normalDir]
        if useSlice:
            cond = surf.slices[topo]
        else:
            iC = topo.mesh.iCompute
            cond = surf.ind[topo][0][iC]
        dirs = np.logical_not(np.arange(self._dim) == normalDir)
        dS = npw.prod(topo.mesh.space_step[dirs])
        result = npw.sum(field.discretize(topo)[component][cond])
        result *= dS
        return result

    def globalResolution(self, parent_topo):
        """
        Compute 'global resolution' of the subplane
        """
        # We create a false topology, with only one proc
        # to get the global resolution for the plane.
        # This could also be done with local computation
        # sum but that would need a lot of communications.
        if parent_topo.rank == 0:
            color = 0
        else:
            color = 1
        subcomm = parent_topo.comm.Split(color)
        dimension = self.domain.dimension
        tmp = None
        if parent_topo.rank == 0:
            resolution = parent_topo.globalMeshResolution
            ghosts = parent_topo.ghosts
            topo = self.domain.getOrCreateTopology(3, resolution,
                                                   ghosts=ghosts, comm=subcomm)
            self.discretize(topo)
            sl = self.slices[topo]
            self.gRes = [sl[i].stop - sl[i].start for i in xrange(dimension)]
            self.gstart = [sl[i].start for i in xrange(dimension)]
            # if the topology has been created just to
            # get the global resolution, we can remove it
            if topo.isNew:
                self.domain.remove(topo)
                self.slices.pop(topo)
                self.ind.pop(topo)
            tmp = self.gRes + self.gstart
        tmp = parent_topo.comm.bcast(tmp)
        self.gRes = tmp[:dimension]
        self.gstart = tmp[dimension:]
        return self.gRes
