"""
@file disk.py
Rigid disk (2D)
"""
from parmepy.domain.obstacle.sphere import Sphere, HemiSphere
import numpy as np
import parmepy.tools.numpywrappers as npw


class Disk(Sphere):
    """
    Disk in a 2D domain.
    """

    def __init__(self, domain, position, radius=1.0, vd=0.0, porousLayers=[]):
        """
        Description of a disk in a domain.
        @param domain : the physical domain that contains the sphere.
        @param position : position of the center
        @param radius : sphere radius, default = 1
        @param porousLayers : a list of thicknesses
        for successive porous layers
        radius is the inside sphere radius and thicknesses are given from
        inside layer to outside one.
        @param vd : velocity of the disk (considered as a rigid body),
        default = 0.
        """
        Sphere.__init__(self, domain, position, radius, vd, porousLayers)
        assert self.domain.dimension == 2

        def dist(x, y, R):
            return npw.asarray(np.sqrt((x - self.position[0]) ** 2
                                       + (y - self.position[1]) ** 2) - R)
        self.chi = [dist]

    def __str__(self):
        s = 'Disk of radius ' + str(self.radius)
        s += ' and center position ' + str(self.position)
        return s


class HalfDisk(HemiSphere):
    """
    Half disk in a 2D domain.
    """
    def __init__(self, domain, position, radius=1.0, vd=0.0, porousLayers=[]):
        """
        Constructor for the semi-disk.
        @param domain : the physical domain that contains the sphere.
        @param position : position of the center
        @param radius : sphere radius, default = 1
        (if box ...)
        @param vd : velocity of the disk (considered as a rigid body),
        default = 0.
        """
        HemiSphere.__init__(self, domain, position, radius, vd, porousLayers)
        assert self.domain.dimension == 2

        def dist(x, y, R):
            """
            """
            return npw.asarray(np.sqrt((x - self.position[0]) ** 2
                                       + (y - self.position[1]) ** 2) - R)
        self.chi = [dist]

        def LeftBox(x, y):
            return x - self.position[0]

        self.LeftBox = LeftBox

    def __str__(self):
        s = 'Half-disk of radius ' + str(self.radius)
        s += ' and center position ' + str(self.position)
        return s
