"""@file obstacle.py

General interface to define a new geometry
inside a domain (sphere, control box ...)
"""
import numpy as np


class Obstacle(object):
    """Ddescription of a physical obstacle.
    An obstacle is the geometrical description of
    a physical sub-domain.
    """
    def __init__(self, domain, formula=None, vd=0.0):
        """ Constructor
        @param domain : the domain that contains this obstacle.
        @param formula : a list of functions that describe
        the geometry of the obstacle, see parmepy.domain.obstacle for details.
        @param vd : velocity of the obstacle (considered as a rigid body),
        default = 0.
        """
        ## Domain.
        self.domain = domain
        from parmepy.domain.box import Box
        assert isinstance(domain, Box),\
            'Obstacle only implemented for box-like domains'
        ## Obstacle dimension.
        self.dimension = domain.dimension
        ## A function that describe the geometry of the obstacle.
        ## see parmepy.domain.obstacle.
        self.chi = []
        if formula is not None:
            if isinstance(formula, list):
                for func in formula:
                    self.chi.append = np.vectorize(func)
            else:
                self.chi = [np.vectorize(formula)]
        ## A dictionnary of lists of indices ...
        ## ind[topo][i] represents the set of points of the domain
        ## discretized with topo that are in the area defined with chi[i].
        self.ind = {}
        ## Velocity of the center of mass of the obstacle
        ## (considered as a rigid body)
        self.vd = vd

    def discretize(self, topo):
        """
        For a given topology, computes the list of points in the domain
        that belongs to the obstacle.
        Add a new index into self.indices.
        @param topo : topology specification for discretization
        @return an array of bool such that array[i,j,k] = True if
        point(i,j,k) is in the obstacle.

        Note FP : there are two ways to 'save' which points are
        in the obstacle : either we set a test function and
        fill a boolean numpy array (case A) or we compute domain.dimension
        lists of indice (B).
        - case A : indices[topo] is a numpy array of the same
        size as topo.mesh.
         +++ : very fast to compute, very fast to apply
         --- : needs more memory (size of bool * topo.mesh.size)
        - case B : indices[topo] is a tupple of lists, like (ix, iy, iz).
        A point inside the obstacle is thus given
        by the indices ix[i], iy[i], iz[i]
         +++ : needs less memory
         --- : very slow to compute/apply compared to case A
        Default choice = case A
        \todo : provide a way for user to choose between case A and B.
        Note FP 2: maybe that would be better to save indices in
        operator (penalization) and not in obstacle, to save memory?
        """
        # first check if we have already compute indices for
        # this topology
        if topo not in self.ind[0].keys():
            # for each indicator function
            self.ind[topo] = []
            for func in self.chi:
                # current function position
                i = self.chi.index(func)
                # apply indicator function on topo local mesh
                self.ind[topo].append(self.chi[i](*topo.mesh.coords) <= 0)

        return self.ind[topo]
