"""
@file planes.py
Plate-like sub-domains at boundaries, normal
to a given direction.
"""
from parmepy.domain.obstacle.obstacle import Obstacle
import numpy as np
import parmepy.tools.numpywrappers as npw


class HalfSpace(Obstacle):
    """
    Divide domain into two sub-spaces, on each side of a plane
    defined by its normal and a point.
    Indices of this.ind describe the half-space below the plane,
    'normal' being the outward normal of the plane.
    """

    def __init__(self, domain, normal, point, epsilon=1e-2, vd=0.0):
        """
        Half space define by the points of the domain on one side
        of a plane.
        @param domain : the physical domain that contains the plane
        @param normal : outward normal
        @param point : coordinates of a point of the plane.
        @param epsilon : tolerance
        @param vd : velocity of obstacle (considered as a rigid body),
        default = 0.
        """
        Obstacle.__init__(self, domain, vd=vd)
        assert epsilon > 0.0, 'Tolerance value must be positive'
        ## Tolerance used to considered that points at the boundary are
        ## in the subspace. Good choice may be grid space_step / 2.
        self.epsilon = epsilon
        ## Direction of the normal to the plate (0:x, 1:y, 2:z))
        ## normal is the 'outer' normal of the 'in' subspace.
        self.normal = npw.integerarray(normal)
        self.point = point

        def Outside(*coords):
            return sum([(coords[i] - self.point[i]) * self.normal[i]
                        for i in xrange(self.domain.dimension)])

        ## Test function for half-space.
        ## Positive value if outside subdomain else negative
        self.chi = [Outside]
        self.slices = {}
        ## Global resolution of the obstacle (plane, sub space ...)
        ## At the time, it's computed only for subspaces, by calling
        ## globalResolution method.
        self.gRes = None
        ## Global index in the original topology of the 'lowest' point
        ## of the obstacle. Only for subspaces.
        self.gstart = None

    def discretize(self, topo):
        # first check if we have already compute indices for
        # this topology

        if topo not in self.ind.keys():
            self.ind[topo] = []
            # apply indicator function on topo local mesh
            cond = npw.asarray(self.chi[0](*topo.mesh.coords) <= self.epsilon)
            self.ind[topo].append(cond)

        return self.ind[topo]

    def __str__(self):
        s = 'Plane normal to vector' + str(self.normal)
        s += ' going through point ' + str(self.point)
        return s


class Plane(HalfSpace):
    """
    A plane in the domain, defined by its normal and a point.
    Indices of plane.ind describe the points belonging to the plane.
    """
    def discretize(self, topo):
        # first check if we have already compute indices for
        # this topology

        if topo not in self.ind.keys():
            self.ind[topo] = []
            # apply indicator function on topo local mesh
            cond = npw.abs(self.chi[0](*topo.mesh.coords)) < self.epsilon
            self.ind[topo].append(cond)

            # assert that the plane is a real surface, i.e.
            # only one value for coords[normalDir].
            # The expr is a bit tricky but it works ...
            ndir = np.where(self.normal != 0)[0][0]
            assert assertSubPlane(ndir, self.ind[topo][0], *topo.mesh.coords),\
                'Your plane is not a surface but a volume.\
                Please reduce epsilon value.'

        return self.ind[topo]


class SubSpace(HalfSpace):
    """
    Define a rectangular space in a plane normal to one
    coord. axis and the subspace below this suface.
    'Below' = direction opposite to the outward normal of the plane
    (input param)
    """
    def __init__(self, domain, normal, point, lengths, epsilon=1e-2, vd=0.0):
        """
        @param domain : the physical domain that contains the space
        @param normal : outward normal
        @param point : coordinates of a point of the plane.
        @param lengths : lengths of the subplane
        @param epsilon : tolerance
        @param vd : velocity of the obstacle (considered as a rigid body),
        default = 0.
        """
        HalfSpace.__init__(self, domain, normal, point, epsilon, vd)

        def dist(cdir, val, *coords):
            return coords[cdir] - val

        self.dist = dist
        self.origin = npw.realarray(point)
        self.max = self.origin + npw.realarray(lengths)
        self.lengths = npw.realarray(lengths)
        ndir = np.where(self.normal != 0)[0][0]
        if normal[ndir] > 0:
            self.max[ndir] = self.origin[ndir]
        elif normal[ndir] < 0:
            self.max[ndir] = self.domain.max[ndir]
        # Only implemented for planes orthogonal to coord. axes
        assert len(self.normal[self.normal == 0]) == self.domain.dimension - 1
        self.coords = {}
        ## Check if some grid points are present inside the current object
        ## for the current mpi proc. If not, isEmpty = True.
        self.isEmpty = {}

    def discretize(self, topo):
        # first check if we have already compute indices for
        # this topology
        condMax = [0] * self.domain.dimension
        condMin = [0] * self.domain.dimension
        if topo not in self.ind.keys():
            self.ind[topo] = []
            # apply indicator function on topo local mesh
            coords = topo.mesh.coords
            cond = npw.asarray(self.chi[0](*coords) < self.epsilon)
            indices = np.where(self.normal == 0)[0]

            for i in indices:
                condMax[i] = self.dist(i, self.max[i], *coords) < self.epsilon
                condMin[i] = self.dist(i, self.origin[i], *coords) > - self.epsilon
                condMin[i] = np.logical_and(condMax[i], condMin[i])
                cond = npw.asarray(np.logical_and(cond, condMin[i]))

            self.ind[topo].append(cond)

        return self.ind[topo]


class SubPlane(SubSpace):
    """
    Define a rectangular surf in a plane normal to one
    coord. axis.
    """
    def discretize(self, topo):
        # first check if we have already compute indices for
        # this topology
        dim = self.domain.dimension
        condMax = [0] * dim
        condMin = [0] * dim
        if topo not in self.ind.keys():
            self.ind[topo] = []
            # apply indicator function on topo local mesh
            coords = topo.mesh.coords
            cond = npw.abs(self.chi[0](*coords)) < self.epsilon
            indices = np.where(self.normal == 0)[0]
            for i in indices:
                condMax[i] = self.dist(i, self.max[i], *coords) < self.epsilon
                condMin[i] = self.dist(i, self.origin[i], *coords) > -self.epsilon
                condMin[i] = np.logical_and(condMax[i], condMin[i])
                cond = npw.asarray(np.logical_and(cond, condMin[i]))

            self.ind[topo].append(cond)
            ilist = np.where(cond)
            if ilist[0].size == 0:
                self.slices[topo] = [slice(0, 0) for i in xrange(dim)]
                self.isEmpty[topo] = True
            else:
                self.isEmpty[topo] = False
                start = [ilist[i].min() for i in xrange(dim)]
                # Ghost points must not be included into surf. points
                ic = topo.mesh.iCompute
                start = [max(start[i], ic[i].start) for i in xrange(dim)]
                end = [ilist[i].max() for i in xrange(dim)]
                end = npw.integerarray([min(end[i], ic[i].stop - 1)
                                        for i in xrange(dim)])
                end += 1
                ndir = np.where(self.normal != 0)[0][0]
                end[ndir] = start[ndir] + 1
                self.slices[topo] = [slice(start[i], end[i])
                                     for i in xrange(dim)]
                assert assertSubPlane(ndir, cond, *topo.mesh.coords),\
                    'Your plane is not a surface but a volume.\
                Please reduce epsilon value.'
            subcoords = []
            # !! Warning : slices will be used for integration,
            # so the last point in each dir is not included.
            # Same thing for coords.
            for i in xrange(dim):
                subcoords.append(coords[i].flat[self.slices[topo][i]])
            subcoords = tuple(subcoords)
            self.coords[topo] = np.ix_(*subcoords)
        return self.ind[topo]

    def globalResolution(self, parent_topo):
        """
        Compute 'global resolution' of the subplane
        """
        # We create a false topology, with only one proc
        # to get the global resolution for the plane.
        # This could also be done with local computation
        # sum but that would need a lot of communications.
        if parent_topo.rank == 0:
            color = 0
        else:
            color = 1
        subcomm = parent_topo.comm.Split(color)
        dimension = self.domain.dimension
        tmp = None
        if parent_topo.rank == 0:
            resolution = parent_topo.globalMeshResolution
            ghosts = parent_topo.ghosts
            topo = self.domain.getOrCreateTopology(3, resolution,
                                                   ghosts=ghosts, comm=subcomm)
            self.discretize(topo)
            sl = self.slices[topo]
            self.gRes = [sl[i].stop - sl[i].start for i in xrange(dimension)]
            self.gstart = [sl[i].start for i in xrange(dimension)]
            # if the topology has been created just to
            # get the global resolution, we can remove it
            if topo.isNew:
                self.domain.remove(topo)
                self.slices.pop(topo)
                self.ind.pop(topo)
            tmp = self.gRes + self.gstart
        tmp = parent_topo.comm.bcast(tmp)
        self.gRes = tmp[:dimension]
        self.gstart = tmp[dimension:]
        return self.gRes


class PlaneBoundaries(Obstacle):
    """
    Defines top and down (meaning for min and max value in
    a given direction) planes at boundaries.
    All points in the spaces above the top plane and below the down plane
    will be included in the PlaneBoundaries list of indices.
    Thickness of the top/down areas is given as an input param.
    Example for z dir:
    \f$ \{x,y,z\} \ for \ z_{max} - \epsilon \leq z \leq z_{max} + \epsilon
    \ or \ z_{min} - \epsilon \leq z \leq z_{min}\f$
    """

    def __init__(self, domain, normal_dir, thickness=0.1):
        """
        Description of a sphere in a domain.
        @param domain : the physical domain that contains the sphere.
        @param thickness : thickness of boundary areas
        @param vd : velocity of obstacle (considered as a rigid body),
        default = 0.
        """
        Obstacle.__init__(self, domain, vd=0.0)
        assert thickness > 0.0, 'Plate thickness must be positive'
        ## Thickness/2
        self.thickness = thickness
        ## Direction of the normal to the plate (0:x, 1:y, 2:z))
        normalUp = np.zeros((self.domain.dimension))
        normalUp[normal_dir] = -1
        pointUp = npw.zeros((self.domain.dimension))
        pointUp[normal_dir] = domain.max[normal_dir] - thickness
        self.upper = HalfSpace(domain, normal=normalUp, point=pointUp,
                               epsilon=1e-3)
        normalDown = np.zeros((self.domain.dimension))
        normalDown[normal_dir] = 1
        pointDown = npw.zeros((self.domain.dimension))
        pointDown[normal_dir] = domain.origin[normal_dir] + thickness
        self.lower = HalfSpace(domain, normal=normalDown, point=pointDown,
                               epsilon=1e-3)

    def discretize(self, topo):
        # first check if we have already compute indices for
        # this topology

        self.lower.discretize(topo)
        self.upper.discretize(topo)
        if topo not in self.ind.keys():
            # Warning FP : ind[topo] must be a list to be coherent
            # with sphere definition, where porous layers are allowed.
            # todo if required : add porous layers for planes.
            self.ind[topo] = []
            self.ind[topo].append(np.logical_or(self.upper.ind[topo][0],
                                                self.lower.ind[topo][0]))

        return self.ind[topo]


def assertSubPlane(ndir, ind, *coords):
    dim = len(coords)
    if dim == 2:
        return assertline(ndir, ind, *coords)
    elif dim == 3:
        return assertsurface(ndir, ind, *coords)


def assertsurface(nd, ind, *coords):

    dim = len(coords)
    shape = np.zeros(dim, dtype=np.int32)
    shape[:] = [coords[i].shape[i] for i in xrange(dim)]
    cshape = coords[nd].shape
    if nd == 0:
        return max([a.max() - a.min()
                    for a in [coords[nd][ind[:, i, j]]
                              for i in xrange(shape[1])
                              for j in xrange(shape[2])
                              if coords[nd][ind[:, i, j]].size
                              > 0]] + [0]) == 0.
    elif nd == 1:
        return max([a.max() - a.min()
                    for a in [coords[nd][ind[i, :, j].reshape(cshape)]
                              for i in xrange(shape[0])
                              for j in xrange(shape[2])
                              if coords[nd][ind[i, :, j].reshape(cshape)].size
                              > 0]] + [0]) == 0.

    else:
        return max([a.max() - a.min()
                    for a in [coords[nd][ind[i, j, :].reshape(cshape)]
                              for i in xrange(shape[0])
                              for j in xrange(shape[1])
                              if coords[nd][ind[i, j, :].reshape(cshape)].size
                              > 0]] + [0]) == 0.


def assertline(nd, ind, *coords):

    dim = len(coords)
    shape = np.zeros(dim, dtype=np.int32)
    shape[:] = [coords[i].shape[i] for i in xrange(dim)]
    cshape = coords[nd].shape
    if nd == 0:
        return max([a.max() - a.min()
                    for a in [coords[nd][ind[:, i]]
                              for i in xrange(shape[1])
                              if coords[nd][ind[:, i]].size
                              > 0]] + [0]) == 0.
    elif nd == 1:
        return max([a.max() - a.min()
                    for a in [coords[nd][ind[i, :].reshape(cshape)]
                              for i in xrange(shape[0])
                              if coords[nd][ind[i, :].reshape(cshape)].size
                              > 0]] + [0]) == 0.
