"""
@file sphere.py
Spherical or hemispherical sub-domain.
"""
from parmepy.domain.obstacle.obstacle import Obstacle
import numpy as np
import parmepy.tools.numpywrappers as npw


class Sphere(Obstacle):
    """
    Spherical domain.
    """

    def __init__(self, domain, position, radius=1.0,
                 vd=0.0, porousLayers=[]):
        """
        Description of a sphere in a domain.
        @param domain : the physical domain that contains the sphere.
        @param position : position of the center
        @param radius : sphere radius, default = 1
        @param porousLayers : a list of thicknesses
        for successive porous layers
        radius is the inside sphere radius and thicknesses are given from
        inside layer to outside one.
        @param vd : velocity of the sphere (considered as a rigid body),
        default = 0.
        """
        Obstacle.__init__(self, domain, vd=vd)
        ## Radius of the sphere
        self.radius = radius
        ## Center position
        self.position = np.asarray(position)

        def dist(x, y, z, R):
            """
            """
            return npw.asarray(np.sqrt((x - self.position[0]) ** 2
                                       + (y - self.position[1]) ** 2
                                       + (z - self.position[2]) ** 2) - R)

        self.chi = [dist]
        ## List of thicknesses for porous layers
        self.layers = porousLayers

    def discretize(self, topo):
        # first check if we have already compute indices for
        # this topology

        if topo not in self.ind.keys():
            currentRadius = self.radius
            self.ind[topo] = []
            # First, internal sphere
            args = (currentRadius,)
            self.ind[topo].append(self.chi[0](*(topo.mesh.coords + args)) <= 0)
            # Then each layer from inside to outside
            # for each indicator function
            for thickness in self.layers:
                # apply indicator function on topo local mesh
                args = (currentRadius,)
                condA = self.chi[0](*(topo.mesh.coords + args)) > 0
                args = (currentRadius + thickness,)
                condB = self.chi[0](*(topo.mesh.coords + args)) <= 0
                self.ind[topo].append(np.logical_and(condA, condB))
                # update current radius
                currentRadius = currentRadius + thickness

        return self.ind[topo]

    def __str__(self):
        """ToString method"""
        s = 'Sphere of radius ' + str(self.radius) + ' and center position '
        s += str(self.position)
        return s


class HemiSphere(Sphere):
    """
    HemiSpherical domain.
    Area defined by the intersection of a sphere and the volume where
    x < xs for xs == x position of the center of the sphere.
    """
    def __init__(self, domain, position, radius=1.0,
                 vd=0.0, porousLayers=[]):
        """
        Description of a sphere in a domain.
        @param domain : the physical domain that contains the sphere.
        @param position : position of the center
        @param radius : sphere radius, default = 1
        @param porousLayers : a list of thicknesses
        for successive porous layers
        radius is the inside sphere radius and thicknesses are given from
        inside layer to outside one.
        @param vd : velocity of the sphere (considered as a rigid body),
        default = 0.
        """
        Sphere.__init__(self, domain, position, radius, vd, porousLayers)

        def LeftBox(x, y, z):
            return x - self.position[0]
        self.LeftBox = LeftBox

    def discretize(self, topo):
        # first check if we have already compute indices for
        # this topology
        if topo not in self.ind.keys():
            currentRadius = self.radius
            self.ind[topo] = []
            # check if we are in the left half-box
            cond0 = self.LeftBox(*(topo.mesh.coords)) <= 0
            # First, internal sphere
            args = (currentRadius,)
            condA = self.chi[0](*(topo.mesh.coords + args)) <= 0
            self.ind[topo].append(np.logical_and(condA, cond0))
            # Then each layer from inside to outside
            # for each indicator function
            for thickness in self.layers:
                # apply indicator function on topo local mesh
                args = (currentRadius,)
                condA = self.chi[0](*(topo.mesh.coords + args)) > 0
                args = (currentRadius + thickness,)
                condB = self.chi[0](*(topo.mesh.coords + args)) <= 0
                np.logical_and(condA, condB, condA)
                np.logical_and(condA, cond0, condA)
                condA = npw.asarray(condA)
                self.ind[topo].append(condA)
                # update current radius
                currentRadius = currentRadius + thickness

        return self.ind[topo]

    def __str__(self):
        s = 'hemisphere of radius ' + str(self.radius)
        s += ' and center position ' + str(self.position)
        return s
