"""
Testing parmepy.domain.obstacle.Obstacle
"""
import parmepy as pp
from parmepy.fields.continuous import Field
from parmepy.mpi.topology import Cartesian
from parmepy.domain.obstacle.sphere import Sphere, HemiSphere
from parmepy.domain.obstacle.disk import Disk, HalfDisk
from parmepy.domain.obstacle.planes import HalfSpace, Plane, SubSpace,\
    SubPlane, PlaneBoundaries
from parmepy.domain.obstacle.controlBox import ControlBox
import numpy as np
from parmepy.constants import CHECK_F_CONT


nb = 129
Lx = Ly = Lz = 2
dom = pp.Box(dimension=3, length=[Lx, Ly, Lz], origin=[-1., -1., -1.])
dom2D = pp.Box(dimension=2, length=[Lx, Ly], origin=[-1., -1.])
resol3D = [nb, nb, nb]
resol2D = [nb, nb]
scal = Field(domain=dom)
scal2D = Field(domain=dom2D)
topo = Cartesian(dom, 3, resol3D)
topo2D = Cartesian(dom2D, 2, resol2D)
coords = topo.mesh.coords
coords2D = topo2D.mesh.coords
scald = scal.discretize(topo).data[0]
scald2D = scal2D.discretize(topo2D).data[0]
h3d = topo.mesh.space_step
h2d = topo2D.mesh.space_step
dvol = np.prod(h3d)
ds = np.prod(h2d)
import math
pi = math.pi
tol = 1e-6
lengths = np.asarray([20 * h3d[0], 22 * h3d[1], 31 * h3d[2]])
rlengths = lengths + h3d
rlengths2d = lengths[:2] + h2d
scald[:] = 1.
scald2D[:] = 1.


def testSphere():
    scald[:] = 1.
    rad = 0.3
    sphere = Sphere(dom, position=[0., 0., 0.],
                    radius=rad, porousLayers=[0.13])

    sphere.discretize(topo)
    ind = sphere.ind[topo][0]
    (ix, iy, iz) = topo.mesh.indices([-0.2, 0, 0.2])
    assert ind[ix, iy, iz]
    (ix, iy, iz) = topo.mesh.indices([0.5, 0.1, 0.2])
    assert not ind[ix, iy, iz]


def testHemiSphere():
    scald[:] = 1.
    rad = 0.3
    sphere = HemiSphere(dom, position=[0., 0., 0.],
                        radius=rad, porousLayers=[0.13])

    sphere.discretize(topo)
    ind = sphere.ind[topo][0]
    (ix, iy, iz) = topo.mesh.indices([-0.3, 0., 0.])
    assert ind[ix, iy, iz]
    (ix, iy, iz) = topo.mesh.indices([0.3, 0., 0.])
    assert not ind[ix, iy, iz]
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testDisk():
    scald2D[:] = 1.
    rad = 0.3
    sphere = Disk(dom2D, position=[0., 0.],
                  radius=rad, porousLayers=[0.13])

    sphere.discretize(topo2D)
    ind = sphere.ind[topo2D][0]
    (ix, iy) = topo2D.mesh.indices([-0.2, 0.])
    assert ind[ix, iy]
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testHalfDisk():
    scald2D[:] = 1.
    rad = 0.3
    sphere = HalfDisk(dom2D, position=[0., 0.],
                      radius=rad, porousLayers=[0.13])

    sphere.discretize(topo2D)
    ind = sphere.ind[topo2D][0]
    (ix, iy) = topo2D.mesh.indices([-0.2, 0.])
    assert ind[ix, iy]
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testHalfSpace2D():
    hsp = HalfSpace(dom2D, [1, 1], [0., 0.])
    hsp.discretize(topo2D)
    ind = hsp.ind[topo2D][0]
    (ix, iy) = topo2D.mesh.indices([-0.8, 0.5])
    assert ind[ix, iy]
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testHalfSpace3D():
    hsp = HalfSpace(dom, [1, 1, 1], [0., 0., 0.])
    hsp.discretize(topo)
    ind = hsp.ind[topo][0]
    (ix, iy, iz) = topo.mesh.indices([-0.8, 0.5, -0.5])
    assert ind[ix, iy, iz]
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testPlane2D():
    plane = Plane(dom2D, [1, 1], [0., 0.])
    plane.discretize(topo2D)
    ind = plane.ind[topo2D][0]
    (ix, iy) = topo2D.mesh.indices([-0.5, 0.5])
    assert ind[ix, iy]
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testPlane3D():
    plane = Plane(dom, [1, 1, 1], [0., 0., 0.])
    plane.discretize(topo)
    ind = plane.ind[topo][0]
    (ix, iy, iz) = topo.mesh.indices([-0.3, 0.5, -0.2])
    assert ind[ix, iy, iz]
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testSubSpace2D():
    ssp = SubSpace(dom2D, [1, 0.], [0., 0.], lengths[:2])
    ssp.discretize(topo2D)
    ind = ssp.ind[topo2D][0]
    (ix, iy) = topo2D.mesh.indices([-0.5, 0.2])
    assert ind[ix, iy]


def testSubSpace3D():
    ssp = SubSpace(dom, [0, 1, 0], [0., 0., 0.], lengths)
    ssp.discretize(topo)
    ind = ssp.ind[topo][0]
    (ix, iy, iz) = topo.mesh.indices([0.3, -0.1, 0.2])
    assert ind[ix, iy, iz]
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testSubPlane2D():
    ssp = SubPlane(dom2D, [1, 0], [0., 0.], lengths[:2])
    ssp.discretize(topo2D)
    ind = ssp.ind[topo2D][0]
    ll = np.sum(scald2D[ind]) * h2d[1]
    rll = rlengths2d[1]
    assert abs(ll - rll) < tol
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testSubPlane3D():
    ssp = SubPlane(dom, [0, 1, 0], [0., 0., 0.], lengths)
    ssp.discretize(topo)
    ind = ssp.ind[topo][0]
    surf = np.sum(scald[ind]) * ds
    rsurf = rlengths[0] * rlengths[2]
    assert abs(surf - rsurf) < tol
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testPlaneBC2D():
    bc = PlaneBoundaries(dom2D, 1, thickness=0.2)
    bc.discretize(topo2D)
    ind = bc.ind[topo2D][0]
    (ix, iy) = topo2D.mesh.indices([-0.5, - Ly * 0.5])
    assert ind[ix, iy]
    (ix, iy) = topo2D.mesh.indices([-0.5, Ly * 0.5 - 2 * h2d[1]])
    assert ind[ix, iy]
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testPlaneBC3D():
    bc = PlaneBoundaries(dom, 1, thickness=0.2)
    bc.discretize(topo)
    ind = bc.ind[topo][0]
    (ix, iy, iz) = topo.mesh.indices([-0.5, -Ly * 0.5, 0.3])
    assert ind[ix, iy, iz]
    (ix, iy, iz) = topo.mesh.indices([-0.5, Ly * 0.5 - 2 * h2d[1], 0.3])
    assert ind[ix, iy, iz]
    assert ind.flags.f_contiguous is CHECK_F_CONT


def testControlBox2D():
    lx = 10 * h3d[0]
    ly = 22 * h3d[1]

    cb = ControlBox(dom2D, [-0.5, -0.5], [lx, ly])
    cb.discretize(topo2D)
    surf = cb.integrate(scal2D, topo2D)
    rsurf = lx * ly
    assert abs(surf - rsurf) < tol
    assert cb.ind[topo2D][0].flags.f_contiguous is CHECK_F_CONT


def testControlBox3D():
    ll = np.asarray([0] * 3)
    ll[0] = 10 * h3d[0]
    ll[1] = 22 * h3d[1]
    ll[2] = 51 * h3d[2]

    cb = ControlBox(dom, [0.5, -0.5, -0.5], ll)
    cb.discretize(topo)
    vol = cb.integrate(scal, topo)
    rvol = np.prod(ll)
    assert abs(rvol - vol) < tol

    vol = cb.integrate(scal, topo, useSlice=False)
    assert abs(rvol - vol) < tol
    ind = np.asarray([0, 1, 2])
    for i in xrange(3):
        surfUp = cb.integrateOnSurface(scal, topo, normalDir=i, up=True)
        surfDown = cb.integrateOnSurface(scal, topo, normalDir=i, up=False)
        j = np.where(ind != i)
        sref = np.prod(ll[j])
        assert abs(surfUp - sref) < tol
        assert abs(surfDown - sref) < tol

    assert cb.ind[topo][0].flags.f_contiguous is CHECK_F_CONT


def testControlBoxSphere():
    lx = 1.5
    ly = 1.5
    lz = 1.5
    rad = 0.2
    cb = ControlBox(dom, [-0.75, -0.75, -0.75], [lx, ly, lz])
    layer = 2 * h3d[0]
    sphere = Sphere(dom, position=[0., 0., 0.],
                    radius=rad, porousLayers=[layer])
    cb.sub(sphere, topo)
    ind = cb.indReduced[topo][0]
    (ix, iy, iz) = topo.mesh.indices([0.1, 0.0, 0.])
    assert not ind[ix, iy, iz]
    (ix, iy, iz) = topo.mesh.indices([0.3, 0.0, 0.])
    assert ind[ix, iy, iz]
    assert ind.flags.f_contiguous is CHECK_F_CONT
