## @package parmepy.fields
# Tools to build and use 1, 2 or 3D fields (scalar, vector or
# any number of components.)
#
# Fields are used to described the physical continuous variables involved in
# a problem on a given domain.
#
# To define a continuous field, one must provide :
# - the domain on which it is defined
# and optionnaly:
# - its number of components
# - a user defined function formula used to initialize the field
# - a name
#
# A continuous field may have several discretisations (i.e. may
# be associated with different parmepy topologies).
# 
# \section fieldsTuto How to define and discretize a field
#
# \b Classes : parmepy.fields.continuous.Field,
#              parmepy.fields.discrete.DiscreteField
#
# First of all, one must have defined a domain :
# \code
# from parmepy.domain.box import Box
# from parmepy.fields.continuous import Field
#
# # First, the domain : 3D and 2D boxes
# dom3D = Box()
# dom2D = Box(dimension=2, length=[1.0, 1.0], origin=[0., 0.])
#\endcode
#
# Definition of a scalar fields on these domains:
# \code
# rho = Field(dom3D)
# rho2 = Field(dom3D, name="second_field")
# scal_2D = Field(dom2D)
# ...
# \endcode
#
# Vector fields :
# \code
# velocity = Field(dom3D, isVector=True)
# # ---> velocity is composed of 3 components.
# velo2D = Field(dom2D, isVector=True)
# # ---> velocity is composed of 2 components.
#\endcode
#
# Other fields :
# \code
# Jacobian = Field(dom3D, nbComponents = 9, name='somename')
# # ---> Creates a 9 components field defined on a 3D domain.
# \endcode
#
# Now, let's turn to fields discretisation. We suppose that
# two parmepy topologies have been defined (see parmepy.mpi.topology)
# on domain dom3D:
# - topo1 : with a local (for each mpi process) resolution "r1"
# - topo2 : r2
# \code
# rho.discretize(topo1)
# # --> allocate memory for numpy array of shape r1
# velocity.discretize(topo2)
# # --> allocate memory for 3 numpy arrays of shape r2
# velocity.discretize(topo1)
# #  --> allocate memory for 3 numpy arrays of shape r1
#\code
# 
# To handle the underlying numpy arrays do :
# \code
# velo_d = velocity.discreteFields[topo1]
# # --> the discrete field for velocity on resolution r1
# print velo2_d.data[0] # first component of velocity on topo1.
# \endcode
#
# 
# \section fieldsInit How to initialize a field
#
# The simplest way :
# \code
# velocity->setTopoInit(topo1)
# velocity->initialize()
# \endcode
# It is necessary to set a default topology on which the field will
# be initialized and then call initialize function.\n
# Remark : if topoinit is not set, the first created discretisation (topo)
# of the field will be used.
#
# All components of the field on topo1 will be set to 0.0.
#
# It's also possible to set an analytical formula (depending on
# domain coordinates) to initialize a field?
#
# There are two possibilities to define a user function:
#  - either a function with numpy arrays arguments, see func1 below,
#  - or a function that will be vectorized during initialize call, see func2.
#
# Notice that :
# - func_1 required arguments : (res, x, y, ...) with res a list of numpy
# arrays which corresponds to the return values of the function and
# x, y ... (depending on the domain dimension) the grid coordinates.
# - func_2 required arguments: (x,y ...) only but will probably be less
# efficient and may result in memory reallocation.
# - optional extra args can be added in the function (see example below)
#
# First example : a function of numpy arrays (func_1 case):
#
# \code
# def func_1_3D(res, x, y, z):
#     res[0][...] = np.sin(x)
#     res[1][...] = np.sin(y)
#     res[2][...] = np.sin(z)
#     return res
#
# def func_1_2D(res, x, y):
#     res[0][...] = np.sin(x)
#     res[1][...] = np.sin(y)
#     return res
#
# velocity = Field(dom3D, isVector=True, formula=func_1_3D)
# velo2D = Field(dom2D, isVector=True, formula=func_1_2D)
# velocity.discretize(topo)
# velo2D.discretize(topo2D)
#
# velocity.setTopoInit(topo)
# velo2D.setTopoInit(topo2D)
#
# velocity.initialize()
# velo2D.initialize()
# \endcode
#
# Obviously, the number of components return by the user function must
# corresponds to the number of components of the field.
#
# Second example : a function of tha will be vectorized (func_2 case).
# The code is almost the same:
# \code
# def func_2_3D(x, y, z):
#     rx = np.sin(x)
#     ry = np.sin(y)
#     rz = np.sin(z)
#     return rx, ry, rz
#
# velocity = Field(dom3D, isVector=True, formula=func_2_3D)
# velocity.discretize(topo)
# velocity.setTopoInit(topo)
# velocity.initialize(doVectorize=True)
# # --> we ask for a vectorization of func_2 so that
# # it can use arrays of coordinates as input and return
# # numpy arrays to compute velocity discrete field.
# \endcode
#
# 
# Third example : add extra parameters.\n
# Let's assume that we need extra arguments to compute the field:
# \code
# def func_1_3D(res, x, y, z, alpha, t):
#     res[0][...] = np.sin(alpha * x + t)
#     res[1][...] = np.sin(alpha * y)
#     res[2][...] = np.sin(z * t)
#     return res
#
# velocity = Field(dom3D, isVector=True, formula=func_2_3D)
# 
# velocity.discretize(topo)
# velocity.setTopoInit(topo)
# # We set the values of extra args :
# alpha = ...
# t = ...
# velocity.setExtraParameters(alpha, t)
# velocity.initialize()
# \endcode
#
# Remarks :
# - extra param in setExtraParameters must be in the same order
#   as in the func_1_3D arg list.
# - setExtraParameters must be called for any new value of alpha/t
# - if the field has to be initialized at each step, or at least several times,
#  you'd rather use parmepy.operator.analytic.Analytic.
# - if you only need to compute the value of the field at a specific point,
#   you can just call:
#\code
# res = velocity.value(res, 1.,2.,0., alpha, t)
#\endcode
#
# - it's possible to reset a formula with :
# \code
# velocity.setFormula(newFormula)
#\endcode
#
# \subsection fieldsTutoDataAccess Fields' data access
#
# Some examples on how to get access or set values of the fields:
#
# \code
# # Get a discretisation for a given continuous field:
# v3 = velocity.discreteFields[topo1]
# # Note that in this case v3 == v1
#
# # Get all the discretisations :
# for topo in velocity.discreteFields.keys()
#     print velocity.discreteFields[topo]
#
# # Get second component value of v2 at point of index 8,8,8
# # in the local mesh of topo2:
# print v2[1][8,8,8]
# # use vorticity formula to initialize v1 values :
# v1.initialize(formula=compute_ref)
# # norm 2 of v1
# nx, ny, nz = v1.norm()
# \endcode
#
#
# \b Remark : usually, in a 'complete' problem, the user does not have
# to bother with the topologies' stuff and the fields discretisations.
# All this is hidden from user : one just has to define
# some operators, bricks of the global problem, associated with
# some required global resolutions and some solving methods.
# Then the setup process of the problem/operators will automatically
# performs the topologies creation and the discretisation of the fields.
# See operators/problem documentation for more details or
# the examples in Parmes/Examples.
#
#
