/**
 * @file advection.cl
 * Advection kernel, vectorized version.
 */

/**
 * Computes particles positions from the velocity field.
 * A work-group is handling a 1D problem. Thus, gidY and gidZ are constants among work-items of a work-group.
 * Each work-item computes <code>NB_I/WI_NB</code> particles positions.
 * Particle are computed through OpenCL vector types of lenght 2, 4 or 8.
 * Velocity data are copied to a local buffer as a cache.
 *
 * @param gvelo Velocity.
 * @param ppos Particle position.
 * @param dt Time step.
 * @param min_position Domain lower coordinate.
 * @param dx Space step.
 *
 * @remark <code>NB_I</code>, <code>NB_II</code>, <code>NB_III</code> : points number in directions from 1st varying index to last.
 * @remark <code>WI_NB</code> corresponds to the work-item number.
 * @remark <code>__N__</code> is expanded at compilation time by vector width.
 * @remark <code>__NN__</code> is expanded at compilation time by a sequence of integer for each vector component.
 * @see parmepy.gpu.tools.parse_file
 */
__kernel void advection_kernel(__global const float* gvelo,
			       __global float* ppos,
			       __local float* velocity_cache,
			       float dt, float min_position, float4 dx, float4 v_dx)
{
  uint gidX = get_global_id(0);	/* OpenCL work-itme global index (X) */
  uint gidY = get_global_id(1); /* OpenCL work-itme global index (Y) */
  uint gidZ = get_global_id(2); /* OpenCL work-itme global index (Z) */
  float invdx = 1.0/dx.x;		/* Space step inverse */
  float v_invdx = 1.0/v_dx.x;		/* Space step inverse */
  uint i;			/* Particle index in 1D problem */
  float__N__ p;				/* Particle position */
  uint line_index = gidY*NB_I+gidZ*NB_I*NB_II; /* Current 1D problem index */

  // Allocated on host : __local float velocity_cache[NB_I]; /* Velocity cache */
  fill_velocity_cache(gvelo, gidX, gidY, gidZ, dx, v_dx, velocity_cache);

  /* Synchronize work-group */
  barrier(CLK_LOCAL_MEM_FENCE);

  for(i=gidX*__N__; i<NB_I; i+=WI_NB*__N__)
    {
      /* Compute position */
      p = advection(i, dt, dx.x, invdx, v_invdx, velocity_cache) + (float__N__)(min_position);
      /* Store result */
      vstore__N__(p, (i+line_index)/__N__, ppos);
    }
}
