/**
 * @file advection_noVec.cl
 * Advection kernel, basic version.
 */

/**
 * Computes particles positions from the velocity field.
 * A work-group is handling a 1D problem. Thus, gidY and gidZ are constants among work-items of a work-group.
 * Each work-item computes NB_I/WI_NB particles positions.
 *
 * @param gvelo Velocity.
 * @param ppos Particle position.
 * @param dt Time step.
 * @param min_position Domain lower coordinate.
 * @param dx Space step.
 *
 * @remark NB_I, NB_II, NB_III : points number in directions from 1st varying index to last.
 * @remark WI_NB corresponds to the work-item number.
 */
__kernel void advection_kernel(__global const float* gvelo,
			       __global float* ppos,
			       __local float* velocity_cache,
			       float dt, float min_position, float4 dx, float4 v_dx)
{
  uint gidX = get_global_id(0);	/* OpenCL work-itme global index (X) */
  uint gidY = get_global_id(1); /* OpenCL work-itme global index (Y) */
  uint gidZ = get_global_id(2); /* OpenCL work-itme global index (Z) */
  float invdx = 1.0/dx.x;		/* Space step inverse */
  float v_invdx = 1.0/v_dx.x;		/* Space step inverse */
  uint i;			/* Particle index in 1D problem */
  uint line_index = gidY*NB_I+gidZ*NB_I*NB_II; /* Current 1D problem index */

  // Allocated on host : __local float velocity_cache[NB_I]; /* Velocity cache */

  fill_velocity_cache(gvelo, gidX, gidY, gidZ, dx, v_dx, velocity_cache);

  /* Synchronize work-group */
  barrier(CLK_LOCAL_MEM_FENCE);

  for(i=gidX; i<NB_I; i+=WI_NB)
    {
      ppos[i+line_index] = advection(i, dt, dx.x, invdx, v_invdx, velocity_cache) + min_position;
    }
}
