/**
 * @file remeshing_vector_2d.cl
 * Remeshing kernel.
 */

/**
 * Performs remeshing of the particles' vector in 2d.
 * A work-group is handling a 1D problem. Thus, gidY and gidZ are constants among work-items of a work-group.
 * Each work-item computes <code>NB_I/WI_NB</code> particles positions. To avoid concurrent witings, in case of strong velocity gradients, work-items computes contiguous particles.
 * Particle are computed through OpenCL vector types of lenght 2, 4 or 8.
 * Scalar results are stored in a local buffer as a cache and then copied to global memory buffer.
 *
 * @param ppos Particle position
 * @param pscal Particle scalar
 * @param gscal Grid scalar
 * @param min_position Domain lower coordinate
 * @param dx Space step
 *
 * @remark <code>NB_I</code>, <code>NB_II</code>, <code>NB_III</code> : points number in directions from 1st varying index to last.
 * @remark <code>WI_NB</code> corresponds to the work-item number.
 * @remark <code>__N__</code> is expanded at compilation time by vector width.
 * @remark <code>__NN__</code> is expanded at compilation time by a sequence of integer for each vector component.
 * @see parmepy.gpu.tools.parse_file
 */
__kernel void remeshing_kernel(__global const float* ppos,
			       __global const float* pvec_X,
			       __global const float* pvec_Y,
			       __global float* gvec_X,
			       __global float* gvec_Y,
			       float min_position, float dx)
{
  uint gidX = get_global_id(0);	/* OpenCL work-itme global index (X) */
  uint gidY = get_global_id(1); /* OpenCL work-itme global index (Y) */
  uint gidZ = get_global_id(2); /* OpenCL work-itme global index (Z) */
  float invdx = 1.0/dx;         /* Space step inverse */
  uint i;			/* Particle index in 1D problem */
  float__N__ p,			/* Particle position */
    v_X, v_Y;			/* Particle quantity */
  uint line_index = gidY*NB_I+ gidZ*NB_I*NB_II; /* Current 1D problem index */

  __local float gvec_X_loc[NB_I]; /* Local buffer for result */
  __local float gvec_Y_loc[NB_I]; /* Local buffer for result */

  for(i=gidX*__N__; i<NB_I; i+=(WI_NB*__N__))
    {
      /* Initialize result buffer */
      gvec_X_loc[i+__NN__] = 0.0;
      gvec_Y_loc[i+__NN__] = 0.0;
    }

  /* Synchronize work-group */
  barrier(CLK_LOCAL_MEM_FENCE);

  for(i=gidX*PART_NB_PER_WI; i<(gidX + 1)*PART_NB_PER_WI; i+=__N__)
    {
      /* Read particle position */
      p = vload__N__((i + line_index)/__N__, ppos) - (float__N__)(min_position);
      /* Read particle scalar */
      v_X = vload__N__((i + line_index)/__N__, pvec_X);
      v_Y = vload__N__((i + line_index)/__N__, pvec_Y);
      /* Remesh particle */
      remesh(i, dx, invdx, v_X, v_Y, p, gvec_X_loc, gvec_Y_loc);
    }

  /* Synchronize work-group */
  barrier(CLK_LOCAL_MEM_FENCE);

  for(i=gidX*__N__; i<NB_I; i+=(WI_NB*__N__))
    {
      /* Store result */
      vstore__N__((float__N__)(gvec_X_loc[noBC_id(i+__NN__)],
			       ),(i + line_index)/__N__, gvec_X);
      vstore__N__((float__N__)(gvec_Y_loc[noBC_id(i+__NN__)],
			       ),(i + line_index)/__N__, gvec_Y);
  }
}
