/**
 * @file private_vector_2d.cl
 * Remeshing function, vectorized, private variable for 2D vector remeshing.
 */

void remesh(uint i, float dx, float invdx,
	    float__N__ v_X, float__N__ v_Y,
	    float__N__ p,
	    __local float* gvec_X_loc, __local float* gvec_Y_loc);


/**
 * Remesh particles in local buffer.
 *
 * Remeshing formula is given a compiling time.
 * Use of builtin OpenCL functions fma and mix. Computations through OpenCL vector types.
 * Use of a private temporary variable for remeshing weights.
 *
 * @param i Particle index
 * @param dx Space step
 * @param invdx 1/dx
 * @param s Particle scalar
 * @param p Particle position
 * @param gscal_loc Local buffer for result
 *
 * @remark <code>NB_I</code>, <code>NB_II</code>, <code>NB_III</code> : points number in directions from 1st varying index to last.
 * @remark <code>__N__</code> is expanded at compilation time by vector width.
 * @remark <code>__NN__</code> is expanded at compilation time by a sequence of integer for each vector component.
 * @remark <code>FORMULA</code> : remeshing formula flag {<code>M4PRIME</code>, <code>M6PRIME</code>, <code>M8PRIME</code>, <code>L6STAR</code>}
 * @remark <code>REMESH</code> is a function-like macro expanding to the proper remeshing formula (i.e.: <code>REMESH(alpha)</code> -> <code>alpha_l2_1</code>)
 * @see parmepy.gpu.tools.parse_file
 * @see parmepy.gpu.cl_src.common
 */
void remesh(uint i, float dx, float invdx,
	    float__N__ v_X, float__N__ v_Y,
	    float__N__ p,
	    __local float* gvec_X_loc, __local float* gvec_Y_loc){
  float__N__ y,			/* Normalized distance to nearest left grid point */
    w;			        /* Temporary remeshing weights */
  int__N__ ind;			/* Integer coordinate */
  uint__N__ index;		/* Remeshing index */

  ind = convert_int__N___rtn(p * invdx);
  y = (p - convert_float__N__(ind) * dx) * invdx;

  index = convert_uint__N__((ind - REMESH_SHIFT + NB_I) % NB_I);

  w = REMESH(alpha)(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  w = REMESH(beta)(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  w = REMESH(gamma)(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  w = REMESH(delta)(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

#if REMESH_SHIFT > 1
  index = (index + 1) % NB_I;
  w = REMESH(eta)(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  w = REMESH(zeta)(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);
#endif

#if REMESH_SHIFT > 2
  index = (index + 1) % NB_I;
  w = REMESH(theta)(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  w = REMESH(iota)(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);
#endif

#if REMESH_SHIFT > 3
  index = (index + 1) % NB_I;
  w = REMESH(kappa)(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  w = REMESH(mu)(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);
#endif
}
