"""
@file config_default.py

OpenCL kernels default configurations.
"""
from parmepy.constants import np
FLOAT_GPU, DOUBLE_GPU = np.float32, np.float64

#build empty dictionaries
kernels_config = {}
kernels_config[2] = {FLOAT_GPU: {}, DOUBLE_GPU: {}}
kernels_config[3] = {FLOAT_GPU: {}, DOUBLE_GPU: {}}

# Copy kernel:
def copy_space_index_2d(size, t_dim, b_rows, vec):
    gwi = (int(size[0] / vec), int(b_rows * size[1] / t_dim), 1)
    lwi = (t_dim / vec, b_rows, 1)
    return gwi, lwi
def copy_space_index_3d(size, t_dim, b_rows, vec):
    gwi = (int(size[0] / vec), int(b_rows * size[1] / t_dim), int(size[2]))
    lwi = (t_dim / vec, b_rows, 1)
    return gwi, lwi
# Configs : sources, tile size, block rows, vector size, index space function
kernels_config[3][FLOAT_GPU]['copy'] = \
    ('kernels/copy_noVec.cl', 32, 2, 1, copy_space_index_3d)
kernels_config[3][DOUBLE_GPU]['copy'] = \
    ('kernels/copy_noVec.cl', 32, 2, 1, copy_space_index_3d)
kernels_config[2][FLOAT_GPU]['copy'] = \
    ('kernels/copy_noVec.cl', 32, 2, 1, copy_space_index_2d)
kernels_config[2][DOUBLE_GPU]['copy'] = \
    ('kernels/copy_noVec.cl', 32, 2, 1, copy_space_index_2d)

# Transpositions kernels:
# XY transposition
# Settings are taken from destination layout as current layout.
# gwi is computed form input layout (appears as transposed layout)
def xy_space_index_2d(size, t_dim, b_rows, vec):
    gwi = (int(size[1] / vec), int(b_rows * size[0] / t_dim), 1)
    lwi = (t_dim / vec, b_rows, 1)
    return gwi, lwi
def xy_space_index_3d(size, t_dim, b_rows, vec):
    gwi = (int(size[1] / vec), int(b_rows * size[0] / t_dim), int(size[2]))
    lwi = (t_dim / vec, b_rows, 1)
    return gwi, lwi
# Configs : sources, tile size, block rows, is padding, vector size,
#              index space function
kernels_config[3][FLOAT_GPU]['transpose_xy'] = \
    ('kernels/transpose_xy_noVec.cl', 32, 2, True, 1, xy_space_index_3d)
kernels_config[3][DOUBLE_GPU]['transpose_xy'] = \
    ('kernels/transpose_xy_noVec.cl', 32, 2, True, 1, xy_space_index_3d)
kernels_config[2][FLOAT_GPU]['transpose_xy'] = \
    ('kernels/transpose_xy_noVec.cl', 32, 2, True, 1, xy_space_index_2d)
kernels_config[2][DOUBLE_GPU]['transpose_xy'] = \
    ('kernels/transpose_xy_noVec.cl', 32, 2, True, 1, xy_space_index_2d)

# XZ transposition
# Settings are taken from destination layout as current layout.
# gwi is computed form input layout (appears as transposed layout)
def xz_space_index_3d(size, t_dim, b_rows, b_deph, vec):
    gwi = (int(size[2] / vec), int(size[1]), int(b_deph * size[0] / t_dim))
    lwi = (t_dim / vec, 1, b_deph)
    return gwi, lwi
# Configs : sources, tile size, block rows, is padding, vector size,
#              index space function
kernels_config[3][FLOAT_GPU]['transpose_xz'] = \
    ('kernels/transpose_xz_slice_noVec.cl', 32, 1, 2, True, 1, xz_space_index_3d)
kernels_config[3][DOUBLE_GPU]['transpose_xz'] = \
    ('kernels/transpose_xz_slice_noVec.cl', 32, 1, 2, True, 1, xz_space_index_3d)

def computational_kernels_index_space(wi, size, vec):
    # Change work-item regarding vector_width
    if wi * vec > size[0]:
        if size[0] % vec > 0:
            raise ValueError(
                "Resolution ({0}) must be a multiple of {1}".format(
                    size[0], vec))
        wi = size[0] // vec

    if len(size) == 3:
        gwi = (int(wi), int(size[1]), int(size[2]))
        lwi = (int(wi), 1, 1)
    else:
        gwi = (int(wi), int(size[1]))
        lwi = (int(wi), 1)
    return gwi, lwi

def advection_index_space_3d(size, vec):
    wi = min(max(32, size[0] / 2), 128)
    return computational_kernels_index_space(wi, size, vec)
def advection_index_space_2d_SP(size, vec):
    wi = min(max(32, size[0] / 2), 128)
    return computational_kernels_index_space(wi, size, vec)
def advection_index_space_2d_DP(size, vec):
    wi = min(max(32, size[0] / 2), 128)
    return computational_kernels_index_space(wi, size, vec)

def remeshing_index_space_3d(size, vec):
    wi = min(max(32, size[0] / 2), 128)
    return computational_kernels_index_space(wi, size, vec)
def remeshing_index_space_2d(size, vec):
    wi = min(max(32, size[0] / 2), 128)
    return computational_kernels_index_space(wi, size, vec)

def advection_and_remeshing_index_space(size, vec):
    wi = min(max(32, size[0] / 2), 128)
    return computational_kernels_index_space(wi, size, vec)


# Advection kernel
# Configs sources, is noBC, vector size, index space function
kernels_config[3][FLOAT_GPU]['advec'] = \
    (["common.cl", "advection/velocity_cache_noVec.cl",
      "advection/builtin_RKN_noVec.cl", "kernels/advection_noVec.cl"],
     False, 1, advection_index_space_3d)
kernels_config[3][DOUBLE_GPU]['advec'] = \
    (["common.cl", "advection/velocity_cache_noVec.cl",
      "advection/builtin_RKN_noVec.cl", "kernels/advection_noVec.cl"],
     False, 1, advection_index_space_3d)
kernels_config[2][FLOAT_GPU]['advec'] = \
    (["common.cl", "advection/velocity_cache_noVec.cl",
      "advection/builtin_RKN_noVec.cl", "kernels/advection_noVec.cl"],
     False, 1, advection_index_space_2d_SP)
kernels_config[2][DOUBLE_GPU]['advec'] = \
    (["common.cl", "advection/velocity_cache_noVec.cl",
      "advection/builtin_RKN_noVec.cl", "kernels/advection_noVec.cl"],
     False, 1, advection_index_space_2d_DP)

# Remeshing kernel
# Configs sources, is noBC, vector size, index space function
kernels_config[3][FLOAT_GPU]['remesh'] = \
    (["common.cl", "remeshing/weights_noVec_builtin.cl",
      "remeshing/basic_noVec.cl", "kernels/remeshing_noVec.cl"],
     False, 1, remeshing_index_space_3d)
kernels_config[3][DOUBLE_GPU]['remesh'] = \
    (["common.cl", "remeshing/weights_noVec_builtin.cl",
      "remeshing/basic_noVec.cl", "kernels/remeshing_noVec.cl"],
     False, 1, remeshing_index_space_3d)
kernels_config[2][FLOAT_GPU]['remesh'] = \
    (["common.cl", "remeshing/weights_noVec_builtin.cl",
      "remeshing/basic_noVec.cl", "kernels/remeshing_noVec.cl"],
     False, 1, remeshing_index_space_2d)
kernels_config[2][DOUBLE_GPU]['remesh'] = \
    (["common.cl", "remeshing/weights_noVec_builtin.cl",
      "remeshing/basic_noVec.cl", "kernels/remeshing_noVec.cl"],
     False, 1, remeshing_index_space_2d)


# Advection and remeshing kernel
# Configs sources, is noBC, vector size, index space function
kernels_config[3][FLOAT_GPU]['advec_and_remesh'] = \
    (["common.cl",
      "remeshing/weights_noVec_builtin.cl", "remeshing/private_noVec.cl",
      "advection/velocity_cache_noVec.cl", "advection/builtin_RKN_noVec.cl",
      "kernels/advection_and_remeshing_noVec.cl"],
     False, 1, advection_and_remeshing_index_space)
kernels_config[3][DOUBLE_GPU]['advec_and_remesh'] = \
    (["common.cl",
      "remeshing/weights_noVec_builtin.cl", "remeshing/private_noVec.cl",
      "advection/velocity_cache_noVec.cl", "advection/builtin_RKN_noVec.cl",
      "kernels/advection_and_remeshing_noVec.cl"],
     False, 1, advection_and_remeshing_index_space)
kernels_config[2][FLOAT_GPU]['advec_and_remesh'] = \
    (["common.cl",
      "remeshing/weights_noVec_builtin.cl", "remeshing/private_noVec.cl",
      "advection/velocity_cache_noVec.cl", "advection/builtin_RKN_noVec.cl",
      "kernels/advection_and_remeshing_noVec.cl"],
     False, 1, advection_and_remeshing_index_space)
kernels_config[2][DOUBLE_GPU]['advec_and_remesh'] = \
    (["common.cl",
      "remeshing/weights_noVec_builtin.cl", "remeshing/private_noVec.cl",
      "advection/velocity_cache_noVec.cl", "advection/builtin_RKN_noVec.cl",
      "kernels/advection_and_remeshing_noVec.cl"],
     False, 1, advection_and_remeshing_index_space)

