"""
@file gpu_discrete.py

Contains class for discrete fields on GPU.
"""
from parmepy import __VERBOSE__
from parmepy.constants import ORDER, np, debug, PARMES_REAL, PARMES_INTEGER, S_DIR
from parmepy.fields.discrete import DiscreteField
from parmepy.gpu import cl, clArray, CL_PROFILE
from parmepy.gpu.gpu_kernel import KernelLauncher, KernelListLauncher
from parmepy.tools.numpywrappers import zeros
from parmepy.mpi.main_var import main_rank, main_size
from parmepy.tools.timers import ManualFunctionTimer

fromLayoutMgrFunc_3D_seq = [
    lambda a, shape: a.reshape(shape, order=ORDER)[...],
    lambda a, shape: a.reshape(shape, order=ORDER).swapaxes(0, 1)[...],
    lambda a, shape: a.reshape(
        shape, order=ORDER).swapaxes(0, 2).swapaxes(0, 1)[...]
    ]
shapeFunc_3D_seq = [
    lambda shape: (shape[0], shape[1], shape[2]),
    lambda shape: (shape[1], shape[0], shape[2]),
    lambda shape: (shape[2], shape[0], shape[1]),
    ]
toLayoutMgrFunc_3D_seq = [
    lambda a: a.ravel(order=ORDER)[...],
    lambda a: a.swapaxes(0, 1).ravel(order=ORDER)[...],
    lambda a: a.swapaxes(0, 1).swapaxes(0, 2).ravel(order=ORDER)[...]
    ]
fromLayoutMgrFunc_3D = [
    lambda a, shape: a.reshape(shape, order=ORDER)[...],
    lambda a, shape: a.reshape(shape, order=ORDER).swapaxes(0, 1)[...],
    lambda a, shape: a.reshape(shape, order=ORDER).swapaxes(0, 2)[...]
    ]
shapeFunc_3D = [
    lambda shape: (shape[0], shape[1], shape[2]),
    lambda shape: (shape[1], shape[0], shape[2]),
    lambda shape: (shape[2], shape[1], shape[0]),
    ]
toLayoutMgrFunc_3D = [
    lambda a: a.ravel(order=ORDER)[...],
    lambda a: a.swapaxes(0, 1).ravel(order=ORDER)[...],
    lambda a: a.swapaxes(0, 2).ravel(order=ORDER)[...]
    ]
fromLayoutMgrFunc_2D = [
    lambda a, shape: a.reshape(shape, order=ORDER)[...],
    lambda a, shape: a.reshape(shape, order=ORDER).swapaxes(0, 1)[...]
    ]
shapeFunc_2D = [
    lambda shape: (shape[0], shape[1]),
    lambda shape: (shape[1], shape[0])
    ]
toLayoutMgrFunc_2D = [
    lambda a: a.ravel(order=ORDER)[...],
    lambda a: a.swapaxes(0, 1).ravel(order=ORDER)[...]
    ]


class GPUDiscreteField(DiscreteField):
    """
    GPU Discrete vector field implementation.
    Allocates OpenCL device memory for the field.
    """
    def __init__(self, cl_env, topology=None, isVector=False, name="?",
                 precision=PARMES_REAL, layout=True,
                 batch_nb=None, batch_d=0):
        """
        Constructor.
        @param queue : OpenCL queue
        @param precision : Floating point precision
        @param parent : Continuous field.
        @param topology : Topology informations
        @param name : Field name
        @param idFromParent : Index in the parent's discrete fields
        @param layout : Boolean indicating if components are arranged in memory
        @param batch_nb : array of integer to specify how to subdivise the
        whole data to work with on the device. Default : [1]*dim
        @param batch_d : direction concerned by the batch_nb array.
        Defaut : all components are considered in the same way.
        @see parmepy.fields.vector.VectorField.__init__
        """
        super(GPUDiscreteField, self).__init__(topology, isVector, name)
        ## OpenCL environment
        self.cl_env = cl_env
        ## Precision for the field
        self.precision = precision
        ## Memory used
        self.mem_size = 0
        ## Initialization OpenCL kernel as KernelLauncher
        self.init_kernel = None
        self._isReleased = False
        ## OpenCL Buffer pointer
        self.gpu_data = [None] * self.nbComponents
        ## Is the device allocations are performed
        self.gpu_allocated = False
        ## OpenCL Events list modifying this field
        self.events = []
        ## Data layout is direction dependant
        self.layout = layout
        ## Layout and shape managers
        if self.domain.dimension == 3:
            if main_size == 1:
                self._shapeFunc = shapeFunc_3D_seq
                self._fromLayoutMgrFunc = fromLayoutMgrFunc_3D_seq
                self._toLayoutMgrFunc = toLayoutMgrFunc_3D_seq
            else:
                self._shapeFunc = shapeFunc_3D
                self._fromLayoutMgrFunc = fromLayoutMgrFunc_3D
                self._toLayoutMgrFunc = toLayoutMgrFunc_3D
        else:
            self._shapeFunc = shapeFunc_2D
            self._fromLayoutMgrFunc = fromLayoutMgrFunc_2D
            self._toLayoutMgrFunc = toLayoutMgrFunc_2D

        ## Flag telling if the data are subdivised or not
        self.isBatch = np.prod(batch_nb) > 1
        ## Batch number that is in the device memory
        self.batch_on_device = [-1] * self.nbComponents
        ## Is the data are on device
        self.batch_is_on_device = [False] * self.nbComponents
        ## Shape of one batch
        self.shape_gpu = [None] * self.domain.dimension
        ## Batch number
        self.batch_nb = [None] * self.domain.dimension
        ## Shift in index going from one batch to an other
        self.batch_shift = [None] * self.domain.dimension
        ## List of slices to access all batches
        self.batch_slices = [None] * self.domain.dimension
        ## Default shape (no batch)
        self.shape_gpu_default = None
        ## Default slice (no batch)
        self.batch_slices_default = None
        if self.isBatch:
            self.shape_gpu_default = np.asarray(
                self.data[0].shape, dtype=PARMES_INTEGER) / batch_nb
            self.batch_slices_default = tuple(
                [slice(0, nb) for nb in self.shape_gpu_default])
            self.batch_setup(batch_nb, batch_d)
            self.allocate = self._allocate_withBatch
        else:
            self.allocate = self._allocate_noBatch
        self.toHost_timer = ManualFunctionTimer(
            name="OpenCL_Transfer_toHost")
        self.toDevice_timer = ManualFunctionTimer(
            name="OpenCL_Transfer_toDevice")

    def batch_setup(self, batch_nb=None, batch_d=0):
        """Batch configuration.
        @param batch_nb : array of integer to specify how to subdivise the
        whole data to work with on the device. Default : [1]*dim
        @param batch_d : direction concerned by the batch_nb array.
        Defaut : all components are considered in the same way.
        """
        # Batch configuration
        if batch_nb is None:
            batch_nb = np.ones((self.domain.dimension, ),
                               dtype=PARMES_INTEGER)
        self.batch_nb[batch_d] = batch_nb
        shape_host = np.asarray(self.data[0].shape, dtype=PARMES_INTEGER)
        shape_host -= 2 * self.topology.ghosts
        self.shape_gpu[batch_d] = shape_host / batch_nb
        self.batch_shift[batch_d] = np.zeros((self.domain.dimension, ),
                                             dtype=PARMES_INTEGER)
        self.batch_shift[batch_d][batch_nb > 1] = \
            self.shape_gpu[batch_d][batch_nb > 1]
        self.batch_slices[batch_d] = []
        gh_batch_d = self.topology.ghosts[batch_d]
        for i in xrange(np.prod(batch_nb)):
            s = [slice(i*bs, nb+i*bs+2*gh_batch_d) for nb, bs
                 in zip(self.shape_gpu[batch_d], self.batch_shift[batch_d])]
            self.batch_slices[batch_d].append(tuple(s))
        self.shape_gpu[batch_d] += 2 * self.topology.ghosts

    def _allocate_noBatch(self):
        """Device memory allocations no batch."""
        if not self.gpu_allocated:
            self._tmp = zeros(np.prod(self.data[0].shape),
                              dtype=self.precision)
            for d in xrange(self.nbComponents):
                self.data[d] = np.asarray(self.data[d],
                                          dtype=self.precision, order=ORDER)
                self.gpu_data[d] = clArray.empty(
                    self.cl_env.queue, self.data[d].shape, self.precision,
                    order=ORDER, allocator=self.cl_env.memPool)
                self.mem_size += self.gpu_data[d].nbytes
                self.cl_env.available_mem -= self.gpu_data[d].nbytes
            self.gpu_allocated = True
            if __VERBOSE__:
                print self.name, self.mem_size, "Bytes (",
                print self.mem_size / (1024 ** 2), "MB)"

    def _allocate_withBatch(self):
        """Device memory allocations with batch."""
        if not self.gpu_allocated:
            max_gpu_shape = self.shape_gpu[0]
            for s in self.shape_gpu[1:]:
                if np.prod(max_gpu_shape) < np.prod(s):
                    max_gpu_shape = s
            self._tmp = zeros(np.prod(max_gpu_shape),
                              dtype=self.precision)
            for d in xrange(self.nbComponents):
                self.data[d] = np.asarray(self.data[d],
                                          dtype=self.precision, order=ORDER)
                # If layout, each component of this vector are used only in
                # directions of component. So it may have different sizes along
                # components. Otherwise, each components (if any) are used
                # in all directions, thus allocating a buffer of max size.
                if self.layout:
                    shape = self.shape_gpu[d]
                else:
                    shape = max_gpu_shape
                if shape is not None:
                    self.gpu_data[d] = clArray.empty(
                        self.cl_env.queue, tuple(shape), self.precision,
                        order=ORDER, allocator=self.cl_env.memPool)
                    self.mem_size += self.gpu_data[d].nbytes
                    self.cl_env.available_mem -= self.gpu_data[d].nbytes
            self.gpu_allocated = True
            if __VERBOSE__:
                print self.name, self.mem_size, "Bytes (",
                print self.mem_size / (1024 ** 2), "MB)"

    @classmethod
    def fromField(cls, cl_env, vfield, precision=PARMES_REAL,
                  layout=True, batch_nb=None, batch_d=0):
        """
        Contructor from a discrete vector field.
        Mutates the given VectorField to a GPUVectorField.
        @param cls : Class of the class method (GPUVectorField)
        @param queue : OpenCL queue
        @param vfield : VectorField
        @param precision : Floating point precision
        @param layout : Boolean indicating if components are arranged in memory
        @param batch_nb : array of integer to specify how to subdivise the
        @param batch_d : direction concerned by the batch_nb array.
        """
        if not isinstance(vfield, GPUDiscreteField):
            vfield.__class__ = cls
            GPUDiscreteField.__init__(
                vfield, cl_env,
                vfield.topology, vfield.isVector, vfield.name,
                precision, layout,
                batch_nb=batch_nb, batch_d=batch_d)
        else:
            if vfield.isBatch:
                vfield.batch_setup(batch_nb, batch_d)

    def setInitializationKernel(self, kernel):
        """
        Set the initialization kernel
        @param kernel : KernelLauncher to use for initialize field.
        """
        self.init_kernel = kernel

    @debug
    def dump(self, filename, mode=None):
        """
        @remark Synchronized OpenCL calls (waiting for event(s) completion)
        """
        self.toHost()
        self.wait()
        DiscreteField.dump(self, filename, mode)

    @debug
    def load(self, filename, fieldname=None):
        """
        @remark Synchronized OpenCL calls (waiting for event(s) completion)
        """
        DiscreteField.load(self, filename, fieldname)
        self.toDevice()

    @debug
    def initialize(self, formula=None, doVectorize=False, currentTime=0.,
                   *args):
        """
        GPU data initialization.
        Performs the initialization from different ways if device not already
        contains up-to-date data:
          - with an OpenCL kernel,
          - with a python formula (as VectorField) and the copy data to device.
        @param formula : Formula to use.
        @param args : formula extra parameters
        @remark Synchronized OpenCL calls (waiting for event(s) completion)
        """
        t = self.precision(currentTime)
        if __VERBOSE__:
            print "{"+str(main_rank)+"}", "Initialize", self.name
        isGPUKernel = isinstance(formula, KernelLauncher) \
            or isinstance(formula, KernelListLauncher)
        if not isGPUKernel and self.init_kernel is None:
            DiscreteField.initialize(self, formula, False, currentTime, *args)
            for d in xrange(self.nbComponents):
                self.data[d] = np.asarray(
                    self.data[d],
                    dtype=self.precision, order=ORDER)
            self.toDevice()
        else:
            if self.isBatch:
                assert np.prod([np.prod(b) for b in self.batch_nb]) == 1,\
                    "Initilization with kernel is impossible with batch. \
                Use Python formula instead."
            if isGPUKernel:
                self.init_kernel = formula
            coord_min = np.ones(4, dtype=self.precision)
            mesh_size = np.ones(4, dtype=self.precision)
            coord_min[0:self.dimension] = np.asarray(
                self.topology.mesh.origin,
                dtype=self.precision)
            mesh_size[0:self.dimension] = np.asarray(
                self.topology.mesh.space_step,
                dtype=self.precision)
            if self.nbComponents == 2:
                evt = self.init_kernel(self.gpu_data[0].data,
                                       self.gpu_data[1].data,
                                       coord_min, mesh_size, t,
                                       *args,
                                       wait_for=self.events)
            elif self.nbComponents == 3:
                evt = self.init_kernel(self.gpu_data[0].data,
                                       self.gpu_data[1].data,
                                       self.gpu_data[2].data,
                                       coord_min, mesh_size, t,
                                       *args,
                                       wait_for=self.events)
            else:
                evt = self.init_kernel(self.gpu_data[0].data,
                                       coord_min, mesh_size, t,
                                       *args,
                                       wait_for=self.events)
            self.events.append(evt)

    def finalize(self):
        if not self._isReleased:
            if __VERBOSE__:
                print "deallocate :", self.name,
                print " (" + str(self.mem_size / (1024. ** 2)) + " MBytes)"
            self.wait()
            for d in xrange(self.nbComponents):
                self.gpu_data[d].data.release()
                self.cl_env.available_mem += self.gpu_data[d].nbytes
            if self.init_kernel is not None:
                for f_timer in self.init_kernel.f_timer:
                    self.timer.addFunctionTimer(f_timer)
            self._isReleased = True

    def toDevice(self, component=None, layoutDir=None, batch_d=None):
        """
        Host to device method.
        @param component : Component to consider (Default : all components)
        @param layoutDir : layout to use
        If the field have a layout per component, layoutDir is unused. Other
        fields can be transfered with a given layout.

        Performs a direct OpenCL copy from numpy arrays
        to OpenCL Buffers.\n
        Arrange memory on device so that vector components are
        contiguous in the direction of the component, if layout flag is True.\n
        Example : A 3D vector field F(x,y,z) is made up of 3
        OpenCL Buffers Fx, Fy, Fz. The memory layout is :
        - Fx : x-major ordering. On device,
        Fx[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fx(i,j,k)
        - Fy : y-major ordering. On device,
        Fy[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fy(j,i,k)
        - Fz : z-major ordering. On device,
        Fz[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fz(k,i,j)
        """
        if component is None:
            range_components = xrange(self.nbComponents)
            evt = [None] * self.nbComponents
        else:
            range_components = [component]
            evt = [None]

        #Ghosts Synchronization before sending
        self.synchro_ghosts()

        self.wait()
        mem_transfered = 0
        for d_id, d in enumerate(range_components):
            if d_id > 0 and evt[d-1] is not None:
                evt[d-1].wait()
            if self.layout:
                layoutDir = d
            if not self.isBatch:
                if layoutDir is None:
                    layoutDir = 0
                if __VERBOSE__:
                    print "{"+str(main_rank)+"}", "host->device :", \
                        self.name, S_DIR[d], layoutDir, 'No Batch'
                self._tmp[...] = self._toLayoutMgrFunc[layoutDir](self.data[d])
                evt[d_id] = cl.enqueue_copy(self.cl_env.queue,
                                            self.gpu_data[d].data, self._tmp)
                mem_transfered += self.gpu_data[d].nbytes
                self.events.append(evt[d_id])
            else:
                if (self.batch_on_device[d] == batch_d) and \
                        self.batch_is_on_device[d]:
                    continue
                if layoutDir is None and batch_d is None:
                    if __VERBOSE__:
                        print "{"+str(main_rank)+"}", "host->device :", \
                            self.name, S_DIR[d], 0, 'Batch_default'
                    # Default transfer, using the shape of the gpu_data
                    # component and the default layout
                    self._tmp[:np.prod(self.shape_gpu_default)] = \
                        self._toLayoutMgrFunc[0](
                            self.data[d][self.batch_slices_default])
                    evt[d_id] = cl.enqueue_copy(
                        self.cl_env.queue,
                        self.gpu_data[d].data,
                        self._tmp[:np.prod(self.shape_gpu_default)])
                    self.batch_on_device[d] = 0
                    self.batch_is_on_device[d] = True
                    mem_transfered += self.gpu_data[d].nbytes
                    self.events.append(evt[d_id])
                else:
                    batch_dir = batch_d
                    if batch_d is None:
                        batch_dir = 0
                    if __VERBOSE__:
                        print "{"+str(main_rank)+"}", "host->device :", \
                            self.name, S_DIR[d], layoutDir, 'Batch', batch_dir
                    if self.shape_gpu[layoutDir] is not None:
                        self._tmp[:np.prod(self.shape_gpu[layoutDir])] = \
                            self._toLayoutMgrFunc[layoutDir](
                                self.data[d][
                                    self.batch_slices[layoutDir][batch_dir]])
                        evt[d_id] = cl.enqueue_copy(
                            self.cl_env.queue,
                            self.gpu_data[d].data,
                            self._tmp[:np.prod(self.shape_gpu[layoutDir])])
                        self.batch_on_device[d] = batch_dir
                        self.batch_is_on_device[d] = True
                        mem_transfered += self.gpu_data[d].nbytes
                        self.events.append(evt[d_id])

        time = 0.
        if CL_PROFILE:
            for e in evt:
                if e is not None:
                    e.wait()
                    time += (e.profile.end - e.profile.start) * 1e-9
            self.toDevice_timer.append_time(time)
        if __VERBOSE__:
            if CL_PROFILE:
                print self.mem_size, "Bytes transfered at ",
                print "{0:.3f} GBytes/sec".format(
                    mem_transfered / (time * 1024 ** 3))

    def toHost(self, component=None, layoutDir=None, batch_d=None):
        """
        Device to host method.
        @param component : Component to consider (Default : all components)
        @param layoutDir : layout to use
        If the field have a layout per component, layoutDir is unused. Other
        fields can be transfered with a given layout.

        Performs a direct OpenCL copy from OpenCL Buffers
        to numpy arrays.\n
        As memory layout, if set, is arranged on device, not only a
        copy is performed but also transpositions to have numpy
        arrays consistent to each other.
        """
        self.wait()
        if component is None:
            range_components = xrange(self.nbComponents)
            evt = [None] * self.nbComponents
        else:
            range_components = [component]
            evt = [None]

        mem_transfered = 0
        for d_id, d in enumerate(range_components):
            if self.layout:
                layoutDir = d
            if not self.isBatch:
                if layoutDir is None:
                    layoutDir = 0
                if __VERBOSE__:
                    print "{"+str(main_rank)+"}", "device->host :", \
                        self.name, S_DIR[d], layoutDir, 'No Batch'
                evt[d_id] = cl.enqueue_copy(self.cl_env.queue,
                                            self._tmp,
                                            self.gpu_data[d].data,
                                            wait_for=self.events)
                mem_transfered += self.gpu_data[d].nbytes
                evt[d_id].wait()
                shape = self._shapeFunc[layoutDir](self.data[d].shape)
                self.data[d][...] = self._fromLayoutMgrFunc[layoutDir](
                    self._tmp, shape)
            elif self.batch_is_on_device[d]:
                if layoutDir is None and batch_d is None:
                    if __VERBOSE__:
                        print "{"+str(main_rank)+"}", "device->host :", \
                            self.name, S_DIR[d], 0, 'Batch_default'
                    # Default transfer, using the shape of the gpu_data
                    # component and the default layout
                    evt[d_id] = cl.enqueue_copy(
                        self.cl_env.queue,
                        self._tmp[:np.prod(self.shape_gpu_default)],
                        self.gpu_data[d].data,
                        wait_for=self.events)
                    mem_transfered += self.gpu_data[d].nbytes
                    evt[d_id].wait()
                    shape = self._shapeFunc[0](self.shape_gpu_default)
                    self.data[d][self.batch_slices_default] = \
                        self._fromLayoutMgrFunc[0](
                            self._tmp[:np.prod(self.shape_gpu_default)],
                            shape)
                    self.batch_is_on_device[d] = False
                else:
                    if batch_d is None:
                        batch_d = 0
                    if __VERBOSE__:
                        print "{"+str(main_rank)+"}", "device->host :", \
                            self.name, S_DIR[d], layoutDir, 'Batch', batch_d
                    evt[d_id] = cl.enqueue_copy(
                        self.cl_env.queue,
                        self._tmp[:np.prod(self.shape_gpu[layoutDir])],
                        self.gpu_data[d].data,
                        wait_for=self.events)
                    mem_transfered += self.gpu_data[d].nbytes
                    evt[d_id].wait()
                    shape = self._shapeFunc[layoutDir](
                        self.shape_gpu[layoutDir])
                    self.data[d][self.batch_slices[layoutDir][batch_d]] = \
                        self._fromLayoutMgrFunc[layoutDir](
                            self._tmp[:np.prod(self.shape_gpu[layoutDir])],
                            shape)
                    self.batch_is_on_device[d] = False

        time = 0.
        if CL_PROFILE:
            for e in evt:
                if e is not None:
                    e.wait()
                    time += (e.profile.end - e.profile.start) * 1e-9
            self.toHost_timer.append_time(time)
        if __VERBOSE__:
            if CL_PROFILE:
                print self.mem_size, "Bytes transfered at ",
                print "{0:.3f} GBytes/sec".format(
                    mem_transfered / (time * 1024 ** 3))

    def wait(self):
        """
        Waiting for all events completion in the field list.
        Resets the events list.
        """
        if __VERBOSE__:
            print "{"+str(main_rank)+"}", "Wait events :", self.name
        for e in self.events:
            e.wait()
        self.events = []

    def clean_events(self):
        """
        Waiting for all events completion in the field list.
        Resets the events list.
        """
        if __VERBOSE__:
            print "{"+str(main_rank)+"}", "Clean events :", \
                self.name, len(self.events)
        c = cl.command_execution_status.COMPLETE
        for e in self.events:
            e.wait()
        self.events = [e for e in self.events
                       if not e.command_execution_status is c]
