"""
@file gpu_particle_advection_2k.py

Discrete advection representation.
"""
from parmepy.constants import debug, PARMES_REAL
from parmepy.gpu.gpu_particle_advection import GPUParticleAdvection
from parmepy.methods_keys import TimeIntegrator, Interpolation, Remesh, \
    Support, Splitting, MultiScale
from parmepy.numerics.integrators.runge_kutta2 import RK2
from parmepy.numerics.interpolation import Linear
from parmepy.numerics.remeshing import L2_1
from parmepy.fields.continuous import Field
from parmepy.gpu.gpu_discrete import GPUDiscreteField
from parmepy.gpu.gpu_kernel import KernelLauncher


class GPUParticleAdvection2k(GPUParticleAdvection):
    """
    Particle advection operator representation on GPU.
    """

    @debug
    def __init__(self, velocity, advectedFields, d,
                 part_position=None, part_advectedFields=None,
                 platform_id=None, device_id=None,
                 device_type=None,
                 method=None,
                 src=None, precision=PARMES_REAL, batch_nb=None,
                 isMultiScale=False):
        """
        Create a Advection operator.
        Work on a given scalar at a given velocity to produce scalar
        distribution at new positions.

        @param velocity : Velocity field
        @param advectedFields : Advected field (only single field is supported)
        @param d : Direction to advect
        @param part_position : DiscreteField to use for particle position.
        @param part_advectedFields : DiscreteField to use for particle
        advected fields.
        @param platform_id : OpenCL platform id (default = 0).
        @param device_id : OpenCL device id (default = 0).
        @param device_type : OpenCL device type (default = 'gpu').
        @param method : the method to use.
        @param src : User OpenCL sources.
        @param precision : Floating point precision required (default = Single
        Precision)
        @param batch_nb : User defined batch number for fields (computed by
        default to fit the device memory).
        @param isMultiScale : Flag to specify if the velocity and advected
        fields are on different grids.
        """
        if method is None:
            method = {TimeIntegrator: RK2,
                      Interpolation: Linear,
                      Remesh: L2_1,
                      Support: 'gpu_2k',
                      Splitting: 'o2'}
        GPUParticleAdvection.__init__(self, velocity, advectedFields, d,
                                      part_position, part_advectedFields,
                                      platform_id, device_id,
                                      device_type,
                                      method, src,
                                      precision, batch_nb,
                                      isMultiScale=isMultiScale)
        self.num_advec, self.num_remesh = None, None

    @debug
    def setUp(self):
        GPUParticleAdvection.setUp(self)

    def globalMemoryUsagePreview(self, v_shape, shape):
        r = (self.velocity.nbComponents * v_shape.prod() +
             (2 * self.advectedFields[0].nbComponents + 1) * shape.prod())
        return r * self.cl_env.prec_size

    def _buffer_allocations(self):
        """
        Allocate OpenCL buffers for velocity and advected field.
        And one more buffer for advected fields quantities on particles.
        And a buffer for storing particle positions.
        """
        ## Velocity.
        alloc = not isinstance(self.velocity, GPUDiscreteField)
        GPUDiscreteField.fromField(self.cl_env, self.velocity,
                                   self.gpu_precision,
                                   batch_nb=self.batch_nb, batch_d=self.dir)
        if alloc:
            self.size_global_alloc += self.velocity.mem_size

        ## Transported field.
        alloc = not isinstance(self.advectedFields[0], GPUDiscreteField)
        GPUDiscreteField.fromField(self.cl_env, self.advectedFields[0],
                                   self.gpu_precision,
                                   layout=False,
                                   batch_nb=self.batch_nb, batch_d=self.dir)
        if alloc:
            self.size_global_alloc += self.advectedFields[0].mem_size

        ## Particle position
        if self.part_position is None:
            self.part_position = \
                Field(self.advectedFields[0].topology.domain,
                      name="Particle_Position",
                      isVector=False
                      ).discretize(self.advectedFields[0].topology)
        alloc = not isinstance(self.part_position, GPUDiscreteField)
        GPUDiscreteField.fromField(self.cl_env, self.part_position,
                                   self.gpu_precision, layout=False,
                                   batch_nb=self.batch_nb, batch_d=self.dir)
        if alloc:
            self.size_global_alloc += self.part_position.mem_size

        ## Result scalar
        if self.part_advectedFields is None:
            self.part_advectedFields = [
                Field(self.advectedFields[0].topology.domain,
                      name="Particle_AdvectedFields",
                      isVector=self.advectedFields[0].isVector
                      ).discretize(self.advectedFields[0].topology)]
        alloc = not isinstance(self.part_advectedFields[0], GPUDiscreteField)
        GPUDiscreteField.fromField(self.cl_env, self.part_advectedFields[0],
                                   self.gpu_precision, layout=False,
                                   batch_nb=self.batch_nb, batch_d=self.dir)
        if alloc:
            self.size_global_alloc += self.part_advectedFields[0].mem_size

        is_batch = (self.velocity.isBatch and
                    self.advectedFields[0].isBatch and
                    self.part_position.isBatch and
                    self.part_advectedFields[0].isBatch)
        is_not_batch = (not self.velocity.isBatch and
                        not self.advectedFields[0].isBatch and
                        not self.part_position.isBatch and
                        not self.part_advectedFields[0].isBatch)
        identic_batch = True
        if is_batch:
            for vb, adb, pposb, padb in zip(
                    self.velocity.batch_nb[self.dir],
                    self.advectedFields[0].batch_nb[self.dir],
                    self.part_position.batch_nb[self.dir],
                    self.part_advectedFields[0].batch_nb[self.dir]):
                identic_batch = identic_batch and \
                    (vb == adb and vb == padb and vb == pposb)

        if not ((is_batch or is_not_batch) and
                (not is_batch or (is_batch and identic_batch))):
            raise RuntimeError("In operator advection on GPU, automatic " +
                               "batch number computations fails: (different " +
                               "batch number for same variables use in " +
                               "different advection operators). " +
                               "User must give an explicit and identical " +
                               "batch_nb parameter for all GPU advection " +
                               "operators")

        self.variables = [self.advectedFields[0], self.velocity,
                          self.part_position, self.part_advectedFields[0]]
        self._work = self.part_advectedFields + [self.part_position]

    def _collect_kernels_cl_src(self):
        """
        Compile OpenCL sources for advection and remeshing kernel.
        """
        build_options = self.build_options + self._size_constants
        src, is_noBC, vec, f_space = self._kernel_cfg['advec']
        gwi, lwi = f_space(self.resol_dir, vec)
        WINb = lwi[0]

        if is_noBC:
            build_options += " -D WITH_NOBC=1"
        build_options += " -D WI_NB=" + str(WINb)
        if self._isMultiScale:
            build_options += " -D MS_FORMULA="
            build_options += self.method[MultiScale].__name__.upper()
        build_options += self._constants[self.dir]
        ## Build code
        src = [s.replace('RKN', self.method[TimeIntegrator].__name__.lower())
               for s in src]
        prg = self.cl_env.build_src(
            src,
            build_options,
            vec,
            nb_remesh_components=self.part_advectedFields[0].nbComponents)
        self.num_advec = KernelLauncher(
            prg.advection_kernel, self.cl_env.queue, gwi, lwi)

        ## remeshing
        build_options = self.build_options + self._size_constants
        src, is_noBC, vec, f_space = self._kernel_cfg['remesh']
        gwi, lwi = f_space(self.resol_dir, vec)
        WINb = lwi[0]

        build_options += " -D FORMULA=" + self.method[Remesh].__name__.upper()
        if is_noBC:
            build_options += " -D WITH_NOBC=1"
        build_options += " -D WI_NB=" + str(WINb)
        build_options += " -D PART_NB_PER_WI="
        build_options += str(self.resol_dir[0] / WINb)
        build_options += self._constants[self.dir]
        ## Build code
        prg = self.cl_env.build_src(
            src, build_options, vec,
            nb_remesh_components=self.part_advectedFields[0].nbComponents)
        self.num_remesh = KernelLauncher(
            prg.remeshing_kernel, self.cl_env.queue, gwi, lwi)

        ## Local memory array for kernels
        advec_nbC = self.part_advectedFields[0].nbComponents
        loc_arrays = [self.v_resol_dir[0]]
        loc_arrays += [self.resol_dir[0]] * (advec_nbC)
        self._num_locMem, self.size_local_alloc = \
            self.cl_env.LocalMemAllocator(loc_arrays)

    def _compute_advec(self, simulation, dtCoeff, split_id, old_dir):
        dt = simulation.timeStep * dtCoeff
        wait_evts = self.velocity.events+self.part_position.events
        # Advection
        evt = self.num_advec(
            self.velocity.gpu_data[self.dir].data,
            self.part_position.gpu_data[0].data,
            self._num_locMem[0],
            self.gpu_precision(dt),
            self.coord_min[self.dir],
            self.mesh_size, self.v_mesh_size,
            wait_for=wait_evts)
        self.part_position.events.append(evt)

    def _compute_1c(self, simulation, dtCoeff, split_id, old_dir):
        self._compute_advec(simulation, dtCoeff, split_id, old_dir)
        wait_evts = self.part_position.events+self.advectedFields[0].events
        evt = self.num_remesh(
            self.part_position.gpu_data[0].data,
            self.part_advectedFields[0].gpu_data[0].data,
            self.advectedFields[0].gpu_data[0].data,
            self._num_locMem[1],
            self.coord_min[self.dir],
            self.mesh_size[self.dir],
            wait_for=wait_evts)
        self.advectedFields[0].events.append(evt)

    def _compute_2c(self, simulation, dtCoeff, split_id, old_dir):
        self._compute_advec(simulation, dtCoeff, split_id, old_dir)
        wait_evts = self.part_position.events+self.advectedFields[0].events
        evt = self.num_remesh(
            self.part_position.gpu_data[0].data,
            self.part_advectedFields[0].gpu_data[0].data,
            self.part_advectedFields[0].gpu_data[1].data,
            self.advectedFields[0].gpu_data[0].data,
            self.advectedFields[0].gpu_data[1].data,
            self._num_locMem[1],
            self._num_locMem[2],
            self.coord_min[self.dir],
            self.mesh_size[self.dir],
            wait_for=wait_evts)
        self.advectedFields[0].events.append(evt)

    def _compute_3c(self, simulation, dtCoeff, split_id, old_dir):
        self._compute_advec(simulation, dtCoeff, split_id, old_dir)
        wait_evts = self.part_position.events+self.advectedFields[0].events
        evt = self.num_remesh(
            self.part_position.gpu_data[0].data,
            self.part_advectedFields[0].gpu_data[0].data,
            self.part_advectedFields[0].gpu_data[1].data,
            self.part_advectedFields[0].gpu_data[2].data,
            self.advectedFields[0].gpu_data[0].data,
            self.advectedFields[0].gpu_data[1].data,
            self.advectedFields[0].gpu_data[2].data,
            self._num_locMem[1],
            self._num_locMem[2],
            self._num_locMem[3],
            self.coord_min[self.dir],
            self.mesh_size[self.dir],
            wait_for=wait_evts)
        self.advectedFields[0].events.append(evt)

    def finalize(self):
        if self.num_advec.f_timer is not None:
            for f_timer in self.num_advec.f_timer:
                self.kernels_timer.addFunctionTimer(f_timer)
            for f_timer in self.num_remesh.f_timer:
                self.kernels_timer.addFunctionTimer(f_timer)
        GPUParticleAdvection.finalize(self)
