"""
@file mesh.py
Cartesian mesh class for local mesh.

"""
from parmepy.constants import np, PARMES_INTEGER, debug
import parmepy.tools.numpywrappers as npw


class SubMesh(object):
    """
    Description of a local mesh, on one process of
    the topology.
    """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, topo, g_start, resolution):
        """
        Example of indices for a 8 point domain with 2 ghosts and 2 procs.
        \code
        X : ghost point
        o : compute point

           |o o o o o o o o|      domain
            0             7       global indices


        X X o o o o X X           -> Proc no1
                X X o o o o X X   -> Proc no2

        Proc no1:
        resolution : (8,)
        global start/end : 0/3
        local start/end : 2/5

        Proc no2:
        resolution: (8,)
        global start/end : 4/7
        local start/end : 2/5
        \endcode
        """
        ## Topology that creates (and owns) this mesh
        self._topology = topo
        ## Local resolution of this mesh, INCLUDING ghost points
        self.resolution = np.asarray(resolution, dtype=PARMES_INTEGER)
        ## Dimension of the mesh
        self.dim = self.resolution.size
        ## index of the lowest "computed" point of this mesh
        ## (in each dir) in the global mesh
        self.global_start = np.asarray(g_start, dtype=PARMES_INTEGER)
        ## index of the upper point (in each dir), global mesh
        self.global_end = self.global_start + self.resolution - 1\
            - 2 * topo.ghosts
        ## Mesh step size
        self.space_step = npw.asarray(topo.domain.length /
                                      (topo.globalMeshResolution - 1))
        ## Mesh local indices, only for "computed" points
        ## (i.e. excluding ghosts)
        self.local_start = topo.ghosts.copy()
        self.local_end = self.resolution - topo.ghosts - 1
        ## List of indices for computational points (i.e. without ghosts)
        ## usage : field[iCompute] returns the array field for all indices
        ## excluding ghost points.
        self.iCompute = []
        ## List of indices to be used for integration over the domain.
        for d in range(self.dim):
            self.iCompute.append(slice(self.local_start[d],
                                       self.local_end[d] + 1))

        ## Coordinates of the "lowest" point of this mesh (including ghost)
        ## Warning FP : this strongly depends on where is the origin
        ## of the domain, of the type of boundary conditions
        ## and if origin is on ghosts or "real" points.
        self.origin = topo.domain.origin.copy()
        self.origin[:] += self.space_step[:] * \
            (self.global_start[:] - topo.ghosts[:])

        self.end = self.origin + self.space_step * (self.resolution - 1)
        if self.dim == 1:
            cx = np.linspace(self.origin, self.end, self.resolution)
            self.coords = tuple([cx, ])
        elif self.dim == 2:
            cx = np.linspace(self.origin[0], self.end[0],
                             self.resolution[0])[:, np.newaxis]
            cy = np.linspace(self.origin[1], self.end[1],
                             self.resolution[1])[np.newaxis, :]
            self.coords = tuple([cx, cy])
        elif self.dim == 3:
            cx = np.linspace(self.origin[0], self.end[0],
                             self.resolution[0])[:, np.newaxis, np.newaxis]
            cy = np.linspace(self.origin[1], self.end[1],
                             self.resolution[1])[np.newaxis, :, np.newaxis]
            cz = np.linspace(self.origin[2], self.end[2],
                             self.resolution[2])[np.newaxis, np.newaxis, :]
            self.coords = tuple([cx, cy, cz])

    def indices(self, tab):
        """
        returns indices of the point of coordinates (close to) tab = x, y, z
        If (x, y, z) is not a grid point, it returns the closest grid point.
        """
        ind = []
        for d in xrange(self.dim):
            cond = np.where(abs(self.coords[d] - tab[d])
                            < (self.space_step[d] * 0.5))
            if cond[0].size > 0:
                ind.append(cond[d][0])
            else:
                return False
        return tuple(ind)

    def __str__(self):
        """ Sub mesh display """
        s = 'Coords (topo):' + str(self._topology.proc_coords[:])
        s += ' Sub mesh resolution:' + str(self.resolution) + '\n'
        s += 'Starting point global indices :' + str(self.global_start) + '\n'
        s += 'Local indices for "computed" points (excluding ghosts) :\n'
        s += '   start = ' + str(self.local_start)
        s += ', end = ' + str(self.local_end) + '\n'
        return s
