"""
@file topology.py

Parmes topologies definitions and related tools.
A Parmes Topology is defined as the association of
a mpi process distribution (mpi topology) and of a set of local meshes
(one per process).

At the time, only cartesian topologies with cartesian meshes are available.

This module provides the following classes :
- Cartesian : 1,2 or 3D parmes topology.
- Bridge : bridge between two parmes topologies : i.e. what must be
exchanged between two topologies and how.

To get more details try :
\code
>> import parmepy.mpi.topology as topo
>> help(topo.Cartesian)
>> help(topo.Bridge)
\endcode

"""

from parmepy.constants import ORDER, np, PARMES_INDEX, debug, PARMES_INTEGER
from parmepy.mpi.mesh import SubMesh
from itertools import count
from parmepy.mpi.main_var import MPI
import parmepy.tools.numpywrappers as npw


class Cartesian(object):
    """
    Define a MPI cartesian topology with an associated
    mesh for each sub-domain.
    Keyword arguments :
    - domain : the geometry; it must be a box.
    - globalMeshResolution : user input. Number of points in the domain
    in each direction. We assume that first point corresponds to origin, and
    last point to boundary point, whatever the boundary type is.
    That is x[0] = domain.origin and
    x[globalMeshResolution-1] = domain.Lengths_x.
    - dim : dimension of the topology
    - shape : topology resolution (i.e process layout in each direction).
    - comm : MPI communicator used to create this topology
    - periods : periodicity of the topology (in each direction)
    - ghosts : number of points in the ghost layer

    """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    ## Counter of topology.Cartesian instances to set a unique id for each
    ## Cartesian topology instance.
    __ids = count(0)

    @debug
    def __init__(self, domain, dim, globalMeshResolution,
                 comm=None, periods=None, ghosts=None, cutdir=None,
                 shape=None,
                 precomputed=False, localres=None, localoffset=None):

        assert domain.dimension == globalMeshResolution.__len__(), \
            'The resolution size differs from the domain dimension.'

        ## Associated domain
        self.domain = domain
        ## An id for the topology
        self.__id = self.__ids.next()
        ## Dim of the cartesian grid for MPI processes.
        self.dim = dim
        # Mind the sequential case : dim from withConstr may be equal
        # to 0 if shape[:] = 1
        ## (Source) Communicator used to build the topology
        if comm is None:
            from parmepy.mpi.main_var import main_comm as comm
        self._comm_origin = comm
        ## number of process in comm_origin
        self._origin_size = self._comm_origin.Get_size()
        ## True if everything has been properly set for this topology
        self.isUpToDate = False
        ## Grid of mpi process layout
        self.shape = np.ones(self.domain.dimension, dtype=PARMES_INDEX)
        ## MPI processes layout.
        if shape is None:
            # If shape is not provided, computation of  the "optimal" processus
            # distribution for each direction of the grid topology.
            shape = np.asarray(MPI.Compute_dims(self._origin_size, self.dim),
                               dtype=PARMES_INDEX)

            if cutdir is not None:
                self.cutdir = np.array(cutdir, dtype=np.bool)
                self.shape[self.cutdir] = shape
            else:
                ## Reorder shape according to the data layout
                # if arrays are in "fortran" order (column major)
                # the last dir has priority for distribution.
                # For C-like (row major) arrays, first dir is the
                # first to be distributed, and so on.
                self.shape[:self.dim] = shape
                self.shape.sort()
                if(ORDER == 'C'):
                    self.shape[:] = self.shape[::-1]
                self.cutdir = self.shape != 1
             # redim if necessary ...
            self.dim = self.shape[self.cutdir].size

        else:
            shape = np.asarray(shape)
            assert shape.size == self.domain.dimension, 'Input shape must be \
                of the same size as the domain dimension.'
            if cutdir is None:
                self.cutdir = shape != 1
            else:
                self.cutdir = cutdir

            self.shape = shape
            self.dim = self.shape[self.cutdir].size

        # Special care for the 1 process case:
        if self._origin_size == 1:
            self.dim = 1
            self.cutdir = shape != 1
            self.cutdir[-1] = True

        ## MPI process grid periodicity; default is true.
        self.periods = np.ones((self.dim), dtype=np.bool)
        if(periods is not None):
            assert (periods.size == self.dim)
            self.periods[:] = periods[:]
        self.comm = self._comm_origin.Create_cart(self.shape[self.cutdir],
                                                  periods=self.periods,
                                                  reorder=True)

        ## Size of the topology (i.e. total number of mpi processes)
        self.size = self.comm.Get_size()
        ## Rank of the current process in the topology
        self.rank = self.comm.Get_rank()
        ## Coordinates of the current process
        self.reduced_coords = np.asarray(self.comm.Get_coords(self.rank),
                                         dtype=PARMES_INDEX)
        ## Coordinates of the current process
        ## What is different between proc_coords and reduced_coords?
        ## --> proc_coords has values even for directions that
        ## are not distributed. If cutdir = [False, False, True]
        ## then reduced_coords = [ nx ] and proc_coords = [0, 0, nx]
        self.proc_coords = np.zeros(self.domain.dimension, dtype=PARMES_INDEX)
        self.proc_coords[self.cutdir] = self.reduced_coords
        ## Neighbours : self.neighbours[0,i] (resp. [1,i])
        ## previous (resp. next) neighbour in direction i
        ## (warning : direction in the grid of process).
        self.neighbours = np.zeros((2, self.dim), dtype=PARMES_INTEGER)
        for direction in range(self.dim):
            self.neighbours[:, direction] = self.comm.Shift(direction, 1)

        ## ghost layer (default = 0)
        if(ghosts is None):
            self.ghosts = np.zeros((self.domain.dimension),
                                   dtype=PARMES_INDEX)
        else:
            self.ghosts = np.asarray(ghosts, dtype=PARMES_INTEGER)
            assert(self.ghosts.size == domain.dimension)
        assert(np.all(self.ghosts >=0))
        self.hasGhosts = not np.all(self.ghosts == 0)

        # --- Computation of the local mesh resolution and indices ---
        ##  Resolution of the global mesh
        self.globalMeshResolution = np.asarray(globalMeshResolution,
                                               dtype=PARMES_INDEX)

        # If local resolution has already been computed
        # (for example from fftw init output)
        if precomputed:
            self.G_start = localoffset
            self.localGridResolution = localres

        # Usual case : we compute local resolution/offset
        else:
            # Number of "computed" points (i.e. excluding ghosts/boundaries).
            pts_noghost = np.zeros((self.domain.dimension), dtype=PARMES_INDEX)
            # Warning FP : we should test boundary conditions type here
            # If periodic, resol_calc = (globalMeshResolution - 1)
            # else, resol_calc =  (globalMeshResolution - 2)
            resolCalc = np.zeros((domain.dimension), dtype=PARMES_INDEX)
            resolCalc[:] = self.globalMeshResolution[:] - 1
            pts_noghost[:] = resolCalc // self.shape

            # If any, remaining points are
            # added on the mesh of the last process.
            remainingPts = np.zeros(self.domain.dimension, dtype=PARMES_INDEX)
            remainingPts[:] = resolCalc % self.shape

            # Then the total number of points (size of arrays to be allocated)
            nbpoints = pts_noghost.copy()
            for i in range(self.domain.dimension):
                if(self.proc_coords[i] == self.shape[i] - 1):
                    nbpoints[i] += remainingPts[i]

            self.localGridResolution = resolCalc.copy()
            self.localGridResolution[:] = nbpoints[:]
            self.localGridResolution[:] += 2 * self.ghosts[:]

            ## Global indices for the local mesh points;
            ## self.G_start[i] is the index of the first
            ## point in the i direction (It includes ghosts/boundary points).
            self.G_start = np.zeros((domain.dimension), dtype=PARMES_INDEX)
            self.G_start[:] = self.proc_coords[:] * pts_noghost[:]

        ## Local mesh on the current mpi process.
        self.mesh = None

        # The topology is register into domain list.
        # If it already exists (in the sense of the comparison operator
        # of the present class) then isNew is turned to false
        # and the present instance will be destroyed.
        self.isNew = True
        self.__id = self.domain.register(self)

    def getParent(self):
        """
        returns the communicator used to build this topology
        """
        return self._comm_origin

    @debug
    def setUp(self):
        """Topology set up.

        Create topology's Local Mesh.
        This function is called during registration in the domain
        (domain.register(topo)).
        """
        if not self.isUpToDate:
            self.mesh = SubMesh(self, self.G_start, self.localGridResolution)
        self.isUpToDate = True

    @classmethod
    def withResolution(cls, domain, shape,
                       globalMeshResolution, ghosts=None, periods=None,
                       comm=None):
        """
        Compute a topology for a given shape of the processus grid.
        If the dimension of the topology is smaller than the domain dimension,
        shape will be reordered according to the following rule :
        (depends on array storage type (C or Fortran))
        - if C : first dir, first distributed.
        - if Fortran : last dir, first distributed.
        @param domain in which the topology is defined
        @param topoResolution : mpi process grid resolution (MUST be of
        the same size as domain), set topoResolution[i] = 1 if
        you don't want to cut the domain in direction i.
        @param ghosts : number of ghost points in each direction, default = 1.
        @param periods : periodicity in each direction, default = true.
        """

        shape = np.asarray(shape, dtype=PARMES_INDEX)
        assert shape.size == domain.dimension, 'Array for topo \
            resolution must be of size domain.dimension'
        dim = shape[shape != 1].size
        assert shape.all(), "Topology error : you try to \
            create a topology with 0 resolution in one direction."
        if dim < domain.dimension:
            shape.sort()
            if(ORDER == 'C'):
                shape[:] = shape[::-1]
        cutdir = shape != 1

        return cls(domain, dim, globalMeshResolution, cutdir=cutdir,
                   periods=periods, ghosts=ghosts, shape=shape,
                   comm=comm)

    @classmethod
    def withResolutionFixed(cls, domain, shape,
                            globalMeshResolution, ghosts=None, periods=None,
                            comm=None):
        """
        Compute a topology for a given shape of the processus grid.
        @param domain in which the topology is defined
        @param topoResolution : mpi process grid resolution (MUST be of
        the same size as domain), set topoResolution[i] = 1 if
        you don't want to cut the domain in direction i.
        @param ghosts : number of ghost points in each direction, default = 1.
        @param periods : periodicity in each direction, default = true.
        """
        shape = np.asarray(shape, dtype=PARMES_INDEX)
        assert shape.size == domain.dimension, 'Array for topo \
            resolution must be of size domain.dimension'
        dim = shape[shape != 1].size
        assert shape.all(), "Topology error : you try to \
            create a topology with 0 resolution in one direction."
        cutdir = shape != 1
        return cls(domain, dim, globalMeshResolution,
                   cutdir=cutdir, periods=periods, ghosts=ghosts, shape=shape,
                   comm=comm)

    @classmethod
    def withCutdir(cls, domain, cutdir,
                   globalMeshResolution, ghosts=None, periods=None,
                   comm=None):
        """
        Compute a topology from a list of directions to be cut.
        @param domain in which the topology is defined
        @param cutdir : array of bool, cutdir[i] = true if mesh must be
        distributed in direction i.
        (cutdir must be of the same size as the domain).
        @param ghosts : number of ghost points in each direction, default = 1.
        @param periods : periodicity in each direction, default = true.
        """
        cutdir = np.asarray(cutdir, dtype=np.bool)
        dim = cutdir[cutdir].size
        return cls(domain, dim, globalMeshResolution,
                   cutdir=cutdir, periods=periods, ghosts=ghosts,
                   comm=comm)

    @classmethod
    def withPrecomputedResolution(cls, domain, shape,
                                  globalMeshResolution, localres,
                                  offset, ghosts=None, periods=None,
                                  comm=None):
        """
        Compute a topology for a given shape of the processus grid
        and a given local mesh resolution/offset (example : output from
        fftw init process).
        @param domain in which the topology is defined
        @param topoResolution : mpi process grid resolution (MUST be of
        the same size as domain), set topoResolution[i] = 1 if
        you don't want to cut the domain in direction i.
        @param localres : local mesh resolution
        @param offset : local offset (i.e. global index of first local mesh
        points)
        @param ghosts : number of ghost points in each direction, default = 1.
        @param periods : periodicity in each direction, default = true.
        """
        shape = np.asarray(shape, dtype=PARMES_INDEX)
        assert shape.size == domain.dimension, 'Array for topo \
            resolution must be of size domain.dimension'
        dim = shape[shape != 1].size
        assert shape.all(), "Topology error : you try to \
            create a topology with 0 resolution in one direction."
        cutdir = shape != 1

        return cls(domain, dim, globalMeshResolution,
                   precomputed=True, localres=localres, localoffset=offset,
                   periods=periods, ghosts=ghosts, cutdir=cutdir, shape=shape,
                   comm=comm)

    def __eq__(self, other):
        """
        Operator for comparison. Based on:
        @li globalMeshResolution
        @li localMeshResolution
        @li shape
        @li domain
        @param other : Topology to compare with.
        @todo completer le test dans parmepy/mpi/tests/test_topology.py
        """
        return np.equal(self.globalMeshResolution,
                        other.globalMeshResolution).all() and \
            np.equal(self.localGridResolution,
                     other.localGridResolution).all() and \
            np.equal(self.shape, other.shape).all() and \
            np.equal(self.ghosts, other.ghosts).all() and \
            self.domain == other.domain

    def __ne__(self, other):
        """
        Not equal operator.
        Seems to be required in addition to __eq__ to
        avoid 'corner-case' behaviors.
        @param other : Topology to compare with.
        """
        result = self.__eq__(other)
        if result is NotImplemented:
            return result
        return not result

    def __str__(self):
        """ Topology info display """
        s = '======== Topology id : ' + str(self.__id) + ' ========\n'
        if(self.rank == 0):
            s += str(self.dim) + 'D cartesian Topology of size ' \
                + str(self.shape) + '\n'
        s += 'Process of coordinates ' + str(self.proc_coords[:])
        s += ' and of ranks (topo/origin) ' + str(self.rank) + '/'
        s += str(self._comm_origin.Get_rank()) + '.\n'
        s += 'Neighbours coordinates : \n'
        s += str(self.neighbours) + '\n'
        s += 'Ghost layer : ' + str(self.ghosts[:]) + '\n'
        s += 'Global mesh resolution : ' + str(self.globalMeshResolution)
        s += '\n'
        s += str(self.mesh)
        s += '\n=================================\n'
        return s

    def getId(self):
        """
        @return the id of the present topology.
        This id is unique among all defined topologies.
        """
        return self.__id

    def toIndexLocal(self, glob_index):
        """
        Return the local mesh indices from
        their global indices.
        Ilocal = Iglobal - G_start + ghost
        @param glob_index an array of size domain.dim*2
        with glob_index = [start_dir1, end_dir1, start_dir2, end_dir2, ...]
        return an array of indices, same 'setup' as glob_index.
        """
        ilocal = np.asarray(glob_index).copy()
        ilocal[0::2] = ilocal[0::2] - self.G_start + self.ghosts
        ilocal[1::2] = ilocal[1::2] - self.G_start + self.ghosts
        return ilocal

    def toIndexGlobal(self, sl):
        """
        Return the global mesh indices from
        their local indices.
        Iglobal = Ilocal + G_start - ghost
        @param loc_index an array of size domain.dim*2
        with loc_index = [start_dir1, end_dir1, start_dir2, end_dir2, ...]
        return an array of indices, same 'setup' as loc_index.
        """
        g_start = self.mesh.global_start
        return [slice(sl[i].start + g_start[i], sl[i].stop + g_start[i])
                for i in xrange(self.domain.dimension)]


class Bridge(object):
    """
    A Bridge is the definition of what must be exchanged between two topologies
    and how to do so.

    For a bridge between topo1 and topo2, it will provide :
    bridge.sendTo = [[rk, i, j, k, l, ...]]
    bridge.recvFrom = [[rk, i, j, k, l, ...]]
    where each line corresponds to a message that must be send
    from current mpi process to process rk for sendTo or
    receive by current mpi process from process rk for recvFrom.
    The other columns give the numbers of the node of the mesh
    that must be exchanged.
    Warning : index numbers corresponds to local nodes numbers.

    In sendTo/recvFrom :
    * First column : dest/source of the message.
    Warning : rank number is given in main_comm.
    * second-third columns : first and last index of nodes in the first dir
    * 4th-5th : first and last index of nodes in the second dir
    * and so on depending on the size of the domain
    associated to the topologies.

    """

    @debug
    def __init__(self, topoFrom, topoTo, parentComm=None):
        """
        Bridge constructor.
        @param source topology
        @param targeted topology
        @param parent communicator, if None, use topoFrom.getParent()
        """
        ## Source topology
        self.topoFrom = topoFrom
        ## Targeted topology
        self.topoTo = topoTo

        # Both topologies must be uptodate
        assert(topoFrom.isUpToDate)
        assert(topoTo.isUpToDate)

        ## recvFrom[rk] returns a list of indices that should be
        ## received from process rk and copied into local field.
        ## Usage :  mpi.recv(rk, ..., yourField[recvFrom[rk]],...)
        self.recvFrom = {}

        ## sendTo[rk] returns a list of indices that should be
        ## sent from current process to process rk.
        ## Usage :  mpi.send(rk, ..., yourField[sendTo[rk]],...)
        self.sendTo = {}

        ## List of local indices that should be copied from data of
        ## topofrom to data of topoTo for the current process.
        ## Usage:
        ## field_topoto[ito] = field_topofrom[ifrom]
        ## --> no mpi messages
        self.ifrom = []
        ## List of local indices that should be updated on the
        ## current process for topoTo from values on the
        ## same process for topoFrom.
        ## Usage:
        ## field_topoto[ito] = field_topofrom[ifrom]
        ## --> no mpi messages
        self.ito = []

        self.hasLocalInter = True

        ## Parent communicator
        ## Todo : define some proper conditions for compatibility
        ## between topoFrom, topoTo and parent:
        ## - same size
        ## - same domain
        ## - common processus ...
        ## At the time we check that both topo have
        ## the same comm_origin.
        if parentComm is None:
            self._parentComm = self.topoFrom.getParent()
        else:
            self._parentComm = parentComm
        self._parentRank = self._parentComm.Get_rank()
        self._parentSize = self._parentComm.Get_size()
        if self.topoFrom == self.topoTo:
            return

        assert(self.topoTo.getParent() is self._parentComm)

        # Both topologies must have the
        # same number of mpi processes.
        assert topoFrom.size == topoTo.size

        dom = topoFrom.domain

        # 1 - Get global mesh indices for each topology
        size1 = topoFrom.size
        start1 = topoFrom.mesh.global_start
        end1 = topoFrom.mesh.global_end
        iglob1 = np.zeros((size1, dom.dimension * 2), dtype=np.int32)
        iglob1[self._parentRank, 0::2] = start1
        iglob1[self._parentRank, 1::2] = end1

        size2 = topoTo.size
        start2 = topoTo.mesh.global_start
        end2 = topoTo.mesh.global_end
        iglob2 = np.zeros((size2, dom.dimension * 2), dtype=np.int32)
        iglob2[self._parentRank, 0::2] = start2
        iglob2[self._parentRank, 1::2] = end2

        # 2 Distribute these indices.
        self._parentComm.Allgather([iglob1[self._parentRank, :], MPI.INT],
                                   [iglob1, MPI.INT])
        self._parentComm.Allgather([iglob2[self._parentRank, :], MPI.INT],
                                   [iglob2, MPI.INT])

        # Connectivity :
        sTo = np.zeros((0, 1 + dom.dimension * 2), dtype=np.int)
        indexFrom = []
        for dim in range(dom.dimension):
            indexFrom.append(range(iglob1[self._parentRank, dim * 2],
                                   iglob1[self._parentRank, dim * 2 + 1] + 1))

        line = np.zeros((1 + 2 * dom.dimension), dtype=np.int32)
        listRanks = [i for i in range(self._parentSize)
                     if i != self._parentRank]

        for i in listRanks:
            line[0] = i
            hasInter = True
            for dim in range(dom.dimension):
                indexTo = range(iglob2[i, dim * 2], iglob2[i, dim * 2 + 1] + 1)
                interRow = [k for k in indexFrom[dim] if k in indexTo]
                interRow.sort()
                if interRow.__len__():
                    line[dim * 2 + 1] = interRow[0]
                    line[dim * 2 + 2] = interRow[-1]
                else:
                    hasInter = False
                    break

            if hasInter:
                sTo = np.vstack((sTo, line))

        localFrom = np.zeros((2 * dom.dimension), dtype=np.int32)

        for dim in range(dom.dimension):
            indexTo = range(iglob2[self._parentRank, dim * 2],
                            iglob2[self._parentRank, dim * 2 + 1] + 1)
            interRow = [k for k in indexFrom[dim] if k in indexTo]
            interRow.sort()

            if interRow.__len__():
                localFrom[dim * 2] = interRow[0]
                localFrom[dim * 2 + 1] = interRow[-1]
            else:
                self.hasLocalInter = False
                break

        if self.hasLocalInter:
            localTo = topoTo.toIndexLocal(localFrom)
            localFrom = topoFrom.toIndexLocal(localFrom)

            for d in range(dom.dimension):
                self.ifrom.append(slice(localFrom[2 * d],
                                        localFrom[2 * d + 1] + 1))
                self.ito.append(slice(localTo[2 * d],
                                      localTo[2 * d + 1] + 1))

        # --- Compute globalConnectivity on process 0 and distribute it ---
        # Global Connectivity : each line corresponds to a message :
        # globalConnectivity[i,:] = [i, j, 0, 3, 2, 6]
        # means that process i must send to process j its mesh
        # from node 0 to 3 in the first direction and from node 2 to
        # 6 in the second dir AND that process j must receive the same mesh
        # from process i. Nodes number given in the global index set.
        # Warning : process ranks are given in mpi_main.
        globalConnectivity = np.zeros((sTo.shape[0], 2 +
                                       dom.dimension * 2), dtype=np.int)

        globalConnectivity[:, 1:] = sTo.copy(order='A')
        globalConnectivity[:, 0] = self._parentRank

        if self._parentRank == 0:
            for i in range(1, self._parentSize):
                temp = self._parentComm.recv(source=i)
                globalConnectivity = np.vstack((globalConnectivity, temp))

        else:
            self._parentComm.ssend(globalConnectivity, dest=0)

        ## recvFrom[:,0] rank of the sending process in main_comm
        ## recvFrom[:,1:] local indices (start,end, ...)
        ## of the points to be receive for each direction.
        ## Example (in 2D)
        ## recvFrom = [[1, 2, 5, 1, 3]]
        ## means that the current process must receive from process 1
        ## an array that will be saved at positions of indices 2:5
        ## in the first dir and 1:3 in second dir
        rFrom = np.empty((), dtype=np.int32)
        if self._parentRank == 0:
            for fromRank in range(self._parentSize):
                cond1 = globalConnectivity[:, 1] == fromRank
                cond2 = globalConnectivity[:, 0] != fromRank
                cond = cond1 & cond2
                sol = np.compress(cond, globalConnectivity, axis=0)
                cond1 = np.ones((sol.shape[1]), dtype=np.bool)
                cond1[1] = False
                sol = np.compress(cond1, sol, axis=1)
                if(fromRank != self._parentRank):
                    self._parentComm.ssend(sol, dest=fromRank)
                else:
                    rFrom = sol.copy()

        else:
            rFrom = self._parentComm.recv(source=0)

        # Final setup for recvFrom and sendTo : shift from
        # global indices to local ones.
        for row in range(rFrom.shape[0]):
            rFrom[row, 1:] = topoTo.toIndexLocal(rFrom[row, 1:])
        for row in range(sTo.shape[0]):
            sTo[row, 1:] = topoFrom.toIndexLocal(sTo[row, 1:])

        row = 0
        for rk in rFrom[:, 0]:
            self.recvFrom[rk] = []
            for d in range(dom.dimension):
                self.recvFrom[rk].append(slice(rFrom[row, 2 * d + 1],
                                               rFrom[row, 2 * d + 2] + 1))
            row += 1

        row = 0
        for rk in sTo[:, 0]:
            self.sendTo[rk] = []
            for d in range(dom.dimension):
                self.sendTo[rk].append(slice(sTo[row, 2 * d + 1],
                                             sTo[row, 2 * d + 2] + 1))
            row += 1

    @debug
    def setUp(self):
        """

        """

    def __str__(self):
        """ TopologyBridge info display """
        s = '======== Bridge from topology ' + str(self.topoFrom.getId())
        s += ' to topology ' + str(self.topoTo.getId()) + ' ========\n'
        s += '[' + str(self._parentRank) + '] sendTo :'
        s += str(self.sendTo) + '\n'
        s += '[' + str(self._parentRank) + '] recvFrom :'
        s += str(self.recvFrom) + '\n'
        s += '[' + str(self._parentRank) + '] iTo :' + str(self.ito) + '\n'
        s += '[' + str(self._parentRank) + '] iFrom :' + str(self.ifrom) + '\n'
        s += '=================================\n'
        return s


class Bridge_intercomm(object):
    """
    This bridge is defined as the Bridge class.

    First, an intercommunicator is created from the tasks identifiers (used
    by user to split the parent communicator).

    The mesh indices are exchanged by process across the intercommunicator.
    Intersections of mesh are computed and indices slices are stored in the
    transfers dictionnary.

    """
    @debug
    def __init__(self, topo, parent_comm, id_from, id_to, proc_tasks):
        """
        Create a bridge between communicators of a single topology.
        @param topo : Topology to consider.
        @param parent_comm : MPI communicator that containts all process
        involved in this bridge.
        @param id_from : Identifier for the input task.
        @param id_to : Identifier for the output task.
        @param proc_task : Tasks assignment for process.
        """
        self.topo = topo
        self.id_from = id_from
        self.id_to = id_to
        self.parent_comm = parent_comm
        self.proc_tasks = proc_tasks
        parent_rank = self.parent_comm.Get_rank()
        my_comm = self.topo.comm
        my_rank = my_comm.Get_rank()
        dim = self.topo.domain.dimension

        # 0 - Create an intercommunicator
        # Create_intercomm attributes are:
        #   - local rank of leader process for current group (always 0)
        #   - parent communicator
        #   - rank of leader process in the remote group
        if self.proc_tasks[parent_rank] == self.id_from:
            self.inter_comm = my_comm.Create_intercomm(
                0,
                self.parent_comm, self.proc_tasks.index(self.id_to))
        if self.proc_tasks[parent_rank] == self.id_to:
            self.inter_comm = my_comm.Create_intercomm(
                0,
                self.parent_comm, self.proc_tasks.index(self.id_from))

        # 1 - Get global mesh indices of my domain
        my_start = self.topo.mesh.global_start
        my_end = self.topo.mesh.global_end
        my_slices = npw.zeros((dim * 2, ), dtype=np.int)
        my_slices[0::2] = my_start
        my_slices[1::2] = my_end

        # 2 - Get global mesh indices of other domain
        if self.proc_tasks[parent_rank] == self.id_from:
            from_size = my_comm.Get_size()
            to_size = self.inter_comm.Get_remote_size()
            other_size = to_size
        if self.proc_tasks[parent_rank] == self.id_to:
            to_size = my_comm.Get_size()
            from_size = self.inter_comm.Get_remote_size()
            other_size = from_size

        # 3 - Get global mesh indices of other topology process
        # Each process broadcast his mesh indices to all process of other group
        # from_slices and to_slices are temporary arrays
        from_slices = npw.zeros((from_size, dim * 2), dtype=np.int)
        to_slices = npw.zeros((to_size, dim * 2), dtype=np.int)

        # 3.1 - each process of the 'from' group broadbast to the 'to' group
        for rk in xrange(from_size):
            if self.proc_tasks[parent_rank] == self.id_from:
                if my_rank == rk:
                    from_slices[rk, :] = my_slices
                    self.inter_comm.bcast(from_slices[rk, :],
                                          root=MPI.ROOT)
                else:
                    self.inter_comm.bcast(from_slices[rk, :],
                                          root=MPI.PROC_NULL)
            elif self.proc_tasks[parent_rank] == self.id_to:
                from_slices[rk, :] = self.inter_comm.bcast(from_slices[rk, :],
                                                           root=rk)
        # 3.2 - each process of the 'to' group broadbast to the 'from' group
        for rk in xrange(to_size):
            if self.proc_tasks[parent_rank] == self.id_to:
                if my_rank == rk:
                    to_slices[rk, :] = my_slices
                    self.inter_comm.bcast(to_slices[rk, :],
                                          root=MPI.ROOT)
                else:
                    self.inter_comm.bcast(to_slices[rk, :],
                                          root=MPI.PROC_NULL)
            elif self.proc_tasks[parent_rank] == self.id_from:
                to_slices[rk, :] = self.inter_comm.bcast(to_slices[rk, :],
                                                         root=rk)
        self.parent_comm.Barrier()

        if self.proc_tasks[parent_rank] == self.id_from:
            other_slices = to_slices
        elif self.proc_tasks[parent_rank] == self.id_to:
            other_slices = from_slices

        # 4. - Build connectivity
        self.transfers = {}
        for other_rk in xrange(other_size):
            intersect_slice = npw.zeros((dim * 2), dtype=np.int)
            intersect_slice[0::2] = np.maximum(my_slices[0::2],
                                               other_slices[other_rk, 0::2])
            intersect_slice[1::2] = np.minimum(my_slices[1::2],
                                               other_slices[other_rk, 1::2])
            # Convert intersection to local slices, exluding ghosts
            if not (intersect_slice[0::2] > intersect_slice[1::2]).any():
                # compatible intersection (else nothing to transfer)
                self.transfers[other_rk] = [
                    (g + i - s, g + j + 1 - s) for i, j, s, g in zip(
                        intersect_slice[0::2],
                        intersect_slice[1::2], my_start, self.topo.ghosts)]

    @debug
    def setUp(self):
        """

        """

# ## Example of broadcast:
# ## Process 0 of group id_from broadcasting value to group id_to
# if proc_tasks[parent_rank] == id_from:
#     if my_rank == 0:
#         value = 123
#         inter_comm.bcast(value, root=MPI.ROOT)
#     else:
#         value = None
#         inter_comm.bcast(value, root=MPI.PROC_NULL)
#     value = comm_s.bcast(value, root=0) # for broad cast within the group
# elif proc_tasks[parent_rank] == id_to:
#     value = None
#     value = inter_comm.bcast(value, root=0)
# assert value == 123

# ## Example of Point-to-Point send/receve:
# x = npw.zeros((10, ))
# if my_rank == 0 and proc_tasks[parent_rank] == id_from:
#     x = npw.realarray(np.arange(10))
# if my_rank == 1 and proc_tasks[parent_rank] == id_to:
#     inter_comm.Recv([x, 10, PARMES_MPI_REAL], source=0, tag=0)
# if my_rank == 0 and proc_tasks[parent_rank] == id_from:
#     inter_comm.Send([x, 10, PARMES_MPI_REAL], dest=1, tag=0)
# if my_rank == 1 and proc_tasks[parent_rank] == id_to:
#     assert (x == npw.realarray(np.arange(10))).all()
