# -*- coding: utf-8 -*-
"""
@file differential_operations.py
Library of functions used to perform classical vector calculus
(diff operations like grad, curl ...)
"""
from parmepy.constants import debug, XDIR, YDIR, ZDIR
from abc import ABCMeta, abstractmethod
from parmepy.numerics.finite_differences import FD_C_4, FD_C_2, FD2_C_2
import numpy as np


class DifferentialOperation(object):
    __metaclass__ = ABCMeta

    ## @debug
    ## def __new__(cls, *args, **kw):
    ##     return object.__new__(cls, *args, **kw)

    @debug
    @abstractmethod
    def __init__(self, topo, work=None):
        if work is None:
            work = []
        self._work = work
        for wk in self._work:
            assert wk.shape == tuple(topo.mesh.resolution)
        self._dim = topo.domain.dimension

    @staticmethod
    def getWorkLengths(nb_components=None, domain_dim=None, fd_method=None):
        """
        Compute the number of required work arrays for this method.
        @param nb_components : number of components of the
        @param domain_dim : dimension of the domain
        fields on which this method operates.
        @return length of list of work arrays of reals.
        """
        assert nb_components is None
        assert domain_dim is None
        assert fd_method is None
        return 0


class Curl(DifferentialOperation):
    """
    Computes \f$ \nabla \times V \f$, V being a vector field.
    """
    def __init__(self, topo, work, method=FD_C_4):

        assert len(work) == self.getWorkLengths()
        DifferentialOperation.__init__(self, topo, work)
        if method is FD_C_4:
            # - 4th ordered FD,
            # - work space provided by used during function call,
            #   i.e. memshape not required,
            # - fd.compute method

            # declare and create fd scheme
            self.fd_scheme = FD_C_4(topo.mesh.space_step)
            # connect to fd function call
            self.fcall = self.FDCentral
            # check ghosts ... must be updated when
            # other fd schemes will be implemented.
            assert (topo.ghosts >= 2).all(),\
                'you need a ghost layer for FD4 scheme.'
        elif method is FD_C_2:
            # - 2nd ordered FD,
            # - work space provided by used during function call,
            #   i.e. memshape not required,
            # - fd.compute method

            # declare and create fd scheme
            self.fd_scheme = FD_C_2(topo.mesh.space_step)
            # connect to fd function call
            self.fcall = self.FDCentral
            # check ghosts ... must be updated when
            # other fd schemes will be implemented.
            assert (topo.ghosts >= 1).all(),\
                'you need a ghost layer for FD4 scheme.'

        else:
            raise ValueError("FD scheme Not yet implemented")

        self._indices = topo.mesh.iCompute
        # initialize fd scheme
        self.fd_scheme.computeIndices(self._indices)

    @staticmethod
    def getWorkLengths(nb_components=None, domain_dim=None, fd_method=None):
        return 1

    def __call__(self, variable, result):
        return self.fcall(variable, result)

    def FDCentral(self, variable, result):
        """
        @param variable : input vector field
        @param result : list of numpy arrays to save result
        """
        #--  d/dy vz -- in result[XDIR]
        self.fd_scheme.compute(variable[ZDIR], YDIR, result[XDIR])
        # -- -d/dz vy -- in work
        self.fd_scheme.compute(variable[YDIR], ZDIR, self._work[0])
        # result_x = d/dy vz - d/dz vy
        result[XDIR][self._indices] -= self._work[0][self._indices]

        #--  d/dz vx -- in result[YDIR]
        self.fd_scheme.compute(variable[XDIR], ZDIR, result[YDIR])
        # -- -d/dx vz -- in work
        self.fd_scheme.compute(variable[ZDIR], XDIR, self._work[0])
        # result_y = d/dz vx - d/dx vz
        result[YDIR][self._indices] -= self._work[0][self._indices]

        #-- d/dx vy in result[ZDIR]
        self.fd_scheme.compute(variable[YDIR], XDIR, result[ZDIR])
        # result_z = d/dx vy - d/dy vx
        self.fd_scheme.compute(variable[XDIR], YDIR, self._work[0])
        result[ZDIR][self._indices] -= self._work[0][self._indices]

        return result


class DivV(DifferentialOperation):
    """
    Computes \f$ \nabla.(\rho V) \f$, \f$ \rho\f$ a scalar field and
    V a vector field.
    Works for any dimension.
    For all methods a list of work arrays must be provided at init.
    Its length is given by the (static) method DivV.getWorkLengths.

    Methods :
    1 - FD_C_4 (default) : 4th order, centered finite differences,
    based on fd.compute.
    init : func = DivV(topo, work)
    call : result = func(var1, scal, result)
    2 - FD_C_4_CAA : 4th order, centered finite differences,
    based on fd.compute_and_add.
    init : func = DivV(topo, work, method='FD_C_4_CAA')
    call : result = func(var1, scal, result)
    shape is a tuple (like np.ndarray.shape), e.g. (nx,ny,nz).

    Note FP:
    - 1 seems to be faster than 2 for large systems (>200*3) and
    less memory consumming. For smaller systems, well, it depends ...
    """
    def __init__(self, topo, work, method=FD_C_4, fd_optim=None):

        DifferentialOperation.__init__(self, topo, work)

        if method is FD_C_4:
            # - 4th ordered FD,
            # - fd.compute_and_add method
            # declare and create fd scheme
            self.fd_scheme = FD_C_4(topo.mesh.space_step)
            # connect to finite differences function ...
            if fd_optim is 'CAA':
                self.fcall = self.FDCentral4_CAA
            else:
                self.fcall = self.FDCentral4
            # check ghosts ... must be updated when
            # other fd schemes will be implemented.
            assert (topo.ghosts >= 2).all(),\
                'you need a ghost layer for FD4 scheme.'

        else:
            raise ValueError("FD scheme Not yet implemented")

        self._indices = topo.mesh.iCompute
        # initialize fd scheme
        self.fd_scheme.computeIndices(self._indices)

    @staticmethod
    def getWorkLengths(nb_components=None, domain_dim=None, fd_method=None):
        """
        fd_method = FD_C_4 --> call fd_compute
        fd_method = FD_C_4_CAA --> call fd.compute_and_add
        """
        if domain_dim == 1 or nb_components == 1:
            return 1
        elif fd_method is not None:
            return 1
        else:
            return 2

    def __call__(self, var1, scal, result):
        return self.fcall(var1, scal, result)

    def FDCentral4(self, var1, scal, result):
        """
        Compute central finite difference scheme, order 4.
        No work vector provided by user --> self._work will be
        used. It must be created at init and thus memshape is required.
        """
        assert len(self._work) == \
            self.getWorkLengths(len(var1), len(scal.shape))
        assert scal is not result
        for i in xrange(len(var1)):
            assert var1[i] is not result

        # _work[0:1] are used as temporary space
        # for computation
        # div computations are accumulate into result.
        # result does not need initialisation to zero.

        # d/dx (scal * var1x), saved into result
        self._work[0][...] = scal * var1[XDIR]
        self.fd_scheme.compute(self._work[0], XDIR, result)
        # other components (if any) saved into _work[0] and added into result
        # d/dy (scal * var1y), saved in work and added into result
        # d/dz(scal * var1z), saved in work and added into result (if 3D)
        for cdir in xrange(1, len(var1)):
            self._work[0][...] = scal * var1[cdir]
            self.fd_scheme.compute(self._work[0], cdir, self._work[1])
            result[self._indices] += self._work[1][self._indices]

        return result

    def FDCentral4_CAA(self, var1, scal, result):
        """
        Compute central finite difference scheme, order 4.
        Use fd_scheme.compute_and_add.
        No work space as input (use self._work).
        It must be created at init and thus memshape is required.
        CAA stands for compute_and_add.
        """
        assert len(self._work) == \
            self.getWorkLengths(len(var1), len(scal.shape))
        assert scal is not result
        for i in xrange(len(var1)):
            assert var1[i] is not result

        # _work[0] is used as temporary space
        # for computation
        # div computations are accumulate into result.
        # result does not need initialisation to zero.

        # d/dx (scal * var1x), saved into result
        self._work[0][...] = scal * var1[XDIR]
        self.fd_scheme.compute(self._work[0], XDIR, result)
        # other components (if any) saved into _work[0] and added into result
        # d/dy (scal * var1y), saved in work and added into result
        # d/dz(scal * var1z), saved in work and added into result (if 3D)
        for cdir in xrange(1, len(var1)):
            self._work[0][...] = scal * var1[cdir]
            self.fd_scheme.compute_and_add(self._work[0], cdir, result)

        return result


class DivWV(DifferentialOperation):
    """
    Computes \f$ \nabla.(W.Vx, W.Vy, W.Vz) \f$,
    \f$W\f$ and V some vector fields.

    Methods :
    1 - FD_C_4 (default) : 4th order, centered finite differences,
    based on fd.compute.
    init : func = DivV(topo, work)
    call : result = func(var1, var2, result)
    If fd_optim == CAA, based on fd.compute_and_add.

    var1 stands for W and var2 for V.
    result must be a vector field (same number of components as var1 and var2),
    each component with the same size as var1[i].
    work must be a two component vector field, (only 1 component for CAA cases)
    each component with the same size as var1[i].

    """
    def __init__(self, topo, work, method=FD_C_4, fd_optim=None):
        DifferentialOperation.__init__(self, topo, work)
        if method is FD_C_4:
            # - 4th ordered FD,
            # - fd.compute_and_add method
            # connect call function
            self.fcall = self.FDCentral4
            # set div(scal.var1) operator
            self.div = DivV(topo, work, method=method, fd_optim=fd_optim)

        else:
            raise ValueError("FD scheme Not yet implemented")

    @staticmethod
    def getWorkLengths(nb_components=None, domain_dim=None, fd_method=None):
        return DivV.getWorkLengths(nb_components, domain_dim, fd_method)

    def __call__(self, var1, var2, result):
        return self.fcall(var1, var2, result)

    def FDCentral4(self, var1, var2, result):
        """
        """
        assert len(result) == len(var1)

        # Note FP var1[dir] and var2[dir] must be different from result.
        # This is checked inside divV call.

        for cdir in xrange(len(var1)):
            result[cdir] = self.div(var1, var2[cdir], result=result[cdir])

        return result


class Laplacian(DifferentialOperation):
    """
    Computes  \f$ \Delta \rho, \rho \f$ a scalar field.
    """
    def __init__(self, topo, work=None, method=FD2_C_2):

        DifferentialOperation.__init__(self, topo, work)
        if method is FD2_C_2:
            # - 2nd ordered FD,
            self.fd_scheme = FD2_C_2(topo.mesh.space_step)
        else:
            raise ValueError("FD scheme Not yet implemented")
        self._indices = topo.mesh.iCompute
        self.fd_scheme.computeIndices(self._indices)
        # dimension of the fields
        self._dim = topo.domain.dimension

    @staticmethod
    def getWorkLengths(nb_components=None, domain_dim=None, fd_method=None):
        return 0

    def __call__(self, var, result):
        self.fd_scheme.compute(var, 0, result)
        for cdir in xrange(1, self._dim):
            self.fd_scheme.compute_and_add(var, cdir, result)
        return result


class GradS(DifferentialOperation):
    """
    Computes \f$ \nabla\rho \f$, \f$ \rho\f$ a scalar field
    """
    def __init__(self, topo, method=FD_C_4):
        DifferentialOperation.__init__(self, topo)
        if method is FD_C_4:
            self.fcall = self.FDCentral
            self.fd_scheme = FD_C_4(topo.mesh.space_step)
            assert (topo.ghosts >= 2).all(),\
                'you need a ghost layer for FD4 scheme.'

        elif method is FD_C_2:
            self.fcall = self.FDCentral
            self.fd_scheme = FD_C_2(topo.mesh.space_step)
            assert (topo.ghosts >= 1).all(),\
                'you need a ghost layer for FD2 scheme.'

        else:
            raise ValueError("FD scheme Not yet implemented")

        self._indices = topo.mesh.iCompute
        self.fd_scheme.computeIndices(self._indices)
        # dimension of the fields
        self._dim = topo.domain.dimension

    def __call__(self, scal, result):
        return self.fcall(scal, result)

    def FDCentral(self, scal, result):
        """
        Compute central finite difference scheme.
        @param scal a numpy array
        @param result, a list of numpy arrays to save resulting
        grad.
        """
        assert len(result) == self._dim
        for cdir in xrange(self._dim):
            #  d/dcdir (scal), saved in data[cdir]
            self.fd_scheme.compute(scal, cdir, result[cdir])

        return result


class GradV(DifferentialOperation):
    """
    Computes \f$ \nabla V \f$, with
    \f$V\f$ a vector field.
    """
    def __init__(self, topo, method=FD_C_4):
        DifferentialOperation.__init__(self, topo)
        if method is FD_C_4 or method is FD_C_2:
            self.fcall = self.FDCentral
            self.grad = GradS(topo, method)
        else:
            raise ValueError("FD scheme Not yet implemented")

    def __call__(self, var1, result):
        return self.fcall(var1, result)

    def FDCentral(self, var1, result):
        nbc = len(var1)
        assert len(result) == nbc * self._dim
        for cdir in xrange(self._dim):
            i1 = cdir * nbc
            i2 = i1 + nbc
            result[i1:i2] = self.grad(var1[cdir], result[i1:i2])

        return result


class GradVxW(DifferentialOperation):
    """
    Computes \f$ [\nabla(V)][W] \f$, with
    \f$V\f$ and \f$W\f$ some vector fields.
    """
    def __init__(self, topo, work, method=FD_C_4):
        DifferentialOperation.__init__(self, topo, work)

        if method is FD_C_4:
            # - 4th ordered FD, with diagnostics in output
            # - fd.compute method

            # declare and create fd scheme
            self.fd_scheme = FD_C_4(topo.mesh.space_step)
            # connect to fd function call
            assert (topo.ghosts >= 2).all(),\
                'you need a ghost layer for FD4 scheme.'
            self.fcall = self.FDCentral4_diag
        else:
            raise ValueError("FD scheme Not yet implemented")

        self._indices = topo.mesh.iCompute
        self.fd_scheme.computeIndices(self._indices)

    @staticmethod
    def getWorkLengths(nb_components=None, domain_dim=None, fd_method=None):
        """
        Compute the number of required work arrays for this method.
        @param nb_components : number of components of the
        @param domain_dim : dimension of the domain
        fields on which this method operates.
        @return length of list of work arrays of reals.
        """
        if domain_dim == 1:
            return 1
        else:
            return 2

    def __call__(self, var1, var2, result, diagnostics):
        return self.fcall(var1, var2, result, diagnostics)

    def FDCentral4_diag(self, var1, var2, result, diagnostics):
        """
        """
        assert len(result) == len(var1)
        nbc = len(var1)
        diagnostics[:] = 0.0
        for comp in xrange(nbc):
            result[comp][...] = 0.0
            self._work[1][...] = 0.0
            for cdir in xrange(self._dim):
                # self._work = d/dcdir (var1_comp)
                self.fd_scheme.compute(var1[comp], cdir, self._work[0])
                # some diagnostics ...
                if cdir == comp:
                    tmp = np.max(abs(self._work[0]))
                    diagnostics[0] = max(diagnostics[0], tmp)
                self._work[1][...] += abs(self._work[0])

                # compute self._work = self._work.var2[cdir]
                self._work[0][...] = self._work[0] * var2[cdir]
                # sum to obtain nabla(var_comp) . var2, saved into result[comp]
                np.add(result[comp], self._work[0], result[comp])
            diagnostics[1] = max(diagnostics[1], np.max(self._work[1]))
        return result, diagnostics
