"""
@file finite_differences.py
Finite difference schemes description
"""
from abc import ABCMeta, abstractmethod
from parmepy.constants import debug
import parmepy.tools.numpywrappers as npw
import numpy as np


class FiniteDifference(object):

    """
    Describe and apply a finite difference scheme to compute
    1st or second derivative of a variable saved in a numpy array.
    usage :\n
    1 - Initialization
    - declare the scheme (by choosing the required FD class)
    and initialize it with a step from domain discretisation description

    \code
    >> step = topo.mesh_space_step
    >> scheme = FD_C_4(step)
    \endcode

    - Compute scheme indices, based on topology grid points :

    \code
    >> scheme.computeIndices(topo.mesh.iCompute)
    \endcode

    2 - Computation

    For a given numpy array (obviously discretized on the topo
    used to compute indices), to compute
    \f$ result = \frac{\partial tab}{\partial dir}\f$ :

    \code
    >> scheme.compute(tab, dir, result)
    \endcode

    or :
    \f$ result = result + \frac{\partial tab}{\partial dir}\f$

    \code
    >> scheme.compute_and_add(tab, dir, result)
    \endcode

    Notes FP :
    Compute method is much more performant than compute and add
    and needs less memory. But in some cases (when fd results need
    to be accumulate into a field) compute_and_add may be useful
    to limit memory usage, if required.
    See Global_tests/testPerfAndMemForFD_and_div.py for perf results.
    """

    __metaclass__ = ABCMeta

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @abstractmethod
    def __init__(self, step):
        """
        Build a finite difference scheme
        @param step : resolution of the mesh
        (list or numpy array of int)
        """
        ## List of slices representing the mesh on which fd scheme is applied
        self.indices = []
        self._step = np.asarray(step)
        #  dim of the field on which scheme will be applied
        # (i.e dim of the domain)
        self._dim = step.size

    @abstractmethod
    def computeIndices(self, indices):
        """
        @param indices : a list of slices (see for example
        parmepy.mpi.mesh.Mesh iCompute) that represent the local
        mesh on which finite-differences will be applied.
        """

    @abstractmethod
    def compute(self, tab, cdir, result):
        """
        @param[in] tab : a numpy array
        @param[in] cdir : direction of differentiation
        @param[in, out] result : numpy array to save result

        Apply finite different scheme on the variable
        in numpy array var, for direction cdir and save
        the result in ... result.
        """

    @abstractmethod
    def compute_and_add(self, tab, cdir, result):
        """
        @param[in] tab : a numpy array
        @param[in] cdir : direction of differentiation
        @param[in, out] result : numpy array to save result

        Apply finite different scheme on the variable
        in numpy array var, for direction cdir and add
        the result into result tab.
        Note tab == result is allowed.

        """


class FD_C_2(FiniteDifference):
    """
    1st derivative, centered scheme, 2nd order.

    """

    def __init__(self, step):
        FiniteDifference.__init__(self, step)

        # list of indices for index - 1
        self._m1 = []
        # index + 1
        self._a1 = []
        self._coeff = npw.asarray(1. / (2. * self._step))

    def computeIndices(self, indices):

        self.indices = indices
        self._m1 = []
        self._a1 = []
        for dim in xrange(self._dim):
            self._m1.append(list(self.indices))
            self._m1[dim][dim] = slice(self.indices[dim].start - 1,
                                       self.indices[dim].stop - 1,
                                       self.indices[dim].step)
            self._a1.append(list(self.indices))
            self._a1[dim][dim] = slice(self.indices[dim].start + 1,
                                       self.indices[dim].stop + 1,
                                       self.indices[dim].step)

    def compute(self, tab, cdir, result):
        assert result is not tab
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        # FD scheme coefficient

        result[self.indices] = tab[self._a1[cdir]]
        result[self.indices] -= tab[self._m1[cdir]]
        result[self.indices] *= self._coeff[cdir]

    def compute_and_add(self, tab, cdir, result):
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        # FD scheme coefficient
        result[self.indices] += self._coeff[cdir] * (tab[self._a1[cdir]] -
                                                     tab[self._m1[cdir]])


class FD2_C_2(FiniteDifference):
    """
    Second derivative, centered scheme, 2nd order.
    """

    def __init__(self, step):
        FiniteDifference.__init__(self, step)

        # list of indices for index - 1
        self._m1 = []
        # index + 1
        self._a1 = []
        self._coeff = npw.asarray(1. / (self._step * self._step))

    def computeIndices(self, indices):

        self.indices = indices
        self._m1 = []
        self._a1 = []
        for dim in xrange(self._dim):
            self._m1.append(list(self.indices))
            self._m1[dim][dim] = slice(self.indices[dim].start - 1,
                                       self.indices[dim].stop - 1,
                                       self.indices[dim].step)
            self._a1.append(list(self.indices))
            self._a1[dim][dim] = slice(self.indices[dim].start + 1,
                                       self.indices[dim].stop + 1,
                                       self.indices[dim].step)

    def compute(self, tab, cdir, result):
        assert result is not tab
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        # FD scheme coefficient
        result[self.indices] = tab[self.indices]
        result[self.indices] *= -2
        result[self.indices] += tab[self._a1[cdir]]
        result[self.indices] += tab[self._m1[cdir]]
        result[self.indices] *= self._coeff[cdir]

    def compute_and_add(self, tab, cdir, result):
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        # FD scheme coefficient

        result[self.indices] += self._coeff[cdir] * (-2 * tab[self.indices] +
                                                     tab[self._m1[cdir]] +
                                                     tab[self._a1[cdir]])


class FD_C_4(FiniteDifference):
    """
    1st derivative, centered scheme, 4th order.
    """

    def __init__(self, step):
        FiniteDifference.__init__(self, step)

        # list of indices for index - 1
        self._m1 = []
        # for index - 2
        self._m2 = []
        # index + 1
        self._a1 = []
        # index + 2
        self._a2 = []
        # FD scheme coefficients
        self._coeff = npw.asarray(1. / (12. * step))

    def computeIndices(self, indices):

        self.indices = indices
        self._m1 = []
        self._m2 = []
        self._a1 = []
        self._a2 = []
        for dim in xrange(self._dim):
            self._m1.append(list(self.indices))
            self._m1[dim][dim] = slice(self.indices[dim].start - 1,
                                       self.indices[dim].stop - 1,
                                       self.indices[dim].step)
            self._m2.append(list(self.indices))
            self._m2[dim][dim] = slice(self.indices[dim].start - 2,
                                       self.indices[dim].stop - 2,
                                       self.indices[dim].step)
            self._a1.append(list(self.indices))
            self._a1[dim][dim] = slice(self.indices[dim].start + 1,
                                       self.indices[dim].stop + 1,
                                       self.indices[dim].step)
            self._a2.append(list(self.indices))
            self._a2[dim][dim] = slice(self.indices[dim].start + 2,
                                       self.indices[dim].stop + 2,
                                       self.indices[dim].step)

    def compute(self, tab, cdir, result):
        assert result is not tab
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        result[self.indices] = tab[self._a1[cdir]]
        result[self.indices] -= tab[self._m1[cdir]]
        result[self.indices] *= 8
        result[self.indices] += tab[self._m2[cdir]]
        result[self.indices] -= tab[self._a2[cdir]]
        result[self.indices] *= self._coeff[cdir]

    def compute_and_add(self, tab, cdir, result):
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        result[self.indices] += self._coeff[cdir] * (8 *
                                                     (tab[self._a1[cdir]] -
                                                      tab[self._m1[cdir]]) +
                                                     tab[self._m2[cdir]] -
                                                     tab[self._a2[cdir]])
