"""
@file remeshing.py
"""
from parmepy.constants import np, PARMES_INDEX
from parmepy.numerics.method import NumMethod


class Remeshing(NumMethod):
    """Remshing"""

    def __init__(self, dim, topo, d, work, iwork):
        """
        Create a remeshing numeric method based on given formula.
        @param dim : problem dimension
        @param dx : mesh space step
        @param origin : mesh lower point
        @param formula : Remeshing formula to use.
        @param work : Work arrays (floats)
        @param iwork : Work arrays (integers)
        work must be a list containing (2 elements):
          - numpy float array like y[0] (used to store particles contributions)
          - numpy float array like y[0] (used to store distance to grid points)
        iwork must be a list containing (dim elements( in the order):):
          - numpy integer array of shape like y[0]
          - numpy integer array of shape like y[0] (if in 2D and 3D)
          - numpy integer array of shape like y[0] (if in 3D)
        Availables formulas :
          - 'l2_1' : Labmda2,1 : (=M'4) 4 point formula, C1 regularity
          - 'l2_2' : Labmda2,2 : 4 point formula, C2 regularity
          - 'l4_2' : Labmda4,2 : (=M'6) 6 point formula, C2 regularity
          - 'l4_3' : Labmda4,3 : 6 point formula, C3 regularity
          - 'l4_4' : Labmda4,4 : 6 point formula, C4 regularity
          - 'l6_3' : Labmda6,3 : 8 point formula, C3 regularity
          - 'l6_4' : Labmda6,4 : 8 point formula, C4 regularity
          - 'l6_5' : Labmda6,5 : 8 point formula, C5 regularity
          - 'l6_6' : Labmda6,6 : 8 point formula, C6 regularity
          - 'l8_4' : Labmda8,4 : 10 point formula, C4 regularity
          - 'm8prime' : M8prime formula
        """
        NumMethod.__init__(self)
        self.dir = d
        self._dim = dim
        self.work = work
        self.iwork = iwork
        assert len(self.work) == 2
        assert len(self.iwork) == self._dim
        self._shift = 0
        self._weights = None
        self.topo = topo
        self._slice_all = [slice(None, None, None)
                           for dd in xrange(dim)]

        if self._dim == 3:
            if self.dir == 0:
                self._affect_working_arrays = self._affect_work_3D_X
            if self.dir == 1:
                self._affect_working_arrays = self._affect_work_3D_Y
            if self.dir == 2:
                self._affect_working_arrays = self._affect_work_3D_Z
        if self._dim == 2:
            if self.dir == 0:
                self._affect_working_arrays = self._affect_work_2D_X
            if self.dir == 1:
                self._affect_working_arrays = self._affect_work_2D_Y
        if self._dim == 1:
            if self.dir == 0:
                self._affect_working_arrays = self._affect_work_1D

    def slice_i_along_d(self, i, d):
        l = list(self._slice_all)
        l[d] = i
        return tuple(l)

    @staticmethod
    def getWorkLengths(nb_components=None, domain_dim=1):
        return 2, domain_dim

    def _affect_work_1D(self, resol):
        return (self.work[0], self.work[1], tuple(self.iwork))

    def _affect_work_2D_X(self, resol):
        self.iwork[1][...] = np.indices((resol[1],))[0].astype(
            PARMES_INDEX)[np.newaxis, :]
        return (self.work[0], self.work[1], tuple(self.iwork))

    def _affect_work_2D_Y(self, resol):
        self.iwork[0][...] = np.indices((resol[0],))[0].astype(
            PARMES_INDEX)[:, np.newaxis]
        return (self.work[0], self.work[1], tuple(self.iwork))

    def _affect_work_3D_X(self, resol):
        self.iwork[1][...] = np.indices((resol[1],))[0].astype(
            PARMES_INDEX)[np.newaxis, :, np.newaxis]
        self.iwork[2][...] = np.indices((resol[2],))[0].astype(
            PARMES_INDEX)[np.newaxis, np.newaxis, :]
        return (self.work[0], self.work[1], tuple(self.iwork))

    def _affect_work_3D_Y(self, resol):
        self.iwork[0][...] = np.indices((resol[0],))[0].astype(
            PARMES_INDEX)[:, np.newaxis, np.newaxis]
        self.iwork[2][...] = np.indices((resol[2],))[0].astype(
            PARMES_INDEX)[np.newaxis, np.newaxis, :]
        return (self.work[0], self.work[1], tuple(self.iwork))

    def _affect_work_3D_Z(self, resol):
        self.iwork[0][...] = np.indices((resol[0],))[0].astype(
            PARMES_INDEX)[:, np.newaxis, np.newaxis]
        self.iwork[1][...] = np.indices((resol[1],))[0].astype(
            PARMES_INDEX)[np.newaxis, :, np.newaxis]
        return (self.work[0], self.work[1], tuple(self.iwork))

    def __call__(self, ppos, pscal, result):
        """
        Remesh particles at position p_pos with scalar p_scal along
        direction d.

        @param p_pos : particle position
        @param p_scal : particle scalar
        @param result : a predefined list of numpy arrays to solve the result
        @return remeshed scalar on grid
        """
        d = self.dir
        mesh = self.topo.mesh
        resolution = self.topo.globalMeshResolution
        origin = self.topo.domain.origin
        dx = mesh.space_step
        tmp, i_y, index = self._affect_working_arrays(mesh.resolution)

        floor = result  # use res array as working array
        floor[...] = (ppos - origin[d]) / dx[d]
        i_y[...] = floor
        floor[...] = np.floor(floor)
        i_y[...] -= floor

        # Gobal indices
        index[d][...] = (floor.astype(PARMES_INDEX) - self._shift) \
            % (resolution[d] - 1)
        result[...] = 0.  # reset res array (no more uses to floor variable)
        for w_id, w in enumerate(self._weights):
            if w_id > 0:
                index[d][...] = (index[d] + 1) % (resolution[d] - 1)
            tmp[...] = w(i_y, pscal)
            for i in xrange(mesh.resolution[d]):
                sl = self.slice_i_along_d(i, d)
                index_sl = tuple([ind[sl] for ind in index])
                result[index_sl] = result[index_sl] + tmp[sl]

        return result


class L2_1(Remeshing):
    """Remshing with L2_1 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 1
        self._weights = [
            lambda y, s: s * ((y * (y * (-y + 2.) - 1.)) / 2.),
            lambda y, s: s * ((y * y * (3. * y - 5.) + 2.) / 2.),
            lambda y, s: s * ((y * (y * (-3. * y + 4.) + 1.)) / 2.),
            lambda y, s: s * ((y * y * (y - 1.)) / 2.)
            ]


class L2_2(Remeshing):
    """Remshing with L2_2 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 1
        self._weights = [
            lambda y, s: s * ((y * (y * (y * (y * (2. * y - 5.) + 3.) + 1.) - 1.)) / 2.),
            lambda y, s: s * ((y * y * (y * (y * (-6. * y + 15.) - 9.) - 2.) + 2.) / 2.),
            lambda y, s: s * ((y * (y * (y * (y * (6. * y - 15.) + 9.) + 1.) + 1.)) / 2.),
            lambda y, s: s * ((y * y * y * (y * (-2. * y + 5.) - 3.)) / 2.)
            ]


class L2_3(Remeshing):
    """Remshing with L2_3 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 1
        self._weights = [
            lambda y, s: s * ((y * (y * (y * y * (y * (y * (-6. * y + 21.) - 25.) + 10.) + 1.) - 1.)) / 2.),
            lambda y, s: s * ((y * y * (y * y * (y * (y * (18. * y - 63.) + 75.) - 30.) - 2.) + 2.) / 2.),
            lambda y, s: s * ((y * (y * (y * y * (y * (y * (-18. * y + 63.) - 75.) + 30.) + 1.) + 1.)) / 2.),
            lambda y, s: s * ((y * y * y * y * (y * (y * (6. * y - 21.) + 25.) - 10.)) / 2.)
            ]


class L2_4(Remeshing):
    """Remshing with L2_4 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 1
        self._weights = [
            lambda y, s: s * ((y * (y * (y * y * y * (y * (y * (y * (20. * y - 90.) + 154.) - 119.) + 35.) + 1.) - 1.)) / 2.),
            lambda y, s: s * ((y * y * (y * y * y * (y * (y * (y * (-60. * y + 270.) - 462.) + 357.) - 105.) - 2.) + 2.) / 2.),
            lambda y, s: s * ((y * (y * (y * y * y * (y * (y * (y * (60. * y - 270.) + 462.) - 357.) + 105.) + 1.) + 1.)) / 2.),
            lambda y, s: s * ((y * y * y * y * y * (y * (y * (y * (-20. * y + 90.) - 154.) + 119.) - 35.)) / 2.)
            ]


class L4_2(Remeshing):
    """Remshing with L4_2 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 2
        self._weights = [
            lambda y, s: s * y * (y * (y * (y * (13. - 5. * y) - 9.) - 1.) + 2.) / 24.,
            lambda y, s: s * y * (y * (y * (y * (25. * y - 64.) + 39.) + 16.) - 16.) / 24.,
            lambda y, s: s * (y * y * (y * (y * (126. - 50. * y) - 70.) - 30.) / 24. + 1.),
            lambda y, s: s * y * (y * (y * (y * (50. * y - 124.) + 66.) + 16.) + 16.) / 24.,
            lambda y, s: s * y * (y * (y * (y * (61. - 25. * y) - 33.) - 1.) - 2.) / 24.,
            lambda y, s: s * y * y * y * (y * (5. * y - 12.) + 7.) / 24.
            ]


class L4_3(Remeshing):
    """Remshing with L4_3 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 2
        self._weights = [
            lambda y, s: s * (2. + (-1. + (-2. + (-22. + (58. + (-49. + 14. * y) * y) * y) * y) * y) * y) * y / 24.,
            lambda y, s: s * (-16. + (16. + (4. + (111. + (-290. + (245. - 70. * y) * y) * y) * y) * y) * y) * y / 24.,
            lambda y, s: s * (1. + (-30. + (-224. + (580. + (-490. + 140. * y) * y) * y) * y * y) * y * y / 24.),
            lambda y, s: s * (16. + (16. + (-4. + (226. + (-580. + (490. - 140. * y) * y) * y) * y) * y) * y) * y / 24.,
            lambda y, s: s * (-2. + (-1. + (2. + (-114. + (290. + (-245. + 70. * y) * y) * y) * y) * y) * y) * y / 24.,
            lambda y, s: s * (23. + (-58. + (49. - 14. * y) * y) * y) * y * y * y * y / 24.
            ]


class L4_4(Remeshing):
    """Remshing with L4_4 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 2
        self._weights = [
            lambda y, s: s * (2. + (-1. + (-2. + (1. + (-80. + (273. + (-354. + (207. - 46. * y) * y) * y) * y) * y) * y) * y) * y) * y / 24.,
            lambda y, s: s * (-16. + (16. + (4. + (-4. + (400. + (-1365. + (1770. + (-1035. + 230. * y) * y) * y) * y) * y) * y) * y) * y) * y / 24.,
            lambda y, s: s * (1. + (-30. + (6. + (-800. + (2730. + (-3540. + (2070. - 460. * y) * y) * y) * y) * y) * y * y) * y * y / 24.),
            lambda y, s: s * (16. + (16. + (-4. + (-4. + (800. + (-2730. + (3540. + (-2070. + 460. * y) * y) * y) * y) * y) * y) * y) * y) * y / 24.,
            lambda y, s: s * (-2. + (-1. + (2. + (1. + (-400. + (1365. + (-1770. + (1035. - 230. * y) * y) * y) * y) * y) * y) * y) * y) * y / 24.,
            lambda y, s: s * (80. + (-273. + (354. + (-207. + 46. * y) * y) * y) * y) * y * y * y * y * y / 24.
            ]


class M8Prime(Remeshing):
    """Remshing with M8Prime kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 3
        self._weights = [
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (-10. * y + 21.) + 28.) - 105.)+ 70.) + 35.) - 56.) + 17.) / 3360.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (70. * y - 175.) - 140.) + 770.)- 560.) - 350.) + 504.) - 102.) / 3360.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (-210. * y + 609.) + 224.) - 2135.)+ 910.) + 2765.) - 2520.) + 255.) / 3360.),
            lambda y, s: s * ((y * y * (y * y * (y * y * (70. * y - 231.) + 588.) - 980.) + 604.)/ 672.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (-70. * y + 259.) - 84.) - 427.)- 182.) + 553.) + 504.) + 51.) / 672.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (210. * y - 861.) + 532.) + 770.)+ 560.) - 350.) - 504.) - 102.) / 3360.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (-70. * y + 315.) - 280.) - 105.)- 70.) + 35.) + 56.) + 17.) / 3360.),
            lambda y, s: s * ((y * y * y * y * y * (y * (10. * y - 49.) + 56.)) / 3360.)
            ]


class L6_3(Remeshing):
    """Remshing with L6_3 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 3
        self._weights = [
            lambda y, s: s * (((-12. + (4. + (15. + (140. + (-370.+ (312. - 89. * y) * y) * y) * y) * y) * y) * y) / 720.),
            lambda y, s: s * (((108. + (-54. + (-120. + (-955. +(2581. + (-2183. + 623. * y) * y) * y) * y) * y) * y ) * y) / 720.),
            lambda y, s: s * (((-180. + (180. + (65. + (950. +(-2574. + (2182. - 623. * y) * y) * y) * y) * y) * y) * y) / 240.),
            lambda y, s: s * ( 1. + ((-196. + (-959. + (2569. +(-2181. + 623. * y) * y) * y) * y * y ) * y * y) / 144.),
            lambda y, s: s * (((108. + (108. + (-39. + (976. + (-2566.+ (2180. - 623. * y) * y) * y) * y) * y) * y) * y) / 144.),
            lambda y, s: s * (((-36. + (-18. + (40. + (-995. + (2565.+ (-2179. + 623. * y) * y) * y) * y) * y) * y) * y) / 240.),
            lambda y, s: s * (((12. + (4. + (-15. + (1010. + (-2566.+ (2178. - 623. * y) * y) * y) * y) * y) * y) * y) / 720.),
            lambda y, s: s * (((-145. + (367. + (-311. + 89. * y)* y) * y) * y * y * y * y) / 720.)
            ]


class L6_4(Remeshing):
    """Remshing with L6_4 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 3
        self._weights = [
            lambda y, s: s * (-12. + (4. + (15. + (-5. + (500. + (-1718. + (2231. + (-1305. + 290. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (108. + (-54. + (-120. + (60. + (-3509. + (12027. + (-15617. + (9135. - 2030. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (-540. + (540. + (195. + (-195. + (10548. + (-36084. + (46851. + (-27405. + 6090. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (1. + (-980. + (280. + (-17605. + (60145. + (-78085. + (45675. - 10150. * y) * y) * y) * y) * y) * y * y) * y * y / 720.),
            lambda y, s: s * (540. + (540. + (-195. + (-195. + (17620. + (-60150. + (78085. + (-45675. + 10150. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (-108. + (-54. + (120. + (60. + (-10575. + (36093. + (-46851. + (27405. - 6090. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (12. + (4. + (-15. + (-5. + (3524. + (-12032. + (15617. + (-9135. + 2030. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (-503. + (1719. + (-2231. + (1305. - 290. * y) * y) * y) * y) * y * y * y * y * y / 720.
            ]


class L6_5(Remeshing):
    """Remshing with L6_5 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 3
        self._weights = [
            lambda y, s: s * (-12. + (4. + (15. + (-5. + (-3. + (1803. + (-7829. + (13785. + (-12285. + (5533. - 1006. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (108. + (-54. + (-120. + (60. + (12. + (-12620. + (54803. + (-96495. + (85995. + (-38731. + 7042. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (-540. + (540. + (195. + (-195. + (-15. + (37857. + (-164409. + (289485. + (-257985. + (116193. - 21126. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (1. + (-980. + (280. + (-63090. + (274015. + (-482475. + (429975. + (-193655. + 35210. * y) * y) * y) * y) * y) * y * y) * y * y) * y * y / 720.),
            lambda y, s: s * (540. + (540. + (-195. + (-195. + (15. + (63085. + (-274015. + (482475. + (-429975. + (193655. - 35210. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (-108. + (-54. + (120. + (60. + (-12. + (-37848. + (164409. + (-289485. + (257985. + (-116193. + 21126. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (12. + (4. + (-15. + (-5. + (3. + (12615. + (-54803. + (96495. + (-85995. + (38731. - 7042. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (-1802. + (7829. + (-13785. + (12285. + (-5533. + 1006. * y) * y) * y) * y) * y) * y * y * y * y * y * y / 720.
            ]


class L6_6(Remeshing):
    """Remshing with L6_6 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 3
        self._weights = [
            lambda y, s: s * (-12. + (4. + (15. + (-5. + (-3. + (1. + (6587. + (-34869. + (77815. + (-93577. + (63866. + (-23426. + 3604. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (108. + (-54. + (-120. + (60. + (12. + (-6. + (-46109. + (244083. + (-544705. + (655039. + (-447062. + (163982. - 25228. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (-540. + (540. + (195. + (-195. + (-15. + (15. + (138327. + (-732249. + (1634115. + (-1965117. + (1341186. + (-491946. + 75684. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (1. + (-980. + (280. + (-20. + (-230545. + (1220415. + (-2723525. + (3275195. + (-2235310. + (819910. - 126140. * y) * y) * y) * y) * y) * y) * y) * y * y) * y * y) * y * y / 720.),
            lambda y, s: s * (540. + (540. + (-195. + (-195. + (15. + (15. + (230545. + (-1220415. + (2723525. + (-3275195. + (2235310. + (-819910. + 126140. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (-108. + (-54. + (120. + (60. + (-12. + (-6. + (-138327. + (732249. + (-1634115. + (1965117. + (-1341186. + (491946. - 75684. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (12. + (4. + (-15. + (-5. + (3. + (1. + (46109. + (-244083. + (544705. + (-655039. + (447062. + (-163982. + 25228. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
            lambda y, s: s * (-6587. + (34869. + (-77815. + (93577. + (-63866. + (23426. - 3604. * y) * y) * y) * y) * y) * y) * y * y * y * y * y * y * y / 720.
            ]


class L8_4(Remeshing):
    """Remshing with L8_4 kernel."""
    def __init__(self, dim, topo, d, work, iwork):
        Remeshing.__init__(self, dim, topo, d, work, iwork)
        self._shift = 4
        self._weights = [
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (y * (y * (-3569. * y + 16061.) - 27454.) + 21126.) - 6125.) + 49.) - 196.) - 36.) + 144.)) / 40320.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (y * (y * (32121. * y - 144548.) + 247074.) - 190092.) + 55125.) - 672.) + 2016.) + 512.) - 1536.)) / 40320.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (y * (y * (-128484. * y + 578188.) - 988256.) + 760312.) - 221060.) + 4732.) - 9464.) - 4032.) + 8064.)) / 40320.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (y * (y * (299796. * y - 1349096.) + 2305856.) - 1774136.) + 517580.) - 13664.) + 13664.) + 32256.) - 32256.)) / 40320.),
            lambda y, s: s * ((y * y * (y * y * (y * (y * (y * (y * (-449694. * y + 2023630.) - 3458700.) + 2661540.) - 778806.) + 19110.) - 57400.) + 40320.) / 40320.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (y * (y * (449694. * y - 2023616.) + 3458644.) - 2662016.) + 780430.) - 13664.) - 13664.) + 32256.) + 32256.)) / 40320.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (y * (y * (-299796. * y + 1349068.) - 2305744.) + 1775032.) - 520660.) + 4732.) + 9464.) - 4032.) - 8064.)) / 40320.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (y * (y * (128484. * y - 578168.) + 988176.) - 760872.) + 223020.) - 672.) - 2016.) + 512.) + 1536.)) / 40320.),
            lambda y, s: s * ((y * (y * (y * (y * (y * (y * (y * (y * (-32121. * y + 144541.) - 247046.) + 190246.) - 55685.) + 49.) + 196.) - 36.) - 144.)) / 40320.),
            lambda y, s: s * ((y * y * y * y * y * (y * (y * (y * (3569. * y - 16060.) + 27450.) - 21140.) + 6181.)) / 40320.)
            ]
