# -*- coding: utf-8 -*-
import parmepy as pp
import numpy as np
from parmepy.fields.continuous import Field
from parmepy.mpi.topology import Cartesian
import parmepy.numerics.differential_operations as diffop
import parmepy.tools.numpywrappers as npw
import math as m
pi = m.pi
cos = np.cos
sin = np.sin


## Function to compute TG velocity
def computeVel(res, x, y, z, t):
    res[0][...] = sin(x) * cos(y) * cos(z)
    res[1][...] = - cos(x) * sin(y) * cos(z)
    res[2][...] = 0.
    return res


## Function to compute reference vorticity
def computeVort(res, x, y, z, t):
    res[0][...] = - cos(x) * sin(y) * sin(z)
    res[1][...] = - sin(x) * cos(y) * sin(z)
    res[2][...] = 2. * sin(x) * sin(y) * cos(z)
    return res


def analyticalDivWV(res, x, y, z, t):
    res[0][...] = - sin(y) * cos(y) * sin(z) * cos(z) * \
        (- sin(x) * sin(x) + cos(x) * cos(x)) + \
        2. * cos(x) * cos(x) * sin(z) * cos(z) * sin(y) * cos(y)
    res[1][...] = - 2. * cos(y) * cos(y) * sin(z) * cos(z) * sin(x) * cos(x) +\
        sin(x) * cos(x) * sin(z) * cos(z) * (cos(y) * cos(y) - sin(y) * sin(y))
    res[2][...] = 0.
    return res


def analyticalGradVxW(res, x, y, z, t):
    res[0][...] = - sin(y) * cos(y) * sin(z) * cos(z)
    res[1][...] = sin(x) * cos(x) * sin(z) * cos(z)
    res[2][...] = 0.
    return res


def analyticalDivStressTensor(res, x, y, z, t):
    res[0][...] = - 3. * sin(x) * cos(y) * cos(z)
    res[1][...] = 3. * cos(x) * sin(y) * cos(z)
    res[2][...] = 0.
    return res


nb = 65
box = pp.Box(3, length=[2.0 * pi, 2.0 * pi, 2.0 * pi])
nbElem = [nb] * 3
topo = Cartesian(box, box.dimension, nbElem,
                 ghosts=[2, 2, 2])
velo = Field(domain=box, formula=computeVel,
             name='Velocity', isVector=True)
vorti = Field(domain=box, formula=computeVort,
              name='Vorticity', isVector=True)
velo.discretize(topo)
vorti.discretize(topo)
velo.initialize(topo=topo)
vorti.initialize(topo=topo)
wd = vorti.discreteFields[topo]
vd = velo.discreteFields[topo]
ind = topo.mesh.iCompute


def testCurl():
    memshape = vd.data[0].shape
    result = [npw.zeros(memshape) for i in xrange(3)]
    lwork = diffop.Curl.getWorkLengths()
    work = [npw.zeros(memshape) for i in xrange(lwork)]
    curlOp = diffop.Curl(topo, work)
    result = curlOp(vd.data, result)

    for i in xrange(3):
        assert np.allclose(wd.data[i][ind], result[i][ind])


def testDivWV():
    # Reference field
    ref = Field(domain=box, formula=analyticalDivWV,
                name='Analytical', isVector=True)
    ref.discretize(topo)
    ref.initialize(topo=topo)
    refd = ref.discreteFields[topo]
    memshape = vd.data[0].shape
    # Div operator
    lwork = diffop.DivWV.getWorkLengths()
    work = [npw.zeros(memshape) for i in xrange(lwork)]
    divOp = diffop.DivWV(topo, work)
    result = [npw.zeros(memshape) for i in xrange(3)]
    result = divOp(vd.data, wd.data, result)

    # Numerical VS analytical
    Lx = box.length[0]
    errX = (Lx / (nb - 1)) ** 4
    for i in xrange(3):
        assert np.allclose(refd[i][ind], result[i][ind], rtol=errX)


def testGradVxW():
    # Reference field
    ref = Field(domain=box, formula=analyticalGradVxW,
                name='Analytical', isVector=True)
    ref.discretize(topo)
    ref.initialize(topo=topo)
    refd = ref.discreteFields[topo]
    memshape = vd.data[0].shape
    lwork = diffop.GradVxW.getWorkLengths()
    work = [npw.zeros(memshape) for i in xrange(lwork)]
    gradOp = diffop.GradVxW(topo, work)
    result = [npw.zeros(memshape) for i in xrange(3)]
    diag = npw.zeros(2)
    result, diag = gradOp(vd.data, wd.data, result, diag)

    # Numerical VS analytical
    Lx = box.length[0]
    errX = (Lx / (nb - 1)) ** 4
    for i in xrange(3):
        assert np.allclose(refd[i][ind], result[i][ind], rtol=errX)
