# -*- coding: utf-8 -*-
"""
@file operator/adapt_timestep.py

Definition of the adaptative time step according to the flow fields.

"""
from parmepy.constants import debug
from parmepy.methods_keys import TimeIntegrator, SpaceDiscretisation,\
    dtCrit
from parmepy.numerics.finite_differences import FD_C_4
from parmepy.operator.discrete.adapt_timestep import AdaptTimeStep_D
from parmepy.operator.continuous import Operator
from parmepy.fields.variable_parameter import VariableParameter
import numpy as np
import parmepy.default_methods as default


class AdaptTimeStep(Operator):
    """
    The adaptative Time Step is computed according
    to the following expression :
    dt_adapt = min (dt_advection, dt_stretching, dt_cfl)
    """

    @debug
    def __init__(self, velocity, vorticity, resolutions, dt_adapt,
                 method=None, topo=None, ghosts=None, io_params=None,
                 task_id=None, comm=None, time_range=None,
                 lcfl=0.125, cfl=0.5):
        """
        Create a timeStep-evaluation operator from given
        velocity and vorticity variables.

        Note : If cfl is None, the computation of the adaptative time step
        is only based on dt_advection and dt_stretching, taking the minimum
        value of the two.

        @param velocity field
        @param vorticity field
        @param resolutions : grid resolution of velocity and vorticity
        @param dt_adapt : adaptative timestep variable
        @param method : solving method
        default = finite differences, 4th order in space for gradU evaluation
        (used in dt_advection computation)
        and Runge-Kutta 3 in time (used in dt_stretching computation) and
        criterion for advection time step evaluation.
        @param topo : a predefined topology to discretize velocity/vorticity
        @param ghosts : number of ghosts points. Default depends on the method.
        Automatically computed if not set.
        @param time_range : [start, end] use to define a 'window' in which
        the current operator is applied. Outside start-end, this operator
        has no effect. Start/end are iteration numbers.
        Default = [2, endofsimu]

        """
        if method is None:
            method = default.ADAPT_TIME_STEP
        Operator.__init__(self, [velocity, vorticity], method, topo=topo,
                          ghosts=ghosts, task_id=task_id, comm=comm)
        ## velocity variable (vector)
        self.velocity = velocity
        ## vorticity variable (vector)
        self.vorticity = vorticity
        ## Grid resolution for each variable (dictionnary)
        self.resolutions = resolutions
        ## adaptative time step variable ("variable" object)
        self.dt_adapt = dt_adapt
        assert isinstance(self.dt_adapt, VariableParameter)
        # Check if 'dt' key is present in dt_adapt dict
        assert 'dt' in self.dt_adapt.data
        ## Numerical methods for time and space discretization
        self.method = method
        self.io_params = io_params
        assert SpaceDiscretisation in self.method.keys()
        assert TimeIntegrator in self.method.keys()
        if not dtCrit in self.method.keys():
            self.method[dtCrit] = 'vort'

        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]
        self.time_range = time_range
        self.lcfl, self.cfl = lcfl, cfl

    def discretize(self):
        if self.method[SpaceDiscretisation] is FD_C_4:
            nbGhosts = 2
        else:
            raise ValueError("Unknown method for space discretization of the\
                time step-evaluation operator.")

        if self._predefinedTopo is not None:
            assert (self._predefinedTopo.ghosts >= nbGhosts).all()
            topo = self._predefinedTopo
            for v in self.variables:
                self.discreteFields[v] = v.discretize(topo)

        else:
            # Variables discretization
            if self.ghosts is not None:
                self.ghosts[self.ghosts < nbGhosts] = nbGhosts
            else:
                self.ghosts = np.ones((self.domain.dimension)) * nbGhosts

            for v in self.variables:
                topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                       self.resolutions[v],
                                                       ghosts=self.ghosts,
                                                       comm=self._comm)
                self.discreteFields[v] = v.discretize(topo)

    @debug
    def setUp(self):
        self.discreteOperator =\
            AdaptTimeStep_D(self.discreteFields[self.velocity],
                            self.discreteFields[self.vorticity],
                            self.dt_adapt, method=self.method,
                            time_range=self.time_range,
                            io_params=self.io_params,
                            lcfl=self.lcfl,
                            cfl=self.cfl)

        self.discreteOperator.setUp()
        self._isUpToDate = True
