"""
@file advection.py

Advection of a field.
"""
from parmepy.constants import debug, np, PARMES_INDEX, S_DIR
from parmepy.operator.continuous import Operator
from parmepy.methods_keys import Scales, TimeIntegrator, Interpolation,\
    Remesh, Support, Splitting, MultiScale
from parmepy.numerics.remeshing import L2_1
from parmepy.fields.continuous import Field
from parmepy.operator.redistribute import Redistribute
from parmepy.operator.advectionDir import AdvectionDir
import parmepy.default_methods as default


class Advection(Operator):
    """
    Advection of a field,
    \f{eqnarray*}
    X = Op(X,velocity)
    \f} for :
    \f{eqnarray*}
    \frac{\partial{X}}{\partial{t}} + velocity.\nabla X = 0
    \f}
    Note : we assume incompressible flow.

    Computations are performed within a dimensional splitting as folows:
      - 2nd order:
        - X-dir, half time step
        - Y-dir, half time step
        - Z-dir, full time step
        - Y-dir, half time step
        - X-dir, half time step
      - 2nd order full half-steps:
        - X-dir, half time step
        - Y-dir, half time step
        - Z-dir, half time step
        - Z-dir, half time step
        - Y-dir, half time step
        - X-dir, half time step
      - 1st order g:
        - X-dir, half time step
        - Y-dir, half time step
        - Z-dir, half time step

    """

    @debug
    def __init__(self, velocity, advectedFields, resolutions,
                 method=None,
                 topo=None, ghosts=None, comm=None,
                 task_id=None, **other_config):
        """
        Create a Transport operator from given variables velocity and scalar.

        @param velocity : velocity variable.
        @param advectedFields : Advected fields (may be a list of Fields).
        @param resolutions : list of resolutions (one per variable)
        @param method : Method used
        @param splittingConfig : Dimensional splitting configuration
        (default 'o2')
        @param other_config : Method specific parameters
        @param topo : a predefined topology to discretize variables
        """
        if method is None:
            method = default.ADVECTION
        v = [velocity]
        if isinstance(advectedFields, list):
            self.advectedFields = advectedFields
        else:
            self.advectedFields = [advectedFields]
        for adF in self.advectedFields:
            v.append(adF)
        vars_str = "_("
        for vv in self.advectedFields:
            vars_str += vv.name + ","
        Operator.__init__(self, v, method, topo=topo,
                          ghosts=ghosts, name_suffix=vars_str[0:-1] + ')',
                          task_id=task_id, comm=comm)
        ## Transport velocity
        self.velocity = velocity
        ## Transported fields
        self.output = self.advectedFields
        self.input = [var for var in self.variables]
        ## Global resolutions of the fields.
        self.resolutions = resolutions
        self._isMultiScale = False
        v_resol = self.resolutions[self.velocity]
        if v_resol != self.resolutions[self.advectedFields[0]]:
            self._isMultiScale = True
        if self._isMultiScale and not MultiScale in method.keys():
            print ("Using default mutiscale interpolation : L2_1")
            self.method[MultiScale] = L2_1
        ## Extra parameters (depend on the method)
        self.config = other_config
        if Scales in self.method.keys():
            self._isSCALES = True
            if not Splitting in self.method.keys():
                self.method[Splitting] = 'strang'
        else:
            self._isSCALES = False
            assert TimeIntegrator in self.method.keys()
            assert Interpolation in self.method.keys()
            assert Remesh in self.method.keys()
            assert Support in self.method.keys()
            if not Splitting in self.method.keys():
                self.method[Splitting] = 'o2'
        self._old_dir = 0
        self.splitting = []
        self._dim = self.velocity.dimension
        self.advecDir = None
        if not self._isSCALES:
            particles_advectedFields = [
                Field(adF.domain, name="Particle_AdvectedFields",
                      isVector=adF.isVector)
                for adF in self.advectedFields]
            if self.method[Support].find('gpu_1k') >= 0:
                particles_positions = None
            else:
                particles_positions = \
                    Field(self.advectedFields[0].domain,
                          name="Particle_Position",  isVector=False
                          )

            # Directional continuous Advection operators
            self.advecDir = [None] * self._dim
            for i in xrange(self._dim):
                self.advecDir[i] = AdvectionDir(
                    self.velocity, self.advectedFields, i,
                    particles_advectedFields, particles_positions,
                    self.resolutions,
                    method=self.method, topo=self._predefinedTopo,
                    ghosts=self.ghosts, isMultiScale=self._isMultiScale,
                    comm=self._comm,
                    task_id=task_id, name_suffix=vars_str[0:-1] + ')',
                    **self.config)

        # function to switch between CPU or GPU setup.
        self._my_setUp = None
    @debug
    def discretize(self):
        """
        Discretisation according to the chosen method.
        Available methods :
        - 'scales' : SCALES fortran routines (3d only, list of vector
        and/or scalar)
          - 'p_O2' : order 4 method, corrected to allow large CFL number,
          untagged particles
          - 'p_O4' : order 4 method, corrected to allow large CFL number,
          untagged particles
          - 'p_L2' : limited and corrected lambda 2
          - 'p_M4' : Lambda_2,1 (=M'4) 4 point formula
          - 'p_M6' (default) : Lambda_4,2 (=M'6) 6 point formula
          - 'p_M8' : M8prime formula
          - 'p_44' : Lambda_4,4 formula
          - 'p_64' : Lambda_6,4 formula
          - 'p_66' : Lambda_6,6 formula
          - 'p_84' : Lambda_8,4 formula
        - 'gpu' : OpenCL kernels (2d and 3d, single field, scalar or vector)
          - Kernels versions:
            - '1k' : Single OpenCL kernel for advection and remeshing
            - '2k' : Separate kernels
          - Integration method:
            - 'rk2' : Runge Kutta 2nd order advection
            - 'rk4' : Runge Kutta 4th order advection
          - remeshing formula:
            - 'm4prime' : = 'l2_1'
            - 'l2_1' : Labmda2,1 : (=M'4) 4 point formula, C1 regularity
            - 'l2_2' : Labmda2,2 : 4 point formula, C2 regularity
            - 'm6prime' : = 'l4_2'
            - 'l4_2' : Labmda4,2 : (=M'6) 6 point formula, C2 regularity
            - 'l4_3' : Labmda4,3 : 6 point formula, C3 regularity
            - 'l4_4' : Labmda4,4 : 6 point formula, C4 regularity
            - 'l6_3' : Labmda6,3 : 8 point formula, C3 regularity
            - 'l6_4' : Labmda6,4 : 8 point formula, C4 regularity
            - 'l6_5' : Labmda6,5 : 8 point formula, C5 regularity
            - 'l6_6' : Labmda6,6 : 8 point formula, C6 regularity
            - 'l8_4' : Labmda8,4 : 10 point formula, C4 regularity
            - 'm8prime' : M8prime formula
        - other : Pure python (2d and 3d, list of vector and/or scalar)
          - Integration method:
            - 'rk2' : Runge Kutta 2nd order advection
            - 'rk4' : Runge Kutta 4th order advection
          - remeshing formula:
            - 'm4prime' : = 'l2_1'
            - 'l2_1' : Labmda2,1 : (=M'4) 4 point formula, C1 regularity
            - 'l2_2' : Labmda2,2 : 4 point formula, C2 regularity
            - 'm6prime' : = 'l4_2'
            - 'l4_2' : Labmda4,2 : (=M'6) 6 point formula, C2 regularity
            - 'l4_3' : Labmda4,3 : 6 point formula, C3 regularity
            - 'l4_4' : Labmda4,4 : 6 point formula, C4 regularity
            - 'l6_3' : Labmda6,3 : 8 point formula, C3 regularity
            - 'l6_4' : Labmda6,4 : 8 point formula, C4 regularity
            - 'l6_5' : Labmda6,5 : 8 point formula, C5 regularity
            - 'l6_6' : Labmda6,6 : 8 point formula, C6 regularity
            - 'l8_4' : Labmda8,4 : 10 point formula, C4 regularity
            - 'm8prime' : M8prime formula
        """
        # --- Advection solver from SCALES ---
        if self._isSCALES:
            if not self._dim == 3:
                raise ValueError("Scales Advection not implemented in 2D.")
            if self.ghosts is not None:
                if (self.ghosts > 0).any():
                    raise ValueError("Scales Advection not\
                    yet implemented with ghosts points.")
            # - Scales imports -
            from parmepy.f2py import scales2py as scales

            # - Extract order form self.method (default p_M6) -
            order = None
            for o in ['p_O2', 'p_O4', 'p_L2',
                      'p_M4', 'p_M6', 'p_M8',
                      'p_44', 'p_64', 'p_66', 'p_84']:
                if self.method[Scales].find(o) >= 0:
                    order = o
            if order is None:
                print ('Unknown advection method, turn to default (p_M6).')
                order = 'p_M6'
            # - Extract splitting form self.method (default strang) -
            splitting = 'strang'
            for s in ['classic', 'strang', 'particle']:
                if self.method[Splitting].find(s) >= 0:
                    splitting = s

            # - Create the topologies (get param from scales) -
            # Scales nbcells equals resolutions - 1
            nbcells = np.asarray(self.resolutions[self.advectedFields[0]],
                                 dtype=PARMES_INDEX) - 1
            if self._predefinedTopo is not None:
                main_size = self._predefinedTopo.size
                comm = self._predefinedTopo.comm
            elif self._comm is not None:
                main_size = self._comm.Get_size()
                comm = self._comm
            else:
                from parmepy.mpi.main_var import main_size
                from parmepy.mpi.main_var import main_comm as comm
            topodims = [1, 1, main_size]
            scalesres, scalesoffset = \
                scales.init_advection_solver(nbcells,
                                             self.domain.length,
                                             topodims, comm.py2f(),
                                             order=order,
                                             dim_split=splitting)
            # Use same topodims as scales to create Cartesian topology
            # in order to discretize our fields
            if self._predefinedTopo is not None:
                topo = self._predefinedTopo
                assert (topo.shape == topodims).all(), 'input topology is\
                    not scales compliant.'
                for v in self.variables:
                    self.discreteFields[v] = v.discretize(topo)
            else:
                for v in self.variables:
                    # the topology for v ...
                    topo = \
                        self.domain.getOrCreateTopology(self._dim,
                                                        self.resolutions[v],
                                                        topodims,
                                                        precomputed=True,
                                                        offset=scalesoffset,
                                                        localres=scalesres,
                                                        ghosts=self.ghosts,
                                                        comm=self._comm)
                    # ... and the corresponding discrete field
                    self.discreteFields[v] = v.discretize(topo)
            if self._isMultiScale:
                self.config['isMultiscale'] = self._isMultiScale
                v_shape = np.asarray(self.resolutions[self.velocity],
                                     dtype=PARMES_INDEX) - 1
                scales.init_multiscale(v_shape[0], v_shape[1], v_shape[2],
                                       self.method[MultiScale])
            self._my_setUp = self.setUp_Scales

        # --- GPU or pure-python advection ---
        else:
            for i in xrange(self._dim):
                self.advecDir[i].discretize()
            self.discreteFields = self.advecDir[0].discreteFields
            self._my_setUp = self.setUp_Python

    @staticmethod
    def getWorkLengths(method=None, domain_dim=None):
        """
        Return the length of working arrays lists required
        for advction discrete operator, depending on :
        - the time integrator (RK2, ...)
        - the interpolation (which depends on domain dimension)
        - the remeshing (which depends on domain dimension)
        @param method : the dict of parameters for the operator.
        Default = parmepy.default_methods.ADVECTION
        """
        if method is None:
            method = default.ADVECTION
        assert Interpolation in method,\
            'An interpolation is required for the advection method.'
        assert TimeIntegrator in method,\
            'A time integrator is required for the advection method.'
        assert Remesh in method,\
            'A remesh is required for the advection method.'
        tw = method[TimeIntegrator].getWorkLengths(1)
        iw, iiw = method[Interpolation].getWorkLengths(domain_dim=domain_dim)
        rw, riw = method[Remesh].getWorkLengths(domain_dim=domain_dim)
        return max(tw + iw, rw), max(iiw, riw)

    def setWorks(self, rwork=None, iwork=None):
        if rwork is None:
            rwork = []
        if iwork is None:
            iwork = []
        if not self._isSCALES:
            for i in xrange(self._dim):
                self.advecDir[i].setWorks(rwork, iwork)

    def setUp(self):
        if not self._isUpToDate:
            self._my_setUp()

    def setUp_Scales(self):
        advectedDiscreteFields = [self.discreteFields[f]
                                  for f in self.advectedFields]
        # - Create the discreteOperator from the
        # list of discrete fields -
        from parmepy.operator.discrete.scales_advection import \
            ScalesAdvection
        self.discreteOperator = ScalesAdvection(
            self.discreteFields[self.velocity],
            advectedDiscreteFields, method=self.method,
            **self.config)

        # -- Final set up --
        self.discreteOperator.setUp()
        self._isUpToDate = True

    def setUp_Python(self):
        for i in xrange(self._dim):
            self.advecDir[i].setUp()
        # If topologies differs between directions,
        # one need Redistribute
        # operators
        main_size = self.advecDir[0].discreteFields[
            self.velocity].topology.size
        if main_size > 1:
            # Build bridges
            self.bridges = self._dim * [None]
            if main_size > 1:
                for dfrom in xrange(self.domain.dimension):
                    self.bridges[dfrom] = self._dim * [None]
                    for dto in xrange(self._dim):
                        if dfrom == dto:
                            self.bridges[dfrom][dto] = None
                        else:
                            self.bridges[dfrom][dto] = Redistribute(
                                self.advectedFields,
                                self.advecDir[dfrom],
                                self.advecDir[dto],
                                name_suffix=str(dfrom)+'_'+str(dto))
                            self.bridges[dfrom][dto].setUp()

        # Splitting configuration
        if self.method[Splitting] == 'o2_FullHalf':
            ## Half timestep in all directions
            [self.splitting.append((i, 0.5))
             for i in xrange(self._dim)]
            [self.splitting.append((self._dim - 1 - i, 0.5))
             for i in xrange(self._dim)]
        elif self.method[Splitting] == 'o1':
            [self.splitting.append((i, 1.)) for i in xrange(self._dim)]
        elif self.method[Splitting] == 'x_only':
            self.splitting.append((0, 1.))
        elif self.method[Splitting] == 'y_only':
            self.splitting.append((1, 1.))
        elif self.method[Splitting] == 'z_only':
            self.splitting.append((2, 1.))
        elif self.method[Splitting] == 'o2':
            ## Half timestep in all directions but last
            [self.splitting.append((i, 0.5))
             for i in xrange(self._dim - 1)]
            self.splitting.append((self._dim - 1, 1.))
            [self.splitting.append((self._dim - 2 - i, 0.5))
             for i in xrange(self._dim - 1)]
        else:
            raise ValueError('Unknown splitting configuration:' +
                             self.method[Splitting])

        if self.method[Support].find('gpu') >= 0:
            splitting_nbSteps = len(self.splitting)
            for d in xrange(self.domain.dimension):
                dOp = self.advecDir[d].discreteOperator
                assert len(dOp.exec_list) == splitting_nbSteps, \
                    "Discrete operator execution " + \
                    "list and splitting steps sizes must be equal " + \
                    str(len(dOp.exec_list)) + " != " + \
                    str(splitting_nbSteps)

        if main_size > 1:
            self.apply = self._apply_Comm
        else:
            self.apply = self._apply_noComm
        self._isUpToDate = True

        if self.method[Support].find('gpu') >= 0:
            s = ""
            device_id = self.advecDir[d].discreteOperator.cl_env._device_id
            gpu_comm = self.advecDir[d].discreteOperator.cl_env.gpu_comm
            gpu_rank = gpu_comm.Get_rank()
            if gpu_rank == 0:
                s += "=== OpenCL buffers allocated"
                s += " on Device:{0} ===\n".format(device_id)
                s += "Global memory used:\n"
            total_gmem = 0
            for d in xrange(self.domain.dimension):
                g_mem_d = 0
                for df in self.advecDir[d].discreteOperator.variables:
                    if not df.gpu_allocated:
                        df.allocate()
                        g_mem_df = gpu_comm.allreduce(df.mem_size)
                        g_mem_d += g_mem_df
                if gpu_rank == 0:
                    s += " Advection" + S_DIR[d] + ": {0:9d}".format(g_mem_d)
                    s += "Bytes ({0:5d} MB)\n".format(g_mem_d / (1024 ** 2))
                total_gmem += g_mem_d
            if gpu_rank == 0:
                s += " Total      : {0:9d}".format(total_gmem)
                s += "Bytes ({0:5d} MB)\n".format(total_gmem / (1024 ** 2))
                s += "Local memory used:\n"
            total_lmem = 0
            for d in xrange(self.domain.dimension):
                l_mem_d = gpu_comm.allreduce(
                    self.advecDir[d].discreteOperator.size_local_alloc)
                if gpu_rank == 0:
                    s += " Advection" + S_DIR[d] + ": {0:9d}".format(l_mem_d)
                    s += "Bytes ({0:5d} MB)\n".format(l_mem_d / (1024 ** 2))
                total_lmem += l_mem_d
            if gpu_rank == 0:
                s += " Total      : {0:9d}".format(total_lmem) + "Bytes"
                print (s)

    @debug
    def _apply_noComm(self, simulation=None):
        """
        Apply this operator to its variables.
        @param simulation : object that describes the simulation
        parameters (time, time step, iteration number ...), see
        parmepy.problem.simulation.Simulation for details.

        Redefinition for advection. Applying a dimensional splitting.
        """
        for req in self.requirements:
            req.wait()
        for split_id, split in enumerate(self.splitting):
            self.advecDir[split[0]].apply(
                simulation, split[1], split_id, self._old_dir)
            self._old_dir = split[0]

    @debug
    def _apply_Comm(self, simulation=None):
        """
        Apply this operator to its variables.
        @param simulation : object that describes the simulation
        parameters (time, time step, iteration number ...), see
        parmepy.problem.simulation.Simulation for details.

        Redefinition for advection. Applying a dimensional splitting.
        """
        for req in self.requirements:
            req.wait()
        for split_id, split in enumerate(self.splitting):
            # Calling the redistribute operators between directions
            if not self._old_dir == split[0]:
                self.bridges[self._old_dir][split[0]].apply()
                self.bridges[self._old_dir][split[0]].wait()
            self.advecDir[split[0]].apply(
                simulation, split[1], split_id, self._old_dir)
            self._old_dir = split[0]

    @debug
    def finalize(self):
        """
        Memory cleaning.
        """
        if self._isSCALES:
            Operator.finalize(self)
        else:
            for dop in self.advecDir:
                dop.finalize()
                self.timer = self.timer + dop.timer

    def __str__(self):
        """
        Common printings for operators
        """
        shortName = str(self.__class__).rpartition('.')[-1][0:-2]
        if self._isSCALES:
            s = Operator.__str__(self)
        else:
            for i in xrange(self._dim):
                if self.advecDir[i].discreteOperator is not None:
                    s = str(self.advecDir[i].discreteOperator)
                else:
                    s = shortName + " operator. Not discretised."
        return s + "\n"
