"""
@file advectionDir.py

Advection of a field in a single direction.
"""
from parmepy.constants import debug, np, S_DIR, PARMES_INDEX
from parmepy.methods_keys import TimeIntegrator, Interpolation, Remesh, \
    Support, MultiScale
from parmepy.numerics.interpolation import Linear
from parmepy.numerics.remeshing import L2_1, L4_2, L4_4
from parmepy.operator.continuous import Operator
from parmepy.tools.timers import Timer
# To get default method values for advection:
import parmepy.default_methods as default


class AdvectionDir(Operator):
    """
    Advection of a field,
    \f{eqnarray*}
    X = Op(X,velocity)
    \f} for :
    \f{eqnarray*}
    \frac{\partial{X}}{\partial{t}} + velocity.\nabla X = 0
    \f}
    Note : we assume incompressible flow.

    Computations are performed in a given direction.
    """

    @debug
    def __init__(self, velocity, advectedFields, d, particle_fields,
                 particle_positions, resolutions,
                 method=None,
                 topo=None, ghosts=None, isMultiScale=False,
                 task_id=None, name_suffix='', comm=None,
                 **other_config):
        if method is None:
            method = default.ADVECTION
        v = [velocity]
        [v.append(f) for f in advectedFields]
        Operator.__init__(self, v, method, topo=topo,
                          ghosts=ghosts, task_id=task_id, comm=comm,
                          name_suffix=name_suffix + S_DIR[d])
        if self._predefinedTopo is not None:
            raise ValueError("User defined topology is not\
               allowed for advecDir.")

        self.output = advectedFields
        self.input = [var for var in self.variables]
        ## Transport velocity
        self.velocity = velocity
        ## direction to advect
        self.dir = d
        ## Resolutions of the fields. Dictionnary with
        ## fields as keys.
        ## Example :
        ## resolutions[velocity] may return [32,32,32].
        self.resolutions = resolutions
        self._isMultiScale = isMultiScale
        self._v_ghosts = None
        if self._isMultiScale:
            if self.method[Support].find('gpu') < 0:
                raise ValueError("Multiscale advection is not supported in "
                                 "Python yet, user should use Scales or GPU.")
            if not MultiScale in method.keys():
                method[MultiScale] = L2_1
            if method[MultiScale] == Linear:
                self._v_ghosts = np.array([1, ] * self.domain.dimension,
                                          dtype=PARMES_INDEX)
            elif method[MultiScale] == L2_1:
                self._v_ghosts = np.array([2, ] * self.domain.dimension,
                                          dtype=PARMES_INDEX)
            elif method[MultiScale] == L4_2 or \
                    method[MultiScale] == L4_4:
                self._v_ghosts = np.array([3, ] * self.domain.dimension,
                                          dtype=PARMES_INDEX)
            else:
                raise ValueError("Unknown multiscale method")

        ## Extra parameters (depend on the method)
        self.config = other_config
        self.particle_fields = particle_fields
        self.particle_positions = particle_positions

    @debug
    def discretize(self):
        """
        Discretisation according to the chosen method.
        Available methods : See Advection.setUp
        """
        if self._comm is not None:
            main_size = self._comm.Get_size()
        else:
            from parmepy.mpi import main_size

        topodims = np.ones((self.domain.dimension))
        # MPI topology depends on direction
        if self.dir == self.domain.dimension - 1:
            # Cut in first dir for last dir computations
            topodims[0] = main_size
        else:
            # Cut in last dir
            topodims[-1] = main_size
        for v in self.variables:
            if v == self.velocity:
                topo = self.domain.getOrCreateTopology(
                    1, self.resolutions[v], topoResolution=topodims,
                    fixedResolution=True, ghosts=self._v_ghosts,
                    comm=self._comm)
            else:
                topo = self.domain.getOrCreateTopology(
                    1, self.resolutions[v], topoResolution=topodims,
                    fixedResolution=True, ghosts=self.ghosts,
                    comm=self._comm)
            self.discreteFields[v] = v.discretize(topo)

    @staticmethod
    def getWorkLengths(method=None, domain_dim=None):
        """
        Return the length of working arrays lists required
        for advction discrete operator, depending on :
        - the time integrator (RK2, ...)
        - the interpolation (which depends on domain dimension)
        - the remeshing (which depends on domain dimension)
        @param method : the dict of parameters for the operator.
        Default = parmepy.default_methods.ADVECTION
        """
        if method is None:
            method = default.ADVECTION
        assert Interpolation in method,\
            'An interpolation is required for the advection method.'
        assert TimeIntegrator in method,\
            'A time integrator is required for the advection method.'
        assert Remesh in method,\
            'A remesh is required for the advection method.'
        tw = method[TimeIntegrator].getWorkLengths(1)
        iw, iiw = method[Interpolation].getWorkLengths(domain_dim=domain_dim)
        rw, riw = method[Remesh].getWorkLengths(domain_dim=domain_dim)
        return max(tw + iw, rw), max(iiw, riw)

    def setWorks(self, rwork=None, iwork=None):
        if rwork is None:
            rwork = []
        if iwork is None:
            iwork = []
        self.discreteOperator.setWorks(rwork, iwork)

    def setUp(self):
        advectedDiscreteFields = [self.discreteFields[v]
                                  for v in self.variables
                                  if not v is self.velocity]
        topo = advectedDiscreteFields[0].topology
        particles_advectedDiscreteFields = [v.discretize(topo)
                                            for v in self.particle_fields]
        particles_positionsDiscreteField = None
        if self.particle_positions is not None:
            particles_positionsDiscreteField =\
                self.particle_positions.discretize(topo)

        if self.method[Support].find('gpu_2k') >= 0:
            from parmepy.gpu.gpu_particle_advection_2k \
                import GPUParticleAdvection2k as advec
        elif self.method[Support].find('gpu_1k') >= 0:
            from parmepy.gpu.gpu_particle_advection_1k \
                import GPUParticleAdvection1k as advec
        else:
            from parmepy.operator.discrete.particle_advection \
                import ParticleAdvection as advec

        self.discreteOperator = advec(
            self.discreteFields[self.velocity],
            advectedDiscreteFields, self.dir, method=self.method,
            part_position=particles_positionsDiscreteField,
            part_advectedFields=particles_advectedDiscreteFields,
            isMultiScale=self._isMultiScale,
            **self.config)

        # -- Final set up --
        self.discreteOperator.setUp()
        self._isUpToDate = True

    @debug
    def apply(self, simulation, dtCoeff, split_id, old_dir=None):
        for req in self.requirements:
            req.wait()
        self.discreteOperator.apply(
            simulation, dtCoeff, split_id, old_dir)

    @debug
    def finalize(self):
        """
        Memory cleaning.
        """
        self.timer = Timer(self, suffix=S_DIR[self.dir])
        self.discreteOperator.finalize()
        self.timer = self.timer + self.discreteOperator.timer
