"""
@file operator/continuous.py

Interface common to all continuous operators.
"""
from abc import ABCMeta, abstractmethod
from parmepy.constants import debug
import numpy as np
from parmepy.tools.timers import Timer


class Operator(object):
    """
    Abstract base class for continuous operators.

    An operator is composed of :
    - a set of continuous variables (at least one)
    - a method which defined how it would be discretized/processed
    - a discrete operator : object build using the method
    and the discretized variables.

    To each variable a 'resolution' is associated, used
    to create a topology and a discrete field.
    See details in 'real' operators (derived classes) description.
    """
    __metaclass__ = ABCMeta

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    @abstractmethod
    def __init__(self, variables, method=None, topo=None, ghosts=None,
                 name_suffix='', task_id=None, comm=None):
        """
        Build the operator.
        The only required parameter is a list of variables.
        @param variables : list of fields on which this operator will apply.
        @param method : a dictionnary of methods.
        See methods.py for authorized values.
        the operator.
        @param name : operator id.
        @param topo : a predefined topology to discretize variables
        @param ghosts : number of points in the ghost layer
        @param comm : MPI communicator to build topologies
        Warning : you cannot set both topo and ghosts.
        """
        ## List of parmepy.continuous.Fields involved in the operator.
        if isinstance(variables, list):
            self.variables = variables
        else:
            self.variables = [variables]
        ## Input variables.
        self.input = []
        ## Output variables.
        self.output = []
        ## Domain of definition.
        self.domain = self.variables[0].domain
        for v in self.variables:
            assert v.domain is self.domain, 'All variables of the operator\
                must be defined on the same domain.'
        ## Object to describe the method of discretization of this operator.
        self.discreteOperator = None
        ## bool to check if the setup function has been called for
        ## this operator
        self._isUpToDate = False
        ## A dictionnary of discreteFields associated with this operator
        ## key = continuous variable \n
        ## Example : discrete_variable = discreteFields[velocity]
        ## returns the discrete fields that corresponds to the continuous
        ## variable velocity.
        self.discreteFields = {}
        ## The method used to discretize the operator.
        ## if method is None:
        ##     method = {}
        self.method = method
        ## Number of points in the ghost layer
        if ghosts is not None:
            self.ghosts = np.asarray(ghosts)
            assert self.ghosts.size == self.domain.dimension
        else:
            self.ghosts = None

        ## A predefined topology to discretize variables.
        self._comm = comm
        self._predefinedTopo = None
        ## Check if only one of topo or comm is given by user
        assert comm is None or topo is None
        if topo is not None:
            self._predefinedTopo = topo
            if self.ghosts is not None:
                assert (self.ghosts == topo.ghosts).all(), 'topo.ghosts and\
                input for ghosts must not be different'
            else:
                self.ghosts = topo.ghosts
        self.name = self.__class__.__name__ + name_suffix
        self.task_id = task_id
        self.timer = Timer(self)
        ## Redistribute operator list that we must wait for.
        self.requirements = []
        ## time monitoring
        self.time_info = None

    @staticmethod
    def getWorkLengths(method=None, domain_dim=None):
        """
        Return the length of working arrays lists required
        for the discrete operator.
        """
        assert method is None
        assert domain_dim is None
        return 0, 0

    def setWorks(self, rwork=None, iwork=None):
        if rwork is None:
            rwork = []
        if iwork is None:
            iwork = []
        self.discreteOperator.setWorks(rwork, iwork)

    @abstractmethod
    def discretize(self):
        """
        Build (or get) topologies required for this operator
        and discretize all its variables.
        """

    def addRedistributeRequirement(self, red):
        self.requirements.append(red)

    def getRedistributeRequirement(self):
        return self.requirements

    @abstractmethod
    def setUp(self):
        """
        Last step of initialization. After this, the operator must be
        ready for apply call.

        Main step : setup for discrete operators.
        """

    @debug
    def finalize(self):
        """
        Memory cleaning.
        """
        if self.discreteOperator is not None:
            self.discreteOperator.finalize()
            self.timer = self.timer + self.discreteOperator.timer

    @debug
    def apply(self, simulation=None):
        """
        Apply this operator to its variables.
        @param simulation : object that describes the simulation
        parameters (time, time step, iteration number ...), see
        parmepy.problem.simulation.Simulation for details.
        """
        for req in self.requirements:
            req.wait()
        self.discreteOperator.apply(simulation)

    def printComputeTime(self):
        """ Time monitoring."""
        if self.discreteOperator is not None:
            self.discreteOperator.printComputeTime()
            self.time_info = self.discreteOperator.time_info
        else:
            from parmepy.mpi.main_var import main_rank
            shortName = str(self.__class__).rpartition('.')[-1][0:-2]
            s = '[' + str(main_rank) + '] ' + shortName
            s += " : operator not discretized --> no computation, time = 0."
            print (s)

    def isUp(self):
        """
        True if ready to be applied (--> setup function has
        been called succesfully)
        """
        return self._isUpToDate

    def __str__(self):
        """
        Common printings for operators
        """
        shortName = str(self.__class__).rpartition('.')[-1][0:-2]
        if self.discreteOperator is not None:
            s = str(self.discreteOperator)
        else:
            s = shortName + " operator. Not discretised."
        return s + "\n"

    def updateGhosts(self):
        """
        Update ghost points values, if any.
        """
        self.discreteOperator.updateGhosts()


class EmptyOperator(object):
    """Empty operator"""
    def __init__(self, *args, **kwargs):
        pass

    def apply(self, *args, **kwargs):
        pass
