# -*- coding: utf-8 -*-
"""
@file operator/discrete/adapt_timestep.py

Evaluation of the adaptative time step according to the flow fields.
"""

from parmepy.constants import debug
from parmepy.methods_keys import TimeIntegrator, SpaceDiscretisation,\
    dtCrit
from parmepy.numerics.finite_differences import FD_C_4
from parmepy.operator.discrete.discrete import DiscreteOperator
from parmepy.numerics.integrators.euler import Euler
from parmepy.numerics.integrators.runge_kutta2 import RK2
from parmepy.numerics.integrators.runge_kutta3 import RK3
from parmepy.numerics.integrators.runge_kutta4 import RK4
from parmepy.numerics.differential_operations import GradV
import parmepy.tools.numpywrappers as npw
from parmepy.numerics.updateGhosts import UpdateGhosts
from parmepy.mpi import MPI
from parmepy.constants import np, PARMES_MPI_REAL
import parmepy.tools.io_utils as io


class AdaptTimeStep_D(DiscreteOperator):
    """
    The adaptative Time Step is computed according
    to the following expression :
    dt_adapt = min (dt_advection, dt_stretching, dt_cfl)
    """

    @debug
    def __init__(self, velocity, vorticity, dt_adapt, method=None,
                 lcfl=0.125, cfl=0.5, io_params=None, time_range=None):
        """
        @param velocity : discrete field
        @param vorticity : discrete field
        @param dt_adapt : adaptative timestep
        (a parmepy.variable_parameter.VariableParameter)
        @param method : numerical method for space/time discretizations
        @param lcfl : the lagrangian CFL coefficient used
        for advection stability
        @param cfl : the CFL coefficient.
        @param filename : output file name
        @param io_params : parameters (dict) to set file output.
        If  None, no output. Set io_params = {} if you want output,
        with default parameters values.
        See parmepy.tools.io_utils.Writer for details
        @param time_range : [start, end] use to define a 'window' in which
        the current operator is applied. Outside start-end, this operator
        has no effect. Start/end are iteration numbers.
        Default = [2, endofsimu]
        """
        ## velocity discrete field
        self.velocity = velocity
        ## vorticity discrete field
        self.vorticity = vorticity
        ## adaptative time step variable
        self.dt_adapt = dt_adapt
        if method is None:
            method = {TimeIntegrator: RK3, SpaceDiscretisation: FD_C_4,
                      dtCrit: 'vort'}
        DiscreteOperator.__init__(self, [self.velocity, self.vorticity],
                                  method=method)

        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]
        topo = self.velocity.topology
        io_params["writebuffshape"] = (1, 7)
        if io_params is None:
            self._writer = None
        else:
            if not "filename" in io_params:
                io_params["filename"] = "dt_adapt"
            if topo is not None:
                self._writer = io.Writer(io_params, topo.comm)

        ## Courant Fredrich Levy constants
        self.lcfl = lcfl
        self.cfl = cfl

        # Definition of stability coefficient for stretching operator
        self.cststretch = 0.
        timeint = self.method[TimeIntegrator]
        classtype = timeint.mro()[0]
        if classtype is Euler:
            self.coef_stretch = 2.0
        elif classtype is RK2:
            self.coef_stretch = 2.0
        elif classtype is RK3:
            self.coef_stretch = 2.5127
        elif classtype is RK4:
            self.coef_stretch = 2.7853

        # Definition of criterion for dt_advec computation
        self.dtCrit = self.method[dtCrit]
        if not isinstance(self.dtCrit, list):
            self.dtCrit = [self.dtCrit]
        ## Time range
        if time_range is None:
            time_range = [2, np.infty]
        self.time_range = time_range

        # local buffer :
        # [time, dt, d1, d2, d3, d4, d5]
        # for d1...d5 see computation details in apply.
        self.diagnostics = npw.zeros((7))
        self._t_diagnostics = np.empty_like(self.diagnostics)

        # All diagnostcs function  definition:
        # (Index in self.diagnostics, function, is gradU needed)
        self._all_functions = {
            'gradU': (2, self._compute_gradU, True),
            'stretch': (3, self._compute_stretch, True),
            'cfl': (4, self._compute_cfl, False),
            'vort': (5, self._compute_vort, False),
            'deform': (6, self._compute_deform, True),
            }

        # Build the user required function list
        self._used_functions = []
        self._is_gradU_needed = False
        for crit in self.dtCrit:
            self._is_gradU_needed = \
                self._is_gradU_needed or self._all_functions[crit][-1]
            self._used_functions.append(self._all_functions[crit])
        assert len(self._used_functions) >= 1, "You must specify at least " + \
            "one criterion among: " + str(self._all_functions.keys())

        # Definition of dt:
        self.get_all_dt = []

        # definition of dt_advection
        if 'gradU' in self.dtCrit:
            # => based on gradU
            self.get_all_dt.append(
                lambda diagnostics: self.lcfl / diagnostics[2])
        if 'deform' in self.dtCrit:
            # => based on the deformations
            self.get_all_dt.append(
                lambda diagnostics: self.lcfl / diagnostics[6])
        if 'vort' in self.dtCrit:
            # => based on the vorticity
            self.get_all_dt.append(
                lambda diagnostics: self.lcfl / diagnostics[5])
        if 'stretch' in self.dtCrit:
            self.get_all_dt.append(
                lambda diagnostics: self.coef_stretch / diagnostics[3])
        if 'cfl' in self.dtCrit:
            h = self.velocity.topology.mesh.space_step[0]
            self.get_all_dt.append(
                lambda diagnostics: (self.cfl * h) / diagnostics[4])

    def setUp(self):

        # prepare ghost points synchro for velocity
        self._synchronize = UpdateGhosts(self.velocity.topology,
                                         self.velocity.nbComponents)
        # gradU function
        self._function = GradV(self.velocity.topology,
                               method=self.method[SpaceDiscretisation])
        memshape = self.velocity.data[0].shape
        worklength = self.velocity.nbComponents ** 2
        # gradU result array.
        self.grad = [npw.zeros(memshape) for i in xrange(worklength)]

        self._isUpToDate = True

    def _gradU(self):
        # Synchronize ghost points of velocity
        self._synchronize(self.velocity.data)
        # gradU computation
        self.grad = self._function(self.velocity.data, self.grad)

    def _compute_gradU(self):
        res = 0.
        nbComponents = self.velocity.nbComponents
        for direc in xrange(nbComponents):
            # maxima of partial derivatives of velocity :
            # needed for advection stability condition (1st option)
            res = max(res, np.max(abs(self.grad[(nbComponents + 1) * direc])))
        return res

    def _compute_stretch(self):
        res = 0.
        nbComponents = self.velocity.nbComponents
        for direc in xrange(nbComponents):
            # maxima of partial derivatives of velocity:
            # needed for stretching stability condition
            tmp = np.max(sum([abs(self.grad[i])
                              for i in xrange(nbComponents * direc,
                                              nbComponents * (direc + 1))
                              ]))
            res = max(res, tmp)
        return res

    def _compute_cfl(self):
        # maxima of velocity : needed for CFL based time step
        return np.max([np.max(np.abs(v_c)) for v_c in self.velocity.data])

    def _compute_vort(self):
        # maxima of vorticity :
        # needed for advection stability condition (2nd option)
        return np.max([np.max(np.abs(w_c)) for w_c in self.vorticity.data])

    def _compute_deform(self):
        # 1/2(gradU + gradU^T) computation
        self.grad[1] += self.grad[3]
        self.grad[2] += self.grad[6]
        self.grad[5] += self.grad[7]
        self.grad[1] *= 0.5
        self.grad[2] *= 0.5
        self.grad[5] *= 0.5
        self.grad[3][...] = self.grad[1][...]
        self.grad[6][...] = self.grad[2][...]
        self.grad[7][...] = self.grad[5][...]
        # maxima of deformation tensor:
        # needed for advection stability condition (3rd option)
        res = 0.
        nbComponents = self.velocity.nbComponents
        for direc in xrange(nbComponents):
            tmp = np.max(sum([abs(self.grad[i])
                              for i in xrange(nbComponents * direc,
                                              nbComponents * (direc + 1))
                              ]))
            res = max(res, tmp)
        return res

    def apply(self, simulation=None):
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")
        # current time
        time = simulation.time
        iteration = simulation.currentIteration
        Nmax = min(simulation.iterMax, self.time_range[1])
        self.diagnostics[0] = time
        if iteration >= self.time_range[0] and iteration <= Nmax:
            # Calling for requirements completion
            DiscreteOperator.apply(self, simulation)
            if self._is_gradU_needed:
                self._gradU()
            for func in self._used_functions:
                # func: (Index in self.diagnostics, function, is gradU needed)
                self.diagnostics[func[0]] = func[1]()

            self.velocity.topology.comm.Allreduce(
                sendbuf=[self.diagnostics, 7, PARMES_MPI_REAL],
                recvbuf=[self._t_diagnostics, 7, PARMES_MPI_REAL],
                op=MPI.MAX)
            self.diagnostics[...] = self._t_diagnostics

            self.dt_adapt['dt'] = np.min(
                [dt(self.diagnostics) for dt in self.get_all_dt])

            self.diagnostics[1] = self.dt_adapt['dt']
            if self._writer is not None and self._writer.doWrite(iteration):
                self._writer.buffer[0, :] = self.diagnostics
                self._writer.write()

            # Update simulation time step with the new dt
            simulation.updateTimeStep(self.dt_adapt['dt'])
