# -*- coding: utf-8 -*-
"""
@file operator/discrete/multiphase.py
Discrete MultiPhase Rot Grad P
"""
from parmepy.operator.discrete.discrete import DiscreteOperator
import parmepy.numerics.differential_operations as diff_op
from parmepy.constants import np, debug, XDIR, YDIR, ZDIR, \
    PARMES_REAL, ORDER
from parmepy.methods_keys import SpaceDiscretisation
from parmepy.numerics.finite_differences import FD_C_4, FD_C_2
from parmepy.numerics.updateGhosts import UpdateGhosts
from parmepy.tools.timers import timed_function


class Baroclinic_d(DiscreteOperator):
    """

    """
    @debug
    def __init__(self, velocity, vorticity, density, viscosity,
                 method={SpaceDiscretisation: FD_C_2}, formula=None):
        """
        Constructor.
        Create the baroclinic term -GradRho/rho x GradP/rho
        in N.S equation
        @param velocity : discrete vector field
        @param vorticity : discrete vector field
        @param density : discrete scalar field
        @param viscosity
        @param method : numerical method for space discretizations
        @param formula : formula to initialize u^(n-1) 
        Note : this should be the formula used to initialize 
        the velocity field

        """
        DiscreteOperator.__init__(self, [velocity, vorticity, density,
                                         viscosity], method=method)

        self.velocity = velocity
        self.velocity_old = np.zeros_like(self.velocity.data, 
                                          dtype=PARMES_REAL, 
                                          order=ORDER)
        self.vorticity = vorticity
        self.density = density
        self.viscosity = viscosity
        self.formula = formula
        self.input = [self.velocity, self.vorticity, self.density]
        self.output = [self.vorticity]

        # prepare ghost points synchro for velocity (vector) 
        # and density (scalar) fields
        self._synchronizeVel = UpdateGhosts(self.velocity.topology,
                                            self.velocity.nbComponents)
        self._synchronizeRho = UpdateGhosts(self.density.topology,
                                            self.density.nbComponents)

    def setUp(self):
        """
        u^(n-1) initialization : u^(n-1) = u(t=0).
        """
        currentTime = 0.
        arg_list = self.velocity.topology.mesh.coords + (currentTime,)
        self.velocity_old = self.formula(self.velocity_old, *arg_list)

    @debug
    @timed_function
    def apply(self, simulation=None):
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")

        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)

        t = simulation.time
        dt = simulation.timeStep
        # Synchronize ghost points of velocity and density
        self._synchronizeVel(self.velocity.data)
        self._synchronizeRho(self.density.data)

        topo = self.velocity.topology
        iCompute = topo.mesh.iCompute
        gradOp = diff_op.GradS(topo, self.method[SpaceDiscretisation])

        result = np.zeros_like(self.velocity.data, 
                               dtype=PARMES_REAL, order=ORDER)

        # result = du/dt
        for d in xrange(self.velocity.dimension):
            result[d][iCompute] = (self.velocity[d][iCompute] - 
                                   self.velocity_old[d][iCompute]) / dt

        # result = result + (u. grad)u
        # (u. grad)u = (u.du/dx + v.du/dy + w.du/dz ;
        #               u.dv/dx + v.dv/dy + w.dv/dz ;
        #               u.dw/dx + v.dw/dy + w.dw/dz)
        UgradU = np.zeros_like(self.velocity.data, 
                               dtype=PARMES_REAL, order=ORDER)
        tempGrad = np.zeros_like(self.velocity.data, 
                                 dtype=PARMES_REAL, order=ORDER)
        tempGrad = gradOp(self.velocity[XDIR], tempGrad)
        for d in xrange(self.velocity.dimension):
            UgradU[XDIR][iCompute] += self.velocity[d][iCompute] \
                                      * tempGrad[d][iCompute]
        tempGrad = gradOp(self.velocity[YDIR], tempGrad)
        for d in xrange(self.velocity.dimension):
            UgradU[YDIR][iCompute] += self.velocity[d][iCompute] \
                                      * tempGrad[d][iCompute]
        tempGrad = gradOp(self.velocity[ZDIR], tempGrad)
        for d in xrange(self.velocity.dimension):
            UgradU[ZDIR][iCompute] += self.velocity[d][iCompute] \
                                      * tempGrad[d][iCompute]

        for d in xrange(self.velocity.dimension):
            result[d][iCompute] += UgradU[d][iCompute]

        ## result = result - nu*\Laplacian u (-g) = gradP/rho
        viscousTerm = np.zeros_like(self.velocity.data, 
                                    dtype=PARMES_REAL, order=ORDER)

        laplacian = diff_op.Laplacian(topo)
        fd_scheme = FD_C_2((topo.mesh.space_step))
        laplacian.fd_scheme.computeIndices(iCompute)

        for d in xrange(self.velocity.dimension):
            viscousTerm[d] = laplacian(self.velocity[d], viscousTerm[d])
        for d in xrange(self.velocity.dimension):
            viscousTerm[d][iCompute] *= self.viscosity
            result[d][iCompute] -= viscousTerm[d][iCompute]
        ## gravity term : result = result - g with g=(0,-1,-0)
#            if d==1:
#                result[d][iCompute] -= -1.0

        ## baroclinicTerm = -(gradRho/rho) x (gradP/rho)
        gradRho_rho = np.zeros_like(self.velocity.data, 
                                    dtype=PARMES_REAL, order=ORDER)
        gradRho_rho = gradOp(self.density[0], gradRho_rho)

        ## To comment out if the advected scalar is log(rho) :
#        for d in xrange(self.velocity.dimension):
#            gradRho_rho[d][iCompute] = gradRho_rho[d][iCompute] / \
#                                       self.density[0][iCompute]

        baroclinicTerm = np.zeros_like(result, dtype=PARMES_REAL, 
                                       order=ORDER)
        baroclinicTerm[0][iCompute] = - gradRho_rho[1][iCompute] * \
                                        result[2][iCompute] + \
                                      gradRho_rho[2][iCompute] * \
                                        result[1][iCompute]
        baroclinicTerm[1][iCompute] = - gradRho_rho[2][iCompute] * \
                                        result[0][iCompute] + \
                                      gradRho_rho[0][iCompute] * \
                                        result[2][iCompute]
        baroclinicTerm[2][iCompute] = - gradRho_rho[0][iCompute] * \
                                        result[1][iCompute] + \
                                      gradRho_rho[1][iCompute] * \
                                        result[0][iCompute]

        ## vorti(n+1) = vorti(n) + dt * baroclinicTerm
        for d in xrange(self.vorticity.dimension):
            self.vorticity[d][iCompute] += dt * baroclinicTerm[d][iCompute]

        ## velo(n-1) update
        for d in xrange(self.velocity.dimension):
            self.velocity_old[d][iCompute] = self.velocity.data[d][iCompute]
