# -*- coding: utf-8 -*-
"""
@file discrete/differential.py

Discretisation of the differential operators (curl, grad ...)
"""
from parmepy.constants import debug
from parmepy.operator.discrete.discrete import DiscreteOperator
from parmepy.numerics.differential_operations import Curl, GradV
import parmepy.tools.numpywrappers as npw
from abc import ABCMeta, abstractmethod
from parmepy.numerics.updateGhosts import UpdateGhosts
from parmepy.methods_keys import SpaceDiscretisation
try:
    from parmepy.f2py import fftw2py
except ImportError:
    from parmepy.fakef2py import fftw2py
import parmepy.default_methods as default


class Differential(DiscreteOperator):
    __metaclass__ = ABCMeta

    ## @debug
    ## def __new__(cls, *args, **kw):
    ##     return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, invar, outvar, method=None):
        """
        @param[in] invar : input field
        @param[in,out] outvar : Grad of the input field.
        Warning : must be a field with dim * invar.nbcomponents,
        dim being the domain dimension.
        @param[in] method : how to compute the grad? Default = finite
        differences, 4th order.
        """
        self.invar = invar
        self.outvar = outvar
        if method is None:
            method = default.DIFFERENTIAL
        DiscreteOperator.__init__(self, [self.invar, self.outvar],
                                  method=method)
        self.input = [self.invar]
        self.output = [self.outvar]
        self._dim = self.invar.topology.domain.dimension
        self._synchronize = None

    @staticmethod
    def getWorkLengths(method=None, domain_dim=None):
        assert method is None
        assert domain_dim is None
        return 0.

    @abstractmethod
    def apply(self, simulation=None):
        """
        """


class CurlFFT(Differential):
    """
    Compute the curl of a discrete field, using Fourier fftw
    """

    def apply(self, simulation=None):
        DiscreteOperator.apply(self, simulation)
        ghosts_in = self.invar.topology.ghosts
        ghosts_out = self.outvar.topology.ghosts
        self.outvar.data[0], self.outvar.data[1], self.outvar.data[2] = \
            fftw2py.solve_curl_3d(self.invar.data[0], self.invar.data[1],
                                  self.invar.data[2], self.outvar.data[0],
                                  self.outvar.data[1], self.outvar.data[2],
                                  ghosts_in, ghosts_out)

    def finalize(self):
        """
        Clean memory (fftw plans and so on)
        """
        fftw2py.clean_fftw_solver(self.outvar.dimension)


class CurlFD(Differential):
    """
    Compute the curl of a discrete field, using finite differences.
    """

    def setUp(self):
        # prepare ghost points synchro for velocity
        self._synchronize = UpdateGhosts(self.invar.topology,
                                         self.invar.nbComponents)

        worklength = self.getWorkLengths()
        memshape = self.invar.data[0].shape
        if not self.hasExternalWork:
            self._rwork = [npw.zeros(memshape) for i in xrange(worklength)]
        else:
            assert len(self._rwork) == worklength
            for wk in self._rwork:
                assert wk.shape == memshape
        assert self.outvar.nbComponents == self.invar.nbComponents

        self._function = Curl(self.invar.topology, self._rwork,
                              self.method[SpaceDiscretisation])

    @staticmethod
    def getWorkLengths(method=None, domain_dim=None):
        return Curl.getWorkLengths()

    def apply(self, simulation=None):
        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)
        self._synchronize(self.invar.data)
        self.outvar.data = self._function(self.invar.data, self.outvar.data)


class GradFD(Differential):
    """
    Compute the grad of a discrete field, using finite differences.
    """

    def setUp(self):
        # prepare ghost points synchro for velocity
        self._synchronize = UpdateGhosts(self.invar.topology,
                                         self.invar.nbComponents)
        assert self.outvar.nbComponents == self._dim * self.invar.nbComponents
        self._function = GradV(self.invar.topology,
                               self.method[SpaceDiscretisation])

    def curlFD(self, simulation=None):
        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)
        self._synchronize(self.invar.data)
        self.outvar.data = self._function(self.invar.data, self.outvar.data)

    def apply(self, simulation=None):
        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)
        self._synchronize(self.invar.data)
        self.outvar.data = self._function(self.invar.data, self.outvar.data)
