# -*- coding: utf-8 -*-
"""
@file diffusion_fft.py
Discrete Diffusion operator using FFTW (fortran)
"""
try:
    from parmepy.f2py import fftw2py
except ImportError:
    from parmepy.fakef2py import fftw2py
from parmepy.operator.discrete.discrete import DiscreteOperator
from parmepy.constants import debug
from parmepy.mpi import MPI


class DiffusionFFT(DiscreteOperator):
    """
    Discretized Poisson operator based on FFTW.
    See details in parmepy.operator.diffusion.

    """
    @debug
    def __init__(self, vorticity, viscosity, method=None):
        """
        Constructor.
        @param[in,out] vorticity :  discretisation of the field \f$ \omega \f$.
        @param[in] viscosity : \f$\nu\f$, viscosity of the considered medium.
        """
        ## Discretisation of the solution field
        self.vorticity = vorticity
        ## Viscosity.
        self.viscosity = viscosity

        if self.vorticity.dimension == 1:
            raise AttributeError("Wrong problem dimension: only 2D \
                and 3D cases are implemented.")
        # Base class initialisation
        DiscreteOperator.__init__(self, [self.vorticity], method)
        self.input = [self.vorticity]
        self.output = [self.vorticity]

    @debug
    def apply(self, simulation=None):
        if simulation is None:
            raise ValueError("Missing dt value for diffusion computation.")
        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)
        ctime = MPI.Wtime()

        dt = simulation.timeStep
        ghosts = self.vorticity.topology.ghosts

        if (self.vorticity.dimension == 2):
            self.vorticity.data = fftw2py.solve_diffusion_2d(
                self.viscosity * dt, self.vorticity.data)

        elif (self.vorticity.dimension == 3):
            self.vorticity.data[0], self.vorticity.data[1],\
                self.vorticity.data[2] = \
                fftw2py.solve_diffusion_3d(self.viscosity * dt,
                                           self.vorticity.data[0],
                                           self.vorticity.data[1],
                                           self.vorticity.data[2],
                                           ghosts)

        else:
            raise ValueError("invalid problem dimension")
        self._apply_timer.append_time(MPI.Wtime() - ctime)

    def finalize(self):
        """
        Clean memory (fftw plans and so on)
        """
        fftw2py.clean_fftw_solver(self.vorticity.dimension)
