"""
@file operator/discrete/discrete.py
Abstract interface for discrete operators.
"""
from abc import ABCMeta, abstractmethod
from parmepy.constants import debug
from parmepy.tools.timers import Timer, ManualFunctionTimer
from parmepy.methods_keys import GhostUpdate


class DiscreteOperator(object):
    """
    Abstract description of a discretized operator.
    It provides a common interface to all discrete
    operators.

    """

    __metaclass__ = ABCMeta

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    @abstractmethod
    def __init__(self, variables, method=None):
        """
        Create an empty discrete operator.
        """
        if isinstance(variables, list):
            ## variables
            self.variables = variables
        else:
            self.variables = [variables]
        ## Input variables
        self.input = []
        ## Output variables
        self.output = []
        ## Operator numerical method.
        if method is None:
            method = {}
        self.method = method
        if not GhostUpdate in method:
            method[GhostUpdate] = True
        self.name = self.__class__.__name__
        ## Object to store computational times of lower level functions
        self.timer = Timer(self)
        ## bool to check if the setup function has been called for
        ## this operator
        self._isUpToDate = False
        ## True if work arrays are provided by external call (self.setWorks)
        self.hasExternalWork = False
        self._apply_timer = ManualFunctionTimer('apply_function')
        self.timer.addFunctionTimer(self._apply_timer)
        # Local (optional) work arrays. Set with setWorks function
        self._rwork = None
        self._iwork = None

    @staticmethod
    def getWorkLengths(nb_components=None, domain_dim=None):
        """
        Compute the number of required work arrays for this method.
        @param nb_components : number of components of the
        @param domain_dim : dimension of the domain
        fields on which this method operates.
        @return length of list of work arrays of reals.
        @return length of list of work arrays of int.
        """
        assert nb_components is None
        assert domain_dim is None
        return 0, 0

    def setWorks(self, rwork=None, iwork=None):

        # Set work arrays for real and int.
        # Warning : no copy! We must have pointer
        # links between work and self.work arrays.
        if rwork is None:
            rwork = []
        if iwork is None:
            iwork = []
        self._rwork = rwork
        self._iwork = iwork
        self.hasExternalWork = True
        # Warning : we do not check work lengths/size.
        # This will be done during discreteOperator.setUp.

    def setUp(self):
        """
        Update the operator --> after calling this function,
        the operator is ready to use.
        Main steps are:
        - allocate (if required) work arrays
        - check work arrays number and shapes
        - create instances of numerical methods required by
        this operator
        """
        self._isUpToDate = True

    def isUp(self):
        """
        True if ready to be applied (--> setup function has
        been called succesfully)
        """
        return self._isUpToDate

    @debug
    @abstractmethod
    def apply(self, simulation=None):
        """
        Apply this operator to its variables.
        @param simulation : object that describes the simulation
        parameters (time, time step, iteration number ...), see
        parmepy.problem.simulation.Simulation for details.
        """
        pass

    @debug
    def finalize(self):
        """
        Cleaning, if required.
        """
        pass

    def __str__(self):
        """Common printings for discrete operators."""
        shortName = str(self.__class__).rpartition('.')[-1][0:-2]
        s = shortName + " discrete operator. \n"
        if self.input is not None:
            s += "Input fields : \n"
            for f in self.input:
                s += str(f) + "\n"
        if self.output is not None:
            s += "Output fields : \n"
            for f in self.output:
                s += str(f) + "\n"
        return s

    def updateGhosts(self):
        """
        Update ghost points values, if any.
        This function must be implemented in the discrete
        operator if it may be useful to ask for ghost
        points update without a call to apply.
        For example for monitoring purpose before
        operator apply.
        """
        pass
