"""
@file particle_advection.py

Advection solver, particular method, pure-python version.

"""
from parmepy.constants import debug, WITH_GUESS, PARMES_INDEX
from parmepy.methods_keys import TimeIntegrator, Interpolation, Remesh
from parmepy.operator.discrete.discrete import DiscreteOperator
from parmepy.fields.continuous import Field
import parmepy.tools.numpywrappers as npw
from parmepy.mpi import MPI
import parmepy.default_methods as default


class ParticleAdvection(DiscreteOperator):
    """
    Particular method for advection of a list of fields in a given direction.
    """

    @debug
    def __init__(self, velocity, advectedFields, d,
                 part_position=None, part_advectedFields=None,
                 method=None,
                 isMultiScale=False):
        """
        Constructor.
        @param velocity discrete field
        @param advectedFields : list of discrete fields to be advected
        @param d : Direction to advect
        @param method : Method used
          - Integration methods:
            - 'rk2' : Runge Kutta 2nd order advection
            - 'rk4' : Runge Kutta 4th order advection
          - remeshing formula:
            - 'm4prime' : = 'l2_1'
            - 'l2_1' : Labmda2,1 : (=M'4) 4 point formula, C1 regularity
            - 'l2_2' : Labmda2,2 : 4 point formula, C2 regularity
            - 'm6prime' : = 'l4_2'
            - 'l4_2' : Labmda4,2 : (=M'6) 6 point formula, C2 regularity
            - 'l4_3' : Labmda4,3 : 6 point formula, C3 regularity
            - 'l4_4' : Labmda4,4 : 6 point formula, C4 regularity
            - 'l6_3' : Labmda6,3 : 8 point formula, C3 regularity
            - 'l6_4' : Labmda6,4 : 8 point formula, C4 regularity
            - 'l6_5' : Labmda6,5 : 8 point formula, C5 regularity
            - 'l6_6' : Labmda6,6 : 8 point formula, C6 regularity
            - 'l8_4' : Labmda8,4 : 10 point formula, C4 regularity
            - 'm8prime' : M8prime formula
        """
        ## Advection velocity
        self.velocity = velocity
        variables = [self.velocity]
        if not isinstance(advectedFields, list):
            self.advectedFields = [advectedFields]
        else:
            self.advectedFields = advectedFields
        # update the list of discrete variables for this operator
        [variables.append(advecF) for advecF in self.advectedFields]
        if method is None:
            method = default.ADVECTION
        DiscreteOperator.__init__(self, variables, method)

        self.input = self.variables
        self.output = [df for df in self.variables if not df is self.velocity]
        self.dir = d
        self.part_position = part_position
        self.part_advectedFields = part_advectedFields
        self._dim = self.velocity.dimension
        self._workspace, self._iworkspace = [], []
        self._isMultiScale = isMultiScale

    @staticmethod
    def getWorkLengths(method=None, domain_dim=None):
        """
        Return the length of working arrays lists required
        for advction discrete operator, depending on :
        - the time integrator (RK2, ...)
        - the interpolation (which depends on domain dimension)
        - the remeshing (which depends on domain dimension)
        @param method : the dict of parameters for the operator.
        Default = parmepy.default_methods.ADVECTION
        """
        if method is None:
            method = default.ADVECTION
        assert Interpolation in method,\
            'An interpolation is required for the advection method.'
        assert TimeIntegrator in method,\
            'A time integrator is required for the advection method.'
        assert Remesh in method,\
            'A remesh is required for the advection method.'
        tw = method[TimeIntegrator].getWorkLengths(1)
        iw, iiw = method[Interpolation].getWorkLengths(domain_dim=domain_dim)
        rw, riw = method[Remesh].getWorkLengths(domain_dim=domain_dim)
        return max(tw + iw, rw), max(iiw, riw)

    @debug
    def setUp(self):
        ## Result position
        if self.part_position is None:
            self.part_position = \
                Field(self.advectedFields[0].topology.domain,
                      "Particle_Position",
                      isVector=False
                      ).discretize(self.advectedFields[0].topology)
        ## Result advectedFields
        if self.part_advectedFields is None:
            self.part_advectedFields = [
                Field(adF.topology.domain, "Particle_AdvectedFields",
                      isVector=adF.isVector).discretize(adF.topology)
                for adF in self.advectedFields]

        RKn = self.method[TimeIntegrator]
        Interpol = self.method[Interpolation]
        Rmsh = self.method[Remesh]
        memshape = self.part_position.data[0].shape
        # Get work lengths
        w, iw = self.getWorkLengths(self.method, self._dim)
        w_rk = RKn.getWorkLengths(nb_components=1)
        w_interp, iw_interp = Interpol.getWorkLengths(domain_dim=self._dim)
        w_remesh, iw_remesh = Rmsh.getWorkLengths(domain_dim=self._dim)
        # Allocate local works
        if not self.hasExternalWork:
            for i in xrange(w):
                self._workspace.append(
                    npw.zeros_like(self.part_position.data[0]))
            for i in xrange(iw):
                self._iworkspace.append(
                    npw.zeros(memshape, dtype=PARMES_INDEX))
        else:
            # Check external works lengths
            assert len(self._workspace) == w
            assert len(self._iworkspace) == iw
            for wk in self._workspace:
                assert wk.shape == tuple(memshape)
            for wk in self._iworkspace:
                assert wk.shape == tuple(memshape)

        # Distribute work arrays
        self._interpol_w = self._workspace[:w_interp]
        self._interpol_iw = self._iworkspace[:iw_interp]
        self._integrator_w = self._workspace[w_interp:w_interp + w_rk]
        self._remesh_w = self._workspace[:w_remesh]
        self._remesh_iw = self._iworkspace[:iw_remesh]

        self.num_interpolate = \
            Interpol(self.velocity, self.dir,
                     self.advectedFields[0].topology,
                     work=self._interpol_w,
                     iwork=self._interpol_iw)
        self.num_advec = RKn(1, work=self._integrator_w,
                             f=self.num_interpolate,
                             topo=self.velocity.topology, optim=WITH_GUESS)
        self.num_remesh = Rmsh(
            self.velocity.dimension,
            self.advectedFields[0].topology,
            self.dir, work=self._remesh_w,
            iwork=self._remesh_iw)

        self._isUpToDate = True

    @debug
    def apply(self, simulation, dtCoeff, split_id, old_dir=None):
        """
        Apply advection operator.

        @param t : current time.
        @param dt : time step.
        @param d : Direction of splitting.

        Advection algorithm:
        @li 1. Particle initialization : \n
                 - by copy advectedFields from grid to particles
                 if previous splitting direction equals current
                 splitting direction.\n
                 - by transposition of advectedFields from grid to particle.
        @li 2. Particle advection :\n
                 - compute particle position in splitting direction as a
                 advectedFields. Performs a RK2 resolution of dx_p/dt = a_p.
        """
        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)
        ctime = MPI.Wtime()

        t, dt = simulation.time, simulation.timeStep * dtCoeff
        # copy for particle advected initialization
        for p_adF, adF in zip(self.part_advectedFields, self.advectedFields):
            for dim in xrange(adF.nbComponents):
                p_adF[dim][...] = adF[dim][...]

        # Get particle initial position (grid points)
        self.part_position[0][...] = \
            self.part_position.topology.mesh.coords[self.dir]

        # Advect particles
        # RK use the first 2 (or 3) works and leave others to interpolation
        # First work contains fist evaluation of ode right hand side.
        self._integrator_w[0][...] = self.velocity.data[self.dir][...]
        self.part_position.data = self.num_advec(
            t, self.part_position.data, dt, result=self.part_position.data)

        # Remesh particles
        # It uses the last dim + 2 workspaces (same as interpolation)
        for adF, p_adF in zip(self.advectedFields,
                              self.part_advectedFields):
            for dim in xrange(adF.nbComponents):
                adF[dim][...] = self.num_remesh(
                    self.part_position.data[0], p_adF.data[dim],
                    result=adF[dim])
        self._apply_timer.append_time(MPI.Wtime() - ctime)

    @debug
    def finalize(self):
        pass
