# -*- coding: utf-8 -*-
"""
@file operator/discrete/penalization.py
Discrete operator for penalization problem.
"""
from parmepy.constants import debug
from discrete import DiscreteOperator
from parmepy.tools.timers import timed_function
import numpy as np


class Penalization_d(DiscreteOperator):
    """
    Discretized penalisation operator.
    See details in parmepy.operator.penalization
    """

    @debug
    def __init__(self, variables, obstacles, factor):
        """
        Constructor.
        @param[in,out] variables : list of discrete fields to be penalized
        @param[in] obstacle : physical domain in which the penalization
        is applied.
        @param[in] factor : penalization factoricient
        """
        DiscreteOperator.__init__(self, variables)

        ## Penalization parameter
        self.factor = np.asarray(factor)
        ## Obstacle
        if isinstance(obstacles, list):
            self.obstacles = obstacles
        else:
            self.obstacles = [obstacles]
        # Discretize obstacles, i.e. create boolean arrays
        topo = self.variables[0].topology
        self.cond = None
        for obstacle in self.obstacles:
            cond = obstacle.discretize(topo)
            if self.cond is None:
                self.cond = cond
            else:
                lsize = min(len(self.cond), len(cond))
                for c in xrange(lsize):
                    np.logical_or(cond[c], self.cond[c], self.cond[c])
                for c in xrange(lsize + 1, len(cond)):
                    self.cond[c] = cond[c]
        assert self.factor.size == len(self.cond)
        self.input = self.output = self.variables

    @debug
    @timed_function
    def apply(self, simulation=None):
        if simulation is None:
            raise ValueError("Missing dt value for penalization computation.")

        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)

        dt = simulation.timeStep
        coef = 1.0 / (1.0 + dt * self.factor)

        for v in self.variables:
            i = 0
            for cond in self.cond:
                for dim in xrange(v.nbComponents):
                    v[dim][cond] *= coef[i]
                i += 1

