# -*- coding: utf-8 -*-
"""
@file poisson_fft.py
Discrete operator for Poisson problem (fftw based)
"""
import parmepy.tools.numpywrappers as npw
try:
    from parmepy.f2py import fftw2py
except ImportError:
    from parmepy.fakef2py import fftw2py

from parmepy.operator.discrete.discrete import DiscreteOperator
from parmepy.constants import debug, prof
from parmepy.mpi import MPI


class PoissonFFT(DiscreteOperator):
    """
    Discretized Poisson operator based on FFTW.
    See details in parmepy.operator.poisson
    """

    @debug
    def __init__(self, velocity, vorticity, projection=None,
                 multires=False, filterSize=None, correction=None):
        """
        Constructor.
        @param[out] velocity : discretisation of the solution field
        @param[in] vorticity : discretisation of the RHS (mind the minus rhs!)
        @param projection :
        @param[in] multires : true if velo/vorticity do not have the same
        resolution
        @param filterSize :
        @param correction : operator used to shift velocity according
        to a given input (fixed) flowrate.
        See parmepy.operator.velocity_correction.
        Default = None.
        """
        ## Solution field
        self.velocity = velocity
        ## RHS field
        self.vorticity = vorticity
        ## Solenoidal projection of vorticity ?
        self.projection = projection
        ## Filter size array = domainLength/(CoarseRes-1)
        self.filterSize = filterSize
        # Base class initialisation
        DiscreteOperator.__init__(self, [velocity, vorticity])

        # If 2D problem, vorticity must be a scalar
        self.dim = self.velocity.domain.dimension
        if self.dim == 2:
            assert self.vorticity.nbComponents == 1

        if self.dim != 2 and self.dim != 3:
            raise AttributeError("Wrong problem dimension: only 2D \
                                        and 3D cases are implemented.")

        self.input = [self.vorticity]
        self.output = [self.velocity]
        ## Multiresolution ?
        self.multires = multires

        # connexion to the required apply function
        # (to avoid 'if' calls during apply)
        if self.dim == 2:
            self._solve = self._solve2D
        elif self.dim == 3:
            # If there is a projection, vorticity is also an output
            if self.projection is not None:
                self.output.append(self.vorticity)
                if self.multires:
                    self._solve = self._solve3D_proj_multires
                else:
                    self._solve = self._solve3D_proj
            else:
                if self.multires:
                    self._solve = self._solve3D_multires
                else:
                    self._solve = self._solve3D

        # Operator to shift velocity according to an input required flowrate
        if correction is not None:
            self.correctionOp = correction
            self.solve = self._solve_and_correct
        else:
            self.solve = self._solve

    def _solve2D(self, simu=None):
        """
        Solve 2D poisson problem
        """
        self.velocity.data[0], self.velocity.data[1] =\
            fftw2py.solve_poisson_2d(self.vorticity.data,
                                     self.velocity.data[0],
                                     self.velocity.data[1])

    def _project(self):
        """
        apply projection onto vorticity
        """
        ghosts_w = self.vorticity.topology.ghosts
        self.vorticity.data[0], self.vorticity.data[1], \
            self.vorticity.data[2] = \
               fftw2py.projection_om_3d(self.vorticity.data[0],
                                        self.vorticity.data[1],
                                        self.vorticity.data[2], ghosts_w)

    def _solve3D_multires(self, simu=None):
        """
        3D, multiresolution
        """
        # Projects vorticity values from fine to coarse grid
        # in frequencies space by nullifying the smallest modes
        vortFilter = npw.copy(self.vorticity.data)
        vortFilter[0], vortFilter[1], vortFilter[2] = \
           fftw2py.multires_om_3d(self.filterSize[0], self.filterSize[1],
                                  self.filterSize[2], self.vorticity.data[0],
                                  self.vorticity.data[1],
                                  self.vorticity.data[2])

        # Solves Poisson equation using filter vorticity
        ghosts_v = self.velocity.topology.ghosts
        ghosts_w = self.vorticity.topology.ghosts
        self.velocity.data[0], self.velocity.data[1], self.velocity.data[2] = \
            fftw2py.solve_poisson_3d(vortFilter[0], vortFilter[1],
                                     vortFilter[2], self.velocity.data[0],
                                     self.velocity.data[1],
                                     self.velocity.data[2], ghosts_w, ghosts_v)

    def _solve3D_proj_multires(self, simu):
        """
        3D, multiresolution, with projection
        """
        ite = simu.currentIteration
        if self.projection.doProjection(ite):
            self._project()
        self._solve3D_multires()

    def _solve3D_proj(self, simu):
        """
        3D, with projection
        """
        ite = simu.currentIteration
        if self.projection.doProjection(ite):
            self._project()
        self._solve3D()

    def _solve3D(self,simu=None):
        """
        Basic solve
        """
        # Solves Poisson equation using usual vorticity
        ghosts_v = self.velocity.topology.ghosts
        ghosts_w = self.vorticity.topology.ghosts
        self.velocity.data[0], self.velocity.data[1], self.velocity.data[2] =\
            fftw2py.solve_poisson_3d(self.vorticity.data[0],
                                     self.vorticity.data[1],
                                     self.vorticity.data[2],
                                     self.velocity.data[0],
                                     self.velocity.data[1],
                                     self.velocity.data[2], ghosts_w, ghosts_v)

    def _solve_and_correct(self, simu):
        self._solve(simu.currentIteration)
        self.correctionOp.apply(simu)

    @debug
    @prof
    def apply(self, simulation=None):
        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)
        ctime = MPI.Wtime()
        self.solve(simulation)
        self._apply_timer.append_time(MPI.Wtime() - ctime)

    def finalize(self):
        """
        Clean memory (fftw plans and so on)
        """
        pass
        #fftw2py.clean_fftw_solver(self.velocity.dimension)
