# -*- coding: utf-8 -*-
"""
@file scales_advection.py
Discrete Advection operator based on scales library (Jean-Baptiste)

"""
try:
    from parmepy.f2py import scales2py
except ImportError:
    from parmepy.fakef2py import scales2py
from parmepy.operator.discrete.discrete import DiscreteOperator
from parmepy.constants import debug
from parmepy.mpi import MPI
import math
ceil = math.ceil


class ScalesAdvection(DiscreteOperator):
    """
    Advection process, based on scales library.
    """

    @debug
    def __init__(self, velocity, advectedFields, method='',
                 isMultiscale=False):
        """
        Constructor.
        @param velocity discrete field
        @param advectedFields : list of discrete fields to be advected
        @param method : Method used.
        """
        ## Advection velocity
        self.velocity = velocity
        variables = [self.velocity]
        if not isinstance(advectedFields, list):
            self.advectedFields = [advectedFields]
        else:
            self.advectedFields = advectedFields
        # update the list of discrete variables for this operator
        [variables.append(advecF) for advecF in self.advectedFields]

        DiscreteOperator.__init__(self, variables, method)

        self.input = [self.velocity]
        self.output = list(variables).remove(self.velocity)

        # Scales functions for each field (depending if vector)
        self._scales_func = []

        for adF in self.advectedFields:
            if adF.nbComponents == 3:
                if isMultiscale:
                    # 3D interpolation of the velocity before advection
                    self._scales_func.append(
                        scales2py.solve_advection_inter_basic_vect)
                    # Other interpolation only 2D interpolation first and
                    # 1D interpolations before advections in each direction
                    # (slower than basic): solve_advection_inter
                else:
                    self._scales_func.append(scales2py.solve_advection_vect)
            else:
                if isMultiscale:
                    self._scales_func.append(
                        scales2py.solve_advection_inter_basic)
                else:
                    self._scales_func.append(scales2py.solve_advection)

    @debug
    def apply(self, simulation=None):
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")

        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)
        ctime = MPI.Wtime()

        dt = simulation.timeStep

        # Call scales advection
        for adF, fun in zip(self.advectedFields, self._scales_func):
            adF = fun(dt, self.velocity.data[0],
                      self.velocity.data[1],
                      self.velocity.data[2],
                      *adF)
        self._apply_timer.append_time(MPI.Wtime() - ctime)

    def finalize(self):
        """
        \todo check memory deallocation in scales???
        """
        DiscreteOperator.finalize(self)
