# -*- coding: utf-8 -*-
"""
@file operator/discrete/stretching.py

Discretisation of the stretching operator for two different formulations.
"""

from parmepy.constants import debug, WITH_GUESS
from parmepy.methods_keys import TimeIntegrator, SpaceDiscretisation
from parmepy.operator.discrete.discrete import DiscreteOperator
from parmepy.numerics.integrators.euler import Euler
from parmepy.numerics.integrators.runge_kutta2 import RK2
from parmepy.numerics.integrators.runge_kutta3 import RK3
from parmepy.numerics.integrators.runge_kutta4 import RK4
import parmepy.numerics.differential_operations as diff_op
import parmepy.tools.numpywrappers as npw
from parmepy.numerics.updateGhosts import UpdateGhosts
from parmepy.mpi import MPI
from abc import ABCMeta, abstractmethod
import math
ceil = math.ceil


class Stretching(DiscreteOperator):
    """
    Abstract interface to stretching discrete operators.
    Two formulations are available:
    - conservative, see operator.discrete.stretching.Conservative
    - 'Grad(UxW), see operator.discrete.stretching.GradUW
    """

    __metaclass__ = ABCMeta

    @debug
    def __init__(self, velocity, vorticity, method=None):
        """
        @param velocity : discrete field
        @param vorticity : discrete field
        @param method : numerical method for space/time discretizations
        Default = {TimeIntegrator: RK3, SpaceDiscretisation: FD_C_4}
        """
        ## velocity discrete field
        self.velocity = velocity
        ## vorticity discrete field
        self.vorticity = vorticity
        if method is None:
            import parmepy.default_methods as default
            method = default.STRETCHING
        DiscreteOperator.__init__(self, [self.velocity, self.vorticity],
                                  method=method)

        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]
        # \todo multiresolution case
        assert self.velocity.topology == self.vorticity.topology,\
            'Multiresolution case not yet implemented.'

        ## Number of components of the operator (result)
        self.nbComponents = 3  # Stretching only in 3D and for vector fields.

        ## stability constant
        self.cststretch = 0.
        # Depends on time integration method
        timeint = self.method[TimeIntegrator]
        classtype = timeint.mro()[0]
        if classtype is Euler:
            self.cststretch = 2.0
        elif classtype is RK2:
            self.cststretch = 2.0
        elif classtype is RK3:
            self.cststretch = 2.5127
        elif classtype is RK4:
            self.cststretch = 2.7853
        # prepare ghost points synchro for velocity
        self._synchronize = UpdateGhosts(self.velocity.topology,
                                         self.velocity.nbComponents
                                         + self.vorticity.nbComponents)
        ## Formulation used to compute stretching (default = DivWV)
        self.formulation = diff_op.DivWV

    def setUp(self):

        memshape = self.velocity.data[0].shape
        # work list length for time-integrator
        self._work_length_ti = self.method[TimeIntegrator].getWorkLengths(3)
        # work list length for DivWV operation.
        self._work_length_str = self.formulation.getWorkLengths()
        if not self.hasExternalWork:
            self._worklength = self._work_length_str + self._work_length_ti
            self._rwork = [npw.zeros(memshape)
                           for i in xrange(self._worklength)]
        else:
            assert len(self._rwork) == self._work_length_ti + \
                self._work_length_str
            for wk in self._rwork:
                assert wk.shape == memshape

        self.time_int_work = self._rwork[:self._work_length_ti]
        self.str_work = self._rwork[self._work_length_ti:]
        # A function to compute the gradient of a vector field
        # Work vector is provided in input.
        self.strFunc = \
            self.formulation(self.velocity.topology,
                             self.str_work, method=
                             self.method[SpaceDiscretisation])

    def updateGhosts(self):
        """
        Update ghost points values
        """
        self._synchronize(self.velocity.data + self.vorticity.data)

    @abstractmethod
    def apply(self, simulation=None):
        """
        """


class Conservative(Stretching):
    """
    Discretisation of the following problem :
    \f{eqnarray*} \frac{\partial\omega}{\partial t} = \nabla.(\omega:v) \f}
    """
    def __init__(self, velocity, vorticity, method=None):
        Stretching.__init__(self, velocity, vorticity, method)
        self.formulation = diff_op.DivWV

    @staticmethod
    def getWorkLengths(method=None, domain_dim=3):
        assert TimeIntegrator in method,\
            'A time integrator is required for the stretching.'
        # Stretching occurs only in 3D
        rwork_length = method[TimeIntegrator].getWorkLengths(domain_dim)
        rwork_length += diff_op.DivWV.getWorkLengths()
        return rwork_length, 0

    def setUp(self):
        Stretching.setUp(self)

        # Time integration scheme.
        def rhs(t, y, result):
            result = self.strFunc(y, self.velocity.data, result)
            return result

        # Create the time integrator
        self.timeIntegrator = \
            self.method[TimeIntegrator](self.nbComponents,
                                        self.time_int_work,
                                        self.velocity.topology,
                                        f=rhs,
                                        optim=WITH_GUESS)
        self._isUpToDate = True

    def apply(self, simulation=None):
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")

        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)
        ctime = MPI.Wtime()

        # time step
        dt = simulation.timeStep
        # current time
        t = simulation.time
        # Synchronize ghost points of velocity and vorticity
        self._synchronize(self.velocity.data + self.vorticity.data)
        # - Call time integrator -
        # Init workspace with a first evaluation of the
        # rhs of the integrator
        self.time_int_work[:self.nbComponents] = \
            self.timeIntegrator.f(t, self.vorticity.data,
                                  self.time_int_work[:self.nbComponents])
        # perform integration and save result in-place
        self.vorticity.data = self.timeIntegrator(t, self.vorticity.data, dt,
                                                  result=self.vorticity.data)
        self._apply_timer.append_time(MPI.Wtime() - ctime)


class GradUW(Stretching):
    """ Discretisation of the following problem:
    \f{eqnarray*} \frac{\partial\omega}{\partial t}=[\nabla(v)][\omega]\f}
    """

    def __init__(self, velocity, vorticity, method=None):
        Stretching.__init__(self, velocity, vorticity, method)
        self.formulation = diff_op.GradVxW

    @staticmethod
    def getWorkLengths(method=None, domain_dim=3):
        """
        Return the length of working arrays lists required
        for stretching discrete operator, depending on :
        - the formulation (Conservative or GradUW)
        - the time integrator (RK3, ...)
        @param method : the dict of parameters for the operator.
        """
        if method is None:
            import parmepy.default_methods as default
            method = default.STRETCHING
        assert TimeIntegrator in method,\
            'A time integrator is required for the stretching.'
        # Stretching occurs only in 3D
        rwork_length = method[TimeIntegrator].getWorkLengths(domain_dim)
        rwork_length += diff_op.GradVxW.getWorkLengths()
        return rwork_length, 0

    def setUp(self):
        Stretching.setUp(self)
        ## a vector to save diagnostics computed from GradVxW (max div ...)
        self.diagnostics = npw.ones(2)

        # Time integration scheme.
        def rhs(t, y, result):
            result, self.diagnostics =\
                self.strFunc(self.velocity.data, y, result, self.diagnostics)
            return result

        # Create the time integrator
        self.timeIntegrator = \
            self.method[TimeIntegrator](self.nbComponents,
                                        self.time_int_work,
                                        self.velocity.topology,
                                        f=rhs,
                                        optim=WITH_GUESS)
        self._isUpToDate = True

    def apply(self, simulation=None):
        # Calling for requirements completion
        DiscreteOperator.apply(self, simulation)
        ctime = MPI.Wtime()

        # time step
        dt = simulation.timeStep
        # current time
        t = simulation.time

        # Synchronize ghost points of velocity
        self._synchronize(self.velocity.data + self.vorticity.data)

        # Compute the number of required subcycles
        ndt, subdt = self.checkStability(dt)
        assert sum(subdt) == dt

        # Init workspace with a first evaluation of the
        # rhs of the integrator
        for i in xrange(ndt):
            self.time_int_work[:self.nbComponents] = \
                self.timeIntegrator.f(t, self.vorticity.data,
                                      self.time_int_work[:self.nbComponents])

            # perform integration and save result in-place
            self.vorticity.data = \
                self.timeIntegrator(t, self.vorticity.data, subdt[i],
                                    result=self.vorticity.data)
        self._apply_timer.append_time(MPI.Wtime() - ctime)

    def checkStability(self, dt):
        """
        Computes a stability condition depending on some
        diagnostics (from GradVxW)
        @param current time step
        @return int,float : the number of required subcycles and
        the subcycle time-step.
        """
        dt_stab = min(dt, self.cststretch / self.diagnostics[1])
        nb_cycles = int(ceil(dt / dt_stab))
        subdt = npw.zeros((nb_cycles))
        subdt[:] = dt_stab
        subdt[-1] = dt - (nb_cycles - 1) * dt_stab
        return nb_cycles, subdt
