"""
@file monitoring.py
Global interface for monitors.
"""
from abc import ABCMeta, abstractmethod
from parmepy.operator.continuous import Operator
from parmepy.tools.timers import Timer
import parmepy.tools.io_utils as io


class Monitoring(Operator):
    """Abstract interface to monitoring operators."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, variables, topo, io_params=None, task_id=None):
        """ Constructor
        @param variables : list of fields to monitor.
        @param topo : topo on which fields are to be monitored
        @param filename output file name. Default is None ==> no output.
        Full or relative path.
        @param io_params : parameters (dict) to set file output.
        If  None, no output. Set io_params = {} if you want output,
        with default parameters values.
        See parmepy.tools.io_utils.Writer for details
        """
        Operator.__init__(self, variables, topo=topo, task_id=task_id)
        ## Object to store computational times of lower level functions
        self.timer = Timer(self)
        self.requirements = []
        self._isUpToDate = False
        # Output file.
        if io_params is None:
            self._writer = None
        else:
            if topo is not None:
                self._writer = io.Writer(io_params, topo.comm)

    def setUp(self):
        """
        Class-method required
        to fit with operator base-class interface.
        """
        self._isUpToDate = True

    def finalize(self):
        if self._writer is not None:
            self._writer.finalize()

    def discretize(self):
        """
        Does nothing.
        No discretization in monitors, but class-method required
        to fit with operator base-class interface.
        """
        pass

    def addRedistributeRequirement(self, red):
        self.requirements.append(red)

    def getRedistributeRequirement(self):
        return self.requirements
