"""
@file printer.py

File output for field(s) value on a grid.
"""
from parmepy.constants import S_DIR, debug, VTK, HDF5, DATA
from parmepy.operator.monitors.monitoring import Monitoring
import parmepy.tools.numpywrappers as npw
import parmepy.tools.io_utils as io
import os
from numpy import newaxis, float64

try:
    import evtk.hl as evtk
except ImportError as evtk_error:
    evtk = None
try:
    import h5py
except ImportError as h5py_error:
    h5py = None
from parmepy.tools.timers import timed_function


class Printer(Monitoring):
    """
    Print field(s) values on a given topo, in HDF5 or VTK format.
    """
    def __init__(self, variables, topo, prefix=None,
                 frequency=1, formattype=None, task_id=None, xmfalways=False,
                 subset=None):
        """
        Create a results printer for given fields, prefix
        prefix (relative path) and an output frequency.

        @param variables : list of variables to export.
        @param frequency : output rate (output every freq iteration)
        @param prefix output file name. Default is None ==> no output.
        Full or relative path.
        @param formattype : output file format, default=vtk.
        @param xmfalways : true if xmf output must be done every time
        an hdf5 file is created
        """
        Monitoring.__init__(self, variables, topo, task_id=task_id)
        assert frequency > 0
        self.frequency = frequency
        ## Default output type
        if formattype is None:
            self.formattype = VTK
        else:
            self.formattype = formattype
        if self.formattype == VTK and evtk is None:
            print ("You set a printer with VTK as format and evtk module ",)
            print ("is not present. You must specify another format",)
            print (" for output (DATA or HDF5)")
            raise evtk_error
        if self.formattype == HDF5 and h5py is None:
            print ("You set a printer with HDF5 as format and h5py module ",)
            print ("is not present. You must specify another extension",)
            print (" (DATA or VTK)")
            raise h5py_error

        self.input = self.variables
        self.output = []
        # If no prefix is given, set it to
        # the concatenation of variables'names.
        if prefix is None:
            prefix = io.io.defaultPath()
            name = ''
            for var in self.input:
                name += var.name
                name += '_'
            prefix = os.path.join(prefix, name)
        else:
            if not os.path.isabs(prefix):
                prefix = os.path.join(io.io.defaultPath(), prefix)
        ## shortcut to topo name
        self.topo = self._predefinedTopo
        if self.topo is not None:
            io.io.checkDir(prefix, 0, self.topo.comm)
        self.prefix = prefix
        self._xmf_data_files = []
        if self.formattype == VTK:
            # filename = prefix_rk_N, rk = current process rank,
            # N = counter value
            self._get_filename = lambda i: self.prefix + \
                "_{0}_{1:05d}".format(self.topo.rank, i)
            self.step = self._step_VTK
        elif self.formattype == HDF5:
            # filename = prefix_N, N = counter value
            self._get_filename = lambda i: self.prefix + \
                "_{0:05d}".format(i) + '.h5'
            self.step = self._step_HDF5
            if xmfalways:
                self.step = self._step_HDF5_XMF
        elif self.formattype == DATA:
            self._get_filename = lambda i: self.prefix + \
                "_{0}_{1:05d}.dat".format(self.topo.rank, i)
            self.step = self._step_DATA
        self._isUpToDate = True
        # count the number of calls
        self._count = 0
        ## Rank of 'leader' mpi process for output
        ## Forced to 0. \todo : add a param for this?
        self.io_rank = 0

        ## Set a subset of the original domain, to reduce
        ## output
        self.subset = subset
        if self.subset is not None:
            self.subset.discretize(self.topo)
            self._slices = self.subset.slices[self.topo]
            # Global resolution for hdf5 output
            self.globalResolution = self.subset.globalResolution(self.topo)
        else:
            self.globalResolution = list(self.topo.globalMeshResolution - 1)
            self._slices = self.topo.mesh.iCompute
        self.globalResolution.reverse()

    @debug
    @timed_function
    def apply(self, simulation=None):
        if simulation is None:
            raise ValueError("Missing simulation value for monitoring.")

        if simulation.currentIteration == -1 or \
                simulation.currentIteration % self.frequency == 0:
            # Transfer from GPU to CPU if required
            for f in self.variables:
                df = f.discreteFields[self.topo]
                try:
                    if not df.isBatch:
                    # To host only if data fit in the device memory
                        df.toHost()
                        df.wait()
                except AttributeError:
                    pass
            self.step(simulation)
            self._count += 1

    def createXMFFile(self):
        if self.formattype == HDF5 and self.topo.rank == self.io_rank:
            # Write the xmf file driving all h5 files.
            # Writing only one file
            # We have a temporal list of Grid => a single xmf file
            # Manual writing of the xmf file because "XDMF library very
            # difficult to compile and to use"
            #  [Advanced HDF5 & XDMF - Groupe Calcul]
            f = open(self.prefix + '.xmf', 'w')
            f.write("<?xml version=\"1.0\" ?>\n")
            f.write("<!DOCTYPE Xdmf SYSTEM \"Xdmf.dtd\">\n")
            f.write("<Xdmf Version=\"2.0\">\n")
            f.write(" <Domain>\n")
            f.write("  <Grid Name=\"CellTime\" GridType=\"Collection\" ")
            f.write("CollectionType=\"Temporal\">\n")
            for ds_names, i, t in self._xmf_data_files:
                f.write(_TemporalGridXMF(
                        self.topo, ds_names, i, t, self._get_filename(i),
                        self.subset))
            f.write("  </Grid>\n")
            f.write(" </Domain>\n")
            f.write("</Xdmf>\n")
            f.close()

    def finalize(self):
        self.createXMFFile()
        Monitoring.finalize(self)

    def _build_vtk_dict(self):
        """Build a dictionary from fields to VTK image."""
        res = {}

        for f in self.variables:
            df = f.discreteFields[self.topo]
            #ind = df.topology.mesh.iCompute
            for d in xrange(df.nbComponents):
                if f.isVector:
                    name = df.name + S_DIR[d]
                else:
                    name = df.name
                if len(df.data[d].shape) == 2:
                    res[name] = npw.realarray(df.data[d][:, :, newaxis])
                else:
                    res[name] = npw.realarray(df.data[d])
        return res

    def _step_VTK(self, simu):
        """
        """
        orig = [0.] * 3
        dim = self.topo.mesh.dim
        orig[:dim] = [self.topo.mesh.origin[i] for i in xrange(dim)]
        #orig = tuple(orig)
        #ind = self.topo.mesh.local_start
        ## orig = tuple([self.topo.mesh.coords[i].flatten()[ind[i]]
        ##               for i in xrange(self.topo.mesh.dim)])
        spacing = [0] * 3
        spacing[:dim] = [self.topo.mesh.space_step[i] for i in xrange(dim)]
        evtk.imageToVTK(self._get_filename(self._count),
                        origin=orig, spacing=spacing,
                        pointData=self._build_vtk_dict())

    def _step_HDF5(self, simu):
        t = simu.time
        filename = self._get_filename(self._count)
        # Write the h5 file
        # (force np.float64, ParaView seems to not be able to read float32)
        # Writing compressed hdf5 files (gzip compression seems the best)
        # In parallel, one file is written, thus filename no longe contains
        # mpi rank.
        # TODO: Why gzip compression not working in parallel ??
        # Remark: h5py must be build with --mpi option
        if self.topo.size == 1:
            f = h5py.File(filename, "w")
            compression = 'gzip'
        else:
            f = h5py.File(filename, 'w', driver='mpio', comm=self.topo.comm)
            compression = None
        # It's necessary to compute the set of indices of the current subset
        # in global notation
        if self.subset is None:
            # Note : g_start and g_end do not include ghost points.
            g_start = self.topo.mesh.global_start
            g_end = self.topo.mesh.global_end + 1
            sl = [slice(g_start[i], g_end[i])
                  for i in xrange(self.domain.dimension)]
        else:
            g_start = self.subset.gstart
            # convert self._slices to global position in topo
            sl = self.topo.toIndexGlobal(self._slices)
            # And shift using global position of the surface
            sl = [slice(sl[i].start - g_start[i], sl[i].stop - g_start[i])
                  for i in xrange(self.domain.dimension)]
        sl.reverse()
        sl = tuple(sl)
        datasetNames = []
        for field in self.variables:
            df = field.discreteFields[self.topo]
            for d in xrange(df.nbComponents):
                # creating datasets for the vector field
                currentName = df.name + S_DIR[d]
                datasetNames.append(currentName)
                ds = f.create_dataset(currentName,
                                      self.globalResolution,
                                      dtype=float64,
                                      compression=compression)
                # In parallel, each proc must write in the proper part
                # Of the dataset (of site global resolution)
                ds[sl] = npw.realarray(df.data[d][self._slices].T)

        self._xmf_data_files.append((datasetNames, self._count, t))

        f.close()

    def _step_HDF5_XMF(self, simu):
        self._step_HDF5(simu)
        self.createXMFFile()

    def _step_DATA(self, simu):
        f = open(self._get_filename(self._count), 'w')
        shape = self.topo.mesh.resolution
        coords = self.topo.mesh.coords
        pbDimension = self.domain.dimension
        if pbDimension == 2:
            if len(shape) == 2:
                for i in xrange(shape[0] - 1):
                    for j in xrange(shape[1] - 1):
                        f.write("{0:8.12} {1:8.12} ".format(
                                coords[0][i, 0], coords[1][0, j]))
                        for field in self.variables:
                            df = field.discreteFields[self.topo]
                            if field.isVector:
                                f.write("{0:8.12} {1:8.12} ".format(
                                        df[0][i, j],
                                        df[1][i, j]))
                            else:
                                f.write("{0:8.12} ".format(
                                        df[0][i, j]))
                        f.write("\n")
        elif pbDimension == 3:
            for i in xrange(shape[0] - 1):
                for j in xrange(shape[1] - 1):
                    for k in xrange(shape[2] - 1):
                        f.write(
                            "{0:8.12} {1:8.12} {2:8.12} ".format(
                                coords[0][i, 0, 0],
                                coords[1][0, j, 0],
                                coords[2][0, 0, k]))
                        for field in self.variables:
                            df = field.discreteFields[self.topo]
                            if field.isVector:
                                f.write(
                                    "{0:8.12} {1:8.12} " +
                                    "{2:8.12} ".format(
                                        df[0][i, j, k],
                                        df[1][i, j, k],
                                        df[2][i, j, k]))
                            else:
                                f.write("{0:8.12} ".format(
                                        df[0][i, j, k]))
                        f.write("\n")
        else:
            for i in xrange(shape[0] - 1):
                f.write("{0:8.12} ".format(coords[0][i]))
                for field in self.variables:
                    df = field.discreteFields[self.topo]
                    if field.isVector:
                        f.write(
                            "{0:8.12} ".format(df[0][i]))
                    else:
                        f.write("{0:8.12} ".format(df[i]))
                        f.write("\n")
        f.close()


def _listFormat(l):
    """
    Format a list to the xml output.
    Removes the '[]()' and replace ',' with ' ' in default str.
    @param l : list to format
    """
    return str(l).replace(',', ' ').replace('[', '').replace(']', '').replace(
        '(', '').replace(')', '')


def _TemporalGridXMF(topo, datasetNames, ite, t, filename, subset=None):
    g = ""
    dimension = topo.mesh.dim
    if dimension == 2:
        topoType = "2DCORECTMesh"
        geoType = "ORIGIN_DXDY"
    elif dimension == 3:
        topoType = "3DCORECTMesh"
        geoType = "ORIGIN_DXDYDZ"
    g += "   <Grid Name=\"Iteration {0:03d}\"".format(ite)
    g += " GridType=\"Uniform\">\n"
    g += "    <Time Value=\"{0}\" />\n".format(t)
    g += "    <Topology TopologyType=\"" + str(topoType) + "\""
    g += " NumberOfElements=\""
    if subset is not None:
        resolution = subset.gRes
    else:
        resolution = list(topo.globalMeshResolution - 1)
        resolution.reverse()
    g += _listFormat(resolution) + " \"/>\n"
    g += "    <Geometry GeometryType=\"" + geoType + "\">\n"
    g += "     <DataItem Dimensions=\"" + str(dimension) + " \""
    g += " NumberType=\"Float\" Precision=\"4\" Format=\"XML\">\n"
    if subset is not None:
        ori = list(subset.origin)
    else:
        ori = list(topo.domain.origin)
    ori.reverse()
    g += "     " + _listFormat(ori) + "\n"
    g += "     </DataItem>\n"
    g += "     <DataItem Dimensions=\"" + str(dimension) + " \""
    g += " NumberType=\"Float\" Precision=\"8\" Format=\"XML\">\n"
    step = list(topo.mesh.space_step)
    step.reverse()
    g += "     " + _listFormat(step) + "\n"
    g += "     </DataItem>\n"
    g += "    </Geometry>\n"
    for name in datasetNames:
        g += "    <Attribute Name=\""
        g += name + "\""
        g += " AttributeType=\"Scalar\" Center=\"Node\">\n"
        g += "     <DataItem Dimensions=\""
        g += _listFormat(resolution) + " \""
        g += " NumberType=\"Float\" Precision=\"8\" Format=\"HDF\""
        g += " Compression=\"Raw\">\n"  #
        g += "      " + filename.split('/')[-1]
        g += ":/" + name
        g += "\n     </DataItem>\n"
        g += "    </Attribute>\n"
    g += "   </Grid>\n"
    return g
