"""
@file reader.py

File output for field(s) value on a grid.
"""
from parmepy.constants import debug, HDF5
from parmepy.operator.monitors.monitoring import Monitoring
import parmepy.tools.io_utils as io
import os

try:
    import h5py
except ImportError as h5py_error:
    h5py = None
from parmepy.tools.timers import timed_function


class Reader(Monitoring):
    """
    Print field(s) values on a given topo, in HDF5 or VTK format.
    """
    def __init__(self, variables, topo, prefix,
                 formattype=None, task_id=None,
                 subset=None, names=None):
        """
        @param variables : list of variables to read.
        @param prefix input file name (without ext)
        @param formattype : input file format, default=HDF5.
        """
        Monitoring.__init__(self, variables, topo, task_id=task_id)
        ## Default output type
        if formattype is None:
            self.formattype = HDF5
        else:
            self.formattype = formattype
        if self.formattype is not HDF5:
            raise ValueError("Format  not allowed : only HDF5\
                readers are implemented.")
        if self.formattype == HDF5 and h5py is None:
            print ("You set a printer with HDF5 as format and h5py module ",)
            print ("is not present. You must specify another extension",)
            print (" (DATA or VTK)")
            raise h5py_error

        self.input = self.variables
        self.output = self.variables
        if not os.path.isabs(prefix):
            prefix = os.path.join(io.io.defaultPath(), prefix)
        ## shortcut to topo name
        self.topo = self._predefinedTopo
        if self.topo is not None:
            io.io.checkDir(prefix, 0, self.topo.comm)
        self.prefix = prefix
        self._isUpToDate = True
        ## Set a subset of the original domain, to reduce
        ## output
        self.subset = subset
        if self.subset is not None:
            self.subset.discretize(self.topo)
            self._slices = self.subset.slices[self.topo]
            # Global resolution for hdf5 output
            self.globalResolution = self.subset.globalResolution(self.topo)
        else:
            self.globalResolution = list(self.topo.globalMeshResolution - 1)
            self._slices = self.topo.mesh.iCompute
        self.globalResolution.reverse()
        self.step = self.readHDF5
        filename = self.prefix + '.h5'
        assert os.path.isfile(filename), 'error, file does not exists'
        if self.topo.size == 1:
            self._file = h5py.File(filename, "r")
        else:
            self._file = h5py.File(filename, 'r', driver='mpio',
                                   comm=self.topo.comm)
        self._globalSlice = []
        if self.subset is None:
            # Note : g_start and g_end do not include ghost points.
            g_start = self.topo.mesh.global_start
            g_end = self.topo.mesh.global_end + 1
            self._globalSlice = [slice(g_start[i], g_end[i])
                                 for i in xrange(self.domain.dimension)]
        else:
            g_start = self.subset.gstart
            # convert self._slices to global position in topo
            self._globalSlice = self.topo.toIndexGlobal(self._slices)
            # And shift using global position of the surface
            self._globalSlice = [slice(self._globalSlice[i].start - g_start[i],
                                       self._globalSlice[i].stop - g_start[i])
                                 for i in xrange(self.domain.dimension)]
        self._globalSlice.reverse()
        self._globalSlice = tuple(self._globalSlice)
        if names is None:
            self.names = {}
            names = self.dataset_names()
            i = 0
            for field in self.variables:
                self.names[field] = []
                for d in xrange(field.nbComponents):
                    self.names[field].append(names[i])
                    i += 1
        else:
            self.names = names
            for field in self.variables:
                fname = self.names[field]
                self.names[field] = [v for v in self.dataset_names()
                                     if fname in v]
               # assert len(self.names[field]) == field.nbComponents

    @debug
    @timed_function
    def apply(self, simulation=None):
        self.step()

    def dataset_names(self):
        """
        Return the list of available names for datasets in
        the required file.
        """
        return self._file.keys()

    def readHDF5(self):
        # It's necessary to compute the set of indices of the current subset
        # in global notation
        for field in self.variables:
            df = field.discreteFields[self.topo]
            for d in xrange(df.nbComponents):
                # creating datasets for the vector field
                #currentName = df.name + S_DIR[d]
                #if not currentName in f.iterkeys():
                #    raise ValueError("The required field name is \
                #        not in HDF5 file.")
                #Note FP : temp method --> use the first name is dataset
                # as data for field. Todo : set a list of names
                # to be downloaded.
                currentName = self.names[field][d]
                ds = self._file[currentName]
                # In parallel, each proc must write in the proper part
                # Of the dataset (of site global resolution)
                df.data[d][self._slices] = ds[self._globalSlice].T

    def finalize(self):
        Monitoring.finalize(self)
        self._file.close()
