# -*- coding: utf-8 -*-
"""
@file operator/penalization.py

Penalisation of a given field
"""
from parmepy.operator.continuous import Operator
from parmepy.operator.discrete.penalization import Penalization_d
from parmepy.constants import debug


class Penalization(Operator):
    """
    Solves
    \f{eqnarray*}
    v = Op(\v)
    \f} with :
    \f{eqnarray*}
    \frac{\partial v}{\partial t} &=& \lambda\chi_s(v_D - v)
    \f}

    """

    @debug
    def __init__(self, variables, obstacles, coeff,
                 resolutions=None, topo=None,
                 ghosts=None, task_id=None, comm=None):
        """
        Constructor.
        @param[in,out] variables : list of fields to be penalized
        @param[in] obstacle : list of domains where p
        enalization must be applied.
        @param[in] resolutions :  list of resolutions (one per variable)

        """
        Operator.__init__(self, variables, topo=topo, ghosts=ghosts,
                          task_id=task_id, comm=comm)

        ## domain where penalization must be applied.
        ## A parmepy.domain.obstacle .
        ## Obstacle
        if isinstance(obstacles, list):
            self.obstacles = obstacles
        else:
            self.obstacles = [obstacles]
        ## lambda penalization function
        self.coeff = coeff
        self.resolution = resolutions.values()[0]
        # all variables must have the same resolution
        for resol in resolutions.values():
            assert resol == self.resolution
        self.input = self.output = self.variables

    def discretize(self):
        if self._predefinedTopo is not None:
            topo = self._predefinedTopo
        else:
            topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                   self.resolution,
                                                   ghosts=self.ghosts,
                                                   comm=self._comm)
        for v in self.variables:
            # the discrete fields
            self.discreteFields[v] = v.discretize(topo)

            # note that since all fields must have the same
            # resol, there must be one and only one topology
            # for all the variables of this operator.

    @debug
    def setUp(self):
        """
        """
        self.discreteOperator = Penalization_d(self.discreteFields.values(),
                                               self.obstacles, self.coeff)

        self.discreteOperator.setUp()
        self._isUpToDate = True
