"""
@file redistribute_intercomm.py
Setup for data transfer/redistribution between a single parmes topology based
on different MPI communicators with null intersection (for example
by Comm_Split). One of the topology is labeled as the source and the other is
the destination.

It relies on a Bridge_intercomm.
"""
from parmepy.constants import debug, PARMES_MPI_REAL, ORDERMPI, S_DIR, np
from parmepy import __VERBOSE__
from parmepy.operator.continuous import Operator
from parmepy.mpi.topology import Bridge_intercomm
from parmepy.methods_keys import Support


class RedistributeIntercomm(Operator):
    """
    Interconnection between two topologies on different sub set of MPI process.
    SetUp will compute a Bridge_intercomm between a single topology.
    Transfers data from topology of id_from to the id_to.
    """
    @debug
    def __init__(self, variables, op_from, op_to,
                 proc_tasks, parent_comm, topo=None,
                 method=None, component=None, name_suffix=''):
        """
        Create an operator to distribute data between two mpi topologies for a
        list of variables.

        @param variables : the set of variables to be redistributed
        @param topo : Parmes topology that differs across process of the
        parent_comm MPI intracommunicator.
        @param id_from : id of the task considered as input.
        @param id_to : id of the task considered as output.
        @param proc_tasks: python array specifying the task id of each of
        the parent_comm MPI intracommunicator.
        @param parent_comm : Parent communicator (Each process that use this
        operator must be a member of the parent_comm)
        @param component : Component to consider.
        @remark : proc_tasks size and number of processus in parent_comm
        must be equal.
        """
        if not isinstance(variables, list):
            variables = [variables]
        vars_str = "_("
        for vv in variables:
            vars_str += vv.name + ","
        vars_str = vars_str[:-1] + ')'
        if not component is None:
            vars_str += S_DIR[component]
        Operator.__init__(self, variables, method,
                          name_suffix=vars_str+name_suffix)
        assert parent_comm.Get_size() == len(proc_tasks), \
            "Parent communicator ({0})".format(parent_comm.Get_size()) + \
            " and size of the task id array " + \
            "({0}) are not equal".format(len(proc_tasks))
        self.topo = topo
        self.opFrom = op_from
        self.opTo = op_to
        self.id_from = self.opFrom.task_id
        self.id_to = self.opTo.task_id
        self.parent_comm = parent_comm
        self._dim = variables[0].domain.dimension
        self.proc_tasks = proc_tasks
        self.input = self.output = self.variables
        self.component = component
        if component is None:
            # All components are considered
            self._range_components = lambda v: range(v.nbComponents)
        else:
            # Only the given component is considered
            self._range_components = lambda v: [component]

        self.bridges = {}
        self.r_request = {}
        self.s_request = {}
        self._r_types = {}
        self._s_types = {}
        for v in self.variables:
            self._r_types[v] = {}
            self._s_types[v] = {}
        self._toHost_fields = []
        self._toDevice_fields = []
        self._parent_rank = self.parent_comm.Get_rank()
        self._my_rank = None

    def discretize(self):

        for v in self.variables:
            if self.topo is None:
                if self.proc_tasks[self._parent_rank] == self.id_from:
                    self.topo = self.opFrom.discreteFields[v].topology
                else:
                    self.topo = self.opTo.discreteFields[v].topology

        self._my_rank = self.topo.comm.Get_rank()
        self._dim = self.topo.domain.dimension

        for v in self.variables:
            self.discreteFields[v] = v.discretize(self.topo)

    @debug
    def setUp(self):
        """
        Computes intersection of topologies and set the MPI intercommunicator.
        """
        assert self.topo.isUpToDate, \
            """You should setup topology
            before any attempt to setup a redistribute operator."""

        # Look for an operator opertating on device.
        try:
            opFrom_is_device = \
                self.opFrom.method[Support].find('gpu') >= 0
        except KeyError:  # op.method is a dict not containing Support in keys
            opFrom_is_device = False
        except IndexError:  # op.method is a sting
            opFrom_is_device = False
        except TypeError:  # op.method is None
            opFrom_is_device = False
        try:
            opTo_is_device = \
                self.opTo.method[Support].find('gpu') >= 0
        except KeyError:  # op.method is a dict not containing Support in keys
            opTo_is_device = False
        except IndexError:  # op.method is a sting
            opTo_is_device = False
        except TypeError:  # op.method is None
            opTo_is_device = False

        if not opFrom_is_device and not opTo_is_device:
            # case: opFrom(host) --bridge--> opTo(host)
            self._the_apply = self._apply_host
        else:
            # Have on device operators
            if opFrom_is_device and not opTo_is_device:
                # case: opFrom(GPU) --toHost--bridge--> opTo(host)
                self._the_apply = self._apply_toHost_host
            elif not opFrom_is_device and opTo_is_device:
                # case: opFrom(host) --bridge--toDevice--> opTo(GPU)
                self._the_apply = self._apply_host_toDevice
            else:
                # case: opFrom(GPU) --toHost--bridge--toDevice--> opTo(host)
                # Transfers are removed if variables are batched
                if np.any([self.opFrom.discreteFields[v].isBatch
                           for v in self.variables] +
                          [self.opTo.discreteFields[v].isBatch
                           for v in self.variables]):
                    self._the_apply = self._host
                else:
                    self._the_apply = self._apply_toHost_host_toDevice

        # Build bridges and toTransfer lists
        self.bridge = Bridge_intercomm(self.topo, self.parent_comm,
                                       self.id_from, self.id_to,
                                       self.proc_tasks)

        for v in self.variables:
            # toTransfer list completion
            if self.proc_tasks[self._parent_rank] == self.id_from:
                if opFrom_is_device:
                    self._toHost_fields.append(self.opFrom.discreteFields[v])
            if self.proc_tasks[self._parent_rank] == self.id_to:
                if opTo_is_device:
                    self._toDevice_fields.append(self.opTo.discreteFields[v])

        for v in self.variables:
            dv = v.discreteFields[self.topo]
            transfers = self.bridge.transfers
            # Set reception
            if self.proc_tasks[self._parent_rank] == self.id_to:
                for from_rk in transfers.keys():
                    subshape = tuple(
                        [transfers[from_rk][i][1] - transfers[from_rk][i][0]
                         for i in range(self._dim)])
                    substart = tuple(
                        [transfers[from_rk][i][0] for i in range(self._dim)])
                    self._r_types[v][from_rk] = \
                        PARMES_MPI_REAL.Create_subarray(dv.data[0].shape,
                                                        subshape,
                                                        substart,
                                                        order=ORDERMPI)
                    self._r_types[v][from_rk].Commit()
            # Set Sending
            if self.proc_tasks[self._parent_rank] == self.id_from:
                for to_rk in transfers.keys():
                    subshape = tuple(
                        [transfers[to_rk][i][1] - transfers[to_rk][i][0]
                         for i in range(self._dim)])
                    substart = tuple(
                        [transfers[to_rk][i][0] for i in range(self._dim)])
                    self._r_types[v][to_rk] = \
                        PARMES_MPI_REAL.Create_subarray(dv.data[0].shape,
                                                        subshape,
                                                        substart,
                                                        order=ORDERMPI)
                    self._r_types[v][to_rk].Commit()
        self._isUpToDate = True

    @debug
    def apply(self, simulation=None):
        """
        Apply this operator to its variables.
        @param simulation : object that describes the simulation
        parameters (time, time step, iteration number ...), see
        parmepy.problem.simulation.Simulation for details.
        """
        for req in self.requirements:
            req.wait()
        self._the_apply(simulation)

    def _apply_toHost_host_toDevice(self, simulation=None):
        if __VERBOSE__:
            print ("{0} APPLY toHOST+HOST+toDEVICE".format(self._parent_rank))
        if self.proc_tasks[self._parent_rank] == self.id_from:
            self._toHost()
            self._wait_device()
        self._host()
        self._wait_host()
        if self.proc_tasks[self._parent_rank] == self.id_to:
            self._toDevice()
            self._wait_device()

    def _apply_toHost_host(self, simulation=None):

        if __VERBOSE__:
            print ("{0} APPLY toHOST+HOST".format(self._parent_rank))
        if self.proc_tasks[self._parent_rank] == self.id_from:
            self._toHost()
            self._wait_device()
        self._host()
        self._wait_host()

    def _apply_host_toDevice(self, simulation=None):
        if __VERBOSE__:
            print ("{0} APPLY HOST+toDEVICE".format(self._parent_rank))
        self._host()
        self._wait_host()
        self.parent_comm.Barrier()
        if self.proc_tasks[self._parent_rank] == self.id_to:
            self._toDevice()
            self._wait_device()
        self.parent_comm.Barrier()

    def _apply_host(self, simulation=None):
        if __VERBOSE__:
            print ("{0} APPLY HOST".format(self._parent_rank))
        self._host()
        self._wait_host()

    def _host(self, simulation=None):
        """
        Proceed with data redistribution from opFrom to opTo
        """
        self.parent_comm.Barrier()
        self.r_request = {}
        self.s_request = {}
        for v in self.variables:
            dv = v.discreteFields[self.topo]
            transfers = self.bridge.transfers
            for d in self._range_components(v):
                v_name = dv.name + S_DIR[d]
                # Set reception
                if self.proc_tasks[self._parent_rank] == self.id_to:
                    for from_rk in transfers.keys():
                        self.r_request[v_name + str(from_rk)] = \
                            self.bridge.inter_comm.Irecv(
                                [dv.data[d], 1, self._r_types[v][from_rk]],
                                source=from_rk, tag=from_rk)
                # Set Sending
                if self.proc_tasks[self._parent_rank] == self.id_from:
                    for to_rk in transfers.keys():
                        self.s_request[v_name + str(to_rk)] = \
                            self.bridge.inter_comm.Issend(
                                [dv.data[d], 1, self._r_types[v][to_rk]],
                                dest=to_rk, tag=self._my_rank)

    def _toHost(self):
        """
        Proceed with data transfer of variables from device to host
        """
        if __VERBOSE__:
            print ("{0} APPLY toHOST".format(self._parent_rank))
        for v in self.variables:
            dv = self.opFrom.discreteFields[v]
            if dv in self._toHost_fields:
                dv.toHost(self.component)

    def _toDevice(self):
        """
        Proceed with data transfer of variables from device to host
        """
        if __VERBOSE__:
            print ("{0} APPLY toDEVICE".format(self._parent_rank))
        for v in self.variables:
            dv = self.opTo.discreteFields[v]
            if dv in self._toDevice_fields:
                dv.toDevice(self.component)

    def _wait_device(self):
        if __VERBOSE__:
            print ("{0} WAIT OPENCL".format(self._parent_rank))
        for dv in self._toDevice_fields + self._toHost_fields:
            dv.wait()

    def _wait_host(self, simulation=None):
        """Wait for requests completion."""
        if __VERBOSE__:
            print ("{0} WAIT MPI".format(self._parent_rank))
        for rk in self.r_request:
            self.r_request[rk].Wait()
        for rk in self.s_request:
            self.s_request[rk].Wait()
        self.parent_comm.Barrier()
        self.r_request = []
        self.s_request = []

    def test(self, rsend=None, rrecv=None):
        """
        if neither rsend or rrecv is given return
        True if all communication request are complete
        else check for sending to rsend or
        receiving from rrecv. Process ranks
        should be given in local communicator.
        @param rsend : variable name + S_DIR + rank
        @param rrecv : variable name + S_DIR + rank
        """
        if(rsend is not None or rrecv is not None):
            send_res = True
            recv_res = True
            if rsend is not None:
                send_res = self.s_request[rsend].Test()
            if rrecv is not None:
                recv_res = self.r_request[rrecv].Test()
            res = send_res and recv_res
        else:
            res = True
            for rk in self.r_request.keys():
                res = self.r_request[rk].Test()
                if not res:
                    return res
            for rk in self.s_request.keys():
                res = self.s_request[rk].Test()
                if not res:
                    return res
        return res
