# -*- coding: utf-8 -*-
"""
@file reprojection_criterion.py
Compute reprojection criterion and divergence maximum
"""
import numpy as np
from parmepy.constants import debug, PARMES_MPI_REAL
from parmepy.methods_keys import SpaceDiscretisation
from parmepy.operator.monitors.monitoring import Monitoring
from parmepy.numerics.finite_differences import FD_C_4
from parmepy.numerics.differential_operations import GradV
import parmepy.tools.numpywrappers as npw
from parmepy.numerics.updateGhosts import UpdateGhosts
from parmepy.mpi import MPI
from parmepy.tools.timers import timed_function


class Reprojection_criterion(Monitoring):
    """
    Computes and prints reprojection criterion.
    See the related PDF called "vorticity_solenoidal_projection.pdf"
    in ParmesDoc for more details.
    """
    def __init__(self, vorticity, reproj_cst, reprojRate,
                 topo, checkCriterion=False, io_params=None,
                 method=None, task_id=None):
        """
        Constructor.
        @param vorticity field
        @param method : finite difference scheme
        @param topo : the topology on which we want to monitor the fields
        @param prefix : output file name.
        @param checkCriterion :  set behavior of this operator :
        if True, compute some criterion and force frequency
        to 1 if reprojection is needed.
        else (false) performs reprojection every reprojRate iterations.
        @param reprojRate : set rate of execution of the reprojection
        @param io_params : parameters (dict) to set file output.
        If  None, no output. Set io_params = {} if you want output,
        with default parameters values. Default file name = 'reproj'
        See parmepy.tools.io_utils.Writer for details
        """
        if io_params is not None:
            if not "filename" in io_params:
                io_params["filename"] = "reproj"
            # Set output buffer shape
            io_params["writebuffshape"] = (1, 5)

        Monitoring.__init__(self, [vorticity], topo, io_params,
                            task_id=task_id)
        # alias for local topo
        self._topo = self._predefinedTopo
        # \todo : rewrite for multiresolution case.
        # Note FP : for multiresolution case, it would probably be
        # better to use two different operators for energy and enstrophy.
        ## Frequency for reprojection
        self.reprojRate = reprojRate
        ## The initial value will be used as default during
        # simulation
        self.default_rate = reprojRate
        ## Set behavior of this operator :
        ## if True, compute some criterion and force frequency
        ## to 1 if reprojection is needed.
        ## else (false) performs reprojection every frequency iterations.
        self.checkCriterion = checkCriterion
        # constant defining the reprojection criterion :
        # if the latter is greater than this constant, then a reprojection
        # is needed
        self.reproj_cst = reproj_cst
        self.proj_counter = 0
        ## local buffer
        ## diag = [time, projCriterion, d1, d2, proj_counter]
        ## See apply for details about d1, d2.
        self.diagnostics = npw.zeros(5)
        # Connect writer buffer to diagnostics, if required
        if self._writer is not None:
            self._writer.buffer = self.diagnostics.reshape(1, 5)

        if self.checkCriterion:
            ## vorticity field
            self.vorticity = vorticity
            if method is None:
                method = {SpaceDiscretisation: FD_C_4}
            ## Numerical methods for space discretization
            assert SpaceDiscretisation in method.keys()
            self.method = method[SpaceDiscretisation]

            self.input = [vorticity]
        else:
            self.input = []

        self.output = []

    def setUp(self):
        if not self._isUpToDate and self.checkCriterion:
            # Get discrete fields for vorticity.
            # Note FP : two options to get the discrete fields:
            # - 'field.discretization(topo)' that just try
            # to get the discrete field and return an error
            # if it does not exist.
            # - 'field.discretizet(topo)
            # try to get the discrete field and create
            # a new discretization if it does not exist.
            # Current choice : no creation.
            self.discreteFields[self.vorticity] =\
                self.vorticity.discretization(self._topo)

            # prepare ghost points synchro for vorticity
            self._synchronize = UpdateGhosts(self._topo,
                                             self.vorticity.nbComponents)

            # grad function
            self._function = GradV(self._topo, method=self.method)

            self.vort_d = self.discreteFields[self.vorticity]

            memshape = self.vort_d.data[0].shape
            worklength = self.vort_d.nbComponents ** 2
            # gradient result array
            self.grad = [npw.zeros(memshape) for i in xrange(worklength)]

        self._isUpToDate = True

    @debug
    @timed_function
    def apply(self, simulation=None):
        """
        Computes and prints reprojection criterion and divergence maximum
        """
        self.diagnostics[0] = simulation.time
        ite = simulation.currentIteration

        # Reset reprojection frequency to default
        self.reprojRate = self.default_rate

        # If required, computation of a criterion
        # and reset frequency
        if self.checkCriterion:
            # Synchronize ghost points of vorticity
            self._synchronize(self.vort_d.data)
            # gradU computation
            self.grad = self._function(self.vort_d.data, self.grad)
            nbComponents = self.vorticity.nbComponents
            # maxima of vorticity divergence (abs)
            self.diagnostics[2] = \
                np.max(abs(sum([(self.grad[(nbComponents + 1) * i])
                               for i in xrange(nbComponents)])))

            # maxima of partial derivatives of vorticity
            for grad_n in self.grad:
                self.diagnostics[3] = max(self.diagnostics[3],
                                          np.max(abs(grad_n)))

            # computation of the reprojection criterion and reduction
            # among all proc.
            projCriterion = self.diagnostics[2] / self.diagnostics[3]
            projCriterion = \
                self._topo.comm.allreduce(projCriterion, PARMES_MPI_REAL,
                                          op=MPI.MAX)
            self.diagnostics[1] = projCriterion

            # is reprojection of vorticity needed for the future time step ?
            if (projCriterion > self.reproj_cst):
                self.reprojRate = 1

        # Note FP : is counter really useful? Maybe it should be increased
        # by the operator that use projection (poisson)?
        if (ite % self.reprojRate == 0):
            self.proj_counter += 1
        self.diagnostics[4] = self.proj_counter

        # Print results, if required
        # Remark : writer buffer is (pointer) connected to diagnostics
        if self._writer is not None and self._writer.doWrite(ite):
            self._writer.write()

    def doProjection(self, ite):
        """
        True if projection must be done
        """
        return ite % self.reprojRate == 0
