# -*- coding: utf-8 -*-
"""
@file operator/stretching.py

Computation of stretch. term in Navier-Stokes.

"""
from parmepy.constants import debug
from parmepy.methods_keys import TimeIntegrator, Formulation, \
    SpaceDiscretisation
from parmepy.numerics.finite_differences import FD_C_4
from parmepy.operator.continuous import Operator
from parmepy.operator.discrete.stretching import Conservative, GradUW
import numpy as np


class Stretching(Operator):
    """
    \todo write latex formulas
    """

    @debug
    def __init__(self, velocity, vorticity, resolutions,
                 method=None, topo=None, ghosts=None,
                 comm=None, task_id=None):
        """
        Create a Stretching operator from given
        velocity and vorticity variables.
        'FD_C4 RK3',

        @param velocity field
        @param vorticity field
        @param resolutions : grid resolution of velocity and vorticity
        @param method : solving method
        (default = finite differences, 4th order, in space
        and Runge-Kutta 3 in time.)
        @param topo : a predefined topology to discretize velocity/vorticity
        @param ghosts : number of ghosts points. Default depends on the method.
        Autom. computed if not set.
        """
        Operator.__init__(self, [velocity, vorticity], method, topo=topo,
                          ghosts=ghosts, task_id=task_id, comm=comm)
        ## velocity variable (vector)
        self.velocity = velocity
        ## vorticity variable (vector)
        self.vorticity = vorticity
        ## Grid resolution for each variable (dictionnary)
        self.resolutions = resolutions
        ## Numerical methods for time and space discretization

        if method is None:
            import parmepy.default_methods as default
            method = default.STRETCHING
        self.method = method
        assert Formulation in self.method.keys()
        assert SpaceDiscretisation in self.method.keys()
        assert TimeIntegrator in self.method.keys()

        ## Formulation used for the stretching equation.
        ## Default = conservative form.
        if self.method[Formulation] == "GradUW":
            self.formulation = GradUW
        else:
            self.formulation = Conservative

        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]

    def discretize(self):
        if self.method[SpaceDiscretisation] is FD_C_4:
            nbGhosts = 2
        else:
            raise ValueError("Unknown method for space discretization of the\
                stretching operator.")

        if self._predefinedTopo is not None:
            assert (self._predefinedTopo.ghosts >= nbGhosts).all()
            topo = self._predefinedTopo
            for v in self.variables:
                self.discreteFields[v] = v.discretize(topo)

        else:
            # Variables discretization
            if self.ghosts is not None:
                self.ghosts[self.ghosts < nbGhosts] = nbGhosts
            else:
                self.ghosts = np.ones((self.domain.dimension)) * nbGhosts

            for v in self.variables:
                topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                       self.resolutions[v],
                                                       ghosts=self.ghosts,
                                                       comm=self._comm)
                self.discreteFields[v] = v.discretize(topo)

    @staticmethod
    def getWorkLengths(method=None, domain_dim=None):
        """
        Return the length of working arrays lists required
        for stretching discrete operator, depending on :
        - the formulation (Conservative or GradUW)
        - the time integrator (RK3, ...)
        @param method : the dict of parameters for the operator.
        Default = {TimeIntegrator: RK3, Formulation: Conservative,
                      SpaceDiscretisation: FD_C_4}
        """
        if method is None:
            import parmepy.default_methods as default
            method = default.STRETCHING
        assert Formulation in method,\
            'A formulation is required for the stretching.'
        assert TimeIntegrator in method,\
            'A time integrator is required for the stretching.'
        if method[Formulation] == "GradUW":
            formulation = GradUW
        else:
            formulation = Conservative
        return formulation.getWorkLengths(method[TimeIntegrator])

    def setWorks(self, rwork=None, iwork=None):
        if rwork is None:
            rwork = []
        if iwork is None:
            iwork = []
        self.discreteOperator.setWorks(rwork, iwork)

    @debug
    def setUp(self):
        self.discreteOperator =\
            self.formulation(self.discreteFields[self.velocity],
                             self.discreteFields[self.vorticity],
                             method=self.method)

        self.discreteOperator.setUp()
        self._isUpToDate = True
