"""
@file parmepy.operator.tests.test_analytic
Test initialization of fields with analytic formula
---> either field.initialize or analytic operator description.
"""
import numpy as np
from parmepy.domain.box import Box
from parmepy.fields.continuous import Field
from parmepy.operator.analytic import Analytic
from parmepy.problem.simulation import Simulation
from parmepy.mpi.topology import Cartesian


def func_scal_1(res, x, y, z, t):
    res[0][...] = x - 0.1 * y + 10. * z * z * t
    return res


def func_scal_2(x, y, z, t):
    f = x - 0.1 * y + 10. * z * z * t
    return f


def func_vec_1(res, x, y, z, t):
    res[0][...] = x
    res[1][...] = 0.1 * y
    res[2][...] = 10. * z * z
    return res


def func_vec_2(x, y, z, t):
    f_x = x
    f_y = 0.1 * y
    f_z = 10. * z * z
    return f_x, f_y, f_z


def func_vec_3(res, x, y, z, t, theta):
    res[0][...] = x + theta
    res[1][...] = 0.1 * y
    res[2][...] = 10. * z * z
    return res


def func_vec_4(x, y, z, t, theta):
    f_x = x + theta
    f_y = 0.1 * y
    f_z = 10. * z * z
    return f_x, f_y, f_z


def func_vec_5(res, x, y, z, t, theta):
    res[0][...] = x + theta
    res[1][...] = 0.1 * y
    res[2][...] = 10. * z * z
    res[3][...] = theta * z
    return res


def func_vec_6(res, x, y, t, theta):
    res[0][...] = x + theta
    res[1][...] = 0.1 * y
    res[2][...] = 10. * y
    res[3][...] = theta * t
    return res


res3D = [33, 33, 33]
res2D = [33, 33]
L2D = [1., 1.]
origin2D = [0., 0.]
nbc = 4
simu = Simulation(tinit=0., tend=0.1, nbIter=1)

## I - Tests with direct calls of field.initialize


# Non-Vectorized formula for a scalar
def test_analytical_field_1():
    box = Box()
    topo = Cartesian(box, 3, res3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_scal_1)
    ref = Field(box)
    caf.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = id(cafd.data[0])
    caf.initialize()
    refd.data = func_scal_1(refd.data, *(coords + (0.,)))
    assert np.allclose(cafd[0], refd.data[0])
    assert id(cafd.data[0]) == ids
    time = 3.0
    caf.initialize(currentTime=time)
    refd.data = func_scal_1(refd.data, *(coords + (time,)))
    assert np.allclose(cafd[0], refd.data[0])
    assert id(cafd.data[0]) == ids


# Vectorized formula
def test_analytical_field_2():
    box = Box()
    topo = Cartesian(box, 3, res3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_scal_2, doVectorize=True)
    ref = Field(box)
    caf.setTopoInit(topo)
    ref.setTopoInit(topo)
    cafd = caf.discretize(topo)
    ids = id(cafd.data[0])
    refd = ref.discretize(topo)
    caf.initialize()
    refd.data = func_scal_1(refd.data, *(coords + (0.,)))
    assert np.allclose(cafd[0], refd.data[0])
    assert id(cafd.data[0]) == ids
    time = 3.0
    caf.initialize(currentTime=time)
    refd.data = func_scal_1(refd.data, *(coords + (time,)))
    assert np.allclose(cafd[0], refd.data[0])
    assert id(cafd.data[0]) == ids


# Non-Vectorized formula for a vector
def test_analytical_field_3():
    box = Box()
    topo = Cartesian(box, 3, res3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_1, isVector=True)
    ref = Field(box, isVector=True)
    caf.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = np.zeros((3))
    for i in xrange(3):
        ids[i] = id(cafd.data[i])
    caf.initialize()
    refd.data = func_vec_1(refd.data, *(coords + (0.,)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(currentTime=time)
    refd.data = func_vec_1(refd.data, *(coords + (time,)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Vectorized formula for a vector
def test_analytical_field_4():
    box = Box()
    topo = Cartesian(box, 3, res3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_2, isVector=True, doVectorize=True)
    ref = Field(box, isVector=True)
    caf.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = np.zeros((3))
    for i in xrange(3):
        ids[i] = id(cafd.data[i])
    caf.initialize()
    refd.data = func_vec_1(refd.data, *(coords + (0.,)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(currentTime=time)
    refd.data = func_vec_1(refd.data, *(coords + (time,)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Non-Vectorized formula for a vector, with extra-arguments
def test_analytical_field_5():
    box = Box()
    topo = Cartesian(box, 3, res3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_3, isVector=True)
    theta = 0.3
    caf.setExtraParameters(theta)
    ref = Field(box, isVector=True)
    caf.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = np.zeros((3))
    for i in xrange(3):
        ids[i] = id(cafd.data[i])
    caf.initialize()
    refd.data = func_vec_3(refd.data, *(coords + (0., theta)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(currentTime=time)
    refd.data = func_vec_3(refd.data, *(coords + (time, theta)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Vectorized formula for a vector, with extra-arguments
def test_analytical_field_6():
    box = Box()
    topo = Cartesian(box, 3, res3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_4, isVector=True, doVectorize=True)
    theta = 0.3
    caf.setExtraParameters(theta)
    ref = Field(box, isVector=True)
    caf.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = np.zeros((3))
    for i in xrange(3):
        ids[i] = id(cafd.data[i])
    caf.initialize()
    refd.data = func_vec_3(refd.data, *(coords + (0., theta)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(currentTime=time)
    refd.data = func_vec_3(refd.data, *(coords + (time, theta)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Non-Vectorized formula for a field with nbComponents
# different from domain dim and  with extra-arguments
def test_analytical_field_7():
    box = Box()
    topo = Cartesian(box, 3, res3D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_5, nbComponents=nbc)
    theta = 0.3
    caf.setExtraParameters(theta)
    ref = Field(box, nbComponents=nbc)
    caf.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = np.zeros((nbc))
    for i in xrange(nbc):
        ids[i] = id(cafd.data[i])

    caf.initialize()
    refd.data = func_vec_5(refd.data, *(coords + (0., theta)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(currentTime=time)
    refd.data = func_vec_5(refd.data, *(coords + (time, theta)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Non-Vectorized formula for a 2D field with nbComponents
# different from domain dim and  with extra-arguments
def test_analytical_field_8():
    box = Box(dimension=2, length=L2D, origin=origin2D)
    topo = Cartesian(box, 2, res2D)
    coords = topo.mesh.coords
    caf = Field(box, formula=func_vec_6, nbComponents=nbc)
    theta = 0.3
    caf.setExtraParameters(theta)
    ref = Field(box, nbComponents=nbc)
    caf.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discretize(topo)
    refd = ref.discretize(topo)
    ids = np.zeros((nbc))
    for i in xrange(nbc):
        ids[i] = id(cafd.data[i])
    caf.initialize()
    refd.data = func_vec_6(refd.data, *(coords + (0., theta)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
    time = 3.0
    caf.initialize(currentTime=time)
    refd.data = func_vec_6(refd.data, *(coords + (time, theta)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


## II - Same initializations as I but with call to analytic operator.

# Non-Vectorized and vectorized formulas for a scalar
def test_analytical_op_1():
    box = Box()
    caf = Field(box, formula=func_scal_1)
    caf2 = Field(box, formula=func_scal_2, doVectorize=True)
    op = Analytic(caf, resolutions={caf: res3D})
    op2 = Analytic(caf2, resolutions={caf2: res3D})
    op.discretize()
    op2.discretize()
    op.setUp()
    op2.setUp()
    topo = op.discreteFields[caf].topology
    coords = topo.mesh.coords
    ref = Field(box)
    caf.setTopoInit(topo)
    caf2.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discreteFields[topo]
    cafd2 = caf2.discreteFields[topo]
    ids = id(cafd.data[0])
    ids2 = id(cafd2.data[0])
    op.apply(simu)
    op2.apply(simu)
    refd.data = func_scal_1(refd.data, *(coords + (simu.time,)))
    assert np.allclose(cafd[0], refd.data[0])
    assert id(cafd.data[0]) == ids
    assert np.allclose(cafd2[0], refd.data[0])
    assert id(cafd2.data[0]) == ids2


# Non-Vectorized and vectorized formulas for a vector
def test_analytical_op_3():
    box = Box()
    caf = Field(box, formula=func_vec_1, isVector=True)
    caf2 = Field(box, formula=func_vec_2, doVectorize=True, isVector=True)
    op = Analytic(caf, resolutions={caf: res3D})
    op2 = Analytic(caf2, resolutions={caf2: res3D})
    op.discretize()
    op2.discretize()
    op.setUp()
    op2.setUp()
    topo = op.discreteFields[caf].topology
    coords = topo.mesh.coords
    ref = Field(box, isVector=True)
    caf.setTopoInit(topo)
    caf2.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discreteFields[topo]
    cafd2 = caf2.discreteFields[topo]
    ids = np.zeros((3))
    ids2 = np.zeros((3))
    for i in xrange(3):
        ids[i] = id(cafd.data[i])
        ids2[i] = id(cafd2.data[i])
    op.apply(simu)
    op2.apply(simu)
    refd.data = func_vec_1(refd.data, *(coords + (simu.time,)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
        assert np.allclose(cafd2[i], refd.data[i])
        assert id(cafd2.data[i]) == ids2[i]


# Non-Vectorized and vectorized formulas for a vector with extra-args
def test_analytical_op_4():
    box = Box()
    caf = Field(box, formula=func_vec_3, isVector=True)
    caf2 = Field(box, formula=func_vec_4, doVectorize=True, isVector=True)
    op = Analytic(caf, resolutions={caf: res3D})
    op2 = Analytic(caf2, resolutions={caf2: res3D})
    op.discretize()
    op2.discretize()
    op.setUp()
    op2.setUp()
    topo = op.discreteFields[caf].topology
    coords = topo.mesh.coords
    ref = Field(box, isVector=True)
    caf.setTopoInit(topo)
    caf2.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discreteFields[topo]
    cafd2 = caf2.discreteFields[topo]
    ids = np.zeros((3))
    ids2 = np.zeros((3))
    for i in xrange(3):
        ids[i] = id(cafd.data[i])
        ids2[i] = id(cafd2.data[i])
    theta = 3.
    caf.setExtraParameters(theta)
    caf2.setExtraParameters(theta)
    op.apply(simu)
    op2.apply(simu)
    refd.data = func_vec_3(refd.data, *(coords + (simu.time, theta)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]
        assert np.allclose(cafd2[i], refd.data[i])
        assert id(cafd2.data[i]) == ids2[i]


# Non-Vectorized formula for a nbc components field with extra-args
def test_analytical_op_5():
    box = Box()
    caf = Field(box, formula=func_vec_5, nbComponents=nbc)
    op = Analytic(caf, resolutions={caf: res3D})
    op.discretize()
    op.setUp()
    topo = op.discreteFields[caf].topology
    coords = topo.mesh.coords
    ref = Field(box, nbComponents=nbc)
    caf.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discreteFields[topo]
    ids = np.zeros((nbc))
    for i in xrange(nbc):
        ids[i] = id(cafd.data[i])
    theta = 3.
    caf.setExtraParameters(theta)
    op.apply(simu)
    refd.data = func_vec_5(refd.data, *(coords + (simu.time, theta)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]


# Non-Vectorized formula for a nbc components field in 2D, with extra-args
def test_analytical_op_6():
    box = Box(dimension=2, length=L2D, origin=origin2D)
    caf = Field(box, formula=func_vec_6, nbComponents=nbc)
    op = Analytic(caf, resolutions={caf: res2D})
    op.discretize()
    op.setUp()
    topo = op.discreteFields[caf].topology
    coords = topo.mesh.coords
    ref = Field(box, nbComponents=nbc)
    caf.setTopoInit(topo)
    ref.setTopoInit(topo)
    refd = ref.discretize(topo)
    cafd = caf.discreteFields[topo]
    ids = np.zeros((nbc))
    for i in xrange(nbc):
        ids[i] = id(cafd.data[i])
    theta = 3.
    caf.setExtraParameters(theta)
    op.apply(simu)
    refd.data = func_vec_6(refd.data, *(coords + (simu.time, theta)))
    for i in xrange(caf.nbComponents):
        assert np.allclose(cafd[i], refd.data[i])
        assert id(cafd.data[i]) == ids[i]

if __name__ == "__main__":
    test_analytical_field_1()
    test_analytical_field_2()
    test_analytical_field_3()
    test_analytical_field_4()
    test_analytical_field_5()
    test_analytical_field_6()
    test_analytical_field_7()
    test_analytical_field_8()
    test_analytical_op_1()
    test_analytical_op_3()
    test_analytical_op_4()
    test_analytical_op_5()
    test_analytical_op_6()
