"""
@file parmepy.operator.tests.test_particle_advection
Testing pure python particle advection with null velocity.
"""
from parmepy.domain.box import Box
from parmepy.fields.continuous import Field
from parmepy.operator.advection import Advection
from parmepy.constants import ORDER, PARMES_REAL
from parmepy.problem.simulation import Simulation
import parmepy.tools.numpywrappers as npw
import numpy as np


def setup_2D():
    box = Box(2, length=[2., 2.], origin=[-1., -1.])
    scal = Field(domain=box, name='Scalar')
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y: (0., 0.), isVector=True)
    return scal, velo


def setup_vector_2D():
    box = Box(2, length=[2., 2.], origin=[-1., -1.])
    scal = Field(domain=box, name='Vector', isVector=True)
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y: (0., 0.), isVector=True)
    return scal, velo


def setup_list_2D():
    box = Box(2, length=[2., 2.], origin=[-1., -1.])
    scal1 = Field(domain=box, name='Scalar1')
    scal2 = Field(domain=box, name='Scalar2')
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y: (0., 0.), isVector=True)
    return [scal1, scal2], velo


def setup_3D():
    box = Box(3, length=[2., 4., 1.], origin=[-1., -2., 0.])
    scal = Field(domain=box, name='Scalar')
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z: (0., 0., 0.), isVector=True)
    return scal, velo


def setup_vector_3D():
    box = Box(3, length=[2., 4., 1.], origin=[-1., -2., 0.])
    scal = Field(domain=box, name='Scalar', isVector=True)
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z: (0., 0., 0.), isVector=True)
    return scal, velo


def setup_list_3D():
    box = Box(3, length=[2., 4., 1.], origin=[-1., -2., 0.])
    scal1 = Field(domain=box, name='Scalar1')
    scal2 = Field(domain=box, name='Scalar2')
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z: (0., 0., 0.), isVector=True)
    return [scal1, scal2], velo


def assertion(scal, advec):
    advec.discretize()
    advec.setUp()
    scal_d = scal.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    scal_init = npw.copy(scal_d.data[0])

    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    return np.allclose(scal_init, scal_d.data[0])


def assertion_vector2D(scal, advec):
    advec.discretize()
    advec.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    scal_d.data[1][...] = np.asarray(np.random.random(scal_d.data[1].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    scal1_init = npw.copy(scal_d.data[0])
    scal2_init = npw.copy(scal_d.data[1])

    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    print (np.max(np.abs((scal1_init - scal_d.data[0]))))
    print (np.max(np.abs((scal2_init - scal_d.data[1]))))
    return np.allclose(scal1_init, scal_d.data[0]) and \
        np.allclose(scal2_init, scal_d.data[1])


def assertion_vector3D(scal, advec):
    advec.discretize()
    advec.setUp()

    scal_d = scal.discreteFields.values()[0]
    scal_d.data[0][...] = np.asarray(np.random.random(scal_d.data[0].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    scal_d.data[1][...] = np.asarray(np.random.random(scal_d.data[1].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    scal_d.data[2][...] = np.asarray(np.random.random(scal_d.data[2].shape),
                                     dtype=PARMES_REAL, order=ORDER)
    scal1_init = npw.copy(scal_d.data[0])
    scal2_init = npw.copy(scal_d.data[1])
    scal3_init = npw.copy(scal_d.data[2])

    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    return np.allclose(scal1_init, scal_d.data[0]) and \
        np.allclose(scal2_init, scal_d.data[1]) and \
        np.allclose(scal3_init, scal_d.data[2])


def assertion_list(scal, advec):
    advec.discretize()
    advec.setUp()

    scal1_d = scal[0].discreteFields.values()[0]
    scal2_d = scal[1].discreteFields.values()[0]
    scal1_d.data[0][...] = np.asarray(np.random.random(scal1_d.data[0].shape),
                                      dtype=PARMES_REAL, order=ORDER)
    scal2_d.data[0][...] = np.asarray(np.random.random(scal2_d.data[0].shape),
                                      dtype=PARMES_REAL, order=ORDER)
    scal1_init = npw.copy(scal1_d.data[0])
    scal2_init = npw.copy(scal2_d.data[0])

    advec.apply(Simulation(tinit=0., tend=0.01, nbIter=1))
    print (scal1_init, scal1_d.data[0])
    print (scal2_init, scal2_d.data[0])
    return np.allclose(scal1_init, scal1_d.data[0]) and \
        np.allclose(scal2_init, scal2_d.data[0])


def test_nullVelocity_2D():
    """
    """
    scal, velo = setup_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17],
                                   scal: [17, 17]},
                      )
    assert assertion(scal, advec)


def test_nullVelocity_vector_2D():
    """
    """
    scal, velo = setup_vector_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17],
                                   scal: [17, 17]},
                      )
    assert assertion_vector2D(scal, advec)


def test_nullVelocity_list_2D():
    """
    """
    scal, velo = setup_list_2D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17],
                                   scal[0]: [17, 17],
                                   scal[1]: [17, 17]},
                      )
    assert assertion_list(scal, advec)


def test_nullVelocity_3D():
    """
    """
    scal, velo = setup_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      )
    assert assertion(scal, advec)


def test_nullVelocity_vector_3D():
    """
    """
    scal, velo = setup_vector_3D()

    advec = Advection(velo, scal,
                      resolutions={velo: [17, 17, 17],
                                   scal: [17, 17, 17]},
                      )
    assert assertion_vector3D(scal, advec)


# def test_nullVelocity_list_3D():
#     """
#     """
#     scal, velo = setup_list_3D()

#     advec = Advection(velo, scal,
#                       resolutions={velo: [17, 17, 17],
#                                    scal[0]: [17, 17, 17],
#                                    scal[1]: [17, 17, 17]},
#                       )
#     assert assertion_list(scal, advec)

if __name__ == '__main__':
    test_nullVelocity_2D()
    #test_nullVelocity_vector_2D()
    test_nullVelocity_list_2D()
    test_nullVelocity_3D()
    test_nullVelocity_vector_3D()
