# -*- coding: utf-8 -*-

import parmepy as pp
from parmepy.operator.velocity_correction import VelocityCorrection
from parmepy.problem.simulation import Simulation
import numpy as np
from parmepy.mpi.topology import Cartesian
import math
pi = math.pi
sin = np.sin
cos = np.cos
import parmepy.tools.numpywrappers as npw

## Physical Domain description
print (" ========= Start Navier-Stokes 3D (Flow past bluff bodies) =========")

## pi constant
pi = np.pi
cos = np.cos
sin = np.sin
# Upstream flow velocity
uinf = 1.0
tol = 1e-12


## Function to compute TG velocity
def computeVel(res, x, y, z, t):
    res[0][...] = sin(x) * cos(y) * cos(z)
    res[1][...] = - cos(x) * sin(y) * cos(z)
    res[2][...] = 0.
    return res


## Function to compute reference vorticity
def computeVort(res, x, y, z, t):
    res[0][...] = - cos(x) * sin(y) * sin(z)
    res[1][...] = - sin(x) * cos(y) * sin(z)
    res[2][...] = 2. * sin(x) * sin(y) * cos(z)
    return res


## Function to compute TG velocity
def computeVel2D(res, x, y, t):
    res[0][...] = sin(x) * cos(y)
    res[1][...] = - cos(x) * sin(y)
    return res


## Function to compute reference vorticity
def computeVort2D(res, x, y, t):
    res[0][...] = - cos(x) * sin(y)
    return res


def test_velocity_correction_3D():
    dim = 3
    ## Domain
    boxlength = npw.realarray([2.0 * pi, pi, pi])
    boxorigin = npw.realarray([0., 0., 0.])
    box = pp.Box(dim, length=boxlength, origin=boxorigin)

    ## Global resolution
    nbElem = [33, 33, 33]

    ## Vector Fields
    velo = pp.Field(domain=box, formula=computeVel,
                    name='Velocity', isVector=True)
    vorti = pp.Field(domain=box, formula=computeVort,
                     name='Vorticity', isVector=True)

    ## Usual Cartesian topology definition
    NBGHOSTS = 2
    ghosts = np.ones((box.dimension)) * NBGHOSTS
    topo = Cartesian(box, box.dimension, nbElem,
                     ghosts=ghosts)

    op = {}
    ref_rate = uinf * box.length[1] * box.length[2]
    op['correction'] = VelocityCorrection(velo, vorti,
                                          resolutions={velo: nbElem,
                                                       vorti: nbElem},
                                          req_flowrate=ref_rate, topo=topo)
    op['correction'].discretize()
    op['correction'].setUp()
    # === Simulation setup ===
    simu = Simulation(tinit=0.0, tend=5., timeStep=0.005, iterMax=1000000)
    # init fields
    velo.initialize(topo=topo)
    vorti.initialize(topo=topo)

    # Apply correction
    op['correction'].apply(simu)
    # check new flowrate values
    sref = op['correction'].discreteOperator.surfRef
    flowrate = velo.integrateOnSurface(sref, topo)
    assert (flowrate - ref_rate) < tol
    for i in xrange(1, dim):
        flowrate2 = velo.integrateOnSurface(sref, topo, component=i)
        assert flowrate2 < tol


def test_velocity_correction_2D():
    dim = 2
    ## Domain
    boxlength = npw.realarray([2.0 * pi, pi])
    boxorigin = npw.realarray([0., 0.])
    box = pp.Box(dim, length=boxlength, origin=boxorigin)

    ## Global resolution
    nbElem = [33, 33]

    ## Vector Fields
    velo = pp.Field(domain=box, formula=computeVel2D,
                    name='Velocity', isVector=True)
    vorti = pp.Field(domain=box, formula=computeVort2D,
                     name='Vorticity', isVector=False)

    ## Usual Cartesian topology definition
    NBGHOSTS = 2
    ghosts = np.ones((box.dimension)) * NBGHOSTS
    topo = Cartesian(box, box.dimension, nbElem,
                     ghosts=ghosts)

    op = {}
    ref_rate = uinf * box.length[1]
    op['correction'] = VelocityCorrection(velo, vorti,
                                          resolutions={velo: nbElem,
                                                       vorti: nbElem},
                                          req_flowrate=ref_rate, topo=topo)
    op['correction'].discretize()
    op['correction'].setUp()
    # === Simulation setup ===
    simu = Simulation(tinit=0.0, tend=5., timeStep=0.005, iterMax=1000000)
    # init fields
    velo.initialize(topo=topo)
    vorti.initialize(topo=topo)

    # Apply correction
    op['correction'].apply(simu)
    # check new flowrate values
    sref = op['correction'].discreteOperator.surfRef
    flowrate = velo.integrateOnSurface(sref, topo)
    assert (flowrate - ref_rate) < tol
    for i in xrange(1, dim):
        flowrate2 = velo.integrateOnSurface(sref, topo, component=i)
        assert flowrate2 < tol

# This may be useful to run mpi tests
if __name__ == "__main__":
    test_velocity_correction_2D()
    test_velocity_correction_3D()
