"""
Module for testing parmepy.operator.transport_d
"""

import unittest
import parmepy as pp
import numpy as np
import pyopencl as cl


class Transport_dTestCase(unittest.TestCase):

    def setUp_OpenCL_basic(self):
        self.b = pp.Box()
        self.nbElem = [32, 32, 32]
        self.topo = pp.CartesianTopology(domain=self.b,
                                         resolution=self.nbElem,
                                         dim=len(self.nbElem))
        build_options = "-cl-single-precision-constant -cl-opt-disable"
        build_options += " -D WIDTH=32"
        build_options += " -D WGN=32"
        build_options += " -D PADDING=0"
        build_options += " -D BASIC=1"
        self.prg_basic = cl.Program(self.ctx, self.gpu_src).build(build_options)
        self.p_positions = pp.fields.continuous.ContinuousField(self.b,
                                                                name="test_field_positions")
        self.p_scalar = pp.fields.continuous.ContinuousField(self.b,
                                                             name="test_field_scalar")
        self.true_p_positions = pp.fields.analytical.AnalyticalField(self.b,
                                                                     formula=lambda x,y,z:x,
                                                                     name="test_field_true_positions")
        self.g_velocity = pp.fields.analytical.AnalyticalField(self.b,
                                                              formula=lambda x, y, z: (0., 1., 0.),
                                                              name="test_field_g_velocity", vector=True)
        self.g_scalar = pp.fields.analytical.AnalyticalField(self.b,
                                                              formula=lambda x, y, z: 1.,
                                                              name="test_field_g_scalar")
        self.b.discretize(self.topo.resolution)
        self.pos, self.pos_id = self.p_positions.discretize(self.topo)
        self.scal, self.velo_id = self.p_scalar.discretize(self.topo)
        self.gvelo, self.pos_id = self.g_velocity.discretize(self.topo)
        self.gscal, self.velo_id = self.g_scalar.discretize(self.topo)
        self.true_pos, self.velo_id = self.true_p_positions.discretize(self.topo)
        self.true_p_positions.initialize()
        self.g_velocity.initialize()
        self.g_scalar.initialize()
        self.pos.data = np.asarray(self.pos.data, dtype=pp.constants.PARMES_REAL_GPU, order='F')
        self.scal.data = np.asarray(self.scal.data, dtype=pp.constants.PARMES_REAL_GPU, order='F')
        self.gvelo.data[0] = np.asarray(self.gvelo.data[0], dtype=pp.constants.PARMES_REAL_GPU, order='F')
        self.gvelo.data[1] = np.asarray(self.gvelo.data[1], dtype=pp.constants.PARMES_REAL_GPU, order='F')
        self.gvelo.data[2] = np.asarray(self.gvelo.data[2], dtype=pp.constants.PARMES_REAL_GPU, order='F')
        self.gscal.data = np.asarray(self.gscal.data, dtype=pp.constants.PARMES_REAL_GPU, order='F')
        self.pos.gpu_data = cl.Buffer(self.ctx, cl.mem_flags.READ_WRITE,
                                      size=self.pos.data.nbytes)
        self.scal.gpu_data = cl.Buffer(self.ctx, cl.mem_flags.READ_WRITE,
                                       size=self.scal.data.nbytes)
        self.gvelo.gpu_data = [cl.Buffer(self.ctx, cl.mem_flags.READ_WRITE,
                                      size=gvelo.nbytes) for gvelo in self.gvelo.data]
        self.gscal.gpu_data = cl.Buffer(self.ctx, cl.mem_flags.READ_WRITE,
                                       size=self.gscal.data.nbytes)
        cl.enqueue_copy(self.queue, self.gscal.gpu_data, self.gscal.data)
        cl.enqueue_copy(self.queue, self.gvelo.gpu_data[0], self.gvelo.data[0])
        cl.enqueue_copy(self.queue, self.gvelo.gpu_data[1], self.gvelo.data[1])
        cl.enqueue_copy(self.queue, self.gvelo.gpu_data[2], self.gvelo.data[2])
        self.transport = pp.Transport(self.g_velocity, self.g_scalar)
        self.transport.discretize(result_position=self.p_positions, result_scalar=self.p_scalar)
        self.transport.setMethod(pp.particular_solvers.gpu.KernelLauncher(self.prg_basic.advection,
                                                                          self.queue,
                                                                          (32, 32, 32), None))
        self.transport.discreteOperator.init_copy = pp.particular_solvers.gpu.KernelLauncher(self.prg_basic.advec_init_copy,
                                                                   self.queue,
                                                                   (32, 32, 32), None)
        self.transport.discreteOperator.init_transpose = pp.particular_solvers.gpu.KernelListLauncher([self.prg_basic.advec_init_transpose_3D_01,
                                                                             self.prg_basic.advec_init_transpose_3D_02],
                                                                                self.queue,
                                                                                [(2 * int(self.nbElem[0]),
                                                                                  int(self.nbElem[1]) / 32,
                                                                                  int(self.nbElem[2])),
                                                                                 (2 * int(self.nbElem[0]),
                                                                                  int(self.nbElem[1]),
                                                                                  int(self.nbElem[2]) / 32)
                                                                                 ],
                                                                                [None,
                                                                                 None])

    def setUp(self):
        try:
            self.platform = cl.get_platforms()[0]
            self.device = self.platform.get_devices(cl.device_type.GPU)[0]
            self.ctx = cl.Context([self.device])
            self.queue = cl.CommandQueue(self.ctx, properties=cl.command_queue_properties.PROFILING_ENABLE)
            f = open(pp.constants.GPU_SRC, 'r')
            self.gpu_src = "".join(f.readlines())
            f.close()
            self.test = True
            print "Testing on " + self.device.name
        except Exception:
            self.test = False
            print "No tests because no GPU"

    def tearDown(self):
        pass

    def test_apply_basic(self):
        """Test advection basic"""
        if self.test:
            self.setUp_OpenCL_basic()
            self.transport.discreteOperator.apply(0., 0.1, 0)
            print self.scal.domain.step
            self.queue.finish()
            cl.enqueue_copy(self.queue, self.pos.data, self.pos.gpu_data)
            cl.enqueue_copy(self.queue, self.scal.data, self.scal.gpu_data)
            self.queue.finish()
            print np.max(self.pos.data),np.min(self.pos.data)
            print self.pos.data[1,1,1], self.true_pos.data[1,1,1]
            np.testing.assert_array_almost_equal(self.pos.data, self.true_pos.data)
            np.testing.assert_array_almost_equal(self.scal.data, self.gscal.data)

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(
        unittest.TestLoader().loadTestsFromTestCase(Transport_dTestCase)
        )
