# -*- coding: utf-8 -*-
import unittest
from parmepy.fields.discrete import *
from parmepy.fields.continuous import *
from parmepy.fields.analytical import *
from parmepy.domain.topology import *
from parmepy.domain.box import *
from parmepy.constants import *
from parmepy.particular_solvers.integrator.integrator import ODESolver
from parmepy.particular_solvers.integrator.euler import Euler
import numpy as np
import numpy.testing as npt
import math


class test_Euler(unittest.TestCase):
    """
    Euler test
    solve u'(t) = f( u(t),t )
    u(t) = a t + U0
    f(u,t) = a + (u - (at + U0))^4
    """
    def setUp(self):
        self.e = 0.0002  # Accepted error between result and analytical result
        self.dim = 3
        self.boxLength = [1., 1., 1.]
        self.boxMin = [ 0., 0., 0.]
        self.nbPts = [6, 6, 6]
        self.box = Box(dimension=self.dim,
                       length=self.boxLength,
                       origin=self.boxMin)
        self.dt = 0.1
        self.FinalTime = 3
        self.U0 = [1.0 , 2.0, 3.0 ]

    def testEulerInt(self):
        # Continuous fields and operator declaration
        self.velo = AnalyticalField(domain=self.box, formula=self.vitesse, name='Velocity', vector=True)
        # Topology definition / Fields and operator discretization
        self.result = [np.ones((self.nbPts), dtype=PARMES_REAL, order=ORDER) for d in xrange(self.dim)]
        self.topo3D = CartesianTopology(domain=self.box, resolution=self.nbPts, dim=self.dim)
        self.velo.discretize(self.topo3D)
        self.velo.initialize()
        self.resol = self.topo3D.mesh.resolution
        #test for t=0 and initialization of the fonction f
        t = 0.
        fctInter = [np.ones((self.nbPts), dtype=PARMES_REAL, order=ORDER) for d in xrange(self.dim)]
        method = Euler()

        #Solution calculated with Euler
        fctInter = self.fctTest(t,self.velo.discreteField[self.velo._fieldId].data[0]\
                                            ,self.velo.discreteField[self.velo._fieldId].data[1]\
                                            ,self.velo.discreteField[self.velo._fieldId].data[2])
        for d in xrange(self.dim) :
            self.result[d][...] = method.integrate(self.velo.discreteField[self.velo._fieldId], fctInter, t, self.dt, d )

        t=t+self.dt

        #Analytical solution
        self.anal=self.analyticalSolution(t,self.velo.discreteField[self.velo._fieldId].data[0]\
                                            ,self.velo.discreteField[self.velo._fieldId].data[1]\
                                            ,self.velo.discreteField[self.velo._fieldId].data[2])
        npt.assert_array_less(abs(self.anal[0]-self.result[0]), self.e)
        npt.assert_array_less(abs(self.anal[1]-self.result[1]), self.e)
        npt.assert_array_less(abs(self.anal[2]-self.result[2]), self.e)
        
        # time loop
        while t < self.FinalTime :
            #print "T=" , t
            fctInter = self.fctTest(t,self.result[0],self.result[1],self.result[2])
            for d in xrange(self.dim) :
                self.result[d][...] = method.integrate(self.result, fctInter, t, self.dt, d )

            t = t + self.dt

            self.anal=self.analyticalSolution(t,self.velo.discreteField[self.velo._fieldId].data[0]\
                                            ,self.velo.discreteField[self.velo._fieldId].data[1]\
                                            ,self.velo.discreteField[self.velo._fieldId].data[2])
        # Comparison with analytical solution 
            npt.assert_array_less(abs(self.anal[0]-self.result[0]), self.e)
            npt.assert_array_less(abs(self.anal[1]-self.result[1]), self.e)
            npt.assert_array_less(abs(self.anal[2]-self.result[2]), self.e)


    def vitesse(self, x, y, z):
        vx = 1.0
        vy = 2.0
        vz = 3.0
        return vx, vy, vz

    def fctTest(self, t, x, y, z):
        wx = 0.2 +( x - self.analyticalSolution(t, 1., 2., 3.)[0])**4
        wy = 0.5  +( y - self.analyticalSolution(t, 1., 2., 3.)[1])**4
        wz = 1. +( z - self.analyticalSolution(t, 1., 2., 3.)[2])**4
        return [wx,wy,wz]

    def analyticalSolution(self, t, x, y, z):
        sx = 0.2 *t + x
        sy = 0.5 *t + y
        sz = 1. *t + z
        return [sx, sy, sz]

    def runTest(self):
        self.setUp()
        self.testOperatorDiv()


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(test_Euler))
    return suite

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(suite())
