"""
@file io_utils.py
Tools related to i/o in parmes.
"""
import os
import scitools.filetable as ft
import parmepy.tools.numpywrappers as npw
from inspect import getouterframes, currentframe
import parmepy.mpi as mpi
from parmepy.tools.sys_utils import SysUtils as su
from re import findall, IGNORECASE


class io(object):
    """
    Static class with utilities for file i/o
    """
    @staticmethod
    def defaultPath():
        """
        Get the default path for io.
        Used name if set. If not, get basename of the
        top exec. file.
        Example : if you run python MyExe.py for your parmes simulation,
        results will be saved in ./MyExe/p1/, ./MyExe/p4/
        the number after 'p' being the number of mpi processus
        set for the simulation.
        """
        # Memo for getouterframe usage:
        #    frame, filename, line_number, function_name, lines, index =\
        # inspect.getouterframes(inspect.currentframe())[-1]
        # Warning FP : the behavior of python and ipython is different for
        # this command.
        a = getouterframes(currentframe())
        ind = -1
        # --- ipython ---
        if su.in_ipython():
            sublist = [i[1] for i in a]
            for val in sublist:
                ll = findall('ipython', val, IGNORECASE)
                if len(ll) > 0:
                    ind = sublist.index(val) - 1
                    break
            if ind > -1:
                # -- interactive ipython but call with execfile--
                if len(findall('io_utils', a[ind][1])) > 0:
                    return './interactive/p' + str(mpi.main_size)
                a = a[ind]
            else:
                # -- interactive ipython without execfile call --
                return './interactive/p' + str(mpi.main_size)
        else:
            # -- python --
            a = a[-1]
            if a[-1] is None or len(findall('py.test', a[1])) > 0:
                # interactive python
                return './interactive/p' + str(mpi.main_size)

        return os.path.join(a[1].split('.')[0], 'p' + str(mpi.main_size))

    @staticmethod
    def checkDir(filename, io_rank=0, comm=None):
        """
        Check if the directory of 'filename' exists.
        Creates it if not.
        @param filename : file name with full or relative path
        @param io_rank : processus rank that does the check.
        @param comm : the mpi communicator that does the check
        """
        # Create output dir if required
        if comm is None:
            comm = mpi.main_comm
        if comm.Get_rank() == io_rank:
            d = os.path.dirname(filename)
            if not os.path.exists(d):
                os.makedirs(d)


class Writer(object):
    """
    Usage :
    \code
    >>> import parmepy.tools.io_utils as io
    >>> defaultPath = io.io.defaultPath()
    >>> params = {"filename": defaultPath + 'r.dat', "writebuffshape": (1, 2)}
    >>> wr = Writer(params)
    >>> ite = 3 # current iteration number
    >>> if wr.doWrite(ite):
    ...    wr.buffer[...] = 3.
    ...    wr.write()
    >>> wr.finalize()

    \endcode
    result : buffer is written into r.dat.
    """
    def __init__(self, params, comm=None):
        """
        @param params : dictionnary to set parameters
        for output file (name, path, ...)
        available keys are :
        - filename : file name with full or relative path.
        Default = data.out, in filepath.
        If filename contains absolute path, filepath is ignored.
        - filepath : path to the i/o file.
        Default =  parmepy.tools.io_utils.io.defaultPath.
        - frequency : how often output file is written. Default = 1
        (i.e. every iteration). N means writes every N iteration and
        0 means write only after last iteration.
        - io_rank : mpi process rank (in comm)
         that performs writting. Default = 0
        - writebuffshape : shape (numpy array) of the output buffer,
         written in filename. Default = 1.
        - safeOutput : boolean. Default = True.
        True --> open/close file everytime data are written.
        False --> open at init and close during finalize. Cost less but if simu
        crashes, data are lost.
        @param comm : mpi communicator that handles this
        writer. Default = parmepy.mpi.main_comm.
        """
        # Directory for output
        if "filepath" in params:
            filepath = params["filepath"]
        else:
            filepath = io.defaultPath()
        # file name
        if "filename" in params:
            filename = params["filename"]
        else:
            filename = 'data.out'
        # # Absolute path + name for i/o file
        # # Note that if filename contains absolute path
        # # filepath is ignored
        self.filename = os.path.join(filepath, filename)

        ## How often i/o must occur
        if "frequency" in params:
            self.frequency = params["frequency"]
        else:
            self.frequency = 1
        ## Rank of the mpi process that performs i/o (if any)
        if "io_rank" in params:
            self.io_rank = params["io_rank"]
        else:
            self.io_rank = 0

        if comm is None:
            comm = mpi.main_comm
        ## A reference communicator, just to identify a
        ## process rank for io.
        self.comm = comm

        # check if output dir exists, create it if not.
        io.checkDir(self.filename, self.io_rank, self.comm)

        ## Shape of the output buffer (must be a 2D numpy array)
        if "writebuffshape" in params:
            self.buffshape = params["writebuffshape"]
        else:
            self.buffshape = (1, 1)
        assert len(self.buffshape) == 2,\
            '2D shape required : set writebuffshape: (x,y)'
        ## The buffer (numpy array) that will be printed to a file
        self.buffer = npw.zeros(self.buffshape)

        ## Defines how often
        ## output file is written :
        ## True --> open/close file everytime
        ## data are written.
        ## False --> open at init and close
        ## during finalize. Cost less but if simu
        ## crashes, data are lost.
        if "safeOutput" in params:
            self.safeOutput = params["safeOutput"]
        else:
            self.safeOutput = True

        if self.safeOutput:
            self.write = self._fullwrite
        else:
            self.write = self._partialwrite
        # Force synchro to be sure that all output dirs
        # have been created.
        self.comm.barrier()
        if self.comm.Get_rank() == self.io_rank:
            self._file = open(self.filename, 'w')

    def doWrite(self, ite):
        """
        True if output is required
        for iteration ite
        @param ite : current iteration number
        """
        rank = self.comm.Get_rank()
        num = ite + 1  # Number of iterations done
        return rank == self.io_rank and (num % self.frequency) == 0

    def _fullwrite(self):
        if self.comm.Get_rank() == self.io_rank:
            self._file = open(self.filename, 'a')
            ft.write(self._file, self.buffer)
            self._file.close()

    def _partialwrite(self):
        if self.comm.Get_rank() == self.io_rank:
            ft.write(self._file, self.buffer)

    def finalize(self):
        if self.comm.Get_rank() == self.io_rank:
            if not self._file.closed:
                self._file.close()
