"""
@file timers.py

Contains class for monitoring computational time in a non-intrusive way
thanks to decorator.
"""
from parmepy.mpi import MPI, main_rank
from parmepy import __VERBOSE__
ftime = MPI.Wtime


def timed_function(f):
    """
    Decorator for function timing. Get the corresponding function timer in the
    operator (represented in args[0]).

    @remark : Works only with methods that belongs to object with Timer
    member named timer.
    """
    def wrapper(*args, **kwargs):
        f_timer = args[0].timer.getFunctionTimer(f)
        res = f_timer(f)(*args, **kwargs)
        return res
    # def wrapper(*args, **kwargs):
    #     return f(*args, **kwargs)
    return wrapper


class FunctionTimer(object):
    """
    Class for timing functions with the timed_function decorator.

    The FunctionTimer is instancied with a given function. Calling the timer
    will call the function timed with execution time computation.
    """

    def __init__(self, func):
        """
        Creates a timer for the given function.

        @param func : function to time.
        """
        ## Timed function name
        self.func_name = func.func_name
        ## Total time spent in the function (in seconds).
        self.t = 0.
        # ## Times per calls
        # self.times = []
        ## Calls number
        self.ncalls = 0

    def __call__(self, func):
        """
        The function given is wrapped with time computations instructions
        through MPI.Wtime function.
        It returns the wrapped function.
        @param func function to time.
        @return wrapped function.
        """
        def f(*args, **kwargs):
            t0 = ftime()
            res = func(*args, **kwargs)
            t = ftime() - t0
            self.t += t
            self.ncalls += 1
            #self.times.append(t)
            if __VERBOSE__:
                print (args[0].__class__.__name__, ' -- ',)
                print (self.func_name, " : ", t, 's')
            return res
        return f

    def __str__(self):
        s = self.func_name
        s += ' : ' + str(self.t) + " s  ({0} calls)".format(self.ncalls)
        return s


class ManualFunctionTimer(FunctionTimer):
    """
    Class for manual timing some code. For instance, this class is desined for
    monitoring OpenCL kernels where function call does not reflect
    computational time.
    """
    def __init__(self, name):
        def fun():
            pass
        fun.func_name = name
        FunctionTimer.__init__(self, fun)

    def append_time(self, t):
        """Manual computational time append"""
        self.t += t
        self.ncalls += 1
        #self.times.append(t)
        if __VERBOSE__:
            print (self.func_name, " : ", t, 's')


class Timer(object):
    """
    Manages a dictionary of FunctionTimer objects for monitoring functions
    member of the given operator.
    """

    def __init__(self, op, suffix=''):
        """
        Creates an Timer.
        @param op : Timer owner.
        """
        ## Parent object name
        self._obj = op
        self.name = self._obj.name + suffix
        ## FunctionTimer dictionary with functions or operators as keys.
        self.f_timers = {}
        ## Total time spent in the operator (in seconds).
        self.t = 0.
        self._isEmpty = True

    def __add__(self, t):
        """
        Overrides operator +.
        @param t : Other Timer object
        """
        t.compute_summary()
        #self.t += t.t
        self._isEmpty = False
        self.f_timers[t] = t
        return self

    def getFunctionTimer(self, func):
        """
        Get the FunctionTimer related to given function, or created
        if not exists.

        @param func : Function to get the corresponding timer.
        @return : FunctionTimer related to func.
        """
        try:
            return self.f_timers[func]
        except KeyError:
            self.f_timers[func] = FunctionTimer(func)
            return self.f_timers[func]

    def addFunctionTimer(self, ft):
        """
        Add a function timer to Timer.

        @param ft : FunctionTimer to add.
        """
        self.f_timers[ft] = ft

    def addSubTimer(self, t, prefix=''):
        if len(t.f_timers.keys()) > 0:
            t.compute_summary()
            t.name += prefix
            self.f_timers[t] = t

    def compute_summary(self):
        """
        Compute a summary of the different functions referenced herein.
        """
        self.t = 0.
        for f in self.f_timers.keys():
            # if self.f_timers[f] is a Timer, these times are already
            # in the current Timer sum.
            if not isinstance(self.f_timers[f], Timer):
                self.t += self.f_timers[f].t
                self._isEmpty = False

    def rTimes(self):
        s = ""
        for f in sorted(self.f_timers.keys()):
            nl = "@@F@@[" + str(main_rank) + "]@@S@@"
            if isinstance(self.f_timers[f], Timer):
                if len(self.f_timers[f].f_timers) > 0:
                    s += nl + '     |-- ' + f.name
                    if self.f_timers[f].t > 0.:
                        s += " : " + str(self.f_timers[f].t)
                    subTimer = self.f_timers[f]
                    s += subTimer.rTimes().replace('@@S@@', '@@S@@    ')
            else:
                s += nl + '     |-- ' + str(self.f_timers[f])
        return s

    def __str__(self):
        self.compute_summary()
        if not self._isEmpty:
            s = "[" + str(main_rank) + "] Time spent in "
            s += self.name + ' = ' + str(self.t) + ' s'
            s += self.rTimes().replace('@@F@@', '\n').replace('@@S@@', '')
        else:
            s = ""
        return s
