!------------------------------------------------------------------------------
!
! PROGRAM : advec_sheartag
!
! DESCRIPTION: 
!> This program provide a numerical illustration of an adveciton problem: a
!! velocity field with radial shear. This test illustrate the effect of
!corrected remeshing formula.
!!
!! @details
!! This example could be used as a benchmark (to determine optimisation
!! efficiency or parallel scalability). The velocity field is choosen to ensure
!! That the solver will imply tag and thus corrected remeshing formula. It is a
!! Typical test cases to evaluate efficiency of such a formula in context of large
!! Cfl number (eg 3 or more).
!! Note that the analyti solution is known and could be check. 
!!
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

program advec_sheartag

    ! External Library
    use mpi
    ! Scales code
    use cart_topology
    use advec
    use vtkxml
    ! Test procedures
    use test_common

    implicit none

    logical, parameter      :: output = .false.     ! Do you want output for
                                                    ! vizualisation purpose ?
    real(WP), parameter     :: M_PI = ACOS(-1.0)    ! Pi value
    real(WP),parameter      :: period = 1

    logical     :: success = .true. ! logical error
    integer     :: ierr             ! mpi error code
    integer     :: rank_world       ! processus rank on "MPI_COMM_WORLD"
    integer     :: nb_proc,nb_procZ ! number of processus

    character(str_short)                        :: order='p_02'     ! space order of the solveur
    real(WP), dimension(:, :, :), allocatable   :: Vx, Vy, VZ       ! the flow
    real(WP), dimension(:, :, :), allocatable   :: scal3D           ! the scalar field
    integer                                     :: i,j,k            ! mesh indice   
    real(WP)                                    :: T_step           ! time where output are done
    real(WP)                                    :: T_end            ! final time
    real(WP)                                    :: T                ! current time
    real(WP)                                    :: dt               ! time step
    real(WP), dimension(:, :, :), allocatable   :: good_scal        ! analytic solution
    integer                                     :: tag_num,tag_err  ! identifiant for io
    integer                                     :: tag_sol          ! identifiant for io
    real(WP)                                    :: rx, ry, rr
    integer, parameter                          :: N_mesh=400       ! number of mesh
   
    ! ===== Initialisation of parallel context =====

    ! Set the verbosity
    verbose_test = .true.
    verbose_more = .true.
    ! Initialise mpi
    call mpi_init(ierr)    
    call mpi_comm_rank(MPI_COMM_WORLD, rank_world, ierr)
    call mpi_comm_size(MPI_COMM_WORLD, nb_proc, ierr)

    ! ===== Cut the domain along y and initialize the topology =====
    nb_procZ = 1
    if (mod(N_mesh, nb_proc)/=0) stop 'wrong number of processes : it have to divide 100'
    call cart_create((/ nb_proc, nb_procZ /), ierr)

    ! ===== Create mesh =====
    call discretisation_create(N_mesh,N_mesh,20,dble(2),dble(2),dble(0.5))
    call set_group_size(5)
    call advec_init(order)
    call mpi_barrier(MPI_COMM_WORLD, ierr)

    ! ===== Field allocation =====
    allocate(scal3D(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_scal(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vx(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vy(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vz(N_proc(1), N_proc(2), N_proc(3)))

    ! ===== Initialization ====
    Vx = 0.0
    Vy = 0.0
    Vz = 0.0
    do k = 1, N_proc(3)
        do j = 1, N_proc(2)
            ry = ((j-1+coord(2)*N_proc(2))*d_sc(2)) - length(2)/2.0
            do i = 1, N_proc(1)
                rx = ((i-1+coord(1)*N_proc(1))*d_sc(1))-length(1)/2.0
                rr = (rx**2+ry**2)
                if (rr<1) then
                    scal3D(i,j,k) = (1-rr)**6
                else
                    scal3D(i,j,k) = 0
                end if
                Vx(i,j,k) = cos(3*M_PI*sqrt(rr)/2)*(-ry)
                Vy(i,j,k) = cos(3*M_PI*sqrt(rr)/2)*(rx)
            end do
        end do
    end do
    good_scal = scal3D
    T = 0.0
    dt = 3*min(d_sc(1),d_sc(2))
    T_step = 0.0
    T_end = 0.8
    call test_substatus('cfl=3 and dt', dt, cart_rank)

    ! === Create output context and solve initial state =====
    if (output) then
        call vtkxml_init_all(9, nb_proc_dim, length, cart_rank, coord,'./adv_res/')
        call vtkxml_init_field(trim(order)//'_tag_num', tag_num)
        call vtkxml_init_field(trim(order)//'_tag_err', tag_err)
        call vtkxml_init_field(trim(order)//'_tag_sol', tag_sol)
        call vtkxml_write(tag_sol, good_scal)
    end if


    ! ===== Solve equation ====
    do while(T< (T_end - dt))
        call advec_step(dt, Vx, Vy, Vz, scal3D)
        T = T + dt
        if (T>T_step) then
            T_step = T_step + period/10
            call test_substatus ('tag, t', T, cart_rank)
            if (output) call vtkxml_write(tag_num, scal3D)
        end if
    end do
    if (T<T_end) then
        dt = T_end - T
        call advec_step(dt, Vx, Vy, Vz, scal3D)
    end if
    if (output) then
        call vtkxml_write(tag_num, scal3D)
        call vtkxml_write(tag_err, scal3D-good_scal)
        call vtkxml_finish()
    end if
    call test_substatus ('time', (T+dt), cart_rank)
    call test_check_success(scal3D, good_scal, success, cart_rank)
    call test_substatus('3D test - with tag', success, cart_rank)

    
    ! --- Free memory --- 
    deallocate(scal3D)
    deallocate(good_scal)
    deallocate(Vx)
    deallocate(Vy)
    deallocate(Vz)

    success = .not.success

    call mpi_finalize(ierr)

end program advec_sheartag
