!------------------------------------------------------------------------------
!
! PROGRAM : advec_turnsphere
!
! DESCRIPTION: 
!> This program provide a numerical illustration of an adveciton problem: a
!! turning sphere.
!!
!! @details
!! This example could be used as a benchmark (to determine optimisation
!! efficiency or parallel scalability) 
!! Note that the analyti solution is known and could be check. 
!!
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

program advec_turnsphere

    ! External Library
    use mpi
    ! Scales code
    use advec
    use cart_topology
    use vtkxml
    ! Test procedures
    use test_common

    implicit none
    
    real(WP), parameter     :: M_PI = ACOS(-1.0)    ! Pi value
    logical, parameter      :: output = .true.     ! Do you want output for
                                                    ! vizualisation purpose ?
    real(WP),parameter      :: period = 1
    integer, parameter      :: mesh_size = 160

    logical     :: success = .true. ! logical error
    integer     :: ierr             ! mpi error code
    integer     :: rank_world       ! processus rank on "MPI_COMM_WORLD"
    integer     :: nb_proc,nb_procZ ! number of processus

    real(WP), dimension(:, :, :), allocatable   :: scal3D           ! the scalar field
    real(WP), dimension(:, :, :), allocatable   :: Vx, Vy, Vz       ! the flow
    integer                                     :: i,j,k            ! mesh indice   
    real(WP)                                    :: rx, ry, rz, rr   ! distance**2 to the mesh center
    real(WP)                                    :: rayon            ! rayon of the initial scalar sphere
    real(WP)                                    :: T                ! time
    real(WP)                                    :: T_end            ! final time
    real(WP)                                    :: T_ite            ! time corresponding to output
    real(WP)                                    :: dt               ! time step
    real(WP), dimension(:, :, :), allocatable   :: good_scal        ! analytic solution
    real(WP), dimension(:, :, :), allocatable   :: good_velo        ! temp field
    integer                                     :: tag_rot          ! tag for visualisation context
    integer                                     :: tag_sol          ! tag for visualisation context
    real(WP)                                    :: times, time1, time2 ! to evaluate computation time.
   
    ! ===== Initialisation of parallel context =====

    ! Set the verbosity
    verbose_test = .true.
    verbose_more = .true.
    ! Initialise mpi
    call mpi_init(ierr)    
    call mpi_comm_rank(MPI_COMM_WORLD, rank_world, ierr)
    call mpi_comm_size(MPI_COMM_WORLD, nb_proc, ierr)

    ! Cut the domain along Y and initialize the toppology
    nb_procZ = 1
    if ((mod(nb_proc,5)==0).and.(mod(mesh_size, nb_proc/5)==0)) then
        nb_procZ = 5
        nb_proc = nb_proc/5
    else if ((mod(nb_proc,2)==0).and.(mod(mesh_size, nb_proc/2)==0)) then
        nb_procZ = 2
        nb_proc = nb_proc/2
    else
        if (mod(mesh_size, nb_proc)/=0) then
            if(cart_rank==0) write(*,'(a,x,i0)') 'wrong number of processes : it have to divide', mesh_size
            stop
        end if
    end if

    ! ===== Create mesh ====
    call cart_create((/ nb_proc, nb_procZ /), ierr)
    call set_group_size(5)
    call discretisation_create(mesh_size,mesh_size,mesh_size,dble(1),dble(1),dble(1))
    call test_substatus('group size', group_size(1,1), cart_rank)
    call mpi_barrier(MPI_COMM_WORLD, ierr)

    ! ===== Initialize the particular solver =====
    call advec_init('p_O2')

    ! ===== Field allocation =====
    allocate(scal3D(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_scal(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vx(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vy(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vz(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_velo(N_proc(1), N_proc(2), N_proc(3)))

    ! ===== Initialization ====
    ! -- Scalar --
    scal3D = 0.
    rayon = (minval(N*d_sc)/10.0)**2
    do k = 1, N_proc(3)
        rz = (d_sc(3)*(k + coord(3)*N_proc(3) - 3.0*N(3)/5.0))**2
        do j = 1, N_proc(2)
            ry = (d_sc(2)*(j + coord(2)*N_proc(2)- 3.0*N(2)/5.0))**2
            do i = 1, N_proc(1)
                rx = (d_sc(1)*(i - 3.0*N(1)/5.0))**2
                rr = rx + ry + rz
                if (rr < rayon) scal3D(i,j,k) = (1 - rr/rayon)**1
            end do
        end do
    end do
    good_scal = scal3D
    ! -- Velocity --
    do k = 1, N_proc(3)
        do j = 1, N_proc(2)
            do i = 1, N_proc(1)
                Vx(i,j,k)=(2*M_PI/period)*(length(2)/2.0-((j+coord(2)*N_proc(2))*d_sc(2)))
                Vy(i,j,k)=(2*M_PI/period)*(((i+coord(1)*N_proc(1))*d_sc(1))-length(1)/2.0)
            end do
        end do
    end do
    Vz = 0.0



    ! ===== Initialize output context ====
    if (output) then
        call vtkxml_init_all(2, nb_proc_dim, length, cart_rank, coord,'./adv_res/')
        call vtkxml_init_field('turning', tag_rot)
        call vtkxml_init_field('turn_sol', tag_sol)
        call vtkxml_write(tag_rot, scal3D, 'turning')
    end if


    ! ===== Compute the numerical solution =====
    dt = 0.02
    T = 0
    T_end = 2*period
    T_ite = 0
    times = 0
    do while (T<=T_end - dt)
        call cpu_time(time1)
        call advec_step(dt, Vx, Vy, Vz, scal3D)
        T = T + dt
        call cpu_time(time2)
        times=times+(time2-time1)
        if (T>T_ite) then
            T_ite = T_ite + T_end/10
            call test_substatus ('time', T, cart_rank)
            if (output) call vtkxml_write(tag_rot, scal3D, 'turning')
        end if
    end do
    if (T<T_end) then
        dt = T_end - T
        call advec_step(dt, Vx, Vy, Vz, scal3D)
    end if
    if (output) then
        call vtkxml_write(tag_rot, scal3D, 'turning')
        call vtkxml_write(tag_sol, good_scal)
        call vtkxml_finish()
    end if

    call test_check_success(scal3D, good_scal, success, cart_rank)
print*, 'times = ', times
    call test_substatus('computational time', times, cart_rank)
    call test_substatus('turning sphere', success, cart_rank)


    ! --- Free memory ---
    deallocate(scal3D)
    deallocate(good_scal)
    deallocate(Vx)
    deallocate(Vy)
    deallocate(Vz)

    success = .not.success


    call mpi_finalize(ierr)

end program advec_turnsphere
