!> @addtogroup part_test
!! @{
    
!------------------------------------------------------------------------------
!
! MODULE: test_advection
!
! DESCRIPTION: 
!> Validation test for advection method.
!!
!! @details
!! This module provide different test to validate the transport solver. 
!! All these test are unit test : they return a logical value to check if 
!! the code version pass it or not. 
!!
!! That is all these test are logical function, they return true if the result
!! is the right one and false otherwise. 
!! All the "test_part_*" function are devoted to validate the particular solver
!! The following test are included :
!!      A - Validate the particular method, step by step
!!      1 -> Test the procedure "AC_obtain_senders" from advec_common
!!      2 -> Validate the redistribution the buffer during the remeshing (XXX todo)
!!      3 -> Validate the remeshing of untagged particles
!!      4 -> Validate the remeshing of tagged particles (XXX todo)
!!      B - Validate an advection solver (XXX todo)
!!      1 -> advec a ball with a constant velocity
!!      2 -> advec a ball with a spheric velocity field (the ball turns)
!!
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

module advec_aux

    use string
    use precision
    implicit none

    real(WP), private   :: epsilon_error = 1e-4     ! Error tolerance


    ! Public procedures

    ! ===== Test for the particles solver =====
    ! Public function
    public              :: test_part_remesh_no_tag
    public              :: test_part_advec_1D
    public              :: test_part_advec_3D
    public              :: test_advecY


    ! ===== Generic test for an advection solver =====

    ! Private procedure

contains

!> Particles method: validation of the remeshing of untagged particles
!!     @param[in]   init_scal   =  optional parameter to initialise the scalar fied to
!!                                  a constant one or to a sphere shape
!!     @param[in]   order_opt   =  optional parameter to choose the remeshing formula
!!     @return      error       = test error (= false if the code pass the test) (= not success)
function test_part_remesh_no_tag (init_scal, order_opt) result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use advecY
    use advec_common    ! Some porcedure common to advection along all directions
    use advec_variables ! contains info about solver parameters and others.
    use cart_topology
    ! Test procedures
    use advec_aux_init
    use test_common

    logical                                     :: success
    character(len=*), intent(in), optional      :: init_scal
    character(len=*), intent(in), optional      :: order_opt        ! space order of the solveur

    character(len=17)                           :: initialisation   ! to choose how to initialise the scalar field
    character(len=str_short)                    :: order            ! space order of the solveur
    real(WP), dimension(:, :, :), allocatable   :: scalar           ! the scalar field
    real(WP), dimension(:,:,:), allocatable     :: p_pos_adim, p_SC ! the adimensionned particles position and the scalar they advect
    logical, dimension(:,:,:), allocatable      :: bl_type, bl_tag  ! type and tag of each particle bloc
    real(WP)                                    :: velocity         ! constant velocity of the flux
    integer                                     :: nb_proc          ! number of processes
    integer                                     :: ierr             ! mpi error code
    integer                                     :: i,j,k            ! mesh indice   
    integer, dimension(2)                       :: ind_group        ! indice of the current group of line
    integer                                     :: T_step           ! time
    integer                                     :: T_end            ! final time
    real(WP), dimension(:,:,:), allocatable     :: good_scal        ! analytic solution
    real(WP), dimension(3)                      :: translat         ! to compute analytic solution

    ! Initialize the particular solver
    order = 'p_O2'
    if (present(order_opt)) order = trim(order_opt)
    call advec_init(order)

    allocate(scalar(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_scal(N_proc(1), N_proc(2), N_proc(3)))
    allocate(p_pos_adim(N_proc(2),1,1))
    allocate(p_SC(N_proc(2),1,1))
    allocate(bl_type(1+N_proc(2)/bl_size,1,1))
    allocate(bl_tag((N_proc(2)/bl_size),1,1))

    if(present(init_scal)) then
        initialisation = init_scal
    else 
        initialisation = 'constant'
    end if
    success = .true.

    !call cart_create((/ 1, nb_proc, 1 /), ierr)
    !call mesh_default()

    ! Choose a velocity
    velocity = N(2)/11.*d_sc(2)
    T_step = 1
    T_end = 1
    T_end = T_step
    translat = 0
    translat(2) = - velocity*T_step
    translat = translat/d_sc

    ! Initialise the scalar field
    call scal_init(initialisation, scalar, good_scal, translat)
    call test_substatus('initialisation', success, cart_rank)

    ! Choose to test remeshing for centered/left and tagged/untagged particles
    bl_type = .true.
    bl_tag  = .false.

    ! Advec it with a particular method
    do k = begin_proc(3), end_proc(3)
    ind_group(2) = 0
    !k = begin_proc(3)
        ind_group(1) = 0
        ind_group(2) = ind_group(2) + 1
        do i = begin_proc(1), end_proc(1)
        !i = begin_proc(1)
            ind_group(1) = ind_group(1) + 1
            ! Initialise the particles
            p_SC(:,1,1) = scalar(i,:,k)
            !scalar(i,:,k)=0
            do j = begin_proc(2), end_proc(2)
                p_pos_adim(j,1,1) = j
            end do
!            do T_step = 1, 11
                ! Advection
                p_pos_adim = p_pos_adim + T_step*velocity/d_sc(2)
                ! Remeshing
                allocate(send_group_min(1,1))
                allocate(send_group_max(1,1))
                if (order == 'p_O4') then
                    call Yremesh_O4(ind_group, (/1,1/), p_pos_adim, bl_type, bl_tag, i, k, scalar)
                else if (order == 'p_M6') then
                    call Yremesh_Mprime6(ind_group, (/1,1/), p_pos_adim, i, k, scalar)
                else
                    call Yremesh_O2_group(ind_group, (/1,1/), p_pos_adim, bl_type, bl_tag, i, k, scalar)
                end if
                deallocate(send_group_min)
                deallocate(send_group_max)
!            end do
        end do
    end do

    ! Check the final scalar field
    call test_check_success(scalar, good_scal, success, cart_rank)

    deallocate(scalar)
    deallocate(p_pos_adim)
    deallocate(p_SC)
    deallocate(bl_type)
    deallocate(bl_tag)

    success = .not.success

end function test_part_remesh_no_tag


!> Particles method: validation of the advection along each direction
!! individually with particle method - no tag case.
!!     @param[in]   init_scal   =  optional parameter to initialise the scalar fied to
!!                                  a constant one or to a sphere shape
!!     @param[in]   shift       = global translation of indices (optional)
!!     @param[in]   order_opt   =  optional parameter to choose the remeshing formula
!!     @return      error       = test error (= false if the code pass the test) (= not success)
!! @details
!!    These tests are devoted to validate the advection solver based on particle
!!    method. They can be used for other advection solvers too. Their specificity
!!    is to test each configuration that could be encoutered in the order 2 solver
!!    based on particle method. Therefore they test all cases without tag (and there-
!!    fore corrected remeshing formula) for order 2 (or order 4) solver.
function test_part_advec_1D(init_scal, shift, order_opt) result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use advecX
    use advecY
    use advecZ
    use cart_topology
    ! Test procedures
    use advec_aux_init
    use test_common

    logical                                     :: success
    character(len=*), intent(in), optional      :: init_scal
    integer, intent(in), optional               :: shift
    character(len=*), intent(in), optional      :: order_opt        ! space order of the solveur

    character(str_short)                        :: initialisation   ! to choose how to initialise the fields
    integer                                     :: shift_bis        ! shift effectly used in the test
    character(str_short)                        :: order            ! space order of the solveur
    real(WP), dimension(:, :, :), allocatable   :: scal3D           ! the scalar field
    real(WP), dimension(:, :, :), allocatable   :: velo             ! the flow
    real(WP), dimension(:, :, :), allocatable   :: Vx, Vy, VZ       ! the flow
    integer                                     :: i,j,k            ! mesh indice   
    real(WP)                                    :: T_step           ! time where output are done
    real(WP)                                    :: T_end            ! final time
    real(WP)                                    :: T                ! current time
    real(WP)                                    :: dt               ! time step
    real(WP), dimension(:, :, :), allocatable   :: good_scal        ! analytic solution
    integer                                     :: direction        ! current direction
    integer                                     :: tag_io,tag_er    ! identifiant for io
    integer                                     :: tag_sol          ! identifiant for io

    ! -- Mesh init --
    call discretisation_create(80,80,80,dble(1),dble(1),dble(1))

    ! -- Allocation --
    allocate(scal3D(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_scal(N_proc(1), N_proc(2), N_proc(3)))
    allocate(velo(N_proc(1), N_proc(2), N_proc(3)))

    ! -- Initialisation --
    if(present(init_scal)) then
        initialisation = init_scal
    else 
        initialisation = 'center'
    end if
    if (present(shift)) then
        shift_bis = shift
    else
        shift_bis = 0
    end if
    success = .true.
    dt = 0.1

    call test_substatus('shift', shift_bis, cart_rank)

    ! Initialize the particular solver
    order = 'p_O2'
    if (present(order_opt)) order = order_opt
    call advec_init(order)

    ! ===== Test along X =====
    ! Initialise the velocity, the scalar field and compute the theoritical solution
    direction = 1
    call test_substatus('direction', direction, cart_rank)
    call scal_velo_init_part(init_scal, dt, shift_bis, scal3D, velo, direction, good_scal)
    call test_substatus('initialisation', success, cart_rank)

    ! Advec it with a particular method
    call advecX_calc(dt, velo, scal3D, order)
    call test_check_success(scal3D, good_scal, success, cart_rank)
    call test_substatus('advection along X', success, cart_rank)


    ! ===== Test along Y =====
    ! Initialise the velocity, the scalar field and compute the theoritical solution
    direction = 2
    call test_substatus('direction', direction, cart_rank)
    call scal_velo_init_part(init_scal, dt, shift_bis, scal3D, velo, direction, good_scal)
    call test_substatus('initialisation', success, cart_rank)

    ! Advec it with a particular method
    call advecY_calc(dt, velo, scal3D, order)
    call test_check_success(scal3D, good_scal, success, cart_rank)
    call test_substatus('advection along Y', success, cart_rank)


    ! ===== Test along Z =====
    ! Initialise the velocity, the scalar field and compute the theoritical solution
    direction = 3
    call test_substatus('direction', direction, cart_rank)
    call scal_velo_init_part(init_scal, dt, shift_bis, scal3D, velo, direction, good_scal)
    call test_substatus('initialisation', success, cart_rank)

    ! Advec it with a particular method
    call advecZ_calc(dt, velo, scal3D, order)
    call test_check_success(scal3D, good_scal, success, cart_rank)
    call test_substatus('advection along Z', success, cart_rank)



    deallocate(scal3D)
    deallocate(good_scal)
    deallocate(velo)

    success = .not.success

end function test_part_advec_1D


!> Particles method: validation of the advection in 3D cases.
!!     @param[in]   order       =  optional parameter to choose between order 2
!!                                  or order 4 particle method.
!!     @param[in]   shift       = global translation of indices (optional)
!!     @return      error       = test error (= false if the code pass the test) (= not success)
!! @details
!!    This tests are devoted to validate the advection solver based on particle
!!    method on 3D cases. They can be used for other advection solvers too. Their specificity
!!    is to test each configuration that could be encoutered in the order 2 and order 4
!!    solver based on particle method. Supposing the 1D tests (test_part_advec_1D) have been
!!    passed successful, then passed this test means the solver does not contain
!!    error anymore.
function test_part_advec_3D(shift, order) result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use cart_topology
    use vtkxml
    ! Test procedures
    use advec_aux_init
    use test_common

    logical                                     :: success
    character(len=*), intent(in), optional      :: order
    integer, intent(in), optional               :: shift

    character(str_short)                        :: initialisation   ! to choose how to initialise the fields
    integer                                     :: shift_bis        ! shift effectly used in the test
    character(str_short)                        :: order_bis        ! space order of the solveur
    real(WP), dimension(:, :, :), allocatable   :: scal3D           ! the scalar field
    real(WP), dimension(:, :, :), allocatable   :: velo             ! the flow
    real(WP), dimension(:, :, :), allocatable   :: Vx, Vy, VZ       ! the flow
    integer                                     :: i,j,k            ! mesh indice   
    real(WP)                                    :: T_step           ! time where output are done
    real(WP)                                    :: T_end            ! final time
    real(WP)                                    :: T                ! current time
    real(WP)                                    :: dt               ! time step
    real(WP), dimension(:, :, :), allocatable   :: good_scal        ! analytic solution
    integer                                     :: direction        ! current direction
    integer                                     :: tag_num,tag_err  ! identifiant for io
    integer                                     :: tag_sol          ! identifiant for io
    real(WP)                                    :: rx, ry, rz, rr


    ! -- Parameters initialisation --
    call discretisation_create(80,80,80,dble(1),dble(1),dble(1))
    call set_group_size(10)
    if(present(order)) then
        order_bis = order
    else 
        order_bis = 'p_O2'
    end if
    ! Initialize the particular solver
    call advec_init(order_bis)
    if (present(shift)) then
        shift_bis = shift
    else
        shift_bis = 0
    end if
    call test_substatus('shift', shift_bis, cart_rank)
    success = .true.
    T_end = 1.0

    ! -- Allocation --
    allocate(scal3D(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_scal(N_proc(1), N_proc(2), N_proc(3)))
    allocate(velo(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vx(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vy(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vz(N_proc(1), N_proc(2), N_proc(3)))

    ! Initialize output context
    call vtkxml_init_all(9, nb_proc_dim, length, cart_rank, coord,'./adv_res/')

    ! -- Fields initialisation --
    ! ===== 3D velocity =====
    ! Initialise the velocity, the scalar field and compute the theoritical solution
    dt = 1
    call scal_velo_init_part('center', dt, shift_bis, scal3D, velo, 1, good_scal)
    dt = 0.1
    good_scal = scal3D
    velo = 1
    call test_substatus('Start test : X,Y,Z,3D,tag', cart_rank)
    

    ! === Velocity along X ===
    T = 0.0
    T_step = 0.0
    Vx = velo
    Vy = 0
    Vz = 0
    call vtkxml_init_field(trim(order_bis)//'_X_num', tag_num)
    call vtkxml_init_field(trim(order_bis)//'_X_err', tag_err)
    call vtkxml_init_field(trim(order_bis)//'_X_sol', tag_sol)
    call vtkxml_write(tag_sol, good_scal)
    do while(T< (T_end - dt))
        call advec_step(dt, Vx, Vy, Vz, scal3D)
        T = T + dt
        if (T>T_step) then
            T_step = T_step + period/10
            call test_substatus ('X, t', T, cart_rank)
            call vtkxml_write(tag_num, scal3D)
        end if
    end do
    if (T<T_end) then
        dt = T_end - T
        call advec_step(dt, Vx, Vy, Vz, scal3D)
    end if
    call vtkxml_write(tag_num, scal3D)
    call vtkxml_write(tag_err, scal3D-good_scal)
    call test_substatus ('X, t', (T+dt), cart_rank)
    call test_check_success(scal3D, good_scal, success, cart_rank)
    call test_substatus('3D test - V along X', success, cart_rank)

    ! === Velocity along Y ===
    T = 0.0
    T_step = 0.0
    Vx = 0
    Vy = velo
    Vz = 0
    call vtkxml_init_field(trim(order_bis)//'_Y_num', tag_num)
    call vtkxml_init_field(trim(order_bis)//'_Y_err', tag_err)
    scal3D = good_scal
    do while(T< (T_end - dt))
        call advec_step(dt, Vx, Vy, Vz, scal3D)
        T = T + dt
        if (T>T_step) then
            T_step = T_step + period/10
            call test_substatus ('Y, t', T, cart_rank)
            call vtkxml_write(tag_num, scal3D)
        end if
    end do
    if (T<T_end) then
        dt = T_end - T
        call advec_step(dt, Vx, Vy, Vz, scal3D)
    end if
    call vtkxml_write(tag_num, scal3D)
    call vtkxml_write(tag_err, scal3D-good_scal)
    call test_check_success(scal3D, good_scal, success, cart_rank)
    call test_substatus('3D test - V along Y', success, cart_rank)

    ! === Velocity along Z ===
    T = 0.0
    T_step = 0.0
    Vx = 0
    Vy = 0
    Vz = velo
    dt = 0.1
    T_end = 1
    scal3D = good_scal
    call vtkxml_init_field(trim(order_bis)//'_Z_num', tag_num)
    call vtkxml_init_field(trim(order_bis)//'_Z_err', tag_err)
    do while(T< (T_end - dt))
        call advec_step(dt, Vx, Vy, Vz, scal3D)
        T = T + dt
        if (T>T_step) then
            T_step = T_step + period/10
            call test_substatus ('Z, t', T, cart_rank)
            call vtkxml_write(tag_num, scal3D)
        end if
    end do
    if (T<T_end) then
        dt = T_end - T
        call advec_step(dt, Vx, Vy, Vz, scal3D)
    end if
    call vtkxml_write(tag_num, scal3D)
    call vtkxml_write(tag_err, scal3D-good_scal)
    call test_check_success(scal3D, good_scal, success, cart_rank)
    call test_substatus('3D test - V along Z', success, cart_rank)

    ! === 3D Velocity  ===
    call test_substatus('3D test - 3D velocity', cart_rank)
    T = 0.0
    T_step = 0.0
    Vx = 2*velo
    Vy = velo
    Vz = 3*velo
    call vtkxml_init_field(trim(order_bis)//'_3Dnum', tag_num)
    call vtkxml_init_field(trim(order_bis)//'_3Derr', tag_err)
    scal3D = good_scal
    do while(T< (T_end - dt))
        call advec_step(dt, Vx, Vy, Vz, scal3D)
        T = T + dt
        if (T>T_step) then
            T_step = T_step + period/10
            call test_substatus ('3D, t', T, cart_rank)
            call vtkxml_write(tag_num, scal3D)
       end if
    end do
   if (T<T_end) then
        dt = T_end - T
        call advec_step(dt, Vx, Vy, Vz, scal3D)
    end if
    call vtkxml_write(tag_num, scal3D)
    call vtkxml_write(tag_err, scal3D-good_scal)
    call test_check_success(scal3D, good_scal, success, cart_rank)
    call test_substatus('3D test - 3D velo', success, cart_rank)

    ! --- Free memory ---
    call vtkxml_finish()
    deallocate(Vx)
    deallocate(Vy)
    deallocate(Vz)
    deallocate(scal3D)
    deallocate(good_scal)
    deallocate(velo)

    ! === Case with corrected remeshing formula ===
    call discretisation_create(400,400,16,dble(2),dble(2),dble(0.5))
    call set_group_size(40)
    call advec_init(order_bis)
    ! -- Allocation --
    allocate(scal3D(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_scal(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vx(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vy(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vz(N_proc(1), N_proc(2), N_proc(3)))
    ! Initialize output context
    call vtkxml_init_all(9, nb_proc_dim, length, cart_rank, coord,'./adv_res/')
    call vtkxml_init_field(trim(order_bis)//'_tag_num', tag_num)
    call vtkxml_init_field(trim(order_bis)//'_tag_err', tag_err)
    call vtkxml_init_field(trim(order_bis)//'_tag_sol', tag_sol)
    do k = 1, N_proc(3)
        do j = 1, N_proc(2)
            ry = ((j-1+coord(2)*N_proc(2))*d_sc(2)) - length(2)/2.0
            do i = 1, N_proc(1)
                rx = ((i-1+coord(1)*N_proc(1))*d_sc(1))-length(1)/2.0
                rr = (rx**2+ry**2)
                if (rr<1) then
                    scal3D(i,j,k) = (1-rr)**6
                else
                    scal3D(i,j,k) = 0
                end if
                Vx(i,j,k) = cos(3*M_PI*sqrt(rr)/2)*(-ry)
                Vy(i,j,k) = cos(3*M_PI*sqrt(rr)/2)*(rx)
            end do
        end do
    end do
    good_scal = scal3D
    call vtkxml_write(tag_sol, good_scal)
    Vz = 0.0
    T = 0.0
    dt = 3*min(d_sc(1),d_sc(2))
    T_step = 0.0
    T_end = 0.8
    call test_substatus('cfl=3 and dt', dt, cart_rank)
    do while(T< (T_end - dt))
        call advec_step(dt, Vx, Vy, Vz, scal3D)
        T = T + dt
        if (T>T_step) then
            T_step = T_step + period/10
            call test_substatus ('tag, t', T, cart_rank)
            call vtkxml_write(tag_num, scal3D)
        end if
    end do
    if (T<T_end) then
        dt = T_end - T
        call advec_step(dt, Vx, Vy, Vz, scal3D)
    end if
    call vtkxml_write(tag_num, scal3D)
    call vtkxml_write(tag_err, scal3D-good_scal)
    call test_substatus ('time', (T+dt), cart_rank)
    call test_check_success(scal3D, good_scal, success, cart_rank)
    call test_substatus('3D test - with tag', success, cart_rank)


    call discretisation_default()
    call advec_init(order_bis)
    
    ! --- Free memory --- 
    call vtkxml_finish()

    success = .not.success

end function test_part_advec_3D



!> Test devoted to validate advection solver along one direction
!!     @param[in]   init_scal   =  optional parameter to initialise the scalar fied to
!!                                  a constant one or to a sphere shape
!!     @return      error       = test error (= false if the code pass the test) (= not success)
function test_advecY(init_scal) result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use advecY
    use cart_topology
    ! Test procedures
    use advec_aux_init
    use test_common

    logical                                     :: success
    character(len=*), intent(in), optional      :: init_scal

    character(len=17)                           :: initialisation   ! to choose how to initialise the scalar field
    character(len=str_short)                    :: order            ! space order of the solveur
    real(WP), dimension(:, :, :), allocatable   :: scal3D           ! the scalar field
    real(WP), dimension(:, :, :), allocatable   :: Vy               ! the flow
    real(WP)                                    :: velocity         ! constant velocity of the flux
    integer                                     :: ierr             ! mpi error code
    integer                                     :: i,j,k            ! mesh indice   
    integer                                     :: T_step           ! time
    integer                                     :: T_end            ! final time
    real(WP)                                    :: dt               ! time step
    real(WP), dimension(:, :, :), allocatable   :: good_scal        ! analytic solution
    real(WP), dimension(3)                      :: translat         ! to compute analytic solution

    allocate(scal3D(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_scal(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vy(N_proc(1), N_proc(2), N_proc(3)))

    if(present(init_scal)) then
        initialisation = init_scal
    else 
        initialisation = 'constant'
    end if
    success = .true.


    ! Initialize the particular solver
    order = 'p_O2'
    call advec_init(order)


    ! Initialise the velocity
    velocity = N(2)/11*d_sc(2)
    Vy = velocity
    T_end = 1
    dt = 0.1
    translat = 0
    translat(2) = - velocity*T_end
    translat = translat/d_sc

    ! Initialise the scalar field
    call scal_init(initialisation, scal3D, good_scal, translat)
    call test_substatus('initialisation', success, cart_rank)

    ! Advec it with a particular method
    dt = 1
    call advecY_calc(dt, Vy, scal3D, 'p_O2')
    call test_check_success(scal3D, good_scal, success, cart_rank)
    call test_substatus('advec along Y', success, cart_rank)


    ! --- Free memory ---
    deallocate(scal3D)
    deallocate(good_scal)
    deallocate(Vy)

    success = .not.success

end function test_advecY


!> Test devoted to validate the 3D advection solver
!!     @return      error       = test error (= false if the code pass the test) (= not success)
function test_advec_rot() result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use cart_topology
    use vtkxml
    ! Test procedures
    use advec_aux_init
    use test_common

    logical                                     :: success

    character(len=17)                           :: init             ! to choose how to initialise the scalar field
    character(len=str_short)                    :: order            ! space order of the solveur
    real(WP), dimension(:, :, :), allocatable   :: scal3D           ! the scalar field
    real(WP), dimension(:, :, :), allocatable   :: VX, Vy, Vz       ! the flow
    real(WP)                                    :: velocity         ! constant velocity of the flux
    integer                                     :: ierr             ! mpi error code
    integer                                     :: i,j,k            ! mesh indice   
    real(WP)                                    :: T                ! time
    real(WP)                                    :: T_end            ! final time
    real(WP)                                    :: T_ite            ! time corresponding to output
    real(WP)                                    :: dt               ! time step
    real(WP), dimension(:, :, :), allocatable   :: good_scal        ! analytic solution
    real(WP), dimension(:, :, :), allocatable   :: good_velo        ! temp field
    real(WP), dimension(:), allocatable         :: pos_adim         ! temp field
    real(WP), dimension(3)                      :: translat         ! to compute analytic solution
    integer                                     :: tag_rot          ! tag for visualisation context
    integer                                     :: tag_sol          ! tag for visualisation context


    allocate(scal3D(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_scal(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vx(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vy(N_proc(1), N_proc(2), N_proc(3)))
    allocate(Vz(N_proc(1), N_proc(2), N_proc(3)))
    allocate(good_velo(N_proc(1), N_proc(2), N_proc(3)))

    init = '2D_rot'
    success = .true.
    ! Initialize output context
    call vtkxml_init_all(2, nb_proc_dim, length, cart_rank, coord,'./adv_res/')


    ! Initialize the particular solver
    order = 'p_O2'
    call advec_init(order)
    call vtkxml_init_field('turning', tag_rot)
    call vtkxml_init_field('turn_sol', tag_sol)


    ! Initialise the velocity
    call scal_init(init, scal3D, good_scal)
    period = 1
    dt = 0.02
    allocate(pos_adim(N_proc(1)))
    pos_adim = 1
    call velo_init(init, Vx, 1, pos_adim, good_velo)
    call velo_init(init, Vy, 2, pos_adim, good_velo)
    call velo_init(init, Vz, 3, pos_adim, good_velo)
    call vtkxml_write(tag_rot, scal3D, 'turning')
    call test_substatus('initialisation', success, cart_rank)

    ! Advec it with a particular method
    T = 0
    T_end = 4*period
    T_ite = 0
    do while (T<=T_end - dt)
        call advec_step(dt, Vx, Vy, Vz, scal3D)
        T = T + dt
        if (T>T_ite) then
            T_ite = T_ite + T_end/30
            call test_substatus ('time', T, cart_rank)
            call vtkxml_write(tag_rot, scal3D, 'turning')
        end if
    end do
    if (T<T_end) then
        dt = T_end - T
        call advec_step(dt, Vx, Vy, Vz, scal3D)
    end if
    call vtkxml_write(tag_rot, scal3D, 'turning')
    call vtkxml_write(tag_sol, good_scal)

    call test_check_success(scal3D, good_scal, success, cart_rank)
    call test_substatus('turning sphere', success, cart_rank)


    ! --- Free memory ---
    call vtkxml_finish()
    deallocate(scal3D)
    deallocate(good_scal)
    deallocate(Vx)
    deallocate(Vy)
    deallocate(Vz)

    success = .not.success

end function test_advec_rot

end module advec_aux

!> @}
