!> @addtogroup part_test
!! @{
    
!------------------------------------------------------------------------------
!
! MODULE: test_part_common
!
! DESCRIPTION: 
!> This module provides tests to validate the particular solver. It is
!! more precisly focused on testing the "common part" (advec_common) used for each
!! directions.
!!
!! @author
!! Jean-Baptiste Lagaert, LEGI
!!
!! @details 
!!  This module is devoted to validate all the procedure from "advec_common". All the
!! tests are unit tests: they return a logical value to check if the code version pass
!! it or not. 
!!
!! That is all these test are logical function, they return true if the result
!! is the right one and false otherwise. 
!! All the "test_part_*" function are devoted to validate the particular
!! solver
!! This module provide the following test: 
!!    - Validate the particular method, step by step
!!        - Test the procedure "AC_obtain_senders" from advec_common
!!        - Validate the redistribution the buffer during the remeshing
!!        - Validate the redistribution the buffer during the remeshing -
!!            debug version : only one processus contains non-zero field.
!!        - Validate the remeshing of untagged particles
!!        - Validate the velocity interpolation (the RK2 scheme use the
!!            velocity at midlle point to advec particles).
!!        - Validate how type (left or center) and tag are computing for a
!!            single line
!!        - Validate how type (left or center) and tag are computing for a
!!            group of line
!!
!
!------------------------------------------------------------------------------

module advec_aux_common

    use string
    use precision
    implicit none



    ! ===== Test for the particles solver =====
    ! Public function
    public              :: test_part_AC_velo_determine_com
    !public              :: test_part_AC_bufferToScalar
    !public              :: test_part_AC_bufferToScalar_Deb
    !public              :: test_part_AC_interpol_velocity
    public              :: test_part_AC_type_and_block_O2_group




contains


!!> Particles method: validation of the procedure "AC_bufferToScalar".
!!!     @return      success = test success (= false if the code pass the test)
!!!     @param[in]   shift   = global translation of indices (optional)
!!! @details
!!! Test the procedure "AC_obtain_senders" wich send local buffer use for remeshing
!!! to the right processes. This procedure belong to "advec_common". 
!function test_part_AC_bufferToScalar(shift) result(success)
!
!    ! Library
!    use mpi
!    ! Scale code
!    use advec
!    use advec_common    ! Some porcedure common to advection along all directions
!    use advec_variables ! contains info about solver parameters and others.
!    use cart_topology
!    ! Test procdure
!    use test_common
!
!    integer, intent(in), optional       :: shift
!    logical                             :: success
!
!    integer                             :: shift_bis    ! global translation of indices
!    character(str_short)                :: order        ! order of the particles solver
!    integer                             :: j_min, j_max ! input argument of the tested procedure
!    real(WP), dimension(:), allocatable :: send_buffer  ! the buffer to redistribute
!    real(WP), dimension(:), allocatable :: scal1D       ! the scalar field
!    integer                             :: direction    ! direction (2 if along Y, 3=along Z)
!    real(WP)                            :: good_scal    ! theoritical value of scal1D
!    integer, dimension(2)               :: ind_group    ! indice of current group of lines
!    integer, dimension(2)               :: rece_proc    ! minimal and maximal gap between my coordinate and the one from which 
!                                                        ! I will receive data
!    integer                             :: proc_min     ! smaller gap between me and the processes to where I send data
!    integer                             :: proc_max     ! smaller gap between me and the processes to where I send data
!
!    ! Some initialisation
!    success = .true.
!    ind_group = 1
!    if (present(shift)) then
!        shift_bis = shift
!    else
!        shift_bis = 0
!    end if
!    call test_substatus('shift', shift_bis, cart_rank)
!    
!    ! Initialize the particular solver
!    order = 'p_O2'
!    call advec_init(order, verbosity=.false.)
!
!    do direction = 1, 3
!        if (nb_proc_dim(direction)>1) then 
!            call test_substatus('direction', direction, cart_rank)
!            good_scal = modulo(coord(direction)-shift_bis, nb_proc_dim(direction))
!            
!            ! Test 1 - unrealistic case with no communication
!            ! Initialize the buffer to remesh
!            call AC_set_part_bound_size(0)
!            j_min = 1 + shift_bis*N_proc(direction)
!            j_max = N_proc(direction)*(shift_bis+1)
!            allocate (send_buffer(j_min:j_max))
!            allocate (scal1D(1:N_proc(direction)))
!            send_buffer = coord(direction)
!            scal1D = 0.0
!            ! Let's go !
!            ! Determine the communication needed : who will communicate whit who?
!            call AC_obtain_senders_line(j_min, j_max, direction, ind_group, proc_min, proc_max, rece_proc)
!            ! And then distribute buffer along the processus
!            call AC_bufferToScalar(direction, ind_group, j_min, j_max, proc_min, proc_max, rece_proc, send_buffer, scal1D)
!            ! Check the success
!            call test_check_success(scal1D, good_scal, success, cart_rank)
!            deallocate (send_buffer)
!            deallocate (scal1D)
!            call test_substatus('just me', success, cart_rank)
!
!            ! I communicate with my two neighbors
!            call advec_init(order, verbosity=.false.)
!            j_min = shift_bis*N_proc(direction)
!            j_max = (1+shift_bis)*N_proc(direction) -1+2*bl_bound_size
!            allocate (send_buffer(j_min:j_max))
!            allocate (scal1D(1:N_proc(direction)))
!            send_buffer = coord(direction)
!            send_buffer(j_min) = modulo(coord(direction)-1, nb_proc_dim(direction))/2.0
!            send_buffer(j_min+1) = send_buffer(j_min+1)/2.
!            send_buffer(j_max) = modulo(coord(direction)+1, nb_proc_dim(direction))/2.0
!            send_buffer(j_max-1) = send_buffer(j_max-1)/2.
!            scal1D = 0.0
!            ! Determine the communication needed : who will communicate whit who?
!            call AC_obtain_senders_line(j_min, j_max, direction, ind_group, proc_min, proc_max, rece_proc)
!            ! And then distribute buffer along the processus
!            call AC_bufferToScalar(direction, ind_group, j_min, j_max, proc_min, proc_max, rece_proc, send_buffer, scal1D)
!            ! Check the success
!            call test_check_success(scal1D, good_scal, success, cart_rank)
!            deallocate (send_buffer)
!            deallocate (scal1D)
!            call test_substatus('me and my neighbors', success, cart_rank)
!        else 
!            call test_substatus('only one proc along direction, no test', cart_rank)
!        end if
!    end do
!
!    success = .not.success
!
!end function test_part_AC_bufferToScalar
!
!
!
!!> Particles method: validation of the procedure "AC_bufferToScalar"
!!!     @return      success = test success (= false if the code pass the test)
!!!     @param[in]   shift   = global translation of indices (optional)
!!!     @param[in]   rank    = rank of the processus which will contains non-zero field.
!!! @details
!!!    Debugging version : only one processus contains non zero field to remesh.
!!!    Each processus communicate with its two neibor. A shift could be add.
!function test_part_AC_bufferToScalar_Deb(rank, shift) result(success)
!
!    ! Library
!    use mpi
!    ! Scale code
!    use advec
!    use advec_common    ! Some porcedure common to advection along all directions
!    use advec_variables ! contains info about solver parameters and others.
!    use cart_topology
!    ! Test procdure
!    use test_common
!
!    integer, intent(in)                 :: rank
!    integer, intent(in), optional       :: shift
!    logical                             :: success
!
!    integer                             :: ierr         ! mpi error code
!    integer                             :: shift_bis    ! global translation of indices
!    character(str_short)                :: order        ! order of the particles solver
!    integer                             :: j_min, j_max ! input argument of the tested procedure
!    real(WP), dimension(:), allocatable :: send_buffer  ! the buffer to redistribute
!    real(WP), dimension(:), allocatable :: scal1D       ! the scalar field
!    real(WP), dimension(:), allocatable :: good_scal    ! theoritical value of scal1D
!    integer                             :: direction    ! direction (2 if along Y, 3=along Z)
!    integer, dimension(3)               :: mycoord      ! my coordonates in the mpi topology
!    integer                             :: rank_shift   ! rank of processus of rank = rank+shift
!    integer, dimension(2)               :: ind_group    ! indice of current group of lines
!    integer, dimension(2)               :: rece_proc    ! minimal and maximal gap between my coordinate and the one from which 
!                                                        ! I will receive data
!    integer                             :: proc_min     ! smaller gap between me and the processes to where I send data
!    integer                             :: proc_max     ! smaller gap between me and the processes to where I send data
!
!    ! Some initialisation
!    success = .true.
!    ind_group = 1
!    if (present(shift)) then
!        shift_bis = shift
!    else
!        shift_bis = 0
!    end if
!    call test_substatus('shift', shift_bis, cart_rank)
!    call test_substatus('rank tested', rank, cart_rank)
!
!    ! Initialize the particular solver
!    order = 'p_O2'
!    call advec_init(order, verbosity=.false.)
!    
!    do direction = 1, 3
!        call test_substatus('direction', direction, cart_rank)
!        if (nb_proc_dim(direction)>1) then 
!            ! Initialise the solver environnement and the field
!            call advec_init(order, verbosity=.false.)
!            j_min = shift_bis*N_proc(direction)
!            j_max = (1+shift_bis)*N_proc(direction) -1+2*bl_bound_size
!            allocate (send_buffer(j_min:j_max))
!            allocate (scal1D(1:N_proc(direction)))
!            send_buffer = 0
!            if (cart_rank == rank) then
!                send_buffer = 2
!                send_buffer(j_min) = 1
!                send_buffer(j_max) = 3
!            end if
!            scal1D = 0.0
!
!            ! -- Compute the analytic solution --
!            allocate (good_scal(1:N_proc(direction)))
!            good_scal = 0.0
!            ! For the processus wich correspond to me after the shift
!            call mpi_cart_coords(cart_comm, rank, 3, mycoord, ierr)
!            mycoord(direction) = mycoord(direction) + shift_bis
!            call mpi_cart_rank(cart_comm, mycoord, rank_shift, ierr)
!            if (cart_rank==rank_shift)   good_scal = 2
!            ! For the next processus in the current direction
!            mycoord(direction) = mycoord(direction) + 1
!            call mpi_cart_rank(cart_comm, mycoord, rank_shift, ierr)
!            if (cart_rank==rank_shift)   good_scal(1) = 3
!            ! For the previous processus in the current direction
!            mycoord(direction) = mycoord(direction) -2
!            call mpi_cart_rank(cart_comm, mycoord, rank_shift, ierr)
!            if (cart_rank==rank_shift)   good_scal(N_proc(direction)) = 1
!
!            ! -- Compute the numerical solution --
!            ! Determine the communication needed : who will communicate whit who?
!            call AC_obtain_senders_line(j_min, j_max, direction, ind_group, proc_min, proc_max, rece_proc)
!            ! And then distribute buffer along the processus
!            call AC_bufferToScalar(direction, ind_group, j_min, j_max, proc_min, proc_max, rece_proc, send_buffer, scal1D)
!            ! Check the success
!            call test_check_success(scal1D, good_scal, success, cart_rank)
!            deallocate (send_buffer)
!            deallocate (scal1D)
!            deallocate (good_scal)
!            call test_substatus('me and my neighbors', success, cart_rank)
!        else 
!            call test_substatus('only one proc along direction, no test', cart_rank)
!        end if
!    end do
!
!    success = .not.success
!
!
!end function test_part_AC_bufferToScalar_Deb


!> Particles method: validation of the procedure "AC_velocity_determine_communication" wich determine
!! who will comunicate with who during the remeshing
!!     @return      success = test success (= false if the code pass the test)
!!     @param[in]   shift   = global translation of indices (optional)
function test_part_AC_velo_determine_com(shift) result(success)

    ! Library
    use mpi
    ! Scale code
    use advec
    use advec_common    ! Some porcedure common to advection along all directions
    use advec_variables ! contains info about solver parameters and others.
    use cart_topology
    ! Test procdure
    use test_common

    integer, intent(in), optional               :: shift
    logical                                     :: success          
    ! Be aware : during this function, success = true if everything is right, but we return not success = error

    integer                                     :: shift_bis    ! global translation of indices
    character(str_short)                        :: order        ! order of the particles solver
    integer                                     :: ierr         ! mpi error code
    integer , dimension(5,5)                    :: rece_ind_min ! minimal indice of mesh involved in remeshing particles (of my local subdomains)
    integer , dimension(5,5)                    :: rece_ind_max ! maximal indice of mesh involved in remeshing particles (of my local subdomains)
    integer, dimension(5,5,2)                   :: rece_gap     ! distance between me and processus wich send me information
    integer, dimension(2 , 2)                   :: send_gap     ! distance between me and processus to wich I send information
    integer, dimension(2)                       :: rece_gap_abs ! min (resp max) value of rece_gap(:,:,i) with i=1 (resp 2)
    integer, dimension(:), allocatable          :: rece_rank    ! rank of processus from which I receive data
    integer, dimension(2 , 2)                   :: send_theoric ! theoritical value of send_gap
    integer, dimension(2)                       :: ind_group    ! indice of current group of lines
    integer                                     :: direction    ! current direction (alonG X, Y or Z)
    integer, dimension(:,:), allocatable        :: cartography  ! cartography(proc_gap) contains the set of the lines indice in the block for wich the 
                                                                ! current processus requiers data from proc_gap and for each of these lines the range 
                                                                ! of mesh points from where it requiers the velocity values.
    integer, dimension(:,:), allocatable        :: carto_th     ! theoritical value of cartography
    integer                                     :: max_size
    integer                                     :: proc_gap     ! gap between two different mpi processus
    integer, dimension(2)                       :: gs           ! group size


    ! Some initialisation
    success = .true.
    ind_group = 1
    gs = (/5,5/)
    direction = 3
    if (present(shift)) then
        shift_bis = shift
    else
        shift_bis = 0
    end if
    call test_substatus('shift', shift_bis, cart_rank)

    ! Initialize the particular solver
    order = 'p_O2'
    call advec_init(order, verbosity=.false.)
    call test_substatus('initialisation solveur', success,  cart_rank)

    ! ===== Each line of the group is initialized considering it second indice =====
    ! -- First line communcate just with myself (ie it host processus) --
    ! Init the indice range for each lines.
    rece_ind_min(:,1) = 1 + shift_bis*N_proc(direction)
    rece_ind_max(:,1) = N_proc(direction)*(shift_bis+1)
    ! -- Second line communicate with me and the previous processus --
    rece_ind_min(:,2) = shift_bis*N_proc(direction)
    rece_ind_max(:,2) = (1+shift_bis)*N_proc(direction)
    ! -- Third line communicate with me and my two neighbors --
    rece_ind_min(:,3) = shift_bis*N_proc(direction)
    rece_ind_max(:,3) = (1+shift_bis)*N_proc(direction) + 1
    ! -- Fourth line communicate with me and the next processus --
    rece_ind_min(:,4) = shift_bis*N_proc(direction) + 1
    rece_ind_max(:,4) = (1+shift_bis)*N_proc(direction) + 1
    ! -- And for the fith line it depend of it firts indice inside the group --
    rece_ind_min(:,5) = 1 + shift_bis*N_proc(direction)
    rece_ind_min(3,5) = shift_bis*N_proc(direction)
    rece_ind_max(:,5) = (1+shift_bis)*N_proc(direction) + 1
    rece_ind_max(2,5) = (1+shift_bis)*N_proc(direction)

    ! Compute the associated processus
    rece_gap(:,:,1) = floor(real(rece_ind_min-1)/N_proc(direction))
    rece_gap(:,:,2) = floor(real(rece_ind_max-1)/N_proc(direction))
    rece_gap_abs(1) = minval(rece_gap(:,:,1))
    rece_gap_abs(2) = maxval(rece_gap(:,:,2))
    allocate(rece_rank(rece_gap_abs(1):rece_gap_abs(2)))
    max_size = 2 + gs(2)*(2+3*gs(1))
    allocate(cartography(max_size,rece_gap_abs(1):rece_gap_abs(2)))


    ! -- And launch the test --
    cartography = -1
    call AC_velocity_determine_communication(direction, ind_group, gs, send_gap,  &
    & rece_gap, rece_gap_abs, rece_rank, cartography)

    ! -- Check result --
    ! send_gap
    send_theoric(1,:) = (/-1-shift_bis,1-shift_bis/)
    send_theoric(2,:) = 25
    call test_check_success(send_gap, send_theoric, success, cart_rank)
    call test_substatus('send_gap', success, cart_rank)
    ! cartography
    allocate(carto_th(max_size, rece_gap_abs(1):rece_gap_abs(2)))
    carto_th = -1
    if (rece_rank(-1+shift_bis) /= D_rank(direction)) then
        carto_th(1:13,-1+shift_bis)= (/0,13,0,2,2,0,2,1,5,1,5,3,3/)
    else
        carto_th(1:7,-1+shift_bis)= (/0,7,0,0,0,0,0/)
    end if
    if (rece_rank(shift_bis) /= D_rank(direction)) then
        carto_th(1:17,shift_bis)= (/0,17,2,2,2,2,2,1,5,1,5,1,5,1,5,1,5/)
    else
        carto_th(1:7,shift_bis)= (/0,7,0,0,0,0,0/)
    end if
    if (rece_rank(1+shift_bis) /= D_rank(direction)) then
        carto_th(1:15,1+shift_bis)= (/0,15,0,0,2,2,4,1,5,1,5,1,1,3,5/)
    else
        carto_th(1:7,1+shift_bis)= (/0,7,0,0,0,0,0/)
    end if
    call test_check_success(cartography, carto_th, success, cart_rank)
    call test_substatus('cartography', success, cart_rank)

!        ! -- For the fourth: Contraction / extention --
!        if (modulo(nb_proc_dim(direction),2)==0) then
!            if (modulo(D_rank(direction),2)== 0) then
!                ! Contraction
!                j_min = 1 + shift_bis*N_proc(direction)
!                j_max = (1+shift_bis)*N_proc(direction)
!                send_begin = 0
!                send_end = 0
!            else
!                ! Dilatation
!                j_min = +1 - 2*bl_bound_size + shift_bis*N_proc(direction)
!                j_max = (1+shift_bis)*N_proc(direction) + 2*bl_bound_size 
!                send_begin = -1
!                send_end = 1
!            end if

    deallocate(rece_rank)
    deallocate(cartography)
    deallocate(carto_th)

    success = .not.success

end function test_part_AC_velo_determine_com


!> Particles method: validation of the velocity interpolation
!!     @return      success     = test success (= false if the code pass the test)
!!     @param[in]   direction   = 1 for along X, 2 for along Y, 3 for along Z
!!     @param[in]   shift       = global translation of indice
function test_part_AC_velo_compute_group(direction, shift) result(success)

    ! External Library
    use mpi
    ! Scales code
    use cart_topology
    use advec_common
    ! Test procedures
    use advec_aux_init
    use test_common

    logical                                     :: success          
    integer, intent(in)                         :: direction, shift

    integer                                                 :: ind,ind1,ind2! indices
    integer                                                 :: nn, clock    ! to generate random number
    integer, dimension(:), allocatable                      :: seed         ! to generate random number
    real(WP), dimension(N_proc(direction), 5, 5)            :: p_pos        ! location where the velocity is interpolated
    real(WP), dimension(N_proc(direction))                  :: r            ! random numbers to initialise p_pos
    real(WP), dimension(N_proc(direction),5,5)              :: velo         ! velocity field
    real(WP), dimension(N_proc(direction),5,5)              :: good_velo    ! analytic interpolation of velocity field in particles position
    real(WP)                                                :: dt           ! time step
    integer                                                 :: T            ! for solver iteration
    integer, dimension(2)                                   :: gs           ! group size

    ! Initialisation
    success = .true.
    call test_substatus('test velocity interpolation', cart_rank)
    call test_substatus('shift', shift, cart_rank)
    ! To generate random number
    call RANDOM_SEED(size = nn)
    allocate(seed(nn))
    call SYSTEM_CLOCK(COUNT=clock)   
    seed = clock + 37 * (/ (ind - 1, ind = 1, nn) /)
    call RANDOM_SEED(PUT = seed)   
    deallocate(seed)
    ! Position where the velocity will be interpolated
    CALL RANDOM_NUMBER(r)
    do ind = 1, N_proc(direction)
        p_pos(ind,:,:) = (ind+shift)
    end do
    gs = group_size(direction,:)
    call test_check_success((/5,5/),gs, success, cart_rank)
    call test_substatus('group is 5', success, cart_rank)
    if (success) then

        ! ===== Init particle positions =====
        ! -- First column : particle are inside current and previous processes --
        p_pos(:,1,1) = p_pos(:,1,1) - 3 + r*d_sc(direction)
        p_pos(:,2,1) = p_pos(:,2,1) - 3
        p_pos(:,3,1) = p_pos(:,3,1) - 2 + r*d_sc(direction)
        p_pos(:,4,1) = p_pos(:,4,1) - 2
        p_pos(:,5,1) = p_pos(:,5,1) - 1 + r*d_sc(direction)
        ! -- Second and third  column : they are on the current processus and my two neighbors --
        p_pos(:,:,2) = p_pos(:,:,1) + 1
        p_pos(:,:,3) = p_pos(:,:,1) + 2
        ! -- Fourth column: paritcles are on the current and th next processes --
        p_pos(:,:,4) = p_pos(:,:,1) + 3


        ! ===== First test: constant velocity =====
        do ind2 = 1, 5
            do ind1 = 1, 5
                call particle_velo_init('constant', velo(:,ind1,ind2), direction, p_pos(:,ind1,ind2), good_velo(:,ind1,ind2))
            end do
        end do
        dt = 0.0
        call AC_particle_velocity(dt, direction, gs, (/1,1/), p_pos, velo)
        ! Check result
        call test_check_success(velo, good_velo, success, cart_rank)
        call test_substatus('constant velocity', success, cart_rank)

        ! ===== Second test: unconstant velocity =====
        do ind2 = 1, 5
            do ind1 = 1, 5
                call particle_velo_init('translation_field', velo(:,ind1,ind2), direction, &
                    & p_pos(:,ind1,ind2), good_velo(:,ind1,ind2))
            end do
        end do
        call AC_particle_velocity(dt, direction, gs, (/1,1/), p_pos, velo)
        ! Check result
        call test_check_success(velo, good_velo, success, cart_rank)
        call test_substatus('unconstant velocity', success, cart_rank)
    end if

    ! Return not success
    success = .not. success

end function test_part_AC_velo_compute_group


!> Particle method: validation of particle tag and block type determination for
!! a group of line
!!     @return      error   = test error (= false if the code pass the test) (= not success)
function test_part_AC_type_and_block_O2_group() result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use advec_common    ! Some porcedure common to advection along all directions
    use advec_variables ! contains info about solver parameters and others.
    use cart_topology
    ! Test procedures
    use test_common

    logical                                     :: success

    integer                                     :: shift_bis    ! global translation of indices
    character(str_short)                        :: order        ! order of the particles solver
    
    logical, dimension(:,:,:), allocatable      :: bl_type      ! correct (analytic) type of block
    logical, dimension(:,:,:), allocatable      :: bl_tag       ! correct (analytic) tag of block
    logical, dimension(:,:,:), allocatable      :: good_type    ! correct (analytic) type of block
    logical, dimension(:,:,:), allocatable      :: good_tag     ! correct (analytic) tag of block
    integer                                     :: direction    ! direction (2 if along Y, 3=along Z)
    real(WP)                                    :: dt, cfl      ! time step and CFL number
    real(WP), dimension(:,:,:), allocatable     :: p_V          ! particle velocity (used to tag particles)
    integer                                     :: ind, ind2    ! indice of the current particle
    integer                                     :: ind_bl       ! indice of the current block
    real(WP)                                    :: lambda_min   ! minimum courant number on a block
    integer                                     :: ind_tag_bl   ! indice of the first tagged block in the test
    integer                                     :: ind_tag_p    ! indice of particle corresponding to the velocity 
                                                                ! variation wich induce a tag of the block ind_tag_bl 
    integer                                     :: ind_tag_bl2  ! indice of the second tagged block in the test
    integer                                     :: ind_tag_p2   ! indice of particle wich induce a tag of the block ind_tag_bl2


    ! Initialisation of context (solver, ...)
    success = .true.
    order ='p_O2'
    call set_group_size(5)
    call advec_init(order, verbosity=.false.)
    direction = 3

    ! Allocation
    allocate(bl_type(bl_nb(direction)+1,group_size(direction,1),group_size(direction,2)))
    allocate(good_type(bl_nb(direction)+1,group_size(direction,1),group_size(direction,2)))
    allocate(bl_tag(bl_nb(direction),group_size(direction,1),group_size(direction,2)))
    allocate(good_tag(bl_nb(direction),group_size(direction,1),group_size(direction,2)))
    allocate(p_V(N_proc(direction),group_size(direction,1),group_size(direction,2)))

    ! Initialize of the different parameter and field
    dt = 1
    cfl = dt/d_sc(direction)
    p_V = 0

    ! Test case = nothing on line of indice different from (2,3)
    ! On that line, first part of block are left one, and second part are center.
    ! Due to periodic condition, there is also 2 tag, one for each type switch
    ! (left => center, and center => left).
    ! Update dt in order to create "chock" without broke the stability condition
    dt = 0.8*d_sc(direction)/sqrt(2*1.2*bl_size)
    cfl = dt/d_sc(direction)
    good_type = .false. 
    good_tag = .false.
    ind_tag_bl = (nb_proc_dim(direction)*bl_nb(direction)/3) - (bl_nb(direction)*coord(direction))
    ind_tag_p = (ind_tag_bl-1)*bl_size + bl_size/2 + 1
    ind_tag_p2 = floor((0.3/0.8)*(N(direction)-N_proc(direction)*coord(direction)-ind_tag_p) + ind_tag_p)
    ind_tag_p2 = ind_tag_p2 + 1
    ind_tag_bl2 = (ind_tag_p2 - 1 - bl_size/2)/bl_size + 1
    if (modulo(ind_tag_p2-1-bl_size/2, bl_size)==0) ind_tag_bl2 = ind_tag_bl2 -1
    if (ind_tag_bl <=bl_nb(direction)) then
        if (ind_tag_bl>=1) then
            ! Tag the right block transition
            good_tag(ind_tag_bl, 2, 3) = .true.
        else 
            ! Change the velocity for the first half block
            lambda_min = max(0,shift_bis) + 10
            do ind = 1, bl_size/2
                p_V(ind, 2, 3) = 0.8*(1.0-float(ind-ind_tag_p) &
                    & /(N(direction)-N_proc(direction)*coord(direction)-ind_tag_p))/cfl
            end do
            if (ind+1<ind_tag_p2) good_type(1, 2, 3) = .true.
        end if
        ! Update velocity and block type
        do ind_bl = max(2, ind_tag_bl+1), bl_nb(direction)
            lambda_min = max(0,shift_bis) + 10
            do ind = 1, bl_size
                ind2 = ind+((ind_bl-2)*bl_size)+bl_size/2 ! the first block is only a half block
                p_V(ind2, 2, 3) = 0.8*(1.-float(ind2-ind_tag_p) &
                        &/(N(direction)-N_proc(direction)*coord(direction)-ind_tag_p))/cfl
            end do
            if (ind2+1<ind_tag_p2) good_type(ind_bl, 2, 3) = .true.
        end do
        ! For the last half block
        lambda_min = max(0,shift_bis) + 10
        ind_bl = bl_nb(direction) + 1
        do ind = 1, bl_size/2
            ind2 = ind+((ind_bl-2)*bl_size)+bl_size/2 ! the first block is only a half block
            p_V(ind2, 2 , 3) = 0.8*(1.-float(ind2-ind_tag_p) &
                    & /(N(direction)-N_proc(direction)*coord(direction)-ind_tag_p))/cfl
        end do
        if (ind2+bl_size/2+1<ind_tag_p2) good_type(ind_bl, 2, 3) = .true.
    end if
    if ((ind_tag_bl2 <= bl_nb(direction)).and.(ind_tag_bl2>0)) good_tag(ind_tag_bl2, 2, 3) = .true.
    call AC_type_and_block(dt, direction, group_size(direction,:), (/1,1/), p_V, bl_type, bl_tag)


    call test_check_success(bl_tag, good_tag, success, cart_rank)
    call test_substatus('test 3 - two tag ', success, cart_rank)
    call test_check_success(bl_type, good_type, success, cart_rank)
    call test_substatus('test 3 - center at first, then left', success, cart_rank)


    ! Deallocation
    deallocate(bl_type)
    deallocate(good_type)
    deallocate(bl_tag)
    deallocate(good_tag)
    deallocate(p_V)

    success = .not. success

end function test_part_AC_type_and_block_O2_group


! ============================================================
! ==========          Test about remeshing          ==========
! ============================================================

!> Particle method: validation of computation of remeshing range for a group of line
!!     @param[in]   shift       = global translation of indice
!!     @return      error   = test error (= false if the code pass the test) (= not success)
function test_part_AC_remesh_range() result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use advec_common    ! Some porcedure common to advection along all directions
    use advec_variables ! contains info about solver parameters and others.
    use cart_topology
    ! Test procedures
    use test_common
    use advec_aux_init

    logical                                 :: success

    integer                                 :: shift_bis    ! global translation of indices
    character(str_short)                    :: order        ! order of the particles solver
    
    logical, dimension(:,:,:), allocatable  :: bl_type      ! type of block
    logical, dimension(:,:,:), allocatable  :: bl_tag       ! tag of block
    integer                                 :: direction    ! direction (2 if along Y, 3=along Z)
    real(WP), dimension(:,:,:), allocatable :: p_pos_adim   ! particle position
    integer, dimension(:,:), allocatable    :: send_min     ! first mesh where particles are remeshed
    integer, dimension(:,:), allocatable    :: good_send_min! theoritical first mesh where particles are remeshed
    integer, dimension(:,:), allocatable    :: send_max     ! last mesh where particles are remeshed
    integer, dimension(:,:), allocatable    :: good_send_max! theoritical last mesh where particles are remeshed
    integer, dimension(:,:,:), allocatable  :: send_gap     ! distance between me and processus wich send me information
    integer, dimension(2)                   :: send_gap_abs ! min (resp max) value of rece_gap(:,:,i) with i=1 (resp 2)
    integer, dimension(2)                   :: gs           ! group size
    integer, dimension(2)                   :: ind_group    ! group indice


    ! Initialisation of context (solver, ...)
    success = .true.
    direction = 2
    shift_bis = 0.0
    ind_group = (/1,1/)
    call set_group_size(5)
    gs = group_size(direction,:)

    ! Allocation
    allocate(bl_type(bl_nb(direction)+1,group_size(direction,1),group_size(direction,2)))
    allocate(bl_tag(bl_nb(direction),group_size(direction,1),group_size(direction,2)))
    allocate(p_pos_adim(N_proc(direction),group_size(direction,1),group_size(direction,2)))
    allocate(send_min(group_size(direction,1),group_size(direction,2)))
    allocate(send_max(group_size(direction,1),group_size(direction,2)))
    allocate(good_send_min(group_size(direction,1),group_size(direction,2)))
    allocate(good_send_max(group_size(direction,1),group_size(direction,2)))
    allocate(send_gap(group_size(direction,1),group_size(direction,2),2))

    ! ----- Initialisation -----
    bl_type = .true.
    bl_tag = .false.
    call pos_gp_init(direction, gs, shift_bis, p_pos_adim, bl_type)

    ! ----- Lambda 2 corrected -----
    order ='p_O2'
    ! Init solver
    call advec_init(order, verbosity=.false.)
    ! Test procedure
    call AC_remesh_range(bl_type, p_pos_adim, direction, send_min, send_max, send_gap, send_gap_abs)
    ! Compute theoric value
    where (bl_type(1,:,:))
        ! First particle is a centered one
        send_min = nint(p_pos_adim(1,:,:))-1
    elsewhere
        ! First particle is a left one
        send_min = floor(p_pos_adim(1,:,:))-1
    end where
    where (bl_type(N_proc(direction)/bl_size +1,:,:))
        ! Last particle is a centered one
        send_max = nint(p_pos_adim(N_proc(direction),:,:))+1
    elsewhere
        ! Last particle is a left one
        send_max = floor(p_pos_adim(N_proc(direction),:,:))+1
    end where
    ! Check results
    call test_check_success(send_min, good_send_min, success, cart_rank)
    call test_check_success(send_max, good_send_max, success, cart_rank)
    call test_substatus('remesh range - order 2 ', success, cart_rank)

    ! ----- Lambda 4 corrected -----
    order ='p_O4'
    ! Init solver
    call advec_init(order, verbosity=.false.)
    ! Test procedure
    call AC_remesh_range(bl_type, p_pos_adim, direction, send_min, send_max, send_gap, send_gap_abs)
    ! Compute theoric value
    where (bl_type(1,:,:))
        ! First particle is a centered one
        send_min = nint(p_pos_adim(1,:,:))-2
    elsewhere
        ! First particle is a left one
        send_min = floor(p_pos_adim(1,:,:))-2
    end where
    where (bl_type(N_proc(direction)/bl_size +1,:,:))
        ! Last particle is a centered one
        send_max = nint(p_pos_adim(N_proc(direction),:,:))+2
    elsewhere
        ! Last particle is a left one
        send_max = floor(p_pos_adim(N_proc(direction),:,:))+2
    end where
    ! Check results
    call test_check_success(send_min, good_send_min, success, cart_rank)
    call test_check_success(send_max, good_send_max, success, cart_rank)
    call test_substatus('remesh range - order 4 ', success, cart_rank)
    
    ! ----- M'6 -----
    order ='p_O2'
    ! Init solver
    call advec_init(order, verbosity=.false.)
    ! Test procedure
    call AC_remesh_range(bl_type, p_pos_adim, direction, send_min, send_max, send_gap, send_gap_abs)
    ! Compute theoric value
    send_min = nint(p_pos_adim(1,:,:))-2
    send_max = floor(p_pos_adim(N_proc(direction),:,:))+3
    ! Check results
    call test_check_success(send_min, good_send_min, success, cart_rank)
    call test_check_success(send_max, good_send_max, success, cart_rank)
    call test_substatus('remesh range - Mprime 6 ', success, cart_rank)
!    call AC_remesh_determine_communication(direction, gs, ind_group, rece_gap, send_gap, send_gap_abs, send_rank, cartography)


    ! Free memory
    deallocate(bl_type)
    deallocate(bl_tag)
    deallocate(p_pos_adim)
    deallocate(send_min)
    deallocate(send_max)
    deallocate(good_send_min)
    deallocate(good_send_max)
    deallocate(send_gap)

    success = .not. success

end function test_part_AC_remesh_range


!> Particle method: validation of computation of remeshing range for a group of line
!!     @return      error   = test error (= false if the code pass the test) (= not success)
function test_part_AC_remesh_determine_communication(shift) result(success)

    ! Library
    use mpi
    ! Scales code
    use advec
    use advec_common    ! Some porcedure common to advection along all directions
    use advec_variables ! contains info about solver parameters and others.
    use cart_topology
    ! Test procedures
    use test_common
    use advec_aux_init

    integer, intent(in), optional           :: shift
    logical                                 :: success

    integer                                 :: shift_bis    ! global translation of indices
    character(str_short)                    :: order        ! order of the particles solver
    
    logical, dimension(:,:,:), allocatable  :: bl_type      ! type of block
    logical, dimension(:,:,:), allocatable  :: bl_tag       ! tag of block
    integer                                 :: direction    ! direction (2 if along Y, 3=along Z)
    real(WP), dimension(:,:,:), allocatable :: p_pos_adim   ! particle position
    integer, dimension(:,:), allocatable    :: send_min     ! first mesh where particles are remeshed
    integer, dimension(:,:), allocatable    :: send_max     ! last mesh where particles are remeshed
    integer, dimension(:,:,:), allocatable  :: send_gap     ! distance between me and processus wich send me information
    integer, dimension(2)                   :: send_gap_abs ! min (resp max) value of rece_gap(:,:,i) with i=1 (resp 2)
    integer, dimension(2 , 2)               :: rece_gap     ! distance between me and processus to wich I send information
    integer, dimension(2)                   :: gs           ! group size
    integer, dimension(2)                   :: ind_group    ! group indice
    integer                                 :: max_size     ! maximal size of cartography(:,proc_gap)
    integer, dimension(:,:), allocatable    :: cartography  ! cartography(proc_gap) contains the set of the lines indice in the block to wich the 
                                                            ! current processus will send data during remeshing and for each of these lines the range 
                                                            ! of mesh points from where it requiers the velocity values.
    integer, dimension(:), allocatable      :: send_rank    ! rank of processus to wich I send information


    ! Initialisation of context (solver, ...)
    success = .true.
    direction = 2
    shift_bis = 0.0
    ind_group = (/1,1/)
    call set_group_size(5)
    gs = group_size(direction,:)
    if (present(shift)) then
        shift_bis = shift
    else
        shift_bis = 0
    end if


    ! Allocation
    allocate(bl_type(bl_nb(direction)+1,group_size(direction,1),group_size(direction,2)))
    allocate(bl_tag(bl_nb(direction),group_size(direction,1),group_size(direction,2)))
    allocate(p_pos_adim(N_proc(direction),group_size(direction,1),group_size(direction,2)))
    allocate(send_min(group_size(direction,1),group_size(direction,2)))
    allocate(send_max(group_size(direction,1),group_size(direction,2)))
    allocate(send_gap(group_size(direction,1),group_size(direction,2),2))

    ! ----- Initialisation -----
    bl_type = .true.
    bl_tag = .false.
    call pos_gp_init(direction, gs, shift_bis, p_pos_adim, bl_type)

    ! ===== Test method =====
    ! -- Init solver --
    order ='p_O2'
    call advec_init(order, verbosity=.false.)
    ! -- Compute range --
    call AC_remesh_range(bl_type, p_pos_adim, direction, send_min, send_max, send_gap, send_gap_abs)
    ! -- Allocation --
    max_size = 2 + gs(2)*(2+3*gs(1))
    allocate(cartography(max_size,send_gap_abs(1):send_gap_abs(2)))
    allocate(send_rank(send_gap_abs(1):send_gap_abs(2)))
    ! -- Test the target procedure: determine which processes communicate together --
    call AC_remesh_determine_communication(direction, gs, ind_group, rece_gap, send_gap, send_gap_abs, send_rank, cartography)

    ! -- Compute theoritical results

    ! -- Check results --
    !call test_check_success(send_min, good_send_min, success, cart_rank)
    !call test_check_success(send_max, good_send_max, success, cart_rank)
    !call test_substatus('remesh range - order 2 ', success, cart_rank)


    ! -- Free memory --
    deallocate(bl_type)
    deallocate(bl_tag)
    deallocate(p_pos_adim)
    deallocate(send_min)
    deallocate(send_max)
    deallocate(send_gap)
    deallocate(cartography)
    deallocate(send_rank)

    success = .not. success

end function test_part_AC_remesh_determine_communication

end module advec_aux_common
!> @}
