!> @addtogroup part_test
!! @{
    
!------------------------------------------------------------------------------
!
! MODULE: test_advection_init
!
! DESCRIPTION: 
!> Initialisation procedure for advection test (and solver based on particle method).
!!
!! @details
!! This module provide different initialisation setup in order to test the transport solver. 
!!
!! The following initialisation are included :
!!      1 -> Constant field for both scalar and velocity
!!      2 -> 2D-rotation of a sphere
!!      3 -> scalar(i,j,k) = i/Nx + 10* j/Ny + 100*k/Nz with periodic boundary condition
!!      4 -> velocity(i,j,k) = i/Nx + 10* j/Ny + 100*k/Nz with periodic boundary condition
!!
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

module advec_aux_init

    use string
    use precision
    implicit none

    ! ===== Initialisation for advection test =====
    ! Public function
    public              :: velo_init
    public              :: particle_velo_init
    public              :: scal_init
    public              :: scal_velo_init_part
    ! Private function
    private             :: compute_velo_tag

    ! ===== Setup parameter =====
    ! Public variables
    !> Period for rotation cases
    real(WP), public    :: period = 1
    ! Private variable
    !> Pi value
    real(WP), parameter   :: M_PI = ACOS(-1.0)




contains

!> Initialisation of the scalar field for the different tests.
!!    @param[in]    init        = parameter to initialise the scalar fied to a constant one or to a sphere shape
!!    @param[in]    translat    = optional argument, translat the field, in order to obtain analytic solution for non zero velocity in test case
!!                                  "translation field"
!!    @param[out]   scalar      = scalar field
!!    @param[out]   good_scal   = analytic solution
subroutine scal_init(init, scalar, good_scal, translat) 

    use test_common
    use cart_topology

    character(len=*), intent(in)                    :: init
    real(WP), dimension(:,:,:), intent(out)         :: scalar
    real(WP), dimension(:,:,:), intent(out)         :: good_scal
    real(WP), dimension(3), intent(in), optional    :: translat

    real(WP)                                        :: rr, rx, ry, rz, rayon
    real(WP)                                        :: dist
    integer                                         :: i, j, k


    select case(init)
        case('constant')
            call test_substatus('constant scal', cart_rank)
            scalar = 1. 
            good_scal = 1.
        case('2D_rot')
            call test_substatus('uncentred sphere', cart_rank)
            scalar = 0.
            rayon = (minval(N*d_sc)/10.0)**2
            do k = 1, N_proc(3)
                rz = (d_sc(3)*(k + coord(3)*N_proc(3) - 3.0*N(3)/5.0))**2
                do j = 1, N_proc(2)
                    ry = (d_sc(2)*(j + coord(2)*N_proc(2)- 3.0*N(2)/5.0))**2
                    do i = 1, N_proc(1)
                        rx = (d_sc(1)*(i - 3.0*N(1)/5.0))**2
                        rr = rx + ry + rz
                        if (rr < rayon) scalar(i,j,k) = (1 - rr/rayon)**4
                    end do
                end do
            end do
            good_scal = scalar
        case('translation_field')
            call test_substatus('translation field', cart_rank)
            if (present(translat)) then
                call test_substatus('translation on X', translat(1), cart_rank)
                call test_substatus('translation on Y', translat(2), cart_rank)
                call test_substatus('translation on Z', translat(3), cart_rank)
            else
                call test_substatus('velocity = zero ', cart_rank)
            end if
            do k = 1, N_proc(3)
                do j = 1, N_proc(2)
                    do i = 1, N_proc(1)
                        dist = (float(i-1+coord(1)*N_proc(1))/N(1)) 
                        scalar(i,j,k) = cos((dist-0.5)*2*M_PI)
                        dist = (float(j-1+coord(2)*N_proc(2))/N(2))
                        scalar(i,j,k) = scalar(i,j,k)*cos((dist-0.5)*2*M_PI)
                        dist = (float(k-1+coord(3)*N_proc(3))/N(3)) 
                        scalar(i,j,k) = scalar(i,j,k)*cos((dist-0.5)*2*M_PI)
                        if (present(translat)) then
                            ! Only periodic condition
                            dist = (i+translat(1)-1+coord(1)*N_proc(1))*d_sc(1) ! distance au bord x = 0
                            dist = modulo(dist, length(1)) ! on utilise la periodicite
                            ! In other case, as the velocity is a linear function of the position, the analytic velocity
                            ! is the same that the interpolate one
                            good_scal(i,j,k) = cos((dist-0.5)*2*M_PI)
                            dist = (j+translat(2)-1+coord(2)*N_proc(2))*d_sc(2)
                            dist = modulo(dist, length(2))
                            good_scal(i,j,k) = good_scal(i,j,k) * cos((dist-0.5)*2*M_PI)
                            dist = (k+translat(3)-1+coord(3)*N_proc(3))*d_sc(3)
                            dist = modulo(dist, length(3))
                            good_scal(i,j,k) = good_scal(i,j,k) * cos((dist-0.5)*2*M_PI)
                        else
                            good_scal(i,j,k) = scalar(i,j,k)
                        end if
                    end do
                end do
            end do
        case default
            scalar = 1.
            good_scal = 1.
    end select

end subroutine scal_init



!> Initialisation of the velocity field to test its interpolation
!!    @param[in]        init        = parameter to initialise the scalar fied to a constant one or to a sphere shape
!!    @param[out]       velo        = velocity field along one direction
!!    @param[in]        direction   = current direction (along X,Y or Z-axis)
!!    @param[in]        p_pos_adim  = adimensionned particles postion (location where the velocity will be interpolated)
!!    @param[in,out]    good_velo   = analytic interpolation of the velocity field (on location p_pos)
subroutine velo_init(init, velo, direction, p_pos_adim, good_velo)

    use test_common
    use cart_topology

    character(len=*), intent(in)                :: init
    real(WP), dimension(:,:,:), intent(inout)   :: velo
    real(WP), dimension(:), intent(in)          :: p_pos_adim
    integer, intent(in)                         :: direction
    real(WP), dimension(:,:,:), intent(inout)   :: good_velo

    real(WP)                                    :: dist         ! distance from original position for translation case
    integer                                     :: i, j, k      ! mesh indices
    integer                                     :: ierr         ! mpi error code


    select case(init)
        case('constant')
            velo = 1. 
            good_velo = 1.
        case('2D_rot')
            select case(direction)
                case(1)
                    do k = 1, N_proc(3)
                        do j = 1, N_proc(2)
                            do i = 1, N_proc(1)
                                velo(i,j,k)=(2*M_PI/period)*(length(2)/2.0-((j+coord(2)*N_proc(2))*d_sc(2)))
                                good_velo(i,j,k)=(2*M_PI/period)*(length(2)/2.0-((j+coord(2)*N_proc(2))*d_sc(2)))
                            end do
                        end do
                    end do
                case(2)
                    do k = 1, N_proc(3)
                        do j = 1, N_proc(2)
                            do i = 1, N_proc(1)
                                velo(i,j,k)=(2*M_PI/period)*(((i+coord(1)*N_proc(1))*d_sc(1))-length(1)/2.0)
                                good_velo(i,j,k)=(2*M_PI/period)*(((i+coord(1)*N_proc(1))*d_sc(1))-length(1)/2.0)
                            end do
                        end do
                    end do
                case(3)
                    do k = 1, N_proc(3)
                        do j = 1, N_proc(2)
                            do i = 1, N_proc(1)
                                velo(i,j,k)=0
                                good_velo(i,j,k)=0
                            end do
                        end do
                    end do
                case default
                    call test_substatus(' XXX error : wrong direction =', direction,  cart_rank)
                    stop
            end select
        case('translation_field')
            select case(direction)
                case(1)
                    do k = 1, N_proc(3)
                        do j = 1, N_proc(2)
                            do i = 1, N_proc(1)
                                velo(i,j,k) = (float(i-1)/N(1)) &
                                    & + 10*(float(j-1+coord(2)*N_proc(2))/N(2)) &
                                    & + 100*(float(k-1+coord(3)*N_proc(3))/N(3)) 
                                if (periods(1) .eqv. .true.) then
                                    dist = (p_pos_adim(i)-1+coord(1)*N_proc(1))*d_sc(1) ! distance au bord x = 0
                                    dist = modulo(dist, length(1)) ! on utilise la periodicite
                                    ! If dist belong to (length - dx, length) then with have to interpolate 
                                    ! the velocity between postion (length-dx) and 0 (due to periodicity boundary condition)
                                    if (dist>length(1)-d_sc(1)) dist = (length(1)-dist)*(length(1)-d_sc(1))/d_sc(1)
                                    ! In other case, as the velocity is a linear function of the position, the analytic velocity
                                    ! is the same that the interpolate one
                                    good_velo(i,j,k) = dist/length(1) &
                                        & + 10*(float(j-1+coord(2)*N_proc(2))/N(2)) &
                                        & + 100*(float(k-1+coord(3)*N_proc(3))/N(3)) 
                                else
                                    call test_substatus(' boundary along X condition not implemented ', cart_rank)
                                end if
                            end do
                        end do
                    end do
                case(2)
                    do k = 1, N_proc(3)
                        do j = 1, N_proc(2)
                            do i = 1, N_proc(1)
                                velo(i,j,k) = (float(i-1)/N(1)) &
                                    & + 10*(float(j-1+coord(2)*N_proc(2))/N(2)) &
                                    & + 100*(float(k-1+coord(3)*N_proc(3))/N(3)) 
                                if (periods(2) .eqv. .true.) then
                                    dist = (p_pos_adim(j)-1+coord(2)*N_proc(2))*d_sc(2)
                                    dist = modulo(dist, length(2))
                                    ! See direction 1 for explaination
                                    if (dist>length(2)-d_sc(2)) dist = (length(2)-dist)*(length(2)-d_sc(2))/d_sc(2)
                                    good_velo(i,j,k) = (float(i-1)/N(1)) &
                                        & + 10*(dist)/length(2) &
                                        & + 100*(float(k-1+coord(3)*N_proc(3))/N(3))
                                else
                                    call test_substatus(' boundary along X condition not implemented ', cart_rank)
                                end if
                            end do
                        end do
                    end do
                case(3)
                    do k = 1, N_proc(3)
                        do j = 1, N_proc(2)
                            do i = 1, N_proc(1)
                                velo(i,j,k) = (float(i-1)/N(1)) &
                                    & + 10*(float(j-1+coord(2)*N_proc(2))/N(2)) &
                                    & + 100*(float(k-1+coord(3)*N_proc(3))/N(3)) 
                                if (periods(3) .eqv. .true.) then
                                    dist = (p_pos_adim(k)-1+coord(3)*N_proc(3))*d_sc(3)
                                    dist = modulo(dist, length(3))
                                    ! See direction 1 for explaination
                                    if (dist>length(3)-d_sc(3)) dist = (length(3)-dist)*(length(3)-d_sc(3))/d_sc(3)
                                    good_velo(i,j,k) = (float(i-1)/N(1)) &
                                        & + 10*(float(j-1+coord(2)*N_proc(2))/N(2)) &
                                        & + 100*dist/length(3)
                                else
                                    call test_substatus(' boundary along X condition not implemented ', cart_rank)
                                end if
                            end do
                        end do
                    end do
                case default
                    call test_substatus(' XXX error : wrong direction =', direction,  cart_rank)
                    stop
                end select
        case default
            velo = 1.
            good_velo = 1.
    end select


end subroutine velo_init


!> Initialisation of particle velocity field to test its interpolation
!!    @param[in]        init        = parameter to initialise the scalar fied to a constant one or to a sphere shape
!!    @param[out]       velo        = initial particles velocity
!!    @param[in]        direction   = current direction (along X,Y or Z-axis)
!!    @param[in]        p_pos_adim  = adimensionned particles postion (location where the velocity will be interpolated)
!!    @param[in,out]    good_velo   = analytic interpolation of the velocity field (on location p_pos)
subroutine particle_velo_init(init, velo, direction, p_pos_adim, good_velo)

    use test_common
    use cart_topology

    character(len=*), intent(in)                :: init
    real(WP), dimension(:), intent(inout)       :: velo
    real(WP), dimension(:), intent(in)          :: p_pos_adim
    integer, intent(in)                         :: direction
    real(WP), dimension(:), intent(inout)       :: good_velo

    real(WP)                                    :: dist         ! distance from original position for translation case
    integer                                     :: ind          ! particle indice


    select case(init)
        case('constant')
            velo = 1. 
            good_velo = 1.
        case('translation_field')
            do ind = 1, size(velo)
                velo(ind) = (float(ind-1+coord(direction)*N_proc(direction))/N(direction))
                if (periods(direction) .eqv. .true.) then
                    dist = (p_pos_adim(ind)-1+coord(direction)*N_proc(direction))*d_sc(direction)
                    dist = modulo(dist, length(direction))
                    ! See direction 1 for explaination
                    if (dist>length(direction)-d_sc(direction)) &
                        & dist = (length(direction)-dist)*(length(direction)-d_sc(direction))/d_sc(direction)
                    good_velo(ind) = dist/length(direction)
                else
                    call test_substatus(' boundary condition not implemented ', cart_rank)
                end if
            end do
        case default
            velo = 1.
            good_velo = 1.
    end select

end subroutine particle_velo_init


!> Initialisation of paticles and velocity field to provide setups to test advection
!!    @param[in]        init        = parameter used to choose between the different setup
!!    @param[in,out]    dt          = time step (used to compute solution)
!!    @param[in]        shift       = global translation of indices (optional)
!!    @param[out]       scal3D      = scalar field
!!    @param[out]       velo        = velocity field along one direction
!!    @param[in]        direction   = current direction (along X,Y or Z-axis)
!!    @param[out]       good_scal   = analytic solution of the advection problem
subroutine scal_velo_init_part(init, dt, shift, scal3D, velo, direction, good_scal)

    ! Scales code
    use cart_topology
    use advec_common    ! Some porcedure common to advection along all directions
    use advec_variables ! contains info about solver parameters and others.
    ! Test procedures
    use test_common

    character(len=*), intent(in)                                    :: init
    real(WP), intent(inout)                                         :: dt
    integer, intent(in)                                             :: shift
    real(WP), dimension(N_proc(1),N_proc(2),N_proc(3)), intent(out) :: velo
    real(WP), dimension(N_proc(1),N_proc(2),N_proc(3)), intent(out) :: scal3D
    real(WP), dimension(N_proc(1),N_proc(2),N_proc(3)), intent(out) :: good_scal
    integer, intent(in)                                             :: direction

    integer                                     :: i, j, k      ! mesh indices
    real(WP)                                    :: cfl          ! ratio between time and space steps
    real(WP)                                    :: t            ! some time indication
    real(WP)                                    :: dt_bis       ! time step used to compute trajectories
                                                                ! with a numerical integration
    real(WP)                                    :: sX, sY, sZ   ! some temp variable
    real(WP), dimension(3)                      :: vect_dir     ! some temp variable
    

    cfl = dt/d_sc(direction)


    ! -- Initialize velocity and compute some trajectories --
    select case(init)
        case('left')
            !  -- Test case 1 : only left block, no tag --
            velo = (shift+0.3)/cfl
            scal3D = -dt*velo
            call test_substatus('particle- no tag, only left block', cart_rank)

        case('tag')
            ! Update dt in order to create "chock" without broke the stability condition
            dt = 0.8*d_sc(direction)/sqrt(2*1.2*bl_size)
            cfl = dt/d_sc(direction)
            ! Compute trajectories
            scal3D = 0.0
            do k = 1, N_proc(3)
                do j = 1, N_proc(2)
                    do i = 1, N_proc(1)
                        velo(i,j,k) = compute_velo_tag(dble(j), shift, cfl, direction)
                        dt_bis = min(0.01_WP, dt/10)
                        t = 0
                        do while(t <= dt-dt_bis)
                            scal3D(i,j,k) = scal3D(i,j,k) &
                                &- dt_bis*compute_velo_tag(j+scal3D(i,j,k), shift, cfl, direction)/d_sc(direction)
                            t = t + dt_bis
                        end do
                        dt_bis = dt - t
                        scal3D(i,j,k) = scal3D(i,j,k) &
                             &- dt_bis*compute_velo_tag(j+scal3D(i,j,k), shift, cfl, direction)/d_sc(direction)
                    end do
                end do
            end do
            call test_substatus('particle- 2 tag, left and center', cart_rank)

        case default ! default = 'center'
            ! Test case 2 : only center block, no tag
            velo = (shift+0.8)/cfl
            scal3D = -dt*velo
            call test_substatus('particle- no tag, only center block', cart_rank)
    end select

    ! -- Compute solution of the advection problem
    vect_dir = 0
    vect_dir(direction) = 1
    do k = 1, N_proc(3)
        sZ = (k-1+(coord(3)*N_proc(3)))*d_sc(3) 
        do j = 1, N_proc(2)
            sY = (j-1+(coord(2)*N_proc(2)))*d_sc(2)
            do i = 1, N_proc(1)
                sX = (i-1+(coord(1)*N_proc(1)))*d_sc(1)
                good_scal(i,j,k) = cos(2*M_PI*(sZ+vect_dir(3)*scal3D(i,j,k))/length(3))
                good_scal(i,j,k) = good_scal(i,j,k)*cos(2*M_PI*(sY+vect_dir(2)*scal3D(i,j,k))/length(2))
                !good_scal(i,j,k) = good_scal(i,j,k)*cos(3*2*M_PI*(sY+vect_dir(2)*scal3D(i,j,k))/length(2))
                good_scal(i,j,k) = good_scal(i,j,k)*cos(2*M_PI*(sX+vect_dir(1)*scal3D(i,j,k))/length(1))
                !good_scal(i,j,k) = good_scal(i,j,k)*cos(5*2*M_PI*(sX+vect_dir(1)*scal3D(i,j,k))/length(1))
                scal3D(i,j,k) = cos(2*M_PI*sZ/length(3))
                scal3D(i,j,k) = scal3D(i,j,k)*cos(2*M_PI*sY/length(2))
                !scal3D(i,j,k) = scal3D(i,j,k)*cos(3*2*M_PI*sY/length(2))
                scal3D(i,j,k) = scal3D(i,j,k)*cos(2*M_PI*sX/length(1))
                !scal3D(i,j,k) = scal3D(i,j,k)*cos(5*2*M_PI*sX/length(1))
            end do
        end do
    end do

end subroutine scal_velo_init_part

!> Compute a velocity field wich produce 2 tagged block during an advection step
!! with the solver based on particle method
!!    @param[in]    pos         = relative position along current direction
!!    @param[in]    shift       = shift  (in number of mesh)
!!    @param[in]    cfl         = time step/space step
!!    @param[in]    direction   = current direction
!!    @return       res         = velocity field
function compute_velo_tag(pos, shift, cfl, direction) result(res)

    ! Topology
    use cart_topology
    ! Solver information
    use advec_common    ! Some porcedure common to advection along all directions
    use advec_variables ! contains info about solver parameters and others.
    ! Test tools
    use test_common

    real(WP), intent(in)    :: pos, cfl
    integer, intent(in)     :: shift, direction
    real(WP)                :: res
    real(WP)                :: pos_abs      ! absolute position (i,j,k are relative position in the current processus)
    integer                 :: ind_tag_bl, ind_tag_p

    ind_tag_bl = (nb_proc_dim(direction)*(N(direction)/bl_size)/3)! - (bl_number(direction)*coord(direction))
    ind_tag_p = (ind_tag_bl-1)*bl_size + bl_size/2 + 1

    pos_abs = pos + coord(direction)*N_proc(direction)
    
    res = shift/cfl
    if ((pos_abs >= ind_tag_p).and.(pos_abs < N(direction))) then
        res = res + 0.8*(1.0-(pos_abs-ind_tag_p)/(N(direction)-ind_tag_p))/cfl
    end if

end function compute_velo_tag


!> Initialisation of particle position for a group
!!    @param[in]    direction   = current direction
!!    @param[in]    gs          = group_size
!!    @param[in]    shift       = shift  (in number of mesh)
!!    @param[out]   p_pos       = scalar field
!!    @param[out]   bl_type     = bloc type (left/center), optional
subroutine pos_gp_init(direction, gs, shift, p_pos, bl_type)

    ! Topology
    use cart_topology
    ! Solver information
    use advec_common    ! Some porcedure common to advection along all directions
    use advec_variables ! contains info about solver parameters and others.
    ! Test tools
    use test_common

    integer, intent(in)                                                             :: shift, direction
    integer, dimension(2), intent(in)                                               :: gs
    real(WP), dimension(N_proc(direction), gs(1), gs(2)), intent(out)               :: p_pos
    logical, dimension(bl_nb(direction)+1, gs(1), gs(2)), intent(inout), optional   :: bl_type

    integer     :: ind, ind2 ! some indice (line coordiante, particle indice)

    ! ===== Each line of the group is initialized considering it second indice =====
    ind2 = 1
    do while(ind2<=gs(2))
        ! Init are done with a cycle of five
        select case(modulo(ind2,5))
        case(1)
            do ind = 1, N_proc(direction)
                p_pos(ind,:,ind2) = ind - 2.1 + shift*N_proc(direction)
            end do
            if(present(bl_type)) bl_type(:,:,ind2) = .true.
        case(2)
            do ind = 1, N_proc(direction)
                p_pos(ind,:,ind2) = ind - 1.1 + shift*N_proc(direction)
            end do
            if(present(bl_type)) bl_type(:,:,ind2) = .true.
        case(3)
            do ind = 1, N_proc(direction)
                p_pos(ind,:,ind2) = ind + shift*N_proc(direction)
            end do
            if(present(bl_type)) bl_type(:,:,ind2) = .false.
        case(4)
            do ind = 1, N_proc(direction)
                p_pos(ind,:,ind2) = ind + 1.1 + shift*N_proc(direction)
            end do
            if(present(bl_type)) bl_type(:,:,ind2) = .false.
        case default
            do ind = 1, N_proc(direction)
                p_pos(ind,:,ind2) = ind + 2.1 + shift*N_proc(direction)
            end do
            if(present(bl_type)) bl_type(:,:,ind2) = .false.
        end select

        ind2 = ind2 +1
    end do

end subroutine pos_gp_init


end module advec_aux_init
!> @}
