!> @addtogroup part_test
!! @{
    
!------------------------------------------------------------------------------
!
! PROGRAM : advec_main
!
! DESCRIPTION: 
!> This program use the function implemented in the module advec_aux to
!! test the advection solver.
!!
!! @author
!! Jean-Baptiste Lagaert, LEGI
!
!> @details
!! All these test are unit test: they return a logical value to check if 
!! the code version pass it or not. 
!!
!! That is all these test are logical function, they return true if the result
!! is the right one and false otherwise. 
!! All the "test_part_*" function are devoted to validate the particular
!! solver
!!
!! The following test are included:
!!    - Validate the particular method, step by step
!!        - Test the procedure "AC_obtain_senders" from advec_common
!!        - Validate the redistribution the buffer during the remeshing
!!        - Validate the redistribution the buffer during the remeshing -
!!            debug version : only one processus contains non-zero field.
!!        - Validate the remeshing of untagged particles
!!        - Validate the velocity interpolation (the RK2 scheme use the
!!            velocity at midlle point to advec particles).
!!        - Validate how type (left or center) and tag are computing for a
!!            single line
!!        - Validate how type (left or center) and tag are computing for a
!!            group of line
!!    - Validate an advection solver (in advec_aux)
!!        - advec a ball with a constant velocity
!!        - advec a ball with a spheric velocity field (the ball turns)
!!        - Advection with radial shear. This test also validate the remeshing 
!!              of tagged particles.
!!
!
!------------------------------------------------------------------------------

program advec_main

    ! External Library
    use mpi
    ! Scales code
    use cart_topology
    ! Test procedures
    use test_common
    use advec_aux
    use advec_aux_common
    use advec_aux_common_line

    implicit none

    logical     :: error = .false.  ! logical error
    integer     :: ierr             ! mpi error code
    integer     :: rank_world       ! processus rank on "MPI_COMM_WORLD"
    integer     :: nb_proc,nb_procZ ! number of processus
    integer     :: i                ! some boucle indice
    integer     :: direction        ! along X (1), Y (2) or Z (3)
   
    ! ===== Initialisation =====

    ! Set the verbosity
    verbose_test = .true.
    verbose_more = .false.
    ! Initialise mpi
    call mpi_init(ierr)    
    call mpi_comm_rank(MPI_COMM_WORLD, rank_world, ierr)
    call mpi_comm_size(MPI_COMM_WORLD, nb_proc, ierr)

    ! Cut the domain along Y and initialize the toppology
    nb_procZ = 1
    if ((mod(nb_proc,5)==0).and.(mod(default_size, nb_proc/5)==0)) then
        nb_procZ = 5
        nb_proc = nb_proc/5
    else if ((mod(nb_proc,2)==0).and.(mod(default_size, nb_proc/2)==0)) then
        nb_procZ = 2
        nb_proc = nb_proc/2
    else
        if (mod(default_size, nb_proc)/=0) stop 'wrong number of processes : it have to divide 100'
    end if
    call cart_create((/ nb_proc, nb_procZ /), ierr)
    call discretisation_default()
    call mpi_barrier(MPI_COMM_WORLD, ierr)

    ! ===== Test about procedures involved in remeshing process =====
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_title('particle method - remeshing', rank_world)
    
    ! Does it well compute who communicate with who during remeshing ?
    error = test_part_AC_obtain_senders()
    call test_status(error, 'obtain_senders sans shift', rank_world)
    error = test_part_AC_obtain_senders(1)
    call test_status(error, 'obtain_senders avec shift', rank_world)
    error = test_part_AC_obtain_senders(3*nb_proc_dim(2)+2)
    call test_status(error, 'obtain_senders avec large shift', rank_world)
    
    ! The remeshing are done in a buffer. Is it well re-distribuate on processes ?
    do i = 0, min(4, nb_proc-1)
        error = test_part_AC_bufferToScalar_Deb(i)
        call test_status(error, 'bufferToScalar deb, no shift, rank =', i, rank_world)
        error = test_part_AC_bufferToScalar_Deb(i,1)
        call test_status(error, 'bufferToScalar deb,shift; rank =', i, rank_world)
    end do
    error = test_part_AC_bufferToScalar()
    call test_status(error, 'bufferToScalar sans shift', rank_world)
    error = test_part_AC_bufferToScalar(1)
    call test_status(error, 'bufferToScalar avec shift', rank_world)
    error = test_part_AC_bufferToScalar(3*nb_proc_dim(2)+2)
    call test_status(error, 'bufferToScalar avec large shift', rank_world)

    ! Test remeshing of untagged particles
    error = test_part_remesh_no_tag()
    call test_status(error, 'remeshing of cst field advected a cst v', rank_world)
    error = test_part_remesh_no_tag('translation_field')
    call test_status(error, 'remeshing of uncst field advected a cst v', rank_world)
    error = test_part_remesh_no_tag(order_opt='p_O4')
    call test_status(error, 'O4: remeshing of cst field advected a cst v', rank_world)
    error = test_part_remesh_no_tag('translation_field', order_opt='p_O4')
    call test_status(error, 'O4: remeshing of uncst field advected a cst v', rank_world)
    error = test_part_remesh_no_tag(order_opt='p_M6')
    call test_status(error, 'Mprime 6: remeshing of cst field advected a cst v', rank_world)
    error = test_part_remesh_no_tag('translation_field', order_opt='p_M6')
    call test_status(error, 'Mprime 6: remeshing of uncst field advected a cst v', rank_world)


    ! ===== Test about procedures involved in computation of particles advection =====
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_title('particle: velocity interpolation', rank_world)

    ! Test auxiliary procedures
    error = test_part_AC_obtain_recevers()
    call test_status(error, 'AC_obtain_recevers', rank_world)

    ! Test velocity interpolation
    error = test_part_AC_interpol_velocity(2, 0)
    call test_status(error, 'velocity interpolation, no shift', rank_world)
    error = test_part_AC_interpol_velocity(2, 1)
    call test_status(error, 'velocity interpolation, shift', rank_world)

    ! Test group variants
    call test_title('particle: velocity interpolation (group)', rank_world)
    call set_group_size(5)
    error = test_part_AC_velo_determine_com(0)
    call test_status(error, 'velocity: communication', rank_world)
    error = test_part_AC_velo_determine_com(1)
    call test_status(error, 'velocity: communication, shift', rank_world)
    do direction = 1, 3
        error = test_part_AC_velo_compute_group(direction,0)
        call test_status(error, 'velocity: interpol', direction, rank_world)
    end do

    ! ===== Test others =====
    call test_title('particle method - tag and bloc type', rank_world)
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    error = test_part_AC_type_and_block_O2()
    call test_status(error, 'determine block type and tag particles', rank_world)
    error = test_part_AC_type_and_block_O2_group()
    call test_status(error, 'block and type for a group of line', rank_world)

   ! ===== Test solvers =====

    ! Test devoted to solver based on particles method
    call test_title('particle method - advection test', rank_world)
    error = test_part_advec_1D()
    call test_status(error, 'advection - center block, no tag', rank_world)
    error = test_part_advec_1D('left')
    call test_status(error, 'advection - left block, no tag', rank_world)
    verbose_more = .true.
    error = test_part_advec_3D()
    call test_status(error, 'advection - 3D tests', rank_world)
    verbose_more = .false.

    ! Idem for order 4 (corrected lambda 4)
    call test_title('particle method - advection test', rank_world)
    error = test_part_advec_1D(order_opt='p_O4')
    call test_status(error, 'advection - center block, no tag', rank_world)
    error = test_part_advec_1D('left', order_opt='p_O4')
    call test_status(error, 'advection - left block, no tag', rank_world)
    verbose_more = .true.
    error = test_part_advec_3D(order='p_O4')
    call test_status(error, 'advection - 3D tests', rank_world)
    verbose_more = .false.

    ! Idem for order M'6 remeshing formula
    call test_title('particle method - advection test', rank_world)
    error = test_part_advec_1D(order_opt='p_M6')
    call test_status(error, 'advection - center block, no tag', rank_world)
    error = test_part_advec_1D('left', order_opt='p_M6')
    call test_status(error, 'advection - left block, no tag', rank_world)
    verbose_more = .true.
    error = test_part_advec_3D(order='p_M6')
    call test_status(error, 'advection - 3D tests', rank_world)
    verbose_more = .false.

    ! Generic test
!    call mpi_barrier(MPI_COMM_WORLD, ierr)
!    call test_title('generic advection test', rank_world)
!    error = test_advec_rot()
!    call test_status(error, 'turning sphere', rank_world)


    call mpi_finalize(ierr)

end program advec_main

!> @}
