!------------------------------------------------------------------------------
!
! MODULE: topo_aux
!
! DESCRIPTION: 
!> This module provides different tests to validate the topology and the
!! interface with the different data structures.
!!
!! @details
!! Different automatic test are developped in order to check the mesh creation
!! and the interface between the two data structures (the one used for the
!! particular method and the one from the spectral part).
!! All these test are unit test : they return a logical value to check if 
!! the ierr version pass it or not. 
!!
!! That is all these test are logical function, they return true if the result
!! is the right one and false otherwise. 
!!
!! The following test are included :
!!      1 -> Initialise the topology, check the number of processes and
!!              the communicators.
!!      2 -> Check the periodicity.
!!      3 -> Check if the subgrid on each processus have the good size
!           
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

module topo_aux

    use mpi
    use cart_topology
    use string
    use precision
    implicit none

    real(WP), private   :: epsilon_success = 1e-4     ! Error tolerance


    ! Public procedures

    ! ===== Test the topology =====
    ! Public function
    public              :: test_topo_init
    public              :: test_topo_perio
    public              :: test_topo_submesh




contains

!> Test the topology initialisation
!!    @return success   = logical success (= false if the ierr pass the test)
!! @details
!!    Test the cartesian topology : check the number of processes and the
!!    communicators.
function test_topo_init() result(success)

    use test_common

    logical     :: success          ! success status

    integer     :: ierr             ! mpi success ierr
    integer     :: nb_proc          ! total number of processus
    integer     :: nb_Y, nb_Z       ! actual number of processus in each direction
    integer, dimension(2)   :: dims ! wanted number of processus in Y and Z direction

    success = .true.
    call mpi_comm_size(MPI_COMM_WORLD, nb_proc, ierr)

    ! Cut the domain along Y and initialize the toppology
    dims = (/ nb_proc, 1 /)
    call cart_create(dims, ierr)

    ! Check the number of process in each communicator
    call mpi_comm_size(Y_comm, nb_Y, ierr)
    if (nb_Y /= nb_proc) then
        call test_substatus('number of processes in Y_comm', nb_Y, myrank)
        call test_substatus('and it must be', nb_proc, myrank)
        success = .false.
    end if
    call mpi_barrier(MPI_COMM_WORLD, ierr)

    call mpi_comm_size(Z_comm, nb_Z, ierr)
    if (nb_Z /= 1) then
        call test_substatus('number of processes in Z_comm', nb_Z, myrank)
        call test_substatus('and it must be', 1, myrank)
        success = .false.
    end if
    call mpi_barrier(MPI_COMM_WORLD, ierr)

    ! Compare it with the one saved in cart_topo
    if (nb_Y /= nb_proc_dim(2)) then
        call test_substatus('number of processes in Y_comm', nb_Y, myrank)
        call test_substatus('and the solver beleave it is', nb_proc_dim(2), myrank)
        success = .false.
    end if
    if (nb_Z /= nb_proc_dim(3)) then
        call test_substatus('number of processes in Y_comm', nb_Z, myrank)
        call test_substatus('and the solver beleave it is', nb_proc_dim(3), myrank)
        success = .false.
    end if

    ! Return error = not succes
    success = .not. success

end function test_topo_init

!> Check if it provide the right cartesian structure with a good periodicity
!!    @return success   = logical success (= false if the ierr pass the test)
function test_topo_perio() result(success)

    use test_common

    logical     :: success            ! success status
    
    integer     :: ierr                     ! mpi success ierr
    integer     :: rankP, rankN             ! rank of previous and next (for shift)
    integer     :: nb_Y, nb_Z               ! number of processus in each direction
    integer, dimension(2)   :: dims         ! number of processus in Y and Z direction
    integer, dimension(3)   :: coord_bis    ! coordonate of another processus
    integer                 :: new_coord    ! theoritical coordinate

    success = .true.
    
    ! Get the size
    call mpi_comm_size(Y_comm, nb_Y, ierr)
    call mpi_comm_size(Z_comm, nb_Z, ierr)

    ! Shift along Y
    ! Positive shift
    call mpi_cart_shift(cart_comm, 2-1, 1, rankP, rankN, ierr)
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call mpi_cart_coords(cart_comm, rankP, 3, coord_bis, ierr)
    new_coord = modulo(coord(2)-1, nb_Y)
    if ((coord_bis(2) /=(new_coord)).OR.(coord_bis(3)/=coord(3)) ) then
        call test_substatus('wrong Y-1 on rank', myrank, printer)
        call test_substatus('theoritical Y-1', new_coord, printer)
        call test_substatus('computed Y-1', coord_bis(2), printer)
call test_substatus('X', coord_bis(1), printer)
call test_substatus('Z', coord_bis(3), printer)
        success = .false.
    end if
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call mpi_cart_coords(cart_comm, rankN, 3, coord_bis, ierr)
    new_coord = modulo(coord(2)+1, nb_Y)
    if ((coord_bis(2) /=(new_coord)).OR.(coord_bis(3)/=coord(3)) ) then
        call test_substatus('wrong Y+1 on rank', myrank, printer)
        success = .false.
    end if
    ! Negative shift
    call mpi_cart_shift(cart_comm, 2-1, -1, rankP, rankN, ierr)
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call mpi_cart_coords(cart_comm, rankN, 3, coord_bis, ierr)
    new_coord = modulo(coord(2)-1, nb_Y)
    if ((coord_bis(2) /=(new_coord)).OR.(coord_bis(3)/=coord(3)) ) then
        call test_substatus('wrong Y+(-1) on rank', myrank, printer)
        success = .false.
    end if
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call mpi_cart_coords(cart_comm, rankP, 3, coord_bis, ierr)
    new_coord = modulo(coord(2)+1, nb_Y)
    if ((coord_bis(2) /=(new_coord)).OR.(coord_bis(3)/=coord(3)) ) then
        call test_substatus('wrong Y-(-1) on rank', myrank, printer)
        success = .false.
    end if
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_substatus('topo and periodicity along Y', success, myrank)
    
    ! Shift along Z
    ! Positive shift
    call mpi_cart_shift(cart_comm, 3-1, 1, rankP, rankN, ierr)
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call mpi_cart_coords(cart_comm, rankP, 3, coord_bis, ierr)
    new_coord = modulo(coord(3)-1, nb_Z)
    if ((coord_bis(3) /=(new_coord)).OR.(coord_bis(2)/=coord(2)) ) then
        call test_substatus('wrong Z-1 on rank', myrank, printer)
        success = .false.
    end if
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call mpi_cart_coords(cart_comm, rankN, 3, coord_bis, ierr)
    new_coord = modulo(coord(3)+1, nb_Z)
    if ((coord_bis(3) /=(new_coord)).OR.(coord_bis(2)/=coord(2)) ) then
        call test_substatus('wrong Z+1 on rank', myrank, printer)
        success = .false.
    end if
    ! Negative shift
    call mpi_cart_shift(cart_comm, 3-1, -1, rankP, rankN, ierr)
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call mpi_cart_coords(cart_comm, rankN, 3, coord_bis, ierr)
    new_coord = modulo(coord(3)-1, nb_Z)
    if ((coord_bis(3) /=(new_coord)).OR.(coord_bis(2)/=coord(2)) ) then
        call test_substatus('wrong Z+(-1) on rank', myrank, printer)
        success = .false.
    end if
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call mpi_cart_coords(cart_comm, rankP, 3, coord_bis, ierr)
    new_coord = modulo(coord(3)+1, nb_Z)
    if ((coord_bis(3) /=(new_coord)).OR.(coord_bis(2)/=coord(2)) ) then
        call test_substatus('wrong Z-(-1)) on rank', myrank, printer)
        success = .false.
    end if
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_substatus('topo and periodicity along Z', success, myrank)

    ! Big shift
    call mpi_cart_coords(cart_comm, myrank, 3, coord, ierr)
    call mpi_cart_shift(cart_comm, 2-1, 1+2*Nb_Y, rankP, rankN, ierr)
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call mpi_cart_coords(cart_comm, rankP, 3, coord_bis, ierr)
    new_coord = modulo(coord(2)-1, nb_Y)
    if ((coord_bis(2) /=(new_coord)).OR.(coord_bis(3)/=coord(3)) ) then
        call test_substatus('wrong Y- on rank', myrank, printer)
        call test_substatus('theoritical Y-', new_coord, printer)
        call test_substatus('computed Y-', coord_bis(2), printer)
        success = .false.
    end if
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call mpi_cart_coords(cart_comm, rankN, 3, coord_bis, ierr)
    new_coord = modulo(coord(2)+1, nb_Y)
    if ((coord_bis(2) /=(new_coord)).OR.(coord_bis(3)/=coord(3)) ) then
        call test_substatus('wrong Y+ on rank', myrank, printer)
        success = .false.
    end if
    call mpi_barrier(MPI_COMM_WORLD, ierr)
    call test_substatus('huge shift along Y', success, myrank)

    ! Return error = not success
    success = .not.success

end function test_topo_perio


!> Test the construction of subdomain and the mesh size in each processus
!!    @return success   = logical success (= false if the ierr pass the test)
!! @details
!!    Check if the subgrid on each processus have the good size
function test_topo_submesh() result(success)

    use test_common

    logical :: success
    integer :: ierr ! mpi success code

    success = .true.

    call discretisation_default()
    
    ! Check the number of mesh
    if (N_proc(1)/= 100) then
        call test_substatus('local number of mesh along X', N_proc(1), myrank)
        success = .false.
    end if
    if (N_proc(2)/= 100/nb_proc_dim(2)) then
        call test_substatus('local number of mesh along Y', N_proc(2), myrank)
        success = .false.
    end if
    if (N_proc(3)/= 100/nb_proc_dim(3)) then
        call test_substatus('local number of mesh along Z', N_proc(3), myrank)
        success = .false.
    end if

    ! Return error = not success
    success = .not.success

    call mpi_barrier(MPI_COMM_WORLD, ierr)

end function test_topo_submesh


end module topo_aux
