!------------------------------------------------------------------------------
!
! MODULE: test_advection
!
! DESCRIPTION: 
!> This module provide different tools useful to perform test.
!
!> @author
!! Jean-Baptiste Lagaert, LEGI
!
!------------------------------------------------------------------------------

module test_common

    use string
    use precision
    
    implicit none
    
    ! ===== Public variables =====
    !> To print some status message during the test
    logical :: verbose_test = .true.
    !> More verbosity !
    logical :: verbose_more = .true.
    !> To choose wich processes lead the screen output
    integer :: printer = 0

    ! ===== Public procedure =====
    ! - To print some information about the test (verbosity case)
    public  :: test_title
    public  :: test_status
    public  :: mpi_test_substatus
    public  :: test_substatus
    public  :: test_check_success

    ! ===== Private procedure =====
    private :: test_status_M
    private :: test_status_MI
    private :: test_substatus_M
    private :: test_substatus_MI
    private :: test_substatus_MR
    private :: test_substatus_ML
    private :: test_substatus_M3I
    private :: test_check_success_S
    private :: test_check_success_F
    private :: test_check_success_F2
    private :: test_check_success_F3
    private :: test_check_success_FI
    private :: test_check_success_FL
    private :: test_check_success_F3L

    
    ! ===== Private variables =====
    !> Error tolerance
    real(WP), private   :: epsilon_success = 1e-4 

    ! ===== Interface =====
    interface test_status
        module procedure test_status_M, test_status_MI
    end interface test_status

    interface test_substatus
        module procedure test_substatus_M, test_substatus_MI, test_substatus_MR &
            & , test_substatus_ML, test_substatus_M3I
    end interface test_substatus

    interface test_check_success
        module procedure test_check_success_S, test_check_success_F, &
            & test_check_success_F2, test_check_success_F3, test_check_success_FI, &
            & test_check_success_F2I, test_check_success_FL, test_check_success_F3L
    end interface test_check_success



contains

!> Diffuse the error status and print the test status
!!  @param[in]      message = information message
!!  @param[in]      rank    = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_title(message, rank)

    use cart_topology

    character(len =*), intent(in)   :: message
    integer, intent(in)             :: rank

    character(len=40)               :: mess_bis         ! message copy


    if((verbose_test).and.(rank==printer)) then
        mess_bis = message
        write(*,'(A1,1X,A40)')'#', mess_bis
        if((verbose_more).and.(rank==printer)) print*,''
    end if

end subroutine test_title


!> Diffuse the error status and print the test status
!!  @param[in, out] error   = logical equal true if there is an error
!!  @param[in]      message = information message
!!  @param[in]      rank    = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_status_M(error, message, rank)

    use mpi
    use cart_topology

    logical, intent(inout)             :: error
    character(len =*), intent(in)   :: message
    integer, intent(in)             :: rank

    character(len=40)               :: mess_bis         ! message copy
    integer                         :: error_int = 0
    integer                         :: error_red = 0
    integer                         :: ierr             ! mpi error code

    if(error .eqv. .true.) error_int = 1
    call mpi_allreduce(error_int, error_red, 1, MPI_INTEGER, MPI_MAX, MPI_COMM_WORLD, ierr)
    if(error_red==1) error=.true.

    if((verbose_test).and.(rank==printer)) then
        mess_bis = message
        write(*,'(5X,A2,2X,A40,X,A2,L2)')'->', mess_bis, '=', .not.error
        if((verbose_more).and.(rank==printer)) print*,''
    end if

end subroutine test_status_M


!> Diffuse the error status and print the test status
!!  @param[in, out] error           = logical equal true if there is an error
!!  @param[in]      message         = information message
!!  @param[in]      message_int     = integer added to the information message
!!  @param[in]      rank            = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_status_MI(error, message, message_int , rank)

    use mpi
    use cart_topology

    logical, intent(inout)          :: error
    character(len =*), intent(in)   :: message
    integer, intent(in)             :: rank, message_int

    character(len=37)               :: mess_bis         ! message copy
    integer                         :: error_int = 0
    integer                         :: error_red = 0
    integer                         :: ierr             ! mpi error code

    if(error .eqv. .true.) error_int = 1
    call mpi_allreduce(error_int, error_red, 1, MPI_INTEGER, MPI_MAX, MPI_COMM_WORLD, ierr)
    if(error_red==1) error=.true.

    if((verbose_test).and.(rank==printer)) then
        mess_bis = message
        write(*,'(5X,A2,2X,A37,X,I2,X,A2,L2)')'->', mess_bis, message_int, '=', .not.error
        if((verbose_more).and.(rank==printer)) print*,''
    end if

end subroutine test_status_MI


!> Use a mpi error code to update the test status and print it
!!  @param[in]  ierr    = mpi error code
!!  @param[in]  error   = logical equal true if there is an error
!!  @param[in]  message = information message
!!  @param[in]  rank    = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine mpi_test_substatus(ierr, error, message, rank)

    use mpi

    integer, intent(in)             :: ierr
    logical, intent(inout)          :: error
    character(len =*), intent(in)   :: message
    integer, intent(in)             :: rank

    if (ierr /= MPI_SUCCESS) then
        error = .false.
    end if

    call test_substatus(message, error, rank)
    error = .not. error

end subroutine mpi_test_substatus

!> Print a sub-status message
!!  @param[in]  message = information message
!!  @param[in]  rank    = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_substatus_M(message, rank)

    character(len =*), intent(in)  :: message
    integer, intent(in)             :: rank

    if((verbose_more).and.(rank==printer)) then
        write(*,'(10X,A2,2X,A40)')'+', message
    end if

end subroutine test_substatus_M


!> Print a sub-status message and a integer
!!  @param[in]  message = information message
!!  @param[in]  i       = integer to print
!!  @param[in]  rank    = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_substatus_MI(message, i, rank)

    character(len =*), intent(in)  :: message
    integer, intent(in)             :: i
    integer, intent(in)             :: rank

    if((verbose_more).and.(rank==printer)) then
        write(*,'(10X,A2,2X,A40,X,A1,X,I5)')'+', message, '=', i
    end if

end subroutine test_substatus_MI


!> Print a sub-status message and a integer
!!  @param[in]  message = information message
!!  @param[in]  i       = integer table of dimension 3 to print
!!  @param[in]  rank    = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_substatus_M3I(message, i, rank)

    character(len =*), intent(in)  :: message
    integer, dimension(3), intent(in)             :: i
    integer, intent(in)             :: rank

    if((verbose_more).and.(rank==printer)) then
        write(*,'(10X,A2,2X,A40,X,A1,X,I3,X,A1,X,I3,X,A1,X,I3)')'+',message,'=',i(1),',',i(2),',',i(3)
    end if

end subroutine test_substatus_M3I


!> Print a sub-status message and a real
!!  @param[in]  message = information message
!!  @param[in]  r       = real to print
!!  @param[in]  rank    = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_substatus_MR(message, r, rank)

    use precision

    character(len=* ), intent(in)   :: message
    real(WP), intent(in)            :: r
    integer, intent(in)             :: rank

    if((verbose_more).and.(rank==printer)) then
        write(*,'(10X,A2,2X,A40,X,A1,X,F8.5)')'+', message, '=', r
    end if

end subroutine test_substatus_MR


!> Print a sub-status message and a logical (after sending its value if false)
!!  @param[in]      message = information message
!!  @param[in,out]  l       = logical to print
!!  @param[in]      rank    = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_substatus_ML(message, l, rank)

    use precision
    use mpi
    use cart_topology

    character(len =*), intent(in)   :: message
    logical, intent(inout)          :: l
    integer, intent(in)             :: rank
    integer                         :: error_int = 0
    integer                         :: error_red = 0
    integer                         :: ierr             ! mpi error code

    if(l .eqv. .false.) error_int = 1
    call mpi_allreduce(error_int, error_red, 1, MPI_INTEGER, MPI_MAX, MPI_COMM_WORLD, ierr)
    if(error_red==1) l=.false.

    if((verbose_more).and.(rank==printer)) then
        write(*,'(10X,A2,2X,A40,X,A1,X,L5)')'+', message, '=', l
    end if

end subroutine test_substatus_ML



!> Check if the numerical success stay under a threshold - constant theoritical
!! solution
!!     @param[in,out]   success     = test success (= false if the code pass the test)
!!     @param[in]       scal1D      = numerical value of the scalar (1D)
!!     @param[in]       good_scal   = theoritical value of the scalar
!!     @param[in]       rank        = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_check_success_S(scal1D, good_scal, success, rank)

    use precision
    use cart_topology

    real(WP), intent(in)                        :: good_scal        ! theoritical value of scal1D
    real(WP), dimension(:),intent(in)           :: scal1D           ! the computed scalar field
    logical, intent(inout)                      :: success
    integer, intent(in)                         :: rank

    integer                                     :: success_inf        ! norm L_inf of the success

    success_inf = maxval(scal1D - good_scal)
    if (success_inf>=epsilon_success) then 
        success = .false.
        call test_substatus('XXX error', rank)
        call test_substatus('max scal0D', maxval(scal1D), rank)
        call test_substatus('min scal0D', minval(scal1D), rank)
        call test_substatus('and it must be', good_scal, rank)
    end if

end subroutine test_check_success_S


!> Check if two integer 1-dimensionnal table are equal.
!!     @param[in,out]   success     = test success (= false if the code pass the test)
!!     @param[in]       scal1D      = numerical value of the scalar (1D)
!!     @param[in]       good_scal   = theoritical value of the scalar
!!     @param[in]       rank        = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_check_success_FI(scal1D, good_scal, success, rank)

    use precision
    use cart_topology

    integer, dimension(:),intent(in)            :: good_scal        ! theoritical value of scal1D
    integer, dimension(:),intent(in)            :: scal1D           ! the computed scalar field
    logical, intent(inout)                      :: success
    integer, intent(in)                         :: rank

    integer                                     :: success_inf        ! norm L_inf of the success

    success_inf = maxval(abs(scal1D - good_scal))

    if (success_inf>=epsilon_success) then 
        success = .false.
        call test_substatus('XXX error', rank)
        call test_substatus('max scal1D', maxval(scal1D), rank)
        call test_substatus('min scal1D', minval(scal1D), rank)
        call test_substatus('max solution', maxval(good_scal), rank)
    end if

end subroutine test_check_success_FI


!> Check if two integer 2-dimensionnal table are equal.
!!     @param[in,out]   success     = test success (= false if the code pass the test)
!!     @param[in]       scal1D      = numerical value of the scalar (1D)
!!     @param[in]       good_scal   = theoritical value of the scalar
!!     @param[in]       rank        = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_check_success_F2I(array2D, good_array, success, rank)

    use precision
    use cart_topology

    integer, dimension(:,:), intent(in)         :: good_array       ! theoritical value of array2D
    integer, dimension(:,:), intent(in)         :: array2D          ! the computed value
    logical, intent(inout)                      :: success
    integer, intent(in)                         :: rank

    integer                                     :: success_inf        ! norm L_inf of the success

    success_inf = maxval(abs(array2D - good_array))

    if (success_inf>=0.5) then ! As we consider integer, this error is enough
        success = .false.
        call test_substatus('XXX error', rank)
        call test_substatus('max array2D', maxval(array2D), rank)
        call test_substatus('min array2D', minval(array2D), rank)
        call test_substatus('max solution', maxval(good_array), rank)
    end if

end subroutine test_check_success_F2I


!> Check if the numerical success stay under a threshold - 1D space-dependant analytic solution
!!     @param[in,out]   success     = test success (= false if the code pass the test)
!!     @param[in]       scal1D      = numerical value of the scalar (1D)
!!     @param[in]       good_scal   = theoritical value of the scalar
!!     @param[in]       rank        = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_check_success_F(scal1D, good_scal, success, rank)

    use precision
    use cart_topology

    real(WP), dimension(:),intent(in)           :: good_scal        ! theoritical value of scal1D
    real(WP), dimension(:),intent(in)           :: scal1D           ! the computed scalar field
    logical, intent(inout)                      :: success
    integer, intent(in)                         :: rank

    real(WP)                                    :: success_inf        ! norm L_inf of the success

    success_inf = maxval(abs(scal1D - good_scal))

    if (success_inf>=epsilon_success) then 
        success = .false.
        call test_substatus('XXX error', rank)
        call test_substatus('max scal1D', maxval(scal1D), rank)
        call test_substatus('min scal1D', minval(scal1D), rank)
        call test_substatus('max solution', maxval(good_scal), rank)
        call test_substatus('min solution', minval(good_scal), rank)
    end if

end subroutine test_check_success_F


!> Check if the numerical success stay under a threshold - 2D space-dependant analytic solution
!!     @param[in,out]   success     = test success (= false if the code pass the test)
!!     @param[in]       scal2D      = numerical value of the scalar (2D)
!!     @param[in]       good_scal   = theoritical value of the scalar
!!     @param[in]       rank        = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_check_success_F2(scal2D, good_scal, success, rank)

    use precision
    use cart_topology

    real(WP), dimension(:,:),intent(in)         :: good_scal        ! theoritical value of scal1D
    real(WP), dimension(:,:),intent(in)         :: scal2D           ! the computed scalar field
    logical, intent(inout)                      :: success
    integer, intent(in)                         :: rank

    real(WP)                                    :: success_inf        ! norm L_inf of the success

    success_inf = maxval(abs(scal2D - good_scal))
    if (success_inf>=epsilon_success) then 
        success = .false.
        call test_substatus('XXX error', rank)
        call test_substatus('max scal2D', maxval(scal2D), rank)
        call test_substatus('min scal2D', minval(scal2D), rank)
        call test_substatus('max solution', maxval(good_scal), rank)
        call test_substatus('min solution', minval(good_scal), rank)
    end if

end subroutine test_check_success_F2


!> Check if the numerical success stay under a threshold - 3D space-dependant analytic solution
!!     @param[in,out]   success     = test success (= false if the code pass the test)
!!     @param[in]       scal3D      = numerical value of the scalar (3D)
!!     @param[in]       good_scal   = theoritical value of the scalar
!!     @param[in]       rank        = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_check_success_F3(scal3D, good_scal, success, rank)

    use precision
    use cart_topology
    use mpi

    real(WP), dimension(:,:,:),intent(in)   :: good_scal        ! theoritical value of scal1D
    real(WP), dimension(:,:,:),intent(in)   :: scal3D           ! the computed scalar field
    logical, intent(inout)                  :: success
    integer, intent(in)                     :: rank

    real(WP)                                :: success_inf      ! norm L_inf of the success
    real(WP)                                :: success_inf_gl   ! norm L_inf of the success
    integer                                 :: ierr             ! mpi error code

integer, dimension(3) :: temp

    success_inf = maxval(abs(scal3D - good_scal))
    success_inf_gl=success_inf
    call mpi_reduce(success_inf, success_inf_gl, 1, MPI_DOUBLE_PRECISION, MPI_MAX, 0, MPI_COMM_WORLD, ierr)
    if (success_inf_gl>=epsilon_success) success = .false.
    call test_substatus('norm inf of error', success_inf_gl, rank)

    if (success_inf>=epsilon_success) then 
        call test_substatus('XXX error', rank)
        temp = minloc(scal3D - good_scal)
        call test_substatus('error min in', temp, rank)
        call test_substatus('scal3D', scal3D(temp(1), temp(2), temp(3)), rank)
        call test_substatus('sol', good_scal(temp(1), temp(2), temp(3)), rank)
        temp = maxloc(scal3D - good_scal)
        call test_substatus('error max in', temp, rank)
        call test_substatus('scal3D', scal3D(temp(1), temp(2), temp(3)), rank)
        call test_substatus('sol', good_scal(temp(1), temp(2), temp(3)), rank)
    end if

end subroutine test_check_success_F3


!> Check if two 1D logical field are identical or not
!!     @param[in,out]   success     = test success (= false if the code pass the test)
!!     @param[in]       scal3D      = numerical value of the logical field (1D)
!!     @param[in]       good_scal   = theoritical value
!!     @param[in]       rank        = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_check_success_FL(scal3D, good_scal, success, rank)

    use precision
    use cart_topology
    use mpi

    logical, dimension(:),intent(in)        :: good_scal        ! theoritical value of scal1D
    logical, dimension(:),intent(in)        :: scal3D           ! the computed scalar field
    logical, intent(inout)                  :: success
    integer, intent(in)                     :: rank

    logical                                     :: success_inf      ! local error
    logical                                     :: success_inf_gl   ! global error
    integer                                     :: ierr             ! mpi error code


    success_inf = all(scal3D .eqv. good_scal)
    success_inf_gl=success_inf
    call mpi_reduce(success_inf, success_inf_gl, 1, MPI_LOGICAL, MPI_LAND, 0, MPI_COMM_WORLD, ierr)

    success = success_inf_gl

end subroutine test_check_success_FL


!> Check if two 3D logical field are identical or not
!!     @param[in,out]   success     = test success (= false if the code pass the test)
!!     @param[in]       scal3D      = numerical value of the logical field (3D)
!!     @param[in]       good_scal   = theoritical value
!!     @param[in]       rank        = mpi rank to avoid to print message for each processes (usefull if there is a lot of them)
subroutine test_check_success_F3L(scal3D, good_scal, success, rank)

    use precision
    use cart_topology
    use mpi

    logical, dimension(:,:,:),intent(in)        :: good_scal        ! theoritical value of scal1D
    logical, dimension(:,:,:),intent(in)        :: scal3D           ! the computed scalar field
    logical, intent(inout)                      :: success
    integer, intent(in)                         :: rank

    logical                                     :: success_inf      ! local error
    logical                                     :: success_inf_gl   ! global error
    integer                                     :: ierr             ! mpi error code


    success_inf = all(scal3D .eqv. good_scal)
    success_inf_gl=success_inf
    call mpi_reduce(success_inf, success_inf_gl, 1, MPI_LOGICAL, MPI_LAND, 0, MPI_COMM_WORLD, ierr)

    success = success_inf_gl

end subroutine test_check_success_F3L


end module test_common
