!> Penalization stuff (init chi, penalize vorticity)
!! Note : drag/lift are also computed with penalize routines. 
module SetsIndicators

  use client_data
  use VectorCalculus
  !use client_topology, only : nsublist
  use mpi,only:MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD
  use Domain
  implicit none

  private

  public :: init_obstacles,compute_control_box,compute_test,nocaForces,laplacian,chi_sphere,chi_boundary,chi_box,&
       getMemoryForIndicators

  ! Indicators functions 
  ! The top and bottom (z axis) boundaries
  integer, dimension(:,:), pointer  :: chi_boundary=>NULL()
  ! The sphere
  integer, dimension(:,:), pointer  :: chi_sphere=>NULL()

  ! Axes z is south-north, x down to upstream and y east-west
  !> control volume function
  integer, dimension(:,:), pointer :: chi_box => NULL() 
  !> north boundary ind. func (ie zmax)
  integer, dimension(:,:), pointer :: chi_north => NULL() 
  !> south boundary ind. func (ie zmin)
  integer, dimension(:,:), pointer :: chi_south => NULL()
  !> east boundary ind. func (ie ymin)
  integer, dimension(:,:), pointer :: chi_east => NULL()
  !> west boundary ind. func (ie ymax)
  integer, dimension(:,:), pointer :: chi_west => NULL()
  !> upstream boundary ind. func (ie xmax)
  integer, dimension(:,:), pointer :: chi_up => NULL() ! upstream
  !> downstream boundary ind. func (ie xmin)
  integer, dimension(:,:), pointer :: chi_down => NULL() ! downstream 
  !> 
  integer,parameter :: Down=1,Up=2,West=3,East=4,South=5,North=6
  !> Normal to each face of the control volume
  real(mk),dimension(dime,2*dime) :: normal
  !> A buffer for force on control volume, used to save force value on previous time step
  real(mk),dimension(dime)::bufferForce
  !> normalisation factor used to compute the drag in the Noca's way
  real(mk) :: coef
  real(mk),parameter :: uinf=1.0

  ! temp to avoid ppm dependence
  integer, parameter :: nsublist = 1


contains

  !> compute chi functions for penalization at the boundaries to enforce dirichlet conditions on zmin and zmax
  subroutine init_obstacles(resolution,step,lower,upper,center,radius,layer,coordMin)
    !> Number of points in each dir 
    integer, dimension(dime), intent(in)  :: resolution
    !> Grid steps sizes
    real(mk), dimension(dime), intent(in) :: step
    !> Dimensions of the boundary layers
    real(mk), dimension(dime), intent(in) :: upper,lower
    !> position of the center of the sphere
    real(mk),dimension(dime),intent(in):: center
    !> Radius of the sphere
    real(mk),intent(in) :: radius
    !> Boundary layer thickness
    real(mk), intent(in) :: layer
    !> Coordinates of the lowest point of the local domain
    real(mk),dimension(dime),intent(in) :: coordMin

    integer, dimension(:,:), allocatable :: tmp_boundary,tmp_sphere
    integer::istat,i,j,k
    real(mk),dimension(dime) :: coords
    real(mk) :: dist
    integer :: sizeMaxChi,count_boundary,count_sphere
    real :: layerMin,layerMax
    sizeMaxChi = product(resolution)
    allocate(tmp_boundary(dime,sizeMaxChi),stat=istat)
    allocate(tmp_sphere(dime,sizeMaxChi),stat=istat)
    
    layerMin = lower(c_Z) + layer
    layerMax = upper(c_Z) - layer

    if(istat.ne.0) stop 'Chi-boundaries function allocation error.'
    count_boundary=0
    count_sphere =0
    do k=1,resolution(c_Z)
       coords(c_Z)=coordMin(c_Z) + (k-1)*step(c_Z)
       do j=1,resolution(c_Y)
          coords(c_Y)=coordMin(c_Y) + (j-1)*step(c_Y)
          do i=1,resolution(c_X)
             coords(c_X)=coordMin(c_X) + (i-1)*step(c_X)
             if( (coords(c_Z)>layerMax).or.(coords(c_Z)<layerMin)) then
                count_boundary=count_boundary+1
                tmp_boundary(c_X,count_boundary)=i
                tmp_boundary(c_Y,count_boundary)=j
                tmp_boundary(c_Z,count_boundary)=k
             end if
             dist = dot_product(coords-center,coords-center) - radius**2
             if(dist <=0.0) then ! We are on or in the sphere ...
                count_sphere=count_sphere+1 
                tmp_sphere(c_X,count_sphere)=i
                tmp_sphere(c_Y,count_sphere)=j
                tmp_sphere(c_Z,count_sphere)=k
             end if
          end do
       end do
    end do
    allocate(chi_boundary(dime,count_boundary))
    chi_boundary=tmp_boundary(:,1:count_boundary)
    allocate(chi_sphere(dime,count_sphere))
    chi_sphere=tmp_sphere(:,1:count_sphere)
    
    deallocate(tmp_boundary,tmp_sphere)
    
  end subroutine init_obstacles

  !> Compute indicator functions for the control box (including a sphere ...)
  ! This routine must fill all chi functions in
  !> \param box_min coordinate of the lower point of the box
  !> \param boxMax coordinate of the upper point of the box
  !! We suppose (require) that boxMin+boxMax corresponds to grid points ...
  subroutine compute_control_box(resolution,step,boxMin,boxMax,center,radius,coordMin)
    !> Number of points in each dir (1st index) for each sub (2nd index)
    integer, dimension(dime),intent(in) :: resolution
    !> Grid steps sizes
    real(mk), dimension(dime), intent(in) :: step
    !> lower point of the box
    real(mk),dimension(dime),intent(in) :: boxMin
    !> upper point of the box
    real(mk),dimension(dime),intent(in) :: boxMax
    !> position of the center of the sphere
    real(mk),dimension(dime),intent(in):: center
    !> radius of the sphere
    real(mk),intent(in) :: radius
    !> Lower point of the domain
    real(mk),dimension(dime),intent(in) :: coordMin

    logical,dimension(dime) :: isMinIn,isMaxIn
    real(mk),dimension(dime) :: coordMax,coords

    integer, dimension(2*dime) :: ind
    integer, dimension(2*dime) :: nbPoints
    ! Size (number of points in each dir) of the control volume
    integer(kind=8), dimension(dime) :: boxDim
    real(mk) :: coord,dist
    integer :: count,i,j,k,direction,count_box
    integer(kind=8) :: nbPointsBox
    integer,dimension(:,:),allocatable :: tmp_box

    !> radius of the sphere
    real(mk) :: radiusBis

    ! Add an assert to check that the sphere radius is shorter thant the box size ...
    radiusBis=radius!-step(c_Z)
    ! First compute normals to the box boundaries
    normal(:,:)=0.0
    normal(c_X,Up)=1.0
    normal(c_X,Down)=-1.0
    normal(c_Y,West)=-1.0
    normal(c_Y,East)=1.0
    normal(c_Z,North)=1.0
    normal(c_Z,South)=-1.0
    ! Coordinates of the upper point 
    coordMax(:)=coordMin(:)+(resolution(:)-1)*step(:) 
    ! First : check if box boundaries are in the current domain
    isMaxIn=.False.
    isMinIn=.False.
    where(coordMin <= boxMin) isMinIn=.True.  ! Lower boundaries
    where(coordMax >= boxMax) isMaxIn=.True.  ! Upper boundaries

    !! Look for local indices corresponding to the box boundaries (i.e. x,y,z = boxMin() and boxMax())
    ! index order: Down,Up,West,East,South,North
    ind=0 !
    where(isMinIn(:).and.isMaxIn(:)) 
       ind(1:2*dime:2)=1
       ind(2:2*dime:2)=1
    end where
    where((.not.isMinIn(:)).and.(isMaxIn(:))) 
       ind(1:2*dime:2)=1
       ind(2:2*dime:2)=1
    end where
    where((isMinIn(:)).and.(.not.isMaxIn(:)))
       ind(1:2*dime:2)=1
       ind(2:2*dime:2)=resolution(:)
    end where

    do direction=1,dime 
       do k=1,resolution(direction)
          coord=coordMin(direction)+(k-1)*step(direction)
          if(isMinIn(direction)) then
             if(coord<boxMin(direction)) ind(2*direction-1)=k+1
          end if
          if(isMaxIn(direction)) then
             if(coord<boxMax(direction)) then
                ind(2*direction)=k+1
             else
                exit
             end if
          end if
       end do
    end do
    ! ind now contains the number of the points that are on the box boundary, in each direction, in the following order :
    ! ind(Down Up West East South North)

    ! Remove last point to integrate properly ...
    ind(2:2*dime:2)=ind(2:2*dime:2)-1
    
    ! Count the number of points on each face and inside the domain
    nbPoints=0
    if(isMinIn(1)) nbPoints(Down)=(ind(East)-ind(West)+1)*(ind(North)-ind(South)+1)
    if(isMaxIn(1))   nbPoints(Up)=(ind(East)-ind(West)+1)*(ind(North)-ind(South)+1)
    if(isMinIn(3)) nbPoints(South)=(ind(East)-ind(West)+1)*(ind(Up)-ind(Down)+1)
    if(isMaxIn(3)) nbPoints(North)=(ind(East)-ind(West)+1)*(ind(Up)-ind(Down)+1)
    if(isMinIn(2))  nbPoints(East)=(ind(Up)-ind(Down)+1)*(ind(North)-ind(South)+1)
    if(isMaxIn(2))  nbPoints(West)=(ind(Up)-ind(Down)+1)*(ind(North)-ind(South)+1)

    boxDim(c_X)=ind(Up)-ind(Down)+1
    boxDim(c_Y)=ind(East)-ind(West)+1
    boxDim(c_Z)=ind(North)-ind(South)+1
    nbPointsBox = boxDim(c_X)*boxDim(c_Y)*boxDim(c_Z)
    
    allocate(tmp_box(dime,nbPointsBox))
    allocate(chi_up(dime,nbPoints(Up)),chi_down(dime,nbPoints(Down)),chi_east(dime,nbPoints(East)),chi_west(dime,nbPoints(West)))
    allocate(chi_south(dime,nbPoints(South)),chi_north(dime,nbPoints(North)))
    count_box=0

    if(all(boxDim>0)) then
       do k=ind(South),ind(North)
          coords(c_Z) = coordMin(c_Z)+(k-1)*step(c_Z)
          do j=ind(West),ind(East)
             coords(c_Y) = coordMin(c_Y)+(j-1)*step(c_Y)
             do i=ind(Down),ind(Up)
                coords(c_X) = coordMin(c_X)+(i-1)*step(c_X)
                dist = dot_product(coords-center,coords-center) - radiusBis**2
                if(dist >=0.0) then ! We are on or outside the sphere ...
                   count_box = count_box+1
                   tmp_box(c_X,count_box)=i
                   tmp_box(c_Y,count_box)=j
                   tmp_box(c_Z,count_box)=k
                end if
             end do
          end do
       end do
    end if
    allocate(chi_box(dime,count_box))
    chi_box=tmp_box(:,1:count_box)
    
    deallocate(tmp_box)
    if(isMinIn(3)) then ! South boundary
       count=1
       chi_south(3,:)=ind(South)
       do j=ind(West),ind(East)
          do i=ind(Down),ind(Up)
             chi_south(1,count)=i
             chi_south(2,count)=j
             count=count+1
          end do
       end do
    end if
    if(isMinIn(2)) then ! East boundary
       count=1
       chi_east(2,:)=ind(East)+1
       do k=ind(South),ind(North)
          do i=ind(Down),ind(Up)
             chi_east(1,count)=i
             chi_east(3,count)=k
             count=count+1
          end do
       end do
    end if
    if(isMinIn(1)) then ! Downstream boundary is in the domain
       count=1
       chi_down(1,:)=ind(Down)
       do k=ind(South),ind(North)
          do j=ind(West),ind(East)
             chi_down(2,count)=j
             chi_down(3,count)=k
             count=count+1
          end do
       end do
    end if

    if(isMaxIn(3)) then ! North boundary is in the domain
       count=1
       chi_north(3,:)=ind(North)+1
       do j=ind(West),ind(East)
          do i=ind(Down),ind(Up)
             chi_north(1,count)=i
             chi_north(2,count)=j
             count=count+1
          end do
       end do
    end if
    if(isMaxIn(2)) then ! West boundary is in the domain
       count=1
       chi_west(2,:)=ind(West)
       do k=ind(South),ind(North)
          do i=ind(Down),ind(Up)
             chi_west(1,count)=i
             chi_west(3,count)=k
             count=count+1
          end do
       end do
    end if
    if(isMaxIn(1)) then ! Upstream boundary is in the domain
       count=1
       chi_up(1,:)=ind(Up)+1
       do k=ind(South),ind(North)
          do j=ind(West),ind(East)
             chi_up(2,count)=j
             chi_up(3,count)=k
             count=count+1
          end do
       end do
    end if

    bufferForce = 0.0    
    ! Compute coef used to calculate the drag in the Nocas's way
    coef = 2./(uinf**2*pi*radius**2)

  end subroutine compute_control_box
  

  !> Set input field to one on the control volume boundaries and to zero elsewhere 
  subroutine compute_test(testfield,chi)
    real(mk), dimension(:,:,:),pointer:: testfield
    integer,dimension(:,:),pointer ::chi
    integer :: k
    do k=1,size(chi,2)
       testfield(chi(1,k),chi(2,k),chi(3,k)) = 1.0
    end do
  end subroutine compute_test

  !> Computation of the drag according to "method B" presented in
  !! Noca99 or Plouhmans, 2002, Journal of Computational Physics
  subroutine nocaForces(force,velo,vort,nu,coordMin,step,dt,dvol)

    !> The force to be computed
    real(mk), dimension(dime),intent(inout) :: force
    !! velocity and vorticity fields, intent(in)
    real(mk), dimension(:,:,:,:,:),pointer :: velo,vort
    !> viscosity
    real(mk),intent(in)::nu
    !! Coordinates of the lowest point in the domain
    real(mk),dimension(dime),intent(in):: coordMin
    !! mesh step sizes
    real(mk),dimension(dime),intent(in)::step
    !> Time step
    real(mk),intent(in)::dt
    !> element. vol
    real(mk),intent(in) ::dvol
    ! Surface element
    real(mk) :: dsurf
    real(mk),dimension(dime)::localForce
    integer :: info
    localForce=0.0
    force=0.0

    ! Downstream and upstream surface
    dsurf=step(c_Y)*step(c_Z)
    call integrateOnSurface(localForce,velo,vort,chi_down,normal(:,Down),c_X,nu,dsurf,coordMin,step)
    call integrateOnSurface(localForce,velo,vort,chi_up,normal(:,Up),c_X,nu,dsurf,coordMin,step)
    ! East and West 
    dsurf=step(c_X)*step(c_Z)
    call integrateOnSurface(localForce,velo,vort,chi_east,normal(:,East),c_Y,nu,dsurf,coordMin,step)
    call integrateOnSurface(localForce,velo,vort,chi_west,normal(:,West),c_Y,nu,dsurf,coordMin,step)
    ! North and south
    dsurf=step(c_Y)*step(c_X)
    call integrateOnSurface(localForce,velo,vort,chi_south,normal(:,South),c_Z,nu,dsurf,coordMin,step)
    call integrateOnSurface(localForce,velo,vort,chi_north,normal(:,North),c_Z,nu,dsurf,coordMin,step)
    ! over the volume ...
    call integrateOnBox(localForce,vort,chi_box,dvol,dt,coordMin,step)

    localForce=localForce*coef
    !write(*,'(a,3f10.5)') ' drag local: ', localForce
    call MPI_Reduce(localForce,force,dime,MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,info)

  end subroutine nocaForces

  !> Return -1/(dime-1)*d/dt int_over_control_box coord X vorticity 
  subroutine integrateOnBox(force,vort,chi,dvol,dt,coordMin,step)
    !! The force to be computed
    real(mk), dimension(dime),intent(inout) :: force
    !! vorticity fields, intent(in)
    real(mk), dimension(:,:,:,:,:),pointer :: vort
    !! Indicator function of the box
    integer, dimension(:,:), pointer :: chi
    !! Element of volume
    real(mk),intent(in)::dvol
    !> Time step
    real(mk),intent(in)::dt
    !! Coordinates of the lowest point in the domain
    real(mk),dimension(dime),intent(in):: coordMin
    !! mesh step sizes
    real(mk),dimension(dime),intent(in)::step

    ! coordinates of the current point
    real(mk),dimension(dime) :: coords,int1
    ! local indices
    integer :: i,j,k,ind
    real(mk)::fact
    fact = -dvol/((dime-1)*dt)
    int1=0.0
    !! For all points in the box ...
    do ind=1,size(chi,2)
       i=chi(1,ind)
       j=chi(2,ind)
       k=chi(3,ind)
       ! coordinates of the current point
       coords = coordMin + (chi(:,ind)-1)*step
       !! part1 of the force
       int1=int1+cross_prod(coords,vort(:,i,j,k,nsublist))
    end do
    force = force + fact*(int1-bufferForce)
    bufferForce = int1 ! Save for next time step ...
  end subroutine integrateOnBox

  !> Compute integrals on surface to calculate forces acting on the body.
  !! See (2.1) of Noca 1999 or (52) of Plouhmans 2002 
  !! Integrals on the sphere are neglected. 
  subroutine integrateOnSurface(force,velo,vort,chi,NormalVec,direction,nu,dsurf,coordMin,step)
    !! The force to be computed
    real(mk), dimension(dime),intent(inout) :: force
    !! velocity and vorticity fields, intent(in)
    real(mk), dimension(:,:,:,:,:),pointer :: velo,vort
    !! Indicator function of the considered face, intent(in)
    integer, dimension(:,:), pointer :: chi
    !! Normal to the considered face (dir : to the outside of the volume)
    real(mk), dimension(dime),intent(in) :: NormalVec
    !! index of the non-null coordinate in normal, must be c_X,c_Y or c_Z
    integer,intent(in) :: direction
    !! viscosity
    real(mk),intent(in)::nu
    !! Element of surface
    real(mk),intent(in)::dsurf
    !! Coordinates of the lowest point in the domain
    real(mk),dimension(dime),intent(in):: coordMin
    !! mesh step sizes
    real(mk),dimension(dime),intent(in)::step

    !! Some local values ...
    real(mk) :: u_u,n_u,n_w,fact
    integer :: i,j,k,ind
    !! local coordinates
    real(mk), dimension(dime) :: coords
    real(mk), dimension(dime) :: int1,int2,nDivT,diff_dir,buffer
    
    fact = 1./(dime-1)
    ! For each point of the current plane ...
    int1=0.0
    int2=0.0
    diff_dir=0.0
    do ind=1,size(chi,2)
       i=chi(1,ind)
       j=chi(2,ind)
       k=chi(3,ind)
       !! part1 = 1/2(velocity.velocity)n - (n.velocity)velocity - 1/(dime-1)((n.velocity)(coord X vorticity) + (n.vorticity)(coord X velocity)
       ! 0.5*velocity.velocity
       u_u=dot_product(velo(:,i,j,k,nsublist),velo(:,i,j,k,nsublist))
       ! normal.velocity
       n_u=dot_product(velo(:,i,j,k,nsublist),NormalVec(:))
       ! normal.vorticity
       n_w=dot_product(vort(:,i,j,k,nsublist),NormalVec(:))
       ! coordinates of the current point
       coords = coordMin + (chi(:,ind)-1)*step
       !! part1 of the force
       int1=int1+0.5*u_u*NormalVec(:)-n_u*velo(:,i,j,k,nsublist)&
            - fact*n_u*cross_prod(coords,vort(:,i,j,k,nsublist))&
            + fact*n_w*cross_prod(coords,velo(:,i,j,k,nsublist))

       !! part 2 of the force, the one concerning T = nu(nabla u + nabla uT)
       !! Considering that the box is a parallepiped, each normal is equal to something like 1 0 0 
       !! and the integral simplifies in :
       !! n.T = nabla velocity_dir + d/ddir velocity, dir being the dir of the normal,
       !! n X nabla.T = function(laplacian of the velocity components)
       nDivT = 0.0 ! n X nabla.T 

       if(direction==c_X) then ! face = Down or Up, d/dx
          diff_dir = diffX(velo,i,j,k,step(direction),nsublist)
          nDivT(2) = -laplacian(velo,c_Z,i,j,k,step,nsublist) ! Laplacian of velocity_y
          nDivT(3) = laplacian(velo,c_Y,i,j,k,step,nsublist) ! Laplacian of velocity_z
       else if(direction==c_Y) then ! face = East or West, d/dy
          diff_dir = diffY(velo,i,j,k,step(direction),nsublist) 
          nDivT(3)=-laplacian(velo,c_X,i,j,k,step,nsublist) ! Laplacian of velocity_x
          nDivT(1)=laplacian(velo,c_Z,i,j,k,step,nsublist) ! Laplacian of velocity_z
       else if(direction==c_Z) then ! face = North or South, d/dz
          diff_dir = diffZ(velo,i,j,k,step(direction),nsublist)
          nDivT(2)=laplacian(velo,c_X,i,j,k,step,nsublist) ! Laplacian of velocity_x
          nDivT(1)=-laplacian(velo,c_Y,i,j,k,step,nsublist) ! Laplacian of velocity_y
       end if
       buffer=nabla(velo,direction,i,j,k,step,nsublist) + diff_dir + fact*cross_prod(coords,nDivT)
       buffer=NormalVec(direction)*nu*buffer
       int2=int2+buffer

    end do

    ! Product with element of surface and sum to the total (input) force
    force = force+(int1+int2)*dsurf
  end subroutine integrateOnSurface

  !> get memory allocated for indicator sets
  function getMemoryForIndicators()
    real(mk) :: getMemoryForIndicators
    
    getMemoryForIndicators = sizeof(chi_boundary)+sizeof(chi_sphere)+sizeof(chi_box)+sizeof(chi_north)&
         + sizeof(chi_west)+sizeof(chi_east)+sizeof(chi_south)+sizeof(chi_up)+sizeof(chi_down)
    getMemoryForIndicators = getMemoryForIndicators*1e-6
    if(verbose) then
       write(*,'(a,i3,a,f10.4,a)') &
            '[',rank,'] memory used for indicator sets:', getMemoryForIndicators, ' MB.'
    end if
    
  end function getMemoryForIndicators

end module SetsIndicators
