!> Functions used to compute fields values on the grid, for specific pre-defined cases. 
module testsFunctions

  use client_data
  implicit none

  real(mk) :: xref
  integer :: np
contains
  
  !> Computes the analytical values for stream function, velocity and vorticity such that
  !>   \f{eqnarray*}{ rhs_{ex} &=& -\omega_{ex} = \Delta\psi_{ex} \\    vel_{ex} &=& \nabla \times \psi_{ex} \f}
  !>   \f{eqnarray*}{\nabla.\psi_{ex} &=& \nabla.\omega_{ex} = \nabla.vel_{ex} = 0.0  \\\nabla\times vel_{ex} &=& \omega_{ex}  \f}
  !>  (see maple file)
  subroutine poisson_analytic(resolution,step,coordMin,rhs_ex,vel_ex,psi_ex)

    !> the local resolution
    integer, dimension(dime),intent(in) :: resolution
    !> size of mesh step in each dir
    real(mk), dimension(dime),intent(in) :: step
    !> Coordinates of the local minimum point
    real(mk),dimension(dime),intent(in) :: coordMin
    ! rhs function values on the grid (i.e. -omega_ex)
    real(mk), dimension(:,:,:,:), pointer :: rhs_ex
    ! velocity function values on the grid
    real(mk), dimension(:,:,:,:), pointer :: vel_ex
    ! Stream function values on the grid
    real(mk), dimension(:,:,:,:), pointer :: psi_ex
    
    
    real(mk) :: x,y,z,cx,cy,cz,c2x,c2y,c2z,sx,sy,sz,s2x,s2y,s2z
    integer :: i,j,k
    real(mk) :: pi
    
    pi = 4.0*atan(1.0_mk)
       
    do k=1,resolution(c_Z)
       z = coordMin(c_Z) + (k-1)*step(c_Z)
       cz=cos(pi*z)
       c2z=cos(2.*pi*z)
       sz=sin(pi*z)
       s2z=sin(2.*pi*z)
       do j=1,resolution(c_Y)
          y = coordMin(c_Y) + (j-1)*step(c_Y)
          cy=cos(pi*y)
          c2y=cos(2.*pi*y)
          sy=sin(pi*y)
          s2y=sin(2.*pi*y)
          do i=1,resolution(c_X)
             x = coordMin(c_X) + (i-1)*step(c_X)
             cx=cos(pi*x)
             c2x=cos(2.*pi*x)
             sx=sin(pi*x)
             s2x=sin(2.*pi*x)
             
             rhs_ex(1,i,j,k) = 8.*s2y*pi**2*s2z
             rhs_ex(2,i,j,k) = 8.*s2x*pi**2*s2z
             rhs_ex(3,i,j,k)=  8.*s2x*pi**2*s2y
             
             vel_ex(1,i,j,k) = 2.*s2x*pi*(c2y - c2z)
             vel_ex(2,i,j,k) = 2.*s2y*pi*(c2z - c2x)
             vel_ex(3,i,j,k) = 2.*s2z*pi*(c2x - c2y)
             
             psi_ex(1,i,j,k) = sy*sz
             psi_ex(2,i,j,k) = sx*sz
             psi_ex(3,i,j,k) = sx*sy
             
          end do
       end do
    end do
  end subroutine poisson_analytic
  
  !> Init vorticity field on a grid, such that :
  !> \f$  \omega(x,y,z) =  \left[\begin{array}{c} 0 \\ -3z/(Lz^2) \\ 0 \end{array}\right]\f$
  subroutine init_vorticity(vorticity,resolution,step,coordMin,lower,upper)

    !> vorticity field
    real(mk), dimension(:,:,:,:), pointer :: vorticity 
    !> the local mesh resolution
    integer,dimension(dime),intent(in) :: resolution
    !> size of mesh step in each dir
    real(mk), dimension(dime),intent(in) :: step
    !> Coordinates of the minimal point of the local domain
    real(mk),dimension(dime),intent(in) :: coordMin
    !> boundaries of the domain
    real(mk),dimension(dime),intent(in):: upper,lower
    integer :: i,j,k
    real(mk) :: x,z,physicalDomainSize_Z
    real(mk) :: coef

    physicalDomainSize_Z = (upper(c_Z)- lower(c_Z))/2.
    vorticity = 0.0
    coef = -3./(physicalDomainSize_Z)**2
    
    do k=1,resolution(c_Z)
       z = coordMin(c_Z) + (k-1)*step(c_Z)
       do j=1,resolution(c_Y)
          !y = coordMin(c_Y) + (j-1)*step(c_Y)
          do i=1,resolution(c_X)
             x = coordMin(c_X) + (i-1)*step(c_X)
             !if( (z < upper(c_Z)).and.(z>lower(c_Z))) then
                vorticity(c_Y,i,j,k) = cos(2.*pi*z)!)coef*z
             !endif
          end do
       end do
    end do
    
    
  end subroutine init_vorticity
  
  !> Computes the analytical values for stream function, velocity and vorticity such that
  !>   \f{eqnarray*}{ rhs_{ex} &=& -\omega_{ex} = \Delta\psi_{ex} \\    vel_{ex} &=& \nabla \times \psi_{ex} \f}
  !>   \f{eqnarray*}{\nabla.\psi_{ex} &=& \nabla.\omega_{ex} = \nabla.vel_{ex} = 0.0  \\\nabla\times vel_{ex} &=& \omega_{ex}  \f}
  !>  (see maple file)
  subroutine rhs_analytic(resolution,step,coordMin,rhs_ex,velocity,vorticity,nu)
    !> the local mesh resolution
    integer,dimension(dime),intent(in) :: resolution
    !> size of mesh step in each dir
    real(mk), dimension(dime),intent(in) :: step
    !> Coordinates of the minimal point of the local domain
    real(mk),dimension(dime),intent(in) :: coordMin
    ! rhs function values on the grid (i.e. -omega_ex)
    real(mk), dimension(:,:,:,:), pointer :: rhs_ex,velocity,vorticity
    !> viscosity
    real(mk),intent(in) :: nu

    real(mk) :: x,y,z,cx,cy,cz,c2x,c2y,c2z,sx,sy,sz,s2x,s2y,s2z
    integer :: i,j,k
    
    do k=1,resolution(c_Z)
       z = coordMin(c_Z) + (k-1)*step(c_Z)
       cz=cos(pi*z)
       c2z=cos(2.*pi*z)
       sz=sin(pi*z)
       s2z=sin(2.*pi*z)
       do j=1,resolution(c_Y)
          y = coordMin(c_Y) + (j-1)*step(c_Y)
          cy=cos(pi*y)
          c2y=cos(2.*pi*y)
          sy=sin(pi*y)
          s2y=sin(2.*pi*y)
          do i=1,resolution(c_X)
             x = coordMin(c_X) + (i-1)*step(c_X)
             cx=cos(pi*x)
             c2x=cos(2.*pi*x)
             sx=sin(pi*x)
             s2x=sin(2.*pi*x)
             
             rhs_ex(1,i,j,k) = 32.*s2y*pi**4*s2z*(-2.*nu+c2x*c2y-c2x*c2z)
             rhs_ex(2,i,j,k) = -32.*s2x*pi**4*s2z*(2.*nu-c2y*c2z+c2x*c2y)
             rhs_ex(3,i,j,k) = 32.*s2x*pi**4*s2y*(-2.*nu+c2x*c2z-c2z*c2y)

             vorticity(1,i,j,k) = 8.*s2y*pi**2*s2z
             vorticity(2,i,j,k) = 8.*s2x*pi**2*s2z
             vorticity(3,i,j,k)=  8.*s2x*pi**2*s2y
             
             velocity(1,i,j,k) = 2.*s2x*pi*(c2y - c2z)
             velocity(2,i,j,k) = 2.*s2y*pi*(c2z - c2x)
             velocity(3,i,j,k) = 2.*s2z*pi*(c2x - c2y)
             
             
          end do
       end do
    end do
  end subroutine rhs_analytic
  !> Test purpose function,
  !> computes \f$  \omega(x,y,z) =  \left[\begin{array}{c} 0 \\ 0 \\ 0.1 \end{array}\right]\f$
  !> and  \f$  velocity(x,y,z) =  \left[\begin{array}{c} 0 \\ 0 \\ 2.0 \end{array}\right]\f$
  subroutine test_particles(vorticity,velocity,resolution,step,coordMin)
    
    !> vorticity field
    real(mk), dimension(:,:,:,:), pointer :: vorticity ,velocity
    !> the local resolution
    integer, dimension(dime),intent(in) :: resolution
    !> size of mesh step in each dir
    real(mk), dimension(dime),intent(in) :: step
    !> Coordinates of the local minimum point
    real(mk),dimension(dime),intent(in) :: coordMin

    integer :: i,j,k,l
    real(mk) :: pi,x,y,z
    
    pi = 4.0*atan(1.0_mk)
    vorticity = 0.
    velocity = 0.
    do k=1,resolution(c_Z)
       z = coordMin(c_Z) + (k-1)*step(c_Z)
       do j=1,resolution(c_Y)
          y = coordMin(c_Y) + (j-1)*step(c_Y)
          do i=1,resolution(c_X)
             x = coordMin(c_X) + (i-1)*step(c_X)
             if( (x .eq.10*step(c_X)).and.(y.eq.0).and.(z.eq.xref)) then
                do l=1,np
                   vorticity(3,i,j,k+l-1) =  0.1;
                   velocity(3,i,j,k+l-1) = 2.0
                enddo
             endif
          end do
       end do
    end do
    xref = xref + step(c_Z)
    !np = np +1
  end subroutine test_particles
  
  !> Test purpose function,
  !> computes \f$  \omega(x,y,z) =  \left[\begin{array}{c} cos(x) \\ cos(x) \\ cos(x) \end{array}\right]\f$
  subroutine test_vorticity(vorticity,resolution,step,coordMin,lower,upper)
    ! vorticity field
    real(mk), dimension(:,:,:,:), pointer :: vorticity 
    !> the local resolution
    integer, dimension(dime),intent(in) :: resolution
    !> size of mesh step in each dir
    real(mk), dimension(dime),intent(in) :: step
    !> Coordinates of the local minimum point
    real(mk),dimension(dime),intent(in) :: coordMin
    !> Boundaries coordinates
    real(mk),dimension(dime),intent(in):: upper, lower

    integer :: i,j,k
    real(mk) :: pi,x,y,z,physicalDomainSize_Z
    real(mk) :: coef

    physicalDomainSize_Z = (upper(3)- lower(3))/2.
    pi = 4.0*atan(1.0_mk)
    vorticity = 0.0
    coef = -3./(physicalDomainSize_Z)**2
    do k=1,resolution(c_Z)
       z = coordMin(c_Z) + (k-1)*step(c_Z)
       do j=1,resolution(c_Y)
          y = coordMin(c_Y) + (j-1)*step(c_Y)
          do i=1,resolution(c_X)
             x = coordMin(c_X) + (i-1)*step(c_X)
             vorticity(:,i,j,k) = cos(x)
          end do
       end do
    end do
  end subroutine test_vorticity


  !> Compute flow through x = xmin surface, for the following stream function :
  !> \f$  \psi(x,y,z) =  \left[\begin{array}{c} 0 \\-U_{inf}z(1 - \frac{R^2}{z^2+x^2})  \\ 0 \end{array}\right] \f$
  !> i.e.
  !> \f[ flowRate_{theo} = \left[ -U_{inf}length_y(z - \frac{R^2z}{z^2+x^2})\right]_{lower_z}^{upper_z}\f]
  function requiredFlowRate3D(radius,length,lower,upper,uinf)

    !> sphere radius
    real(mk),intent(in) :: radius
    !> domain dimensions
    real(mk),dimension(dime),intent(in) :: length
    !> physical domain lower point
    real(mk),dimension(dime),intent(in)::lower
    !> physical domain upper point
    real(mk),dimension(dime),intent(in)::upper
    !> velocity inf
    real(mk),intent(in) :: uinf
    !> required flow rate
    real(mk) :: requiredFlowRate3D
    
    real(mk) :: dom
    ! position of the surface for flow rate computation
    real(mk)::xPos

    xPos = lower(c_X)
    dom = upper(c_Z)**2+xPos**2
    if(abs(dom) < epsilon(dom)) then ! if dom == 0
       requiredFlowRate3D = upper(c_Z)
    else
       requiredFlowRate3D = upper(c_Z)*(1.-radius**2/dom)
    end if
    dom = lower(c_Z)**2+xPos**2
    if(abs(dom) < epsilon(dom)) then
       requiredFlowRate3D = requiredFlowRate3D - lower(c_Z)
    else 
       requiredFlowRate3D = requiredFlowRate3D - lower(c_Z)*(1. - radius**2/dom)
    end if
    requiredFlowRate3D = requiredFlowRate3D * uinf *length(c_Y)
    return
  end function requiredFlowRate3D

    !> Compute flow through x = xmin surface, for the following stream function :
  !> \f$  \psi(x,y,z) =  \left[\begin{array}{c} 0 \\-U_{inf}z(1 - \frac{R^2}{z^2+x^2})  \\ 0 \end{array}\right] \f$
  !> i.e.
  !> \f[ flowRate_{theo} = \left[ -U_{inf}length_y(z - \frac{R^2z}{z^2+x^2})\right]_{lower_z}^{upper_z}\f]
  function requiredFlowRate2D(radius,lower,upper,uinf)

    !> sphere radius
    real(mk),intent(in) :: radius
    !> physical domain lower point
    real(mk),dimension(dime),intent(in)::lower
    !> physical domain upper point
    real(mk),dimension(dime),intent(in)::upper
    !> velocity inf
    real(mk),intent(in) :: uinf
    !> required flow rate
    real(mk) :: requiredFlowRate2D
    
    real(mk) :: dom
    ! position of the surface for flow rate computation
    real(mk)::xPos

    xPos = lower(c_X)
    dom = upper(c_Y)**2+xPos**2
    if(abs(dom) < epsilon(dom)) then ! if dom == 0
       requiredFlowRate2D = upper(c_Y)
    else
       requiredFlowRate2D = upper(c_Y)*(1.-radius**2/dom)
    end if
    dom = lower(c_Y)**2+xPos**2
    if(abs(dom) < epsilon(dom)) then
       requiredFlowRate2D = requiredFlowRate2D - lower(c_Y)
    else
       requiredFlowRate2D = requiredFlowRate2D - lower(c_Y)*(1. - radius**2/dom)
    end if
    requiredFlowRate2D = requiredFlowRate2D * uinf
    return
  end function requiredFlowRate2D


  subroutine Gaussian2D(field,resolution,step,coordMin,center)
    
    !> Field initialized with a Gaussian
    real(mk), dimension(:,:,:), pointer :: field
    !> Space step
    real(mk), dimension(dime) :: step
    !> local resolution
    integer, dimension(dime) :: resolution
    !> Coordinates of the lowest point in the current subdomain
    real(mk),dimension(dime),intent(in) :: coordMin
    !> 
    real(mk), dimension(dime),intent(in) :: center
           
    real(mk), parameter :: sigma = 0.2
    real(mk) :: expo
    real(mk),dimension(dime) :: coord
    integer :: i,j

    field = 0.0
    do j = 1,resolution(c_Y)
       coord(c_Y) = coordMin(c_Y)+(j-1)*step(c_Y)-center(c_Y)
       do i = 1, resolution(c_X)
          coord(c_X) = coordMin(c_X) + (i-1)*step(c_X)-center(c_X)
          expo=dot_product(coord,coord)*0.5/sigma**2
          field(i,j,:) = exp(-expo)
       end do
    end do

  end subroutine Gaussian2D

  subroutine Gaussian1D(field,resolution,step,coordMin,dir,shift)
    
    !> Field initialized with a Gaussian
    real(mk), dimension(:,:,:), pointer :: field
    !> Space step
    real(mk), dimension(dime) :: step
    !> local resolution
    integer, dimension(dime) :: resolution
    !> Coordinates of the lowest point in the current subdomain
    real(mk),dimension(dime),intent(in) :: coordMin
    !> 
    real(mk), intent(in) :: shift
    !> Advection direction
    integer, intent(in) :: dir

    real(mk), parameter :: sigma = 0.2
    real(mk) :: coord,coeff
    integer :: i

    !coord(c_Y) = coordMin(c_Y) + (j-1)*step(c_Y)
    coeff = 1./(sigma*sqrt(2.*pi))
    print * ,"shift",shift
    field = 0.
    do i = 1, resolution(dir)
       coord = coordMin(dir) + (i-1)*step(dir)-shift
       field(i,:,1) = coeff*exp(-0.5*(coord/sigma)**2.)
    end do

  end subroutine Gaussian1D
  

end module testsFunctions
