!> This module is used to run a simulation using 
!! ppm (core and numerics).
!! Solve Navier-Stokes (vorticity) for a 2D flow around a cylinder.
!!
module NavierStokes2D
  
  ! All required ppm modules ...
  use ppm_module_init, only : ppm_init
  use ppm_module_data, only : ppm_kind_double,ppm_param_bcdef_periodic
  use ppm_module_finalize, only : ppm_finalize
  use ppm_module_map_field_ghost, only:ppm_map_field_ghost_get, ppm_map_field_ghost_put
  use ppm_module_map_field
  use ppm_module_is_initialized
  ! use client_io
  use client_data
  ! some tools
  use parmesTools
  ! Physical domain and grid
  use Domain
  ! Fields on the grid
  use Fields, only: initFields, velocity, vorticity2D, gauss,rhs,scalar!,testFunc!,getMemoryUsedForFields,shiftVelocityX, gauss!vel_ex)!, stream_function
  use PPMFields
  ! Topology
  use client_topology, only: PPMinitTopo,topo,getPPMLocalResolution,meshNum
  ! Poisson (ppm) solver
  use Solver, only : init_poisson_solver, solve_poisson, ppm_poisson_drv_none, ppm_poisson_drv_curl_fd2, ppm_poisson_drv_curl_fd4
  ! Penalisation stuff
  use penalisation, only : penalise_velocity
  ! Functions used to identify surfaces, volumes ... 
  use SetsIndicators,only:chi_sphere,chi_boundary,compute_control_box,init_obstacles,nocaForces,&
       compute_test,chi_box,getMemoryForIndicators
  ! curl, prod ...
  use vectorcalculus
  ! everything dealing with particles
  use Particles!, only : initNSSolver_particles,getMemoryUsedForParticles,countAndUpdateParticles,countAndCreateParticles,RK4_2D,&
  !     createParticlesEverywhere,PPMupdateParticles2D,Ppmremesh2d,RK2_2D
  ! file io 
  use io_vtk
  
  use mpi
  
  ! user-defined functions (for tests or fields initialization)
  use testsFunctions

  use poisson

!  use solverDiffusion

  implicit none

  ! Some global vars ... 
  !> counter for memory usage
  real(mk) :: memoryUsed
  !> required flow rate (used to shift velocity after Poisson)
  real(mk) :: reqFlowRate
  
contains

  !> All required initialisations :
  !! MPI, ppm
  !! Domain, grid
  !! Fields
  !! Particles
  subroutine init_client(Re,localResolution,coordMin)

    use client_topology, only: isubl
    !> Reynolds numner
    real(mk),intent(in) ::Re
    !> local mesh resolution
    integer, dimension(dim3),intent(out) :: localResolution
    !> Local (proc.) minimum point coordinates
    real(mk),dimension(dim3),intent(out) ::coordMin

    ! Precision and tolerance (for ppm init)
    integer :: prec,tol
    ! MPI comm
    integer :: comm
    ! debug mode (for ppm)
    integer :: debug
    
    !> Size of the control box (user defined)
    real(mk) :: sizeOfTheBox  
    !> Dimensions and positions of the control volume
    real(mk),dimension(dime):: boxMin,boxMax
    
    !> For ppm fftw solver
    integer :: derive
    
    !> Sphere radius
    real(mk) :: sphere_radius 
    !> Sphere position
    real(mk),dimension(dime) :: sphere_pos
    !> velocity inf
    real(mk),parameter :: uinf = 1.0
    integer :: nbPoints

    !> Thickness of the boundary layer used to enforce Dirichlet BC with penalisation
    real(mk) :: layer

    logical :: testInit = .False.
    integer :: info

    ! "read" mpi parameters
    call MPI_COMM_RANK(MPI_COMM_WORLD,rank,info)
    call MPI_COMM_SIZE(MPI_COMM_WORLD,nbprocs,info)
    comm = MPI_COMM_WORLD  
    
    write (*,'(a,i5,a)') '[',rank,'] --------------------------------------------------------------> start initialization '

    !======================
    ! Init ppm 
    !======================
    prec = ppm_kind_double 
    debug = 0
    tol = -10
    info = 0 !!! Warning : info MUST be 0 else ppm does not init and tells nothing ...
    call ppm_init(dime,prec,tol,comm,debug,info)
    
    !======================
    ! Geometry and grid
    !======================
    ! Set domain size, resolution, boundaries, number of ghost points ... 
    ! At the time, many values are set in Domain.f90 file
    call init_geometry(ppm_param_bcdef_periodic)
    call init_grid()

    !======================
    ! Creates the topology
    !======================
    ! Based on ppm. 
    call PPMinitTopo(physDomainLowerPoint,physDomainUpperPoint,domain_bc,domain_ghostsize,grid_resolution)
    
    !> Get the coordinates of the lowest point for the current domain
    coordMin(:)=topo%min_subd(:,isubl) 
    coordMin(c_Z) = physDomainUpperPoint(c_Z)
    !======================
    ! Fields allocation
    !======================
    ! Local number of nodes
    localResolution = getPPMLocalResolution(topo,meshNum)

    ! Allocate fields on the local mesh (velocity, vorticity ...)
    call initFields(localResolution,domain_ghostsize)
    
    !=======================================================
    ! Set obstacle (sphere) and other specific parameters.
    ! Compute a control volume for diagnostics computation
    !=======================================================
    !> Set sphere size and position
!!$    sphere_radius=0.5
!!$    sphere_pos(c_Y)=(physDomainUpperPoint(c_Y)+physDomainLowerPoint(c_Y))/2.0
!!$    sphere_pos(c_X) = 0.0
!!$ !   sphere_pos(c_Z) = 0.0
!!$    
!!$    if(rank == 0) then
!!$       reqFlowRate = requiredFlowRate2D(sphere_radius,domainLength,physDomainLowerPoint,physDomainUpperPoint,uinf)
!!$    end if
!!$    call MPI_Bcast(reqFlowRate,1,mpi_mk,0,MPI_COMM_WORLD,info);

    !! We set the size of the box (must be greater than the sphere diameter ...)
    !    sizeOfTheBox = domainLength(c_Y)-0.3 ! 6*sphere_radius
    
    !! Position of the upper and lower points of the box
    !! Compute the box position on the grid
    !nbPoints = floor(abs(0.5*(domainLength(c_Z)-sizeOfTheBox)/grid_step(c_Z)))
    !boxMin=physDomainLowerPoint!+3*grid_step
    !boxMax=physDomainUpperPoint!-3*grid_step
!!$    nbPoints = 200
!!$
!!$    boxMin = physDomainLowerPoint
!!$    boxMax = physDomainUpperPoint
!!$    boxMin(c_Y)=physDomainLowerPoint(c_Y)+ nbPoints*grid_step(c_Y)
!!$    boxMax(c_Y)=physDomainUpperPoint(c_Y)- nbPoints*grid_step(c_Y)
!!$    boxMin(c_X)=physDomainLowerPoint(c_X)+ nbPoints*grid_step(c_X)
!!$    boxMax(c_X)=physDomainUpperPoint(c_X)- (nbPoints+401)*grid_step(c_X)
!!$    
    !! compute indicator functions for the control box and the sphere
  !  call compute_control_box(localResolution,grid_step,boxMin,boxMax,sphere_pos,sphere_radius,coordMin)
    
    !====================================================================
    ! Compute indicator function for the boundaries in z dir 
    ! (will be used to enforce dirichlet conditions on these boundaries.)
    !====================================================================
    layer = 0.0
   ! call init_obstacles(localResolution,grid_step,physDomainLowerPoint,physDomainUpperPoint,sphere_pos,sphere_radius,layer,coordMin)
    ! Test : export test function to vtk file.
    ! compute_test set 1 at all points belonging to chi_... 
 !   testFunc=0.0
    !call compute_test(testFunc,chi_boundary)
    !call compute_test(testFunc,chi_box)
    !testFunc = 1.0
    !call printChiToVTK("box",testFunc,localResolution,grid_step,coordMin)
    !! display the memory used for indicator sets and fields
 !   memoryUsed=getMemoryUsedForFields()+getMemoryForIndicators()
!    deallocate(testFunc)
   
    !================================
    ! Solvers (grid and particles) 
    !================================
    
    ! --- Grid solver for Poisson ---
    ! derive : means that finite difference, order 4, are used to compute velocity from stream function, in ppm
    !derive = ppm_poisson_drv_curl_fd4 
    ! init fftw solver, from ppm
    !call init_poisson_solver(vorticity,velocity,topo%ID,meshNum,derive)
    
    ! Initialisation of all the fftw stuff (plans ...).
    ! Warnings : 
    ! - this must be done before any initialisation of the fields : plans creation may rewrite data of fieldIn/fieldOut.
    ! - last argument (resolution) must be the global resolution : dedicated mpi-topologies are created for fftw. 
    call initFFTW2D(grid_resolution,domainLength)
    
    !call diffusionInitFFT(topo%ID,velocity,vorticity,1./Re,domainLength,grid_resolution,physDomainLowerPoint,&
     !    physDomainUpperPoint,domain_bc)
    ! --- Particles ---
    ! Set particles parameters (kernel type, cutoff ...)
    call initNSSolver_particles()
    
    if(verbose) then
       write(*,'(a,i5,a)') '[',rank,'] ======================================== Summary ========================================'
       if (rank==0) then
          write(*,'(a,i5,a,3f10.4)') '[',rank,'] Computational domain (geom) dimensions : ', domainLength
          write(*,'(a,i5,a,3d20.10)') '[',rank,'] Space step : ', grid_step
          write(*,'(a,i5,a,f10.4)') '[',rank,'] Reynolds : ', Re
          write(*,'(a,i5,a,f10.4,3f10.4)') '[',rank,'] Sphere radius and position : ', sphere_radius,sphere_pos
          write(*,'(a,i5,a,f10.4)') '[',rank,'] The required flow rate is : ', reqFlowRate
          write(*,'(a,i5,a,i5)') '[',rank,'] Number of points in the boundary layer: ', nbpoints
       end if
       write(*,'(a,i5,a,3f10.5)') '[',rank,'] Current subdomain (geom) position : ', coordMin
       write(*,'(a,i5,a,3i10)') '[',rank,'] Current subdomain resolution : ', localResolution
       write(*,'(a,i5,a)') '[',rank,'] ========================================================================================='
    end if
    
    write (*,'(a,i5,a)') '[',rank,'] --------------------------------------------------------------> end of initialization'

  end subroutine init_client
 
  subroutine main_client() bind(c,name='NavierStokes2D')

    real(mk) :: initial_time,final_time
    real(mk) :: Re
    real(mk) :: t1
    real(mk), dimension(dim3) :: coordMin
    !> local mesh resolution
    integer, dimension(dim3) :: localResolution

    real(mk) :: shift = 0.0
    real(mk),dimension(dime)::coord
    integer :: i,info,j
    character(len=30) :: buffer
    real(mk),dimension(dime)::center
    Re = 133.4! 210.6
    initial_time = 0.0
    final_time = 100.0
   ! Initialisation stuff 
    t1 = MPI_WTIME()
    call init_client(Re,localResolution,coordMin)
    write (*,'(a,i5,a,f10.5)')  '[',rank,'] Initialisation time: ', MPI_WTIME()-t1

    
    do j=1,localResolution(c_Y)
       coord(c_Y) = coordMin(c_Y) + (j-1)*grid_step(c_Y)
       do i =1,localResolution(c_X)
          coord(c_X) = coordMin(c_X) + (i-1)*grid_step(c_X)
          vorticity2D(i,j,1) = -4.*pi*pi*cos(2.*pi*coord(c_X)/domainLength(c_X))/(domainLength(c_X)**2)! + cos(2.*pi*coord(c_Y)/domainLength(c_Y)) 
       end do
    end do

    PPMvelocity2D(c_Y,:,:,:) = 0.0
    !rhs(:,:,:,:) = velocity(:,:,:,:)
    !vorticity2D(:,:,:) = 1.0!velocity(:,:,:,:)
    !vorticity2D(1:localResolution(c_X)-1,1:localResolution(c_Y)-1,1) = 0.0
    !print *, "N2",norm2(velocity,localResolution,grid_step)!)velocity(c_X,:,:,:)
    !print *, "N2 v", norm2(vorticity2D,localResolution,grid_step)!)vorticity(c_X,:,:,:)
    call solvePoisson(vorticity2D,velocity,grid_resolution,topo%ID,meshNum)
    print *, "end of poisson solve"
!!$    call MPI_BARRIER(MPI_COMM_WORLD,info)
!!$
!!$    print *, "post N2",norm2(velocity,localResolution,grid_step)!)velocity(c_X,:,:,:)
!!$    !print *, "post N2 v", norm2(vorticity2D,localResolution,grid_step)!)vorticity(c_X,:,:,:)
!!$    do i = 1, localResolution(c_X)
!!$       print *,"------------", velocity(c_X,i,:,1)
!!$       print *,"************", vorticity2D(i,:,1)
!!$    end do
!!$
!!$
    
    !vorticity2D = vorticity - rhs
    !print *, "results :", norm1(vorticity,localResolution,grid_step)




!    call Mpi_barrier(MPI_COMM_WORLD,info)


    write(buffer,*) rank
    buffer = adjustl(buffer)
    buffer = "gauss"//buffer
    open(43,file=buffer) ! Output only for one process
    do j=1,localResolution(c_Y)
       coord(c_Y) = coordMin(c_Y) + (j-1)*grid_step(c_Y)
    !   do i =1,localResolution(c_X)
          !coord(c_X) = coordMin(c_X) + (i-1)*grid_step(c_X)
       write(43,'(4e14.5)') coord(c_Y),velocity(c_X,5,j,1) 
     !  end do
    end do
    
   close(43)
!!$    shift = 0!-500.*1e-3
!!$    ! init a Gaussian
!!$    center = 0.0
!!$    !call Gaussian1D(scalar,localResolution,grid_step,coordMin,c_X,shift)
!!$!!    call Gaussian2D(scalar,localResolution,grid_step,coordMin,center)
!!$    center(c_X) = 100.*1e-3
    !shift = 100.*1e-3
    !    call Gaussian1D(gauss,localResolution,grid_step,coordMin,c_X,shift)
!!$    call Gaussian2D(gauss,localResolution,grid_step,coordMin,center)
!!$    call ppm_map_field_ghost_get(topo%ID,meshNum,domain_ghostsize,info)
!!$    call ppm_map_field_push(topo%ID,meshNum,PPMscalar2D,info)
!!$    call ppm_map_field_send(info)
!!$    call ppm_map_field_pop(topo%ID,meshNum,PPMscalar2D,domain_ghostsize,info)
    ! Time loop 
!    call timeLoop(initial_time,final_time,Re,localResolution,coordMin)
!!$    rhs = 0.0
!!$    rhs(1,:,:,:) = gauss - scalar
!!$    
!!$    print *, rank,"norms ", norm1(rhs,localResolution,grid_step),norm2(rhs,localResolution,grid_step),normInf(rhs,localResolution)
!!$    !print *, "jzaiziaui", rank,rhs(1,65,12,1),rhs(1,1,12,1)
!!$    !do j = 1, localResolution(c_Y)
!!$    j = 43
!!$       coord(c_Y) =coordMin(c_Y) + (j-1)*grid_step(c_Y) 
!!$       do i = 1,localResolution(c_X)
!!$          coord(c_X) = coordMin(c_X) + (i-1)*grid_step(c_X)
!!$          if(gauss(i,j,1)< 1e-5) gauss(i,j,1)=0.0
!!$          if(scalar(i,j,1)< 1e-5) scalar(i,j,1)=0.0
!!$          write(43,'(4e14.5)') coord(c_X),coord(c_Y),real(gauss(i,j,1)),real(PPMscalar2D(i,j,1))
!!$       end do
    !end do

!!$    print *,"max diff", maxloc(rhs),maxval(rhs)
!!$    close(43)
    !call printToVTK("gauss",0,gauss,vorticity,localResolution,grid_step,coordMin)
    
    ! Close everything related to ppm
    call ppm_finalize(info)
    write (*,'(a,i5,a)')  '[',rank,'] ==================================== End of simulation. ===================================='

  end subroutine main_client


  !> Simulation over time ...
  subroutine timeLoop(initial_time,final_time,Re,resolution,coordMin)

    !> Starting time
    real(mk), intent(in) :: initial_time
    !> Ending time
    real(mk), intent(in) :: final_time
    !> Reynolds number
    real(mk), intent(in) :: Re
    !> Local mesh resolution
    integer, dimension(dim3) :: resolution
    !> coordinates of the lowest point of the domain
    real(mk),dimension(dim3),intent(in) :: coordMin

    real(mk) :: current_time,elapsed_time,time_step
    integer :: iter,i,j,k
    logical :: resetpos
    real(mk) :: nu,dtMax
    !    real(mk), dimension(:,:), pointer :: err1, err2, errinf
    ! diagnostics 1:3 -> drag/lifts "porous" drag, 4:6 Winkelmans paper method
    !    real(mk), dimension(dime) :: diagnostics 
    real(mk),dimension(dime)::force

    real(mk)::dvol,x,z,y
    real(mk),dimension(dime) :: norm2Vel,nref
    integer :: info
    integer,parameter :: maxiter = 1
    dvol=product(grid_step)
    nu = 1./Re!0.5*FolkeRatio/Re!*uinf
    current_time = initial_time
    iter = 1
    
    ! Max value allowed for the time step
    dtMax = 1e-3!0.01!grid_step(1)**2/(6.*nu)
    print *, "dtmax", dtMax
    ! offset for velocity to set a desired average value for flow rate.
    ! Initial time step
    time_step=dtMax

    if(rank==0) open(10,file='diagnostics') ! Output only for one process
    
    !! Compute a first distribution of velocity and vorticity
    !call computeInitialFields(time_step,resolution,grid_step,coordMin)
!
    ! Synchronise all procs after initialization process ...
    call MPI_BARRIER(MPI_COMM_WORLD,info)
    write (*,'(a,i5,a)')  '[',rank,'] -------------------------------------------------------------->  start simulation'
    
    resetpos = .TRUE.
    !current_time=final_time
    !do while(current_time <= final_time)
    do while(iter <= maxiter)
       elapsed_time = MPI_Wtime()
       ! Perturbation step, only required for high Reynolds number 
       ! if( current_time>3.0 .and. current_time<4.0) then
       !   call perturb(velocity, current_time)
       ! end if
       
       !============================================================ 
       ! Compute the diagnostics, Noca method
       !============================================================ 
       !call nocaForces(force,velocity,vorticity,nu,coordMin,grid_step,time_step,dvol)
       
       if(rank == 0) then
          write(*,'(i5,a,3f10.5)') iter, ' drag: ', force
          write(10,'(11e14.5)') current_time,force
       end if

       !============================================================ 
       ! Solve Navier-Stokes using particular method
       !============================================================ 
       call PPMupdateParticles2DScalar(PPMscalar2D,.true.,topo%ID,meshNum,PPMvelocity2D)
       call RK2_2DScalar(time_step,topo%ID,meshNum,domain_ghostsize,PPMvelocity2D)
       call PPMremesh2DScalar(topo%ID,meshNum,domain_ghostsize,PPMscalar2D)
         
       !============================================================ 
       ! Solve Poisson for the new vorticity --> velocity
       !============================================================ 
       ! Compute velocity from vorticity
       ! Two steps:
       ! - solve Poisson for stream_function and update velocity from stream_function : everything is supposed
       ! to be done in ppm routine; indeed we do not have to deal with stream_function.
       ! - update velocity to fit with a required flow rate
       ! Solve poisson to find velocity 
!!$       call solve_poisson(vorticity,velocity,topo%ID,meshNum,domain_ghostsize)
!!$       call ppm_map_field_ghost_get(topo%ID,meshNum,domain_ghostsize,info)
!!$       call ppm_map_field_push(topo%ID,meshNum,velocity,3,info)
!!$       call ppm_map_field_send(info)
!!$       call ppm_map_field_pop(topo%ID,meshNum,velocity,3,domain_ghostsize,info)
!!$       print *, "POST poisson", minval(vorticity),maxval(vorticity),minval(velocity),maxval(velocity)
!!$       print *, 'veloc', sum(velocity(c_X,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(velocity(c_Y,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(velocity(c_Z,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1))
!!$       print *, 'vort',sum(vorticity(c_X,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(vorticity(c_Y,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(vorticity(c_Z,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1))

!!$
!!$       call shiftVelocityX(reqFlowRate,grid_step,resolution,domainLength(c_Y)*domainLength(c_Z),coordMin,physDomainLowerPoint)
!!$       call ppm_map_field_ghost_get(topo%ID,meshNum,domain_ghostsize,info)
!!$       call ppm_map_field_push(topo%ID,meshNum,velocity, 3, info)
!!$       call ppm_map_field_send(info)
!!$       call ppm_map_field_pop(topo%ID,meshNum,velocity,3,domain_ghostsize,info)

!!$       print *, "POST shift", minval(vorticity),maxval(vorticity),minval(velocity),maxval(velocity)
!!$       print *, 'veloc', sum(velocity(c_X,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(velocity(c_Y,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(velocity(c_Z,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1))
!!$       print *, 'vort',sum(vorticity(c_X,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(vorticity(c_Y,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(vorticity(c_Z,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1))

       !============================================================ 
       ! Penalize velocity on the sphere and on the boundaries
!!$       !============================================================ 
!!$       call penalise_velocity(velocity,time_step,chi_sphere,chi_boundary)
!!$
!!$       call ppm_map_field_ghost_get(topo%ID,meshNum,domain_ghostsize,info)
!!$       call ppm_map_field_push(topo%ID,meshNum,velocity, 3, info)
!!$       call ppm_map_field_send(info)
!!$       call ppm_map_field_pop(topo%ID,meshNum,velocity,3,domain_ghostsize,info)
!!$       
!!$       print *, "POST penal", minval(vorticity),maxval(vorticity),minval(velocity),maxval(velocity)
!!$       print *, 'veloc', sum(velocity(c_X,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(velocity(c_Y,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(velocity(c_Z,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1))
!       call printToVTK("post-penal",iter,velocity,vorticity,resolution,grid_step,coordMin)
       
       
       !============================================================ 
       ! Compute the new "penalized" vorticity
       !============================================================ 
!!$       call curlDF4(velocity,vorticity,resolution,grid_step)
!!$       call ppm_map_field_ghost_get(topo%ID,meshNum,domain_ghostsize,info)
!!$       call ppm_map_field_push(topo%ID,meshNum,vorticity,3,info)
!!$       call ppm_map_field_send(info)
!!$       call ppm_map_field_pop(topo%ID,meshNum,vorticity,3,domain_ghostsize,info)
!!$       
!!$
       !============================================================ 
       ! Compute stretch/diffusion from current velocity/vorticity
       !============================================================ 
!!$       call computeRHS(velocity,vorticity,rhs,resolution,grid_step,nu)
!!$       !!call computeStretch(velocity,vorticity,rhs,resolution,grid_step)
!!$       call ppm_map_field_ghost_get(topo%ID,meshNum,domain_ghostsize,info)
!!$       call ppm_map_field_push(topo%ID,meshNum,rhs,3,info)
!!$       call ppm_map_field_send(info)
!!$       call ppm_map_field_pop(topo%ID,meshNum,rhs,3,domain_ghostsize,info)
       
!!$       print *, 'veloc', sum(velocity(c_X,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(velocity(c_Y,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(velocity(c_Z,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1))
!!$       print *, 'vort',sum(vorticity(c_X,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(vorticity(c_Y,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(vorticity(c_Z,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1))
!!$       print *, 'stretch',sum(rhs(c_X,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(rhs(c_Y,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1)),&
!!$            sum(rhs(c_Z,1:resolution(c_X)-1,1:resolution(c_Y)-1,1:resolution(c_Z)-1,nsublist))&
!!$            /((resolution(c_X)-1)*(resolution(c_Y)-1)*(resolution(c_Z)-1))
!!$       print *, 'max v...', maxval(velocity(c_X,:,:,:,:)), maxval(velocity(c_Y,:,:,:,:)), maxval(velocity(c_Z,:,:,:,:))
!!$       print *, 'min v...', minval(velocity(c_X,:,:,:,:)), minval(velocity(c_Y,:,:,:,:)), minval(velocity(c_Z,:,:,:,:))
!!$       print *, 'max w...', maxval(vorticity(c_X,:,:,:,:)), maxval(vorticity(c_Y,:,:,:,:)), maxval(vorticity(c_Z,:,:,:,:))
!!$       print *, 'min w...', minval(vorticity(c_X,:,:,:,:)), minval(vorticity(c_Y,:,:,:,:)), minval(vorticity(c_Z,:,:,:,:))
!!$       print *, 'max s...', maxval(rhs(c_X,:,:,:,:)), maxval(rhs(c_Y,:,:,:,:)), maxval(rhs(c_Z,:,:,:,:))
!!$       print *, 'min s...', minval(rhs(c_X,:,:,:,:)), minval(rhs(c_Y,:,:,:,:)), minval(rhs(c_Z,:,:,:,:))
!!$
!!$



!!$       print *, 'indices ...', maxloc(vorticity(:,:,:,:,:)),'toto',minloc(vorticity(c_Y,:,:,:,:))
!!$
!!$
!!$       print *, 'vort', vorticity(c_Y,-1,-1,4,1)
!!$       print *, 'vel', velocity(c_X,1,1,4,1), velocity(c_X,1,1,128,1)
!!$       print *, 'vel +1', velocity(c_X,1,1,5,1), velocity(c_X,1,1,129,1)
!!$       print *, 'vel', velocity(c_X,1,1,6,1), velocity(c_X,1,1,130,1)
!!$       print *, 'vel -1', velocity(c_X,1,1,3,1), velocity(c_X,1,1,127,1)
!!$       print *, 'vel', velocity(c_X,1,1,2,1), velocity(c_X,1,1,126,1)
 
       
       ! Update time
       current_time = current_time + time_step
       
       if(verbose) then 
          !memoryUsed = memoryUsed + getMemoryUsedForParticles()
          write(*,'(a,i5,a,i10)') "[", rank,"] end of iter ",iter
          write(*,'(a,i5,a,f10.4)')  "[", rank,"] simulation time : ", MPI_WTime()-elapsed_time
          write(*,'(a,i5,a,f14.10)')  "[", rank,"] current time step : ", time_step
          write(*,'(a,i5,a,f10.4)')  "[", rank,"] Memory used : ",  memoryUsed
          write(*,'(a,i5,a,f14.10)')  "[", rank,"] current time : ", current_time
       end if

       ! Compute the new time step according to vorticity maximum value
       !    call updateTimeStep(time_step,vorticity,dtMax)
       
       ! Output every 20 time unit ...
!       if(mod(current_time,20.).lt.time_step) then
 !         call printToVTK("run",iter,velocity,vorticity,resolution,grid_step,coordMin)
  !     end if
       
       iter = iter+1
       !call MPI_BARRIER(MPI_COMM_WORLD,info)
    enddo
    
    if(rank==0) close(10)

  end subroutine timeLoop

  subroutine updateTimeStep(dt,ref_field,dtMax)
    
    real(mk), dimension(:,:,:,:,:), pointer :: ref_field
    real(mk), intent(inout) :: dt
    real(mk),intent(in)::dtMax

    real(mk) :: local_max,omega_max
    integer :: info

    local_max = maxval(ref_field(:,:,:,:,nsublist)) ! We ignore the ppm nsubs stuff ...
    call MPI_ALLReduce(local_max,omega_max,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)
    dt = min(0.25/omega_max,dtMax)
    
  end subroutine updateTimeStep




end module NavierStokes2D

