!> This module is used to run a simulation using 
!! ppm (core and numerics)
!!
module ppmNavierStokes3D
  
  ! All required ppm modules ...
  use ppm_module_init, only : ppm_init
  use ppm_module_data, only : ppm_kind_double
  use ppm_module_finalize, only : ppm_finalize
  use ppm_module_map_field_ghost, only:ppm_map_field_ghost_get, ppm_map_field_ghost_put
  use ppm_module_map_field
  use ppm_module_data, only: ppm_param_bcdef_periodic 

  ! use client_io
  use client_data
  ! some tools
  use parmesTools
  ! Physical domain and grid
  use Domain
  ! Fields on the grid
  use Fields, only: init_fields, velocity, vorticity, rhs,testFunc,getMemoryUsedForFields!, vel_ex!, stream_function
  ! Topology
  use client_topology, only: init_topo,topo,getPPMLocalResolution
  ! Poisson (ppm) solver
  use Solver, only : init_poisson_solver, solve_poisson, ppm_poisson_drv_none, ppm_poisson_drv_curl_fd2, ppm_poisson_drv_curl_fd4
  ! Penalisation stuff
  use penalisation, only : penalise_velocity
  ! Functions used to identify surfaces, volumes ... 
  use SetsIndicators,only:chi_sphere,chi_boundary,compute_control_box,init_boundary_layer,nocaForces,&
       compute_test,chi_box,getMemoryForIndicators
  ! curl, prod ...
  use vectorcalculus, only: curldf4, computeRHS, norm1, norm2, normInf
  ! everything dealing with particles
  use Particles, only : init_particles,update_particles,push_particles,remesh,getMemoryUsedForParticles
  ! file io 
  use io_vtk
  
  use mpi
  
  ! user-defined functions (for tests or fields initialization)
  use testsFunctions

  implicit none

  integer, private :: info 

  ! Some global vars ... 
  !> Sphere radius
  real(mk),private :: sphere_radius
  !> Sphere position
  real(mk),dimension(dime),private :: sphere_pos
  !> counter for memory usage
  real(mk) :: memoryUsed
  
contains

  !> All required initialisations :
  !! MPI, ppm
  !! Domain, grid
  !! Fields
  !! Particles
  subroutine init_client(Re,localResolution,coordMin)

    use client_topology, only: isubl
    !> Reynolds numner
    real(mk),intent(in) ::Re
    !> local mesh resolution
    integer, dimension(dime),intent(out) :: localResolution
    !> Local (proc.) minimum point coordinates
    real(mk),dimension(dime),intent(out) ::coordMin

    ! Precision and tolerance (for ppm init)
    integer :: prec,tol
    ! MPI comm
    integer :: comm
    ! debug mode (for ppm)
    integer :: debug
    

    !> Dimensions and positions of the control volume
    real(mk),dimension(dime):: boxMin,boxMax

    !> For ppm fftw solver
    integer :: derive

    memoryUsed = 0

    ! "read" mpi parameters
    call MPI_COMM_RANK(MPI_COMM_WORLD,rank,info)
    call MPI_COMM_SIZE(MPI_COMM_WORLD,nbprocs,info)
    comm = MPI_COMM_WORLD  
    
    write (*,'(a,i5,a)') '[',rank,'] --------------------------------------------------------------> start initialization '

    !======================
    ! Init ppm 
    !======================
    prec = ppm_kind_double ! Defined in ppm_param.h
    debug = 0
    tol = -10
    info = -1
    call ppm_init(dime,prec,tol,comm,debug,info)
    
    !======================
    ! Geometry and grid
    !======================
    ! Set domain size, resolution, boundaries, number of ghost points ... 
    ! At the time, many values are set in Domain.f90 file
    call init_geometry(ppm_param_bcdef_periodic)
    call init_grid()

    !======================
    ! Creates the topology
    !======================
    ! Based on ppm. 
    call init_topo(falseLowerPoint,falseUpperPoint,domain_bc,domain_ghostsize,grid_resolution)
    
    !> Get the coordinates of the lowest point for the current domain
    coordMin(:)=topo%min_subd(:,isubl) 

    !======================
    ! Fields allocation
    !======================
    ! Local number of nodes
    localResolution = getPPMLocalResolution()
    ! Allocate fields (velocity, vorticity ...)
    call init_fields(localResolution,domain_ghostsize)
    
    !=======================================================
    ! Compute a control volume for diagnostics computation
    !=======================================================
    !! Position of the upper and lower points of the box
    boxMin=falseLowerPoint!+1.0!15*grid_step
    boxMax=falseUpperPoint!15*grid_step
    !print *, rank,'sizes',boxMax-boxMin
    !print *, rank, 'sx', 0.5*(boxMax(3)-boxMin(3))*(boxMax(2)-boxMin(2)),(boxMax(3)-boxMin(3))*(boxMax(2)-boxMin(2))
    !print *, rank, 'sy', 0.5*(boxMax(3)-boxMin(3))*(boxMax(1)-boxMin(1)),(boxMax(3)-boxMin(3))*(boxMax(1)-boxMin(1))
    !print *, rank, 'sz', 0.5*(boxMax(1)-boxMin(1))*(boxMax(2)-boxMin(2)),(boxMax(1)-boxMin(1))*(boxMax(2)-boxMin(2))
    !! Sphere radius and position
    sphere_radius=0.5
    sphere_pos=(upperPoint+lowerPoint)/2.0

    !! compute indicator functions for the control box and the sphere
    call compute_control_box(localResolution,grid_step,boxMin,boxMax,sphere_pos,sphere_radius,coordMin)
    !! Check :  == 1 inside the box (excluding the sphere), 0 elsewhere
    
    !=============================================================
    ! Compute indicator function for the boundaries in z dir 
    ! (will be used to enforce dirichlet conditions on these bound.)
    !=============================================================
    call init_boundary_layer(localResolution,grid_step,lowerPoint,upperPoint,coordMin)

    ! Test : export test function to vtk file.
    ! compute_test set 1 at all points belonging to chi_... 
    testFunc=0.0
    call compute_test(testFunc,chi_boundary)
    call compute_test(testFunc,chi_box)
    call printChiToVTK("box",testFunc,localResolution,grid_step,coordMin)
    !! display the memory used for indicator sets and fields
    memoryUsed=getMemoryUsedForFields()+getMemoryForIndicators()
    deallocate(testFunc)
   
    !================================
    ! Solvers (grid and particles) 
    !================================
    
    ! --- Grid solver for Poisson ---
    ! derive : means that finite difference, order 4, are used to compute velocity from stream function, in ppm
    derive = ppm_poisson_drv_curl_fd4 
    ! init fftw solver, from ppm
    call init_poisson_solver(vorticity,velocity,topo%ID,topo%mesh(1)%ID,derive)
    
    ! --- Particles ---
    ! Set particles parameters (kernel type, cutoff ...)
    call init_particles()
      
    if(verbose) then
       !   write(*,'(a,i5,a100)') '[',rank,'] ============================================= Summary ============================================='
       write(*,'(a,i5,a)') '[',rank,'] ======================================== Summary ========================================'
       if (rank==0) then
          write(*,'(a,i5,a,3f10.4)') '[',rank,'] Computational domain (geom) dimensions : ', lengths
          write(*,'(a,i5,a,3d20.10)') '[',rank,'] Space step : ', grid_step
          write(*,'(a,i5,a,f10.4)') '[',rank,'] Reynolds : ', Re
       end if
       write(*,'(a,i5,a,3f10.5)') '[',rank,'] Current subdomain (geom) position : ', coordMin
       write(*,'(a,i5,a,3i10)') '[',rank,'] Current subdomain resolution : ', localResolution
 !      write(*,'(a,i5,a)') '[',rank,'] ===================================================================================================='
       write(*,'(a,i5,a)') '[',rank,'] ========================================================================================='
    end if
    
    write (*,'(a,i5,a)') '[',rank,'] --------------------------------------------------------------> end of initialization'

  end subroutine init_client
  
  subroutine computeInitialFields(fit_velo,dt,resolution,step,coordMin)
    !> offset value for velocity to fit with a given flow rate
    real(mk),intent(in) :: fit_velo
    !> current time step
    real(mk),intent(in) ::dt
    !> local mesh resolution
    integer,dimension(dime),intent(in) :: resolution
    !> mesh step sizes
    real(mk),dimension(dime),intent(in)::step
    !> Local (proc.) minimum point coordinates
    real(mk),dimension(dime),intent(in) ::coordMin
    ! set velocity to 0. and compute a full vorticity distribution, such that
    ! Omega_y = -3z/(Lz^2)
    velocity = 0.0
    call init_vorticity(vorticity,resolution,step,coordMin,lowerPoint,upperPoint)
    ! Init ghost values for vorticity
    call ppm_map_field_ghost_get(topo%ID,topo%mesh(1)%ID,domain_ghostsize,info)
    call ppm_map_field_push(topo%ID,topo%mesh(1)%ID,vorticity, 3, info)
    call ppm_map_field_send(info)
    call ppm_map_field_pop(topo%ID,topo%mesh(1)%ID,vorticity,3,domain_ghostsize,info)
    ! Solve poisson for velocity
    call solve_poisson(vorticity,velocity,topo%ID,topo%mesh(1)%ID,domain_ghostsize)
    velocity(c_X,:,:,:,:) = velocity(c_X,:,:,:,:) + fit_velo
    ! Penalize to take sphere and boundaries into account
    call penalise_velocity(velocity,dt,chi_sphere,chi_boundary)
    call ppm_map_field_ghost_get(topo%ID,topo%mesh(1)%ID,domain_ghostsize,info)
    call ppm_map_field_push(topo%ID,topo%mesh(1)%ID,velocity, 3, info)
    call ppm_map_field_send(info)
    call ppm_map_field_pop(topo%ID,topo%mesh(1)%ID,velocity,3,domain_ghostsize,info)
    ! Compute and map the new "penalized" vorticity
    call curlDF4(velocity,vorticity,resolution,step)
    call ppm_map_field_ghost_get(topo%ID,topo%mesh(1)%ID,domain_ghostsize,info)
    call ppm_map_field_push(topo%ID,topo%mesh(1)%ID,vorticity,3,info)
    call ppm_map_field_send(info)
    call ppm_map_field_pop(topo%ID,topo%mesh(1)%ID,vorticity,3,domain_ghostsize,info)
    
  end subroutine computeInitialFields

  subroutine main_client() bind(c,name='NavierStokes3D')

    real(mk) :: initial_time,final_time
    real(mk) :: Re
    real(mk) :: t1
    real(mk), dimension(dime) :: coordMin
    !> local mesh resolution
    integer, dimension(dime) :: localResolution
    Re = 133.4! 210.6
    initial_time = 0.0
    final_time = initial_time!20.
    
    ! Initialisation stuff 
    t1 = MPI_WTIME()
    call init_client(Re,localResolution,coordMin)
    write (*,'(a,i5,a,f10.5)')  '[',rank,'] Initialisation time: ', MPI_WTIME()-t1
    
    ! Time loop 
    call timeLoop(initial_time,final_time,Re,localResolution,coordMin)

    ! Close everything related to ppm
    call ppm_finalize(info)
    write (*,'(a,i5,a)')  '[',rank,'] ================== End of simulation. =================='

  end subroutine main_client


  !> Simulation over time ...
  subroutine timeLoop(initial_time,final_time,Re,resolution,coordMin)

    !> Starting time
    real(mk), intent(in) :: initial_time
    !> Ending time
    real(mk), intent(in) :: final_time
    !> Reynolds number
    real(mk), intent(in) :: Re
    !> Local mesh resolution
    integer, dimension(dime) :: resolution
    !> coordinates of the lowest point of the domain
    real(mk),dimension(dime),intent(in) :: coordMin

    real(mk) :: current_time,elapsed_time,time_step
    integer :: iter
    logical :: resetpos
    real(mk) :: fit_velo,nu,dtMax
    !    real(mk), dimension(:,:), pointer :: err1, err2, errinf
    ! diagnostics 1:3 -> drag/lifts "porous" drag, 4:6 Winkelmans paper method
    !    real(mk), dimension(dime) :: diagnostics 
    real(mk),dimension(dime)::force

    real(mk)::dvol

    dvol=product(grid_step)
    nu =1./Re
    current_time = initial_time
    iter = 1

    ! Max value allowed for the time step
    dtMax = grid_step(1)**2/(6.*nu)
    
    ! offset for velocity to set a desired average value for flow rate.
    fit_velo =  (upperPoint(c_Z)-lowerPoint(c_Z))/lengths(c_Z)
    ! Initial time step
    time_step=dtMax

    if(rank==0) open(10,file='diagnostics') ! Output only for one process

    !! Compute a first distribution of velocity and vorticity
    call computeInitialFields(fit_velo,time_step,resolution,grid_step,coordMin)
    call printToVTK("init",iter,velocity,vorticity,topo,grid_step)
    force=0.0
    call nocaForces(force,velocity,vorticity,1./Re,coordMin,grid_step,time_step,dvol)
    if(rank==0) print*, 'Initial drag ', force
    ! Synchronise all procs after initialization process ...
    call MPI_BARRIER(MPI_COMM_WORLD,info)
    write (*,'(a,i5,a)')  '[',rank,'] -------------------------------------------------------------->  start simulation'
    resetpos = .TRUE.
    do while(current_time <= final_time)
       elapsed_time = MPI_Wtime()
       ! Perturbation step, only required for high Reynolds number 
       ! if( current_time>3.0 .and. current_time<4.0) then
       !   call perturb(velocity, current_time)
       ! end if
       
       !============================================================ 
       ! Compute stretch/diffusion from current velocity/vorticity
       !============================================================ 


       print *, rank, 'vel X', velocity(1,10,10,-1:3,1)
       print *, rank, 'vel X', velocity(1,10,10,resolution(3)-2:resolution(3)+2,1)
       print *, rank, 'vort X', vorticity(1,10,10,-1:3,1)
       print *, rank, 'vort X', vorticity(1,10,10,resolution(3)-2:resolution(3)+2,1)

       call computeRHS(velocity,vorticity,rhs,resolution,grid_step,nu)
       print *, rank, 'rhs pre', minval(rhs),maxval(rhs)
       print *, rank, 'rhs preX', rhs(1,10,10,-1:3,1)
       print *, rank, 'rhs preX', rhs(1,10,10,resolution(3)-2:resolution(3)+2,1)
!       print *, rank, 'rhs preY', rhs(2,10,10,-1:3,1),rhs(2,10,10,resolution(3)-2:resolution(3)+2,1)
 !      print *, rank, 'rhs pre2', rhs(3,10,10,-1:3,1),rhs(3,10,10,resolution(3)-2:resolution(3)+2,1)
       
       call ppm_map_field_ghost_get(topo%ID,topo%mesh(1)%ID,domain_ghostsize,info)
       call ppm_map_field_push(topo%ID,topo%mesh(1)%ID,rhs,3,info)
       call ppm_map_field_send(info)
       call ppm_map_field_pop(topo%ID,topo%mesh(1)%ID,rhs,3,domain_ghostsize,info)
          
       print *, rank, 'rhs', minval(rhs),maxval(rhs)
       print *, rank, 'rhs postX', rhs(1,10,10,-1:3,1)
       print *, rank, 'rhs postX', rhs(1,10,10,resolution(3)-2:resolution(3)+2,1)
  !     print *, rank, 'rhs postY', rhs(2,10,10,-1:3,1),rhs(2,10,10,resolution(3)-2:resolution(3)+2,1)
   !    print *, rank, 'rhs post2', rhs(3,10,10,-1:3,1),rhs(3,10,10,resolution(3)-2:resolution(3)+2,1)

       !============================================================ 
       ! Solve Navier-Stokes using particular method
       !============================================================ 
       ! Initialize the particles according to the last computed vorticity, velocity and rhs
       call update_particles(vorticity,resetpos,topo%ID,topo%mesh(1)%ID,velocity)
       ! Integrate
       call push_particles(time_step,topo%ID,topo%mesh(1)%ID,domain_ghostsize,velocity,rhs)
       ! Remesh
       call remesh(topo%ID, topo%mesh(1)%ID,domain_ghostsize,vorticity)
       ! Ghost values for vorticity
       call ppm_map_field_ghost_get(topo%ID,topo%mesh(1)%ID,domain_ghostsize,info)
       call ppm_map_field_push(topo%ID,topo%mesh(1)%ID,vorticity, 3, info)
       call ppm_map_field_send(info)
       call ppm_map_field_pop(topo%ID,topo%mesh(1)%ID,vorticity,3,domain_ghostsize,info)

       !============================================================ 
       ! Solve Poisson for the new vorticity --> velocity
       !============================================================ 
       ! Compute velocity from vorticity
       ! Two steps:
       ! - solve Poisson for stream_function and update velocity from stream_function : everything is supposed
       ! to be done in ppm routine; indeed we do not have to deal with stream_function.
       ! - update velocity to fit with required boundary values
       ! Solve poisson to find velocity 
       call solve_poisson(vorticity,velocity,topo%ID,topo%mesh(1)%ID,domain_ghostsize)
       velocity(c_X,:,:,:,:) = velocity(c_X,:,:,:,:) + fit_velo

       !============================================================ 
       ! Penalize velocity on the sphere and on the boundaries
       !============================================================ 
       call penalise_velocity(velocity,time_step,chi_boundary,chi_sphere)
       call ppm_map_field_ghost_get(topo%ID,topo%mesh(1)%ID,domain_ghostsize,info)
       call ppm_map_field_push(topo%ID,topo%mesh(1)%ID,velocity, 3, info)
       call ppm_map_field_send(info)
       call ppm_map_field_pop(topo%ID,topo%mesh(1)%ID,velocity,3,domain_ghostsize,info)
       
       !============================================================ 
       ! Compute the new "penalized" vorticity
       !============================================================ 
       call curlDF4(velocity,vorticity,resolution,grid_step)
       call ppm_map_field_ghost_get(topo%ID,topo%mesh(1)%ID,domain_ghostsize,info)
       call ppm_map_field_push(topo%ID,topo%mesh(1)%ID,vorticity,3,info)
       call ppm_map_field_send(info)
       call ppm_map_field_pop(topo%ID,topo%mesh(1)%ID,vorticity,3,domain_ghostsize,info)

       !============================================================ 
       ! Compute the diagnostics, Noca method
       !============================================================ 
       call nocaForces(force,velocity,vorticity,1./Re,coordMin,grid_step,time_step,dvol)
       
       if(rank == 0) then
          write(*,'(i5,a,3f10.5)') iter, ' drag: ', force
          write(10,'(11e14.5)') current_time,force
       end if

       if(verbose) then 
          memoryUsed = memoryUsed + getMemoryUsedForParticles()
          write(*,'(a,i5,a,i10)') "[", rank,"] end of iter ",iter
          write(*,'(a,i5,a,f10.4)')  "[", rank,"] simulation time : ", MPI_WTime()-elapsed_time
          write(*,'(a,i5,a,f10.5)')  "[", rank,"] current time step : ", time_step
          write(*,'(a,i5,a,f10.4)')  "[", rank,"] Memory used : ",  memoryUsed
       end if
       
       ! Update time
       current_time = current_time + time_step
       
       ! Compute the new time step according to vorticity maximum value
       call updateTimeStep(time_step,vorticity,dtMax)
       
       ! Output every 10 time unit ...
       if(mod(current_time,20.).lt.time_step) then
          call printToVTK("run",iter,velocity,vorticity,topo,grid_step)
       end if
       
       !current_time = final_time
       iter = iter+1
       call MPI_BARRIER(MPI_COMM_WORLD,info)
    enddo
     
    if(rank==0) close(10)

  end subroutine timeLoop

  subroutine updateTimeStep(dt,ref_field,dtMax)
    
    real(mk), dimension(:,:,:,:,:), pointer :: ref_field
    real(mk), intent(inout) :: dt
    real(mk),intent(in)::dtMax

    real(mk) :: local_max,omega_max
    
    local_max = maxval(ref_field(:,:,:,:,1)) ! We ignore the ppm nsubs stuff ...
    call MPI_Reduce(local_max,omega_max,1,MPI_DOUBLE_PRECISION,MPI_MAX,0,MPI_COMM_WORLD,info)
    if(rank == 0) then
       dt = min(0.25/omega_max,dtMax)
    end if
    call MPI_bcast(dt,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,info)
  end subroutine updateTimeStep

end module ppmNavierStokes3D
