!> Declaration, allocation of all the fields on the grid.
module Fields
  
  use client_data
  use mpi
  use PPMFields

  implicit none

  !> Velocity 
  real(mk), dimension(:,:,:,:), pointer :: velocity => NULL()
  !> Vorticity
  real(mk), dimension(:,:,:,:), pointer :: vorticity =>NULL()
  !> Vorticity (a scalar in the 2D case)
  real(mk), dimension(:,:,:), pointer :: vorticity2D =>NULL()
  !> Stream function - Test purpose. Useless if ppm fft solver works as it is supposed to ...
  ! real(mk), dimension(:,:,:,:,:), pointer :: stream_function =>NULL()
  !> rhs of vorticity eq (i.e. stretch + diffusion terms)
  real(mk), dimension(:,:,:,:), pointer :: rhs =>NULL()
  real(mk), dimension(:,:,:), pointer :: gauss =>NULL()
  !> 
  !real(mk), dimension(:,:,:,:,:), pointer :: vel_ex => NULL()
  !> Scalar on the grid, test purpose for chi functions
  real(mk),dimension(:,:,:),pointer::testFunc=>NULL()
  !> Scalar on the grid
  real(mk),dimension(:,:,:),pointer::scalar=>NULL()
  
contains
  
  !> Fields allocation. 
  !! Warning : ghostpoints must be included in field (i.e. size = "real size" + 2*number of ghostpoints)
  subroutine initFields(resolution,ghostsize)
    
    !> Required resolution for the fields (without ghosts)
    integer, dimension(dime), intent(in) :: resolution
    !> number of ghost points in each direction (ghost(c_X) = 2 means resolution(c_X)+ 4 points for the field)
    integer, dimension(:),pointer:: ghostsize
    
    integer::istat
    ! Lower and upper bounds for fields
    integer, dimension(dime) :: ldl, ldu
    ! nsublist from ppm topo. Assumed to be equal to 1 see Topology.f90

    ldl = 1 - ghostsize
    ldu = resolution + ghostsize


    if(dime==2) then
       call initPPMFields2D(resolution,ghostsize)
       velocity => PPMvelocity2D
       vorticity2D => PPMvorticity2D
       scalar => PPMscalar2D
       allocate(gauss(ldl(c_X):ldu(c_X),ldl(c_Y):ldu(c_Y),1), stat = istat)
       allocate(rhs(dime,ldl(c_X):ldu(c_X),ldl(c_Y):ldu(c_Y),1), stat = istat)
       
    else if(dime == 3) then
       call initPPMFields3D(resolution,ghostsize)
       velocity => PPMvelocity3D(:,:,:,:,1)
       vorticity => PPMvorticity3D(:,:,:,:,1)
       rhs => PPMrhs3D(:,:,:,:,1)
       allocate(gauss(ldl(c_X):ldu(c_X),ldl(c_Y):ldu(c_Y),ldl(c_Z):ldu(c_Z)), stat = istat)
    end if

    allocate(testFunc(resolution(c_X),resolution(c_Y),resolution(c_Z)))


!!$    ! Velocity ...
!!$    allocate(velocity(dime,ldl(1):ldu(1),ldl(2):ldu(2),ldl(3):ldu(3),nsublist),stat = istat)
!!$    if(istat.ne.0) stop 'Field allocation error for velocity'
!!$    ! Vorticity
!!$    allocate(vorticity(dime,ldl(1):ldu(1),ldl(2):ldu(2),ldl(3):ldu(3),nsublist), stat = istat)
!!$    if(istat.ne.0) stop 'Field allocation error for vorticity'
!!$    ! rhs
!!$    allocate(rhs(dime,ldl(1):ldu(1),ldl(2):ldu(2),ldl(3):ldu(3),nsublist), stat = istat)
 
!!$    if(istat.ne.0) stop 'Field allocation error for rhs'
!!$    
!!$    !allocate(stream_function(dime,ldl(1):ldu(1),ldl(2):ldu(2),ldl(3):ldu(3),nsublist), stat = istat)
!!$    !if(istat.ne.0) stop 'stream_function allocation error for rhs'
    
!!    allocate(testFunc(ldl(1):ldu(1),ldl(2):ldu(2),ldl(3):ldu(3)), stat = istat)
 !!   if(istat.ne.0) stop 'Field allocation error for testFunc'
    ! Scalar
    !allocate(scalar(ldl(1):ldu(1),ldl(2):ldu(2),ldl(3):ldu(3),nsublist), stat = istat)
    !if(istat.ne.0) stop 'Field allocation error for scalar'
  end subroutine initFields

  !> compute the size of the memory used to save fields
  function getMemoryUsedForFields()
    real(mk) :: getMemoryUsedForFields 
    getMemoryUsedForFields = sizeof(velocity)+sizeof(vorticity)+sizeof(rhs)+sizeof(testfunc)+sizeof(scalar)
    getMemoryUsedForFields = getMemoryUsedForFields*1.e-6
    if(verbose) then
       write(*,'(a,i3,a,f10.4,a)') &
            '[',rank,'] Fields have been initialized.  Memory used :', getMemoryUsedForFields, ' MB.'
    end if
  end function getMemoryUsedForFields

  !> Shift the velocity x component according to a required flow rate
  !> \f[ velocity_x = velocity_x + shift \f]
  !> with
  !> \f[shift = \frac{reqFlowRate-currentFlowRate}{S_x} \f]
  !> \f$S_x\f$ being the surface where x=xmin (incoming flow)
  subroutine shiftVelocityX(reqFlowRate,step,resolution,surf,coordMin,lower)
    !> Required flow rate
    real(mk), intent(in) :: reqFlowRate
    !> Grid step
    real(mk), dimension(dime), intent(in) :: step
    !> local resolution
    integer,dimension(dime),intent(in) :: resolution
    !> Area of the surface for integration 
    real(mk) ,intent(in) :: surf
    !> Coordinates of the minimal point of the current domain
    real(mk),dimension(dime),intent(in) :: coordMin
    !> lower bound of the physical domain
    real(mk),dimension(dime),intent(in) :: lower

    real(mk) :: localShift,globalShift
    integer :: info
    localShift = 0 
    
    ! We compute the current flow rate through surface x = xmin, for the x component of the velocity : 
    ! FlowRate = sum(surf x=xmin)( velocity(c_X) ) * step(c_Z)*step(c_Y)
    ! And use this value to shift the velocity(c_X) to have :
    ! RequiredFlowRate = FlowRate + Surf(x=xmin)*globalShift
 
    ! Step 1 : compute flowRate
    if(abs(coordMin(c_X)-lower(c_X)) <= 2.*epsilon(globalShift) ) then
       ! Compute mean value of the velocity through a face ortho. to x dir.
       ! Warning: ghost points must be excluded
       localShift = sum(velocity(c_X,1,1:resolution(c_Y)-1,1:resolution(c_Z)-1))
    end if
    ! Step 2 : reduction over all mpi processus ...
    call MPI_ALLReduce(localShift,globalShift,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
    ! Step 3 : set global shift
    globalShift = (reqFlowRate-globalShift*step(c_Y)*step(c_Z))/surf 
    
    ! Step 4 : update velocity(c_X)
    velocity(c_X,:,:,:) = velocity(c_X,:,:,:) + globalShift

  end subroutine shiftVelocityX

end module Fields
