!> Declaration, allocation of all the fields on the grid.
module PPMFields
  
  use client_data
  use client_topology, only: nsublist

  implicit none

  !> Velocity 3D (PPM-style storage) 
  real(mk), dimension(:,:,:,:,:), pointer :: PPMvelocity3D => NULL()
  !> Velocity 2D (PPM-style storage)
  real(mk), dimension(:,:,:,:), pointer :: PPMvelocity2D => NULL()
  !> Vorticity 3D (PPM-style storage) 
  real(mk), dimension(:,:,:,:,:), pointer :: PPMvorticity3D => NULL()
  !> Vorticity 2D (PPM-style storage)
  real(mk), dimension(:,:,:), pointer :: PPMvorticity2D => NULL()
  !> RHS 3D (PPM-style storage) 
  real(mk), dimension(:,:,:,:,:), pointer :: PPMrhs3D => NULL()
  !> A Scalar on a 3D grid (PPM-style storage) 
  real(mk), dimension(:,:,:,:), pointer :: PPMscalar3D => NULL()
  !> A Scalar on a 2D grid (PPM-style storage) 
  real(mk), dimension(:,:,:), pointer :: PPMscalar2D => NULL()
  
contains
  
  !> Fields allocation. 
  !! Warning : ghostpoints must be included in field (i.e. size = "real size" + 2*number of ghostpoints)
  subroutine initPPMFields3D(resolution,ghostsize)
    
    !> Required resolution for the fields (without ghosts)
    integer, dimension(dime), intent(in) :: resolution
    !> number of ghost points in each direction (ghost(c_X) = 2 means resolution(c_X)+ 4 points for the field)
    integer, dimension(:),pointer:: ghostsize
    
    integer::istat
    ! Lower and upper bounds for fields
    integer, dimension(dime) :: ldl, ldu
    ! nsublist from ppm topo. Assumed to be equal to 1 see Topology.f90
    
    ldl = 1 - ghostsize
    ldu = resolution + ghostsize
    ! Velocity ...
    allocate(PPMvelocity3D(dime,ldl(1):ldu(1),ldl(2):ldu(2),ldl(3):ldu(3),nsublist),stat = istat)
    if(istat.ne.0) stop 'Field allocation error for velocity (PPM)'
    ! Vorticity
    allocate(PPMvorticity3D(dime,ldl(1):ldu(1),ldl(2):ldu(2),ldl(3):ldu(3),nsublist), stat = istat)
    if(istat.ne.0) stop 'Field allocation error for vorticity (PPM))'
    ! rhs
    allocate(PPMrhs3D(dime,ldl(1):ldu(1),ldl(2):ldu(2),ldl(3):ldu(3),nsublist), stat = istat)
    if(istat.ne.0) stop 'Field allocation error for rhs (PPM))'

  end subroutine initPPMFields3D

  subroutine initPPMFields2D(resolution,ghostsize)
    
    !> Required resolution for the fields (without ghosts)
    integer, dimension(dime), intent(in) :: resolution
    !> number of ghost points in each direction (ghost(c_X) = 2 means resolution(c_X)+ 4 points for the field)
    integer, dimension(:),pointer:: ghostsize
    
    integer::istat
    ! Lower and upper bounds for fields
    integer, dimension(dime) :: ldl, ldu
    ! nsublist from ppm topo. Assumed to be equal to 1 see Topology.f90
    
    ldl = 1 - ghostsize
    ldu = resolution + ghostsize
    ! Velocity ...
    allocate(PPMvelocity2D(dime,ldl(1):ldu(1),ldl(2):ldu(2),1),stat = istat)
    if(istat.ne.0) stop 'Field allocation error for velocity (PPM)'
    ! Vorticity  (a sacalar in 2D)
    allocate(PPMvorticity2D(ldl(1):ldu(1),ldl(2):ldu(2),1), stat = istat)
    if(istat.ne.0) stop 'Field allocation error for vorticity (PPM))'
    ! Scalar
    allocate(PPMscalar2D(ldl(1):ldu(1),ldl(2):ldu(2),1), stat = istat)
    if(istat.ne.0) stop 'Field allocation error for scalar (PPM))'

  end subroutine initPPMFields2D

end module PPMFields
