!> Functions dealing with particles :
!> - initialisation, creation
!> - remesh and interpolation
!> - integration (push) 
module Particles
  
  use ppm_module_rmsh, only : ppm_rmsh_create_part,ppm_interp_m2p, ppm_interp_p2m !, ppm_rmsh_remesh
  use ppm_module_map_field_ghost, only:ppm_map_field_ghost_get!, ppm_map_field_ghost_put
  use ppm_module_map_field, only : ppm_map_field_push,ppm_map_field_send,ppm_map_field_pop
  use ppm_module_impose_part_bc
  use client_data
  use ppm_module_data, only : ppm_param_rmsh_kernel_mp4
  use ppm_module_map_part
  use ppm_module_util_dbg

  implicit none

  private

  public initNSSolver_particles,getMemoryUsedForParticles,ScalarSolver_particles,init_parts,&
       countAndCreateParticles, countAndUpdateParticles,PPMupdateParticles3D, PPMupdateParticles2D,&
       freeParticles,createParticlesEverywhere,npart,RK4_2D,Ppmremesh2d,RK2_2D,PPMupdateParticles2DScalar,&
       RK2_2DScalar,PPMremesh2DScalar,remesh2D,createParticlesEverywhereScalar,RK4_2Dscalar

  !> cutoff values (threshold for vorticity values for which particles are created)
  real(mk), dimension(2) :: cutoff
  !> Current number of particles
  integer :: npart
  !> Particles positions
  real(mk), dimension(:,:), pointer :: xp=>NULL()
  !> Particles  strength (ie var carried by part, vorticity indeed)
  real(mk), dimension(:,:), pointer :: omp=>NULL()
  !> Particle velocities
  real(mk), dimension(:,:), pointer :: velop=>NULL()
  !> Particles RHS term
  real(mk), dimension(:,:), pointer :: rhsp => NULL()
  !> Particles scalar term
  real(mk), dimension(:), pointer :: scalar_p => NULL()
  !> Backup vector for RK schemes
  real(mk), dimension(:,:), pointer :: buffer=>NULL(),buffer2=>NULL(),buffer3=>NULL()
  
  !> Size of buffer, i.e. more or less number of particles at the previous time step
  integer :: buffer_size
  !> Kernel for remesh ...
  integer :: kernel

contains

  subroutine ScalarSolver_particles(scalar,velocity,dt,topoid,meshid,ghostsize,resetpos,step,coordMin,resolution)
    
    !> Vorticity field, on the grid
    real(mk),dimension(:,:,:),pointer :: scalar
    !> Velocity field, on the grid
    real(mk),dimension(:,:,:,:),pointer :: velocity
    !> Current time step
    real(mk),intent(in) :: dt
    !> Current topo id
    integer, intent(in) :: topoid
    !> Current mesh id
    integer, intent(in) :: meshid
    !> Number of ghost points
    integer,  dimension(:),pointer:: ghostsize 
    !> bool to reset (or not) particles positions
    logical, intent(in) :: resetpos
    real(mk),dimension(dim3),intent(in) :: coordMin,step
    integer, dimension(dim3),intent(in) :: resolution
    !> Error status
    integer :: dir
    
    do dir=1,dime
       !call ppm_rmsh_create_part(topoid,meshid,xp,npart,scalar_p,scalar,cutoff,info,resetpos,&
       !  field_wp=velocity,wp=velop,lda2=dime)
       call reset_parts(scalar,resolution,coordMin,step)
       write(*,'(a,i5,a,i8)') '[',rank,'] initialisation with ', npart,' particles'
       ! Integrate
       call push_split_particles(dir,dt,topoid,topoid,ghostsize,velocity,coordMin,step)
       ! Remesh
       call remesh_split_mp6(scalar,dir,step,coordMin)
          
       if(dir == c_X) then
          scalar(-ghostsize(c_X)+1:ghostsize(c_X)+1,:,:) = scalar(-ghostsize(c_X)+1:ghostsize(c_X)+1,:,:) &
               + scalar(resolution(c_X)-ghostsize(c_X):resolution(c_X)+ghostsize(c_X),:,:)
          scalar(resolution(c_X)-ghostsize(c_X):resolution(c_X)+ghostsize(c_X),:,:) = scalar(-ghostsize(c_X)&
               +1:ghostsize(c_X)+1,:,:)
       else if(dir == c_Y) then 
          scalar(:,-ghostsize(c_Y)+1:ghostsize(c_Y)+1,:) = scalar(:,-ghostsize(c_Y)+1:ghostsize(c_Y)+1,:) &
               + scalar(:,resolution(c_Y)-ghostsize(c_Y):resolution(c_Y)+ghostsize(c_Y),:)
          scalar(:,resolution(c_Y)-ghostsize(c_Y):resolution(c_Y)+ghostsize(c_Y),:) = scalar(:,-ghostsize(c_Y)&
               +1:ghostsize(c_Y)+1,:)
       else
          scalar(:,:,-ghostsize(c_Z)+1:ghostsize(c_Z)+1) = scalar(:,:,-ghostsize(c_Z)+1:ghostsize(c_Z)+1) &
               + scalar(:,:,resolution(c_Z)-ghostsize(c_Z):resolution(c_Z)+ghostsize(c_Z))
          scalar(:,:,resolution(c_Z)-ghostsize(c_Z):resolution(c_Z)+ghostsize(c_Z)) = scalar(:,:,-ghostsize(c_Z)&
               +1:ghostsize(c_Z)+1)
       end if
    end do
    ! Ghost values for vorticity
!!$    call ppm_map_field_ghost_get(topoid,meshid,ghostsize,info)
!!$    call ppm_map_field_push(topoid,meshid,scalar,info)
!!$    call ppm_map_field_send(info)
!!$    call ppm_map_field_pop(topoid,meshid,scalar,ghostsize,info)
    
  end subroutine ScalarSolver_particles

  subroutine init_parts(scalar,velocity,resolution,coordMin,step)
    real(mk),dimension(:,:,:), pointer :: scalar
    real(mk),dimension(:,:,:,:),pointer :: velocity
    real(mk),dimension(dim3),intent(in)::coordMin,step
    integer, dimension(dim3),intent(in)::resolution

    integer :: i,j,k,current_part
    real(mk),dimension(dim3) :: coord
    
    current_part = 1
    npart = product(resolution-1)
    allocate(xp(dime,npart),velop(dime,npart),scalar_p(npart),buffer(dime,npart))
    do k=1,resolution(c_Z)-1
       coord(c_Z) = coordMin(c_Z) + (k-1)*step(c_Z)
       do j=1,resolution(c_Y)-1
          coord(c_Y) = coordMin(c_Y) + (j-1)*step(c_Y)
          do i=1,resolution(c_X)-1
             coord(c_X) = coordMin(c_X) + (i-1)*step(c_X)
             xp(:,current_part) = coord(:)
             velop(:,current_part) = velocity(:,i,j,k)
             scalar_p(current_part) = scalar(i,j,k)
             current_part = current_part + 1

          end do
       end do
    end do

    
  endsubroutine init_parts

  subroutine reset_parts(scalar,resolution,coordMin,step)
    
    real(mk),dimension(:,:,:), pointer :: scalar
    real(mk),dimension(dime),intent(in)::coordMin,step
    integer, dimension(dime),intent(in)::resolution

    integer :: i,j,k,current_part
    real(mk),dimension(dim3) :: coord
    
    current_part = 1
    do k=1,resolution(c_Z)-1
       coord(c_Z) = coordMin(c_Z) + (k-1)*step(c_Z)
       do j=1,resolution(c_Y)-1
          coord(c_Y) = coordMin(c_Y) + (j-1)*step(c_Y)
          do i=1,resolution(c_X)-1
             coord(c_X) = coordMin(c_X) + (i-1)*step(c_X)
             xp(:,current_part) = coord(:)
             scalar_p(current_part) = scalar(i,j,k)
             current_part = current_part + 1
          end do
       end do
    end do

  endsubroutine reset_parts


  !> Set required parameters for particles creations
  subroutine initNSSolver_particles()
    
    ! Cutoff : lower and upper bound for cutoff, i.e. for each value of the ref. field between cutoff(1) and cutoff(2), a particle will be
    ! created.
    cutoff(1) = 1.d-8
    cutoff(2) = 1e9!00000
    kernel = ppm_param_rmsh_kernel_mp4
    npart = 0
    buffer_size = npart
  end subroutine initNSSolver_particles

  
  
  !> PPM-based creation/update of particles distribution
  subroutine PPMupdateParticles3D(field_on_grid,resetpos,topoid,meshid,vel)
    !> The field used to create particles (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:,:), pointer :: field_on_grid
    !>  true to reset distribution else false
    logical, intent(in) :: resetpos
    !> topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer :: info
    !> velocity on grid (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:,:), pointer :: vel
    info = 0

    ! -------->  The following call will allocate memory for xp, omp and velop
    ! -------->  And deallocation will be handled by ppm. (?)
    ! -------->  Other variables carried by particles but not allocated during ppm_rmsh_create call
    ! -------->  is there a better way to do this, with map_push or other ppm routines? No according to Omar. 
    ! Call ppm func to create the particles distribution
    call ppm_rmsh_create_part(topoid,meshid,xp,npart,omp,dime,field_on_grid,cutoff,info,resetpos,&
         field_wp=vel,wp=velop,lda2=dime)
    write(*,'(a,i5,a,i8,a)') '[',rank,'] initialisation with ', npart,' particles'

    !if(associated(xp)) print *, "xp shape:", rank, " ", shape(xp), npart
    ! -------->  Ok, from here all vars carried by particles must have the following shape : dime,npart
    !> -------->  mesh to particles for velocity  => done in ppm_rmsh_create_part 
    !    call ppm_interp_m2p(topoid, meshid, xp, npart, velop, dime, kernel, ghostsize, vel,info)
    !    print *, 'End of update particles'

  end subroutine PPMupdateParticles3D

  subroutine PPMupdateParticles2DScalar(field_on_grid,resetpos,topoid,meshid,vel)
    !> The field used to create particles (must be in PPM-style storage)
    real(mk), dimension(:,:,:), pointer :: field_on_grid
    !>  true to reset distribution else false
    logical, intent(in) :: resetpos
    !> topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer :: info,i
    !> velocity on grid (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:), pointer :: vel
    info = 0

    ! -------->  The following call will allocate memory for xp, omp and velop
    ! -------->  And deallocation will be handled by ppm. (?)
    ! -------->  Other variables carried by particles but not allocated during ppm_rmsh_create call
    ! -------->  is there a better way to do this, with map_push or other ppm routines? No according to Omar. 
    ! Call ppm func to create the particles distribution
    call ppm_rmsh_create_part(topoid,meshid,xp,npart,scalar_p,field_on_grid,cutoff,info,.true.,&
         field_wp=vel,wp=velop,lda2=dime)
    write(*,'(a,i5,a,i8,a)') '[',rank,'] initialisation with ', npart,' particles'
 
    open(45,file="omp") ! Output only for one process
    do i = 1,npart
       write(45,'(3(11e14.5))') xp(c_X,i),scalar_p(i)
    end do
    close(45)


   !if(associated(xp)) print *, "xp shape:", rank, " ", shape(xp), npart
    ! -------->  Ok, from here all vars carried by particles must have the following shape : dime,npart
    !> -------->  mesh to particles for velocity  => done in ppm_rmsh_create_part 
    !    call ppm_interp_m2p(topoid, meshid, xp, npart, velop, dime, kernel, ghostsize, vel,info)
    !    print *, 'End of update particles'

  end subroutine PPMupdateParticles2DScalar

  subroutine PPMupdateParticles2D(field_on_grid,resetpos,topoid,meshid,vel)
    !> The field used to create particles (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:), pointer :: field_on_grid
    !>  true to reset distribution else false
    logical, intent(in) :: resetpos
    !> topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer :: info
    !> velocity on grid (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:), pointer :: vel
    info = 0

    ! -------->  The following call will allocate memory for xp, omp and velop
    ! -------->  And deallocation will be handled by ppm. (?)
    ! -------->  Other variables carried by particles but not allocated during ppm_rmsh_create call
    ! -------->  is there a better way to do this, with map_push or other ppm routines? No according to Omar. 
    ! Call ppm func to create the particles distribution
    call ppm_rmsh_create_part(topoid,meshid,xp,npart,omp,dime,field_on_grid,cutoff,info,resetpos,&
         field_wp=vel,wp=velop,lda2=dime)
    write(*,'(a,i5,a,i8,a)') '[',rank,'] initialisation with ', npart,' particles'

    !if(associated(xp)) print *, "xp shape:", rank, " ", shape(xp), npart
    ! -------->  Ok, from here all vars carried by particles must have the following shape : dime,npart
    !> -------->  mesh to particles for velocity  => done in ppm_rmsh_create_part 
    !    call ppm_interp_m2p(topoid, meshid, xp, npart, velop, dime, kernel, ghostsize, vel,info)
    !    print *, 'End of update particles'

  end subroutine PPMupdateParticles2D

  !> time integration
  !> Advection of particles, more or less a copy of Adrien's code.
  subroutine push_particles(dt,topoid,meshid,ghostsize,vel,rhs)
    !> time step
    real(mk), intent(in) ::dt
    ! topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer,  dimension(:),pointer:: ghostsize
    ! velocity on grid (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:,:), pointer :: vel
     ! Secondary field to be mapped to particles (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:,:), pointer :: rhs
    
    ! Compute new particles positions ...
    ! Integrate ... ===> update omp
    ! Todo: switch process between the different methods, leaded by a user-defined var? Later ...
    ! call runge_kutta_2(dt,topoid,meshid,ghostsize,vel,rhs)
    call RK4_3D(dt,topoid,meshid,ghostsize,vel,rhs)
  end subroutine push_particles
  
  subroutine PPMremesh3D(topoid,meshid,ghostsize,field)
    ! topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer,  dimension(:),pointer :: ghostsize
    ! vorticity on grid, (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:,:), pointer :: field
    
    integer info
    info = -1
    !call ppm_dbg_print_d(topoid,ghostlayer,1,1,info,xp,npart)
    call ppm_interp_p2m(topoid,meshid,xp,npart,omp,dime,kernel,ghostsize,field,info)
    if(info.ne.0) then
       stop 'Particles: ppm remesh error '
    end if
  end subroutine PPMremesh3D
  
  subroutine PPMremesh2D(topoid,meshid,ghostsize,field)
    ! topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer,  dimension(:),pointer :: ghostsize
    ! vorticity on grid, (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:), pointer :: field
    
    integer info
    info = -1
    !call ppm_dbg_print_d(topoid,ghostlayer,1,1,info,xp,npart)
    call ppm_interp_p2m(topoid,meshid,xp,npart,omp,dime,kernel,ghostsize,field,info)
    if(info.ne.0) then
       stop 'Particles: ppm remesh error '
    end if
  end subroutine PPMremesh2D
  subroutine PPMremesh2DScalar(topoid,meshid,ghostsize,field)
    ! topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer,  dimension(:),pointer :: ghostsize
    ! vorticity on grid, (must be in PPM-style storage)
    real(mk), dimension(:,:,:), pointer :: field
    
    integer info
    info = -1
    !call ppm_dbg_print_d(topoid,ghostlayer,1,1,info,xp,npart)
    call ppm_interp_p2m(topoid,meshid,xp,npart,scalar_p,kernel,ghostsize,field,info)
    if(info.ne.0) then
       stop 'Particles: ppm remesh error '
    end if
  end subroutine PPMremesh2DScalar
  
  !> Runge Kutta 2 for positions and 1 for vorticity
  subroutine RK2_3D(dt,topoid,meshid,ghostsize,vel,rhs)
    !> time step
    real(mk), intent(in) ::dt
    ! topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer,  dimension(:),pointer:: ghostsize
    ! velocity on grid (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:,:), pointer :: vel
     ! Secondary field to be mapped to particles (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:,:), pointer :: rhs
    
    !> local loop indices
    integer :: i,newNpart,k
    integer :: info
    
    ! buffer must be of the same size as xp so if the number of particles has increased in comparison with previous step,
    ! we reallocate secondary fields. 
    ! Note Franck : Either we allocate buffer at each time step or we reallocate only when buffer_size increase. 
    ! Since memory is our main problem, we allocate/deallocate at each time step, for the moment. 
    !    if(buffer_size.lt.npart) then
    !      deallocate(buffer) 
    allocate(buffer(dime,npart))
    buffer_size = npart

    do i=1,npart
       do k=1,dime
          buffer(k,i)=xp(k,i)
          xp(k,i)=buffer(k,i)+0.5*dt*velop(k,i)
       end do
    end do
    ! Particles positions have changed ... we must map between domains
    !    call ppm_impose_part_bc(topoid,xp,npart,info) ! Useless according to doc, required according to Omar ...
    newNpart = 0
    call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    !call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    ! Update sizes ...
    npart = newNpart
    !    if(size(rhsp,2).lt.npart) then
    !      deallocate(rhsp) 
    allocate(rhsp(dime,npart))
    !  end if
!!    print *, 'max min omp start', maxval(omp),minval(omp),maxval(rhsp),minval(rhsp)

    !! Mesh to particles for the new particles positions ...
    !> for velocity
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)
    !> for rhs of vorticity eq.
    call ppm_interp_m2p(topoid,meshid,xp,npart,rhsp,dime,kernel,ghostsize,rhs,info)
    ! Update positions according to the new velocity and vorticity with new rhs
    
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+dt*velop(k,i)
          omp(k,i)=omp(k,i)+dt*rhsp(k,i)
       end do
    end do
!!    print *, 'max min omp stop', maxval(omp),minval(omp), maxval(rhsp),minval(rhsp)

    ! Free memory as soon as possible ...
    deallocate(buffer,rhsp)
    ! Vorticity mapping ...
    newNpart = 0

    call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    !call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    npart = newNpart

  end subroutine RK2_3D

  !> Runge Kutta 4
  subroutine RK4_3D(dt,topoid,meshid,ghostsize,vel,rhs)
    !> time step
    real(mk), intent(in) ::dt
    ! topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer,  dimension(:),pointer::ghostsize
    ! velocity on grid (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:,:), pointer :: vel
     ! Secondary field to be mapped to particles (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:,:), pointer :: rhs
    
    !> local loop indices
    integer :: i, newNpart,k
    real(mk) :: alpha
    !> error status
    integer :: info
    ! buffer is used to save intermediate values for xp and omp, thus buffersize=2*npart

    allocate(buffer(dime,npart),buffer2(dime,npart),buffer3(dime,npart),rhsp(dime,npart))
    buffer_size=npart
    
    !> Compute current rhs on particles
    call ppm_interp_m2p(topoid,meshid,xp,npart,rhsp,dime,kernel,ghostsize,rhs,info)
    ! First RK stage
    ! Velocity is up to date, following the call to update_particles
    alpha=0.5*dt
    do i=1,npart
       do k=1,dime
          buffer(k,i)=xp(k,i)
          xp(k,i)=buffer(k,i)+alpha*velop(k,i)
          buffer2(k,i)=velop(k,i)
          buffer3(k,1)=rhsp(k,i)
       end do
    end do
    ! Particles positions have changed ... we must map between domains
    !    call ppm_impose_part_bc(topoid,xp,npart,info) ! Useless according to doc, required according to Omar ...
    newNpart = 0
    call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    !call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_push(buffer2,dime,npart,info)    
    call ppm_map_part_push(buffer3,dime,npart,info)    
    call ppm_map_part_push(rhsp,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(rhsp,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer3,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer2,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    ! Update sizes ...
    npart = newNpart
 
    !! Mesh to particles for the new particles positions ...
    !> for velocity and rhs
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)
    call ppm_interp_m2p(topoid,meshid,xp,npart,rhsp,dime,kernel,ghostsize,rhs,info)

    !! Second RK4 stage, with the updated velocity
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+alpha*velop(k,i)
          buffer2(k,i)=buffer2(k,i)+2.*velop(k,i)
          buffer3(k,i)=buffer3(k,i)+2.*rhsp(k,i)
       end do
    end do
    newNpart = 0
    call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    !call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_push(buffer2,dime,npart,info)    
    call ppm_map_part_push(buffer3,dime,npart,info)    
    call ppm_map_part_push(rhsp,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(rhsp,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer3,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer2,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    ! Update sizes ...
    npart = newNpart

    !! Mesh to particles for the new particles positions ...
    !> for velocity and rhs
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)
    call ppm_interp_m2p(topoid,meshid,xp,npart,rhsp,dime,kernel,ghostsize,rhs,info)

    !! Third RK4 stage
    alpha=dt
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+alpha*velop(k,i)
          buffer2(k,i)=buffer2(k,i)+2.*velop(k,i)
          buffer3(k,i)=buffer3(k,i)+2.*rhsp(k,i)
       end do
    end do
    newNpart = 0
    call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    !call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_push(buffer2,dime,npart,info)    
    call ppm_map_part_push(buffer3,dime,npart,info)    
    call ppm_map_part_push(rhsp,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(rhsp,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer3,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer2,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    ! Update sizes ...
    npart = newNpart

    !! Mesh to particles for the new particles positions ...
    !> for velocity and rhs
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)
    call ppm_interp_m2p(topoid,meshid,xp,npart,rhsp,dime,kernel,ghostsize,rhs,info)
    
    !! Last RK4 stage
    alpha=dt/6.
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+alpha*buffer2(k,i)+alpha*velop(k,i)
          omp(k,i)=omp(k,i)+alpha*buffer3(k,i)+alpha*rhsp(k,i)
       end do
    end do
    
    ! Free memory as soon as possible ...
    deallocate(buffer,buffer2,buffer3,rhsp)
    ! Vorticity mapping ...
    newNpart = 0
    call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    !call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    npart = newNpart
    
  end subroutine RK4_3D

  !> Runge Kutta 4, for 2D domain (i.e rhs == 0).
  subroutine RK4_2D(dt,topoid,meshid,ghostsize,vel)
    !> time step
    real(mk), intent(in) ::dt
    ! topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer,  dimension(:),pointer::ghostsize
    ! velocity on grid (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:), pointer :: vel
    
    !> local loop indices
    integer :: i, newNpart,k
    real(mk) :: alpha
    !> error status
    integer :: info
    ! buffer is used to save intermediate values for xp and omp, thus buffersize=2*npart

    allocate(buffer(dime,npart),buffer2(dime,npart))
    buffer_size=npart
   
    ! First RK stage
    ! Velocity is up to date, following the call to update_particles
    alpha=0.5*dt
    do i=1,npart
       do k=1,dime
          buffer(k,i)=xp(k,i)
          xp(k,i)=buffer(k,i)+alpha*velop(k,i)
          buffer2(k,i)=velop(k,i)
       end do
    end do
    ! Particles positions have changed ... we must map between domains
    call ppm_impose_part_bc(topoid,xp,npart,info) ! Useless according to doc, required according to Omar ...
    newNpart = 0
    !call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_push(buffer2,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(buffer2,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    ! Update sizes ...
    print *, "NEW", npart,newNpart
    npart = newNpart

 
    !! Mesh to particles for the new particles positions ...
    !> for velocity and rhs
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)

    !! Second RK4 stage, with the updated velocity
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+alpha*velop(k,i)
          buffer2(k,i)=buffer2(k,i)+2.*velop(k,i)
       end do
    end do
    call ppm_impose_part_bc(topoid,xp,npart,info) 
    newNpart = 0
    !call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_push(buffer2,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(buffer2,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    print *, "NEW2", npart,newNpart
    ! Update sizes ...
    npart = newNpart

    !! Mesh to particles for the new particles positions ...
    !> for velocity and rhs
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)

    !! Third RK4 stage
    alpha=dt
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+alpha*velop(k,i)
          buffer2(k,i)=buffer2(k,i)+2.*velop(k,i)
       end do
    end do
    newNpart = 0
    call ppm_impose_part_bc(topoid,xp,npart,info) 
    !call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_push(buffer2,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(buffer2,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    print *, "NEW3", npart,newNpart
    ! Update sizes ...
    npart = newNpart

    !! Mesh to particles for the new particles positions ...
    !> for velocity and rhs
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)

    !! Last RK4 stage
    alpha=dt/6.
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+alpha*buffer2(k,i)+alpha*velop(k,i)
       end do
    end do
    
    ! Free memory as soon as possible ...
    deallocate(buffer,buffer2)
    ! Vorticity mapping ...
    call ppm_impose_part_bc(topoid,xp,npart,info) 
    newNpart = 0
    !call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    print *, "NEW4", npart,newNpart
    npart = newNpart
    
  end subroutine RK4_2D
  
  !> Runge Kutta 4, for 2D domain (i.e rhs == 0).
  subroutine RK4_2DScalar(dt,topoid,meshid,ghostsize,vel)
    !> time step
    real(mk), intent(in) ::dt
    ! topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer,  dimension(:),pointer::ghostsize
    ! velocity on grid (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:), pointer :: vel
    
    !> local loop indices
    integer :: i, newNpart,k
    real(mk) :: alpha
    !> error status
    integer :: info
    ! buffer is used to save intermediate values for xp and omp, thus buffersize=2*npart

    allocate(buffer(dime,npart),buffer2(dime,npart))
    buffer_size=npart
   
    ! First RK stage
    ! Velocity is up to date, following the call to update_particles
    alpha=0.5*dt
    do i=1,npart
       do k=1,dime
          buffer(k,i)=xp(k,i)
          xp(k,i)=buffer(k,i)+alpha*velop(k,i)
          buffer2(k,i)=velop(k,i)
       end do
    end do
    ! Particles positions have changed ... we must map between domains
    !call ppm_impose_part_bc(topoid,xp,npart,info) ! Useless according to doc, required according to Omar ...
    newNpart = 0
    !call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(scalar_p,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_push(buffer2,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(buffer2,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(scalar_p,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    ! Update sizes ...
    npart = newNpart

    !! Mesh to particles for the new particles positions ...
    !> for velocity and rhs
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)
    !! Second RK4 stage, with the updated velocity
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+alpha*velop(k,i)
          buffer2(k,i)=buffer2(k,i)+2.*velop(k,i)
       end do
    end do
!    call ppm_impose_part_bc(topoid,xp,npart,info) 
    newNpart = 0
    !call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(scalar_p,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_push(buffer2,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(buffer2,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(scalar_p,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    ! Update sizes ...
    npart = newNpart

    !! Mesh to particles for the new particles positions ...
    !> for velocity and rhs
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)

    !! Third RK4 stage
    alpha=dt
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+alpha*velop(k,i)
          buffer2(k,i)=buffer2(k,i)+2.*velop(k,i)
       end do
    end do
    newNpart = 0
    !call ppm_impose_part_bc(topoid,xp,npart,info) 
    !call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(scalar_p,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_push(buffer2,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(buffer2,dime,npart,newNpart,info)
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(scalar_p,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    ! Update sizes ...
    npart = newNpart

    !! Mesh to particles for the new particles positions ...
    !> for velocity and rhs
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)

    !! Last RK4 stage
    alpha=dt/6.
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+alpha*buffer2(k,i)+alpha*velop(k,i)
       end do
    end do
    
    ! Scalar mapping ...
    !call ppm_impose_part_bc(topoid,xp,npart,info) 
    newNpart = 0
    !call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(scalar_p,npart,info)    ! vorticity
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(scalar_p,npart,newNpart,info)
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    npart = newNpart
    ! Free memory as soon as possible ...
    deallocate(buffer,buffer2)

  end subroutine RK4_2DSCALAR

  !> Runge Kutta 4, for 2D domain (i.e rhs == 0).
  subroutine RK2_2D(dt,topoid,meshid,ghostsize,vel)
    !> time step
    real(mk), intent(in) ::dt
    ! topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer,  dimension(:),pointer::ghostsize
    ! velocity on grid (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:), pointer :: vel
    
    !> local loop indices
    integer :: i, newNpart,k
    real(mk) :: alpha
    !> error status
    integer :: info
    ! buffer is used to save intermediate values for xp and omp, thus buffersize=2*npart

    allocate(buffer(dime,npart))
    buffer_size=npart
   
    ! First RK stage
    ! Velocity is up to date, following the call to update_particles
    alpha=0.5*dt
    do i=1,npart
       do k=1,dime
          buffer(k,i)=xp(k,i)
          xp(k,i)=buffer(k,i)+alpha*velop(k,i)
       end do
    end do
    ! Particles positions have changed ... we must map between domains
  !  call ppm_impose_part_bc(topoid,xp,npart,info) ! Useless according to doc, required according to Omar ...
    newNpart = 0
    !call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    ! Update sizes ...
    npart = newNpart

    !! Mesh to particles for the new particles positions ...
    !> for velocity and rhs
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)

    !! Second RK2 stage, with the updated velocity
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+dt*velop(k,i)
       end do
    end do
    ! Vorticity mapping ...
    !    call ppm_impose_part_bc(topoid,xp,npart,info) 
    newNpart = 0
    call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    !call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(omp,dime,npart,info)    ! vorticity
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(omp,dime,npart,newNpart,info)
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    npart = newNpart
    ! Free memory as soon as possible ...
    deallocate(buffer)

  end subroutine RK2_2D
  
  !> Runge Kutta 2, for 2D domain (i.e rhs == 0).
  subroutine RK2_2DScalar(dt,topoid,meshid,ghostsize,vel)
    !> time step
    real(mk), intent(in) ::dt
    ! topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer,  dimension(:),pointer::ghostsize
    ! velocity on grid (must be in PPM-style storage)
    real(mk), dimension(:,:,:,:), pointer :: vel
    
    !> local loop indices
    integer :: i, newNpart,k
    real(mk) :: alpha
    !> error status
    integer :: info,np0
    ! buffer is used to save intermediate values for xp and omp, thus buffersize=2*npart

    allocate(buffer(dime,npart))
    buffer_size=npart
    
    np0 = npart

    ! First RK stage
    ! Velocity is up to date, following the call to update_particles
    alpha=0.5*dt
    do i=1,npart
       do k=1,dime
          buffer(k,i)=xp(k,i)
          xp(k,i)=buffer(k,i)+alpha*velop(k,i)
       end do
    end do
    ! Particles positions have changed ... we must map between domains
    !call ppm_impose_part_bc(topoid,xp,npart,info) ! Useless according to doc, required according to Omar ...
    newNpart = 0
    call ppm_map_part_global(topoid,xp,npart,info) ! positions
    !call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(scalar_p,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(scalar_p,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    ! Update sizes ...
    npart = newNpart
    !! Mesh to particles for the new particles positions ...
    !> for velocity and rhs
    call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)
    !! Second RK2 stage, with the updated velocity
    do i=1,npart
       do k=1,dime
          xp(k,i)=buffer(k,i)+dt*velop(k,i)
       end do
    end do
 !   call ppm_impose_part_bc(topoid,xp,npart,info) 
    newNpart = 0
    !    call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(scalar_p,npart,info)    ! vorticity
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(scalar_p,npart,newNpart,info)
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    npart = newNpart
  end subroutine RK2_2DSCALAR

  !> Runge Kutta 2 for positions and 1 for vorticity
  subroutine push_split_particles(dir,dt,topoid,meshid,ghostsize,vel,coordMin,step)
    !> splitting direction
    integer, intent(in) :: dir
    !> time step
    real(mk), intent(in) ::dt
    ! topo and mesh ids
    integer, intent(in) :: topoid
    integer, intent(in) :: meshid
    integer,  dimension(:),pointer:: ghostsize
    ! velocity on grid
    real(mk), dimension(:,:,:,:), pointer :: vel
    real(mk),dimension(dim3),intent(in) :: coordMin,step
    !> local loop indices
    integer :: i,newNpart
    integer :: info

    do i=1,npart
       buffer(dir,i)=xp(dir,i)
       xp(dir,i)=buffer(dir,i)+0.5*dt*velop(dir,i)
    end do
    ! Particles positions have changed ... we must map between domains
    !    call ppm_impose_part_bc(topoid,xp,npart,info) ! Useless according to doc, required according to Omar ...
    newNpart = 0
    call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    !call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(velop,dime,npart,info)    ! velocity
    call ppm_map_part_push(scalar_p,npart,info)    ! vorticity
    call ppm_map_part_push(buffer,dime,npart,info)    
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(buffer,dime,npart,newNpart,info)
    call ppm_map_part_pop(scalar_p,npart,newNpart,info)
    call ppm_map_part_pop(velop,dime,npart,newNpart,info)  
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    ! Update sizes ...
    npart = newNpart
    !    if(size(rhsp,2).lt.npart) then

    !> for velocity
    !call ppm_interp_m2p(topoid,meshid,xp,npart,velop,dime,kernel,ghostsize,vel,info)
    call mesh2particles(vel,coordMin,step,dir)
    
    do i=1,npart
       xp(dir,i)=buffer(dir,i)+dt*velop(dir,i)
    end do

    ! Vorticity mapping ...
    newNpart = 0
    
    call ppm_map_part_partial(topoid,xp,npart,info) ! positions
    call ppm_map_part_global(topoid,xp,npart,info) ! positions
    call ppm_map_part_push(scalar_p,npart,info)    ! vorticity
    call ppm_map_part_send(npart,newNpart,info)          ! send
    call ppm_map_part_pop(scalar_p,npart,newNpart,info)
    call ppm_map_part_pop(xp,dime,npart,newNpart,info) 
    npart = newNpart
  end subroutine push_split_particles
  
  subroutine mesh2particles(velocity,coordMin,step,dir)

    real(mk),dimension(:,:,:,:),pointer :: velocity
    integer, intent(in) :: dir
    real(mk),dimension(dim3),intent(in) :: step
    real(mk),dimension(dim3),intent(in) :: coordMin

    real(mk) :: coord,dist
    real(mk),dimension(2) :: weights
    integer :: i,j
    integer, dimension(2,dime) :: indGrid

    do i = 1,npart
       do j = 1,2
          indGrid(j,:) = ((xp(:,i) - coordMin(:))/step(:) + epsilon(pi))+1
       enddo
       indGrid(2,dir) = indGrid(1,dir) + 1
     
       coord = coordMin(dir) + (indGrid(1,dir)-1)*step(dir)
       dist = (xp(dir,i) - coord)/step(dir)
       weights(1) = 1.-dist
       weights(2) = dist
       
       velop(dir,i) =  weights(1)*velocity(dir,indGrid(1,1),indGrid(1,2),indGrid(1,3))&
            + weights(2)*velocity(dir,indGrid(2,1),indGrid(2,2),indGrid(2,3))
    end do

  end subroutine mesh2particles

  subroutine remesh_split_mp6(scalar,dir,step,coordMin)
    real(mk),dimension(:,:,:), pointer :: scalar
    integer, intent(in) :: dir
    real(mk),dimension(dim3),intent(in) :: step
    real(mk),dimension(dim3),intent(in) :: coordMin

    real(mk) :: coord,dist
    real(mk),dimension(6) :: val,weights
    integer :: i,j

    integer, dimension(6,dime) :: indGrid

    scalar = 0.0

    do i = 1,npart
       do j = 1,6
          indGrid(j,:) = ((xp(:,i) - coordMin(:))/step(:) + epsilon(pi))+1
       enddo
       indGrid(1,dir) = indGrid(3,dir) - 2
       indGrid(2,dir) = indGrid(3,dir) - 1
       indGrid(4,dir) = indGrid(3,dir) + 1
       indGrid(5,dir) = indGrid(3,dir) + 2
       indGrid(6,dir) = indGrid(3,dir) + 3

       coord = coordMin(dir) + (indGrid(3,dir)-1)*step(dir)
       dist = (xp(dir,i) - coord)/step(dir)
       weights(1) = -(dist)*(5.*(dist + 2.)-8.)*(dist - 1.)**3/24.
       weights(2) = (dist)*(dist - 1.)*(25.*(dist + 1.)**3-114.*(dist + 1.)**2+153.*(dist + 1.)-48.)/24.
       weights(3) = -(dist-1.)*(25.*dist**4-38.*dist**3-3.*dist**2+12.*dist+12)/12.
       weights(4) = (dist)*(25.*(1. - dist)**4-38.*(1. - dist)**3-3.*(1. - dist)**2+12.*(1. - dist)+12)/12.
       weights(5) = (1. - dist)*(-dist)*(25.*(2. - dist)**3-114.*(2. - dist)**2+153.*(2. - dist)-48.)/24.
       weights(6) = -(1. - dist)*(5.*(3. - dist)-8.)*(-dist)**3/24.

       val(:) = scalar_p(i)*weights(:)
        do j = 1, 6
           scalar(indGrid(j,c_X),indGrid(j,c_Y),indGrid(j,c_Z)) = &
               scalar(indGrid(j,c_X),indGrid(j,c_Y),indGrid(j,c_Z)) + val(j)
       end do

    end do
 
  end subroutine remesh_split_mp6
  

  !> Return the memory used to save var. attached to particles
  function getMemoryUsedForParticles()
    real(mk) :: getMemoryUsedForParticles
    
    getMemoryUsedForParticles = sizeof(xp)+sizeof(velop)+sizeof(rhsp)+sizeof(buffer)+&
         sizeof(buffer2)+sizeof(buffer3)
    getMemoryUsedForParticles = getMemoryUsedForParticles*1e-6
    if(verbose) then
       write(*,'(a,i3,a,f10.4,a)') &
            '[',rank,'] memory used for particles:', getMemoryUsedForParticles, ' MB.'
    end if

  end function getMemoryUsedForParticles

  !> Create particles on the grid points where refField>cutoff.
  subroutine countAndCreateParticles(refField,velocity,resolution,step,coordMin)
    
    !> Field (on grid) used to set particles
    real(mk), dimension(:,:,:,:), pointer :: refField
    !> velocity on the grid
    real(mk), dimension(:,:,:,:), pointer :: velocity
    !> Local resolution
    integer, dimension(dim3),intent(in) :: resolution
    !> Space step
    real(mk), dimension(dim3),intent(in) :: step
    !> Coordinates of the lowest point of the current domain
    real(mk), dimension(dim3),intent(in) :: coordMin
    
    ! particles counter
    integer :: count
    ! norm of refField at each point of the grid
    real(mk) :: strength
    ! coordinates
    real(mk), dimension(dim3) :: coord
    integer, dimension(dim3) :: nbCells
    
    integer :: i,j,k
    
    nbCells = max(resolution-1,1)

    print *, "uuuu",maxval(refField)
    
    ! Count the number of particles within cutoff bounds.
    ! Loop over grid points. We use 'max' to allow the '1 point in one direction' case 
    ! (i.e. to deal with 2D case in a 3D context)
    ! Remark : there is no resize function in fortran so we must count first, allocate and then fill fields. 
    count = 0

    print *, "str", cutoff
   do k=1,nbCells(c_Z)
       do j=1,nbCells(c_Y)
          do i=1,nbCells(c_X)
             strength = sqrt(sum(refField(:,i,j,k)**2))
             if((strength.gt.cutoff(1)).and.(strength.lt.cutoff(2))) then
                count = count + 1
             end if
          end do
       end do
    end do
    
    ! Allocations
    allocate(xp(dime,count),omp(dime,count),velop(dime,count))
    
    print *, 'nb parts ...', count, shape(xp)

    ! and set values
    coord = coordMin
    count = 0
    do k=1,nbCells(c_Z)
       do j=1,nbCells(c_Y)
          do i=1,nbCells(c_X)
             strength = sqrt(sum(refField(:,i,j,k)**2))
             if((strength.gt.cutoff(1)).and.(strength.lt.cutoff(2))) then
                count = count + 1
                omp(:,count) = refField(:,i,j,k)
                velop(:,count) = velocity(:,i,j,k)
                xp(:,count) = coord(1:dime)
             end if
             coord(c_X) = coord(c_X) + step(c_X)
          end do
          coord(c_Y) = coord(c_Y) + step(c_Y)
       end do
       coord(c_Z) = coord(c_Z) + step(c_X)
    end do
    
    npart = count
    write(*,'(a,i5,a,i8,a)') '[',rank,'] initialisation with ', npart,' particles'

  end subroutine countAndCreateParticles

  !> Update particles on the grid points where refField>cutoff.
  !! Note : There are deallocation/reallocation if the number of particles as increased.
  subroutine countAndUpdateParticles(refField,velocity,resolution,step,coordMin)
    
    !> Field (on grid) used to set particles
    real(mk), dimension(:,:,:,:), pointer :: refField
    !> velocity on the grid
    real(mk), dimension(:,:,:,:), pointer :: velocity
    !> Local resolution
    integer, dimension(dim3),intent(in) :: resolution
    !> Space step
    real(mk), dimension(dim3),intent(in) :: step
    !> Coordinates of the lowest point of the current domain
    real(mk), dimension(dim3),intent(in) :: coordMin
    
    ! particles counter
    integer :: count
    ! norm of refField at each point of the grid
    real(mk) :: strength
    ! coordinates
    real(mk), dimension(dim3) :: coord
    integer, dimension(dim3) :: nbCells
    integer :: i,j,k
       
    nbCells = max(resolution-1,1)
    
    ! Count the number of particles within cutoff bounds.
    ! Loop over grid points. We use 'max' to allow the '1 point in one direction' case 
    ! (i.e. to deal with 2D case in a 3D context)
    ! Remark : there is no resize function in fortran so we must count first, allocate and then fill fields. 
    count = 0
    do k=1,max(resolution(c_Z)-1,1)
       do j=1,max(resolution(c_Y)-1,1)
          do i=1,max(resolution(c_X)-1,1)
             strength = sqrt(sum(refField(:,i,j,k)**2))
             if((strength.gt.cutoff(1)).and.(strength.lt.cutoff(2))) then
                count = count + 1
             end if
          end do
       end do
    end do
    
    ! Check if reallocation is required
    if(count > npart) then
       ! Free old memory
       if(associated(xp)) deallocate(xp)
       if(associated(omp)) deallocate(omp)
       if(associated(velop)) deallocate(velop)
    
       ! Allocations
       allocate(xp(dime,count),omp(dime,count),velop(dime,count))
    end if

    ! and set values
    coord = coordMin
    count = 0
    do k=1,max(resolution(c_Z)-1,1)
       do j=1,max(resolution(c_Y)-1,1)
          do i=1,max(resolution(c_X)-1,1)
             strength = sqrt(sum(refField(:,i,j,k)**2))
             if((strength.gt.cutoff(1)).and.(strength.lt.cutoff(2))) then
                count = count + 1
                omp(:,count) = refField(:,i,j,k)
                velop(:,count) = velocity(:,i,j,k)
                xp(:,count) = coord
             end if
             coord(c_X) = coord(c_X) + step(c_X)
          end do
          coord(c_Y) = coord(c_Y) + step(c_Y)
       end do
       coord(c_Z) = coord(c_Z) + step(c_X)
    end do
    
    npart = count
    write(*,'(a,i5,a,i8,a)') '[',rank,'] initialisation with ', npart,' particles'
  end subroutine countAndUpdateParticles

  !> Create particles on the grid points where refField>cutoff.
  subroutine createParticlesEverywhere(refField,velocity,resolution,step,coordMin)
    
    !> Field (on grid) used to set particles
    real(mk), dimension(:,:,:,:), pointer :: refField
    !> velocity on the grid
    real(mk), dimension(:,:,:,:), pointer :: velocity
    !> Local resolution
    integer, dimension(dim3),intent(in) :: resolution
    !> Space step
    real(mk), dimension(dim3),intent(in) :: step
    !> Coordinates of the lowest point of the current domain
    real(mk), dimension(dim3),intent(in) :: coordMin
    
    ! coordinates
    real(mk), dimension(dim3) :: coord
    integer, dimension(dim3) :: nbCells
    integer :: i,j,k,count
    
    nbCells = max(resolution-1,1)
    coord = coordMin
    count = 0
    npart = product(nbCells)
    ! Allocations
    allocate(xp(dime,npart),omp(dime,npart),velop(dime,npart))
    do k=1,nbCells(c_Z)
       do j=1,nbCells(c_Y)
          do i=1,nbCells(c_X)
             count = count + 1
             omp(:,count) = refField(:,i,j,k)
             velop(:,count) = velocity(:,i,j,k)
             xp(:,count) = coord
             coord(c_X) = coord(c_X) + step(c_X)
          end do
          coord(c_X) = coordMin(c_X)
          coord(c_Y) = coord(c_Y) + step(c_Y)
       end do
       coord(c_Y) = coordMin(c_Y)
       coord(c_Z) = coord(c_Z) + step(c_X)
    end do
    
    write(*,'(a,i5,a,i8,a)') '[',rank,'] initialisation with ', npart,' particles'

  end subroutine createParticlesEverywhere
  
  !> Create particles on the grid points where refField>cutoff.
  subroutine createParticlesEverywhereScalar(refField,velocity,resolution,step,coordMin)
    
    !> Field (on grid) used to set particles
    real(mk), dimension(:,:,:), pointer :: refField
    !> velocity on the grid
    real(mk), dimension(:,:,:,:), pointer :: velocity
    !> Local resolution
    integer, dimension(dim3),intent(in) :: resolution
    !> Space step
    real(mk), dimension(dim3),intent(in) :: step
    !> Coordinates of the lowest point of the current domain
    real(mk), dimension(dim3),intent(in) :: coordMin
    
    ! coordinates
    real(mk), dimension(dim3) :: coord
    integer, dimension(dim3) :: nbCells
    integer :: i,j,k,count
    
    nbCells = max(resolution-1,1)
    coord = coordMin
    count = 0
    npart = product(nbCells)
    ! Allocations
    allocate(xp(dime,npart),scalar_p(npart),velop(dime,npart))
    do k=1,nbCells(c_Z)
       do j=1,nbCells(c_Y)
          do i=1,nbCells(c_X)
             count = count + 1
             scalar_p(count) = refField(i,j,k)
             velop(:,count) = velocity(:,i,j,k)
             xp(:,count) = coord
             coord(c_X) = coord(c_X) + step(c_X)
          end do
          coord(c_X) = coordMin(c_X)
          coord(c_Y) = coord(c_Y) + step(c_Y)
       end do
       coord(c_Y) = coordMin(c_Y)
       coord(c_Z) = coord(c_Z) + step(c_X)
    end do
    
    write(*,'(a,i5,a,i8,a)') '[',rank,'] initialisation with ', npart,' particles'
    
    open(45,file="scalp") ! Output only for one process
    do i = 1,npart
       write(45,'(6e14.5)') xp(c_X,i),scalar_p(i)
    end do
    close(45)
  end subroutine createParticlesEverywhereScalar


  !> Free all arrays carried by particles
  !! useful only for tests.
  ! --------> is it required to call some specific routines to clean anything related to particles, at the end of the simulation?
  ! or ppm_finalize do it all?  
  ! According to Omar: do not use ppm "internal" routines but clean it yourself. 
  subroutine freeParticles()
    if(associated(xp)) deallocate(xp)
    if(associated(omp)) deallocate(omp)
    if(associated(velop)) deallocate(velop)
    if(associated(rhsp)) deallocate(rhsp)
    if(associated(buffer)) deallocate(buffer)
    if(associated(buffer2)) deallocate(buffer2)
    if(associated(scalar_p)) deallocate(scalar_p)
    npart = 0
    buffer_size = 0
  end subroutine freeParticles

  subroutine remesh2D(field,coordMin,step,resolution)
    
    real(mk),dimension(:,:,:),pointer :: field
    real(mk),dimension(dim3),intent(in) :: coordMin
    real(mk),dimension(dim3),intent(in) :: step
    integer, dimension(dim3),intent(in) :: resolution

    real(mk),dimension(dime) :: invStep
    !! List of grid point indices for each particle : [Left-1 Left Right Right+1]
    !! Left/Right == left/right points of the grid around the particle
    integer, dimension(4,dime) :: indGrid
    !! current point coordinates and distance to the left boundary of the local domain
    real(mk),dimension(dime) :: coord,dist
    !! weights
    real(mk),dimension(4,dime) :: weights
    integer:: i,j
    invStep = 1./step(1:dime)
    do i = 1,npart
       indGrid(2,:) = ((xp(:,i) - coordMin(:))*invStep(:) + epsilon(pi))+1
       indGrid(1,:) = indGrid(2,:) - 1
       indGrid(3,:) = indGrid(2,:) + 1
       indGrid(4,:) = indGrid(2,:) + 2
       !enforce periodicity
       do j =1,4
          indGrid(j,:) = mod(indGrid(j,:)+resolution(1:dime),resolution(1:dime))+1
       end do

       coord = coordMin(1:dime) + (indGrid(2,:)-1)*step(1:dime)
       dist = (xp(:,i) - coord)*invStep
       
       weights(1,:) = 0.5*dist*((1.-dist)**2)
       weights(2,:) = 1.-2.5*dist**2+1.5*dist**3
       weights(3,:) = 1.-2.5*(1.-dist)**2+1.5*(1.-dist)**3
       weights(4,:) = 0.5*dist**2*(dist-1.)
       
       do j =1,4
          field(indGrid(j,c_X),indGrid(:,c_Y),1) = field(indGrid(j,c_X),indGrid(:,c_Y),1)&
               + weights(j,c_X)*weights(:,c_Y)*scalar_p(i)
       end do

    end do
    
  end subroutine remesh2D



end module Particles
