module ppmljk_poisson

  use client_data
  use ppm_module_poisson
  use ppm_module_mktopo
  use ppm_module_topo_get
  use ppm_module_mesh_define
  use ppm_module_map_field
  use ppm_module_map_field_global
  use ppm_module_map
  use mpi

  implicit none

  integer,dimension(dim3), parameter :: zeros = (/0,0,0/)

  !#define __ZEROSI (/0,0,0/)

contains

  !> Initialisation of Poisson solver (fft), based on ppm and on a copy on the 
  !! equivalent file in ppm
  subroutine mypoisson_init(topoid,meshid,ppmpoisson,fieldin,fieldout,info)
!!! * ppm_poisson_grn_pois_per - Poisson equation, periodic boundaries
!!!
!!! [NOTE]
!!! fieldin is not preserved by this routine!
!!! fieldin and fieldout must NOT be the same fields. In-place FFTs have
!!! not been implemented.
!!!

    !> the topology id
    integer, intent(in) :: topoid
    !> corresponding mesh id
    integer, intent(in) :: meshid
    !> fftw plan interface
    type(ppm_poisson_plan),intent(out) :: ppmpoisson
    !> input field (i.e. rhs of poisson eq.)
    !@ strictly speaking fieldin is not being used in the init routine
    real(mk), dimension(:,:,:,:,:), pointer :: fieldin
    !> output field
    real(mk), dimension(:,:,:,:,:), pointer :: fieldout
    !> error status
    integer, intent(out) :: info

    !-------------------------------------------------------------------------
    ! Local variables
    !-------------------------------------------------------------------------
    !  real(mk)                            :: t0
    real(mk),dimension(:,:),pointer     :: xp=>NULL()      !particle positions
    TYPE(ppm_t_topo),pointer                :: topology=>NULL()
    TYPE(ppm_t_equi_mesh)                   :: mesh
    integer ,dimension(dime)               :: indl,indu
    real(mk),PARAMETER                  :: PI=ACOS(-1.0_mk) !@ use ppm pi
    real(mk)                            :: normfac
!!!factor for the Greens function, including FFT normalization
    integer                                 :: i,j,k
    integer                                 :: kx,ky,kz
    integer                                 :: isubl,isub
    integer,dimension(dime*2)              :: bcdef
    integer                                 :: assigning
    integer                                 :: decomposition
    integer,SAVE                            :: ttopoid
    integer                                 :: tmeshid
    real(mk)                            :: Lx2,Ly2,Lz2

    real(mk),dimension(dime)           :: tmpmin,tmpmax
    integer, dimension(:),pointer           :: maxndataxy=>NULL(),maxndataz=>NULL()
    integer, dimension(:  ), pointer        :: dummynmxy=>NULL(),dummynmz=>NULL()

    !-------------------------------------------------------------------------
    ! Initialise routine
    !-------------------------------------------------------------------------
    !  CALL substart('ppm_poisson_init',t0,info)

    ppmpoisson%case  = ppm_poisson_grn_pois_per

    !-------------------------------------------------------------------------
    ! Nullify pointers from the ppmpoisson plans and the fftplans
    !-------------------------------------------------------------------------
    NULLIFY(xp)
    NULLIFY(ppmpoisson%costxy)
    NULLIFY(ppmpoisson%istartxy)
    NULLIFY(ppmpoisson%ndataxy)
    NULLIFY(ppmpoisson%istartxyc)
    NULLIFY(ppmpoisson%ndataxyc)
    NULLIFY(ppmpoisson%costz)
    NULLIFY(ppmpoisson%istartz)
    NULLIFY(ppmpoisson%ndataz)
    NULLIFY(ppmpoisson%planfxy%plan)
    NULLIFY(ppmpoisson%planbxy%plan)
    NULLIFY(ppmpoisson%planfz%plan)
    NULLIFY(ppmpoisson%planbz%plan)

    !-------------------------------------------------------------------------
    ! Get topology and mesh values of input/output
    !-------------------------------------------------------------------------
    call ppm_topo_get(topoid,topology,info)
    mesh  = topology%mesh(meshid)

    !-------------------------------------------------------------------------
    ! Setup mesh sizes for intermediate meshes/topologies
    !-------------------------------------------------------------------------
    !size of real slabs
    ppmpoisson%nmxy (1) =  mesh%nm(1)
    ppmpoisson%nmxy (2) =  mesh%nm(2)
    ppmpoisson%nmxy (3) =  mesh%nm(3)
    !size of complex slabs
    ppmpoisson%nmxyc(1) = (mesh%nm(1)-1)/2+1
    !!ppmpoisson%nmxyc(1) =  mesh%nm(1)
    ppmpoisson%nmxyc(2) =  mesh%nm(2)
    ppmpoisson%nmxyc(3) =  mesh%nm(3)
    !size of complex pencils
    ppmpoisson%nmz  (1) = (ppmpoisson%nmxyc(1))
    ppmpoisson%nmz  (2) = (ppmpoisson%nmxyc(2))
    ppmpoisson%nmz  (3) = (ppmpoisson%nmxyc(3))
    !size of the fft
    ppmpoisson%nmfft(1) =  mesh%nm(1)-1
    ppmpoisson%nmfft(2) =  mesh%nm(2)-1
    ppmpoisson%nmfft(3) =  mesh%nm(3)-1
    !Inverse of the size of the domain squared
    Lx2 = 1.0_mk/(topology%max_physd(1)-topology%min_physd(1))**2
    Ly2 = 1.0_mk/(topology%max_physd(2)-topology%min_physd(2))**2
    Lz2 = 1.0_mk/(topology%max_physd(3)-topology%min_physd(3))**2

    !-------------------------------------------------------------------------
    ! Create temporary derivation arrays if necessary
    !-------------------------------------------------------------------------
    ppmpoisson%derivatives = ppm_poisson_drv_curl_sp
    
    !-------------------------------------------------------------------------
    ! Create spectral scaling components always. Just in case some 
    ! reprojection comes up
    ! The conditionals need to be for not just the Poisson equation
    !-------------------------------------------------------------------------
    ppmpoisson%normkx = &
         & 2.0_mk*PI/(topology%max_physd(1)-topology%min_physd(1))
    ppmpoisson%normky = &
         & 2.0_mk*PI/(topology%max_physd(2)-topology%min_physd(2))
    ppmpoisson%normkz = &
         & 2.0_mk*PI/(topology%max_physd(3)-topology%min_physd(3))

    !-------------------------------------------------------------------------
    ! Create new slab topology
    !-------------------------------------------------------------------------
    ttopoid = 0
    tmeshid = -1
    decomposition       = ppm_param_decomp_xy_slab
    assigning           = ppm_param_assign_internal
    bcdef               = ppm_param_bcdef_periodic
    tmpmin              = topology%min_physd
    tmpmax              = topology%max_physd

    CALL ppm_mktopo(ttopoid,tmeshid,xp,0,&
         & decomposition,assigning,&
         & tmpmin,tmpmax,bcdef,&
         & zeros,ppmpoisson%costxy,&
         & ppmpoisson%nmxy,info)

    ppmpoisson%topoidxy = ttopoid
    ppmpoisson%meshidxy = tmeshid
    !-------------------------------------------------------------------------
    ! Get additional xy-mesh information
    !-------------------------------------------------------------------------
    CALL ppm_topo_get_meshinfo(ppmpoisson%topoidxy,ppmpoisson%meshidxy, &
         & dummynmxy,ppmpoisson%istartxy,ppmpoisson%ndataxy,maxndataxy, &
         & ppmpoisson%isublistxy,ppmpoisson%nsublistxy,info)


    !-------------------------------------------------------------------------
    ! Create complex slab mesh
    !-------------------------------------------------------------------------
    ttopoid = ppmpoisson%topoidxy
    tmeshid = -1
    CALL ppm_mesh_define(ttopoid,tmeshid,&
         & ppmpoisson%nmxyc,ppmpoisson%istartxyc,ppmpoisson%ndataxyc,info)
    ppmpoisson%meshidxyc = tmeshid


    !-------------------------------------------------------------------------
    ! Create new pencil topology
    !-------------------------------------------------------------------------
    ttopoid = 0
    tmeshid = -1
    bcdef               = ppm_param_bcdef_periodic
    assigning       = ppm_param_assign_internal
    decomposition   = ppm_param_decomp_zpencil

    CALL ppm_mktopo(ttopoid,tmeshid,xp,0,&
         & decomposition,assigning,&
         & tmpmin,tmpmax,bcdef,&
         & zeros,ppmpoisson%costz,&
         & ppmpoisson%nmz,info)

    ppmpoisson%topoidz = ttopoid
    ppmpoisson%meshidz = tmeshid
    !-------------------------------------------------------------------------
    ! Get additional z-mesh information
    !-------------------------------------------------------------------------
    CALL ppm_topo_get_meshinfo(ppmpoisson%topoidz,ppmpoisson%meshidz, &
         & dummynmz,ppmpoisson%istartz,ppmpoisson%ndataz,maxndataz, &
         & ppmpoisson%isublistz,ppmpoisson%nsublistz,info)

    !-------------------------------------------------------------------------
    ! Set and get minimum and maximum indicies
    !-------------------------------------------------------------------------
    indl(1) = 1
    indl(2) = 1
    indl(3) = 1

    !-------------------------------------------------------------------------
    ! Allocate real xy slabs
    !-------------------------------------------------------------------------
    ALLOCATE(ppmpoisson%fldxyr(dime,&
         & indl(1):maxndataxy(1),indl(2):maxndataxy(2),indl(3):maxndataxy(3),&
         & 1:ppmpoisson%nsublistxy),stat=info)


    !-------------------------------------------------------------------------
    ! Set and get minimum and maximum indicies of COMPLEX xy slabs
    !-------------------------------------------------------------------------
    indl(1) = 1
    indl(2) = 1
    indl(3) = 1
    indu(1) = 0
    indu(2) = 0
    indu(3) = 0
    DO isub=1,ppmpoisson%nsublistxy
       isubl = ppmpoisson%isublistxy(isub)
       indu(1) = MAX(indu(1),ppmpoisson%ndataxyc(1,isubl))
       indu(2) = MAX(indu(2),ppmpoisson%ndataxyc(2,isubl))
       indu(3) = MAX(indu(3),ppmpoisson%ndataxyc(3,isubl))
    ENDDO


    !-------------------------------------------------------------------------
    ! Allocate complex xy slabs
    !-------------------------------------------------------------------------
    ALLOCATE(ppmpoisson%fldxyc(dime,&
         & indl(1):indu(1),indl(2):indu(2),indl(3):indu(3),&
         & 1:ppmpoisson%nsublistxy),stat=info)


    !-------------------------------------------------------------------------
    ! Allocate two complex z pencils + Greens fcn array !@check return vars.
    !-------------------------------------------------------------------------
    ALLOCATE(ppmpoisson%fldzc1(dime,&
         & indl(1):maxndataz(1),indl(2):maxndataz(2),indl(3):maxndataz(3),&
         & 1:ppmpoisson%nsublistz),stat=info)

    ALLOCATE(ppmpoisson%fldzc2(dime,&
         & indl(1):maxndataz(1),indl(2):maxndataz(2),indl(3):maxndataz(3),&
         & 1:ppmpoisson%nsublistz),stat=info)


    !-------------------------------------------------------------------------
    ! The complex Greens function is always kept on the z-pencil topology
    !-------------------------------------------------------------------------
    ALLOCATE(ppmpoisson%fldgrnr(&
         & indl(1):maxndataz(1),indl(2):maxndataz(2),indl(3):maxndataz(3),&
         & 1:ppmpoisson%nsublistz),stat=info)

    !-------------------------------------------------------------------------
    ! Set up xy FFT plans
    ! The inverse plan takes the returning topology since it has the full size
    !-------------------------------------------------------------------------
    CALL ppm_fft_forward_2d(ppmpoisson%topoidxy,ppmpoisson%meshidxy,&
         & ppmpoisson%planfxy,ppmpoisson%fldxyr,&
         & ppmpoisson%fldxyc,info)

    CALL ppm_fft_backward_2d(ppmpoisson%topoidxy,ppmpoisson%meshidxy,&
         & ppmpoisson%planbxy,ppmpoisson%fldxyc,&
         & ppmpoisson%fldxyr,info)


    !-------------------------------------------------------------------------
    ! Set up z FFT plans
    !-------------------------------------------------------------------------
    CALL ppm_fft_forward_1d(ppmpoisson%topoidz,ppmpoisson%meshidz,&
         & ppmpoisson%planfz,ppmpoisson%fldzc1,&
         & ppmpoisson%fldzc2,info)

    CALL ppm_fft_backward_1d(ppmpoisson%topoidz,ppmpoisson%meshidz,&
         & ppmpoisson%planbz,ppmpoisson%fldzc2,&
         & ppmpoisson%fldzc1,info)


    !-------------------------------------------------------------------------
    ! Compute Greens function. Analytic, periodic
    !
    ! (d2_/dx2 + d2_/dy2 + d2_/dz2)psi = -omega     =>
    ! -4*pi2(kx2 + ky2 + kz2)PSI       = -OMEGA     =>
    ! PSI                              = 1/(4*pi2)*1/(kx2 + ky2 + kz2)OMEGA
    !-------------------------------------------------------------------------
    ! Scaling the spectral coefficients... 
    ! one minus due to (i*k)^2 and another due to the Poisson equation
    normfac = 1.0_mk/(4.0_mk*PI*PI * &
                                !and normalisation of FFTs (full domain) !vertex
         & real((ppmpoisson%nmfft(1))* &
         &      (ppmpoisson%nmfft(2))* &
         &      (ppmpoisson%nmfft(3)),mk))
    DO isub=1,ppmpoisson%nsublistz
       isubl=ppmpoisson%isublistz(isub)
       DO k=1,ppmpoisson%ndataz(3,isubl)
          DO j=1,ppmpoisson%ndataz(2,isubl)
             DO i=1,ppmpoisson%ndataz(1,isubl)
                kx = i-1 + (ppmpoisson%istartz(1,isubl)-1)
                ky = j-1 + (ppmpoisson%istartz(2,isubl)-1)
                kz = k-1 + (ppmpoisson%istartz(3,isubl)-1)
                !This is a nasty way to do this but it is only done once so...:
                IF (kx .GT. (ppmpoisson%nmfft(1)/2)) kx = kx-(ppmpoisson%nmfft(1))
                IF (ky .GT. (ppmpoisson%nmfft(2)/2)) ky = ky-(ppmpoisson%nmfft(2))
                IF (kz .GT. (ppmpoisson%nmfft(3)/2)) kz = kz-(ppmpoisson%nmfft(3))
                ppmpoisson%fldgrnr(i,j,k,isub) = &
                     & normfac/(real(kx*kx,mk)*Lx2 &
                     &        + real(ky*ky,mk)*Ly2 &
                     &        + real(kz*kz,mk)*Lz2)
                !Take care of singularity
                !This is nasty as well
                IF ((kx*kx+ky*ky+kz*kz) .EQ. 0) THEN
                   ppmpoisson%fldgrnr(i,j,k,isub) = 0.0_mk
                ENDIF
             ENDDO
          ENDDO
       ENDDO
    ENDDO
  end subroutine mypoisson_init

  subroutine mypoisson_solve(topoid,meshid,ppmpoisson,fieldin,fieldout,gstw,info)


    INTEGER, INTENT(IN)                                         :: topoid
!!! Topology ID
    INTEGER, INTENT(IN)                                         :: meshid
!!! Mesh ID
    TYPE(ppm_poisson_plan),INTENT(INOUT)                        :: ppmpoisson
!!! The PPM Poisson plan
    REAL(mk),DIMENSION(:,:,:,:,:),POINTER                   :: fieldin
!!! Input data field
    REAL(mk),DIMENSION(:,:,:,:,:),POINTER                   :: fieldout
!!! Output data field
    INTEGER,DIMENSION(dime),INTENT(IN)                         :: gstw
!!! Ghost layer width
    INTEGER, INTENT(OUT)                                        :: info
!!! Return status, 0 upon succes

    !-------------------------------------------------------------------------
    ! Local variables
    !-------------------------------------------------------------------------

    REAL(mk)                      :: t0
    INTEGER                           :: isub,isubl
    INTEGER                           :: i,j,k
    INTEGER                           :: info2
    INTEGER                           :: presentcase
    COMPLEX(mk)                   :: divomega
    INTEGER                           :: gi,gj,gk
    COMPLEX(mk)                   :: kx,ky,kz
    COMPLEX(mk)                   :: phix,phiy,phiz
    REAL(mk)                      :: normfac

    !-------------------------------------------------------------------------
    ! Check if we run a different/temporary case
    !-------------------------------------------------------------------------

    presentcase = ppmpoisson%case

    !-----------------------------------------------------------------------
    ! Map data globally to the slabs (XY)
    ! This is where the vorticity is extended and padded with 0 for free-space
    !-----------------------------------------------------------------------
    !Initialise
    CALL ppm_map_field_global(topoid,ppmpoisson%topoidxy,meshid,ppmpoisson%meshidxy,info)
    !Push the data
    CALL ppm_map_field_push(topoid,meshid,fieldin,3,info)
    CALL ppm_map_field_send(info)
    !Retrieve
    CALL ppm_map_field_pop(ppmpoisson%topoidxy,ppmpoisson%meshidxy,ppmpoisson%fldxyr,3,zeros,info)

    !-----------------------------------------------------------------------
    ! Do slab FFT (XY) - use the xy topology as its extent has not been halved
    !-----------------------------------------------------------------------
    CALL ppm_fft_execute_2d(ppmpoisson%topoidxy,&
         & ppmpoisson%meshidxy, ppmpoisson%planfxy, &
         & ppmpoisson%fldxyr, ppmpoisson%fldxyc, &
         & info)

    !-----------------------------------------------------------------------
    ! Map to the pencils (Z)
    !-----------------------------------------------------------------------
    !Initialise
    CALL ppm_map_field_global(&
         & ppmpoisson%topoidxy, &
         & ppmpoisson%topoidz, &
         & ppmpoisson%meshidxyc, &
         & ppmpoisson%meshidz,info)

    !Push the data
    CALL ppm_map_field_push(&
         & ppmpoisson%topoidxy, &
         & ppmpoisson%meshidxyc,ppmpoisson%fldxyc,3,info)
    !Send
    CALL ppm_map_field_send(info)

    !Retrieve
    CALL ppm_map_field_pop(&
         & ppmpoisson%topoidz, &
         & ppmpoisson%meshidz,ppmpoisson%fldzc1, &
         & 3,zeros,info)

    !-----------------------------------------------------------------------
    ! Do pencil FFT (Z)
    !-----------------------------------------------------------------------
    CALL ppm_fft_execute_1d(ppmpoisson%topoidz,&
         & ppmpoisson%meshidz, ppmpoisson%planfz, &
         & ppmpoisson%fldzc1, ppmpoisson%fldzc2, &
         & info)


    !-----------------------------------------------------------------------
    ! Apply the periodic Greens function
    !-----------------------------------------------------------------------
    DO isub=1,ppmpoisson%nsublistz
       isubl=ppmpoisson%isublistz(isub)
       DO k=1,ppmpoisson%ndataz(3,isubl)
          DO j=1,ppmpoisson%ndataz(2,isubl)
             DO i=1,ppmpoisson%ndataz(1,isubl)
                ppmpoisson%fldzc2(1,i,j,k,isub) = ppmpoisson%fldgrnr( i,j,k,isub)*&
                     & ppmpoisson%fldzc2(1,i,j,k,isub)
                ppmpoisson%fldzc2(2,i,j,k,isub) = ppmpoisson%fldgrnr( i,j,k,isub)*&
                     & ppmpoisson%fldzc2(2,i,j,k,isub)
                ppmpoisson%fldzc2(3,i,j,k,isub) = ppmpoisson%fldgrnr( i,j,k,isub)*&
                     & ppmpoisson%fldzc2(3,i,j,k,isub)
             ENDDO
          ENDDO
       ENDDO
    ENDDO


    !-----------------------------------------------------------------------
    ! Spectral derivatives
    ! normkx, etc contains 2pi/Lx
    !-----------------------------------------------------------------------
    normfac = 1.0_MK/ REAL((ppmpoisson%nmfft(1))* & !vertex
         & (ppmpoisson%nmfft(2))* &
         & (ppmpoisson%nmfft(3)),MK)
    DO isub=1,ppmpoisson%nsublistz
       isubl=ppmpoisson%isublistz(isub)
       DO k=1,ppmpoisson%ndataz(3,isubl)
          gk = k - 1 + (ppmpoisson%istartz(3,isubl)-1)
          IF (gk .GT. (ppmpoisson%nmfft(3)/2)) gk = gk-(ppmpoisson%nmfft(3))
          kz = CMPLX(0.0_MK,REAL(gk,MK),MK)*ppmpoisson%normkz
          DO j=1,ppmpoisson%ndataz(2,isubl)
             gj = j - 1 + (ppmpoisson%istartz(2,isubl)-1)
             IF (gj .GT. (ppmpoisson%nmfft(2)/2)) gj = gj-(ppmpoisson%nmfft(2))
             ky = CMPLX(0.0_MK,REAL(gj,MK),MK)*ppmpoisson%normky
             DO i=1,ppmpoisson%ndataz(1,isubl)
                gi = i - 1 + (ppmpoisson%istartz(1,isubl)-1)
                IF (gi .GT. (ppmpoisson%nmfft(1)/2)) gi = gi-(ppmpoisson%nmfft(1))
                kx = CMPLX(0.0_MK,REAL(gi,MK),MK)*ppmpoisson%normkx
                
                phix = ppmpoisson%fldzc2(1,i,j,k,isub)
                phiy = ppmpoisson%fldzc2(2,i,j,k,isub)
                phiz = ppmpoisson%fldzc2(3,i,j,k,isub)
                
                ppmpoisson%fldzc2(1,i,j,k,isub) = (ky*phiz-kz*phiy)
                ppmpoisson%fldzc2(2,i,j,k,isub) = (kz*phix-kx*phiz)
                ppmpoisson%fldzc2(3,i,j,k,isub) = (kx*phiy-ky*phix)
             ENDDO
          ENDDO
       ENDDO
    ENDDO

    !-----------------------------------------------------------------------
    ! IFFT pencil (Z)
    !-----------------------------------------------------------------------
    CALL ppm_fft_execute_1d(ppmpoisson%topoidz,&
         & ppmpoisson%meshidz, ppmpoisson%planbz, &
         & ppmpoisson%fldzc2, ppmpoisson%fldzc1, &
         & info)

    !-----------------------------------------------------------------------
    ! Map back to slabs (XY)
    !-----------------------------------------------------------------------
    !Initialise
    CALL ppm_map_field_global(&
         & ppmpoisson%topoidz, &
         & ppmpoisson%topoidxy, &
         & ppmpoisson%meshidz, &
         & ppmpoisson%meshidxyc,info)
    !Push the data
    CALL ppm_map_field_push(&
         & ppmpoisson%topoidz, &
         & ppmpoisson%meshidz,ppmpoisson%fldzc1,3,info)

    !Send
    CALL ppm_map_field_send(info)

    !Retrieve
    CALL ppm_map_field_pop(&
         & ppmpoisson%topoidxy, &
         & ppmpoisson%meshidxyc,ppmpoisson%fldxyc, &
         & 3,zeros,info)

    !-----------------------------------------------------------------------
    ! IFFT (XY) use the non-reduced topology
    !-----------------------------------------------------------------------
    CALL ppm_fft_execute_2d(ppmpoisson%topoidxy,&
         & ppmpoisson%meshidxy, ppmpoisson%planbxy, &
         & ppmpoisson%fldxyc, ppmpoisson%fldxyr, &
         & info)


    !-----------------------------------------------------------------------
    ! Map back to standard topology (XYZ)
    !-----------------------------------------------------------------------
    !Initialise
    CALL ppm_map_field_global(&
         & ppmpoisson%topoidxy, &
         & topoid, &
         & ppmpoisson%meshidxy, &
         & meshid,info)
    !Push the data
    CALL ppm_map_field_push(&
         & ppmpoisson%topoidxy, &
         & ppmpoisson%meshidxy,ppmpoisson%fldxyr,3,info)

    !Send
    CALL ppm_map_field_send(info)

    !-------------------------------------------------------------------------
    ! FINAL RETRIEVE - Here we do different things depending on the task
    ! i.e. the receiver varies
    !-------------------------------------------------------------------------
    IF ((ppmpoisson%derivatives .EQ. ppm_poisson_drv_curl_fd2 .OR. &
         &  ppmpoisson%derivatives .EQ. ppm_poisson_drv_curl_fd4) ) THEN
       CALL ppm_map_field_pop(&
            & topoid, &
            & meshid,ppmpoisson%drv_vr, &
            & 3,gstw,info)
       !-------------------------------------------------------------------------
       ! Ghost the temporary array for derivatives (drv_vr)
       !-------------------------------------------------------------------------
       CALL ppm_map_field_ghost_get(topoid,meshid,gstw,info)
       CALL ppm_map_field_push(topoid,meshid,ppmpoisson%drv_vr,3,info)
       CALL ppm_map_field_send(info)
       CALL ppm_map_field_pop(topoid,meshid,ppmpoisson%drv_vr,3,gstw,info)

    ELSE
       CALL ppm_map_field_pop(&
            & topoid, &
            & meshid,fieldout, &
            & 3,gstw,info)
    ENDIF

    !-------------------------------------------------------------------------
    ! Treat ghost layer to make FD stencils work
    !-------------------------------------------------------------------------
    IF (ppmpoisson%derivatives .EQ. ppm_poisson_drv_curl_fd2) THEN
       CALL ppm_poisson_extrapolateghost(topoid,meshid,ppmpoisson%drv_vr,&
            & 2,4,gstw,info)
    ENDIF
    IF (ppmpoisson%derivatives .EQ. ppm_poisson_drv_curl_fd4 .AND.&
         & (presentcase .EQ. ppm_poisson_grn_pois_fre)) THEN
       CALL ppm_poisson_extrapolateghost(topoid,meshid,ppmpoisson%drv_vr,&
            & 2,4,gstw,info)
    ENDIF

    !-------------------------------------------------------------------------
    ! Optionally do derivatives
    ! Perhaps make ppm_poisson_fd take _none as argument. Then maybe no
    ! if-statement is required
    !-------------------------------------------------------------------------
    IF (presentcase .NE. ppm_poisson_grn_reprojec) THEN
       IF (ppmpoisson%derivatives .EQ. ppm_poisson_drv_curl_fd2) THEN
          CALL ppm_poisson_fd(topoid,meshid,ppmpoisson%drv_vr,fieldout,&
               & ppm_poisson_drv_curl_fd2,info)
       ENDIF
       IF (ppmpoisson%derivatives .EQ. ppm_poisson_drv_curl_fd4) THEN
          CALL ppm_poisson_fd(topoid,meshid,ppmpoisson%drv_vr,fieldout,&
               & ppm_poisson_drv_curl_fd4,info)
       ENDIF
    ENDIF

    !-------------------------------------------------------------------------
    ! Finally ghost the velocity/stream function field before returning it
    ! Also extrapolate if freespace
    !-------------------------------------------------------------------------
    CALL ppm_map_field_ghost_get(topoid,meshid,gstw,info)
    CALL ppm_map_field_push(topoid,meshid,fieldout,3,info)
    CALL ppm_map_field_send(info)
    CALL ppm_map_field_pop(topoid,meshid,fieldout,3,gstw,info)

  end subroutine mypoisson_solve


end module ppmljk_poisson
