# -*- coding: utf-8 -*-
"""
@package Problem
Problem representation
"""
from ..Param import *
from DiscreteProblem import DiscreteProblem


class DiscreteTransportProblem(DiscreteProblem):
    """
    Transport problem discrete representation.
    """

    def __init__(self, advection, domains=None, variables=None, operators=None):
        """
        Constructor.
        Create a transport problem instance.

        @param advection : Advection operator for transport problem.
        @param domains : Additionals domains.
        @param variables : Additionnals variables.
        @param operators : Additionnals operators.
        """
        DiscreteProblem.__init__(self)
        ## Current simulation time
        self.t = 0
        ## Advection operator for the transport problem
        self.advecOp = advection
        self.addOperator([self.advecOp])
        if domains != None:
            self.addDomain(domains)
        if variables != None:
            self.addVariable(variables)
        if operators != None:
            self.addOperator(operators)
        ## Splitting informations
        self.splittingStep = None

    def step(self, dt):
        [c_step[0].applyOperator(self.t, dt * c_step[1], c_step[2]) for c_step in self.operator_list]
        self.t += dt

    def solve(self, T, dt):
        """
        Solving method.
        Solve by applying each operators. For a given splitting.

        @param T : Simulation final time.
        @param dt : Simulation time step.
        """
        print "\n===    Solving    ==="
        ite = 0
        if not self.printer is None:
            while self.t < T:
                self.step(dt)
                ite += 1
                #print "iteration ", ite, "\t t =", self.t
                self.printer.printStep()
        else:
            while self.t < T:
                self.step(dt)
                ite += 1
                #print "Iteration ", ite, "\t t =", self.t, "\t(No output mode)"
        print "\n===    End Solving    ==="
        print "\n===    Timings    ==="
        c_time = 0.
        q_time = 0.
        s_time = 0.
        c_flop = 0
        c_bytes_accessed = 0
        for op in self.operators:
            op_compute_time = 0.
            print "Time :", op
            for i in xrange(self.domains[0].dimension):
                print "  - direction ", i, " :", op.compute_time[i], "s"
                op_compute_time += op.compute_time[i]
            print "  - TOTAL          ", op_compute_time, "s \t", (op.total_flop / op_compute_time) * 1e-9, "GFlop/s \t", (op.total_bytes_accessed / op_compute_time) * 1e-9, "GB/s"
            c_time += op_compute_time
            c_flop += op.total_flop
            q_time += op.queued_time
            s_time += op.submit_time
            c_bytes_accessed += op.total_bytes_accessed
        print "Computing time : ", c_time, "s \t", (c_flop / c_time) * 1e-9, "GFlop/s \t", (c_bytes_accessed / c_time) * 1e-9, "GB/s"
        if not self.operators[0].gpu_kernel is None:
            print "OpenCL queued time  : ", q_time, "s"
            print "OpenCL submit time  : ", s_time, "s"

    def __str__(self):
        """ToString method"""
        s = "DiscreteTransportProblem (DiscreteProblem) : {0}\n".format(id(self))
        s += " {0} domains, {1} variables and {2} operators.\n".format(len(self.domains), len(self.variables), len(self.operators))
        s += " Domains : \n"
        for d in self.domains:
            s += str(d)
        s += "\n Variables : \n"
        for v in self.variables:
            s += str(v)
        s += "\n Operators : \n"
        for o in self.operators:
            s += str(o)
        s += "\n Solver : {0}\n".format(self.solver)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : DiscreteTransportProblem"
    print DiscreteTransportProblem.__doc__
