# -*- coding: utf-8 -*-
"""
@package Utils
GPU Particular solver description.
"""
import os
import pyopencl as cl
from ..Param import *
from ParticularSolver import ParticularSolver
from ..Domain.ParticleField import ParticleField
from ..Variable.DiscreteVariable import DiscreteVariable
from ..Operator.RemeshingDOp import RemeshingDOp
from ..Operator.TagDOp import TagDOp
from ..Operator.InterpolationDOp import InterpolationDOp
from ..Operator.RemeshingDOp import RemeshingDOp


class GPUParticularSolver(ParticularSolver):
    """
    GPU Particular solver description.
    Link with differents numericals methods used. Prepare GPU side (memory, kernels, ...)
    """
    def __init__(self, problem,
                 ODESolver,
                 InterpolationMethod,
                 RemeshingMethod,
                 splittingMethod='strang',
                 platform_id=0, device_id=0,
                 device_type='gpu',
                 src=None):
        """
        Constructor.
        Create a solver description.
        """
        ParticularSolver.__init__(self, problem, ODESolver, InterpolationMethod, RemeshingMethod, splittingMethod)
        self.parameters_initialization()
        self.user_src = src
        #Get platform.
        self.platform = cl.get_platforms()[platform_id]
        #Get device.
        if device_type == 'gpu':
            self.device = self.platform.get_devices(cl.device_type.GPU)[device_id]
        else:
            self.device = self.platform.get_devices()[device_id]
        print "Running on", self.device.name, "of", self.platform.name, "platform."
        #Creates GPU Context
        self.ctx = cl.Context([self.device])
        #Create CommandQueue on the GPU Context
        self.queue = cl.CommandQueue(self.ctx, properties=cl.command_queue_properties.PROFILING_ENABLE)

    def parameters_initialization(self):
        # Modifying domains parameters to fit with float4 or int4 opencl objects
        self.gpu_shape = tuple(self.grid.elementNumber)
        padding_float = np.zeros(4 - self.grid.dimension, dtype=dtype_real_gpu)
        padding_uint = np.zeros(4 - self.grid.dimension, dtype=dtype_integer)
        self.grid.min = np.concatenate((self.grid.min, padding_float)).astype(dtype_real_gpu)
        self.grid.max = np.concatenate((self.grid.max, padding_float)).astype(dtype_real_gpu)
        self.grid.length = np.concatenate((self.grid.length, padding_float)).astype(dtype_real_gpu)
        self.grid.elementSize = np.concatenate((self.grid.elementSize, padding_float)).astype(dtype_real_gpu)
        self.grid.elementNumber = np.concatenate((self.grid.elementNumber, padding_uint + 1)).astype(dtype_integer)
        self.ppos.values.resize(self.gpu_shape, refcheck=False)

    def buffer_allocations(self):
        self.gpu_used_mem = 0
        for v in self.discreteProblem.variables:
            v.values = np.asarray(v.values, dtype=dtype_real_gpu)
            v.gpu_mem_object = cl.Buffer(self.ctx,
                                         cl.mem_flags.READ_WRITE,
                                         size=v.values.nbytes
                )
            self.gpu_used_mem += v.gpu_mem_object.size
            print "Allocation " + v.name + " on gpu, size:", v.gpu_mem_object.size * 1e-6, 'MBytes'

        if debug > 0:
            debug_size = 0
            self.discreteProblem.advecOp.debug_integer = np.zeros(debug, dtype=dtype_integer)
            self.discreteProblem.advecOp.debug_integer_buffer = cl.Buffer(
                self.ctx, cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR,
                hostbuf=self.discreteProblem.advecOp.debug_integer)
            debug_size += self.discreteProblem.advecOp.debug_integer_buffer.size
            self.discreteProblem.advecOp.debug_float = np.zeros(debug, dtype=dtype_real_gpu)
            self.discreteProblem.advecOp.debug_float_buffer = cl.Buffer(
                self.ctx, cl.mem_flags.READ_WRITE | cl.mem_flags.COPY_HOST_PTR,
                hostbuf=self.discreteProblem.advecOp.debug_float)
            debug_size += self.discreteProblem.advecOp.debug_float_buffer.size
            self.gpu_used_mem += debug_size
            print "Allocation debug buffer on gpu, size:", debug_size, 'B'

    def initialize(self):
        """
        Solver initialisation for a given DiscreteProblem.DiscreteProblem.

        @param discreteProblem : Problem to initialize.
        """
        # OpenCL Buffer allocations
        print "\n===    Device memory allocations    ==="
        self.buffer_allocations()
        p = self.gpu_used_mem * 100 / (self.device.global_mem_size * 1.)
        print "Total memory allocated on gpu :", self.gpu_used_mem * 1e-6, "MBytes (", p, " % of total available memory)"

        # Kernels creation/compilation
        print "\n===    Kernel sources compiling    ==="
        f = open(os.path.join(os.path.split(os.path.abspath(__file__))[0], "gpu_src.cl"), 'r')
        self.gpu_src = "".join(f.readlines())
        f.close()
        if not self.user_src is None:
            f = open(self.user_src, 'r')
            self.gpu_src += "".join(f.readlines())
            f.close()
        build_options = "-cl-single-precision-constant -cl-opt-disable"
        build_options += " -D DIM=" + str(self.dim)
        build_options += " -D DEBUG=" + str(debug)
        build_options += " -D MAX_LINE_POINTS=" + str(np.max(self.grid.elementNumber))
        self.prg = cl.Program(self.ctx, self.gpu_src).build(build_options)
        print self.prg.get_build_info(self.device, cl.program_build_info.LOG)
        print self.prg.get_build_info(self.device, cl.program_build_info.OPTIONS)
        print self.prg.get_build_info(self.device, cl.program_build_info.STATUS)

        for op in self.discreteProblem.operators:
            for k in self.prg.all_kernels():
                if op.gpu_kernel_name == k.get_info(cl.kernel_info.FUNCTION_NAME):
                    op.gpu_kernel = self.prg
                    op.gpu_queue = self.queue
                    op.gpu_shape = self.gpu_shape

        print "\n===    Data initialization    ==="
        for v in self.discreteProblem.variables:
            v_is_init = False
            print v.name,
            for i,k in enumerate(self.prg.all_kernels()):
                if ('init' + v.name) == k.get_info(cl.kernel_info.FUNCTION_NAME):
                    print 'Initialization Kernel found ',
                    evt = eval("self.prg."+"init"+v.name)(self.queue,
                                                    self.gpu_shape,
                                                    None,
                                                    v.gpu_mem_object,
                                                    dtype_real_gpu(0.),
                                                    v.domain.min.astype(dtype_real_gpu),
                                                    v.domain.elementNumber.astype(dtype_integer),
                                                    v.domain.elementSize.astype(dtype_real_gpu)
                        )
                    self.queue.finish()
                    print 'compute time :', (evt.profile.end - evt.profile.start) * 1e-9, 's',
                    v.contains_data, v_is_init = False, True
            if not v_is_init:
                v.init()
            print '... Done'

        print "\n===    Data Transfers host->device    ==="
        copy_evts = []
        transfered_size = 0
        for v in self.discreteProblem.variables:
            if v.contains_data:
                print v.name,
                if v.dimension == self.dim:
                    print "Memoire arrangée",
                    offset = 0
                    evt = cl.enqueue_copy(self.queue, v.gpu_mem_object,
                                          v.values[..., 0].ravel(),
                                          device_offset=np.long(offset))
                    copy_evts.append(evt)
                    offset += v.values[..., 0].nbytes
                    evt = cl.enqueue_copy(self.queue, v.gpu_mem_object,
                                          v.values[..., 1].swapaxes(1, 0).ravel(),
                                          device_offset=np.long(offset))
                    copy_evts.append(evt)
                    if self.dim == 3:
                        offset += v.values[..., 1].nbytes
                        evt = cl.enqueue_copy(self.queue, v.gpu_mem_object,
                                          v.values[..., 2].swapaxes(1, 0).swapaxes(2, 0).ravel(),
                                          device_offset=np.long(offset))
                        copy_evts.append(evt)
                else:
                    print "Normal layout",
                    evt = cl.enqueue_copy(self.queue, v.gpu_mem_object, v.values)
                    copy_evts.append(evt)
                print "Transfering", v.values.nbytes * 1e-6, "MBytes ..."
                transfered_size += v.values.nbytes
        self.queue.finish()
        if len(copy_evts)>0:
            t = 0.
            for evt in copy_evts:
                t += (evt.profile.end - evt.profile.start)
            print "Transfering host -> device :", t * 1e-9, "s \t", transfered_size / t, "GB/s"
        else:
            print "No transfers"

    def collect_data(self):
        self.queue.finish()
        print "\n===    Data Transfers device->host    ==="
        copy_evts = []
        transfered_size = 0
        for v in self.discreteProblem.variables:
            if v.contains_data:
                print v.name,
                if v.dimension == self.dim:
                    print "Memoire arrangée",
                    temp = np.empty(self.gpu_shape, dtype=dtype_real_gpu)
                    offset = 0
                    evt = cl.enqueue_copy(self.queue, temp, v.gpu_mem_object, device_offset=np.long(offset))
                    copy_evts.append(evt)
                    v.values[..., 0] = temp.reshape(self.gpu_shape)
                    offset += temp.nbytes
                    evt = cl.enqueue_copy(self.queue, temp, v.gpu_mem_object, device_offset=np.long(offset))
                    copy_evts.append(evt)
                    v.values[..., 1] = temp.reshape(self.gpu_shape).swapaxes(1, 0)
                    if self.dim == 3:
                        offset += temp.nbytes
                        evt = cl.enqueue_copy(self.queue, temp, v.gpu_mem_object, device_offset=np.long(offset))
                        copy_evts.append(evt)
                        v.values[..., 2] = temp.reshape(self.gpu_shape).swapaxes(2, 0).swapaxes(1, 0)
                else:
                    print "Normal layout",
                    evt = cl.enqueue_copy(self.queue, v.values, v.gpu_mem_object)
                    copy_evts.append(evt)
                print "Transfering", v.values.nbytes * 1e-6, "MBytes ..."
                transfered_size += v.values.nbytes
        self.queue.finish()
        t = 0.
        for evt in copy_evts:
            t += (evt.profile.end - evt.profile.start)
        print "Transfering device -> host :", t * 1e-9, "s \t", transfered_size / t, "GB/s"

    def terminate(self):
        self.queue.finish()

    def __str__(self):
        """ToString method"""
        s = "GPU  Particular solver "
        s += "\n   - ODESolver : " + str(self.ODESolver)
        s += "\n   - Interpolation : " + str(self.InterpolationMethod)
        s += "\n   - Remeshing : " + str(self.RemeshingMethod)
        s += "\n   - GPU informations : "
        s += "\n       - platform : " + self.platform.name
        s += " OpenCL version: " + self.platform.version
        s += "\n       - device : " + self.device.name
        s += " Global Memory available : " + str(self.device.global_mem_size // 1024 // 1024) + " MB  ( " + str(self.device.global_mem_size) + " B)"
        s += "\n                  Space index max : " + str(self.device.max_work_item_sizes)
        s += " Max workgroup size : " + str(self.device.max_work_group_size)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : GPUParticularSolver"
    print GPUParticularSolver.__doc__
