# -*- coding: utf-8 -*-
"""
@package Utils
GPU Particular solver description OpenGL rendering.
"""
import os
import sys
import pyopencl as cl
from ..Param import *
from ParticularSolver import ParticularSolver
from GPUParticularSolver import GPUParticularSolver
from ..Domain.ParticleField import ParticleField
from ..Variable.DiscreteVariable import DiscreteVariable
from ..Operator.RemeshingDOp import RemeshingDOp
from ..Operator.TagDOp import TagDOp
from ..Operator.InterpolationDOp import InterpolationDOp
from ..Operator.RemeshingDOp import RemeshingDOp
from ..Operator.VolumeDOp import VolumeDOp
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
import time


class GPUParticularSolver_GLRender(GPUParticularSolver):
    """
    GPU Particular solver description.
    Link with differents numericals methods used. Prepare GPU side (memory, kernels, ...)
    """
    def __init__(self, problem,
                 ODESolver,
                 InterpolationMethod,
                 RemeshingMethod,
                 splittingMethod='strang',
                 platform_id=0,
                 device_id=0,
                 device_type='gpu',
                 dt=1.,
                 src=None,
                 endTime=None):
        """
        Constructor.
        Create a solver description.
        """
        ParticularSolver.__init__(self, problem, ODESolver, InterpolationMethod, RemeshingMethod, splittingMethod)
        self.dt = dt
        self.parameters_initialization()
        self.user_src = src
        self.endTime = endTime
        if not endTime is None:
            self.endTime += self.dt * 1.5

        self.volume = DiscreteVariable(domain=self.grid, spec=self.gpu_shape[1:], name="Volume")
        self.volumeDOp = VolumeDOp(self.volume, self.pscal,
                                  np.prod(self.grid.elementSize[0:self.dim]), level=0.5)
        self.discreteProblem.addOperator(self.volumeDOp)
        self.discreteProblem.operator_list.append((self.volumeDOp, 1.0, 0))

        glutInit(sys.argv)
        glutInitWindowSize(self.width, self.height)
        glutInitWindowPosition(0, 0)
        glutCreateWindow('OpenCL/OpenGL Parmepy (' + str(np.prod(self.grid.elementNumber[0:self.grid.dimension])) + ' particles)')

        from pyopencl.tools import get_gl_sharing_context_properties
        if sys.platform == "darwin":
            self.ctx = cl.Context(properties=get_gl_sharing_context_properties(),
                    devices=[])
        else:
            # Some OSs prefer clCreateContextFromType, some prefer
            # clCreateContext. Try both.
            try:
                self.ctx = cl.Context(properties=[
                    (cl.context_properties.PLATFORM, platform)]
                    + get_gl_sharing_context_properties())
            except:
                self.ctx = cl.Context(properties=[
                    (cl.context_properties.PLATFORM, platform)]
                    + get_gl_sharing_context_properties(),
                    devices=[platform.get_devices()[0]])
        self.device = self.ctx.devices[0]
        self.queue = cl.CommandQueue(self.ctx, properties=cl.command_queue_properties.PROFILING_ENABLE)

        self.points = DiscreteVariable(domain=self.parts, dimension=self.dim)
        self.points.values = np.asarray(self.points.values, dtype=dtype_real_gpu)
        self.vbo_points = glGenBuffers(1)
        glBindBuffer(GL_ARRAY_BUFFER, self.vbo_points)
        glBufferData(GL_ARRAY_BUFFER, self.points.values.nbytes, None, GL_STREAM_DRAW)
        glEnableClientState(GL_VERTEX_ARRAY)
        glVertexPointer(self.dim, GL_FLOAT, 0, None)

        self.pcolor = DiscreteVariable(domain=self.parts, dimension=4)
        self.pcolor.values = np.asarray(self.pcolor.values, dtype=dtype_real_gpu)
        self.vbo_color = glGenBuffers(1)
        glBindBuffer(GL_ARRAY_BUFFER, self.vbo_color)
        glBufferData(GL_ARRAY_BUFFER, self.pcolor.values.nbytes, None, GL_STREAM_DRAW)
        glEnableClientState(GL_COLOR_ARRAY)
        glColorPointer(4, GL_FLOAT, 0, None)

    def parameters_initialization(self):
        GPUParticularSolver.parameters_initialization(self)
        #mouse handling for transforming scene
        self.mouse_down = False
        self.mouse_old = np.array([0., 0.])
        self.rotate = np.array([0., 0., 0.])
        self.translate = -(self.grid.min + self.grid.length / 2.)[0:3]
        self.initrans = np.array([0., 0., -2.])
        self.width, self.height = 800, 800

    def buffer_allocations(self):
        GPUParticularSolver.buffer_allocations(self)
        # GL buffer allocation
        self.points.gpu_mem_object = cl.GLBuffer(self.ctx,
            cl.mem_flags.READ_WRITE,
            int(self.vbo_points))
        self.gpu_used_mem += self.ppos.gpu_mem_object.size
        print "Allocation points colors to render, size:", self.ppos.gpu_mem_object.size, 'B'
        self.pcolor.gpu_mem_object = cl.GLBuffer(self.ctx,
            cl.mem_flags.READ_WRITE,
            int(self.vbo_color))
        self.gpu_used_mem += self.pscal.gpu_mem_object.size
        print "Allocation particles color on gpu, size:", self.pscal.gpu_mem_object.size, 'B'
        self.gl_objects = [self.points.gpu_mem_object, self.pcolor.gpu_mem_object]

###GL CALLBACKS
    def timer(self, t):
        glutTimerFunc(t, self.timer, t)
        glutPostRedisplay()

    def on_key(self, *args):
        ESCAPE = '\033'
        if args[0] == ESCAPE or args[0] == 'q':
            sys.exit()

    def on_click(self, button, state, x, y):
        if state == GLUT_DOWN:
            self.mouse_down = True
            self.button = button
        else:
            self.mouse_down = False
        self.mouse_old[0] = x
        self.mouse_old[1] = y

    def on_mouse_motion(self, x, y):
        dx = x - self.mouse_old[0]
        dy = y - self.mouse_old[0]
        if self.mouse_down and self.button == 0:  # left button
            self.rotate[0] += dy * .02
            self.rotate[1] += dx * .02
        elif self.mouse_down and self.button == 2:  # right button
            self.translate[2] -= dy * .001
        self.mouse_old[0] = x
        self.mouse_old[1] = y
    ###END GL CALLBACKS

    def display(self):
        if self.endTime is None or (not self.endTime is None and self.discreteProblem.t <= self.endTime):
            cl.enqueue_acquire_gl_objects(self.queue, self.gl_objects)
            display_step = time.time() - self.display_clock
            self.display_clock = time.time()
            t = time.time()
            self.discreteProblem.step(self.dt)
            l = list(self.gpu_shape)
            splitting_dir = 1
            nb_ligne = l.pop(splitting_dir)
            self.prg.colorize(self.queue, tuple(l), None,
                          self.pscal.gpu_mem_object,
                          self.pcolor.gpu_mem_object,
                          self.points.gpu_mem_object,
                          dtype_integer(splitting_dir),
                          dtype_integer(nb_ligne),
                          self.grid.min,
                          self.grid.elementNumber.astype(dtype_integer),
                          self.grid.elementSize
                )
            glutSetWindowTitle("OpenCL/OpenGL Parmepy T={0:8.4f} ( {1} particles)".format(
                self.discreteProblem.t, np.prod(self.grid.elementNumber[0:self.grid.dimension])))
            cl.enqueue_release_gl_objects(self.queue, self.gl_objects)
            self.queue.finish()
            print "Compute time :", (time.time() - t) * 1e3, "ms, \t over ", display_step * 1e3, "ms, \t", 1. / (time.time() - t), "fps"

            glFlush()
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
            glMatrixMode(GL_MODELVIEW)
            glLoadIdentity()

            #handle mouse transformations
            glTranslatef(self.initrans[0], self.initrans[1], self.initrans[2])
            glRotatef(self.rotate[0], 1, 0, 0)
            glRotatef(self.rotate[1], 0, 1, 0)  # we switched around the axis so make this rotate_z
            glTranslatef(self.translate[0], self.translate[1], self.translate[2])

            glEnable(GL_POINT_SMOOTH)
            glPointSize(1)
            glEnable(GL_BLEND)
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

            glBindBuffer(GL_ARRAY_BUFFER, self.vbo_color)
            glColorPointer(4, GL_FLOAT, 0, None)
            glBindBuffer(GL_ARRAY_BUFFER, self.vbo_points)
            glVertexPointer(self.dim, GL_FLOAT, 0, None)

            glEnableClientState(GL_VERTEX_ARRAY)
            glEnableClientState(GL_COLOR_ARRAY)
            glDrawArrays(GL_POINTS, 0, np.prod(self.gpu_shape))
            glDisableClientState(GL_COLOR_ARRAY)
            glDisableClientState(GL_VERTEX_ARRAY)

            #glDisable(GL_BLEND)
            orig = np.array([0, 0, 0])
            x_one = np.array([1, 0, 0])
            y_one = np.array([0, 1, 0])
            z_one = np.array([0, 0, 1])
            glColor3f(1, 0, 0)    # red
            self.draw_line(orig, x_one)
            glColor3f(0, 1, 0)    # green
            self.draw_line(orig, y_one)
            glColor3f(0, 0, 1)    # blue
            self.draw_line(orig, z_one)

    def draw_line(self, v1, v2):
        glBegin(GL_LINES)
        glVertex3f(v1[0], v1[1], v1[2])
        glVertex3f(v2[0], v2[1], v2[2])
        glEnd()

    def initialize(self):
        """
        Solver initialisation for a given DiscreteProblem.DiscreteProblem.

        @param discreteProblem : Problem to initialize.
        """
        GPUParticularSolver.initialize(self)

        glutDisplayFunc(self.display)
        glViewport(0, 0, self.width, self.height)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(np.max(self.grid.length) * 30., self.width / float(self.height), 0.1, 4.)
        glMatrixMode(GL_MODELVIEW)

        #handle user input
        glutKeyboardFunc(self.on_key)
        glutMouseFunc(self.on_click)
        glutMotionFunc(self.on_mouse_motion)

        #this will call draw every 10ms
        self.display_clock = time.time()
        self.display_time_step = 10
        glutTimerFunc(self.display_time_step, self.timer, self.display_time_step)

        glClearColor(1, 1, 1, 1)
        glColor(0, 0, 1)

        print "\n===    Solving    ==="
        print "Début de la boucle GLUT. ('q' ou 'ESC' pour quiter)"
        self.discreteProblem.solve = glutMainLoop()

    def __str__(self):
        """ToString method"""
        s = "GPU  Particular solver OpenGL rendering"
        s += "\n   - ODESolver : " + str(self.ODESolver)
        s += "\n   - Interpolation : " + str(self.InterpolationMethod)
        s += "\n   - Remeshing : " + str(self.RemeshingMethod)
        s += "\n   - GPU informations : "
        s += "\n       - platform : " + self.platform.name
        s += " OpenCL version: " + self.platform.version
        s += "\n       - device : " + self.device.name
        s += " Global Memory available : " + str(self.device.global_mem_size // 1024 // 1024) + " MB  ( " + str(self.device.global_mem_size) + " B)"
        s += "\n                  Space index max : " + str(self.device.max_work_item_sizes)
        s += " Max workgroup size : " + str(self.device.max_work_group_size)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : GPUParticularSolver_GLRender"
    print GPUParticularSolver_GLRender.__doc__
