# -*- coding: utf-8 -*-
"""
@package Operator
Operator representation
"""
from ..Param import *
from DiscreteOperator import DiscreteOperator
import time


class InterpolationDOp(DiscreteOperator):
    """
    Interpolation operator representation.
    DiscreteOperator.DiscreteOperator specialization.
    """

    def __init__(self, position, velocityField, resvelo, method):
        """
        Constructor.
        Create a Interpolation operator on a given discrete domain.
        Work on a position to interpolate velocity from a grid.

        @param position : particles positions.
        @param velocityField : grid velocity.
        @param resvelo DiscreteVariable.DiscreteVariable : new velocity
        @param method : Numerical method.
        """
        DiscreteOperator.__init__(self)
        ## Particles positions.
        self.ppos = position
        ## Grid velocity values.
        self.gvelo = velocityField
        ## Particles velocity
        self.resultVelocity = resvelo
        self.addVariable([self.ppos, self.gvelo, self.resultVelocity])
        self.numMethod = method
        self.compute_time = 0.
        self.queued_time = 0.
        self.submit_time = 0.
        self.total_flop = 0

    def setResultVariable(self, resvelo):
        """
        Set results variables

        @param resvelo DiscreteVariable.DiscreteVariable : new velocity
        """
        ## Particles velocity
        self.resultVelocity = resvelo
        self.addVariable([self.resultVelocity])

    def applyOperator(self, t, dt, splittingDirection):
        """
        Apply advection operator.

        @param t : current time.
        @param dt : time step for advection.
        @param splittingDirection : direction to advect.
        """
        #for i in self.gvelo.domain.explore():
        #    self.resultVelocity.values[i] = self.numMethod.interpolate(t, self.ppos.values[i], splittingDirection)
        c_time = time.time()
        ## ------------------- NumPy Optimisation
        self.resultVelocity.values = self.numMethod.interpolate(t, self.ppos.values, splittingDirection)
        ## -------------------
        self.compute_time += (time.time() - c_time)
        self.total_flop += self.numMethod.nb_flop * np.prod(self.gvelo.values.shape[0:-1])

    def __str__(self):
        """ToString method"""
        return "InterpolationDOp (DiscreteOperator)"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : InterpolationDOp"
    print InterpolationDOp.__doc__
