# -*- coding: utf-8 -*-
"""
@package Operator
Operator representation
"""
from DiscreteOperator import DiscreteOperator

providedClass = "TagDOp"


class TagDOp(DiscreteOperator):
    """
    Tag operator representation.
    DiscreteOperator.DiscreteOperator specialization.
    """

    def __init__(self, partPositions, partVelocities, grid, resbloc, restag):
        """
        Constructor.
        Create a Tag operator.
        Work on given particles positions and velocities to compute particles' type and tag.

        @param ppos : particles positions.
        @param pvelo : particles velocities.
        @param grid : underlying grid.
        @param resbloc DiscreteVariable.DiscreteVariable : particles bloc type.
        @param restag DiscreteVariable.DiscreteVariable : particles tag.
        """
        DiscreteOperator.__init__(self)
        ## Particles positions.
        self.ppos = partPositions
        ## Particles velocities.
        self.pvelo = partVelocities
        ## grid
        self.grid = grid
        ## Tag
        self.resultTag = restag
        ## Bloc
        self.resultBloc = resbloc
        self.addVariable([self.ppos, self.pvelo, self.resultBloc, self.resultTag])

    def setResultVariable(self, resbloc, restag):
        """
        Set results variables

        @param resbloc DiscreteVariable.DiscreteVariable : particles bloc type.
        @param restag DiscreteVariable.DiscreteVariable : particles tag.
        """
        ## Tag
        self.resultTag = restag
        ## Bloc
        self.resultBloc = resbloc
        self.addVariable([self.resultBloc, self.resultTag])

    def _nint(self, x):
        """
        Nearest interger function.

        @param x : real number.
        @return : x nearest integer.
        """
        if x > -0.5:
            return int(x + 0.5)
        else:
            return int(x + 0.5) - 1

    def applyOperator(self, t, dt, splittingDirection):
        """
        Apply advection operator.
        """
        #Used dimension
        d = splittingDirection
        bloc_size = 2
        for line in self.grid.explore(d):
            nb_bloc = len(line) / bloc_size
            d_bloc = bloc_size * self.grid.elementSize[d]
            eps_bl = self.grid.elementSize[d] * 0.000000001
            cfl = dt / self.grid.elementSize[d]
            max_V = self.pvelo.values[line[0]][d]
            for i in line:
                if max_V < self.pvelo.values[i][d]:
                    max_V = self.pvelo.values[i][d]
            bl_nbPart = [0 for i in xrange(nb_bloc)]
            bl_type = [0 for i in xrange(nb_bloc)]
            bl_ind = [0 for i in xrange(nb_bloc)]
            bl_final_ind = [0 for i in xrange(nb_bloc)]
            bl_lambdaMin = [max_V * cfl for i in xrange(nb_bloc)]
            p_ind_bl = {}
            for i in line:
                self.resultBloc.values[i] = 0
                self.resultTag.values[i] = -1
            # Update bloc infos
            for i in line:
                ind_bl = int((self.grid.points[i][d] - self.grid.min[d] + eps_bl) / d_bloc)
                bl_lambdaMin[ind_bl] = min([bl_lambdaMin[ind_bl], self.pvelo.values[i][d] * cfl])
                bl_nbPart[ind_bl] += 1
                bl_final_ind[ind_bl] = i
            # Compute bl_lambdaMin from next bloc
            for i in xrange(nb_bloc - 1):
                ind = tuple([bl_final_ind[i][dd] + 1 if dd == d else bl_final_ind[i][dd] for dd in xrange(len(line[0]))])
                bl_lambdaMin[i] = min([bl_lambdaMin[i], self.pvelo.values[ind][d] * cfl])
            # Last bloc
            bl_lambdaMin[nb_bloc - 1] = min([bl_lambdaMin[nb_bloc - 1], self.pvelo.values[line[0]][d] * cfl])
            # Bloc type
            for i in xrange(nb_bloc):
                ind = self._nint(bl_lambdaMin[i])
                #print i, ind, bl_lambdaMin[i]
                if bl_lambdaMin[i] - ind + eps_bl < 0.:
                    # Center bloc
                    #print "Centré"
                    bl_type[i] = 1
                    bl_ind[i] = ind + 1
                else:
                    # Left bloc
                    #print "Gauche"
                    bl_type[i] = 0
                    bl_ind[i] = ind
            # Update particles results
            for i in line:
                ind_bl = int((self.grid.points[i][d] - self.grid.min[d] + eps_bl) / d_bloc)
                p_ind_bl[i] = bl_ind[ind_bl]
                self.resultBloc.values[i] = bl_type[ind_bl]
            for i in xrange(len(line) - 1):
                if self.resultTag.values[line[i]] == -1 and (line[i][d] < (self.resultTag.domain.elementNumber[d] - 1)) and line[i][d] > 0:
                    if (p_ind_bl[line[i]] != p_ind_bl[line[i + 1]]) and (self.resultBloc.values[line[i]] != self.resultBloc.values[line[i + 1]]):
                        # tag particle i and i+1
                        print "Tag de la particule", line[i], "et", line[i + 1]
                        self.resultTag.values[line[i]] = 1
                        self.resultTag.values[line[i + 1]] = 2
                    else:
                        # No tag particle i
                        self.resultTag.values[line[i]] = 0
            if (p_ind_bl[line[0]] != p_ind_bl[line[-1]]) and (self.resultBloc.values[line[0]] != self.resultBloc.values[line[-1]]):
                # tag particle 0 and -1
                print "Tag de la particule", line[0], "et", line[-1]
                self.resultTag.values[line[0]] = 2
                self.resultTag.values[line[-1]] = 1
            else:
                # No tag particle 0 and -1
                self.resultTag.values[line[0]] = 0
                self.resultTag.values[line[-1]] = 0

    def __str__(self):
        """ToString method"""
        s = "TagDOp (DiscreteOperator) : {0} \n".format(id(self))
        s += "  particles positions = {0}, particles scalar = {1}\n".format(id(self.ppos), id(self.pscal))
        s += "  Variables : \n"
        for v in self.variables:
            s += str(id(v)) + "\n"
        s += "  Domains : \n"
        for d in self.domains:
            s += str(id(d)) + "\n"
        return s + "\n"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : " + providedClass
    print eval(providedClass).__doc__
