# -*- coding: utf-8 -*-
"""
@package Operator
Operator representation
"""
from ..Param import *
from DiscreteOperator import DiscreteOperator
import time
import pyopencl as cl

providedClass = "VelocityDOp"


class VelocityDOp(DiscreteOperator):
    """
    Advection operator representation.
    DiscreteOperator.DiscreteOperator specialization.
    """

    def __init__(self, velo_op):
        """
        Constructor.
        Create a Transport operator on a given continuous domain.
        Work on a given scalar at a given velocity to produce scalar distribution at new positions.

        @param advec : Advection operator.
        """
        DiscreteOperator.__init__(self)
        self.is_splittable = False
        ## Velocity.
        self.velocity = velo_op.velocity.discreteVariable
        ## Transported scalar.
        self.formula = velo_op.formula
        self.period = velo_op.period
        self.addVariable([self.velocity])
        self.gpu_kernel_name = "velocity"
        self.compute_time = [0., 0., 0.]
        self.submit_time = 0.
        self.queued_time = 0.
        self.total_flop = 0
        self.total_bytes_accessed = 0

    def applyOperator(self, t, dt, splittingDirection):
        """
        Apply advection operator.

        @param t : current time.
        @param dt : time step for advection.
        @param splittingDirection : direction to advect.
        """
        if self.gpu_kernel is None:
            raise NotImplementedError("No implementation for velocity opérator on CPU.")
        else:
            evt = self.gpu_kernel.velocity(self.gpu_queue, tuple(self.gpu_shape), None,
                    self.velocity.gpu_mem_object,
                    dtype_real_gpu(t), dtype_real_gpu(self.period),
                    self.velocity.domain.min.astype(dtype_real_gpu),
                    self.velocity.domain.elementNumber.astype(dtype_integer),
                    self.velocity.domain.elementSize.astype(dtype_real_gpu)
                    )
            # self.gpu_queue.finish()
            # cl.enqueue_copy(self.gpu_queue, self.resultPosition.values, self.resultPosition.gpu_mem_object)
            # print self.resultPosition.values
            self.gpu_queue.finish()
            #temp_tab = np.zeros_like(self.velocity.values)
            #cl.enqueue_copy(self.gpu_queue, temp_tab, self.velocity.gpu_mem_object)
            #self.gpu_queue.finish()
            #print temp_tab
            self.queued_time += (evt.profile.submit - evt.profile.queued) * 1e-9
            self.submit_time += (evt.profile.start - evt.profile.submit) * 1e-9
            self.compute_time[splittingDirection] += (evt.profile.end - evt.profile.start) * 1e-9
            self.total_flop += 32 * np.prod(self.gpu_shape)
            self.total_bytes_accessed += 4 * 2 * np.prod(self.gpu_shape)
            #print "Advection:", self.compute_time * 1e-9

    def __str__(self):
        """ToString method"""
        return "VelocityDOp (DiscreteOperator)"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : VelocityDOp"
    print VelocityDOp.__doc__
