# -*- coding: utf-8 -*-
"""
@package Operator
Operator representation
"""
from ..Param import *
from DiscreteOperator import DiscreteOperator
import time
import pyopencl as cl
import pylab as pl

providedClass = "VolumeDOp"


class VolumeDOp(DiscreteOperator):
    """
    Advection operator representation.
    DiscreteOperator.DiscreteOperator specialization.
    """

    def __init__(self, volume, pscal, pvolume, level=0.5, filename='./volume.dat'):
        """
        Constructor.
        Create a Transport operator on a given continuous domain.
        Work on a given scalar at a given velocity to produce scalar distribution at new positions.

        @param advec : Advection operator.
        """
        DiscreteOperator.__init__(self)
        self.is_splittable = False
        ## Volume.
        self.volume = volume
        self.pscal = pscal
        self.pvolume = pvolume
        self.addVariable([self.volume, self.pscal])
        self.level = level
        self.volume_evolution = []
        self.gpu_kernel_name = "volume"
        self.compute_time = [0., 0., 0.]
        self.submit_time = 0.
        self.queued_time = 0.
        self.total_flop = 0
        self.total_bytes_accessed = 0
        self.file = open(filename, 'w', 1)

    def applyOperator(self, t, dt, splittingDirection):
        """
        Apply advection operator.

        @param t : current time.
        @param dt : time step for advection.
        @param splittingDirection : direction to advect.
        """
        if self.gpu_kernel is None:
            raise NotImplementedError("No implementation for velocity opérator on CPU.")
        else:
            l = list(self.gpu_shape)
            nb_ligne = l.pop(splittingDirection)
            if debug == 0:
                evt = self.gpu_kernel.volume(self.gpu_queue, tuple(l), None,
                    self.pscal.gpu_mem_object,
                    self.volume.gpu_mem_object,
                    dtype_real_gpu(self.level), dtype_real_gpu(self.pvolume), dtype_integer(splittingDirection), dtype_integer(nb_ligne),
                    self.volume.domain.min.astype(dtype_real_gpu),
                    self.volume.domain.length.astype(dtype_real_gpu),
                    self.volume.domain.elementNumber.astype(dtype_integer),
                    self.volume.domain.elementSize.astype(dtype_real_gpu)
                    )
            else:
                evt = self.gpu_kernel.volume(self.gpu_queue, tuple(l), None,
                    self.pscal.gpu_mem_object,
                    self.volume.gpu_mem_object,
                    self.debug_float_buffer, self.debug_integer_buffer,
                    dtype_real_gpu(self.length), dtype_real_gpu(self.pvolume), dtype_integer(splittingDirection), dtype_integer(nb_ligne),
                    self.volume.domain.min.astype(dtype_real_gpu),
                    self.volume.domain.length.astype(dtype_real_gpu),
                    self.volume.domain.elementNumber.astype(dtype_integer),
                    self.volume.domain.elementSize.astype(dtype_real_gpu)
                    )
            self.gpu_queue.finish()
            cl.enqueue_copy(self.gpu_queue, self.volume.values, self.volume.gpu_mem_object)
            self.file.write(str(self.volume.values.sum()) + '\n')
            if debug > 0:
                cl.enqueue_copy(self.gpu_queue, self.debug_float, self.debug_float_buffer)
                cl.enqueue_copy(self.gpu_queue, self.debug_integer, self.debug_integer_buffer)
                print " :: DEBUG :: ", self.debug_float
                print " :: DEBUG :: ", self.debug_integer
            self.queued_time += (evt.profile.submit - evt.profile.queued) * 1e-9
            self.submit_time += (evt.profile.start - evt.profile.submit) * 1e-9
            self.compute_time[splittingDirection] += (evt.profile.end - evt.profile.start) * 1e-9
            self.total_flop += nb_ligne * np.prod(l)
            self.total_bytes_accessed += nb_ligne * np.prod(l) * 4
            #print "Advection:", self.compute_time * 1e-9

    def __str__(self):
        """ToString method"""
        return "VolumeDOp (DiscreteOperator)"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : VolumeDOp"
    print VolumeDOp.__doc__
