"""
Discrete obstacles description
@todo write test in parmepy/domain/test/test_obstacle.py
"""
from parmepy.constants import np


class Obstacle_d(object):
    """
    Discrete obstacles representation.
    """

    def __init__(self, parent, topology=None, name='sphere', idFromParent=None):
        """
        Creates discrete obsctacle and z-boundaries and returns 3 arrays corresponding
        to the three different porous media.
        porosity in the solid = 0
        porosity in the fluid = infinity

        @param parent : Continuous obstacle.
        @param topology : Topology informations
        @param name : Obstacle name
        @param idFromParent : Index in the parent's discrete obstacles
        """

        ## Topology of the obstacle
        if(topology is not None):
            self.topology = topology
        else:
            raise NotImplementedError()
        ## Name of the obstacle
        self.name = name
        ## Obstacle dimension
        self.dimension = topology.domain.dimension
        ## Obstacle resolution
        self.resolution = topology.mesh.resolution
        ## @private Continuous obstacle
        self.__parentObstacle = parent
        ## @private Index in parent's discrete obstacle
        self.__idFromParent = idFromParent
        ## Width of z-boundaries layer
        self.zlayer = self.__parentObstacle.zlayer
        ## Radius of the obstacle
        self.radius = self.__parentObstacle.radius
        ## Position of the obstacle center
        self.center = self.__parentObstacle.center
        ## Orientation of the obstacle
        self.orientation = self.__parentObstacle.orientation
        ## Thickness of the porous layer at the obstacle surface
        self.porousLayer = self.__parentObstacle.porousLayer
        ## Chi function (array) for z-boundaries
        self.chiBoundary = None
        ## Chi function (array) for the solid
        self.chiSolid = None
        ## Chi function (array) for the porous area
        self.chiPorous = None
        ## Determine characteristic function Chi of Obstacle 
        self.chiFunctions()

    def chiFunctions(self):
        """
        Compute the chi functions associated to z-boundaries and solid
        """
#        boolOrientation=[np.bool(0) for i in range (6)]
#        boolOrientation[self.indOrientation(self.orientation)]=True

        ## Temporary arrays
        chiBoundary_i = []
        chiBoundary_j = []
        chiBoundary_k = []
        chiSolid_i = []
        chiSolid_j = []
        chiSolid_k = []
        chiPorous_i = []
        chiPorous_j = []
        chiPorous_k = []

        step = self.topology.mesh.space_step_size
        ghosts = self.topology.ghosts
        coord_start = self.topology.mesh.origin
        coord_end = self.topology.mesh.end * step + self.topology.domain.origin - (ghosts * step)
        layerMin = coord_start[2] + ghosts[2] * step[2] + self.zlayer
        layerMax = coord_end[2] - ghosts[2] * step[2] - self.zlayer
        radiusMinuslayer = self.radius- self.porousLayer
        print 'start, end, layerMin, layerMax', coord_start, coord_end, layerMin, layerMax
        for k in xrange (ghosts[2], self.resolution[2] - ghosts[2]):
            rz = coord_start[2] + step[2] * k
            for j in xrange (ghosts[1], self.resolution[1] - ghosts[1]):
                ry = coord_start[1] + step[1] * j
                for i in xrange (ghosts[0], self.resolution[0] - ghosts[0]):
                    if (rz > layerMax or rz < layerMin):
                    # we are in the z-layer (North or South):
#                         chiBoundaryTmp.append([i,j,k])
                        chiBoundary_i.append(i)
                        chiBoundary_j.append(j)
                        chiBoundary_k.append(k)
                    else :
                        rx = coord_start[0] + step[0] * i
                        dist = np.sqrt((rx - self.center[0]) **2 +
                               (ry - self.center[1]) **2 +
                               (rz - self.center[2]) **2)
#                        if (self.name=='sphere'):
                        if (radiusMinuslayer < dist and dist <= self.radius + 1E-12 and self.porousLayer != 0.):
                            # we are in the porous area of the sphere:
#                                chiPorousTmp.append([i,j,k])
                            chiPorous_i.append(i)
                            chiPorous_j.append(j)
                            chiPorous_k.append(k)
                        if (dist <= radiusMinuslayer + 1E-12):
                            # we are in the solid area of the sphere:
#                                chiSolidTmp.append([i,j,k])
                            chiSolid_i.append(i)
                            chiSolid_j.append(j)
                            chiSolid_k.append(k)
#                        else :
#                            if (radiusMinuslayer < dist and dist <= self.radius and rx <= self.center[0]):
#                            # we are in the porous area of the hemisphere:
##                                chiPorousTmp.append([i,j,k])
#                                chiPorous_i.append(i)
#                                chiPorous_j.append(j)
#                                chiPorous_k.append(k)
#                            if (dist <= radiusMinuslayer and rx <= self.center[0]):
#                            # we are in the solid area of the hemisphere:
##                                chiSolidTmp.append([i,j,k])
#                                chiSolid_i.append(i)
#                                chiSolid_j.append(j)
#                                chiSolid_k.append(k)

        ## Characteristic function of penalized boundaries
        chiBoundary_i = np.asarray(chiBoundary_i)
        chiBoundary_j = np.asarray(chiBoundary_j)
        chiBoundary_k = np.asarray(chiBoundary_k)
        self.chiBoundary = chiBoundary(chiBoundary_i, chiBoundary_j, chiBoundary_k)

        ## Characteristic function of solid areas
        chiSolid_i = np.asarray(chiSolid_i)
        chiSolid_j = np.asarray(chiSolid_j)
        chiSolid_k = np.asarray(chiSolid_k)
        self.chiSolid = chiSolid(chiSolid_i, chiSolid_j, chiSolid_k)

        ## Characteristic function of porous areas
        chiPorous_i = np.asarray(chiPorous_i)
        chiPorous_j = np.asarray(chiPorous_j)
        chiPorous_k = np.asarray(chiPorous_k)
        self.chiPorous = chiPorous(chiPorous_i, chiPorous_j, chiPorous_k)


    def indOrientation(orientation):
        return {'West':0, 'East':1, 'South':2, 'North':3, 'Forward':4, 'Backward':5}.get(orientation, 0)

    def __str__(self):
        """ToString method"""
        return "Obstacle_d (discrete obstacles)"

if __name__ == "__main__" :
    print "This module defines the following classes:"
    print "Obstacle_d: ", Obstacle_d.__doc__
