"""
@package parmepy.particular_solvers.gpu

Particular solver description. Computations are performed on GPU device.
"""
from ..constants import *
from basic import ParticleSolver
from ..fields.continuous import ContinuousField
from ..operator.transport import Transport
from ..operator.remeshing import Remeshing
from ..operator.splitting import Splitting
from ..tools.timer import Timer
from ..tools.printer import Printer
import pyopencl as cl
from ..tools.gpu_data_transfer import hostToDevice
from ..tools.gpu_data_transfer import deviceToHost


class GPUParticleSolver(ParticleSolver):
    """
    GPU Particular solver description.

    Link with differents numericals methods used.
    Prepare GPU side (memory, kernels, ...)
    """
    def __init__(self, problem, t_end, dt,
                 ODESolver=None,
                 InterpolationMethod=None,
                 RemeshingMethod=None,
                 splittingMethod='strang',
                 timer=None,
                 io=None,
                 platform_id=0, device_id=0,
                 device_type='gpu',
                 src=None,
                 precision=PARMES_REAL_GPU):
        """
        @copydoc ParticleSolver.__init__()
        @param platform_id : OpenCL platform id to use.
        @param device_id : OpenCL device id to use.
        @param device_type='gpu' : OpenCL device type to use.
        @param src : OpenCL kernel sources given by user.

        Discover OpenCL environment.
        """
        ParticleSolver.__init__(self, problem, t_end, dt,
                                ODESolver=ODESolver,
                                InterpolationMethod=InterpolationMethod,
                                RemeshingMethod=RemeshingMethod,
                                splittingMethod=splittingMethod,
                                timer=timer,
                                io=io)
        ## User kernel sources filename
        self.user_gpu_src = src
        self.gpu_precision = precision
        print "=== OpenCL environment ==="
        #Get platform.
        try:
            ## OpenCL platform
            self.platform = cl.get_platforms()[platform_id]
        except IndexError:
            print "  Incorrect platform_id :", platform_id, ".",
            print " Only ", len(cl.get_platforms()), " available.",
            print " Getting defalut platform. "
            self.platform = cl.get_platforms()[0]
        print "  Platform   "
        print "  - Name       :", self.platform.name
        print "  - Version    :", self.platform.version
        #Get device.
        try:
            ## OpenCL device
            self.device = self.platform.get_devices(
                eval("cl.device_type." + device_type.upper()))[device_id]
        except cl.RuntimeError as e:
            print "RuntimeError:", e
            self.device = cl.create_some_context().devices[0]
        except AttributeError as e:
            print "AttributeError:", e
            self.device = cl.create_some_context().devices[0]
        print "  Device"
        print "  - Name                :",
        print self.device.name
        print "  - Type                :",
        print cl.device_type.to_string(self.device.type)
        print "  - C Version           :",
        print self.device.opencl_c_version
        print "  - Global mem size     :",
        print self.device.global_mem_size / (1024 ** 3), "GB"
        print "===\n"
        #Creates GPU Context
        ## OpenCL context
        self.ctx = cl.Context([self.device])
        #Create CommandQueue on the GPU Context
        ## OpenCL command queue
        self.queue = cl.CommandQueue(
            self.ctx, properties=cl.command_queue_properties.PROFILING_ENABLE)

    def initialize(self):
        """
        @copydoc ParticleSolver.initialize(self)
        Compile kernel sources. \n
        Allocate device memory as OpenCL Buffers.\n
        Initialize memory on device. \n
        Link kernels ans parmepy.operators.
        @todo : Take non square resolutions
        """
        ParticleSolver.initialize(self)
        # kernels compilation
        print "=== Kernel sources compiling ==="
        self.gpu_src = ""
        if self.user_gpu_src is not None:
            if isinstance(self.user_gpu_src, list):
                for user_src in self.user_gpu_src:
                    print "Sources files (user): ", user_src
                    f = open(user_src, 'r')
                    self.gpu_src += "".join(f.readlines())
                    f.close()
            else:
                print "Sources files (user): ", self.user_gpu_src
                f = open(self.user_gpu_src, 'r')
                self.gpu_src += "".join(f.readlines())
                f.close()
        resolution = self.problem.topology.mesh.resolution
        ## Optimal work item number
        if len(resolution) == 3:
            workItemNumber = 64
        else:
            workItemNumber = 256
        print " Work-Item number: ", workItemNumber
        ## Precision supported
        print " Precision capability  ",
        self.correctlyroundeddividesqrt = True
        if self.gpu_precision is np.float32:
            print "for Single Precision: "
            for v in ['DENORM', 'INF_NAN',
                      'ROUND_TO_NEAREST', 'ROUND_TO_ZERO', 'ROUND_TO_INF',
                      'FMA', 'CORRECTLY_ROUNDED_DIVIDE_SQRT', 'SOFT_FLOAT']:
                try:
                    if eval('(self.device.single_fp_config &' +
                            ' cl.device_fp_config.' +
                            v + ') == cl.device_fp_config.' + v):
                        print v
                except AttributeError as ae:
                    if v is 'CORRECTLY_ROUNDED_DIVIDE_SQRT':
                        self.correctlyroundeddividesqrt = False
                    print v, 'is not supported in OpenCL C 1.2.\n',
                    print '   Exception catched : ', ae
        else:
            print "for Double Precision: "
            if self.device.double_fp_config > 0:
                for v in ['DENORM', 'INF_NAN', 'FMA',
                          'ROUND_TO_NEAREST', 'ROUND_TO_ZERO', 'ROUND_TO_INF',
                          'CORRECTLY_ROUNDED_DIVIDE_SQRT', 'SOFT_FLOAT']:
                    try:
                        if eval('(self.device.double_fp_config &' +
                                ' cl.device_fp_config.' +
                                v + ') == cl.device_fp_config.' + v):
                            print v
                    except AttributeError as ae:
                        if v is 'CORRECTLY_ROUNDED_DIVIDE_SQRT':
                            self.correctlyroundeddividesqrt = False
                        print  v, 'is supported in OpenCL C 1.2.\n',
                        print '   Exception catched : ', ae
            else:
                raise ValueError("Double Precision is not supported by device")
        ## Building options and kernel settings
        build_options = ''
        if self.correctlyroundeddividesqrt:
            build_options += " -cl-fp32-correctly-rounded-divide-sqrt "
        if self.gpu_precision is np.float32:
            build_options += " -cl-single-precision-constant"
        build_options += " -D WIDTH=" + str(resolution[0])
        build_options += " -D WGN=" + str(workItemNumber)

        ## Collect OpenCL source code
        print "\n Source code:"
        ## advection
        if self.advection.include_remesh:
            if len(resolution) == 3:
                if self.gpu_precision is np.float32:
                    if self.RemeshingMethod is 'm6prime':
                        src_remesh_weights = 'weights_m6prime_opt4_builtin.cl'
                        src_advec_and_remesh = 'advection_and_remesh_3D_opt4_builtin_private4.cl'
                    else:
                        src_remesh_weights = 'weights_m8prime_opt4_builtin.cl'
                        src_advec_and_remesh = 'advection_and_remesh_m8prime_3D_opt4_builtin_private4.cl'
                else:
                    if self.RemeshingMethod is 'm6prime':
                        src_remesh_weights = 'weights_m6prime_opt4_builtin.cl'
                        src_advec_and_remesh = 'advection_and_remesh_3D_opt4_builtin_private4_noBC.cl'
                    else:
                        src_remesh_weights = 'weights_m8prime_opt4_builtin.cl'
                        src_advec_and_remesh = 'advection_and_remesh_m8prime_3D_opt4_builtin_private4_noBC.cl'
                advec_and_remesh_gwi = (int(workItemNumber), int(resolution[1]), int(resolution[2]))
                advec_and_remesh_lwi = (int(workItemNumber), 1, 1)
            else:
                if self.gpu_precision is np.float32:
                    if self.RemeshingMethod is 'm6prime':
                        src_remesh_weights = 'weights_m6prime_builtin.cl'
                        src_advec_and_remesh = 'advection_and_remesh_2D_opt8_builtin_noBC.cl'
                    else:
                        src_remesh_weights = 'weights_m8prime_builtin.cl'
                        src_advec_and_remesh = 'advection_and_remesh_m8prime_2D_opt8_builtin_noBC.cl'
                else:
                    if self.RemeshingMethod is 'm6prime':
                        src_remesh_weights = 'weights_m6prime_builtin.cl'
                        src_advec_and_remesh = 'advection_and_remesh_2D_opt4_builtin_noBC.cl'
                    else:
                        src_remesh_weights = 'weights_m8prime_builtin.cl'
                        src_advec_and_remesh = 'advection_and_remesh_m8prime_2D_opt4_builtin_noBC.cl'
                advec_and_remesh_gwi = (int(workItemNumber), int(resolution[1]))
                advec_and_remesh_lwi = (int(workItemNumber), 1)
            f = open(os.path.join(GPU_SRC, src_remesh_weights))
            self.gpu_src += "".join(f.readlines())
            print  "   - remeshing weights:", f.name
            f.close()
            f = open(os.path.join(GPU_SRC, src_advec_and_remesh))
            print  "   - advection and remesh:", f.name
            self.gpu_src += "".join(f.readlines())
            f.close()
        else:
            if len(resolution) == 3:
                if self.gpu_precision is np.float32:
                    src_advec = 'advection_3D_opt4_builtin.cl'
                else:
                    src_advec = 'advection_3D_opt2_builtin.cl'
                advec_gwi = (int(workItemNumber), int(resolution[1]), int(resolution[2]))
                advec_lwi = (int(workItemNumber), 1, 1)
            else:
                if self.gpu_precision is np.float32:
                    src_advec = 'advection_2D_opt4_builtin.cl'
                else:
                    src_advec = 'advection_2D_builtin.cl'
                advec_gwi = (int(workItemNumber), int(resolution[1]))
                advec_lwi = (int(workItemNumber), 1)
            f = open(os.path.join(GPU_SRC, src_advec))
            print  "   - advection:", f.name
            self.gpu_src += "".join(f.readlines())
            f.close()
            ## remeshing
            if len(resolution) == 3:
                if self.RemeshingMethod is 'm6prime':
                    src_remesh = 'remeshing_3D_opt4_private4.cl'
                    src_remesh_weights = 'weights_m6prime_opt4_builtin.cl'
                else:
                    src_remesh = 'remeshing_m8prime_3D_opt4_private4.cl'
                    src_remesh_weights = 'weights_m8prime_opt4_builtin.cl'
                remesh_gwi = (int(workItemNumber), int(resolution[1]), int(resolution[2]))
                remesh_lwi = (int(workItemNumber), 1, 1)
            else:
                if self.RemeshingMethod is 'm6prime':
                    src_remesh = 'remeshing_2D_opt4_noBC.cl'
                    src_remesh_weights = 'weights_m6prime_builtin.cl'
                else:
                    src_remesh = 'remeshing_m8prime_2D_opt4_noBC.cl'
                    src_remesh_weights = 'weights_m8prime_builtin.cl'
                remesh_gwi = (int(workItemNumber), int(resolution[1]))
                remesh_lwi = (int(workItemNumber), 1)
            f = open(os.path.join(GPU_SRC, src_remesh_weights))
            self.gpu_src += "".join(f.readlines())
            print  "   - remeshing weights:", f.name
            f.close()
            f = open(os.path.join(GPU_SRC, src_remesh))
            print  "   - remeshing:", f.name
            self.gpu_src += "".join(f.readlines())
            f.close()
        ## copy
        if len(resolution) == 3:
            if self.gpu_precision is np.float32:
                src_copy = 'copy_3D_opt4.cl'
                build_options += " -D TILE_DIM_COPY=16 -D BLOCK_ROWS_COPY=8"
                copy_gwi = (int(resolution[0] / 4), int(resolution[1] / 2), int(resolution[2]))
                copy_lwi = (4, 8, 1)
            else:
                src_copy = 'copy_3D_locMem.cl'
                build_options += " -D TILE_DIM_COPY=32 -D BLOCK_ROWS_COPY=8"
                copy_gwi = (int(resolution[0]), int(resolution[1] / 4), int(resolution[2]))
                copy_lwi = (32, 8, 1)
        else:
            if self.gpu_precision is np.float32:
                src_copy = 'copy_2D_opt2.cl'
                build_options += " -D TILE_DIM_COPY=16 -D BLOCK_ROWS_COPY=8"
                copy_gwi = (int(resolution[0] / 2), int(resolution[1] / 2))
                copy_lwi = (8, 8)
            else:
                src_copy = 'copy_2D_opt2.cl'
                build_options += " -D TILE_DIM_COPY=32 -D BLOCK_ROWS_COPY=2"
                copy_gwi = (int(resolution[0] / 2), int(resolution[1] / 16))
                copy_lwi = (16, 2)
        f = open(os.path.join(GPU_SRC, src_copy))
        print  "   - copy:", f.name
        self.gpu_src += "".join(f.readlines())
        f.close()
        ## transpose
        if len(resolution) == 3:
            if self.gpu_precision is np.float32:
                src_transpose_xy = 'transpose_3D_xy_coalesced_locPad_Diag_2Dslice_opt2.cl'
                build_options += " -D TILE_DIM_XY=16 -D BLOCK_ROWS_XY=8"
                transpose_xy_gwi = (int(resolution[0] / 2), int(resolution[1] / 2), int(resolution[2]))
                transpose_xy_lwi = (8, 8, 1)
                src_transpose_xz = 'transpose_3D_xz_coalesced_locPad_Diag_bis_3DBlock.cl'
                build_options += " -D TILE_DIM_XZ=16 -D BLOCK_ROWS_XZ=4"
                transpose_xz_gwi = (int(resolution[0]), int(resolution[1] / 4), int(resolution[2] / 4))
                transpose_xz_lwi = (16, 4, 4)
            else:
                src_transpose_xy = 'transpose_3D_xy_coalesced_locPad_Diag_2Dslice_opt4.cl'
                build_options += " -D TILE_DIM_XY=32 -D BLOCK_ROWS_XY=4"
                transpose_xy_gwi = (int(resolution[0] / 4), int(resolution[1] / 8), int(resolution[2]))
                transpose_xy_lwi = (8, 4, 1)
                src_transpose_xz = 'transpose_3D_xz_coalesced_Diag_bis_3DBlock.cl'
                build_options += " -D TILE_DIM_XZ=8 -D BLOCK_ROWS_XZ=2"
                transpose_xz_gwi = (int(resolution[0]), int(resolution[1] / 4), int(resolution[2] / 4))
                transpose_xz_lwi = (8, 2, 2)
        else:
            if self.gpu_precision is np.float32:
                src_transpose_xy = 'transpose_2D_xy_coalesced_locPad_Diag_opt4.cl'
                build_options += " -D TILE_DIM_XY=32 -D BLOCK_ROWS_XY=8"
                transpose_xy_gwi = (int(resolution[0] / 4), int(resolution[1]) / 4)
                transpose_xy_lwi = (8, 8)
            else:
                src_transpose_xy = 'transpose_2D_xy_coalesced_locPad_Diag_opt4.cl'
                build_options += " -D TILE_DIM_XY=32 -D BLOCK_ROWS_XY=2"
                transpose_xy_gwi = (int(resolution[0] / 4), int(resolution[1]) / 16)
                transpose_xy_lwi = (8, 2)
        f = open(os.path.join(GPU_SRC, src_transpose_xy))
        print  "   - transpose_xy:", f.name
        self.gpu_src += "".join(f.readlines())
        f.close()
        if len(resolution) == 3:
            f = open(os.path.join(GPU_SRC, src_transpose_xz))
            self.gpu_src += "".join(f.readlines())
            f.close()

        ## Build code
        if self.gpu_precision is np.float32:
            prg = cl.Program(self.ctx, self.gpu_src.replace('.0', '.0f'))
        else:
            prg = cl.Program(self.ctx, self.gpu_src.replace('float', 'double'))

        ## OpenCL program
        self.prg = prg.build(build_options)
        print "Compiler options : ",
        print self.prg.get_build_info(self.device, cl.program_build_info.OPTIONS)
        print "Compiler status : ",
        print self.prg.get_build_info(self.device, cl.program_build_info.STATUS)
        print "Compiler log : ",
        print self.prg.get_build_info(self.device, cl.program_build_info.LOG)
        print "===\n"
        # Convert fields to recquired precision (self.gpu_precision) and ensure order
        for f in self.problem.variables:
            for df in f.discreteField:
                if f.vector:
                    for d in xrange(len(df.data)):
                        df.data[d] = np.asarray(df.data[d], dtype=self.gpu_precision, order='F')
                else:
                    df.data = np.asarray(df.data, dtype=self.gpu_precision, order='F')
        # Create OpenCL Buffers from fields
        print "=== OpenCL Buffer allocations ==="
        total_mem_used = 0
        for f in self.problem.variables:
            print f.name, "allocate ",
            temp_mem_used = 0
            for df in f.discreteField:
                if f.vector:
                    for d in xrange(len(df.data)):
                        df.gpu_data[d] = cl.Buffer(self.ctx,
                                                   cl.mem_flags.READ_WRITE,
                                                   size=df.data[d].nbytes)
                        temp_mem_used += df.gpu_data[d].size
                else:
                    df.gpu_data = cl.Buffer(self.ctx,
                                                cl.mem_flags.READ_WRITE,
                                                size=df.data.nbytes)
                    temp_mem_used += df.gpu_data.size
            total_mem_used += temp_mem_used
            print temp_mem_used, "Bytes (", temp_mem_used / (1024 ** 2), "MB)"
        print "Total Global Memory used : ", total_mem_used, "Bytes (", total_mem_used / (1024 ** 2), "MB)",
        print "({0:.3f} %)".format(100 * total_mem_used / (self.device.global_mem_size * 1.))
        print "===\n"
        print "=== OpenCL Buffer initialisation ==="
        data, transfer_time, compute_time = 0, 0., 0.
        for f in self.problem.variables:
            initKernel = None
            # Looking for initKernel
            for k in self.prg.all_kernels():
                if k.get_info(cl.kernel_info.FUNCTION_NAME) == ('init' + f.name):
                    initKernel = cl.Kernel(self.prg, 'init' + f.name)
            if initKernel is not None:
                for df in f.discreteField:
                    print f.name, ": kernel init",
                    global_wg = df.resolution.tolist()
                    if global_wg[0] / workItemNumber > 0:
                        global_wg[0] = workItemNumber
                        if len(global_wg) == 3:
                            local_wg = (workItemNumber, 1, 1)
                        else:
                            local_wg = (workItemNumber, 1)
                    else:
                        local_wg = None
                    global_wg = tuple(global_wg)
                    print global_wg, local_wg,
                    dim = df.topology.dim
                    coord_min = np.ones(4, dtype=self.gpu_precision)
                    mesh_size = np.ones(4, dtype=self.gpu_precision)
                    coord_min[0:dim] = df.topology.mesh.origin
                    mesh_size[0:dim] = df.topology.mesh.size
                    print "...",
                    if f.vector:
                        if dim == 3:
                            evt = initKernel(self.queue,
                                             global_wg,
                                             local_wg,
                                             df.gpu_data[XDIR], df.gpu_data[YDIR], df.gpu_data[ZDIR],
                                             coord_min,
                                             mesh_size)
                        else:
                            evt = initKernel(self.queue,
                                             global_wg,
                                             local_wg,
                                             df.gpu_data[XDIR], df.gpu_data[YDIR],
                                             coord_min,
                                             mesh_size)
                    else:
                        evt = initKernel(self.queue,
                                         global_wg,
                                         local_wg,
                                         df.gpu_data,
                                         coord_min,
                                         mesh_size)
                    self.queue.finish()
                    temp_time = (evt.profile.end - evt.profile.start) * 1e-9
                    print "Done in ", temp_time, "sec"
                    compute_time += temp_time
            else:
                for df in f.discreteField:
                    if df.contains_data:
                        print f.name, ":",
                        temp_data, temp_time = hostToDevice(self.queue, df)
                        data += temp_data
                        transfer_time += temp_time
        if data > 0:
            print "Total Transfers : ", data, "Bytes transfered at {0:.3f} GBytes/sec".format((data * 1e-9) / transfer_time)
        if compute_time > 0.:
            print "Total Computing  : ", compute_time, "sec"
        print "===\n"
        # Setting advection and remeshing kernels:
        print "=== OpenCL Kernels setting ==="
        for op in self.problem.operators:
            if op.discreteOperator.name == "splitting":
                for sop in op.operators:
                    print sop.discreteOperator.name,
                    sop.discreteOperator.gpu_precision = self.gpu_precision
                    if sop.discreteOperator.name == 'advection':
                        sop.setMethod(KernelLauncher(self.prg.advection,
                                                     self.queue,
                                                     advec_gwi,
                                                     advec_lwi))
                    elif sop.discreteOperator.name == 'remeshing':
                        sop.setMethod(KernelLauncher(self.prg.remeshing,
                                                     self.queue,
                                                     remesh_gwi,
                                                     remesh_lwi))
                    elif sop.discreteOperator.name == 'advection_and_remeshing':
                        sop.setMethod(KernelLauncher(self.prg.advection_and_remeshing,
                                                 self.queue,
                                                 advec_and_remesh_gwi,
                                                 advec_and_remesh_lwi))
                    else:
                        raise ValueError("Unknown operator : " + sop.discreteOperator.name)
            else:
                raise ValueError("Unknown operator : " + op.discreteOperator.name)
        print self.advection.discreteOperator.name,
        self.advection.discreteOperator.init_copy = KernelLauncher(self.prg.copy,
                                                                   self.queue,
                                                                   copy_gwi,
                                                                   copy_lwi)
        if len(resolution) == 3:
            print self.advection.discreteOperator.name,
            k = KernelListLauncher([self.prg.transpose_xy,
                                    self.prg.transpose_xz],
                                   self.queue,
                                   [transpose_xy_gwi,
                                    transpose_xz_gwi],
                                   [transpose_xy_lwi,
                                    transpose_xz_lwi])
            self.advection.discreteOperator.init_transpose = k
        else:
            print self.advection.discreteOperator.name,
            k = KernelLauncher(self.prg.transpose_xy,
                               self.queue,
                               transpose_xy_gwi,
                               transpose_xy_lwi)
            self.advection.discreteOperator.init_transpose = k
        self.problem.io.set_get_data_method(self.get_data_from_device)
        print "===\n"

    def end(self):
        print "=== OpenCL Buffer deallocations ==="
        for f in self.problem.variables:
            print f.name, "deallocate "
            temp_mem_used = 0
            for df in f.discreteField:
                if f.vector:
                    for d in xrange(len(df.data)):
                        if isinstance(df.gpu_data[d], cl.MemoryObject):
                            df.gpu_data[d].release()
                else:
                    if isinstance(df.gpu_data, cl.MemoryObject):
                        df.gpu_data.release()
        print "===\n"

    def get_data_from_device(self, field):
        """Data collect method."""
        deviceToHost(self.queue, field)

    def __str__(self):
        """ToString method"""
        s = " Particular solver "
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : ParticularSolver"
    print ParticularSolver.__doc__


class KernelListLauncher:
    """
    OpenCL kernel list launcher.

    Manage launching of OpenCL kernels as a list.
    """
    def __init__(self, kernel, queue, gsize=None, lsize=None):
        """
        Create a kernel list launcher.
        @param kernel : kernel list.
        @param queue : OpenCL command queue.
        @param gsize : OpenCL global size index.
        @param lsize : OpenCL local size index.
        """
        ## OpenCL Kernel list
        self.kernel = kernel
        print [k.function_name for k in self.kernel]
        ## OpenCL command queue
        self.queue = queue
        ## OpenCL global size index.
        self.global_size = gsize
        ## OpenCL local size index.
        self.local_size = lsize
        self.call_number = [0 for k in self.kernel]

    def launch(self, d, *args):
        """
        Launch a kernel.

        @param d : kernel index in kernel list.
        @param args : kernel arguments.
        @return OpenCL Event

        OpenCL global size and local sizes are not given in args. Class member are used.
        """
        return KernelListLauncher.launch_sizes_in_args(self, d, self.global_size[d], self.local_size[d], *args)

    def launch_sizes_in_args(self, d, *args):
        """
        Launch a kernel.

        @param d : kernel index in kernel list.
        @param args : kernel arguments.
        @return OpenCL Event.

        Opencl global and local sizes are given in args.
        """
        #print self.kernel[d].function_name, args[0], args[1]
        self.call_number[d] += 1
        evt = self.kernel[d](self.queue, *args)
        return evt

    def finish(self):
        self.queue.finish()

    def function_name(self, d=None):
        """Prints OpenCL Kernels function names informations"""
        if d is not None:
            return self.kernel[d].get_info(cl.kernel_info.FUNCTION_NAME)
        else:
            return [k.get_info(cl.kernel_info.FUNCTION_NAME) for k in self.kernel]


class KernelLauncher(KernelListLauncher):
    """
    OpenCL kernel launcher.

    Manage launching of one OpenCL kernel as a KernelListLauncher with a list of one kernel.
    """
    def __init__(self, kernel, queue, gsize, lsize):
        """
        Create a KernelLauncher.

        @param kernel : kernel.
        @param queue : OpenCL command queue.
        @param gsize : OpenCL global size index.
        @param lsize : OpenCL local size index.

        Create a KernelListLauncher with a list of one kernel.
        """
        KernelListLauncher.__init__(self, [kernel], queue, [gsize], [lsize])

    def launch_sizes_in_args(self, *args):
        """
        Launch the kernel.

        @param args : kernel arguments.
        @return OpenCL Event.

        Opencl global and local sizes are given in args.
        """
        return KernelListLauncher.launch_sizes_in_args(self, 0, *args)

    def finish(self):
        """Wrapping OpenCL queue.finish() method."""
        self.queue.finish()

    def launch(self, *args):
        """
        Launch the kernel.

        @param args : kernel arguments.
        @return OpenCL Event

        OpenCL global size and local sizes are not given in args. Class member are used.
        """
        return KernelListLauncher.launch(self, 0, *args)

    def function_name(self):
        """Prints OpenCL Kernel function name informations"""
        res = KernelListLauncher.function_name(self, 0)
        return res
