/**
 * 2D advection kernel.
 *
 * Memory reads and writes are performed using 8 element OpenCL vectors (float8 or double8).
 * Computations are done using builtin functions on vector types.
 * A local array is used un local memory as a cache for velocity.
 *
 * @param gvelo Velocity.
 * @param ppos Particle position.
 * @param dt Time step.
 * @param min_position Domain lower point.
 * @param dx Space step.
 */
__kernel void advection(__global const float* gvelo,
			__global float* ppos,
			float dt, float min_position, float dx)
{
  uint gidX = get_global_id(0);
  uint gidY = get_global_id(1);
  int8 ind, i_ind, i_ind_p;
  float invdx = 1.0/dx;
  uint i;
  float8 v, vp, y, p, coord;
  __local float velocity_cache[WIDTH];

  for(i=gidX*8; i<WIDTH; i+=WGN*8)
    {
      v = vload8((i+gidY*WIDTH)/8, gvelo);
      velocity_cache[i] = v.s0;
      velocity_cache[i+1] = v.s1;
      velocity_cache[i+2] = v.s2;
      velocity_cache[i+3] = v.s3;
      velocity_cache[i+4] = v.s4;
      velocity_cache[i+5] = v.s5;
      velocity_cache[i+6] = v.s6;
      velocity_cache[i+7] = v.s7;
    }

  barrier(CLK_LOCAL_MEM_FENCE);

  for(i=gidX*8; i<WIDTH; i+=WGN*8)
    {
      v = vload8(i/8, velocity_cache);
      coord = (float8)(i*dx,
		       (i+1)*dx,
		       (i+2)*dx,
		       (i+3)*dx,
		       (i+4)*dx,
		       (i+5)*dx,
		       (i+6)*dx,
		       (i+7)*dx);
      p = fma((float8)(0.5*dt),v,coord);
      ind = convert_int8_rtn(p * invdx);
      y = (fma(- convert_float8(ind),dx,p))* invdx ;
      i_ind = ((ind + WIDTH) % WIDTH);
      i_ind_p = ((i_ind + 1) % WIDTH);
      v = (float8)(velocity_cache[i_ind.s0],
                   velocity_cache[i_ind.s1],
                   velocity_cache[i_ind.s2],
                   velocity_cache[i_ind.s3],
                   velocity_cache[i_ind.s4],
                   velocity_cache[i_ind.s5],
                   velocity_cache[i_ind.s6],
                   velocity_cache[i_ind.s7]);
      vp = (float8)(velocity_cache[i_ind_p.s0],
                    velocity_cache[i_ind_p.s1],
                    velocity_cache[i_ind_p.s2],
                    velocity_cache[i_ind_p.s3],
                    velocity_cache[i_ind_p.s4],
                    velocity_cache[i_ind_p.s5],
                    velocity_cache[i_ind_p.s6],
                    velocity_cache[i_ind_p.s7]);
      p = fma(mix(v,vp,y),dt,coord + (float8)(min_position));
      vstore8(p, (i+gidY*WIDTH)/8, ppos);
    }
}
