/**
 * 3D advection kernel.
 *
 * Memory reads and writes are performed using 2 element OpenCL vectors (float2 or double2).
 * Computations are done using builtin functions on vector types.
 * A local array is used un local memory as a cache for velocity.
 *
 * @param gvelo Velocity.
 * @param ppos Particle position.
 * @param dt Time step.
 * @param min_position Domain lower point.
 * @param dx Space step.
 */
__kernel void advection(__global const float* gvelo,
			__global float* ppos,
			float dt, float min_position, float dx)
{
  uint gidX = get_global_id(0);
  uint gidY = get_global_id(1);
  uint gidZ = get_global_id(2);
  int2 ind, i_ind, i_ind_p;
  float invdx = 1.0/dx;
  uint i;
  float2 v, vp, y, p, coord;
  __local float velocity_cache[WIDTH];

  for(i=gidX*2; i<WIDTH; i+=WGN*2)
    {
      v = vload2((i+gidY*WIDTH+gidZ*WIDTH*WIDTH)/2, gvelo);
      velocity_cache[i] = v.x;
      velocity_cache[i+1] = v.y;
    }

  barrier(CLK_LOCAL_MEM_FENCE);

  for(i=gidX*2; i<WIDTH; i+=WGN*2)
    {
      v = vload2(i/2, velocity_cache);
      coord = (float2)(i*dx, (i+1)*dx);
      p = fma((float2)(0.5*dt),v,coord);
      ind = convert_int2_rtn(p * invdx);
      y = (fma(- convert_float2(ind),dx,p))* invdx ;
      i_ind = ((ind + WIDTH) % WIDTH);
      i_ind_p = ((i_ind + 1) % WIDTH);
      v = (float2)(velocity_cache[i_ind.x], velocity_cache[i_ind.y]);
      vp = (float2)(velocity_cache[i_ind_p.x], velocity_cache[i_ind_p.y]);
      p=fma(mix(v,vp,y),dt,coord + (float2)(min_position));
      vstore2(p, (i+gidY*WIDTH+gidZ*WIDTH*WIDTH)/2, ppos);
    }
}
