# -*- coding: utf-8 -*-
"""
@package Utils
InterpolationMethod interface.
"""
from ..Param import *
from InterpolationMethod import InterpolationMethod


providedClass = "Linear"


class Linear(InterpolationMethod):
    """
    Lambda1 remeshing method implementation.
    """
    def __init__(self, grid, gvalues):
        """
        Constructor.

        @param grid : Grid values
        @param gvalues : Values to interpolate
        """
        InterpolationMethod.__init__(self)
        self.grid = grid
        self.gvalues = gvalues
        self.dim_range = xrange(self.grid.dimension)
        self.nb_flop = 9

    def interpolate(self, t, ppos, dir):
        """
        Abstract method, apply operaton on a variable.
        Must be implemented by sub-class.

        @param t : current time
        @param ppos : particle position.
        @param dir : interpolation direction
        @return particle velocity, velocity at position ppos.
        """
        ind = np.empty(self.grid.shape, dtype=dtype_integer)
        a0 = ((ppos - self.grid.min) / self.grid.elementSize)
        ind[...] = (np.round(np.abs(a0))) * np.sign(a0)
        ind[..., dir] = np.floor(a0[..., dir])
        #ind[...] = (np.round(np.abs(a0))) * np.sign(a0)
        index = [ind[..., i] for i in self.dim_range]
        index[dir] = (index[dir] + self.grid.elementNumber[dir]) % self.grid.elementNumber[dir]
        y = (ppos[..., dir] - self.grid[index][..., dir]) / self.grid.elementSize[dir]
        res = np.copy(self.gvalues.values)
        res[..., dir] = res[..., dir] * (1. - y)
        index[dir] = (index[dir] + 1) % self.grid.elementNumber[dir]
        res[..., dir] += self.gvalues.values[index][..., dir] * y
        return res

    def __call__(self, ppos, t, dir):
        return self.interpolate(ppos, t, dir)

    def __str__(self):
        """ToString method"""
        return "Linear"


if __name__ == "__main__":
    print __doc__
    print "- Provided class : " + providedClass
    print eval(providedClass).__doc__


class Linear2D(InterpolationMethod):
    """
    Lambda1 remeshing method implementation 2D.
    """
    def __init__(self, grid, gvalues):
        """
        Constructor.
        """
        self.grid = grid
        self.gvalues = gvalues

    def interpolate(self, t, ppos):
        """
        Linear interpolation.
        Interpolate the velocity at the given particle position.

        @param position : position to compute
        @param grid Grid.Grid : grid data to interpolate
        @return velocity : velocity at particle position
        """
        ii = xrange(len(ppos))
        # Calcul des indices
        ind = [int((ppos[i] - self.grid.min[i]) / self.grid.elementSize[i]) for i in ii]
        iXY = [(ind[i] % self.grid.elementNumber[i]) for i in ii]
        iXpY = [iXY[i] for i in ii]
        iXpY[0] = (iXpY[0] + 1) % self.grid.elementNumber[0]
        iXYp = [iXY[i] for i in ii]
        iXYp[1] = (iXYp[1] + 1) % self.grid.elementNumber[1]
        iXpYp = [iXY[i] for i in ii]
        iXpYp[0] = (iXpYp[0] + 1) % self.grid.elementNumber[0]
        iXpYp[1] = (iXpYp[1] + 1) % self.grid.elementNumber[1]
        # Calcul de la distance
        y = [((ppos[i] - self.grid[iXY][i]) / self.grid.elementSize[i]) for i in ii]
        # Calcul des poids
        wXY = (1. - y[0]) * (1. - y[1])
        wXpY = (y[0]) * (1. - y[1])
        wXYp = (1. - y[0]) * (y[1])
        wXpYp = (y[0]) * (y[1])
        # Calcul du resultat
        res = [0. for i in ii]
        res = [res[i] + wXY * self.gvalues.values[tuple(iXY)][i] for i in ii]
        res = [res[i] + wXpY * self.gvalues.values[tuple(iXpY)][i] for i in ii]
        res = [res[i] + wXYp * self.gvalues.values[tuple(iXYp)][i] for i in ii]
        res = [res[i] + wXpYp * self.gvalues.values[tuple(iXpYp)][i] for i in ii]
        return res

    def __call__(self, ppos, t):
        return self.interpolate(ppos, t)

    def __str__(self):
        """ToString method"""
        return "Linear"


class Linear3D(InterpolationMethod):
    """
    Lambda1 remeshing method implementation 2D.
    """
    def __init__(self, grid, gvalues):
        """
        Constructor.
        """
        self.grid = grid
        self.gvalues = gvalues

    def interpolate(self, t, ppos):
        """
        Linear interpolation.
        Interpolate the velocity at the given particle position.

        @param position : position to compute
        @param grid Grid.Grid : grid data to interpolate
        @return velocity : velocity at particle position
        """
        ii = xrange(len(ppos))
        # Calcul des indices
        ind = [int((ppos[i] - self.grid.min[i]) / self.grid.elementSize[i]) for i in ii]
        iXYZ = [(ind[i] % self.grid.elementNumber[i]) for i in ii]
        iXpYZ = [iXYZ[i] for i in ii]
        iXpYZ[0] = (iXpYZ[0] + 1) % self.grid.elementNumber[0]
        iXYpZ = [iXYZ[i] for i in ii]
        iXYpZ[1] = (iXYpZ[1] + 1) % self.grid.elementNumber[1]
        iXYZp = [iXYZ[i] for i in ii]
        iXYZp[2] = (iXYZp[2] + 1) % self.grid.elementNumber[2]
        iXpYpZ = [iXYZ[i] for i in ii]
        iXpYpZ[0] = (iXpYpZ[0] + 1) % self.grid.elementNumber[0]
        iXpYpZ[1] = (iXpYpZ[1] + 1) % self.grid.elementNumber[1]
        iXpYZp = [iXYZ[i] for i in ii]
        iXpYZp[0] = (iXpYZp[0] + 1) % self.grid.elementNumber[0]
        iXpYZp[2] = (iXpYZp[2] + 1) % self.grid.elementNumber[2]
        iXYpZp = [iXYZ[i] for i in ii]
        iXYpZp[1] = (iXYpZp[1] + 1) % self.grid.elementNumber[1]
        iXYpZp[2] = (iXYpZp[2] + 1) % self.grid.elementNumber[2]
        iXpYpZp = [iXYZ[i] for i in ii]
        iXpYpZp[0] = (iXpYpZp[0] + 1) % self.grid.elementNumber[0]
        iXpYpZp[1] = (iXpYpZp[1] + 1) % self.grid.elementNumber[1]
        iXpYpZp[2] = (iXpYpZp[2] + 1) % self.grid.elementNumber[2]
        # Calcul de la distance
        y = [(ppos[i] - self.grid[iXYZ][i]) / self.grid.elementSize[i] for i in ii]
        # Calcul des poids
        wXYZ = (1. - y[0]) * (1. - y[1]) * (1. - y[2])
        wXpYZ = (y[0]) * (1. - y[1]) * (1. - y[2])
        wXYpZ = (1. - y[0]) * (y[1]) * (1. - y[2])
        wXYZp = (1. - y[0]) * (1. - y[1]) * (y[2])
        wXpYpZ = (y[0]) * (y[1]) * (1. - y[2])
        wXpYZp = (y[0]) * (1. - y[1]) * (y[2])
        wXYpZp = (1. - y[0]) * (y[1]) * (y[2])
        wXpYpZp = (y[0]) * (y[1]) * (y[2])
        # Calcul du resultat
        res = [0. for i in ii]
        res = [res[i] + wXYZ * self.gvalues.values[tuple(iXYZ)][i] for i in ii]
        res = [res[i] + wXpYZ * self.gvalues.values[tuple(iXpYZ)][i] for i in ii]
        res = [res[i] + wXYpZ * self.gvalues.values[tuple(iXYpZ)][i] for i in ii]
        res = [res[i] + wXYZp * self.gvalues.values[tuple(iXYZp)][i] for i in ii]
        res = [res[i] + wXpYpZ * self.gvalues.values[tuple(iXpYpZ)][i] for i in ii]
        res = [res[i] + wXpYZp * self.gvalues.values[tuple(iXpYZp)][i] for i in ii]
        res = [res[i] + wXYpZp * self.gvalues.values[tuple(iXYpZp)][i] for i in ii]
        res = [res[i] + wXpYpZp * self.gvalues.values[tuple(iXpYpZp)][i] for i in ii]
        return res

    def __call__(self, ppos, t):
        return self.interpolate(ppos, t)

    def __str__(self):
        """ToString method"""
        return "Linear"
