# -*- coding: utf-8 -*-
"""
@package Utils
RK2 method interface.
"""
from ..Param import *
from ODESolver import ODESolver


class RK2(ODESolver):
    """
    ODESolver implementation for solving an equation system with RK2 method.
    y'(t) = f(t,y)

    y(t_n+1)= y(t_n) + dt*y'[y(t_n)+dt/2*y'(y(t_n))].

    """
    def __init__(self, f=None, conditions=lambda x: x, dim=3):
        """
        Constructor.

        @param f function f(t,y) : Right hand side of the equation to solve.
        @param dim : dimensions
        @param conditions : function to apply boundary conditions.
        """
        ODESolver.__init__(self, f)
        # Boundary conditions function.
        self.boundaryConditions = conditions
        self.gpu_kernel = """
        """
        self.nb_flop = 5

    def integrate(self, y, fy, yp, t, dt, dir):
        """
        Integration step for RK2 Method.
         yp= y + dt*y'[y+dt/2*y'(y)].

        @param y : position at time t.
        @param fy : y'(y) value at time t.
        @param yp : result position at time t + dt
        @param t : current time.
        @param dt : time step.
        """
        ## ------------------- NumPy Optimisation
        p1 = np.copy(y[..., dir])
        yp[..., dir] = p1 + fy[..., dir] * dt * 0.5
        self.boundaryConditions(yp, dir)  # TODO: A tester aussi avec return dans la fonction
        vp1 = self.f(t, yp, dir)
        yp[..., dir] = p1 + vp1[..., dir] * dt
        self.boundaryConditions(yp, dir)
        ## -------------------

    def __str__(self):
        """ToString method"""
        return "RK2 Method"


if __name__ == "__main__":
    print __doc__
    print "- Provided class : RK2"
    print RK2.__doc__
