# -*- coding: utf-8 -*-
"""
@package Utils
RK3 method interface.
"""
from .integrator import ODESolver
from ...operator.differentialOperator_d import DifferentialOperator_d
from ...operator.differentialOperator import DifferentialOperator
from parmepy.constants import *
import numpy as np
import copy

class RK3Stretch(ODESolver):
    """
    ODESolver implementation for solving an equation system with RK3 method.
    y'(t) = f(t,y)

    y(t_n+1)= y(t_n) + dt*y'[y(t_n)+dt/2*y'(y(t_n))].

    """
    def __init__(self, f=None, dim=3):
        """
        Constructor.

        @param f function f(t,y) : Right hand side of the equation to solve.
        @param dim : dimensions
        @param conditions : function to apply boundary conditions.
        """
        ODESolver.__init__(self, f)


    def integrate(self, f, fd, field1, field2, res_vorticity, t, dt, dimension):
        """
        Integration step for RK3 Method
         yp= y + dt*y'[y+dt/2*y'(y)].

        @param y : position at time t.
        @param f : operator to calcul y'(y) value at time t.
        @param fd : result position at time t + dt
        @param field1 : vorticity field discretized
        @param field2 : velocity field discretized
        @param t : current time.
        @param dt : time step.
        """
        # RK3 TVD
        maxgersh = np.zeros(1, dtype=PARMES_REAL, order=ORDER)
        K1=np.asarray(fd)
        K2=np.asarray([np.zeros((field2.topology.resolution), dtype=PARMES_REAL, order=ORDER) for d in xrange(field2.topology.dim)])
        K1[:,:,:,:] = field1[:,:,:] + dt * K1[:,:,:,:]
        f.discreteOperator = DifferentialOperator_d(K1, field2, K2, choice='divConservation',maxgersh=maxgersh)
        f.discreteOperator.apply()
        K2[:,:,:,:] = 3./4. * field1[:,:,:] + 1./4. * (K1[:,:,:,:] + dt * K2[:,:,:,:])
        f.discreteOperator = DifferentialOperator_d(K2, field2, res_vorticity, choice='divConservation',maxgersh=maxgersh)
        f.discreteOperator.apply()
        res_vorticity[:,:,:] = 1./3. * field1[:,:,:] + 2./3. * (K2[:,:,:,:] + dt * res_vorticity[:,:,:])
        return res_vorticity

        # RK3 V1
#        maxgersh = np.zeros(1, dtype=PARMES_REAL, order=ORDER)
#        K1=np.asarray(fd)
#        K2=np.asarray([np.zeros((field2.topology.resolution), dtype=PARMES_REAL, order=ORDER) for d in xrange(field2.topology.dim)])
#        K3=np.asarray([np.zeros((field2.topology.resolution), dtype=PARMES_REAL, order=ORDER) for d in xrange(field2.topology.dim)])
#        resultTmp = np.asarray([np.zeros((field2.topology.resolution), dtype=PARMES_REAL, order=ORDER) for d in xrange(field2.topology.dim)])
#        K1[:,:,:,:] = dt * K1[:,:,:,:]
#        resultTmp[:,:,:] = field1[:,:,:] + 1./3. * K1[:,:,:,:]
#        f.discreteOperator = DifferentialOperator_d(f, resultTmp, field2, K2, choice='divConservation',maxgersh=maxgersh)
#        f.discreteOperator.apply()
#        K2[:,:,:,:] = field1[:,:,:] + 2./3. * dt * K2[:,:,:,:]
#        f.discreteOperator = DifferentialOperator_d(f, K2, field2, K3, choice='divConservation',maxgersh=maxgersh)
#        f.discreteOperator.apply()
#        K3[:,:,:,:] = dt * K3[:,:,:,:]
#        res_vorticity[:,:,:] = field1[:,:,:] + 1./4. * (K1[:,:,:,:] + 3 * K3[:,:,:,:])
#        return res_vorticity


    def __str__(self):
        """ToString method"""
        return "RK3 Method"


if __name__ == "__main__":
    print __doc__
    print "- Provided class : RK3"
    print RK3.__doc__
