"""
@file operator/discrete/synchronizeGhosts.py

Update ghost points for some fields defined on a specific topology.
"""
from parmepy.constants import debug, ORDER, PARMES_INTEGER, PARMES_REAL
from parmepy.mpi.main_var import main_comm, main_rank
from discrete import DiscreteOperator
from parmepy.tools.timers import timed_function
import numpy as np


class SynchronizeGhosts_d(DiscreteOperator):
    """
    Ghost points synchronization.
    """

    @debug
    def __init__(self, fieldslist, topology, transferMethod='greaterSend'):
        """
        Defines a way to send/recv values at ghosts points for a list
        of fields, discretized on a given topology.

        @param  fieldslist : a list of fields
        @param topology : the topology common to all fields.
        @param transferMethod : which type of exchange is used.
        """
        DiscreteOperator.__init__(self, fieldslist, method=transferMethod,
                                  name="Ghosts Synchronization")
        self.input = fieldslist
        self.output = fieldslist
        self.compute_time = 0.
        self.fieldslist = fieldslist
        self.topology = topology
        if (main_rank == 0):
            print 'CUT SPACE', self.topology.dims
        self.transferMethod = transferMethod

    @debug
    def setUp(self):

        self.topoMPI = \
            self.topology.comm.Create_cart(
                self.topology.dims,
                periods=self.topology.periods)
        ghosts = self.topology.ghosts
        resolution = self.topology.mesh.resolution
        self.fieldslist = np.asarray([self.fieldslist])
        if (self.transferMethod == 'SendRecv'):
            # Allocation of space for reciev data
            self.dataRecvX = np.zeros((
                ghosts[0], resolution[1], resolution[2]),
                dtype=PARMES_REAL, order=ORDER)
            self.dataRecvY = np.zeros((
                resolution[0], ghosts[1], resolution[2]),
                dtype=PARMES_REAL, order=ORDER)
            self.dataRecvZ = np.zeros((
                resolution[0], resolution[1], ghosts[2]),
                dtype=PARMES_REAL, order=ORDER)
            self.SendRecvList = np.zeros(
                self.topology.dim * 2,
                order=ORDER, dtype=PARMES_INTEGER)
            self.SendRecvList[0], self.SendRecvList[1] = \
                self.topology.Shift(0, 1)
            self.SendRecvList[2], self.SendRecvList[3] = \
                self.topology.Shift(1, 1)
            if (self.topology.dim > 2):
                self.SendRecvList[4], self.SendRecvList[5] = \
                    self.topology.Shift(2, 1)

        if (self.transferMethod == 'greaterSend'):
            tab = self.topology.neighbours
            self.tabSort = np.zeros(
                [self.topology.dim * 2, 2],
                order=ORDER, dtype=PARMES_INTEGER)
            self.tabSort[:, 1] = range(self.topology.dim * 2)
            self.tabSort[0, 0] = tab[0, 0]
            self.tabSort[1, 0] = tab[1, 0]
            self.tabSort[2, 0] = tab[0, 1]
            self.tabSort[3, 0] = tab[1, 1]
            if(self.topology.dim > 2):
                self.tabSort[4, 0] = tab[0, 2]
                self.tabSort[5, 0] = tab[1, 2]
            tabSort1 = self.tabSort[self.tabSort[:, 0] <= main_rank]
            tabSort2 = self.tabSort[self.tabSort[:, 0] > main_rank]
            if (np.asarray(tabSort1[:, 0].shape) > 0):
                tabSort1 = tabSort1[tabSort1[:, 0].argsort()]
            if (np.asarray(tabSort2[:, 0].shape) > 0):
                tabSort2 = tabSort2[tabSort2[:, 0].argsort()]
                tabSort2 = self.TabInvSort(tabSort2)
            self.tabSort = np.concatenate((tabSort1, tabSort2))

        if ((self.transferMethod == 'greaterSend') or
                (self.transferMethod == 'SendRecv')):
            # Allocation of space for send and reciev data
            self.dataSendX = np.zeros((
                ghosts[0], resolution[1], resolution[2]),
                dtype=PARMES_REAL, order=ORDER)
            self.dataSendY = np.zeros((
                resolution[0], ghosts[1], resolution[2]),
                dtype=PARMES_REAL, order=ORDER)
            self.dataSendZ = np.zeros((
                resolution[0], resolution[1], ghosts[2]),
                dtype=PARMES_REAL, order=ORDER)
        # 1 - get discrete fields arrays
        # 2 - get list of ghosts area
        # 3 - get neighbours
        # 4 - find what is to be send/recv, the size of the arrays and
        # allocate buffers if required
        # 5 - set send/recv order
        #self.discreteOperator.setUp()

    @debug
    @timed_function
    def apply(self, simulation=None):
        self.resolution = self.topology.mesh.resolution
        resolution = self.topology.mesh.resolution
        ghosts = self.topology.ghosts
        dim = self.topology.dim
        # Do the send/recv as defined in setup.
        #### if args is None: \todo : A verifier.
#        args = self.fieldslist
#        self.fieldslist = np.asarray([self.fieldslist])
        if (self.transferMethod == 'SendRecv'):
            for f in fieldslist:
                # DOWN
                if (main_rank != self.SendRecvList[0]):
                    self.dataSendZ = f[j][:, :, ghosts[2]:2. * ghosts[2]]
                    main_comm.Sendrecv(
                        self.dataSendZ,
                        dest=self.SendRecvList[0],
                        sendtag=22,
                        recvbuf=self.dataRecvZ,
                        source=main_rank,
                        recvtag=11, status=None)
                    f[j][:, :, resolution[2] - ghosts[2]: resolution[2]] = \
                        self.dataRecvZ
                else:
                    f[j][:, :, resolution[2] - ghosts[2]: resolution[2]
                         ] = f[j][:, :, ghosts[2]:2 * ghosts[2]]
                # UP
                if (main_rank != self.SendRecvList[1]):
                    self.dataSendZ = np.array(np.copy(f[j][
                        :, :,
                        resolution[2] - ghosts[2]:resolution[2]],
                        order=ORDER, dtype=PARMES_REAL))
                    main_comm.Sendrecv(
                        self.dataSendZ,
                        dest=self.SendRecvList[1],
                        sendtag=11,
                        recvbuf=self.dataRecvZ,
                        source=main_rank,
                        recvtag=22, status=None)
                else:
                    f[j][:, :, ghosts[2]:2 * ghosts[2]
                         ] = \
                        f[j][:, :, resolution[2] - ghosts[2]:resolution[2]]
                # SOUTH
                if (main_rank != self.SendRecvList[2]):
                    self.dataSendY = f[j][:, ghosts[1]:2. * ghosts[1], :]
                    main_comm.Sendrecv(
                        self.dataSendY,
                        dest=self.SendRecvList[2],
                        sendtag=44,
                        recvbuf=self.dataRecvY,
                        source=main_rank,
                        recvtag=33, status=None)
                    f[j][:, resolution[2] - ghosts[2]:resolution[2], :] \
                        = self.dataRecvY
                else:
                    f[j][:, resolution[1] - ghosts[1]:resolution[1], :
                         ] = f[j][:, ghosts[1]:2 * ghosts[1], :]
                # NORTH
                if (main_rank != self.SendRecvList[3]):
                    self.dataSendY = np.array(np.copy(f[j][
                        :,
                        resolution[1] - ghosts[1]:resolution[1],
                        :],
                        order=ORDER, dtype=PARMES_REAL))
                    main_comm.Sendrecv(
                        self.dataSendY,
                        dest=self.SendRecvList[3],
                        sendtag=33,
                        recvbuf=self.dataRecvY,
                        source=main_rank,
                        recvtag=44, status=None)
                else:
                    f[j][:, ghosts[1]:2 * ghosts[1], :] =\
                        f[j][:, resolution[1] - ghosts[1]:resolution[1], :]
                # EAST
                if (main_rank != self.SendRecvList[4]):
                    self.dataSendX = f[j][ghosts[0]:2. * ghosts[0], :, :]
                    main_comm.Sendrecv(
                        self.dataSendX,
                        dest=self.SendRecvList[4],
                        sendtag=66,
                        recvbuf=self.dataRecvX,
                        source=main_rank,
                        recvtag=55, status=None)
                    f[j][resolution[0] - ghosts[0]:resolution[0], :, :] \
                        = self.dataRecvX
                else:
                    f[j][resolution[0] - ghosts[0]:resolution[0], :, :
                         ] = f[j][:, :, ghosts[2]:2 * ghosts[2]]
                # WEST
                if (main_rank != self.SendRecvList[5]):
                    self.dataSendX = np.array(np.copy(f[j][
                        resolution[0] - ghosts[0]:resolution[0],
                        :, :],
                        order=ORDER, dtype=PARMES_REAL))
                    main_comm.Sendrecv(
                        self.dataSendX,
                        dest=self.SendRecvList[5],
                        sendtag=55,
                        recvbuf=self.dataRecvX,
                        source=main_rank,
                        recvtag=66, status=None)
                else:
                    f[j][ghosts[0]:2 * ghosts[0], :, :] =\
                        f[j][resolution[0] - ghosts[0]:resolution[0], :, :]

        if (self.transferMethod == 'greaterSend'):
            for f in self.fieldslist:
#                print 'tabsort', main_rank, self.tabSort
                for i in xrange(self.tabSort[:, 0].shape[0]):
#                    print 'Cours', main_rank, self.tabSort[i, :]
                    if (main_rank == self.tabSort[i, 0]):
                        ## Without communication
                        # WEST
                        if (self.tabSort[i, 1] == 0):
                            for j in xrange(dim):
                                #print '\nTEST', resolution,  f[j][...].shape, '\t' , ghosts
                                f[j][resolution[0] - ghosts[0]: resolution[0],
                                     :, :] =\
                                    f[j][ghosts[0]:2 * ghosts[0], :, :]
                        # EAST
                        if (self.tabSort[i, 1] == 1):
                            for j in xrange(dim):
                                f[j][0:ghosts[0], :, :] =\
                                    f[j][
                                        resolution[0] - 2 *
                                        ghosts[0]:resolution[0] - ghosts[0],
                                        :, :]

                        # SOUTH
                        if (self.tabSort[i, 1] == 2):
                            for j in xrange(dim):
                                f[j][:, resolution[1] -
                                     ghosts[1]:resolution[1], :] =\
                                    f[j][:, ghosts[1]:2 * ghosts[1], :]
                        #NORTH
                        if(self.tabSort[i, 1] == 3):
                            for j in xrange(dim):
                                f[j][:, 0:ghosts[1], :] =\
                                    f[j][
                                        :,
                                        resolution[1] - 2 *
                                        ghosts[1]:resolution[1] - ghosts[1],
                                        :]

                        # DOWN
                        if(self.tabSort[i, 1] == 4):
                            for j in xrange(dim):
                                f[j][:, :, resolution[2] -
                                     ghosts[2]:resolution[2]] =\
                                    f[j][:, :, ghosts[2]:2 * ghosts[2]]
                        # UP
                        if(self.tabSort[i, 1] == 5):
                            for j in xrange(dim):
                                f[j][:, :, 0:ghosts[2]] =\
                                    f[j][
                                        :, :,
                                        resolution[2] - 2 *
                                        ghosts[2]:resolution[2] - ghosts[2]]
#                        print 'ok', main_rank, self.tabSort[i, :]
                    else:
                       ### PART COMMUNICATION
                        # WEST
                        if(self.tabSort[i, 1] == 0):
                            if (main_rank < self.tabSort[i, 0]):
                                self.WESTsend(self.tabSort[i, 0], f)
                            else:
                                self.WESTrecv(self.tabSort[i, 0], f)
                        # EST
                        if (self.tabSort[i, 1] == 1):
                            if (main_rank < self.tabSort[i, 0]):
                                self.EASTsend(self.tabSort[i, 0], f)
                            else:
                                self.EASTrecv(self.tabSort[i, 0], f)
                        # SOUTH
                        if(self.tabSort[i, 1] == 2):
                            if (main_rank < self.tabSort[i, 0]):
                                self.SOUTHsend(self.tabSort[i, 0], f)
                            else:
                                self.SOUTHrecv(self.tabSort[i, 0], f)
                        # NORTH
                        if (self.tabSort[i, 1] == 3):
                            if (main_rank < self.tabSort[i, 0]):
                                self.NORTHsend(self.tabSort[i, 0], f)
                            else:
                                self.NORTHrecv(self.tabSort[i, 0], f)
                        # DOWN
                        if (self.tabSort[i, 1] == 4):
                            if (main_rank < self.tabSort[i, 0]):
                                self.DOWNsend(self.tabSort[i, 0], f)
                            else:
                                self.DOWNrecv(self.tabSort[i, 0], f)
                        # UP
                        if (self.tabSort[i, 1] == 5):
                            if (main_rank < self.tabSort[i, 0]):
                                self.UPsend(self.tabSort[i, 0], f)
                            else:
                                self.UPrecv(self.tabSort[i, 0], f)
#                        print 'ok', main_rank, self.tabSort[i, :]
#        return self.discreteOperator.apply(*args)

    def TabInvSort(self, tab):
        tmp = tab[0, 0]
        deb = 0
        tabcont = 0
        tabfinal = np.copy(tab)
        if tab[:, 0].shape[0] == 1:
            return tab
        for i in xrange(1, tab[:, 0].shape[0]):
            if tmp == tab[i, 0]:
                tabcont = tabcont + 1
            else:
                tmp = tab[i, 0]
                if tabcont == 0:
                    tabfinal[deb, :] = tab[deb, :]
                else:
                    tabNew = tab[deb:i, :]
#                    for l in xrange(int(tabNew[:,1].shape[0] / 2)):
#                        tmp = tabNew[2 * l, 1]
#                        tabNew[2 * l, 1] = tabNew[2 * l + 1,1]
#                        tabNew[2 * l + 1, 1] = tmp
                    tabNew = tabNew[np.invert(tabNew[:, 1].argsort())]
                    tabfinal[deb:i, :] = tabNew
                deb = i
                tabcont = 0
                if deb == tab[:, 0].shape[0]:
                    tabfinal[deb, :] = tab[deb, :]
        if tabcont > 0:
            tabNew = tab[deb:tab[:, 0].shape[0], :]
            tabNew = tabNew[np.invert(tabNew[:, 1].argsort())]
            tabfinal[deb:tab[:, 0].shape[0], :] = tabNew
        return tabfinal

    def WESTsend(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            self.dataSendX = np.array(np.copy(listFields[i][
                ghosts[0]:ghosts[0] * 2, :, :]), order=ORDER)
            main_comm.Ssend(self.dataSendX, dest=proc, tag=66)
            main_comm.Recv(self.dataSendX, source=proc, tag=55)
            listFields[i][0:ghosts[0], :, :] = self.dataSendX

    def EASTsend(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            self.dataSendX = np.array(np.copy(listFields[i][
                resolution[0] - 2 * ghosts[0]:resolution[0] - ghosts[0],
                :, :]),
                order=ORDER)
            main_comm.Ssend(self.dataSendX, dest=proc, tag=66)
            main_comm.Recv(self.dataSendX, source=proc, tag=55)
            listFields[i][resolution[0] - ghosts[0]:resolution[0], :, :] =\
                self.dataSendX

    def SOUTHsend(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            self.dataSendY = np.array(np.copy(listFields[i][
                :, ghosts[1]:2 * ghosts[1], :]), order=ORDER)
            main_comm.Ssend(self.dataSendY, dest=proc, tag=44)
            main_comm.Recv(self.dataSendY, source=proc, tag=33)
            listFields[i][:, 0:ghosts[1], :] = self.dataSendY

    def NORTHsend(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            self.dataSendY = np.array(np.copy(listFields[i][
                :,
                resolution[1] - 2 * ghosts[1]:resolution[1] - ghosts[1], :]),
                order=ORDER)
            main_comm.Ssend(self.dataSendY, dest=proc, tag=33)
            main_comm.Recv(self.dataSendY, source=proc, tag=44)
            listFields[i][:, resolution[1] - ghosts[1]:resolution[1], :] =\
                self.dataSendY

    def DOWNsend(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            self.dataSendZ = np.array(np.copy(listFields[i][
                :, :, ghosts[2]:2 * ghosts[2]]), order=ORDER)
            main_comm.Ssend(self.dataSendZ, dest=proc, tag=11)
            main_comm.Recv(self.dataSendZ, source=proc, tag=22)
            listFields[i][:, :, 0:ghosts[2]] = self.dataSendZ

    def UPsend(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            self.dataSendZ = np.array(np.copy(listFields[i][
                :, :,
                resolution[2] - 2 * ghosts[2]:resolution[2] - ghosts[2]
            ]), order=ORDER)
            main_comm.Ssend(self.dataSendZ, dest=proc, tag=22)
            main_comm.Recv(self.dataSendZ, source=proc, tag=11)
            listFields[i][
                :, :,
                resolution[2] - ghosts[2]:resolution[2]] =\
                self.dataSendZ

    def WESTrecv(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            main_comm.Recv(self.dataSendX, source=proc, tag=66)
            listFields[i][0:ghosts[0], :, :] = self.dataSendX
            self.dataSendX = np.array(np.copy(listFields[i][
                ghosts[0]:2 * ghosts[0], :, :]), order=ORDER)
            main_comm.Ssend(self.dataSendX, dest=proc, tag=55)

    def EASTrecv(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            main_comm.Recv(self.dataSendX, source=proc, tag=66)
            listFields[i][resolution[0] - ghosts[0]:resolution[0], :, :] =\
                self.dataSendX
            self.dataSendX = np.array(np.copy(listFields[i][
                resolution[0] - 2 * ghosts[0]:resolution[0] - ghosts[0],
                :, :]),
                order=ORDER)
            main_comm.Ssend(self.dataSendX, dest=proc, tag=55)

    def SOUTHrecv(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            main_comm.Recv(self.dataSendY, source=proc, tag=33)
            listFields[i][:, 0:ghosts[1], :] = self.dataSendY
            self.dataSendY = np.array(np.copy(listFields[i][
                :, ghosts[1]:2 * ghosts[1], :]), order=ORDER)
            main_comm.Ssend(self.dataSendY, dest=proc, tag=44)

    def NORTHrecv(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            main_comm.Recv(self.dataSendY, source=proc, tag=44)
            listFields[i][:, resolution[1] - ghosts[1]:resolution[1], :] =\
                self.dataSendY
            self.dataSendY = np.array(np.copy(listFields[i][
                :, resolution[1] - 2 * ghosts[1]:resolution[1] -
                ghosts[1], :]),
                order=ORDER)
            main_comm.Ssend(self.dataSendY, dest=proc, tag=33)

    def DOWNrecv(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            main_comm.Recv(self.dataSendZ, source=proc, tag=22)
            listFields[i][:, :, 0:ghosts[2]] = self.dataSendZ
            data = np.array(np.copy(listFields[i][
                :, :, ghosts[2]:ghosts[2] * 2]), order=ORDER)
            main_comm.Ssend(self.dataSendZ, dest=proc, tag=11)

    def UPrecv(self, proc, listFields):
        ghosts = self.topology.ghosts
        resolution = self.resolution
        for i in xrange(self.topology.dim):
            main_comm.Recv(self.dataSendZ, source=proc, tag=11)
            listFields[i][:, :, resolution[2] - ghosts[2]:resolution[2]] =\
                self.dataSendZ
            self.dataSendZ = np.array(np.copy(listFields[i][
                :, :,
                resolution[2] - 2 * ghosts[2]:resolution[2] - ghosts[2]
            ]), order=ORDER)
            main_comm.Ssend(self.dataSendZ, dest=proc, tag=22)


if (__name__ == "__main__"):
    print __doc__
    print "- Provided class : SynchronizeGhosts"
    print SynchronizeGhosts_d.__doc__
